@@ -1,5 +1,5 @@
 %---------------------------------------------------------------------------%
-% Copyright (C) 1995 University of Melbourne.
+% Copyright (C) 1996-1997 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %---------------------------------------------------------------------------%
@@ -15,7 +15,7 @@
 :- interface.
 
 :- import_module hlds_module, bytecode.
-:- import_module io.
+:- import_module io, list.
 
 :- pred bytecode_gen__module(module_info::in, list(byte_code)::out,
 	io__state::di, io__state::uo) is det.
@@ -24,11 +24,12 @@
 
 :- implementation.
 
-:- import_module hlds_pred, hlds_goal, hlds_data, prog_data, llds.
-:- import_module passes_aux, call_gen, mode_util, code_util, goal_util, tree.
+:- import_module hlds_pred, hlds_goal, hlds_data, prog_data, llds, arg_info.
+:- import_module passes_aux, call_gen, mode_util, code_util, goal_util.
+:- import_module globals, tree.
 
 :- import_module bool, int, string, list, assoc_list, set, map, varset.
-:- import_module std_util, require.
+:- import_module std_util, require, term.
 
 bytecode_gen__module(ModuleInfo, Code) -->
 	{ module_info_predids(ModuleInfo, PredIds) },
@@ -51,11 +52,19 @@ bytecode_gen__preds([PredId | PredIds], ModuleInfo, Code) -->
 			ProcsCode),
 		{ predicate_name(ModuleInfo, PredId, PredName) },
 		{ list__length(ProcIds, ProcsCount) },
-		{ PredCode = tree(
-			tree(
-				node([enter_pred(PredName, ProcsCount)]),
-				ProcsCode),
-			node([endof_pred])) }
+		{ pred_info_arity(PredInfo, Arity) },
+		{ pred_info_get_is_pred_or_func(PredInfo, PredOrFunc) },
+		{ 
+			(PredOrFunc = predicate ->
+				IsFunc = 0
+			;
+				IsFunc = 1
+			)
+		},
+		{ EnterCode = node([enter_pred(PredName, Arity, IsFunc,
+			ProcsCount)]) },
+		{ EndofCode = node([endof_pred]) },
+		{ PredCode = tree(EnterCode, tree(ProcsCode, EndofCode)) }
 	),
 	bytecode_gen__preds(PredIds, ModuleInfo, OtherCode),
 	{ Code = tree(PredCode, OtherCode) }.
@@ -82,16 +91,19 @@ bytecode_gen__proc(ProcId, PredInfo, ModuleInfo, Code) :-
 	proc_info_vartypes(ProcInfo, VarTypes),
 	proc_info_variables(ProcInfo, VarSet),
 	proc_info_interface_determinism(ProcInfo, Detism),
+	determinism_to_code_model(Detism, CodeModel),
 
-	goal_util__goal_vars(Goal, Vars),
+	goal_util__goal_vars(Goal, GoalVars),
+	proc_info_headvars(ProcInfo, ArgVars),
+	set__insert_list(GoalVars, ArgVars, Vars),
 	set__to_sorted_list(Vars, VarList),
 	map__init(VarMap0),
 	bytecode_gen__create_varmap(VarList, VarSet, VarTypes, 0,
 		VarMap0, VarMap, VarInfos),
 
-	bytecode_gen__init_byte_info(ModuleInfo, VarMap, VarTypes, ByteInfo),
+	bytecode_gen__init_byte_info(ModuleInfo, VarMap, VarTypes, ByteInfo0),
+	bytecode_gen__get_next_label(ByteInfo0, ZeroLabel, ByteInfo1),
 
-	proc_info_headvars(ProcInfo, ArgVars),
 	proc_info_arg_info(ProcInfo, ArgInfo),
 	assoc_list__from_corresponding_lists(ArgVars, ArgInfo, Args),
 
@@ -101,15 +113,15 @@ bytecode_gen__proc(ProcId, PredInfo, ModuleInfo, Code) :-
 	call_gen__output_arg_locs(Args, OutputArgs),
 	bytecode_gen__gen_places(OutputArgs, ByteInfo, PlaceCode),
 
-	bytecode_gen__goal(Goal, ByteInfo, 1, N, GoalCode),
+	bytecode_gen__goal(Goal, ByteInfo1, ByteInfo, GoalCode),
+	bytecode_gen__get_counts(ByteInfo, LabelCount, TempCount),
 
-	BodyTree = tree(
-			tree(
-				PickupCode,
-				node([label(0)])),
-			tree(
-				GoalCode,
-				PlaceCode)),
+	ZeroLabelCode = node([label(ZeroLabel)]),
+	BodyTree =
+		tree(PickupCode,
+		tree(ZeroLabelCode,
+		tree(GoalCode,
+		     PlaceCode))),
 	tree__flatten(BodyTree, BodyList),
 	list__condense(BodyList, BodyCode0),
 	( list__member(not_supported, BodyCode0) ->
@@ -117,107 +129,116 @@ bytecode_gen__proc(ProcId, PredInfo, ModuleInfo, Code) :-
 	;
 		BodyCode = node(BodyCode0)
 	),
-	Code = tree(
-			node([enter_proc(ProcId, Detism, N, VarInfos)]),
-			tree(
-				BodyCode,
-				node([endof_proc]))).
+	proc_id_to_int(ProcId, ProcInt),
+	EnterCode = node([enter_proc(ProcInt, Detism, LabelCount, TempCount,
+		VarInfos)]),
+	( CodeModel = model_semi ->
+		EndofCode = node([semidet_succeed, endof_proc])
+	;
+		EndofCode = node([endof_proc])
+	),
+	Code = tree(EnterCode, tree(BodyCode, EndofCode)).
 
 %---------------------------------------------------------------------------%
 
-:- pred bytecode_gen__goal(hlds__goal::in, byte_info::in, int::in, int::out,
+:- pred bytecode_gen__goal(hlds_goal::in, byte_info::in, byte_info::out,
 	byte_tree::out) is det.
 
-bytecode_gen__goal(GoalExpr - GoalInfo, ByteInfo, N0, N, Code) :-
-	bytecode_gen__goal_expr(GoalExpr, ByteInfo, N0, N, GoalCode),
+bytecode_gen__goal(GoalExpr - GoalInfo, ByteInfo0, ByteInfo, Code) :-
+	bytecode_gen__goal_expr(GoalExpr, GoalInfo, ByteInfo0, ByteInfo,
+		GoalCode),
 	goal_info_get_context(GoalInfo, term__context(_, Line)),
 	Code = tree(node([context(Line)]), GoalCode).
 
-:- pred bytecode_gen__goal_expr(hlds__goal_expr::in, byte_info::in,
-	int::in, int::out, byte_tree::out) is det.
+:- pred bytecode_gen__goal_expr(hlds_goal_expr::in, hlds_goal_info::in,
+	byte_info::in, byte_info::out, byte_tree::out) is det.
 
-bytecode_gen__goal_expr(GoalExpr, ByteInfo, N0, N, Code) :-
+bytecode_gen__goal_expr(GoalExpr, GoalInfo, ByteInfo0, ByteInfo, Code) :-
 	(
-		GoalExpr = higher_order_call(_, _, _, _, _),
-		error("we do not handle higher order calls yet")
+		GoalExpr = higher_order_call(PredVar, ArgVars, ArgTypes,
+			ArgModes, Detism, _IsPredOrFunc),
+		bytecode_gen__higher_order_call(PredVar, ArgVars,
+			ArgTypes, ArgModes, Detism, ByteInfo0, Code),
+		ByteInfo = ByteInfo0
 	;
-		GoalExpr = call(PredId, ProcId, Args, IsBuiltin, _, _),
-		( hlds__is_builtin_is_internal(IsBuiltin) ->
-			bytecode_gen__builtin(PredId, ProcId, Args, ByteInfo,
-				Code)
+		GoalExpr = call(PredId, ProcId, ArgVars, BuiltinState, _, _),
+		( BuiltinState = not_builtin ->
+			goal_info_get_determinism(GoalInfo, Detism),
+			bytecode_gen__call(PredId, ProcId, ArgVars, Detism,
+				ByteInfo0, Code)
 		;
-			bytecode_gen__call(PredId, ProcId, Args, ByteInfo,
-				Code)
+			bytecode_gen__builtin(PredId, ProcId, ArgVars,
+				ByteInfo0, Code)
 		),
-		N = N0
+		ByteInfo = ByteInfo0
 	;
 		GoalExpr = unify(Var, RHS, _Mode, Unification, _),
-		bytecode_gen__unify(Unification, Var, RHS, ByteInfo, Code),
-		N = N0
+		bytecode_gen__unify(Unification, Var, RHS, ByteInfo0, Code),
+		ByteInfo = ByteInfo0
 	;
 		GoalExpr = not(Goal),
-		bytecode_gen__goal(Goal, ByteInfo, N0, N1, SomeCode),
-		N is N1 + 1,
-		Code = tree(
-			tree(
-				node([enter_negation(N1)]),
-				SomeCode),
-			node([endof_negation]))
+		bytecode_gen__goal(Goal, ByteInfo0, ByteInfo1, SomeCode),
+		bytecode_gen__get_next_label(ByteInfo1, EndLabel, ByteInfo),
+		EnterCode = node([enter_negation(EndLabel)]),
+		EndofCode = node([endof_negation, label(EndLabel)]),
+		Code = tree(EnterCode, tree(SomeCode, EndofCode))
 	;
 		GoalExpr = some(_, Goal),
-		bytecode_gen__goal(Goal, ByteInfo, N0, N, SomeCode),
-		Code = tree(
-			tree(
-				node([enter_commit]),
-				SomeCode),
-			node([endof_commit]))
+		bytecode_gen__goal(Goal, ByteInfo0, ByteInfo1, SomeCode),
+		bytecode_gen__get_next_temp(ByteInfo1, Temp, ByteInfo),
+		EnterCode = node([enter_commit(Temp)]),
+		EndofCode = node([endof_commit(Temp)]),
+		Code = tree(EnterCode, tree(SomeCode, EndofCode))
 	;
 		GoalExpr = conj(GoalList),
-		bytecode_gen__conj(GoalList, ByteInfo, N0, N, Code)
+		bytecode_gen__conj(GoalList, ByteInfo0, ByteInfo, Code)
 	;
 		GoalExpr = disj(GoalList, _),
-		bytecode_gen__disj(GoalList, ByteInfo, N0, N1, DisjCode),
-		N is N1 + 1,
-		Code = tree(
-			tree(
-				node([enter_disjunction(N1)]),
-				DisjCode),
-			node([endof_disjunction, label(N1)]))
+		( GoalList = [] ->
+			Code = node([fail]),
+			ByteInfo = ByteInfo0
+		;
+			bytecode_gen__get_next_label(ByteInfo0, EndLabel,
+				ByteInfo1),
+			bytecode_gen__disj(GoalList, ByteInfo1, EndLabel,
+				ByteInfo, DisjCode),
+			EnterCode = node([enter_disjunction(EndLabel)]),
+			EndofCode = node([endof_disjunction, label(EndLabel)]),
+			Code = tree(EnterCode, tree(DisjCode, EndofCode))
+		)
 	;
 		GoalExpr = switch(Var, _, CasesList, _),
-		bytecode_gen__switch(CasesList, Var, ByteInfo, N0, N1,
-			SwitchCode),
+		bytecode_gen__get_next_label(ByteInfo0, EndLabel, ByteInfo1),
+		bytecode_gen__switch(CasesList, Var, ByteInfo1, EndLabel,
+			ByteInfo, SwitchCode),
 		bytecode_gen__map_var(ByteInfo, Var, ByteVar),
-		N is N1 + 1,
-		Code = tree(
-			tree(
-				node([enter_switch(ByteVar, N1)]),
-				SwitchCode),
-			node([endof_switch, label(N1)]))
+		EnterCode = node([enter_switch(ByteVar, EndLabel)]),
+		EndofCode = node([endof_switch, label(EndLabel)]),
+		Code = tree(EnterCode, tree(SwitchCode, EndofCode))
 	;
 		GoalExpr = if_then_else(_Vars, Cond, Then, Else, _),
-		bytecode_gen__goal(Cond, ByteInfo, N0, N1, CondCode),
-		bytecode_gen__goal(Then, ByteInfo, N1, N2, ThenCode),
-		N3 is N2 + 1,
-		bytecode_gen__goal(Else, ByteInfo, N3, N4, ElseCode),
-		N is N4 + 1,
-		Code = tree(
-			tree(
-				tree(
-					node([enter_if(N2, N4)]),
-					CondCode),
-				tree(
-					node([enter_then]),
-					ThenCode)),
-			tree(
-				node([endof_then, label(N2)]),
-				tree(
-					ElseCode,
-					node([endof_else, label(N4)]))))
+		bytecode_gen__get_next_label(ByteInfo0, EndLabel, ByteInfo1),
+		bytecode_gen__get_next_label(ByteInfo1, ElseLabel, ByteInfo2),
+		bytecode_gen__get_next_temp(ByteInfo2, FrameTemp, ByteInfo3),
+		bytecode_gen__goal(Cond, ByteInfo3, ByteInfo4, CondCode),
+		bytecode_gen__goal(Then, ByteInfo4, ByteInfo5, ThenCode),
+		bytecode_gen__goal(Else, ByteInfo5, ByteInfo, ElseCode),
+		EnterIfCode = node([enter_if(ElseLabel, EndLabel, FrameTemp)]),
+		EnterThenCode = node([enter_then(FrameTemp)]),
+		EndofThenCode = node([endof_then(EndLabel), label(ElseLabel)]),
+		EndofIfCode = node([endof_if, label(EndLabel)]),
+		Code =
+			tree(EnterIfCode,
+			tree(CondCode,
+			tree(EnterThenCode,
+			tree(ThenCode,
+			tree(EndofThenCode,
+			tree(ElseCode,
+			     EndofIfCode))))))
 	;
-		GoalExpr = pragma_c_code(_, _, _, _, _, _),
+		GoalExpr = pragma_c_code(_, _, _, _, _, _, _, _),
 		Code = node([not_supported]),
-		N = N0
+		ByteInfo = ByteInfo0
 	).
 
 %---------------------------------------------------------------------------%
@@ -229,7 +250,7 @@ bytecode_gen__gen_places([], _, empty).
 bytecode_gen__gen_places([Var - Loc | OutputArgs], ByteInfo, Code) :-
 	bytecode_gen__gen_places(OutputArgs, ByteInfo, OtherCode),
 	bytecode_gen__map_var(ByteInfo, Var, ByteVar),
-	Code = tree(node([place_arg(r(Loc), ByteVar)]), OtherCode).
+	Code = tree(node([place_arg(r, Loc, ByteVar)]), OtherCode).
 
 :- pred bytecode_gen__gen_pickups(list(pair(var, arg_loc))::in, byte_info::in,
 	byte_tree::out) is det.
@@ -238,30 +259,72 @@ bytecode_gen__gen_pickups([], _, empty).
 bytecode_gen__gen_pickups([Var - Loc | OutputArgs], ByteInfo, Code) :-
 	bytecode_gen__gen_pickups(OutputArgs, ByteInfo, OtherCode),
 	bytecode_gen__map_var(ByteInfo, Var, ByteVar),
-	Code = tree(node([pickup_arg(r(Loc), ByteVar)]), OtherCode).
+	Code = tree(node([pickup_arg(r, Loc, ByteVar)]), OtherCode).
 
 %---------------------------------------------------------------------------%
 
+	% Generate bytecode for a higher order call.
+
+:- pred bytecode_gen__higher_order_call(var::in, list(var)::in, list(type)::in,
+	list(mode)::in, determinism::in, byte_info::in, byte_tree::out) is det.
+
+bytecode_gen__higher_order_call(PredVar, ArgVars, ArgTypes, ArgModes, Detism,
+		ByteInfo, Code) :-
+	determinism_to_code_model(Detism, CodeModel),
+	bytecode_gen__get_module_info(ByteInfo, ModuleInfo),
+	module_info_globals(ModuleInfo, Globals),
+	globals__get_args_method(Globals, ArgsMethod),
+	make_arg_infos(ArgsMethod, ArgTypes, ArgModes, CodeModel, ModuleInfo,
+		ArgInfo),
+	assoc_list__from_corresponding_lists(ArgVars, ArgInfo, ArgVarsInfos),
+
+	call_gen__partition_args(ArgVarsInfos, InVars, OutVars),
+	list__length(InVars, NInVars),
+	list__length(OutVars, NOutVars),
+
+	call_gen__input_arg_locs(ArgVarsInfos, InputArgs),
+	bytecode_gen__gen_places(InputArgs, ByteInfo, PlaceArgs),
+
+	call_gen__output_arg_locs(ArgVarsInfos, OutputArgs),
+	bytecode_gen__gen_pickups(OutputArgs, ByteInfo, PickupArgs),
+
+	bytecode_gen__map_var(ByteInfo, PredVar, BytePredVar),
+	Call = node([higher_order_call(BytePredVar, NInVars, NOutVars,
+		Detism)]),
+	( CodeModel = model_semi ->
+		Check = node([semidet_success_check])
+	;
+		Check = empty
+	),
+	Code = tree(PlaceArgs, tree(Call, tree(Check, PickupArgs))).
+
 	% Generate bytecode for an ordinary call.
 
 :- pred bytecode_gen__call(pred_id::in, proc_id::in, list(var)::in,
-	byte_info::in, byte_tree::out) is det.
+	determinism::in, byte_info::in, byte_tree::out) is det.
 
-bytecode_gen__call(PredId, ProcId, ArgVars, ByteInfo, Code) :-
+bytecode_gen__call(PredId, ProcId, ArgVars, Detism, ByteInfo, Code) :-
 	bytecode_gen__get_module_info(ByteInfo, ModuleInfo),
 	module_info_pred_proc_info(ModuleInfo, PredId, ProcId, _, ProcInfo),
 	proc_info_arg_info(ProcInfo, ArgInfo),
-	assoc_list__from_corresponding_lists(ArgVars, ArgInfo, Args),
+	assoc_list__from_corresponding_lists(ArgVars, ArgInfo, ArgVarsInfos),
 
-	call_gen__input_arg_locs(Args, InputArgs),
+	call_gen__input_arg_locs(ArgVarsInfos, InputArgs),
 	bytecode_gen__gen_places(InputArgs, ByteInfo, PlaceArgs),
 
-	call_gen__output_arg_locs(Args, OutputArgs),
+	call_gen__output_arg_locs(ArgVarsInfos, OutputArgs),
 	bytecode_gen__gen_pickups(OutputArgs, ByteInfo, PickupArgs),
 
 	predicate_id(ModuleInfo, PredId, ModuleName, PredName, Arity),
-	Call = node([call(ModuleName, PredName, Arity, ProcId)]),
-	Code = tree(PlaceArgs, tree(Call, PickupArgs)).
+	proc_id_to_int(ProcId, ProcInt),
+	Call = node([call(ModuleName, PredName, Arity, ProcInt)]),
+	determinism_to_code_model(Detism, CodeModel),
+	( CodeModel = model_semi ->
+		Check = node([semidet_success_check])
+	;
+		Check = empty
+	),
+	Code = tree(PlaceArgs, tree(Call, tree(Check, PickupArgs))).
 
 	% Generate bytecode for a call to a builtin.
 
@@ -304,7 +367,7 @@ bytecode_gen__map_test(ByteInfo, Rval, Code) :-
 		bytecode_gen__map_arg(ByteInfo, X, ByteX),
 		Code = node([builtin_untest(Unop, ByteX)])
 	;
-		error("builtin test is not a unary or binary operator")
+		error("builtin test is not in a recognized form")
 	).
 
 :- pred bytecode_gen__map_assign(byte_info::in, var::in, rval::in,
@@ -320,8 +383,12 @@ bytecode_gen__map_assign(ByteInfo, Var, Rval, Code) :-
 		bytecode_gen__map_arg(ByteInfo, X, ByteX),
 		bytecode_gen__map_var(ByteInfo, Var, ByteVar),
 		Code = node([builtin_unop(Unop, ByteX, ByteVar)])
+	; Rval = var(X) ->
+		bytecode_gen__map_var(ByteInfo, X, ByteX),
+		bytecode_gen__map_var(ByteInfo, Var, ByteVar),
+		Code = node([assign(ByteVar, ByteX)])
 	;
-		error("builtin assignment is not a unary or binary operator")
+		error("builtin assignment is not in a recognized form")
 	).
 
 :- pred bytecode_gen__map_arg(byte_info::in, rval::in, byte_arg::out) is det.
@@ -351,12 +418,24 @@ bytecode_gen__unify(construct(Var, ConsId, Args, UniModes), _, _, ByteInfo,
 	bytecode_gen__map_vars(ByteInfo, Args, ByteArgs),
 	bytecode_gen__map_cons_id(ByteInfo, Var, ConsId, ByteConsId),
 	(
-		bytecode_gen__map_uni_modes(UniModes, Args, ByteInfo, Dirs),
-		bytecode_gen__all_dirs_same(Dirs, to_var)
+		ByteConsId = pred_const(_, _, _, _)
 	->
 		Code = node([construct(ByteVar, ByteConsId, ByteArgs)])
 	;
-		error("invalid mode for construction unification")
+		% Don't call bytecode_gen__map_uni_modes until after
+		% the pred_const test fails, since the arg-modes on
+		% unifications that create closures aren't like other arg-modes.
+		bytecode_gen__map_uni_modes(UniModes, Args, ByteInfo, Dirs),
+		(
+			bytecode_gen__all_dirs_same(Dirs, to_var)
+		->
+			Code = node([construct(ByteVar, ByteConsId, ByteArgs)])
+		;
+			assoc_list__from_corresponding_lists(ByteArgs, Dirs,
+				Pairs),
+			Code = node([complex_construct(ByteVar, ByteConsId,
+				Pairs)])
+		)
 	).
 bytecode_gen__unify(deconstruct(Var, ConsId, Args, UniModes, _), _, _, ByteInfo,
 		Code) :-
@@ -381,7 +460,7 @@ bytecode_gen__unify(simple_test(Var1, Var2), _, _, ByteInfo, Code) :-
 	bytecode_gen__map_var(ByteInfo, Var2, ByteVar2),
 	Code = node([test(ByteVar1, ByteVar2)]).
 bytecode_gen__unify(complicated_unify(_, _), _Var, _RHS, _ByteInfo, _Code) :-
-	error("we do not handle complicated unifications yet").
+	error("complicated unifications should have been handled by polymorphism.m").
 
 :- pred bytecode_gen__map_uni_modes(list(uni_mode)::in, list(var)::in,
 	byte_info::in, list(byte_dir)::out) is det.
@@ -405,12 +484,12 @@ bytecode_gen__map_uni_modes([UniMode | UniModes], [Arg | Args], ByteInfo,
 	->
 		Dir = to_var
 	;
-		VarMode = top_out,
-		ArgMode = top_in
+		VarMode = top_unused,
+		ArgMode = top_unused
 	->
 		Dir = to_none
 	;
-		error("invalid mode for deconstruct unification")
+		error("invalid mode for (de)construct unification")
 	),
 	bytecode_gen__map_uni_modes(UniModes, Args, ByteInfo, Dirs).
 bytecode_gen__map_uni_modes([], [_|_], _, _) :-
@@ -429,56 +508,63 @@ bytecode_gen__all_dirs_same([Dir | Dirs], Dir) :-
 
 	% Generate bytecode for a conjunction
 
-:- pred bytecode_gen__conj(list(hlds__goal)::in, byte_info::in,
-	int::in, int::out, byte_tree::out) is det.
+:- pred bytecode_gen__conj(list(hlds_goal)::in, byte_info::in, byte_info::out,
+	byte_tree::out) is det.
 
-bytecode_gen__conj([], _, N, N, empty).
-bytecode_gen__conj([Goal | Goals], ByteInfo, N0, N, Code) :-
-	bytecode_gen__goal(Goal, ByteInfo, N0, N1, ThisCode),
-	bytecode_gen__conj(Goals, ByteInfo, N1, N, OtherCode),
+bytecode_gen__conj([], ByteInfo, ByteInfo, empty).
+bytecode_gen__conj([Goal | Goals], ByteInfo0, ByteInfo, Code) :-
+	bytecode_gen__goal(Goal, ByteInfo0, ByteInfo1, ThisCode),
+	bytecode_gen__conj(Goals, ByteInfo1, ByteInfo, OtherCode),
 	Code = tree(ThisCode, OtherCode).
 
 %---------------------------------------------------------------------------%
 
 	% Generate bytecode for each disjunct of a disjunction.
 
-:- pred bytecode_gen__disj(list(hlds__goal)::in, byte_info::in,
-	int::in, int::out, byte_tree::out) is det.
+:- pred bytecode_gen__disj(list(hlds_goal)::in, byte_info::in,
+	int::in, byte_info::out,  byte_tree::out) is det.
 
-bytecode_gen__disj([], _, N, N, empty).
-bytecode_gen__disj([Disjunct | Disjuncts], ByteInfo, N0, N, Code) :-
-	bytecode_gen__goal(Disjunct, ByteInfo, N0, N1, ThisCode),
-	N2 is N1 + 1,
-	bytecode_gen__disj(Disjuncts, ByteInfo, N2, N, OtherCode),
-	Code = tree(
-		tree(
-			node([enter_disjunct(N1)]),
-			ThisCode),
-		tree(
-			node([endof_disjunct, label(N1)]),
-			OtherCode)).
+bytecode_gen__disj([], _, _, _, _) :-
+	error("empty disjunction in bytecode_gen__disj").
+bytecode_gen__disj([Disjunct | Disjuncts], ByteInfo0, EndLabel,
+		ByteInfo, Code) :-
+	bytecode_gen__goal(Disjunct, ByteInfo0, ByteInfo1, ThisCode),
+	( Disjuncts = [] ->
+		EnterCode = node([enter_disjunct(-1)]),
+		EndofCode = node([endof_disjunct(EndLabel)]),
+		Code = tree(EnterCode, tree(ThisCode, EndofCode)),
+		ByteInfo = ByteInfo1
+	;
+		bytecode_gen__disj(Disjuncts, ByteInfo1, EndLabel,
+			ByteInfo2, OtherCode),
+		bytecode_gen__get_next_label(ByteInfo2, NextLabel, ByteInfo),
+		EnterCode = node([enter_disjunct(NextLabel)]),
+		EndofCode = node([endof_disjunct(EndLabel), label(NextLabel)]),
+		Code =
+			tree(EnterCode,
+			tree(ThisCode,
+			tree(EndofCode,
+			     OtherCode)))
+	).
 
 %---------------------------------------------------------------------------%
 
 	% Generate bytecode for each arm of a switch.
 
 :- pred bytecode_gen__switch(list(case)::in, var::in, byte_info::in,
-	int::in, int::out, byte_tree::out) is det.
+	int::in, byte_info::out, byte_tree::out) is det.
 
-bytecode_gen__switch([], _, _, N, N, empty).
-bytecode_gen__switch([case(ConsId, Goal) | Cases], Var, ByteInfo, N0, N, Code)
-		:-
+bytecode_gen__switch([], _, ByteInfo, _, ByteInfo, empty).
+bytecode_gen__switch([case(ConsId, Goal) | Cases], Var, ByteInfo0, EndLabel,
+		ByteInfo, Code) :-
 	bytecode_gen__map_cons_id(ByteInfo, Var, ConsId, ByteConsId),
-	bytecode_gen__goal(Goal, ByteInfo, N0, N1, ThisCode),
-	N2 is N1 + 1,
-	bytecode_gen__switch(Cases, Var, ByteInfo, N2, N, OtherCode),
-	Code = tree(
-		tree(
-			node([enter_switch_arm(ByteConsId, N1)]),
-			ThisCode),
-		tree(
-			node([endof_switch_arm, label(N1)]),
-			OtherCode)).
+	bytecode_gen__goal(Goal, ByteInfo0, ByteInfo1, ThisCode),
+	bytecode_gen__switch(Cases, Var, ByteInfo1, EndLabel,
+		ByteInfo2, OtherCode),
+	bytecode_gen__get_next_label(ByteInfo2, NextLabel, ByteInfo),
+	EnterCode = node([enter_switch_arm(ByteConsId, NextLabel)]),
+	EndofCode = node([endof_switch_arm(EndLabel), label(NextLabel)]),
+	Code = tree(EnterCode, tree(ThisCode, tree(EndofCode, OtherCode))).
 
 %---------------------------------------------------------------------------%
 
@@ -490,17 +576,24 @@ bytecode_gen__map_cons_id(ByteInfo, Var, ConsId, ByteConsId) :-
 	(
 		ConsId = cons(Functor, Arity),
 		(
-			Functor = unqualified(FunctorName)
+			% Everything other than characters should
+			% be module qualified.
+			Functor = unqualified(FunctorName),
+			string__to_char_list(FunctorName, FunctorList),
+			( FunctorList = [Char] ->
+				ByteConsId = char_const(Char)
+			;	
+				error("bytecode_gen__map_cons_id: unqualified cons_id is not a char_const")
+			)
 		;
-			% Should have been transformed into a function call
-			% or pred_const.
-			Functor = qualified(_, _),
-			error("bytecode_gen__map_cons_id: qualified cons_id")
-		),
-		bytecode_gen__get_var_type(ByteInfo, Var, Type),
-		code_util__cons_id_to_tag(ConsId, Type, ModuleInfo, ConsTag),
-		bytecode_gen__map_cons_tag(ConsTag, ByteConsTag),
-		ByteConsId = cons(FunctorName, Arity, ByteConsTag)
+			Functor = qualified(ModuleName, FunctorName),
+			bytecode_gen__get_var_type(ByteInfo, Var, Type),
+			code_util__cons_id_to_tag(ConsId,
+				Type, ModuleInfo, ConsTag),
+			bytecode_gen__map_cons_tag(ConsTag, ByteConsTag),
+			ByteConsId = cons(ModuleName, FunctorName,
+				Arity, ByteConsTag)
+		)
 	;
 		ConsId = int_const(IntVal),
 		ByteConsId = int_const(IntVal)
@@ -513,12 +606,14 @@ bytecode_gen__map_cons_id(ByteInfo, Var, ConsId, ByteConsId) :-
 	;
 		ConsId = pred_const(PredId, ProcId),
 		predicate_id(ModuleInfo, PredId, ModuleName, PredName, Arity),
-		ByteConsId = pred_const(ModuleName, PredName, Arity, ProcId)
+		proc_id_to_int(ProcId, ProcInt),
+		ByteConsId = pred_const(ModuleName, PredName, Arity, ProcInt)
 	;
 		ConsId = code_addr_const(PredId, ProcId),
 		predicate_id(ModuleInfo, PredId, ModuleName, PredName, Arity),
+		proc_id_to_int(ProcId, ProcInt),
 		ByteConsId = code_addr_const(ModuleName, PredName, Arity,
-			ProcId)
+			ProcInt)
 	;
 		ConsId = base_type_info_const(ModuleName, TypeName, TypeArity),
 		ByteConsId = base_type_info_const(ModuleName, TypeName,
@@ -535,8 +630,7 @@ bytecode_gen__map_cons_tag(complicated_constant_tag(Primary, Secondary),
 	complicated_constant_tag(Primary, Secondary)).
 bytecode_gen__map_cons_tag(string_constant(_), _) :-
 	error("string_constant cons tag for non-string_constant cons id").
-bytecode_gen__map_cons_tag(int_constant(_), _) :-
-	error("int_constant cons tag for non-int_constant cons id").
+bytecode_gen__map_cons_tag(int_constant(IntVal), enum_tag(IntVal)).
 bytecode_gen__map_cons_tag(float_constant(_), _) :-
 	error("float_constant cons tag for non-float_constant cons id").
 bytecode_gen__map_cons_tag(pred_closure_tag(_, _), _) :-
@@ -565,25 +659,29 @@ bytecode_gen__create_varmap([Var | VarList], VarSet, VarTypes, N0,
 
 %---------------------------------------------------------------------------%(
 
-:- type byte_info	--->	byte_info(
-					map(var, byte_var),
-					map(var, type),
-					module_info).
+:- type byte_info
+	--->	byte_info(
+			map(var, byte_var),
+			map(var, type),
+			module_info,
+			int,		% next label number to use
+			int		% next temp number to use
+		).
 
 :- pred bytecode_gen__init_byte_info(module_info::in, map(var, byte_var)::in,
 	map(var, type)::in, byte_info::out) is det.
 
 bytecode_gen__init_byte_info(ModuleInfo, VarMap, VarTypes, ByteInfo) :-
-	ByteInfo = byte_info(VarMap, VarTypes, ModuleInfo).
+	ByteInfo = byte_info(VarMap, VarTypes, ModuleInfo, 0, 0).
 
 :- pred bytecode_gen__get_module_info(byte_info::in, module_info::out) is det.
 
-bytecode_gen__get_module_info(byte_info(_, _, ModuleInfo), ModuleInfo).
+bytecode_gen__get_module_info(byte_info(_, _, ModuleInfo, _, _), ModuleInfo).
 
 :- pred bytecode_gen__map_vars(byte_info::in,
 	list(var)::in, list(byte_var)::out) is det.
 
-bytecode_gen__map_vars(byte_info(VarMap, _, _), Vars, ByteVars) :-
+bytecode_gen__map_vars(byte_info(VarMap, _, _, _, _), Vars, ByteVars) :-
 	bytecode_gen__map_vars_2(VarMap, Vars, ByteVars).
 
 :- pred bytecode_gen__map_vars_2(map(var, byte_var)::in,
@@ -596,12 +694,32 @@ bytecode_gen__map_vars_2(VarMap, [Var | Vars], [ByteVar | ByteVars]) :-
 
 :- pred bytecode_gen__map_var(byte_info::in, var::in, byte_var::out) is det.
 
-bytecode_gen__map_var(byte_info(VarMap, _, _), Var, ByteVar) :-
+bytecode_gen__map_var(byte_info(VarMap, _, _, _, _), Var, ByteVar) :-
 	map__lookup(VarMap, Var, ByteVar).
 
 :- pred bytecode_gen__get_var_type(byte_info::in, var::in, (type)::out) is det.
 
-bytecode_gen__get_var_type(byte_info(_, VarTypes, _), Var, Type) :-
+bytecode_gen__get_var_type(byte_info(_, VarTypes, _, _, _), Var, Type) :-
 	map__lookup(VarTypes, Var, Type).
 
+:- pred bytecode_gen__get_next_label(byte_info::in, int::out, byte_info::out)
+	is det.
+
+bytecode_gen__get_next_label(ByteInfo0, Label0, ByteInfo) :-
+	ByteInfo0 = byte_info(VarMap, VarTypes, ModuleInfo, Label0, Temp),
+	Label is Label0 + 1,
+	ByteInfo  = byte_info(VarMap, VarTypes, ModuleInfo, Label,  Temp).
+
+:- pred bytecode_gen__get_next_temp(byte_info::in, int::out, byte_info::out)
+	is det.
+
+bytecode_gen__get_next_temp(ByteInfo0, Temp0, ByteInfo) :-
+	ByteInfo0 = byte_info(VarMap, VarTypes, ModuleInfo, Label, Temp0),
+	Temp is Temp0 + 1,
+	ByteInfo  = byte_info(VarMap, VarTypes, ModuleInfo, Label, Temp).
+
+:- pred bytecode_gen__get_counts(byte_info::in, int::out, int::out) is det.
+
+bytecode_gen__get_counts(byte_info(_, _, _, Label, Temp), Label, Temp).
+
 %---------------------------------------------------------------------------%
