@@ -1,5 +1,5 @@
 %---------------------------------------------------------------------------%
-% Copyright (C) 1995 University of Melbourne.
+% Copyright (C) 1994-1997 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %---------------------------------------------------------------------------%
@@ -18,36 +18,29 @@
 
 :- interface.
 
-:- import_module hlds_pred, hlds_goal, llds, code_info.
+:- import_module hlds_pred, llds, code_info.
+:- import_module term.
 
 :- pred call_gen__generate_higher_order_call(code_model, var, list(var),
-				list(type), list(mode), determinism,
-				code_tree, code_info, code_info).
-:- mode call_gen__generate_higher_order_call(in, in, in, in, in, in, out,
+			list(type), list(mode), determinism, hlds_goal_info,
+			code_tree, code_info, code_info).
+:- mode call_gen__generate_higher_order_call(in, in, in, in, in, in, in, out,
 				in, out) is det.
 
-:- pred call_gen__generate_det_call(pred_id, proc_id, list(var),
-					code_tree, code_info, code_info).
-:- mode call_gen__generate_det_call(in, in, in, out, in, out) is det.
-
-:- pred call_gen__generate_semidet_call(pred_id, proc_id, list(var),
-					code_tree, code_info, code_info).
-:- mode call_gen__generate_semidet_call(in, in, in, out, in, out) is det.
-
-:- pred call_gen__generate_nondet_call(pred_id, proc_id, list(var),
-					code_tree, code_info, code_info).
-:- mode call_gen__generate_nondet_call(in, in, in, out, in, out) is det.
+:- pred call_gen__generate_call(code_model, pred_id, proc_id, list(var),
+			hlds_goal_info, code_tree, code_info, code_info).
+:- mode call_gen__generate_call(in, in, in, in, in, out, in, out) is det.
 
 :- pred call_gen__generate_det_builtin(pred_id, proc_id, list(var),
-					code_tree, code_info, code_info).
+			code_tree, code_info, code_info).
 :- mode call_gen__generate_det_builtin(in, in, in, out, in, out) is det.
 
 :- pred call_gen__generate_semidet_builtin(pred_id, proc_id, list(var),
-					code_tree, code_info, code_info).
+			code_tree, code_info, code_info).
 :- mode call_gen__generate_semidet_builtin(in, in, in, out, in, out) is det.
 
 :- pred call_gen__generate_nondet_builtin(pred_id, proc_id, list(var),
-					code_tree, code_info, code_info).
+			code_tree, code_info, code_info).
 :- mode call_gen__generate_nondet_builtin(in, in, in, out, in, out) is
 					erroneous.
 
@@ -58,6 +51,10 @@
 	is det.
 */
 
+:- pred call_gen__partition_args(assoc_list(var, arg_info),
+						list(var), list(var)).
+:- mode call_gen__partition_args(in, out, out) is det.
+
 :- pred call_gen__input_arg_locs(list(pair(var, arg_info)), 
 				list(pair(var, arg_loc))).
 :- mode call_gen__input_arg_locs(in, out) is det.
@@ -74,146 +71,246 @@
 
 :- implementation.
 
-:- import_module hlds_module, hlds_data, prog_data, code_util, globals.
-:- import_module arg_info, type_util, mode_util, shapes, unify_proc.
+:- import_module hlds_module, hlds_goal, hlds_data, prog_data, code_util.
+:- import_module arg_info, type_util, mode_util, unify_proc, instmap.
+:- import_module trace, globals.
 :- import_module bool, int, list, assoc_list, tree, set, map.
 :- import_module std_util, require.
 
-	% To generate a call to a deterministic predicate, first
-	% we get the arginfo for the callee.
-	% We then save any live variables onto the stack, clear any
-	% "reserved" registers (which get allocated as temporaries
-	% during expression generation), and then setup the registers
-	% for the procedure call. We then branch to the procedure
-	% and rebuild the register information to reflect the state
-	% when the callee returns.
-call_gen__generate_det_call(PredId, ModeId, Arguments, Code) -->
-	code_info__get_pred_proc_arginfo(PredId, ModeId, ArgInfo),
-	{ assoc_list__from_corresponding_lists(Arguments, ArgInfo, Args) },
-	{ call_gen__select_out_args(Args, OutArgs) },
-	call_gen__save_variables(OutArgs, CodeA),
-	code_info__setup_call(Args, caller, CodeB),
-	code_info__get_next_label(ReturnLabel),
-	code_info__get_module_info(ModuleInfo),
-	{ call_gen__input_args(ArgInfo, InputArguments) },
-	call_gen__generate_call_livevals(OutArgs, InputArguments, CodeC0),
-	{ call_gen__output_arg_locs(Args, OutputArguments) },
-	call_gen__generate_return_livevals(OutArgs, OutputArguments,
-		OutLiveVals),
-	code_info__make_entry_label(ModuleInfo, PredId, ModeId, yes, Address),
-	{ CodeC1 = node([
-		call(Address, label(ReturnLabel), OutLiveVals, det)
-			- "branch to det procedure",
-		label(ReturnLabel) - "Continuation label"
-	]) },
-	{ Code = tree(CodeA, tree(CodeB, tree(CodeC0, CodeC1))) },
-	call_gen__rebuild_registers(Args).
-
 %---------------------------------------------------------------------------%
 
-	% Check whether the call was really deterministic or whether
-	% it's a nondet call for which we are pruning all solutions except the
-	% first.
+call_gen__generate_call(CodeModel, PredId, ModeId, Arguments, GoalInfo, Code)
+		-->
 
-call_gen__generate_semidet_call(PredId, ProcId, Arguments, Code) -->
-	code_info__get_module_info(ModuleInfo),
-	{ module_info_preds(ModuleInfo, Preds) },
-	{ map__lookup(Preds, PredId, PredInfo) },
-	{ pred_info_procedures(PredInfo, Procs) },
-	{ map__lookup(Procs, ProcId, ProcInfo) },
-	{ proc_info_interface_code_model(ProcInfo, CodeModel) },
-	( { CodeModel = model_semi } ->
-		call_gen__generate_semidet_call_2(PredId, ProcId, Arguments,
-			Code)
-	;
-		call_gen__generate_nondet_call(PredId, ProcId, Arguments, Code)
-	).
-
-:- pred call_gen__generate_semidet_call_2(pred_id, proc_id, list(var),
-					code_tree, code_info, code_info).
-:- mode call_gen__generate_semidet_call_2(in, in, in, out, in, out) is det.
-
-	% To generate a call to a semideterministic predicate, first
-	% we get the arginfo for the callee.
-	% We then save any live variables onto the stack, clear any
-	% "reserved" registers (which get allocated as temporaries
-	% during expression generation), and then setup the registers
-	% for the procedure call. We then branch to the procedure
-	% and rebuild the register information to reflect the state
-	% when the callee returns.
-	% On return we test the value in register r1 to see if the
-	% callee succeeded or failed. In the event of failure
-	% we branch to the appropriate continuation as generated by
-	% code_info__generate_failure.
-
-call_gen__generate_semidet_call_2(PredId, ModeId, Arguments, Code) -->
+		% find out which arguments are input and which are output
 	code_info__get_pred_proc_arginfo(PredId, ModeId, ArgInfo),
-	{ assoc_list__from_corresponding_lists(Arguments, ArgInfo, Args) },
-	{ call_gen__select_out_args(Args, OutArgs) },
-	call_gen__save_variables(OutArgs, CodeA),
-	code_info__setup_call(Args, caller, CodeB),
-	code_info__get_next_label(ReturnLabel),
-	code_info__get_module_info(ModuleInfo),
-	{ call_gen__input_args(ArgInfo, InputArguments) },
-	call_gen__generate_call_livevals(OutArgs, InputArguments, CodeC0),
-	{ call_gen__output_arg_locs(Args, OutputArguments) },
-	call_gen__generate_return_livevals(OutArgs, OutputArguments,	
-		OutLiveVals),
-	code_info__make_entry_label(ModuleInfo, PredId, ModeId, yes, Address),
-	{ CodeC1 = node([
-		call(Address, label(ReturnLabel), OutLiveVals, semidet)
-			- "branch to semidet procedure",
-		label(ReturnLabel)
-			- "Continuation label"
-	]) },
-	call_gen__rebuild_registers(Args),
-	code_info__generate_failure(FailCode),
-	code_info__get_next_label(ContLab),
-	{ CodeD = tree(node([
-		if_val(lval(reg(r(1))), label(ContLab)) -
-			"Test for success"
-		]), tree(FailCode, node([ label(ContLab) - "" ]))) },
+	{ assoc_list__from_corresponding_lists(Arguments, ArgInfo, ArgsInfos) },
 
-	{ Code = tree(CodeA, tree(CodeB, tree(tree(CodeC0, CodeC1), CodeD))) }.
-
-%---------------------------------------------------------------------------%
-
-call_gen__generate_nondet_call(PredId, ModeId, Arguments, Code) -->
-	code_info__get_pred_proc_arginfo(PredId, ModeId, ArgInfo),
-	{ assoc_list__from_corresponding_lists(Arguments, ArgInfo, Args) },
-	{ call_gen__select_out_args(Args, OutArgs) },
+		% save the known variables on the stack, except those
+		% generated by this call
+	{ call_gen__select_out_args(ArgsInfos, OutArgs) },
 	call_gen__save_variables(OutArgs, SaveCode),
-	code_info__unset_failure_cont(FlushCode),
-	code_info__setup_call(Args, caller, SetupCode),
-	code_info__get_next_label(ReturnLabel),
-	code_info__get_module_info(ModuleInfo),
+
+		% save possibly unknown variables on the stack as well
+		% if they may be needed on backtracking, and figure out the
+		% call model
+	call_gen__prepare_for_call(CodeModel, FlushCode, CallModel, _),
+
+		% move the input arguments to their registers
+	code_info__setup_call(ArgsInfos, caller, SetupCode),
+
+	code_info__get_maybe_trace_info(MaybeTraceInfo),
+	( { MaybeTraceInfo = yes(TraceInfo) } ->
+		{ trace__generate_depth_reset_code(TraceInfo, TraceCode) }
+	;
+		{ TraceCode = empty }
+	),
+
+		% figure out what locations are live at the call point,
+		% for use by the value numbering optimization
 	{ call_gen__input_args(ArgInfo, InputArguments) },
 	call_gen__generate_call_livevals(OutArgs, InputArguments, LiveCode),
-	{ call_gen__output_arg_locs(Args, OutputArguments) },
+
+		% figure out what variables will be live at the return point,
+		% and where, for use in the accurate garbage collector
+	code_info__get_instmap(InstMap),
+	{ goal_info_get_instmap_delta(GoalInfo, InstMapDelta) },
+	{ instmap__apply_instmap_delta(InstMap, InstMapDelta,
+		AfterCallInstMap) },
+	{ call_gen__output_arg_locs(ArgsInfos, OutputArguments) },
 	call_gen__generate_return_livevals(OutArgs, OutputArguments,
-		OutLiveVals),
+		AfterCallInstMap, OutLiveVals),
+
+		% make the call
+	code_info__get_module_info(ModuleInfo),
 	code_info__make_entry_label(ModuleInfo, PredId, ModeId, yes, Address),
-	code_info__may_use_nondet_tailcall(TailCall),
-	{ CallModel = nondet(TailCall) },
+	code_info__get_next_label(ReturnLabel),
+	{ call_gen__call_comment(CodeModel, CallComment) },
 	{ CallCode = node([
 		call(Address, label(ReturnLabel), OutLiveVals, CallModel)
-			- "branch to nondet procedure",
+			- CallComment,
 		label(ReturnLabel)
-			- "Continuation label"
+			- "continuation label"
 	]) },
-	call_gen__rebuild_registers(Args),
+
+	call_gen__rebuild_registers(ArgsInfos),
+	call_gen__handle_failure(CodeModel, FailHandlingCode),
+
 	{ Code =
 		tree(SaveCode,
 		tree(FlushCode,
 		tree(SetupCode,
+		tree(TraceCode,
 		tree(LiveCode,
-		     CallCode))))
+		tree(CallCode,
+		     FailHandlingCode))))))
 	}.
 
+%---------------------------------------------------------------------------%
+
+	%
+	% for a higher-order call,
+	% we split the arguments into inputs and outputs, put the inputs
+	% in the locations expected by do_call_<detism>_closure in
+	% runtime/call.mod, generate the call to do_call_<detism>_closure,
+	% and pick up the outputs from the locations that we know
+	% runtime/call.mod leaves them in.
+	%
+	% lambda.m transforms the generated lambda predicates to
+	% make sure that all inputs come before all outputs, so that
+	% runtime/call.mod doesn't have trouble figuring out which registers
+	% the arguments go in.
+	%
+
+call_gen__generate_higher_order_call(_OuterCodeModel, PredVar, Args, Types,
+		Modes, Det, GoalInfo, Code) -->
+	{ determinism_to_code_model(Det, CodeModel) },
+	code_info__get_globals(Globals),
+	code_info__get_module_info(ModuleInfo),
+	{ globals__get_args_method(Globals, ArgsMethod) },
+	{ make_arg_infos(ArgsMethod, Types, Modes, CodeModel, ModuleInfo,
+		ArgInfos) },
+	{ assoc_list__from_corresponding_lists(Args, ArgInfos, ArgsInfos) },
+	{ call_gen__partition_args(ArgsInfos, InVars, OutVars) },
+	code_info__succip_is_used,
+	{ set__list_to_set(OutVars, OutArgs) },
+	call_gen__save_variables(OutArgs, SaveCode),
+
+	call_gen__prepare_for_call(CodeModel, FlushCode, CallModel,
+		RuntimeAddr),
+
+		% place the immediate input arguments in registers
+		% starting at r4.
+	call_gen__generate_immediate_args(InVars, 4, InLocs, ImmediateCode),
+	code_info__generate_stack_livevals(OutArgs, LiveVals0),
+	{ set__insert_list(LiveVals0,
+		[reg(r, 1), reg(r, 2), reg(r, 3) | InLocs], LiveVals) },
+	(
+		{ CodeModel = model_semi }
+	->
+		{ FirstArg = 2 }
+	;
+		{ FirstArg = 1 }
+	),
+
+	{ call_gen__outvars_to_outargs(OutVars, FirstArg, OutArguments) },
+	{ call_gen__output_arg_locs(OutArguments, OutLocs) },
+	code_info__get_instmap(InstMap),
+	{ goal_info_get_instmap_delta(GoalInfo, InstMapDelta) },
+	{ instmap__apply_instmap_delta(InstMap, InstMapDelta,
+		AfterCallInstMap) },
+	call_gen__generate_return_livevals(OutArgs, OutLocs, AfterCallInstMap, 
+		OutLiveVals),
+
+	code_info__produce_variable(PredVar, PredVarCode, PredRVal),
+	(
+		{ PredRVal = lval(reg(r, 1)) }
+	->
+		{ CopyCode = empty }
+	;
+		{ CopyCode = node([
+			assign(reg(r, 1), PredRVal) - "Copy pred-term"
+		]) }
+	),
+
+	{ list__length(InVars, NInVars) },
+	{ list__length(OutVars, NOutVars) },
+	{ ArgNumCode = node([
+		assign(reg(r, 2), const(int_const(NInVars))) -
+			"Assign number of immediate input arguments",
+		assign(reg(r, 3), const(int_const(NOutVars))) -
+			"Assign number of output arguments"
+	]) },
+
+	code_info__get_maybe_trace_info(MaybeTraceInfo),
+	( { MaybeTraceInfo = yes(TraceInfo) } ->
+		{ trace__generate_depth_reset_code(TraceInfo, TraceCode) }
+	;
+		{ TraceCode = empty }
+	),
+
+	code_info__get_next_label(ReturnLabel),
+	{ CallCode = node([
+		livevals(LiveVals)
+			- "",
+		call(RuntimeAddr, label(ReturnLabel), OutLiveVals, CallModel)
+			- "setup and call higher order pred",
+		label(ReturnLabel)
+			- "Continuation label"
+	]) },
+
+	call_gen__rebuild_registers(OutArguments),
+	call_gen__handle_failure(CodeModel, FailHandlingCode),
+
+	{ Code =
+		tree(SaveCode,
+		tree(FlushCode,
+		tree(ImmediateCode,
+		tree(PredVarCode,
+		tree(CopyCode,
+		tree(ArgNumCode,
+		tree(TraceCode,
+		tree(CallCode,
+		     FailHandlingCode))))))))
+	}.
+
+%---------------------------------------------------------------------------%
+
+:- pred call_gen__prepare_for_call(code_model, code_tree, call_model,
+	code_addr, code_info, code_info).
+:- mode call_gen__prepare_for_call(in, out, out, out, in, out) is det.
+
+call_gen__prepare_for_call(CodeModel, FlushCode, CallModel, RuntimeAddr) -->
+	(
+		{ CodeModel = model_det },
+		{ CallModel = det },
+		{ RuntimeAddr = do_det_closure },
+		{ FlushCode = empty }
+	;
+		{ CodeModel = model_semi },
+		{ CallModel = semidet },
+		{ RuntimeAddr = do_semidet_closure },
+		{ FlushCode = empty }
+	;
+		{ CodeModel = model_non },
+		code_info__may_use_nondet_tailcall(TailCall),
+		{ CallModel = nondet(TailCall) },
+		{ RuntimeAddr = do_nondet_closure },
+		code_info__unset_failure_cont(FlushCode)
+	).
+
+:- pred call_gen__handle_failure(code_model, code_tree, code_info, code_info).
+:- mode call_gen__handle_failure(in, out, in, out ) is det.
+
+call_gen__handle_failure(CodeModel, FailHandlingCode) -->
+	( { CodeModel = model_semi } ->
+		code_info__generate_failure(FailCode),
+		code_info__get_next_label(ContLab),
+		{ FailTestCode = node([
+			if_val(lval(reg(r, 1)), label(ContLab))
+				- "test for success"
+		]) },
+		{ ContLabelCode = node([label(ContLab) - ""]) },
+		{ FailHandlingCode =
+			tree(FailTestCode,
+			tree(FailCode, 
+			     ContLabelCode))
+		}
+	;
+		{ FailHandlingCode = empty }
+	).
+
+:- pred call_gen__call_comment(code_model, string).
+:- mode call_gen__call_comment(in, out) is det.
+
+call_gen__call_comment(model_det,  "branch to det procedure").
+call_gen__call_comment(model_semi, "branch to semidet procedure").
+call_gen__call_comment(model_non,  "branch to nondet procedure").
+
 %---------------------------------------------------------------------------%
 
 call_gen__save_variables(Args, Code) -->
-	code_info__get_live_variables(Variables0),
+	code_info__get_known_variables(Variables0),
 	{ set__list_to_set(Variables0, Vars0) },
 	{ set__difference(Vars0, Args, Vars1) },
 	code_info__get_globals(Globals),
@@ -222,7 +319,7 @@ call_gen__save_variables(Args, Code) -->
 		{ GC_Method = accurate }
 	->
 		code_info__get_proc_info(ProcInfo),
-		{ proc_info_get_used_typeinfos_setwise(ProcInfo, Vars1, 
+		{ proc_info_get_typeinfo_vars_setwise(ProcInfo, Vars1, 
 			TypeInfoVars) },
 		{ set__union(Vars1, TypeInfoVars, Vars) }
 	;
@@ -260,7 +357,7 @@ call_gen__rebuild_registers_2([Var - arg_info(ArgLoc, Mode) | Args]) -->
 		{ Mode = top_out }
 	->
 		{ code_util__arg_loc_to_register(ArgLoc, Register) },
-		code_info__set_var_location(Var, reg(Register))
+		code_info__set_var_location(Var, Register)
 	;
 		{ true }
 	),
@@ -303,7 +400,7 @@ call_gen__generate_semidet_builtin(PredId, ProcId, Args, Code) -->
 		;
 			{ error("Unknown builtin predicate") }
 		),
-		code_info__generate_test_and_fail(Rval, TestCode),
+		code_info__fail_if_rval_is_false(Rval, TestCode),
 		( { Assign = yes(Var - AssignRval) } ->
 			code_info__cache_expression(Var, AssignRval)
 		;
@@ -337,12 +434,8 @@ call_gen__generate_builtin_arg(Rval0, Rval, Code) -->
 %---------------------------------------------------------------------------%
 %---------------------------------------------------------------------------%
 
-:- pred call_gen__partition_args(assoc_list(arg_info, var),
-						list(var), list(var)).
-:- mode call_gen__partition_args(in, out, out) is det.
-
 call_gen__partition_args([], [], []).
-call_gen__partition_args([arg_info(_Loc,Mode) - V | Rest], Ins, Outs) :-
+call_gen__partition_args([V - arg_info(_Loc,Mode) | Rest], Ins, Outs) :-
 	(
 		Mode = top_in
 	->
@@ -378,7 +471,7 @@ call_gen__generate_complicated_unify(Var1, Var2, UniMode, CanFail, Code) -->
 			CodeC0),
 		{ call_gen__output_arg_locs(Args, OutputArguments) },
 		call_gen__generate_return_livevals(OutArgs, OutputArguments, 
-						OutLiveVals),
+				GoalInfo, OutLiveVals),
 		{ code_util__make_uni_label(ModuleInfo, VarTypeId, ModeNum,
 			UniLabel) },
 		{ Address = imported(UniLabel) },
@@ -514,21 +607,23 @@ call_gen__generate_call_livevals(OutArgs, InputArgs, Code) -->
 call_gen__insert_arg_livevals([], LiveVals, LiveVals).
 call_gen__insert_arg_livevals([L | As], LiveVals0, LiveVals) :-
 	code_util__arg_loc_to_register(L, R),
-	set__insert(LiveVals0, reg(R), LiveVals1),
+	set__insert(LiveVals0, R, LiveVals1),
 	call_gen__insert_arg_livevals(As, LiveVals1, LiveVals).
 
 %---------------------------------------------------------------------------%
 
 :- pred call_gen__generate_return_livevals(set(var), list(pair(var, arg_loc)),
-					list(liveinfo), code_info, code_info).
-:- mode call_gen__generate_return_livevals(in, in, out, in, out) is det.
+		instmap, list(liveinfo), code_info, code_info).
+:- mode call_gen__generate_return_livevals(in, in, in, out, in, out) is det.
 
-call_gen__generate_return_livevals(OutArgs, OutputArgs, LiveVals) -->
-	code_info__generate_stack_livelvals(OutArgs, LiveVals0),
+call_gen__generate_return_livevals(OutArgs, OutputArgs, AfterCallInstMap, 
+		LiveVals) -->
+	code_info__generate_stack_livelvals(OutArgs, AfterCallInstMap, 
+		LiveVals0),
 	code_info__get_globals(Globals),
 	{ globals__get_gc_method(Globals, GC_Method) },
-	call_gen__insert_arg_livelvals(OutputArgs, GC_Method, LiveVals0, 
-		LiveVals).
+	call_gen__insert_arg_livelvals(OutputArgs, GC_Method, AfterCallInstMap,
+		LiveVals0, LiveVals).
 	
 
 % Maybe a varlist to type_id list would be a better way to do this...
@@ -536,170 +631,27 @@ call_gen__generate_return_livevals(OutArgs, OutputArgs, LiveVals) -->
 %---------------------------------------------------------------------------%
 
 :- pred call_gen__insert_arg_livelvals(list(pair(var, arg_loc)), gc_method, 
-	list(liveinfo), list(liveinfo), code_info, code_info).
-:- mode call_gen__insert_arg_livelvals(in, in, in, out, in, out) is det.
+	instmap, list(liveinfo), list(liveinfo), code_info, code_info).
+:- mode call_gen__insert_arg_livelvals(in, in, in, in, out, in, out) is det.
 
-call_gen__insert_arg_livelvals([], _, LiveVals, LiveVals) --> [].
-call_gen__insert_arg_livelvals([Var - L | As], GC_Method, LiveVals0,
-		LiveVals) -->
+call_gen__insert_arg_livelvals([], _, _, LiveVals, LiveVals) --> [].
+call_gen__insert_arg_livelvals([Var - L | As], GC_Method, AfterCallInstMap, 
+		LiveVals0, LiveVals) -->
 	(
 		{ GC_Method = accurate }
 	->
-		code_info__get_shapes(S_Tab0),
-		code_info__variable_type(Var, Type),
-		code_info__get_module_info(Module_Info),
-		{ module_info_types(Module_Info, Type_Table) },
+		{ instmap__lookup_var(AfterCallInstMap, Var, Inst) },
 
-		% XXX we really should check that the inst of Var is
-		% XXX ground - but that would probably break things when
-		% XXX partial insts get implemented.
-		{ shapes__request_shape_number(Type - ground(shared, no), 
-			Type_Table, S_Tab0, S_Tab1, S_Number) },
+		code_info__variable_type(Var, Type),
 		{ type_util__vars(Type, TypeVars) },
-		(
-			% if not polymorphic
-			{ TypeVars = [] }
-		->
-			{ TypeParams = no }
-		;
-			code_info__find_type_infos(TypeVars, Lvals),
-			{ TypeParams = yes(Lvals) }
-		),
-		code_info__set_shapes(S_Tab1)
+		code_info__find_type_infos(TypeVars, TypeParams),
+		{ LiveVal = live_lvalue(R, var(Type, Inst), TypeParams) }
 	;
-		{ TypeParams = no },
-		{ S_Number = 0 }
+		{ LiveVal = live_lvalue(R, unwanted, []) }
 	),
 	{ code_util__arg_loc_to_register(L, R) },
-	{ LiveVal = live_lvalue(reg(R), num(S_Number), TypeParams) },
-	call_gen__insert_arg_livelvals(As, GC_Method, [LiveVal | LiveVals0], 
-		LiveVals).
-
-%---------------------------------------------------------------------------%
-
-	%
-	% for a higher-order call,
-	% we split the arguments into inputs and outputs, put the inputs
-	% in the locations expected by do_call_<detism>_closure in
-	% runtime/call.mod, generate the call to do_call_<detism>_closure,
-	% and pick up the outputs from the locations that we know
-	% runtime/call.mod leaves them in.
-	%
-	% lambda.m transforms the generated lambda predicates to
-	% make sure that all inputs come before all outputs, so that
-	% runtime/call.mod doesn't have trouble figuring out which registers
-	% the arguments go in.
-	%
-
-call_gen__generate_higher_order_call(_OuterCodeModel, PredVar, Args, Types,
-		Modes, Det, Code) -->
-	{ determinism_to_code_model(Det, InnerCodeModel) },
-	code_info__get_globals(Globals),
-	code_info__get_module_info(ModuleInfo),
-	{ globals__get_args_method(Globals, ArgsMethod) },
-	{ make_arg_infos(ArgsMethod, Types, Modes, InnerCodeModel, ModuleInfo,
-		ArgInfo) },
-	{ assoc_list__from_corresponding_lists(ArgInfo, Args, ArgsAndArgInfo) },
-	{ call_gen__partition_args(ArgsAndArgInfo, InVars, OutVars) },
-	call_gen__generate_higher_call(InnerCodeModel, PredVar, InVars, OutVars,
-		Code).
-
-:- pred call_gen__generate_higher_call(code_model, var, list(var), list(var),
-				code_tree, code_info, code_info).
-:- mode call_gen__generate_higher_call(in, in, in, in, out, in, out) is det.
-
-call_gen__generate_higher_call(CodeModel, PredVar, InVars, OutVars, Code) -->
-	code_info__set_succip_used(yes),
-	{ set__list_to_set(OutVars, OutArgs) },
-	call_gen__save_variables(OutArgs, SaveCode),
-	( { CodeModel = model_non } ->
-		code_info__unset_failure_cont(FlushCode)
-	;
-		{ FlushCode = empty }
-	),
-		% place the immediate input arguments in registers
-		% starting at r4.
-	call_gen__generate_immediate_args(InVars, 4, InLocs, ImmediateCode),
-	code_info__generate_stack_livevals(OutArgs, LiveVals0),
-	{ set__insert_list(LiveVals0,
-		[reg(r(1)), reg(r(2)), reg(r(3)) | InLocs], LiveVals) },
-	(
-		{ CodeModel = model_semi }
-	->
-		{ FirstArg = 2 }
-	;
-		{ FirstArg = 1 }
-	),
-	{ call_gen__outvars_to_outargs(OutVars, FirstArg, OutArguments) },
-	{ call_gen__output_arg_locs(OutArguments, OutLocs) },
-	call_gen__generate_return_livevals(OutArgs, OutLocs, OutLiveVals),
-	code_info__produce_variable(PredVar, PredVarCode, PredRVal),
-	(
-		{ PredRVal = lval(reg(r(1))) }
-	->
-		{ CopyCode = empty }
-	;
-		{ CopyCode = node([
-			assign(reg(r(1)), PredRVal) - "Copy pred-term"
-		])}
-	),
-	{ list__length(InVars, NInVars) },
-	{ list__length(OutVars, NOutVars) },
-	{ SetupCode = tree(CopyCode, node([
-			assign(reg(r(2)), const(int_const(NInVars))) -
-				"Assign number of immediate input arguments",
-			assign(reg(r(3)), const(int_const(NOutVars))) -
-				"Assign number of output arguments"
-		])
-	) },
-	code_info__get_next_label(ReturnLabel),
-	(
-		{ CodeModel = model_det },
-		{ CallModel = det },
-		{ RuntimeAddr = do_det_closure }
-	;
-		{ CodeModel = model_semi },
-		{ CallModel = semidet },
-		{ RuntimeAddr = do_semidet_closure }
-	;
-		{ CodeModel = model_non },
-		code_info__may_use_nondet_tailcall(TailCall),
-		{ CallModel = nondet(TailCall) },
-		{ RuntimeAddr = do_nondet_closure }
-	),
-	{ TryCallCode = node([
-		livevals(LiveVals) - "",
-		call(RuntimeAddr, label(ReturnLabel), OutLiveVals, CallModel)
-			- "setup and call higher order pred",
-		label(ReturnLabel) - "Continuation label"
-	]) },
-	call_gen__rebuild_registers(OutArguments),
-	(
-		{ CodeModel = model_semi }
-	->
-		code_info__generate_failure(FailCode),
-		code_info__get_next_label(ContLab),
-		{ TestSuccessCode = node([
-			if_val(lval(reg(r(1))), label(ContLab)) -
-				"Test for success"
-		]) },
-		{ ContLabelCode = node([label(ContLab) - ""]) },
-		{ CallCode =
-			tree(TryCallCode,
-			tree(TestSuccessCode,
-			tree(FailCode,
-			     ContLabelCode))) }
-	;
-		{ CallCode = TryCallCode }
-	),
-	{ Code =
-		tree(SaveCode,
-		tree(FlushCode,
-		tree(ImmediateCode,
-		tree(PredVarCode,
-		tree(SetupCode,
-		     CallCode)))))
-	}.
+	call_gen__insert_arg_livelvals(As, GC_Method, AfterCallInstMap, 
+		[LiveVal | LiveVals0], LiveVals).
 
 %---------------------------------------------------------------------------%
 
@@ -709,7 +661,7 @@ call_gen__generate_higher_call(CodeModel, PredVar, InVars, OutVars, Code) -->
 
 call_gen__generate_immediate_args([], _N, [], empty) --> [].
 call_gen__generate_immediate_args([V | Vs], N0, [Lval | Lvals], Code) -->
-	{ Lval = reg(r(N0)) },
+	{ Lval = reg(r, N0) },
 	code_info__place_var(V, Lval, Code0),
 	{ N1 is N0 + 1 },
 	call_gen__generate_immediate_args(Vs, N1, Lvals, Code1),
