@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1995 University of Melbourne.
+% Copyright (C) 1995-1997 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -11,6 +11,18 @@
 % abstract 'exprn_info' structure which maintains information about
 % the contents of registers, and manages the cached expressions for
 % variables.
+%
+% We make the following assumptions:
+%
+%	o  Variables can be stored in any number of distinct places.
+%	o  Registers may contain a value corresponding to more than
+%		one variable.
+%	o  Procedures are in superhomogeneous form. This means that
+%		construction unifications and builtins are not nested.
+%	o  Evaluation of arguments in construction and deconstruction
+%		unifications is lazy. This means that arguments in a
+%		`don't care' mode are ignored, and that assignments
+%		are cached.
 
 %------------------------------------------------------------------------------%
 %------------------------------------------------------------------------------%
@@ -19,23 +31,39 @@
 
 :- interface.
 
-:- import_module llds, list, varset, std_util, assoc_list, tree, options.
+:- import_module hlds_goal, llds, list, varset, assoc_list, options.
 
 :- type exprn_info.
 
-%	code_exprn__init_state(Arguments, Varset, Opts, ExprnInfo)
+%	code_exprn__init_state(Arguments, Varset, StackSlots, FollowVars,
+%			Opts, ExprnInfo)
 %		Produces an initial state of the ExprnInfo given
 %		an association list of variables and lvalues. The initial
 %		state places the given variables at their corresponding
 %		locations. The Varset parameter contains a mapping from
 %		variables to names, which is used when code is generated
-%		to provide meaningful comments. Opts gives the table of
-%		options; this is used to decide what expressions are
-%		considered constants.
+%		to provide meaningful comments. StackSlots maps each variable
+%		to its stack slot (if it has one). FollowVars is the initial
+%		follow_vars set; such sets give guidance as to what lvals
+%		(if any) each variable will be needed in next. Opts gives
+%		the table of options; this is used to decide what expressions
+%		are considered constants.
 
-:- pred code_exprn__init_state(assoc_list(var, rval), varset, option_table,
-	exprn_info).
-:- mode code_exprn__init_state(in, in, in, out) is det.
+:- pred code_exprn__init_state(assoc_list(var, rval), varset, stack_slots,
+	follow_vars, option_table, exprn_info).
+:- mode code_exprn__init_state(in, in, in, in, in, out) is det.
+
+%	code_exprn__reinit_state(VarLocs, ExprnInfo0, ExprnInfo)
+%		Produces a new state of the ExprnInfo in which the static
+%		and mostly static information (stack slot map, follow vars map,
+%		varset, option settings) comes from ExprnInfo0 but the dynamic
+%		state regarding variable locations is thrown away and then
+%		rebuilt from the information in VarLocs, an association list
+%		of variables and lvalues. The new state places the given
+%		variables at their corresponding locations.
+
+:- pred code_exprn__reinit_state(assoc_list(var, rval), exprn_info, exprn_info).
+:- mode code_exprn__reinit_state(in, in, out) is det.
 
 %	code_exprn__clobber_regs(CriticalVars, ExprnInfo0, ExprnInfo)
 %		Modifies the state ExprnInfo0 to produce ExprnInfo
@@ -105,17 +133,17 @@
 %		provide it as an Rval of the form lval(reg(_)).
 
 :- pred code_exprn__produce_var_in_reg(var, rval, code_tree,
-						exprn_info, exprn_info).
+	exprn_info, exprn_info).
 :- mode code_exprn__produce_var_in_reg(in, out, out, in, out) is det.
 
-%	code_exprn__produce_var_in_reg_or_stack(Var, Rval, Code,
+%	code_exprn__produce_var_in_reg_or_stack(Var, FollowVars, Rval, Code,
 %			ExprnInfo0, ExprnInfo)
 %		Produces a code fragment Code to evaluate Var and
 %		provide it as an Rval of the form lval(reg(_)),
 %		lval(stackvar(_)), or lval(framevar(_)).
 
 :- pred code_exprn__produce_var_in_reg_or_stack(var, rval, code_tree,
-						exprn_info, exprn_info).
+	exprn_info, exprn_info).
 :- mode code_exprn__produce_var_in_reg_or_stack(in, out, out, in, out) is det.
 
 %	code_exprn__materialize_vars_in_rval(Rval0, Rval, Code, ExprnInfo0,
@@ -130,17 +158,26 @@
 		exprn_info, exprn_info).
 :- mode code_exprn__materialize_vars_in_rval(in, out, out, in, out) is det.
 
-%	code_exprn__acquire_reg(Reg, ExprnInfo0, ExprnInfo)
-%		Finds an unused register and marks it as 'in use'.
+%	code_exprn__acquire_reg(Type, Lval, ExprnInfo0, ExprnInfo)
+%		Finds an unused register of the given type and marks it
+%		as 'in use'.
 %
-%	code_exprn__release_reg(Reg, ExprnInfo, ExprnInfo)
+%	code_exprn__acquire_reg_prefer_given(Type, Pref, Lval, ExprnI0, ExprnI)
+%		Finds an unused register of the same type as Pref and marks it
+%		as 'in use'. If Pref itself is free, assigns that.
+%
+%	code_exprn__release_reg(Lval, ExprnInfo, ExprnInfo)
 %		Marks a previously acquired reg and releases it so
 %		that it can be reused.
 
-:- pred code_exprn__acquire_reg(reg, exprn_info, exprn_info).
-:- mode code_exprn__acquire_reg(out, in, out) is det.
+:- pred code_exprn__acquire_reg(reg_type, lval, exprn_info, exprn_info).
+:- mode code_exprn__acquire_reg(in, out, in, out) is det.
 
-:- pred code_exprn__release_reg(reg, exprn_info, exprn_info).
+:- pred code_exprn__acquire_reg_prefer_given(reg_type, int, lval,
+	exprn_info, exprn_info).
+:- mode code_exprn__acquire_reg_prefer_given(in, in, out, in, out) is det.
+
+:- pred code_exprn__release_reg(lval, exprn_info, exprn_info).
 :- mode code_exprn__release_reg(in, in, out) is det.
 
 %	XXX These should be local, or their function should be folded into
@@ -154,10 +191,10 @@
 %	code_exprn__unlock_reg(Reg, ExprnInfo0, ExprnInfo)
 %		Undoes a lock operation.
 
-:- pred code_exprn__lock_reg(reg, exprn_info, exprn_info).
+:- pred code_exprn__lock_reg(lval, exprn_info, exprn_info).
 :- mode code_exprn__lock_reg(in, in, out) is det.
 
-:- pred code_exprn__unlock_reg(reg, exprn_info, exprn_info).
+:- pred code_exprn__unlock_reg(lval, exprn_info, exprn_info).
 :- mode code_exprn__unlock_reg(in, in, out) is det.
 
 %	code_exprn__clear_r1(Code)
@@ -175,13 +212,34 @@
 :- pred code_exprn__get_varlocs(exprn_info, map(var, set(rval))).
 :- mode code_exprn__get_varlocs(in, out) is det.
 
+%	code_exprn__get_stack_slots(StackSlots)
+%		Returns the table mapping each variable to its stack slot
+%		(if any).
+
+:- pred code_exprn__get_stack_slots(stack_slots, exprn_info, exprn_info).
+:- mode code_exprn__get_stack_slots(out, in, out) is det.
+
+%	code_exprn__get_follow_vars(FollowVars)
+%		Returns the table mapping each variable to the lval (if any)
+%		where it is desired next.
+
+:- pred code_exprn__get_follow_vars(follow_vars, exprn_info, exprn_info).
+:- mode code_exprn__get_follow_vars(out, in, out) is det.
+
+%	code_exprn__set_follow_vars(FollowVars)
+%		Sets the table mapping each variable to the lval (if any)
+%		where it is desired next.
+
+:- pred code_exprn__set_follow_vars(follow_vars, exprn_info, exprn_info).
+:- mode code_exprn__set_follow_vars(in, in, out) is det.
+
 %------------------------------------------------------------------------------%
 %------------------------------------------------------------------------------%
 
 :- implementation.
 
-:- import_module exprn_aux.
-:- import_module bool, map, bag, set, require, int, term, string.
+:- import_module exprn_aux, tree.
+:- import_module bool, map, bag, set, require, int, term, string, std_util.
 
 :- type var_stat	--->	evaled(set(rval))
 			;	cached(rval).
@@ -192,23 +250,37 @@
 		exprn_info(
 			varset,		% all the variables and their names
 			var_map,	% what each variable stands for
-			bag(reg),	% the 'in use' markers for regs
-			set(reg),	% extra markers for acquired regs
+			bag(lval),	% the 'in use' markers for regs
+			set(lval),	% extra markers for acquired regs
+			stack_slots,	% map vars to the stack slot (if any)
+			follow_vars,	% where vars are needed next
 			exprn_opts	% options needed for constant checks
 		).
 
 %------------------------------------------------------------------------------%
 
-code_exprn__init_state(Initializations, Varset, Options, ExprnInfo) :-
+code_exprn__init_state(VarLocs, Varset, StackSlots, FollowVars,
+		Options, ExprnInfo) :-
 	map__init(Vars0),
 	bag__init(Regs0),
-	code_exprn__init_state_2(Initializations, Vars0, Vars, Regs0, Regs),
+	code_exprn__init_state_2(VarLocs, Vars0, Vars, Regs0, Regs),
 	set__init(Acqu),
 	exprn_aux__init_exprn_opts(Options, ExprnOpts),
-	ExprnInfo = exprn_info(Varset, Vars, Regs, Acqu, ExprnOpts).
+	ExprnInfo = exprn_info(Varset, Vars, Regs, Acqu,
+		StackSlots, FollowVars, ExprnOpts).
+
+code_exprn__reinit_state(VarLocs, ExprnInfo0, ExprnInfo) :-
+	map__init(Vars0),
+	bag__init(Regs0),
+	code_exprn__init_state_2(VarLocs, Vars0, Vars, Regs0, Regs),
+	set__init(Acqu),
+	ExprnInfo0 = exprn_info(Varset, _, _, _,
+		StackSlots, FollowVars, ExprnOpts),
+	ExprnInfo = exprn_info(Varset, Vars, Regs, Acqu,
+		StackSlots, FollowVars, ExprnOpts).
 
 :- pred code_exprn__init_state_2(assoc_list(var, rval), var_map, var_map,
-							bag(reg), bag(reg)).
+	bag(lval), bag(lval)).
 :- mode code_exprn__init_state_2(in, in, out, in, out) is det.
 
 code_exprn__init_state_2([], Vars, Vars, Regs, Regs).
@@ -216,15 +288,17 @@ code_exprn__init_state_2([V - L | Rest], Vars0, Vars, Regs0, Regs) :-
 	(
 		map__search(Vars0, V, evaled(Vals0))
 	->
-		set__insert(Vals0, L, Vals)
+		set__insert(Vals0, L, Vals),
+		map__det_update(Vars0, V, evaled(Vals), Vars1)
 	;
-		set__singleton_set(Vals, L)
+		set__singleton_set(Vals, L),
+		map__det_insert(Vars0, V, evaled(Vals), Vars1)
 	),
-	map__set(Vars0, V, evaled(Vals), Vars1),
 	(
-		L = lval(reg(R))
+		L = lval(Loc),
+		Loc = reg(_, _)
 	->
-		bag__insert(Regs0, R, Regs1)
+		bag__insert(Regs0, Loc, Regs1)
 	;
 		Regs1 = Regs0
 	),
@@ -342,8 +416,7 @@ code_exprn__rval_depends_on_reg(binop(_Op, Rval0, Rval1), Vars) :-
 :- pred code_exprn__lval_depends_on_reg(lval, var_map).
 :- mode code_exprn__lval_depends_on_reg(in, in) is semidet.
 
-code_exprn__lval_depends_on_reg(reg(_), _Vars) :-
-	true.
+code_exprn__lval_depends_on_reg(reg(_, _), _Vars).
 code_exprn__lval_depends_on_reg(lvar(Var), Vars) :-
 	map__lookup(Vars, Var, Stat),
 	(
@@ -510,10 +583,10 @@ code_exprn__find_rvals_without_lval([Rval0 | Rvals0], Lval, Rvals) :-
 
 code_exprn__add_lval_reg_dependencies(Lval) -->
 	(
-		{ Lval = reg(Reg) }
+		{ Lval = reg(_, _) }
 	->
 		code_exprn__get_regs(Regs0),
-		{ bag__insert(Regs0, Reg, Regs) },
+		{ bag__insert(Regs0, Lval, Regs) },
 		code_exprn__set_regs(Regs)
 	;
 		{ Lval = field(_Tag, Rval0, Rval1) }
@@ -549,6 +622,17 @@ code_exprn__add_rval_reg_dependencies(unop(_Op, Rval)) -->
 code_exprn__add_rval_reg_dependencies(binop(_Op, Rval0, Rval1)) -->
 	code_exprn__add_rval_reg_dependencies(Rval0),
 	code_exprn__add_rval_reg_dependencies(Rval1).
+code_exprn__add_rval_reg_dependencies(mem_addr(MemRef)) -->
+	code_exprn__add_mem_ref_reg_dependencies(MemRef).
+
+:- pred code_exprn__add_mem_ref_reg_dependencies(mem_ref,
+	exprn_info, exprn_info).
+:- mode code_exprn__add_mem_ref_reg_dependencies(in, in, out) is det.
+
+code_exprn__add_mem_ref_reg_dependencies(stackvar_ref(_)) --> [].
+code_exprn__add_mem_ref_reg_dependencies(framevar_ref(_)) --> [].
+code_exprn__add_mem_ref_reg_dependencies(heap_ref(Rval, _, _)) -->
+	code_exprn__add_rval_reg_dependencies(Rval).
 
 :- pred code_exprn__add_arg_reg_dependencies(list(maybe(rval)),
 						exprn_info, exprn_info).
@@ -572,10 +656,10 @@ code_exprn__add_arg_reg_dependencies([M | Ms]) -->
 
 code_exprn__rem_lval_reg_dependencies(Lval) -->
 	(
-		{ Lval = reg(Reg) }
+		{ Lval = reg(_, _) }
 	->
 		code_exprn__get_regs(Regs0),
-		{ bag__remove(Regs0, Reg, Regs) },
+		{ bag__delete(Regs0, Lval, Regs) },
 		code_exprn__set_regs(Regs)
 	;
 		{ Lval = field(_Tag, Rval0, Rval1) }
@@ -611,6 +695,17 @@ code_exprn__rem_rval_reg_dependencies(unop(_Op, Rval)) -->
 code_exprn__rem_rval_reg_dependencies(binop(_Op, Rval0, Rval1)) -->
 	code_exprn__rem_rval_reg_dependencies(Rval0),
 	code_exprn__rem_rval_reg_dependencies(Rval1).
+code_exprn__rem_rval_reg_dependencies(mem_addr(MemRef)) -->
+	code_exprn__rem_mem_ref_reg_dependencies(MemRef).
+
+:- pred code_exprn__rem_mem_ref_reg_dependencies(mem_ref,
+	exprn_info, exprn_info).
+:- mode code_exprn__rem_mem_ref_reg_dependencies(in, in, out) is det.
+
+code_exprn__rem_mem_ref_reg_dependencies(stackvar_ref(_)) --> [].
+code_exprn__rem_mem_ref_reg_dependencies(framevar_ref(_)) --> [].
+code_exprn__rem_mem_ref_reg_dependencies(heap_ref(Rval, _, _)) -->
+	code_exprn__rem_rval_reg_dependencies(Rval).
 
 :- pred code_exprn__rem_arg_reg_dependencies(list(maybe(rval)),
 						exprn_info, exprn_info).
@@ -740,25 +835,15 @@ code_exprn__update_dependent_vars_3([R0 | Rs0], Var, Rval, [R | Rs]) -->
 :- mode code_exprn__select_rval(in, out) is det.
 
 code_exprn__select_rval(Rvals, Rval) :-
-	(
-		Rvals = []
-	->
+	( Rvals = [] ->
 		error("code_exprn__select_rval: no rvals")
-	;
-		code_exprn__select_reg(Rvals, Rval0)
-	->
+	; code_exprn__select_reg(Rvals, Rval0) ->
 		Rval = Rval0
-	;
-		code_exprn__select_simple_const(Rvals, Rval1)
-	->
+	; code_exprn__select_simple_const(Rvals, Rval1) ->
 		Rval = Rval1
-	;
-		code_exprn__select_stackvar(Rvals, Rval2)
-	->
+	; code_exprn__select_stackvar(Rvals, Rval2) ->
 		Rval = Rval2
-	;
-		Rvals = [Rval3 | _]
-	->
+	; Rvals = [Rval3 | _] ->
 		Rval = Rval3
 	;
 		error("code_exprn__select_rval: cosmic rays strike again")
@@ -769,7 +854,7 @@ code_exprn__select_rval(Rvals, Rval) :-
 
 code_exprn__select_reg([R | Rs], Rval) :-
 	(
-		R = lval(reg(_))
+		R = lval(reg(_, _))
 	->
 		Rval = R
 	;
@@ -882,18 +967,17 @@ code_exprn__cache_exprn(Var, Rval) -->
 		{ string__append("code_exprn__cache_exprn: existing definition of variable ", Name, Msg) },
 		{ error(Msg) }
 	;
-		[]
-	),
-	code_exprn__add_rval_reg_dependencies(Rval),
-	(
-		{ exprn_aux__vars_in_rval(Rval, []) }
-	->
-		{ set__singleton_set(Rvals, Rval) },
-		{ map__set(Vars0, Var, evaled(Rvals), Vars) }
-	;
-		{ map__set(Vars0, Var, cached(Rval), Vars) }
-	),
-	code_exprn__set_vars(Vars).
+		code_exprn__add_rval_reg_dependencies(Rval),
+		(
+			{ exprn_aux__vars_in_rval(Rval, []) }
+		->
+			{ set__singleton_set(Rvals, Rval) },
+			{ map__det_insert(Vars0, Var, evaled(Rvals), Vars) }
+		;
+			{ map__det_insert(Vars0, Var, cached(Rval), Vars) }
+		),
+		code_exprn__set_vars(Vars)
+	).
 
 %------------------------------------------------------------------------------%
 
@@ -934,7 +1018,7 @@ code_exprn__place_cached(Rval0, Var, Lval, Code) -->
 			% but we do need to reserve the registers
 			% needed to access Lval
 		code_exprn__add_lval_reg_dependencies(Lval),
-		{ map__set(Vars0, Var, Stat0, Vars) },
+		{ map__det_update(Vars0, Var, Stat0, Vars) },
 		code_exprn__set_vars(Vars),
 		{ Code = empty }
 	;
@@ -969,7 +1053,7 @@ code_exprn__place_evaled(Rvals0, Var, Lval, Code) -->
 		{ set__to_sorted_list(Rvals0, RvalList) },
 		{ code_exprn__select_rval(RvalList, Rval0) },
 		{
-			Rval0 = lval(reg(_))
+			Rval0 = lval(reg(_, _))
 		;
 			Rval0 = lval(stackvar(_))
 		;
@@ -1071,8 +1155,7 @@ code_exprn__place_exprn(MaybeLval, MaybeVar, Rval0, StandAlone, IsConst,
 		% this produces in the form of the expression
 		code_exprn__clear_lval(Lval, Rval1, Rval2, ClearCode)
 	;
-		code_exprn__get_spare_reg(Reg),
-		{ Lval = reg(Reg) },
+		code_exprn__get_spare_reg(r, Lval),
 		{ Rval2 = Rval1 },
 		{ ClearCode = empty }
 	),
@@ -1109,15 +1192,15 @@ code_exprn__place_exprn(MaybeLval, MaybeVar, Rval0, StandAlone, IsConst,
 
 code_exprn__maybe_fix_clearcode(ClearCode, ExprnCode, Code) -->
 	(
-		{ ClearCode = node([assign(reg(Reg), lval(Lval)) - _]) },
+		{ ClearCode = node([assign(reg(RegT, RegN), lval(Lval)) - _]) },
 		{ tree__flatten(ExprnCode, ExprnListList) },
 		{ list__condense(ExprnListList, ExprnList) },
 		{ ExprnList = [assign(Target, Rval0) - Comment] },
-		code_exprn__clear_lval_of_synonyms(reg(Reg)),
-		\+ code_exprn__lval_in_use(reg(Reg))
+		code_exprn__clear_lval_of_synonyms(reg(RegT, RegN)),
+		\+ code_exprn__lval_in_use(reg(RegT, RegN))
 	->
-		{ exprn_aux__substitute_rval_in_rval(lval(reg(Reg)), lval(Lval),
-			Rval0, Rval) },
+		{ exprn_aux__substitute_rval_in_rval(lval(reg(RegT, RegN)),
+			lval(Lval), Rval0, Rval) },
 		{ Code = node([assign(Target, Rval) - Comment]) }
 	;
 		{ Code = tree(ClearCode, ExprnCode) }
@@ -1189,13 +1272,13 @@ code_exprn__construct_code(Lval, VarName, Rval0, Code) -->
 			]) }
 		;
 			( { Lval = field(_, _, _) } ->
-				code_exprn__acquire_reg(Reg),
-				code_exprn__construct_cell(reg(Reg),
+				code_exprn__acquire_reg(r, Reg),
+				code_exprn__construct_cell(Reg,
 					VarName, Tag, Arity, Rvals, Code0),
 				{ string__append(VarName, " placement",
 					Comment) },
 				{ Code1 = node([
-					assign(Lval, lval(reg(Reg))) - Comment
+					assign(Lval, lval(Reg)) - Comment
 				]) },
 				{ Code = tree(Code0, Code1) },
 				code_exprn__release_reg(Reg)
@@ -1260,7 +1343,7 @@ code_exprn__free_arg_dependenciess([Target | Targets]) -->
 %------------------------------------------------------------------------------%
 
 :- pred code_exprn__produce_vars(list(var), assoc_list(var, rval), code_tree,
-						exprn_info, exprn_info).
+	exprn_info, exprn_info).
 :- mode code_exprn__produce_vars(in, out, out, in, out) is det.
 
 code_exprn__produce_vars([], [], empty) --> [].
@@ -1292,8 +1375,7 @@ code_exprn__produce_var(Var, Rval, Code) -->
 		{ code_exprn__select_rval(RvalList, Rval) },
 		{ Code = empty }
 	;
-		code_exprn__get_spare_reg(Reg),
-		{ Lval = reg(Reg) },
+		code_exprn__select_preferred_reg(Var, Lval),
 		{ Rval = lval(Lval) },
 		code_exprn__place_var(Var, Lval, Code)
 	).
@@ -1301,38 +1383,110 @@ code_exprn__produce_var(Var, Rval, Code) -->
 %------------------------------------------------------------------------------%
 
 code_exprn__produce_var_in_reg(Var, Rval, Code) -->
-	code_exprn__produce_var(Var, Rval0, Code0),
+	code_exprn__get_var_status(Var, Stat),
 	(
-		{ Rval0 = lval(reg(_)) }
+		{ Stat = evaled(Rvals) },
+		{ set__to_sorted_list(Rvals, RvalList) },
+		{ code_exprn__select_reg_rval(RvalList, Rval0) }
 	->
-		{ Code = Code0 },
+		{ Code = empty },
 		{ Rval = Rval0 }
 	;
-		code_exprn__get_spare_reg(Reg),
-		{ Lval = reg(Reg) },
-		code_exprn__place_var(Var, Lval, Code1),
-		{ Rval = lval(Lval) },
-		{ Code = tree(Code0, Code1) }
+		code_exprn__select_preferred_reg(Var, Lval),
+		code_exprn__place_var(Var, Lval, Code),
+		{ Rval = lval(Lval) }
 	).
 
 code_exprn__produce_var_in_reg_or_stack(Var, Rval, Code) -->
-	code_exprn__produce_var(Var, Rval0, Code0),
+	code_exprn__get_var_status(Var, Stat),
 	(
-		{ Rval0 = lval(Loc) },
-		{ Loc = reg(_) ; Loc = stackvar(_) ; Loc = framevar(_) }
+		{ Stat = evaled(Rvals) },
+		{ set__to_sorted_list(Rvals, RvalList) },
+		{ code_exprn__select_reg_or_stack_rval(RvalList, Rval0) }
 	->
-		{ Code = Code0 },
+		{ Code = empty },
 		{ Rval = Rval0 }
 	;
-		code_exprn__get_spare_reg(Reg),
-		{ Lval = reg(Reg) },
-		code_exprn__place_var(Var, Lval, Code1),
-		{ Rval = lval(Lval) },
-		{ Code = tree(Code0, Code1) }
+		code_exprn__select_preferred_lval(Var, Lval),
+		code_exprn__place_var(Var, Lval, Code),
+		{ Rval = lval(Lval) }
 	).
 
 %------------------------------------------------------------------------------%
 
+:- pred code_exprn__select_reg_rval(list(rval), rval).
+:- mode code_exprn__select_reg_rval(in, out) is semidet.
+
+code_exprn__select_reg_rval([Rval0 | Rvals0], Rval) :-
+	( Rval0 = lval(reg(_, _)) ->
+		Rval = Rval0
+	;
+		code_exprn__select_reg_rval(Rvals0, Rval)
+	).
+
+:- pred code_exprn__select_reg_or_stack_rval(list(rval), rval).
+:- mode code_exprn__select_reg_or_stack_rval(in, out) is semidet.
+
+code_exprn__select_reg_or_stack_rval([Rval0 | Rvals0], Rval) :-
+	(
+		Rval0 = lval(Lval),
+		( Lval = reg(_, _) ; Lval = stackvar(_) ; Lval = framevar(_) )
+	->
+		Rval = Rval0
+	;
+		code_exprn__select_reg_or_stack_rval(Rvals0, Rval)
+	).
+
+%------------------------------------------------------------------------------%
+
+:- pred code_exprn__select_preferred_lval(var, lval, exprn_info, exprn_info).
+:- mode code_exprn__select_preferred_lval(in, out, in, out) is det.
+
+code_exprn__select_preferred_lval(Var, Lval) -->
+	code_exprn__get_follow_vars(FollowVars),
+	(
+		{ map__search(FollowVars, Var, PrefLval) }
+	->
+		(
+			\+ { unreal_lval(PrefLval) },
+			\+ code_exprn__lval_in_use(PrefLval)
+		->
+			{ Lval = PrefLval }
+		;
+			code_exprn__get_spare_reg(r, Lval)
+		)
+	;
+		code_exprn__get_spare_reg(r, Lval)
+	).
+
+:- pred code_exprn__select_preferred_reg(var, lval, exprn_info, exprn_info).
+:- mode code_exprn__select_preferred_reg(in, out, in, out) is det.
+
+code_exprn__select_preferred_reg(Var, Lval) -->
+	code_exprn__get_follow_vars(FollowVars),
+	(
+		{ map__search(FollowVars, Var, PrefLval) },
+		{ PrefLval = reg(_, _) }
+	->
+		(
+			\+ { unreal_lval(PrefLval) },
+			\+ code_exprn__lval_in_use(PrefLval)
+		->
+			{ Lval = PrefLval }
+		;
+			code_exprn__get_spare_reg(r, Lval)
+		)
+	;
+		code_exprn__get_spare_reg(r, Lval)
+	).
+
+:- pred unreal_lval(lval).
+:- mode unreal_lval(in) is semidet.
+
+unreal_lval(reg(_, 0)).
+
+%------------------------------------------------------------------------------%
+
 :- pred code_exprn__produce_args(list(rval), assoc_list(rval, rval),
 	code_tree, exprn_info, exprn_info).
 :- mode code_exprn__produce_args(in, out, out, in, out) is det.
@@ -1356,8 +1510,8 @@ code_exprn__produce_arg(Rval0, Rval, Code) -->
 
 code_exprn__release_arglocs([]) --> [].
 code_exprn__release_arglocs([_ - ArgLoc | ArgLocs]) -->
-	( { ArgLoc = lval(reg(Reg)) } ->
-		code_exprn__unlock_reg(Reg)
+	( { ArgLoc = lval(Lval), Lval = reg(_, _) } ->
+		code_exprn__unlock_reg(Lval)
 	;
 		{ error("non-register subargument location in code_exprn") }
 	),
@@ -1368,7 +1522,7 @@ code_exprn__release_arglocs([_ - ArgLoc | ArgLocs]) -->
 	% Move whatever is in r1 out of the way.
 
 code_exprn__clear_r1(Code) -->
-	code_exprn__clear_lval_return_shuffle(reg(r(1)), _, Code).
+	code_exprn__clear_lval_return_shuffle(reg(r, 1), _, Code).
 
 	% Move whatever is in Lval out of the way.
 	% It is possible that the value we want to put into Lval
@@ -1401,26 +1555,26 @@ code_exprn__clear_lval_return_shuffle(Lval, MaybeShuffle, Code) -->
 	(
 		code_exprn__lval_in_use(Lval)
 	->
-		code_exprn__get_spare_reg(Reg),
+		code_exprn__get_spare_reg(r, Reg),
 		code_exprn__get_vars(Vars0),
 		{ map__to_assoc_list(Vars0, VarsList0) },
-		code_exprn__relocate_lval(VarsList0, Lval, reg(Reg), VarsList),
+		code_exprn__relocate_lval(VarsList0, Lval, Reg, VarsList),
 		{ map__from_assoc_list(VarsList, Vars) },
 		code_exprn__set_vars(Vars),
-		{ MaybeShuffle = yes(reg(Reg)) },
-		{ Code = node([assign(reg(Reg), lval(Lval)) - "shuffle lval"]) }
+		{ MaybeShuffle = yes(Reg) },
+		{ Code = node([assign(Reg, lval(Lval)) - "shuffle lval"]) }
 	;
 		{ MaybeShuffle = no },
 		{ Code = empty }
 	).
 
 :- pred code_exprn__relocate_lval(assoc_list(var, var_stat), lval, lval,
-			assoc_list(var, var_stat), exprn_info, exprn_info).
+	assoc_list(var, var_stat), exprn_info, exprn_info).
 :- mode code_exprn__relocate_lval(in, in, in, out, in, out) is det.
 
 code_exprn__relocate_lval([], _OldVal, _NewVal, []) --> [].
-code_exprn__relocate_lval([V - Stat0 | Rest0], OldVal,
-				NewVal, [V - Stat | Rest]) -->
+code_exprn__relocate_lval([V - Stat0 | Rest0], OldVal, NewVal,
+		[V - Stat | Rest]) -->
 	(
 		{ Stat0 = cached(Exprn0) },
 		(
@@ -1445,7 +1599,7 @@ code_exprn__relocate_lval([V - Stat0 | Rest0], OldVal,
 	code_exprn__relocate_lval(Rest0, OldVal, NewVal, Rest).
 
 :- pred code_exprn__relocate_lval_2(list(rval), lval, lval, list(rval),
-							exprn_info, exprn_info).
+	exprn_info, exprn_info).
 :- mode code_exprn__relocate_lval_2(in, in, in, out, in, out) is det.
 
 code_exprn__relocate_lval_2([], _OldVal, _NewVal, []) --> [].
@@ -1515,49 +1669,61 @@ code_exprn__maybe_add_evaled(yes(Var), NewRval) -->
 
 	% Warning: if you get a reg, you must mark it as in use yourself.
 
-:- pred code_exprn__get_spare_reg(reg, exprn_info, exprn_info).
-:- mode code_exprn__get_spare_reg(out, in, out) is det.
+:- pred code_exprn__get_spare_reg(reg_type, lval, exprn_info, exprn_info).
+:- mode code_exprn__get_spare_reg(in, out, in, out) is det.
 
-code_exprn__get_spare_reg(Reg) -->
+code_exprn__get_spare_reg(RegType, Lval) -->
 	code_exprn__get_regs(Regs),
-	{ code_exprn__get_spare_reg_2(1, Regs, Reg) }.
+	{ code_exprn__get_spare_reg_2(RegType, 1, Regs, Lval) }.
 
-:- pred code_exprn__get_spare_reg_2(int, bag(reg), reg).
-:- mode code_exprn__get_spare_reg_2(in, in, out) is det.
+:- pred code_exprn__get_spare_reg_2(reg_type, int, bag(lval), lval).
+:- mode code_exprn__get_spare_reg_2(in, in, in, out) is det.
 
-code_exprn__get_spare_reg_2(N0, Regs, Reg) :-
-	Reg0 = r(N0),
-	(
-		bag__contains(Reg0, Regs)
-	->
+code_exprn__get_spare_reg_2(RegType, N0, Regs, Lval) :-
+	TrialLval = reg(RegType, N0),
+	( bag__contains(Regs, TrialLval) ->
 		N1 is N0 + 1,
-		code_exprn__get_spare_reg_2(N1, Regs, Reg)
+		code_exprn__get_spare_reg_2(RegType, N1, Regs, Lval)
 	;
-		Reg = Reg0
+		Lval = TrialLval
 	).
 
 %------------------------------------------------------------------------------%
 
-code_exprn__acquire_reg(Reg) -->
-	code_exprn__get_spare_reg(Reg),
+code_exprn__acquire_reg(Type, Lval) -->
+	code_exprn__get_spare_reg(Type, Lval),
 	code_exprn__get_regs(Regs0),
-	{ bag__insert(Regs0, Reg, Regs) },
+	{ bag__insert(Regs0, Lval, Regs) },
 	code_exprn__set_regs(Regs),
 	code_exprn__get_acquired(Acqu0),
-	{ set__insert(Acqu0, Reg, Acqu) },
+	{ set__insert(Acqu0, Lval, Acqu) },
 	code_exprn__set_acquired(Acqu).
 
-code_exprn__release_reg(Reg) -->
+code_exprn__acquire_reg_prefer_given(Type, Pref, Lval) -->
+	code_exprn__get_regs(Regs0),
+	{ PrefLval = reg(Type, Pref) },
+	( { bag__contains(Regs0, PrefLval) } ->
+		code_exprn__get_spare_reg(Type, Lval)
+	;
+		{ Lval = PrefLval }
+	),
+	{ bag__insert(Regs0, Lval, Regs) },
+	code_exprn__set_regs(Regs),
+	code_exprn__get_acquired(Acqu0),
+	{ set__insert(Acqu0, Lval, Acqu) },
+	code_exprn__set_acquired(Acqu).
+
+code_exprn__release_reg(Lval) -->
 	code_exprn__get_acquired(Acqu0),
 	(
-		{ set__member(Reg, Acqu0) }
+		{ set__member(Lval, Acqu0) }
 	->
-		{ set__delete(Acqu0, Reg, Acqu) },
+		{ set__delete(Acqu0, Lval, Acqu) },
 		code_exprn__set_acquired(Acqu),
 		code_exprn__get_regs(Regs0),
-		{ bag__remove(Regs0, Reg, Regs) },
+		{ bag__delete(Regs0, Lval, Regs) },
 		(
-			{ bag__contains(Reg, Regs) }
+			{ bag__contains(Regs, Lval) }
 		->
 			{ error("code_exprn__release_reg: reg still has references") }
 		;
@@ -1577,7 +1743,7 @@ code_exprn__lock_reg(Reg) -->
 
 code_exprn__unlock_reg(Reg) -->
 	code_exprn__get_regs(Regs0),
-	{ bag__remove(Regs0, Reg, Regs) },
+	{ bag__delete(Regs0, Reg, Regs) },
 	code_exprn__set_regs(Regs).
 
 %------------------------------------------------------------------------------%
@@ -1611,51 +1777,61 @@ code_exprn__get_var_name(Var, Name) -->
 :- pred code_exprn__set_vars(var_map, exprn_info, exprn_info).
 :- mode code_exprn__set_vars(in, in, out) is det.
 
-:- pred code_exprn__get_regs(bag(reg), exprn_info, exprn_info).
+:- pred code_exprn__get_regs(bag(lval), exprn_info, exprn_info).
 :- mode code_exprn__get_regs(out, in, out) is det.
 
-:- pred code_exprn__set_regs(bag(reg), exprn_info, exprn_info).
+:- pred code_exprn__set_regs(bag(lval), exprn_info, exprn_info).
 :- mode code_exprn__set_regs(in, in, out) is det.
 
-:- pred code_exprn__get_acquired(set(reg), exprn_info, exprn_info).
+:- pred code_exprn__get_acquired(set(lval), exprn_info, exprn_info).
 :- mode code_exprn__get_acquired(out, in, out) is det.
 
-:- pred code_exprn__set_acquired(set(reg), exprn_info, exprn_info).
+:- pred code_exprn__set_acquired(set(lval), exprn_info, exprn_info).
 :- mode code_exprn__set_acquired(in, in, out) is det.
 
 :- pred code_exprn__get_options(exprn_opts, exprn_info, exprn_info).
 :- mode code_exprn__get_options(out, in, out) is det.
 
 code_exprn__get_varset(Varset, ExprnInfo, ExprnInfo) :-
-	ExprnInfo = exprn_info(Varset, _Vars, _Regs, _Acqu, _Opt).
+	ExprnInfo = exprn_info(Varset, _Vars, _Regs, _Acqu, _SS, _FV, _Opt).
 
 code_exprn__set_varset(Varset, ExprnInfo0, ExprnInfo) :-
-	ExprnInfo0 = exprn_info(_Varset, Vars, Regs, Acqu, Opt),
-	ExprnInfo = exprn_info(Varset, Vars, Regs, Acqu, Opt).
+	ExprnInfo0 = exprn_info(_Varset, Vars, Regs, Acqu, SS, FV, Opt),
+	ExprnInfo = exprn_info(Varset, Vars, Regs, Acqu, SS, FV, Opt).
 
 code_exprn__get_vars(Vars, ExprnInfo, ExprnInfo) :-
-	ExprnInfo = exprn_info(_Varset, Vars, _Regs, _Acqu, _Opt).
+	ExprnInfo = exprn_info(_Varset, Vars, _Regs, _Acqu, _SS, _FV, _Opt).
 
 code_exprn__set_vars(Vars, ExprnInfo0, ExprnInfo) :-
-	ExprnInfo0 = exprn_info(Varset, _Vars, Regs, Acqu, Opt),
-	ExprnInfo = exprn_info(Varset, Vars, Regs, Acqu, Opt).
+	ExprnInfo0 = exprn_info(Varset, _Vars, Regs, Acqu, SS, FV, Opt),
+	ExprnInfo = exprn_info(Varset, Vars, Regs, Acqu, SS, FV, Opt).
 
 code_exprn__get_regs(Regs, ExprnInfo, ExprnInfo) :-
-	ExprnInfo = exprn_info(_Varset, _Vars, Regs, _Acqu, _Opt).
+	ExprnInfo = exprn_info(_Varset, _Vars, Regs, _Acqu, _SS, _FV, _Opt).
 
 code_exprn__set_regs(Regs, ExprnInfo0, ExprnInfo) :-
-	ExprnInfo0 = exprn_info(Varset, Vars, _Regs, Acqu, Opt),
-	ExprnInfo = exprn_info(Varset, Vars, Regs, Acqu, Opt).
+	ExprnInfo0 = exprn_info(Varset, Vars, _Regs, Acqu, SS, FV, Opt),
+	ExprnInfo = exprn_info(Varset, Vars, Regs, Acqu, SS, FV, Opt).
 
 code_exprn__get_acquired(Acqu, ExprnInfo, ExprnInfo) :-
-	ExprnInfo = exprn_info(_Varset, _Vars, _Regs, Acqu, _Opt).
+	ExprnInfo = exprn_info(_Varset, _Vars, _Regs, Acqu, _SS, _FV, _Opt).
 
 code_exprn__set_acquired(Acqu, ExprnInfo0, ExprnInfo) :-
-	ExprnInfo0 = exprn_info(Varset, Vars, Regs, _Acqu, Opt),
-	ExprnInfo = exprn_info(Varset, Vars, Regs, Acqu, Opt).
+	ExprnInfo0 = exprn_info(Varset, Vars, Regs, _Acqu, SS, FV, Opt),
+	ExprnInfo = exprn_info(Varset, Vars, Regs, Acqu, SS, FV, Opt).
+
+code_exprn__get_stack_slots(SS, ExprnInfo, ExprnInfo) :-
+	ExprnInfo = exprn_info(_Varset, _Vars, _Regs, _Acqu, SS, _FV, _Opt).
+
+code_exprn__get_follow_vars(FV, ExprnInfo, ExprnInfo) :-
+	ExprnInfo = exprn_info(_Varset, _Vars, _Regs, _Acqu, _SS, FV, _Opt).
+
+code_exprn__set_follow_vars(FV, ExprnInfo0, ExprnInfo) :-
+	ExprnInfo0 = exprn_info(Varset, Vars, Regs, Acqu, SS, _FV, Opt),
+	ExprnInfo = exprn_info(Varset, Vars, Regs, Acqu, SS, FV, Opt).
 
 code_exprn__get_options(Opt, ExprnInfo, ExprnInfo) :-
-	ExprnInfo = exprn_info(_Varset, _Vars, _Regs, _Acqu, Opt).
+	ExprnInfo = exprn_info(_Varset, _Vars, _Regs, _Acqu, _SS, _FV, Opt).
 
 %------------------------------------------------------------------------------%
 %------------------------------------------------------------------------------%
