@@ -1,5 +1,5 @@
 %---------------------------------------------------------------------------%
-% Copyright (C) 1995 University of Melbourne.
+% Copyright (C) 1994-1997 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %---------------------------------------------------------------------------%
@@ -54,10 +54,10 @@
 :- pred code_util__make_proc_label(module_info, pred_id, proc_id, proc_label).
 :- mode code_util__make_proc_label(in, in, in, out) is det.
 
-:- pred code_util__make_uni_label(module_info, type_id, int, proc_label).
+:- pred code_util__make_uni_label(module_info, type_id, proc_id, proc_label).
 :- mode code_util__make_uni_label(in, in, in, out) is det.
 
-:- pred code_util__arg_loc_to_register(arg_loc, reg).
+:- pred code_util__arg_loc_to_register(arg_loc, lval).
 :- mode code_util__arg_loc_to_register(in, out) is det.
 
 	% Determine whether a goal might allocate some heap space,
@@ -68,10 +68,10 @@
 	% goals which do allocate heap space, such as construction
 	% of boxed constants.
 
-:- pred code_util__goal_may_allocate_heap(hlds__goal).
+:- pred code_util__goal_may_allocate_heap(hlds_goal).
 :- mode code_util__goal_may_allocate_heap(in) is semidet.
 
-:- pred code_util__goal_list_may_allocate_heap(list(hlds__goal)).
+:- pred code_util__goal_list_may_allocate_heap(list(hlds_goal)).
 :- mode code_util__goal_list_may_allocate_heap(in) is semidet.
 
 	% Negate a condition.
@@ -86,15 +86,15 @@
 :- pred code_util__compiler_generated(pred_info).
 :- mode code_util__compiler_generated(in) is semidet.
 
-:- pred code_util__predinfo_is_builtin(module_info, pred_info).
-:- mode code_util__predinfo_is_builtin(in, in) is semidet.
+:- pred code_util__predinfo_is_builtin(pred_info).
+:- mode code_util__predinfo_is_builtin(in) is semidet.
 
-:- pred code_util__is_builtin(module_info, pred_id, proc_id, is_builtin).
-:- mode code_util__is_builtin(in, in, in, out) is det.
+:- pred code_util__builtin_state(module_info, pred_id, proc_id, builtin_state).
+:- mode code_util__builtin_state(in, in, in, out) is det.
 
 	% Given a module name, a predicate name, a proc_id and a list of
 	% variables as the arguments, find out if that procedure of that
-	% predicate is a builtin. If yes, the last two arguments
+	% predicate is an inline builtin. If yes, the last two arguments
 	% return two things:
 	%
 	% - an rval to execute as a test if the builtin is semidet; and
@@ -103,9 +103,12 @@
 	%
 	% At least one of these will be present.
 	%
-	% Each rval returned is guaranteed to be either a unop or a binop,
+	% Each test rval returned is guaranteed to be either a unop or a binop,
 	% applied to arguments that are either variables (from the argument
 	% list) or constants.
+	%
+	% Each to be assigned rval is guaranteed to be either in a form
+	% acceptable for a test rval, or in the form of a variable.
 
 :- pred code_util__translate_builtin(string, string, proc_id, list(var),
 	maybe(rval), maybe(pair(var, rval))).
@@ -124,7 +127,7 @@
 	% point we need for the resume point is the one with the resume
 	% variables in their original locations.
 
-:- pred code_util__cannot_stack_flush(hlds__goal).
+:- pred code_util__cannot_stack_flush(hlds_goal).
 :- mode code_util__cannot_stack_flush(in) is semidet.
 
 	% Succeed if the given goal cannot fail before encountering a context
@@ -132,14 +135,23 @@
 	% If such a goal needs a resume point, the only entry point we need
 	% is the stack entry point.
 
-:- pred code_util__cannot_fail_before_stack_flush(hlds__goal).
+:- pred code_util__cannot_fail_before_stack_flush(hlds_goal).
 :- mode code_util__cannot_fail_before_stack_flush(in) is semidet.
 
+	% code_util__count_recursive_calls(Goal, PredId, ProcId, Min, Max)
+	% Given that we are in predicate PredId and procedure ProcId,
+	% return the minimum and maximum number of recursive calls that
+	% an execution of Goal may encounter.
+
+:- pred code_util__count_recursive_calls(hlds_goal, pred_id, proc_id,
+	int, int).
+:- mode code_util__count_recursive_calls(in, in, in, out, out) is det.
+
 %---------------------------------------------------------------------------%
 
 :- implementation.
 :- import_module prog_data, type_util, special_pred.
-:- import_module bool, char, int, string, map, varset, require, std_util.
+:- import_module bool, char, int, string, map, term, varset, require, std_util.
 
 %---------------------------------------------------------------------------%
 
@@ -152,7 +164,7 @@ code_util__make_entry_label(ModuleInfo, PredId, ProcId, Immed, PredAddress) :-
 		;
 			pred_info_is_pseudo_imported(PredInfo),
 			% only the (in, in) mode of unification is imported
-			ProcId = 0
+			hlds_pred__in_in_unification_proc_id(ProcId)
 		)
 	->
 		code_util__make_proc_label(ModuleInfo, PredId, ProcId,
@@ -174,7 +186,7 @@ code_util__make_local_entry_label(ModuleInfo, PredId, ProcId, Immed, Label) :-
 		;
 			pred_info_is_pseudo_exported(PredInfo),
 			% only the (in, in) mode of a unification is exported
-			ProcId = 0
+			hlds_pred__in_in_unification_proc_id(ProcId)
 		)
 	->
 		(
@@ -230,241 +242,281 @@ code_util__make_internal_label(ModuleInfo, PredId, ProcId, LabelNum, Label) :-
 	Label = local(ProcLabel, LabelNum).
 
 code_util__make_proc_label(ModuleInfo, PredId, ProcId, ProcLabel) :-
-	predicate_module(ModuleInfo, PredId, ModuleName),
-	predicate_name(ModuleInfo, PredId, PredName),
 	module_info_pred_info(ModuleInfo, PredId, PredInfo),
+	pred_info_module(PredInfo, PredModule),
+	pred_info_name(PredInfo, PredName),
+	module_info_name(ModuleInfo, ThisModule),
 	(
 		code_util__compiler_generated(PredInfo)
 	->
 		pred_info_arg_types(PredInfo, _TypeVarSet, ArgTypes),
 		(
 			special_pred_get_type(PredName, ArgTypes, Type),
-			type_to_type_id(Type, TypeId0, _)
+			type_to_type_id(Type, TypeId, _),
+			% All type_ids here should be module qualified,
+			% since builtin types are handled separately in
+			% polymorphism.m.
+			TypeId = qualified(TypeModule, TypeName) - Arity
 		->
-			TypeId = TypeId0
+			(
+				ThisModule \= TypeModule,
+				PredName = "__Unify__",
+				\+ hlds_pred__in_in_unification_proc_id(ProcId)
+			->
+				DefiningModule = ThisModule
+			;
+				DefiningModule = TypeModule
+			),
+			ProcLabel = special_proc(DefiningModule, PredName,
+				TypeModule, TypeName, Arity, ProcId)
 		;
 			string__append_list(["code_util__make_proc_label:\n",
 				"cannot make label for special pred `",
 				PredName, "'"], ErrorMessage),
 			error(ErrorMessage)
-		),
-		TypeId = TypeName - Arity,
-		ProcLabel = special_proc(ModuleName, PredName, TypeName,
-				Arity, ProcId)
+		)
 	;
+		(
+			% Work out which module supplies the code for
+			% the predicate.
+			ThisModule \= PredModule,
+			\+ pred_info_is_imported(PredInfo)
+		->
+			% This predicate is a specialized version of 
+			% a pred from a `.opt' file.
+			DefiningModule = ThisModule
+		;	
+			DefiningModule = PredModule
+		),
 		pred_info_get_is_pred_or_func(PredInfo, PredOrFunc),
 		pred_info_arity(PredInfo, Arity),
-		ProcLabel = proc(ModuleName, PredOrFunc, PredName, Arity,
-				ProcId)
+		ProcLabel = proc(DefiningModule, PredOrFunc,
+			PredModule, PredName, Arity, ProcId)
 	).
 
 code_util__make_uni_label(ModuleInfo, TypeId, UniModeNum, ProcLabel) :-
 	module_info_name(ModuleInfo, ModuleName),
-	TypeId = TypeName - Arity,
-	ProcLabel = special_proc(ModuleName, "__Unify__", TypeName, Arity,
-				UniModeNum).
+	( TypeId = qualified(TypeModule, TypeName) - Arity ->
+		( hlds_pred__in_in_unification_proc_id(UniModeNum) ->
+			Module = TypeModule
+		;
+			Module = ModuleName
+		),
+		ProcLabel = special_proc(Module, "__Unify__", TypeModule,
+			TypeName, Arity, UniModeNum)
+	;
+		error("code_util__make_uni_label: unqualified type_id")
+	).
 
 %-----------------------------------------------------------------------------%
 
-code_util__arg_loc_to_register(ArgLoc, r(ArgLoc)).
+code_util__arg_loc_to_register(ArgLoc, reg(r, ArgLoc)).
 
 %-----------------------------------------------------------------------------%
 
-code_util__predinfo_is_builtin(_ModuleInfo, PredInfo) :-
+code_util__predinfo_is_builtin(PredInfo) :-
 	pred_info_module(PredInfo, ModuleName),
 	pred_info_name(PredInfo, PredName),
 %	code_util__translate_builtin(ModuleName, PredName, _, _, _, _).
 	pred_info_arity(PredInfo, Arity),
-	( code_util__builtin(ModuleName, PredName, 0, Arity)
-	; code_util__builtin(ModuleName, PredName, 10000, Arity)
+	( code_util__inline_builtin(ModuleName, PredName, 0, Arity)
+	; code_util__inline_builtin(ModuleName, PredName, 10000, Arity)
 	).
 
-code_util__is_builtin(ModuleInfo, PredId0, ProcId, IsBuiltin) :-
+code_util__builtin_state(ModuleInfo, PredId0, ProcId, BuiltinState) :-
 	predicate_module(ModuleInfo, PredId0, ModuleName),
 	predicate_name(ModuleInfo, PredId0, PredName),
 	predicate_arity(ModuleInfo, PredId0, Arity),
-	(
-		code_util__builtin(ModuleName, PredName, ProcId, Arity)
-	->
-		hlds__is_builtin_make_builtin(yes, yes, IsBuiltin)
+	proc_id_to_int(ProcId, ProcInt),
+	( code_util__inline_builtin(ModuleName, PredName, ProcInt, Arity) ->
+		BuiltinState = inline_builtin
 	;
-		hlds__is_builtin_make_builtin(no, no, IsBuiltin)
+		BuiltinState = not_builtin
 	).
 
-:- pred code_util__builtin(string, string, proc_id, int).
-:- mode code_util__builtin(in, in, in, in) is semidet.
+:- pred code_util__inline_builtin(string, string, int, int).
+:- mode code_util__inline_builtin(in, in, in, in) is semidet.
 
-code_util__builtin(ModuleName, PredName, ProcId, Arity) :-
+code_util__inline_builtin(ModuleName, PredName, ProcId, Arity) :-
 	Arity =< 3,
 	varset__init(VarSet),
 	varset__new_vars(VarSet, Arity, Args, _),
-	code_util__translate_builtin(ModuleName, PredName, ProcId, Args, _, _).
+	code_util__translate_builtin_2(ModuleName, PredName, ProcId, Args, _, _).
 
-code_util__translate_builtin("mercury_builtin", "builtin_int_gt", 0, [X, Y],
+code_util__translate_builtin(Module, PredName, ProcId, Args, BinOp, AsgOp) :-
+	proc_id_to_int(ProcId, ProcInt),
+	code_util__translate_builtin_2(Module, PredName, ProcInt, Args,
+		BinOp, AsgOp).
+
+:- pred code_util__translate_builtin_2(string, string, int, list(var),
+	maybe(rval), maybe(pair(var, rval))).
+:- mode code_util__translate_builtin_2(in, in, in, in, out, out) is semidet.
+
+code_util__translate_builtin_2("mercury_builtin", "builtin_int_gt", 0, [X, Y],
 	yes(binop((>), var(X), var(Y))), no).
-code_util__translate_builtin("mercury_builtin", "builtin_int_lt", 0, [X, Y],
+code_util__translate_builtin_2("mercury_builtin", "builtin_int_lt", 0, [X, Y],
 	yes(binop((<), var(X), var(Y))), no).
 
-code_util__translate_builtin("int", "builtin_plus", 10000, [X, Y, Z],
+code_util__translate_builtin_2("int", "builtin_plus", 10000, [X, Y, Z],
 	no, yes(Z - binop((+), var(X), var(Y)))).
-code_util__translate_builtin("int", "builtin_plus", 10001, [X, Y, Z],
+code_util__translate_builtin_2("int", "builtin_plus", 10001, [X, Y, Z],
 	no, yes(X - binop((-), var(Z), var(Y)))).
-code_util__translate_builtin("int", "builtin_plus", 10002, [X, Y, Z],
+code_util__translate_builtin_2("int", "builtin_plus", 10002, [X, Y, Z],
 	no, yes(Y - binop((-), var(Z), var(X)))).
-code_util__translate_builtin("int", "+", 10000, [X, Y, Z],
+code_util__translate_builtin_2("int", "+", 10000, [X, Y, Z],
 	no, yes(Z - binop((+), var(X), var(Y)))).
-code_util__translate_builtin("int", "+", 10001, [X, Y, Z],
+code_util__translate_builtin_2("int", "+", 10001, [X, Y, Z],
 	no, yes(X - binop((-), var(Z), var(Y)))).
-code_util__translate_builtin("int", "+", 10002, [X, Y, Z],
+code_util__translate_builtin_2("int", "+", 10002, [X, Y, Z],
 	no, yes(Y - binop((-), var(Z), var(X)))).
-code_util__translate_builtin("int", "builtin_minus", 10000, [X, Y, Z],
+code_util__translate_builtin_2("int", "builtin_minus", 10000, [X, Y, Z],
 	no, yes(Z - binop((-), var(X), var(Y)))).
-code_util__translate_builtin("int", "builtin_minus", 10001, [X, Y, Z],
+code_util__translate_builtin_2("int", "builtin_minus", 10001, [X, Y, Z],
 	no, yes(X - binop((+), var(Y), var(Z)))).
-code_util__translate_builtin("int", "builtin_minus", 10002, [X, Y, Z],
+code_util__translate_builtin_2("int", "builtin_minus", 10002, [X, Y, Z],
 	no, yes(Y - binop((-), var(X), var(Z)))).
-code_util__translate_builtin("int", "-", 10000, [X, Y, Z],
+code_util__translate_builtin_2("int", "-", 10000, [X, Y, Z],
 	no, yes(Z - binop((-), var(X), var(Y)))).
-code_util__translate_builtin("int", "-", 10001, [X, Y, Z],
+code_util__translate_builtin_2("int", "-", 10001, [X, Y, Z],
 	no, yes(X - binop((+), var(Y), var(Z)))).
-code_util__translate_builtin("int", "-", 10002, [X, Y, Z],
+code_util__translate_builtin_2("int", "-", 10002, [X, Y, Z],
 	no, yes(Y - binop((-), var(X), var(Z)))).
-code_util__translate_builtin("int", "builtin_times", 10000, [X, Y, Z],
+code_util__translate_builtin_2("int", "builtin_times", 10000, [X, Y, Z],
 	no, yes(Z - binop((*), var(X), var(Y)))).
-code_util__translate_builtin("int", "builtin_times", 10001, [X, Y, Z],
+code_util__translate_builtin_2("int", "builtin_times", 10001, [X, Y, Z],
 	no, yes(X - binop((/), var(Z), var(Y)))).
-code_util__translate_builtin("int", "builtin_times", 10002, [X, Y, Z],
+code_util__translate_builtin_2("int", "builtin_times", 10002, [X, Y, Z],
 	no, yes(Y - binop((/), var(Z), var(X)))).
-code_util__translate_builtin("int", "*", 10000, [X, Y, Z],
+code_util__translate_builtin_2("int", "*", 10000, [X, Y, Z],
 	no, yes(Z - binop((*), var(X), var(Y)))).
-code_util__translate_builtin("int", "*", 10001, [X, Y, Z],
+code_util__translate_builtin_2("int", "*", 10001, [X, Y, Z],
 	no, yes(X - binop((/), var(Z), var(Y)))).
-code_util__translate_builtin("int", "*", 10002, [X, Y, Z],
+code_util__translate_builtin_2("int", "*", 10002, [X, Y, Z],
 	no, yes(Y - binop((/), var(Z), var(X)))).
-code_util__translate_builtin("int", "builtin_div", 10000, [X, Y, Z],
+code_util__translate_builtin_2("int", "builtin_div", 10000, [X, Y, Z],
 	no, yes(Z - binop((/), var(X), var(Y)))).
-code_util__translate_builtin("int", "builtin_div", 10001, [X, Y, Z],
+code_util__translate_builtin_2("int", "builtin_div", 10001, [X, Y, Z],
 	no, yes(X - binop((*), var(Y), var(Z)))).
-code_util__translate_builtin("int", "builtin_div", 10002, [X, Y, Z],
+code_util__translate_builtin_2("int", "builtin_div", 10002, [X, Y, Z],
 	no, yes(Y - binop((/), var(X), var(Z)))).
-code_util__translate_builtin("int", "//", 10000, [X, Y, Z],
+code_util__translate_builtin_2("int", "//", 10000, [X, Y, Z],
 	no, yes(Z - binop((/), var(X), var(Y)))).
-code_util__translate_builtin("int", "//", 10001, [X, Y, Z],
+code_util__translate_builtin_2("int", "//", 10001, [X, Y, Z],
 	no, yes(X - binop((*), var(Y), var(Z)))).
-code_util__translate_builtin("int", "//", 10002, [X, Y, Z],
+code_util__translate_builtin_2("int", "//", 10002, [X, Y, Z],
 	no, yes(Y - binop((/), var(X), var(Z)))).
-code_util__translate_builtin("int", "builtin_mod", 10000, [X, Y, Z],
+code_util__translate_builtin_2("int", "builtin_mod", 10000, [X, Y, Z],
 	no, yes(Z - binop((mod), var(X), var(Y)))).
-code_util__translate_builtin("int", "mod", 10000, [X, Y, Z],
+code_util__translate_builtin_2("int", "rem", 10000, [X, Y, Z],
 	no, yes(Z - binop((mod), var(X), var(Y)))).
-code_util__translate_builtin("int", "builtin_left_shift", 10000, [X, Y, Z],
+code_util__translate_builtin_2("int", "builtin_left_shift", 10000, [X, Y, Z],
 	no, yes(Z - binop((<<), var(X), var(Y)))).
-code_util__translate_builtin("int", "<<", 10000, [X, Y, Z],
+code_util__translate_builtin_2("int", "<<", 10000, [X, Y, Z],
 	no, yes(Z - binop((<<), var(X), var(Y)))).
-code_util__translate_builtin("int", "builtin_right_shift", 10000, [X, Y, Z],
+code_util__translate_builtin_2("int", "builtin_right_shift", 10000, [X, Y, Z],
 	no, yes(Z - binop((>>), var(X), var(Y)))).
-code_util__translate_builtin("int", ">>", 10000, [X, Y, Z],
+code_util__translate_builtin_2("int", ">>", 10000, [X, Y, Z],
 	no, yes(Z - binop((>>), var(X), var(Y)))).
-code_util__translate_builtin("int", "builtin_bit_and", 10000, [X, Y, Z],
+code_util__translate_builtin_2("int", "builtin_bit_and", 10000, [X, Y, Z],
 	no, yes(Z - binop((&), var(X), var(Y)))).
-code_util__translate_builtin("int", "/\\", 10000, [X, Y, Z],
+code_util__translate_builtin_2("int", "/\\", 10000, [X, Y, Z],
 	no, yes(Z - binop((&), var(X), var(Y)))).
-code_util__translate_builtin("int", "builtin_bit_or", 10000, [X, Y, Z],
+code_util__translate_builtin_2("int", "builtin_bit_or", 10000, [X, Y, Z],
 	no, yes(Z - binop(('|'), var(X), var(Y)))).
-code_util__translate_builtin("int", "\\/", 10000, [X, Y, Z],
+code_util__translate_builtin_2("int", "\\/", 10000, [X, Y, Z],
 	no, yes(Z - binop(('|'), var(X), var(Y)))).
-code_util__translate_builtin("int", "builtin_bit_xor", 10000, [X, Y, Z],
+code_util__translate_builtin_2("int", "builtin_bit_xor", 10000, [X, Y, Z],
 	no, yes(Z - binop((^), var(X), var(Y)))).
-code_util__translate_builtin("int", "^", 10000, [X, Y, Z],
+code_util__translate_builtin_2("int", "^", 10000, [X, Y, Z],
 	no, yes(Z - binop((^), var(X), var(Y)))).
-code_util__translate_builtin("int", "builtin_unary_plus", 10000, [X, Y],
+code_util__translate_builtin_2("int", "builtin_unary_plus", 10000, [X, Y],
 	no, yes(Y - var(X))).
-code_util__translate_builtin("int", "+", 10000, [X, Y],
+code_util__translate_builtin_2("int", "+", 10000, [X, Y],
 	no, yes(Y - var(X))).
-code_util__translate_builtin("int", "builtin_unary_minus", 10000, [X, Y],
+code_util__translate_builtin_2("int", "builtin_unary_minus", 10000, [X, Y],
 	no, yes(Y - binop((-), const(int_const(0)), var(X)))).
-code_util__translate_builtin("int", "-", 10000, [X, Y],
+code_util__translate_builtin_2("int", "-", 10000, [X, Y],
 	no, yes(Y - binop((-), const(int_const(0)), var(X)))).
-code_util__translate_builtin("int", "builtin_bit_neg", 10000, [X, Y],
+code_util__translate_builtin_2("int", "builtin_bit_neg", 10000, [X, Y],
 	no, yes(Y - unop(bitwise_complement, var(X)))).
-code_util__translate_builtin("int", "\\", 10000, [X, Y],
+code_util__translate_builtin_2("int", "\\", 10000, [X, Y],
 	no, yes(Y - unop(bitwise_complement, var(X)))).
-code_util__translate_builtin("int", ">", 0, [X, Y],
+code_util__translate_builtin_2("int", ">", 0, [X, Y],
 	yes(binop((>), var(X), var(Y))), no).
-code_util__translate_builtin("int", "<", 0, [X, Y],
+code_util__translate_builtin_2("int", "<", 0, [X, Y],
 	yes(binop((<), var(X), var(Y))), no).
-code_util__translate_builtin("int", ">=", 0, [X, Y],
+code_util__translate_builtin_2("int", ">=", 0, [X, Y],
 	yes(binop((>=), var(X), var(Y))), no).
-code_util__translate_builtin("int", "=<", 0, [X, Y],
+code_util__translate_builtin_2("int", "=<", 0, [X, Y],
 	yes(binop((<=), var(X), var(Y))), no).
 
-code_util__translate_builtin("float", "builtin_float_plus", 10000, [X, Y, Z],
+code_util__translate_builtin_2("float", "builtin_float_plus", 10000, [X, Y, Z],
 	no, yes(Z - binop(float_plus, var(X), var(Y)))).
-code_util__translate_builtin("float", "builtin_float_plus", 10001, [X, Y, Z],
+code_util__translate_builtin_2("float", "builtin_float_plus", 10001, [X, Y, Z],
 	no, yes(X - binop(float_minus, var(Z), var(Y)))).
-code_util__translate_builtin("float", "builtin_float_plus", 10002, [X, Y, Z],
+code_util__translate_builtin_2("float", "builtin_float_plus", 10002, [X, Y, Z],
 	no, yes(Y - binop(float_minus, var(Z), var(X)))).
-code_util__translate_builtin("float", "+", 10000, [X, Y, Z],
+code_util__translate_builtin_2("float", "+", 10000, [X, Y, Z],
 	no, yes(Z - binop(float_plus, var(X), var(Y)))).
-code_util__translate_builtin("float", "+", 10001, [X, Y, Z],
+code_util__translate_builtin_2("float", "+", 10001, [X, Y, Z],
 	no, yes(X - binop(float_minus, var(Z), var(Y)))).
-code_util__translate_builtin("float", "+", 10002, [X, Y, Z],
+code_util__translate_builtin_2("float", "+", 10002, [X, Y, Z],
 	no, yes(Y - binop(float_minus, var(Z), var(X)))).
-code_util__translate_builtin("float", "builtin_float_minus", 10000, [X, Y, Z],
+code_util__translate_builtin_2("float", "builtin_float_minus", 10000, [X, Y, Z],
 	no, yes(Z - binop(float_minus, var(X), var(Y)))).
-code_util__translate_builtin("float", "builtin_float_minus", 10001, [X, Y, Z],
+code_util__translate_builtin_2("float", "builtin_float_minus", 10001, [X, Y, Z],
 	no, yes(X - binop(float_plus, var(Y), var(Z)))).
-code_util__translate_builtin("float", "builtin_float_minus", 10002, [X, Y, Z],
+code_util__translate_builtin_2("float", "builtin_float_minus", 10002, [X, Y, Z],
 	no, yes(Y - binop(float_minus, var(X), var(Z)))).
-code_util__translate_builtin("float", "-", 10000, [X, Y, Z],
+code_util__translate_builtin_2("float", "-", 10000, [X, Y, Z],
 	no, yes(Z - binop(float_minus, var(X), var(Y)))).
-code_util__translate_builtin("float", "-", 10001, [X, Y, Z],
+code_util__translate_builtin_2("float", "-", 10001, [X, Y, Z],
 	no, yes(X - binop(float_plus, var(Y), var(Z)))).
-code_util__translate_builtin("float", "-", 10002, [X, Y, Z],
+code_util__translate_builtin_2("float", "-", 10002, [X, Y, Z],
 	no, yes(Y - binop(float_minus, var(X), var(Z)))).
-code_util__translate_builtin("float", "builtin_float_times", 10000, [X, Y, Z],
+code_util__translate_builtin_2("float", "builtin_float_times", 10000, [X, Y, Z],
 	no, yes(Z - binop(float_times, var(X), var(Y)))).
-code_util__translate_builtin("float", "builtin_float_times", 10001, [X, Y, Z],
+code_util__translate_builtin_2("float", "builtin_float_times", 10001, [X, Y, Z],
 	no, yes(X - binop(float_divide, var(Z), var(Y)))).
-code_util__translate_builtin("float", "builtin_float_times", 10002, [X, Y, Z],
+code_util__translate_builtin_2("float", "builtin_float_times", 10002, [X, Y, Z],
 	no, yes(Y - binop(float_divide, var(Z), var(X)))).
-code_util__translate_builtin("float", "*", 10000, [X, Y, Z],
+code_util__translate_builtin_2("float", "*", 10000, [X, Y, Z],
 	no, yes(Z - binop(float_times, var(X), var(Y)))).
-code_util__translate_builtin("float", "*", 10001, [X, Y, Z],
+code_util__translate_builtin_2("float", "*", 10001, [X, Y, Z],
 	no, yes(X - binop(float_divide, var(Z), var(Y)))).
-code_util__translate_builtin("float", "*", 10002, [X, Y, Z],
+code_util__translate_builtin_2("float", "*", 10002, [X, Y, Z],
 	no, yes(Y - binop(float_divide, var(Z), var(X)))).
-code_util__translate_builtin("float", "builtin_float_divide", 10000, [X, Y, Z],
+code_util__translate_builtin_2("float", "builtin_float_divide", 10000, [X, Y, Z],
 	no, yes(Z - binop(float_divide, var(X), var(Y)))).
-code_util__translate_builtin("float", "builtin_float_divide", 10001, [X, Y, Z],
+code_util__translate_builtin_2("float", "builtin_float_divide", 10001, [X, Y, Z],
 	no, yes(X - binop(float_times, var(Y), var(Z)))).
-code_util__translate_builtin("float", "builtin_float_divide", 10002, [X, Y, Z],
+code_util__translate_builtin_2("float", "builtin_float_divide", 10002, [X, Y, Z],
 	no, yes(Y - binop(float_divide, var(X), var(Z)))).
-code_util__translate_builtin("float", "/", 10000, [X, Y, Z],
+code_util__translate_builtin_2("float", "/", 10000, [X, Y, Z],
 	no, yes(Z - binop(float_divide, var(X), var(Y)))).
-code_util__translate_builtin("float", "/", 10001, [X, Y, Z],
+code_util__translate_builtin_2("float", "/", 10001, [X, Y, Z],
 	no, yes(X - binop(float_times, var(Y), var(Z)))).
-code_util__translate_builtin("float", "/", 10002, [X, Y, Z],
+code_util__translate_builtin_2("float", "/", 10002, [X, Y, Z],
 	no, yes(Y - binop(float_divide, var(X), var(Z)))).
-code_util__translate_builtin("float", "+", 10000, [X, Y],
+code_util__translate_builtin_2("float", "+", 10000, [X, Y],
 	no, yes(Y - var(X))).
-code_util__translate_builtin("float", "-", 10000, [X, Y],
+code_util__translate_builtin_2("float", "-", 10000, [X, Y],
 	no, yes(Y - binop(float_minus, const(float_const(0.0)), var(X)))).
-code_util__translate_builtin("float", "builtin_float_gt", 0, [X, Y],
+code_util__translate_builtin_2("float", "builtin_float_gt", 0, [X, Y],
 	yes(binop(float_gt, var(X), var(Y))), no).
-code_util__translate_builtin("float", ">", 0, [X, Y],
+code_util__translate_builtin_2("float", ">", 0, [X, Y],
 	yes(binop(float_gt, var(X), var(Y))), no).
-code_util__translate_builtin("float", "builtin_float_lt", 0, [X, Y],
+code_util__translate_builtin_2("float", "builtin_float_lt", 0, [X, Y],
 	yes(binop(float_lt, var(X), var(Y))), no).
-code_util__translate_builtin("float", "<", 0, [X, Y],
+code_util__translate_builtin_2("float", "<", 0, [X, Y],
 	yes(binop(float_lt, var(X), var(Y))), no).
-code_util__translate_builtin("float", "builtin_float_ge", 0, [X, Y],
+code_util__translate_builtin_2("float", "builtin_float_ge", 0, [X, Y],
 	yes(binop(float_ge, var(X), var(Y))), no).
-code_util__translate_builtin("float", ">=", 0, [X, Y],
+code_util__translate_builtin_2("float", ">=", 0, [X, Y],
 	yes(binop(float_ge, var(X), var(Y))), no).
-code_util__translate_builtin("float", "builtin_float_le", 0, [X, Y],
+code_util__translate_builtin_2("float", "builtin_float_le", 0, [X, Y],
 	yes(binop(float_le, var(X), var(Y))), no).
-code_util__translate_builtin("float", "=<", 0, [X, Y],
+code_util__translate_builtin_2("float", "=<", 0, [X, Y],
 	yes(binop(float_le, var(X), var(Y))), no).
 
 %-----------------------------------------------------------------------------%
@@ -475,12 +527,7 @@ code_util__translate_builtin("float", "=<", 0, [X, Y],
 code_util__compiler_generated(PredInfo) :-
 	pred_info_name(PredInfo, PredName),
 	pred_info_arity(PredInfo, PredArity),
-	( PredName = "__Unify__", PredArity = 2
-	; PredName = "__Compare__", PredArity = 3
-	; PredName = "__Index__", PredArity = 2
-	; PredName = "__Term_To_Type__", PredArity = 2
-	; PredName = "__Type_To_Term__", PredArity = 2
-	).
+	special_pred_name_arity(_, _, PredName, PredArity).
 
 %-----------------------------------------------------------------------------%
 
@@ -490,12 +537,12 @@ code_util__compiler_generated(PredInfo) :-
 code_util__goal_may_allocate_heap(Goal - _GoalInfo) :-
 	code_util__goal_may_allocate_heap_2(Goal).
 
-:- pred code_util__goal_may_allocate_heap_2(hlds__goal_expr).
+:- pred code_util__goal_may_allocate_heap_2(hlds_goal_expr).
 :- mode code_util__goal_may_allocate_heap_2(in) is semidet.
 
-code_util__goal_may_allocate_heap_2(higher_order_call(_, _, _, _, _)).
+code_util__goal_may_allocate_heap_2(higher_order_call(_, _, _, _, _, _)).
 code_util__goal_may_allocate_heap_2(call(_, _, _, Builtin, _, _)) :-
-	\+ hlds__is_builtin_is_inline(Builtin).
+	Builtin \= inline_builtin.
 code_util__goal_may_allocate_heap_2(unify(_, _, _, construct(_,_,Args,_), _)) :-
 	Args = [_|_].
 code_util__goal_may_allocate_heap_2(some(_Vars, Goal)) :-
@@ -597,52 +644,15 @@ code_util__cons_id_to_tag(code_addr_const(P,M), _, _, code_addr_constant(P,M)).
 code_util__cons_id_to_tag(pred_const(P,M), _, _, pred_closure_tag(P,M)).
 code_util__cons_id_to_tag(base_type_info_const(M,T,A), _, _,
 		base_type_info_constant(M,T,A)).
-code_util__cons_id_to_tag(cons(qualified(_, _), _), _, _, _) :-
-	% should have been transformed into a function call or pred_const.
-	error("code_util__cons_id_to_tag - qualified cons_id").
-code_util__cons_id_to_tag(cons(unqualified(Name), Arity),
-			Type, ModuleInfo, Tag) :-
+code_util__cons_id_to_tag(cons(Name, Arity), Type, ModuleInfo, Tag) :-
 	(
 			% handle the `character' type specially
 		Type = term__functor(term__atom("character"), [], _),
-	 	string__char_to_string(Char, Name)
+		Name = unqualified(ConsName),
+	 	string__char_to_string(Char, ConsName)
 	->
 		char__to_int(Char, CharCode),
 		Tag = int_constant(CharCode)
-	;
-			% handle higher-order types specially
-		type_is_higher_order(Type, PredOrFunc, PredArgTypes)
-	->
-		list__length(PredArgTypes, PredArity),
-		module_info_get_predicate_table(ModuleInfo, PredicateTable),
-		TotalArity is Arity + PredArity,
-		(
-			predicate_table_search_pf_name_arity(
-				PredicateTable, PredOrFunc, Name, TotalArity,
-				PredIds)
-		->
-			( PredIds = [PredId] ->
-				predicate_table_get_preds(PredicateTable,
-					Preds),
-				map__lookup(Preds, PredId, PredInfo),
-				pred_info_procedures(PredInfo, Procs),
-				map__keys(Procs, ProcIds),
-				( ProcIds = [ProcId] ->
-					Tag = pred_closure_tag(PredId, ProcId)
-				;
-					error("sorry, not implemented: taking address of predicate or function with multiple modes")
-				)
-			;
-					% cons_id ought to include the module
-					% prefix, so that we could use
-					% predicate_table__search_pf_m_n_a to 
-					% prevent this from happening
-				error("code_util__cons_id_to_tag: ambiguous pred or func")
-			)
-		;
-			% the type-checker should ensure that this never happens
-			error("code_util__cons_id_to_tag: invalid pred or func")
-		)
 	;
 			% Use the type to determine the type_id
 		( type_to_type_id(Type, TypeId0, _) ->
@@ -658,7 +668,7 @@ code_util__cons_id_to_tag(cons(unqualified(Name), Arity),
 		map__lookup(TypeTable, TypeId, TypeDefn),
 		hlds_data__get_type_defn_body(TypeDefn, TypeBody),
 		(
-			TypeBody = du_type(_, ConsTable0, _)
+			TypeBody = du_type(_, ConsTable0, _, _)
 		->
 			ConsTable = ConsTable0
 		;
@@ -668,7 +678,7 @@ code_util__cons_id_to_tag(cons(unqualified(Name), Arity),
 			)
 		),
 			% Finally look up the cons_id in the table
-		map__lookup(ConsTable, cons(unqualified(Name), Arity), Tag)
+		map__lookup(ConsTable, cons(Name, Arity), Tag)
 	).
 
 %-----------------------------------------------------------------------------%
@@ -676,20 +686,19 @@ code_util__cons_id_to_tag(cons(unqualified(Name), Arity),
 code_util__cannot_stack_flush(GoalExpr - _) :-
 	code_util__cannot_stack_flush_2(GoalExpr).
 
-:- pred code_util__cannot_stack_flush_2(hlds__goal_expr).
+:- pred code_util__cannot_stack_flush_2(hlds_goal_expr).
 :- mode code_util__cannot_stack_flush_2(in) is semidet.
 
 code_util__cannot_stack_flush_2(unify(_, _, _, Unify, _)) :-
 	Unify \= complicated_unify(_, _).
-code_util__cannot_stack_flush_2(call(_, _, _, IsBuiltin, _, _)) :-
-	hlds__is_builtin_is_internal(IsBuiltin),
-	hlds__is_builtin_is_inline(IsBuiltin).
+code_util__cannot_stack_flush_2(call(_, _, _, BuiltinState, _, _)) :-
+	BuiltinState = inline_builtin.
 code_util__cannot_stack_flush_2(conj(Goals)) :-
 	code_util__cannot_stack_flush_goals(Goals).
 code_util__cannot_stack_flush_2(switch(_, _, Cases, _)) :-
 	code_util__cannot_stack_flush_cases(Cases).
 
-:- pred code_util__cannot_stack_flush_goals(list(hlds__goal)).
+:- pred code_util__cannot_stack_flush_goals(list(hlds_goal)).
 :- mode code_util__cannot_stack_flush_goals(in) is semidet.
 
 code_util__cannot_stack_flush_goals([]).
@@ -716,21 +725,24 @@ code_util__cannot_fail_before_stack_flush(GoalExpr - GoalInfo) :-
 		code_util__cannot_fail_before_stack_flush_2(GoalExpr)
 	).
 
-:- pred code_util__cannot_fail_before_stack_flush_2(hlds__goal_expr).
+:- pred code_util__cannot_fail_before_stack_flush_2(hlds_goal_expr).
 :- mode code_util__cannot_fail_before_stack_flush_2(in) is semidet.
 
 code_util__cannot_fail_before_stack_flush_2(conj(Goals)) :-
 	code_util__cannot_fail_before_stack_flush_conj(Goals).
 
-:- pred code_util__cannot_fail_before_stack_flush_conj(list(hlds__goal)).
+:- pred code_util__cannot_fail_before_stack_flush_conj(list(hlds_goal)).
 :- mode code_util__cannot_fail_before_stack_flush_conj(in) is semidet.
 
 code_util__cannot_fail_before_stack_flush_conj([]).
 code_util__cannot_fail_before_stack_flush_conj([Goal | Goals]) :-
 	Goal = GoalExpr - GoalInfo,
 	(
-		( GoalExpr = call(_, _, _, _, _, _)
-		; GoalExpr = higher_order_call(_, _, _, _, _)
+		(
+			GoalExpr = call(_, _, _, BuiltinState, _, _),
+			BuiltinState \= inline_builtin
+		;
+			GoalExpr = higher_order_call(_, _, _, _, _, _)
 		)
 	->
 		true
@@ -744,3 +756,100 @@ code_util__cannot_fail_before_stack_flush_conj([Goal | Goals]) :-
 	).
 
 %-----------------------------------------------------------------------------%
+
+code_util__count_recursive_calls(Goal - _, PredId, ProcId, Min, Max) :-
+	code_util__count_recursive_calls_2(Goal, PredId, ProcId, Min, Max).
+
+:- pred code_util__count_recursive_calls_2(hlds_goal_expr, pred_id, proc_id,
+	int, int).
+:- mode code_util__count_recursive_calls_2(in, in, in, out, out) is det.
+
+code_util__count_recursive_calls_2(not(Goal), PredId, ProcId, Min, Max) :-
+	code_util__count_recursive_calls(Goal, PredId, ProcId, Min, Max).
+code_util__count_recursive_calls_2(some(_, Goal), PredId, ProcId, Min, Max) :-
+	code_util__count_recursive_calls(Goal, PredId, ProcId, Min, Max).
+code_util__count_recursive_calls_2(unify(_, _, _, _, _), _, _, 0, 0).
+code_util__count_recursive_calls_2(higher_order_call(_, _,_, _, _, _), _, _,
+		0, 0).
+code_util__count_recursive_calls_2(pragma_c_code(_,_,_,_, _, _, _, _), _, _,
+		0, 0).
+code_util__count_recursive_calls_2(call(CallPredId, CallProcId, _, _, _, _),
+		PredId, ProcId, Count, Count) :-
+	(
+		PredId = CallPredId,
+		ProcId = CallProcId
+	->
+		Count = 1
+	;
+		Count = 0
+	).
+code_util__count_recursive_calls_2(conj(Goals), PredId, ProcId, Min, Max) :-
+	code_util__count_recursive_calls_conj(Goals, PredId, ProcId, 0, 0,
+		Min, Max).
+code_util__count_recursive_calls_2(disj(Goals, _), PredId, ProcId, Min, Max) :-
+	code_util__count_recursive_calls_disj(Goals, PredId, ProcId, Min, Max).
+code_util__count_recursive_calls_2(switch(_, _, Cases, _), PredId, ProcId,
+		Min, Max) :-
+	code_util__count_recursive_calls_cases(Cases, PredId, ProcId, Min, Max).
+code_util__count_recursive_calls_2(if_then_else(_, Cond, Then, Else, _),
+		PredId, ProcId, Min, Max) :-
+	code_util__count_recursive_calls(Cond, PredId, ProcId, CMin, CMax),
+	code_util__count_recursive_calls(Then, PredId, ProcId, TMin, TMax),
+	code_util__count_recursive_calls(Else, PredId, ProcId, EMin, EMax),
+	CTMin is CMin + TMin,
+	CTMax is CMax + TMax,
+	int__min(CTMin, EMin, Min),
+	int__max(CTMax, EMax, Max).
+
+:- pred code_util__count_recursive_calls_conj(list(hlds_goal),
+	pred_id, proc_id, int, int, int, int).
+:- mode code_util__count_recursive_calls_conj(in, in, in, in, in, out, out)
+	is det.
+
+code_util__count_recursive_calls_conj([], _, _, Min, Max, Min, Max).
+code_util__count_recursive_calls_conj([Goal | Goals], PredId, ProcId,
+		Min0, Max0, Min, Max) :-
+	code_util__count_recursive_calls(Goal, PredId, ProcId, Min1, Max1),
+	Min2 is Min0 + Min1,
+	Max2 is Max0 + Max1,
+	code_util__count_recursive_calls_conj(Goals, PredId, ProcId,
+		Min2, Max2, Min, Max).
+
+:- pred code_util__count_recursive_calls_disj(list(hlds_goal),
+	pred_id, proc_id, int, int).
+:- mode code_util__count_recursive_calls_disj(in, in, in, out, out) is det.
+
+code_util__count_recursive_calls_disj([], _, _, 0, 0).
+code_util__count_recursive_calls_disj([Goal | Goals], PredId, ProcId,
+		Min, Max) :-
+	( Goals = [] ->
+		code_util__count_recursive_calls(Goal, PredId, ProcId,
+			Min, Max)
+	;
+		code_util__count_recursive_calls(Goal, PredId, ProcId,
+			Min0, Max0),
+		code_util__count_recursive_calls_disj(Goals, PredId, ProcId,
+			Min1, Max1),
+		int__min(Min0, Min1, Min),
+		int__max(Max0, Max1, Max)
+	).
+
+:- pred code_util__count_recursive_calls_cases(list(case),
+	pred_id, proc_id, int, int).
+:- mode code_util__count_recursive_calls_cases(in, in, in, out, out) is det.
+
+code_util__count_recursive_calls_cases([], _, _, _, _) :-
+	error("empty cases in code_util__count_recursive_calls_cases").
+code_util__count_recursive_calls_cases([case(_, Goal) | Cases], PredId, ProcId,
+		Min, Max) :-
+	( Cases = [] ->
+		code_util__count_recursive_calls(Goal, PredId, ProcId,
+			Min, Max)
+	;
+		code_util__count_recursive_calls(Goal, PredId, ProcId,
+			Min0, Max0),
+		code_util__count_recursive_calls_cases(Cases, PredId, ProcId,
+			Min1, Max1),
+		int__min(Min0, Min1, Min),
+		int__max(Max0, Max1, Max)
+	).
