@@ -1,11 +1,11 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1995 University of Melbourne.
+% Copyright (C) 1996-1997 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
 %
-% The job of this module is to delete dead procedures and base_gen_info
-% structures from the HLDS.
+% The job of this module is to delete dead predicates, procedures 
+% and base_gen_info structures from the HLDS.
 %
 % It also computes the usage counts that inlining.m uses for the
 % `--inline-single-use' option.
@@ -30,6 +30,14 @@
 	io__state, io__state).
 :- mode dead_proc_elim__eliminate(in, in, out, di, uo) is det.
 
+	% This is performed immediately after make_hlds.m to avoid doing
+	% semantic checking and optimization on predicates from `.opt' 
+	% files which are not used in the current module. This assumes that
+	% the clauses_info is still valid, so it cannot be run after mode
+	% analysis.
+:- pred dead_pred_elim(module_info, module_info).
+:- mode dead_pred_elim(in, out) is det.
+
 :- type entity		--->	proc(pred_id, proc_id)
 			;	base_gen_info(string, string, int).
 
@@ -40,7 +48,7 @@
 
 :- implementation.
 :- import_module hlds_pred, hlds_goal, hlds_data, prog_data, llds.
-:- import_module passes_aux, globals, options.
+:- import_module passes_aux, globals, options, code_util.
 :- import_module int, list, set, queue, map, bool, std_util, require.
 
 %-----------------------------------------------------------------------------%
@@ -159,11 +167,26 @@ dead_proc_elim__initialize_base_gen_infos([], Queue, Queue, Needed, Needed).
 dead_proc_elim__initialize_base_gen_infos([BaseGenInfo | BaseGenInfos],
 		Queue0, Queue, Needed0, Needed) :-
 	BaseGenInfo = base_gen_info(_TypeId, ModuleName, TypeName,
-		Arity, Status, _Elim, _Procs),
+		Arity, _Status, _Elim, _Procs),
 	(
-		( Status = exported
-		; Status = abstract_exported
-		)
+		% XXX: We'd like to do this, but there are problems.
+		% ( Status = exported
+		% ; Status = abstract_exported
+		% )
+		% We need to do more thorough analysis of the
+		% reachability of the special predicates, in general,
+		% because using arg/3 allows us to get at base_type_info
+		% via the base_type_layout. The base_type_infos of
+		% arguments of functors may have had their special preds
+		% eliminated, but they can still be called. In addition,
+		% it would be nice for pragma C code to have some
+		% support for using compiler generated data structures
+		% and preds, so that they aren't just eliminated.
+		%
+		% So presently, all base_type_infos will be treated
+		% as exported, and hence no special preds will be
+		% eliminated.
+		semidet_succeed
 	->
 		Entity = base_gen_info(ModuleName, TypeName, Arity),
 		queue__put(Queue0, Entity, Queue1),
@@ -282,7 +305,7 @@ dead_proc_elim__examine_proc(proc(PredId, ProcId), ModuleInfo, Queue0, Queue,
 		Needed = Needed0
 	).
 
-:- pred dead_proc_elim__examine_goals(list(hlds__goal), pred_proc_id,
+:- pred dead_proc_elim__examine_goals(list(hlds_goal), pred_proc_id,
 	entity_queue, entity_queue, needed_map, needed_map).
 :- mode dead_proc_elim__examine_goals(in, in, in, out, in, out) is det.
 
@@ -306,7 +329,7 @@ dead_proc_elim__examine_cases([case(_, Goal) | Cases], CurrProc, Queue0, Queue,
 	dead_proc_elim__examine_cases(Cases, CurrProc, Queue1, Queue,
 		Needed1, Needed).
 
-:- pred dead_proc_elim__examine_goal(hlds__goal, pred_proc_id,
+:- pred dead_proc_elim__examine_goal(hlds_goal, pred_proc_id,
 	entity_queue, entity_queue, needed_map, needed_map).
 :- mode dead_proc_elim__examine_goal(in, in, in, out, in, out) is det.
 
@@ -315,7 +338,7 @@ dead_proc_elim__examine_goal(GoalExpr - _, CurrProc, Queue0, Queue,
 	dead_proc_elim__examine_expr(GoalExpr, CurrProc, Queue0, Queue,
 		Needed0, Needed).
 
-:- pred dead_proc_elim__examine_expr(hlds__goal_expr, pred_proc_id,
+:- pred dead_proc_elim__examine_expr(hlds_goal_expr, pred_proc_id,
 	entity_queue, entity_queue, needed_map, needed_map).
 :- mode dead_proc_elim__examine_expr(in, in, in, out, in, out) is det.
 
@@ -347,7 +370,7 @@ dead_proc_elim__examine_expr(if_then_else(_, Cond, Then, Else, _),
 		Needed1, Needed2),
 	dead_proc_elim__examine_goal(Else, CurrProc, Queue2, Queue,
 		Needed2, Needed).
-dead_proc_elim__examine_expr(higher_order_call(_,_,_,_,_), _,
+dead_proc_elim__examine_expr(higher_order_call(_,_,_,_,_,_), _,
 		Queue, Queue, Needed, Needed).
 dead_proc_elim__examine_expr(call(PredId, ProcId, _,_,_,_),
 		CurrProc, Queue0, Queue, Needed0, Needed) :-
@@ -355,7 +378,8 @@ dead_proc_elim__examine_expr(call(PredId, ProcId, _,_,_,_),
 	( proc(PredId, ProcId) = CurrProc ->
 		% if it's reachable and recursive, then we can't
 		% eliminate or inline it
-		NewNotation = no
+		NewNotation = no,
+		map__set(Needed0, proc(PredId, ProcId), NewNotation, Needed)
 	; map__search(Needed0, proc(PredId, ProcId), OldNotation) ->
 		(
 			OldNotation = no,
@@ -364,12 +388,14 @@ dead_proc_elim__examine_expr(call(PredId, ProcId, _,_,_,_),
 			OldNotation = yes(Count0),
 			Count is Count0 + 1,
 			NewNotation = yes(Count)
-		)
+		),
+		map__det_update(Needed0, proc(PredId, ProcId), NewNotation,
+			Needed)
 	;
-		NewNotation = yes(1)
-	),
-	map__set(Needed0, proc(PredId, ProcId), NewNotation, Needed).
-dead_proc_elim__examine_expr(pragma_c_code(_, _, PredId, ProcId, _, _),
+		NewNotation = yes(1),
+		map__set(Needed0, proc(PredId, ProcId), NewNotation, Needed)
+	).
+dead_proc_elim__examine_expr(pragma_c_code(_, _, PredId, ProcId, _, _, _, _),
 		_CurrProc, Queue0, Queue, Needed0, Needed) :-
 	queue__put(Queue0, proc(PredId, ProcId), Queue),
 	map__set(Needed0, proc(PredId, ProcId), no, Needed).
@@ -397,52 +423,62 @@ dead_proc_elim__examine_expr(unify(_,_,_, Uni, _), _CurrProc, Queue0, Queue,
 
 %-----------------------------------------------------------------------------%
 
-dead_proc_elim__eliminate(ModuleInfo0, Needed, ModuleInfo, State0, State) :-
+
+		% information used during the elimination phase.
+
+:- type elim_info
+	--->	elimination_info(
+			needed_map,	% collected usage counts 
+			module_info,	% ye olde module_info
+			pred_table	% table of predicates in this module:
+					% preds and procs in this table
+					% may be eliminated
+		).
+			
+dead_proc_elim__eliminate(ModuleInfo0, Needed0, ModuleInfo, State0, State) :-
 	module_info_predids(ModuleInfo0, PredIds),
 	module_info_preds(ModuleInfo0, PredTable0),
-	dead_proc_elim__eliminate_preds(PredIds, Needed, ModuleInfo0,
-		PredTable0, PredTable, State0, State),
-	module_info_set_preds(ModuleInfo0, PredTable, ModuleInfo1),
-	module_info_base_gen_infos(ModuleInfo1, BaseGenInfos0),
+
+	ElimInfo0 = elimination_info(Needed0, ModuleInfo0, PredTable0),
+	list__foldl2(dead_proc_elim__eliminate_pred, PredIds, ElimInfo0, 
+		ElimInfo, State0, State),
+	ElimInfo = elimination_info(Needed, ModuleInfo1, PredTable),
+
+	module_info_set_preds(ModuleInfo1, PredTable, ModuleInfo2),
+	module_info_base_gen_infos(ModuleInfo2, BaseGenInfos0),
 	dead_proc_elim__eliminate_base_gen_infos(BaseGenInfos0, Needed,
 		BaseGenInfos),
-	module_info_set_base_gen_infos(ModuleInfo1, BaseGenInfos, ModuleInfo).
+	module_info_set_base_gen_infos(ModuleInfo2, BaseGenInfos, ModuleInfo).
 
-:- pred dead_proc_elim__eliminate_preds(list(pred_id), needed_map, module_info,
-	pred_table, pred_table, io__state, io__state).
-:- mode dead_proc_elim__eliminate_preds(in, in, in, in, out, di, uo) is det.
 
-dead_proc_elim__eliminate_preds([], _Needed, _, PredTable, PredTable) --> [].
-dead_proc_elim__eliminate_preds([PredId | PredIds], Needed, ModuleInfo,
-		PredTable0, PredTable) -->
-	dead_proc_elim__eliminate_pred(PredId, Needed, ModuleInfo,
-		PredTable0, PredTable1),
-	dead_proc_elim__eliminate_preds(PredIds, Needed, ModuleInfo,
-		PredTable1, PredTable).
+		% eliminate any unused procedures for this pred
 
-:- pred dead_proc_elim__eliminate_pred(pred_id, needed_map, module_info,
-	pred_table, pred_table, io__state, io__state).
-:- mode dead_proc_elim__eliminate_pred(in, in, in, in, out, di, uo) is det.
+:- pred dead_proc_elim__eliminate_pred(pred_id, elim_info, elim_info,
+	io__state, io__state).
+:- mode dead_proc_elim__eliminate_pred(in, in, out, di, uo) is det.
 
-dead_proc_elim__eliminate_pred(PredId, Needed, ModuleInfo,
-		PredTable0, PredTable, State0, State) :-
+dead_proc_elim__eliminate_pred(PredId, ElimInfo0, ElimInfo, State0, State) :-
+	ElimInfo0 = elimination_info(Needed, ModuleInfo, PredTable0),
 	map__lookup(PredTable0, PredId, PredInfo0),
 	pred_info_import_status(PredInfo0, Status),
 	(
 		% Find out if the predicate is defined in this module.
 		% If yes, find out also whether any of its procedures
 		% must be kept.
-		( Status = local, Keep = no
-		; Status = pseudo_exported, Keep = yes(0)
+		( Status = local,
+			Keep = no
+		; Status = pseudo_imported,
+			Keep = no
+		; Status = pseudo_exported,
+			hlds_pred__in_in_unification_proc_id(InitProcId),
+			Keep = yes(InitProcId)
 		)
 	->
-		pred_info_procids(PredInfo0, ProcIds0),
+		pred_info_procids(PredInfo0, ProcIds),
 		pred_info_procedures(PredInfo0, ProcTable0),
-		pred_info_name(PredInfo0, Name),
-		pred_info_arity(PredInfo0, Arity),
-		dead_proc_elim__eliminate_procs(PredId, ProcIds0, Needed, Keep,
-			Name, Arity, ModuleInfo, ProcTable0, ProcTable,
-			State0, State),
+		list__foldl2(dead_proc_elim__eliminate_proc(PredId, Keep, 
+			ElimInfo0),
+			ProcIds, ProcTable0, ProcTable, State0, State),
 		pred_info_set_procedures(PredInfo0, ProcTable, PredInfo),
 		map__det_update(PredTable0, PredId, PredInfo, PredTable)
 	;
@@ -480,18 +516,19 @@ dead_proc_elim__eliminate_pred(PredId, Needed, ModuleInfo,
 		% This predicate is not defined in this module.
 		State = State0,
 		PredTable = PredTable0
-	).
+	),
+	ElimInfo = elimination_info(Needed, ModuleInfo, PredTable).
 
-:- pred dead_proc_elim__eliminate_procs(pred_id, list(proc_id),
-	needed_map, maybe(proc_id), string, int, module_info,
-	proc_table, proc_table, io__state, io__state).
-:- mode dead_proc_elim__eliminate_procs(in, in, in, in, in, in, in, in, out,
-	di, uo) is det.
 
-dead_proc_elim__eliminate_procs(_, [], _, _, _, _, _, ProcTable, ProcTable)
-		--> [].
-dead_proc_elim__eliminate_procs(PredId, [ProcId | ProcIds], Needed, Keep, Name,
-		Arity, ModuleInfo, ProcTable0, ProcTable) -->
+		% eliminate a procedure, if unused
+
+:- pred dead_proc_elim__eliminate_proc(pred_id, maybe(proc_id), elim_info,
+	proc_id, proc_table, proc_table, io__state, io__state).
+:- mode dead_proc_elim__eliminate_proc(in, in, in, in, in, out, di, uo) is det.
+
+dead_proc_elim__eliminate_proc(PredId, Keep, ElimInfo, ProcId, 
+		ProcTable0, ProcTable) -->
+	{ ElimInfo = elimination_info(Needed, ModuleInfo, _PredTable) },
 	(
 		% Keep the procedure if it is in the needed map
 		% or if it is to be kept because it is exported.
@@ -499,7 +536,7 @@ dead_proc_elim__eliminate_procs(PredId, [ProcId | ProcIds], Needed, Keep, Name,
 		; { Keep = yes(ProcId) }
 		)
 	->
-		{ ProcTable1 = ProcTable0 }
+		{ ProcTable = ProcTable0 }
 	;
 		globals__io_lookup_bool_option(very_verbose, VeryVerbose),
 		( { VeryVerbose = yes } ->
@@ -509,10 +546,8 @@ dead_proc_elim__eliminate_procs(PredId, [ProcId | ProcIds], Needed, Keep, Name,
 		;
 			[]
 		),
-		{ map__delete(ProcTable0, ProcId, ProcTable1) }
-	),
-	dead_proc_elim__eliminate_procs(PredId, ProcIds, Needed, Keep, Name,
-		Arity, ModuleInfo, ProcTable1, ProcTable).
+		{ map__delete(ProcTable0, ProcId, ProcTable) }
+	).
 
 :- pred dead_proc_elim__eliminate_base_gen_infos(list(base_gen_info),
 	needed_map, list(base_gen_info)).
@@ -524,7 +559,7 @@ dead_proc_elim__eliminate_base_gen_infos([BaseGenInfo0 | BaseGenInfos0], Needed,
 	dead_proc_elim__eliminate_base_gen_infos(BaseGenInfos0, Needed,	
 		BaseGenInfos1),
 	BaseGenInfo0 = base_gen_info(TypeId, ModuleName, TypeName,
-		Arity, Status, _Elim, Procs),
+		Arity, Status, Elim0, Procs),
 	(
 		Entity = base_gen_info(ModuleName, TypeName, Arity),
 		map__search(Needed, Entity, _)
@@ -532,9 +567,173 @@ dead_proc_elim__eliminate_base_gen_infos([BaseGenInfo0 | BaseGenInfos0], Needed,
 		BaseGenInfos = [BaseGenInfo0 | BaseGenInfos1]
 	;
 		list__length(Procs, ProcsLength),
+
+			% Procs may have been eliminated elsewhere, if so
+			% we sum the eliminated procs together.
+		(
+			Elim0 = yes(NumProcs0)
+		->
+			NumProcs is ProcsLength + NumProcs0
+		;
+			NumProcs = ProcsLength
+		),
 		NeuteredBaseGenInfo = base_gen_info(TypeId, ModuleName, 
-			TypeName, Arity, Status, yes(ProcsLength), []),
+			TypeName, Arity, Status, yes(NumProcs), []),
 		BaseGenInfos = [NeuteredBaseGenInfo | BaseGenInfos1]
 	).
 
 %-----------------------------------------------------------------------------%
+%-----------------------------------------------------------------------------%
+
+:- type dead_pred_info
+	--->	dead_pred_info(
+			module_info,
+			queue(pred_id),	% preds to examine.
+			set(pred_id),	% preds examined.
+			set(pred_id),	% needed pred_ids.
+			set(sym_name)	% pred names needed.
+		).
+
+dead_pred_elim(ModuleInfo0, ModuleInfo) :-
+	module_info_predids(ModuleInfo0, PredIds),
+	queue__init(Queue0),
+	set__init(Preds0),
+	set__init(Names0),
+	DeadInfo0 = dead_pred_info(ModuleInfo0, Queue0, 
+		Preds0, Preds0, Names0),
+	list__foldl(dead_pred_elim_initialize, PredIds, 
+		DeadInfo0, DeadInfo1),
+	dead_pred_elim_analyze(DeadInfo1, DeadInfo),
+	DeadInfo = dead_pred_info(ModuleInfo1, _, _, NeededPreds, _),
+	set__list_to_set(PredIds, PredIdSet),
+	set__difference(PredIdSet, NeededPreds, DeadPreds),
+	set__to_sorted_list(DeadPreds, DeadPredList),
+	list__foldl(module_info_remove_predicate, DeadPredList,
+		ModuleInfo1, ModuleInfo).
+
+:- pred dead_pred_elim_initialize(pred_id::in, dead_pred_info::in,
+		dead_pred_info::out) is det.
+
+dead_pred_elim_initialize(PredId, DeadInfo0, DeadInfo) :-
+	DeadInfo0 = dead_pred_info(ModuleInfo, Q0, Ex, Needed, NeededNames0),
+	module_info_pred_info(ModuleInfo, PredId, PredInfo),
+	( 
+		pred_info_module(PredInfo, PredModule),
+		(
+			% Don't eliminate special preds since they won't
+			% be actually called from the HLDS until after 
+			% polymorphism.
+			code_util__compiler_generated(PredInfo)
+		;
+			% Don't eliminate preds from mercury_builtin.m since
+			% polymorphism.m needs unify/2 and friends.
+			PredModule = "mercury_builtin"
+		;
+			% Don't attempt to eliminate local preds here, since we
+			% want to do semantic checking on those even if they 
+			% aren't used.
+			\+ pred_info_is_imported(PredInfo), 
+			\+ pred_info_import_status(PredInfo, opt_imported)
+		)
+	->
+		pred_info_name(PredInfo, PredName),
+		set__insert(NeededNames0, qualified(PredModule, PredName), 
+			NeededNames),
+		queue__put(Q0, PredId, Q)
+	;
+		NeededNames = NeededNames0,
+		Q = Q0
+	),
+	DeadInfo = dead_pred_info(ModuleInfo, Q, Ex, Needed, NeededNames).
+
+:- pred dead_pred_elim_analyze(dead_pred_info::in, 
+		dead_pred_info::out) is det.
+
+dead_pred_elim_analyze(DeadInfo0, DeadInfo) :-
+	DeadInfo0 = dead_pred_info(ModuleInfo, Q0, Ex0, Needed0, NeededNames),
+	( queue__get(Q0, PredId, Q) ->
+		( set__member(PredId, Ex0) ->
+			DeadInfo2 = dead_pred_info(ModuleInfo, Q,
+				Ex0, Needed0, NeededNames)
+		;
+			set__insert(Needed0, PredId, Needed),
+			set__insert(Ex0, PredId, Ex),
+			DeadInfo1 = dead_pred_info(ModuleInfo, Q, Ex, 
+				Needed, NeededNames),
+			module_info_pred_info(ModuleInfo, PredId, PredInfo),
+			pred_info_clauses_info(PredInfo, ClausesInfo),
+			ClausesInfo = clauses_info(_,_,_,_, Clauses),
+			list__foldl(dead_pred_elim_process_clause, Clauses,
+				DeadInfo1, DeadInfo2)
+		),
+		dead_pred_elim_analyze(DeadInfo2, DeadInfo)
+	;
+		DeadInfo = DeadInfo0
+	).
+
+:- pred dead_pred_elim_process_clause(clause::in, dead_pred_info::in, 
+		dead_pred_info::out) is det.
+
+dead_pred_elim_process_clause(clause(_, Goal, _)) -->
+	pre_modecheck_examine_goal(Goal).
+
+:- pred pre_modecheck_examine_goal(hlds_goal::in, 
+		dead_pred_info::in, dead_pred_info::out) is det.
+
+pre_modecheck_examine_goal(conj(Goals) - _) -->
+	list__foldl(pre_modecheck_examine_goal, Goals).
+pre_modecheck_examine_goal(disj(Goals, _) - _) -->
+	list__foldl(pre_modecheck_examine_goal, Goals).
+pre_modecheck_examine_goal(if_then_else(_, If, Then, Else, _) - _) -->
+	list__foldl(pre_modecheck_examine_goal, [If, Then, Else]).
+pre_modecheck_examine_goal(switch(_, _, Cases, _) - _) -->
+	{ ExamineCase = lambda([Case::in, Info0::in, Info::out] is det, (
+		Case = case(_, Goal),
+		pre_modecheck_examine_goal(Goal, Info0, Info)
+	)) },
+	list__foldl(ExamineCase, Cases).
+pre_modecheck_examine_goal(higher_order_call(_,_,_,_,_,_) - _) --> [].
+pre_modecheck_examine_goal(not(Goal) - _) -->
+	pre_modecheck_examine_goal(Goal).
+pre_modecheck_examine_goal(some(_, Goal) - _) -->
+	pre_modecheck_examine_goal(Goal).
+pre_modecheck_examine_goal(call(_, _, _, _, _, PredName) - _) -->
+	dead_pred_info_add_pred_name(PredName).
+pre_modecheck_examine_goal(pragma_c_code(_, _, _, _, _, _, _, _) - _) --> [].
+pre_modecheck_examine_goal(unify(_, Rhs, _, _, _) - _) -->
+	pre_modecheck_examine_unify_rhs(Rhs).
+
+:- pred pre_modecheck_examine_unify_rhs(unify_rhs::in, 
+		dead_pred_info::in, dead_pred_info::out) is det.
+
+pre_modecheck_examine_unify_rhs(var(_)) --> [].
+pre_modecheck_examine_unify_rhs(functor(Functor, _)) -->
+	( { Functor = cons(Name, _) } ->
+		dead_pred_info_add_pred_name(Name)
+	;
+		[]
+	).
+pre_modecheck_examine_unify_rhs(lambda_goal(_, _, _, _, Goal)) -->
+	pre_modecheck_examine_goal(Goal).
+
+:- pred dead_pred_info_add_pred_name(sym_name::in, dead_pred_info::in, 
+		dead_pred_info::out) is det.
+
+dead_pred_info_add_pred_name(Name, DeadInfo0, DeadInfo) :-
+	DeadInfo0 = dead_pred_info(ModuleInfo, Q0, Ex, Needed, NeededNames0),
+	( set__member(Name, NeededNames0) ->	
+		DeadInfo = DeadInfo0
+	;
+		module_info_get_predicate_table(ModuleInfo, PredicateTable),
+		set__insert(NeededNames0, Name, NeededNames),
+		( predicate_table_search_sym(PredicateTable, Name, PredIds) ->
+			queue__put_list(Q0, PredIds, Q)
+		;
+			Q = Q0
+		),
+		DeadInfo = dead_pred_info(ModuleInfo, Q, Ex,
+			Needed, NeededNames)
+	).
+
+%-----------------------------------------------------------------------------%
+%-----------------------------------------------------------------------------%
