@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1995 University of Melbourne.
+% Copyright (C) 1994-1997 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -52,7 +52,7 @@
 
 :- interface.
 
-:- import_module hlds_module, hlds_pred, hlds_data, io.
+:- import_module hlds_module, hlds_pred, hlds_data, det_report, io.
 
 	% Perform determinism inference for local predicates with no
 	% determinism declarations, and determinism checking for all other
@@ -67,14 +67,20 @@
 	io__state, io__state).
 :- mode determinism_check_proc(in, in, in, out, di, uo) is det.
 
+	% Infer the determinism of a procedure.
+
+:- pred det_infer_proc(pred_id, proc_id, module_info, module_info, globals,
+	determinism, determinism, list(det_msg)).
+:- mode det_infer_proc(in, in, in, out, in, out, out, out) is det.
+
 	% The tables for computing the determinism of compound goals
 	% from the determinism of their components.
 
-:- pred det_conjunction_maxsoln(soln_count, soln_count, soln_count).
-:- mode det_conjunction_maxsoln(in, in, out) is det.
+:- pred det_conjunction_detism(determinism, determinism, determinism).
+:- mode det_conjunction_detism(in, in, out) is det.
 
-:- pred det_conjunction_canfail(can_fail, can_fail, can_fail).
-:- mode det_conjunction_canfail(in, in, out) is det.
+:- pred det_par_conjunction_detism(determinism, determinism, determinism).
+:- mode det_par_conjunction_detism(in, in, out) is det.
 
 :- pred det_disjunction_maxsoln(soln_count, soln_count, soln_count).
 :- mode det_disjunction_maxsoln(in, in, out) is det.
@@ -96,9 +102,9 @@
 :- implementation.
 
 :- import_module hlds_goal, prog_data, det_report, det_util.
-:- import_module mode_util, globals, options, passes_aux.
+:- import_module type_util, mode_util, globals, options, passes_aux.
 :- import_module hlds_out, mercury_to_mercury, instmap.
-:- import_module bool, list, map, set, std_util, require.
+:- import_module bool, list, map, set, std_util, require, term.
 
 %-----------------------------------------------------------------------------%
 
@@ -206,12 +212,6 @@ global_final_pass(ModuleInfo0, ProcList, Debug, ModuleInfo) -->
 
 :- type soln_context	--->	all_solns ; first_soln.
 
-	% Infer the determinism of a procedure.
-
-:- pred det_infer_proc(pred_id, proc_id, module_info, module_info, globals,
-	determinism, determinism, list(det_msg)).
-:- mode det_infer_proc(in, in, in, out, in, out, out, out) is det.
-
 det_infer_proc(PredId, ProcId, ModuleInfo0, ModuleInfo, Globals,
 		Detism0, Detism, Msgs) :-
 
@@ -247,6 +247,10 @@ det_infer_proc(PredId, ProcId, ModuleInfo0, ModuleInfo, Globals,
 		% Take the worst of the old and new detisms.
 		% This is needed to prevent loops on p :- not(p)
 		% at least if the initial assumed detism is det.
+		% This may also be needed to ensure that we don't change
+		% the interface determinism of procedures, if we are
+		% re-running determinism analysis.
+
 	determinism_components(Detism0, CanFail0, MaxSoln0),
 	determinism_components(Detism1, CanFail1, MaxSoln1),
 	det_switch_canfail(CanFail0, CanFail1, CanFail),
@@ -258,9 +262,9 @@ det_infer_proc(PredId, ProcId, ModuleInfo0, ModuleInfo, Globals,
 	proc_info_set_inferred_determinism(Proc1, Detism, Proc),
 
 		%  Put back the new proc_info structure.
-	map__set(Procs0, ProcId, Proc, Procs),
+	map__det_update(Procs0, ProcId, Proc, Procs),
 	pred_info_set_procedures(Pred0, Procs, Pred),
-	map__set(Preds0, PredId, Pred, Preds),
+	map__det_update(Preds0, PredId, Pred, Preds),
 	module_info_set_preds(ModuleInfo0, Preds, ModuleInfo).
 
 %-----------------------------------------------------------------------------%
@@ -269,8 +273,8 @@ det_infer_proc(PredId, ProcId, ModuleInfo0, ModuleInfo, Globals,
 	% It annotates the goal and all its subgoals with their determinism
 	% and returns the annotated goal in `Goal'.
 
-:- pred det_infer_goal(hlds__goal, instmap, soln_context, det_info,
-	hlds__goal, determinism, list(det_msg)).
+:- pred det_infer_goal(hlds_goal, instmap, soln_context, det_info,
+	hlds_goal, determinism, list(det_msg)).
 :- mode det_infer_goal(in, in, in, in, out, out, out) is det.
 
 det_infer_goal(Goal0 - GoalInfo0, InstMap0, SolnContext0, DetInfo,
@@ -361,9 +365,9 @@ det_infer_goal(Goal0 - GoalInfo0, InstMap0, SolnContext0, DetInfo,
 
 %-----------------------------------------------------------------------------%
 
-:- pred det_infer_goal_2(hlds__goal_expr, hlds__goal_info, instmap,
+:- pred det_infer_goal_2(hlds_goal_expr, hlds_goal_info, instmap,
 	soln_context, det_info, set(var), instmap_delta,
-	hlds__goal_expr, determinism, list(det_msg)).
+	hlds_goal_expr, determinism, list(det_msg)).
 :- mode det_infer_goal_2(in, in, in, in, in, in, in, out, out, out) is det.
 
 	% The determinism of a conjunction is the worst case of the elements
@@ -384,14 +388,30 @@ det_infer_goal_2(disj(Goals0, SM), _, InstMap0, SolnContext, DetInfo, _, _,
 	% then it is semideterministic or worse - this is determined
 	% in switch_detection.m and handled via the SwitchCanFail field.
 
-det_infer_goal_2(switch(Var, SwitchCanFail, Cases0, SM), _,
+det_infer_goal_2(switch(Var, SwitchCanFail, Cases0, SM), GoalInfo,
 		InstMap0, SolnContext, DetInfo, _, _,
 		switch(Var, SwitchCanFail, Cases, SM), Detism, Msgs) :-
 	det_infer_switch(Cases0, InstMap0, SolnContext, DetInfo,
-		cannot_fail, at_most_zero, Cases, CasesDetism, Msgs),
+		cannot_fail, at_most_zero, Cases, CasesDetism, Msgs0),
 	determinism_components(CasesDetism, CasesCanFail, CasesSolns),
+	% The switch variable tests are in a first_soln context if and only
+	% if the switch goal as a whole was in a first_soln context and the
+	% cases cannot fail.
+	(
+		CasesCanFail = cannot_fail,
+		SolnContext = first_soln
+	->
+		SwitchSolnContext = first_soln
+	;
+		SwitchSolnContext = all_solns
+	),
+	ExaminesRep = yes,
+	det_check_for_noncanonical_type(Var, ExaminesRep, SwitchCanFail,
+		SwitchSolnContext, GoalInfo, switch, DetInfo, Msgs0,
+		SwitchSolns, Msgs),
 	det_conjunction_canfail(SwitchCanFail, CasesCanFail, CanFail),
-	determinism_components(Detism, CanFail, CasesSolns).
+	det_conjunction_maxsoln(SwitchSolns, CasesSolns, NumSolns),
+	determinism_components(Detism, CanFail, NumSolns).
 
 	% For calls, just look up the determinism entry associated with
 	% the called predicate.
@@ -417,10 +437,12 @@ det_infer_goal_2(call(PredId, ModeId, A, B, C, N), GoalInfo, _, SolnContext,
 		Msgs = []
 	).
 
-det_infer_goal_2(higher_order_call(PredVar, ArgVars, Types, Modes, Det),
+det_infer_goal_2(higher_order_call(PredVar, ArgVars, Types, Modes, Det,
+			IsPredOrFunc),
 		GoalInfo, _InstMap0, SolnContext,
 		_MiscInfo, _NonLocalVars, _DeltaInstMap,
-		higher_order_call(PredVar, ArgVars, Types, Modes, Det),
+		higher_order_call(PredVar, ArgVars, Types, Modes, Det,
+			IsPredOrFunc),
 		Det, Msgs) :-
 	determinism_components(Det, _, NumSolns),
 	(
@@ -434,7 +456,7 @@ det_infer_goal_2(higher_order_call(PredVar, ArgVars, Types, Modes, Det),
 
 	% unifications are either deterministic or semideterministic.
 	% (see det_infer_unify).
-det_infer_goal_2(unify(LT, RT0, M, U, C), GoalInfo, InstMap0, _SolnContext,
+det_infer_goal_2(unify(LT, RT0, M, U, C), GoalInfo, InstMap0, SolnContext,
 		DetInfo, _, _, unify(LT, RT, M, U, C), UnifyDet, Msgs) :-
 	(
 		RT0 = lambda_goal(PredOrFunc, Vars, Modes, LambdaDeclaredDet,
@@ -448,18 +470,26 @@ det_infer_goal_2(unify(LT, RT0, M, U, C), GoalInfo, InstMap0, _SolnContext,
 		;	
 			LambdaSolnContext = all_solns
 		),
-		det_infer_goal(Goal0, InstMap0, LambdaSolnContext, DetInfo,
+		det_info_get_module_info(DetInfo, ModuleInfo),
+		instmap__pre_lambda_update(ModuleInfo, Vars, Modes,
+			InstMap0, InstMap1),
+		det_infer_goal(Goal0, InstMap1, LambdaSolnContext, DetInfo,
 				Goal, LambdaInferredDet, Msgs1),
 		det_check_lambda(LambdaDeclaredDet, LambdaInferredDet,
 				Goal, GoalInfo, DetInfo, Msgs2),
-		list__append(Msgs1, Msgs2, Msgs),
+		list__append(Msgs1, Msgs2, Msgs3),
 		RT = lambda_goal(PredOrFunc, Vars, Modes, LambdaDeclaredDet,
 				Goal)
 	;
 		RT = RT0,
-		Msgs = []
+		Msgs3 = []
 	),
-	det_infer_unify(U, UnifyDet).
+	det_infer_unify_canfail(U, UnifyCanFail),
+	det_infer_unify_examines_rep(U, ExaminesRepresentation),
+	det_check_for_noncanonical_type(LT, ExaminesRepresentation,
+		UnifyCanFail, SolnContext, GoalInfo, unify(C), DetInfo, Msgs3,
+		UnifyNumSolns, Msgs),
+	determinism_components(UnifyDet, UnifyCanFail, UnifyNumSolns).
 
 det_infer_goal_2(if_then_else(Vars, Cond0, Then0, Else0, SM), _GoalInfo0,
 		InstMap0, SolnContext, DetInfo, _NonLocalVars, _DeltaInstMap,
@@ -501,8 +531,7 @@ det_infer_goal_2(if_then_else(Vars, Cond0, Then0, Else0, SM), _GoalInfo0,
 	% Finally combine the results from the three parts
 	( CondCanFail = cannot_fail ->
 		% A -> B ; C is equivalent to A, B if A cannot fail
-		det_conjunction_maxsoln(CondMaxSoln, ThenMaxSoln, MaxSoln),
-		det_conjunction_canfail(CondCanFail, ThenCanFail, CanFail)
+		det_conjunction_detism(CondDetism, ThenDetism, Detism)
 	; CondMaxSoln = at_most_zero ->
 		% A -> B ; C is equivalent to ~A, C if A cannot succeed
 		det_negation_det(CondDetism, MaybeNegDetism),
@@ -512,16 +541,14 @@ det_infer_goal_2(if_then_else(Vars, Cond0, Then0, Else0, SM), _GoalInfo0,
 		;
 			MaybeNegDetism = yes(NegDetism)
 		),
-		determinism_components(NegDetism, NegCanFail, NegMaxSoln),
-		det_conjunction_maxsoln(NegMaxSoln, ElseMaxSoln, MaxSoln),
-		det_conjunction_canfail(NegCanFail, ElseCanFail, CanFail)
+		det_conjunction_detism(NegDetism, ElseDetism, Detism)
 	;
 		det_conjunction_maxsoln(CondMaxSoln, ThenMaxSoln, CTMaxSoln),
 		det_switch_maxsoln(CTMaxSoln, ElseMaxSoln, MaxSoln),
-		det_switch_canfail(ThenCanFail, ElseCanFail, CanFail)
+		det_switch_canfail(ThenCanFail, ElseCanFail, CanFail),
+		determinism_components(Detism, CanFail, MaxSoln)
 	),
 
-	determinism_components(Detism, CanFail, MaxSoln),
 	list__append(ThenMsgs, ElseMsgs, AfterMsgs),
 	list__append(CondMsgs, AfterMsgs, Msgs).
 
@@ -556,14 +583,25 @@ det_infer_goal_2(some(Vars, Goal0), _, InstMap0, SolnContext, DetInfo, _, _,
 
 	% pragma c_codes are handled in the same way as predicate calls
 det_infer_goal_2(pragma_c_code(C_Code, IsRecursive, PredId, ProcId, Args,
-			ArgNameMap), 
+			ArgNameMap, OrigArgTypes, Extra), 
 		GoalInfo, _, SolnContext, DetInfo, _, _,
 		pragma_c_code(C_Code, IsRecursive, PredId, ProcId, Args,
-			ArgNameMap),
+			ArgNameMap, OrigArgTypes, Extra),
 		Detism, Msgs) :-
-	det_lookup_detism(DetInfo, PredId, ProcId, Detism),
-	determinism_components(Detism, _, NumSolns),
-	( NumSolns = at_most_many_cc, SolnContext \= first_soln ->
+	det_lookup_detism(DetInfo, PredId, ProcId, Detism0),
+	determinism_components(Detism0, CanFail, NumSolns0),
+	( Extra = extra_pragma_info(_, _) ->
+		% pragma C codes that specify saved variables and labels
+		% can have more than one solution
+		NumSolns = at_most_many
+	;
+		NumSolns = NumSolns0
+	),
+	determinism_components(Detism, CanFail, NumSolns),
+	(
+		NumSolns = at_most_many_cc,
+		SolnContext \= first_soln
+	->
 		Msgs = [cc_pred_in_wrong_context(GoalInfo, Detism,
 				PredId, ProcId)]
 	;
@@ -572,8 +610,8 @@ det_infer_goal_2(pragma_c_code(C_Code, IsRecursive, PredId, ProcId, Args,
 
 %-----------------------------------------------------------------------------%
 
-:- pred det_infer_conj(list(hlds__goal), instmap, soln_context, det_info,
-		list(hlds__goal), determinism, list(det_msg)).
+:- pred det_infer_conj(list(hlds_goal), instmap, soln_context, det_info,
+		list(hlds_goal), determinism, list(det_msg)).
 :- mode det_infer_conj(in, in, in, in, out, out, out) is det.
 
 det_infer_conj([], _InstMap0, _SolnContext, _DetInfo, [], det, []).
@@ -593,7 +631,7 @@ det_infer_conj([Goal0 | Goals0], InstMap0, SolnContext, DetInfo,
 	update_instmap(Goal0, InstMap0, InstMap1),
 	det_infer_conj(Goals0, InstMap1, SolnContext, DetInfo,
 			Goals, DetismB, MsgsB),
-	determinism_components(DetismB, CanFailB, MaxSolnsB),
+	determinism_components(DetismB, CanFailB, _MaxSolnsB),
 
 	%
 	% Next, work out whether the first conjunct is in a first_soln context
@@ -616,18 +654,15 @@ det_infer_conj([Goal0 | Goals0], InstMap0, SolnContext, DetInfo,
 	%
 	det_infer_goal(Goal0, InstMap0, SolnContextA, DetInfo,
 			Goal, DetismA, MsgsA),
-	determinism_components(DetismA, CanFailA, MaxSolnsA),
 
 	%
 	% Finally combine the results computed above.
 	%
-	det_conjunction_canfail(CanFailA, CanFailB, CanFail),
-	det_conjunction_maxsoln(MaxSolnsA, MaxSolnsB, MaxSolns),
-	determinism_components(Detism, CanFail, MaxSolns),
+	det_conjunction_detism(DetismA, DetismB, Detism),
 	list__append(MsgsA, MsgsB, Msgs).
 
-:- pred det_infer_disj(list(hlds__goal), instmap, soln_context, det_info,
-	can_fail, soln_count, list(hlds__goal), determinism, list(det_msg)).
+:- pred det_infer_disj(list(hlds_goal), instmap, soln_context, det_info,
+	can_fail, soln_count, list(hlds_goal), determinism, list(det_msg)).
 :- mode det_infer_disj(in, in, in, in, in, in, out, out, out) is det.
 
 det_infer_disj([], _InstMap0, _SolnContext, _DetInfo, CanFail, MaxSolns,
@@ -668,29 +703,127 @@ det_infer_switch([Case0 | Cases0], InstMap0, SolnContext, DetInfo, CanFail0,
 		MaxSolns2, Cases, Detism, Msgs2),
 	list__append(Msgs1, Msgs2, Msgs).
 
-	% Deconstruction unifications are deterministic if the type
+%-----------------------------------------------------------------------------%
+
+:- pred det_check_for_noncanonical_type(var, bool, can_fail, soln_context,
+		hlds_goal_info, cc_unify_context, det_info, list(det_msg),
+		soln_count, list(det_msg)).
+:- mode det_check_for_noncanonical_type(in, in, in, in,
+		in, in, in, in, out, out) is det.
+
+det_check_for_noncanonical_type(Var, ExaminesRepresentation, CanFail,
+		SolnContext, GoalInfo, GoalContext, DetInfo, Msgs0,
+		NumSolns, Msgs) :-
+	(
+		%
+		% check for unifications that attempt to examine
+		% the representation of a type that does not have
+		% a single representation for each abstract value
+		%
+		ExaminesRepresentation = yes,
+		det_get_proc_info(DetInfo, ProcInfo),
+		proc_info_vartypes(ProcInfo, VarTypes),
+		map__lookup(VarTypes, Var, Type),
+		det_type_has_user_defined_equality_pred(DetInfo, Type,
+			_TypeContext)
+	->
+		( CanFail = can_fail ->
+			proc_info_variables(ProcInfo, VarSet),
+			Msgs = [cc_unify_can_fail(GoalInfo, Var, Type,
+				VarSet, GoalContext) | Msgs0]
+		; SolnContext \= first_soln ->
+			proc_info_variables(ProcInfo, VarSet),
+			Msgs = [cc_unify_in_wrong_context(GoalInfo, Var,
+				Type, VarSet, GoalContext) | Msgs0]
+		;
+			Msgs = Msgs0
+		),
+		( SolnContext = first_soln ->
+			NumSolns = at_most_many_cc
+		;
+			NumSolns = at_most_many
+		)
+	;
+		NumSolns = at_most_one,
+		Msgs = Msgs0
+	).
+
+% return true iff there was a `where equality is <predname>' declaration
+% for the specified type.
+:- pred det_type_has_user_defined_equality_pred(det_info::in, (type)::in,
+		term__context::out) is semidet.
+det_type_has_user_defined_equality_pred(DetInfo, Type, TypeContext) :-
+	det_info_get_module_info(DetInfo, ModuleInfo),
+	module_info_types(ModuleInfo, TypeTable),
+	type_to_type_id(Type, TypeId, _TypeArgs),
+	map__search(TypeTable, TypeId, TypeDefn),
+	hlds_data__get_type_defn_body(TypeDefn, TypeBody),
+	TypeBody = du_type(_, _, _, yes(_)),
+	hlds_data__get_type_defn_context(TypeDefn, TypeContext).
+
+% return yes iff the results of the specified unification might depend on
+% the concrete representation of the abstract values involved.
+:- pred det_infer_unify_examines_rep(unification::in, bool::out) is det.
+det_infer_unify_examines_rep(assign(_, _), no).
+det_infer_unify_examines_rep(construct(_, _, _, _), no).
+det_infer_unify_examines_rep(deconstruct(_, _, _, _, _), yes).
+det_infer_unify_examines_rep(simple_test(_, _), yes).
+det_infer_unify_examines_rep(complicated_unify(_, _), no).
+	% Some complicated modes of complicated unifications _do_
+	% examine the representation...
+	% but we will catch those by reporting errors in the
+	% compiler-generated code for the complicated unification.
+
+
+	% Deconstruction unifications cannot fail if the type
 	% only has one constructor, or if the variable is known to be
 	% already bound to the appropriate functor.
 	% 
 	% This is handled (modulo bugs) by modes.m, which sets
-	% the determinism field in the deconstruct(...) to semidet for
+	% the appropriate field in the deconstruct(...) to can_fail for
 	% those deconstruction unifications which might fail.
-	% But switch_detection.m may set it back to det again, if it moves
-	% the functor test into a switch instead.
+	% But switch_detection.m may set it back to cannot_fail again,
+	% if it moves the functor test into a switch instead.
 
-:- pred det_infer_unify(unification, determinism).
-:- mode det_infer_unify(in, out) is det.
+:- pred det_infer_unify_canfail(unification, can_fail).
+:- mode det_infer_unify_canfail(in, out) is det.
 
-det_infer_unify(deconstruct(_, _, _, _, CanFail), Detism) :-
-	determinism_components(Detism, CanFail, at_most_one).
-det_infer_unify(assign(_, _), det).
-det_infer_unify(construct(_, _, _, _), det).
-det_infer_unify(simple_test(_, _), semidet).
-det_infer_unify(complicated_unify(_, CanFail), Detism) :-
-	determinism_components(Detism, CanFail, at_most_one).
+det_infer_unify_canfail(deconstruct(_, _, _, _, CanFail), CanFail).
+det_infer_unify_canfail(assign(_, _), cannot_fail).
+det_infer_unify_canfail(construct(_, _, _, _), cannot_fail).
+det_infer_unify_canfail(simple_test(_, _), can_fail).
+det_infer_unify_canfail(complicated_unify(_, CanFail), CanFail).
 
 %-----------------------------------------------------------------------------%
 
+% When figuring out the determinism of a conjunction,
+% if the second goal is unreachable, then then the
+% determinism of the conjunction is just the determinism
+% of the first goal.
+
+det_conjunction_detism(DetismA, DetismB, Detism) :-
+	determinism_components(DetismA, CanFailA, MaxSolnA),
+	( MaxSolnA = at_most_zero ->
+		Detism = DetismA
+	;
+		determinism_components(DetismB, CanFailB, MaxSolnB),
+		det_conjunction_canfail(CanFailA, CanFailB, CanFail),
+		det_conjunction_maxsoln(MaxSolnA, MaxSolnB, MaxSoln),
+		determinism_components(Detism, CanFail, MaxSoln)
+	).
+
+% Figuring out the determinism of a parallel conjunction is much
+% easier than for a sequential conjunction, since you simply
+% ignore the case where the second goal is unreachable.  Just do
+% a normal solution count.
+
+det_par_conjunction_detism(DetismA, DetismB, Detism) :-
+	determinism_components(DetismA, CanFailA, MaxSolnA),
+	determinism_components(DetismB, CanFailB, MaxSolnB),
+	det_conjunction_canfail(CanFailA, CanFailB, CanFail),
+	det_conjunction_maxsoln(MaxSolnA, MaxSolnB, MaxSoln),
+	determinism_components(Detism, CanFail, MaxSoln).
+
 % For the at_most_zero, at_most_one, at_most_many,
 % we're just doing abstract interpretation to count
 % the number of solutions.  Similarly, for the can_fail
@@ -701,6 +834,9 @@ det_infer_unify(complicated_unify(_, CanFail), Detism) :-
 % pruning, but that the goal occurs in a single-solution
 % context, so only the first solution will be returned.
 
+:- pred det_conjunction_maxsoln(soln_count, soln_count, soln_count).
+:- mode det_conjunction_maxsoln(in, in, out) is det.
+
 det_conjunction_maxsoln(at_most_zero,    at_most_zero,    at_most_zero).
 det_conjunction_maxsoln(at_most_zero,    at_most_one,     at_most_zero).
 det_conjunction_maxsoln(at_most_zero,    at_most_many_cc, at_most_zero).
@@ -724,6 +860,9 @@ det_conjunction_maxsoln(at_most_many,    at_most_one,     at_most_many).
 det_conjunction_maxsoln(at_most_many,    at_most_many_cc, at_most_many).
 det_conjunction_maxsoln(at_most_many,    at_most_many,    at_most_many).
 
+:- pred det_conjunction_canfail(can_fail, can_fail, can_fail).
+:- mode det_conjunction_canfail(in, in, out) is det.
+
 det_conjunction_canfail(can_fail,    can_fail,    can_fail).
 det_conjunction_canfail(can_fail,    cannot_fail, can_fail).
 det_conjunction_canfail(cannot_fail, can_fail,    can_fail).
