@@ -1,13 +1,11 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1995 University of Melbourne.
+% Copyright (C) 1995-1997 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
 
 % det_report.m - reporting of determinism errors and warnings.
 
-% XXX all warnings should be disable-able.  Many of the warnings here aren't.
-
 % author: zs.
 
 %-----------------------------------------------------------------------------%
@@ -16,32 +14,50 @@
 
 :- interface.
 
-:- import_module hlds_module, hlds_pred, hlds_goal, det_util.
+:- import_module hlds_module, hlds_pred, hlds_goal, det_util, prog_data.
 :- import_module io.
 
 :- type det_msg	--->
 			% warnings
-			multidet_disj(hlds__goal_info, list(hlds__goal))
-		;	det_disj(hlds__goal_info, list(hlds__goal))
-		;	semidet_disj(hlds__goal_info, list(hlds__goal))
-		;	zero_soln_disj(hlds__goal_info, list(hlds__goal))
-		;	zero_soln_disjunct(hlds__goal_info)
-		;	ite_cond_cannot_fail(hlds__goal_info)
-		;	ite_cond_cannot_succeed(hlds__goal_info)
-		;	negated_goal_cannot_fail(hlds__goal_info)
-		;	negated_goal_cannot_succeed(hlds__goal_info)
-		;	warn_obsolete(pred_id, hlds__goal_info)
+			multidet_disj(term__context, list(term__context))
+		;	det_disj(term__context, list(term__context))
+		;	semidet_disj(term__context, list(term__context))
+		;	zero_soln_disj(term__context, list(term__context))
+		;	zero_soln_disjunct(term__context)
+		;	ite_cond_cannot_fail(term__context)
+		;	ite_cond_cannot_succeed(term__context)
+		;	negated_goal_cannot_fail(term__context)
+		;	negated_goal_cannot_succeed(term__context)
+		;	warn_obsolete(pred_id, term__context)
 				% warning about calls to predicates
 				% for which there is a `:- pragma obsolete'
 				% declaration.
+		;	warn_infinite_recursion(term__context)
+				% warning about recursive calls
+				% which would cause infinite loops.
+		;	duplicate_call(seen_call_id, term__context,
+				term__context)
+				% multiple calls with the same input args.
 			% errors
-		;	cc_pred_in_wrong_context(hlds__goal_info, determinism,
+		;	cc_unify_can_fail(hlds_goal_info, var, type, varset,
+				cc_unify_context)
+		;	cc_unify_in_wrong_context(hlds_goal_info, var, type,
+				varset, cc_unify_context)
+		;	cc_pred_in_wrong_context(hlds_goal_info, determinism,
 				pred_id, proc_id)
-		;	higher_order_cc_pred_in_wrong_context(hlds__goal_info,
+		;	higher_order_cc_pred_in_wrong_context(hlds_goal_info,
 				determinism)
 		;	error_in_lambda(
 				determinism, determinism, % declared, inferred
-				hlds__goal, hlds__goal_info, pred_id, proc_id).
+				hlds_goal, hlds_goal_info, pred_id, proc_id).
+
+:- type seen_call_id
+	--->	seen_call(pred_id, proc_id)
+	;	higher_order_call.
+
+:- type cc_unify_context
+	--->	unify(unify_context)
+	;	switch.
 
 %-----------------------------------------------------------------------------%
 
@@ -55,7 +71,7 @@
 	% Check a lambda goal with the specified declared and inferred
 	% determinisms.
 
-:- pred det_check_lambda(determinism, determinism, hlds__goal, hlds__goal_info,
+:- pred det_check_lambda(determinism, determinism, hlds_goal, hlds_goal_info,
 			det_info, list(det_msg)).
 :- mode det_check_lambda(in, in, in, in, in, out) is det.
 
@@ -63,7 +79,7 @@
 	% and update the module info accordingly.
 
 :- pred det_report_and_handle_msgs(list(det_msg), module_info, module_info,
-	io__state, io__state).
+		io__state, io__state).
 :- mode det_report_and_handle_msgs(in, in, out, di, uo) is det.
 
 	% Print some determinism warning and/or error messages,
@@ -76,6 +92,13 @@
 
 %-----------------------------------------------------------------------------%
 
+:- type det_comparison	--->	tighter ; sameas ; looser.
+
+:- pred compare_determinisms(determinism, determinism, det_comparison).
+:- mode compare_determinisms(in, in, out) is det.
+
+%-----------------------------------------------------------------------------%
+
 :- implementation.
 
 :- import_module hlds_data, type_util, mode_util, inst_match.
@@ -88,23 +111,33 @@
 
 global_checking_pass([], ModuleInfo, ModuleInfo) --> [].
 global_checking_pass([proc(PredId, ProcId) | Rest], ModuleInfo0, ModuleInfo) -->
-	{
-		module_info_preds(ModuleInfo0, PredTable),
-		map__lookup(PredTable, PredId, PredInfo),
-		pred_info_procedures(PredInfo, ProcTable),
-		map__lookup(ProcTable, ProcId, ProcInfo),
-		proc_info_declared_determinism(ProcInfo, MaybeDetism),
-		proc_info_inferred_determinism(ProcInfo, InferredDetism)
-	},
+	{ module_info_pred_proc_info(ModuleInfo0, PredId, ProcId,
+		PredInfo, ProcInfo) },
+	check_determinism(PredId, ProcId, PredInfo, ProcInfo,
+		ModuleInfo0, ModuleInfo1),
+	check_if_main_can_fail(PredId, ProcId, PredInfo, ProcInfo,
+		ModuleInfo1, ModuleInfo2),
+	check_for_multisoln_func(PredId, ProcId, PredInfo, ProcInfo,
+		ModuleInfo2, ModuleInfo3),
+	global_checking_pass(Rest, ModuleInfo3, ModuleInfo).
+
+:- pred check_determinism(pred_id, proc_id, pred_info, proc_info,
+		module_info, module_info, io__state, io__state).
+:- mode check_determinism(in, in, in, in, in, out, di, uo) is det.
+
+check_determinism(PredId, ProcId, _PredInfo, ProcInfo,
+		ModuleInfo0, ModuleInfo) -->
+	{ proc_info_declared_determinism(ProcInfo, MaybeDetism) },
+	{ proc_info_inferred_determinism(ProcInfo, InferredDetism) },
 	(
 		{ MaybeDetism = no },
-		{ ModuleInfo1 = ModuleInfo0 }
+		{ ModuleInfo = ModuleInfo0 }
 	;
 		{ MaybeDetism = yes(DeclaredDetism) },
 		{ compare_determinisms(DeclaredDetism, InferredDetism, Cmp) },
 		(
 			{ Cmp = sameas },
-			{ ModuleInfo1 = ModuleInfo0 }
+			{ ModuleInfo = ModuleInfo0 }
 		;
 			{ Cmp = looser },
 			globals__io_lookup_bool_option(
@@ -118,24 +151,33 @@ global_checking_pass([proc(PredId, ProcId) | Rest], ModuleInfo0, ModuleInfo) -->
 			;
 				[]
 			),
-			{ ModuleInfo1 = ModuleInfo0 }
+			{ ModuleInfo = ModuleInfo0 }
 		;
 			{ Cmp = tighter },
-			{ module_info_incr_errors(ModuleInfo0, ModuleInfo1) },
+			{ module_info_incr_errors(ModuleInfo0, ModuleInfo) },
 			{ Message = "  error: determinism declaration not satisfied.\n" },
 			report_determinism_problem(PredId,
-				ProcId, ModuleInfo1, Message,
+				ProcId, ModuleInfo, Message,
 				DeclaredDetism, InferredDetism),
 			{ proc_info_goal(ProcInfo, Goal) },
 			globals__io_get_globals(Globals),
-			{ det_info_init(ModuleInfo1, PredId, ProcId, Globals,
+			{ det_info_init(ModuleInfo, PredId, ProcId, Globals,
 				DetInfo) },
 			det_diagnose_goal(Goal, DeclaredDetism, [], DetInfo, _)
 			% XXX with the right verbosity options, we want to
 			% call report_determinism_problem only if diagnose
 			% returns false, i.e. it didn't print a message.
 		)
-	),
+	).
+
+:- pred check_if_main_can_fail(pred_id, proc_id, pred_info, proc_info,
+		module_info, module_info, io__state, io__state).
+:- mode check_if_main_can_fail(in, in, in, in, in, out, di, uo) is det.
+
+check_if_main_can_fail(_PredId, _ProcId, PredInfo, ProcInfo,
+		ModuleInfo0, ModuleInfo) -->
+	{ proc_info_declared_determinism(ProcInfo, MaybeDetism) },
+	{ proc_info_inferred_determinism(ProcInfo, InferredDetism) },
 	% check that `main/2' cannot fail
 	( 
 		{ pred_info_name(PredInfo, "main") },
@@ -155,12 +197,72 @@ global_checking_pass([proc(PredId, ProcId) | Rest], ModuleInfo0, ModuleInfo) -->
 			% that would probably just confuse people.
 		io__write_string(
 			"Error: main/2 must be `det' or `cc_multi'.\n"),
-		{ module_info_incr_errors(ModuleInfo1,
-			ModuleInfo2) }
+		{ module_info_incr_errors(ModuleInfo0, ModuleInfo) }
 	;
-		{ ModuleInfo2 = ModuleInfo1 }
-	),
-	global_checking_pass(Rest, ModuleInfo2, ModuleInfo).
+		{ ModuleInfo = ModuleInfo0 }
+	).
+
+:- pred check_for_multisoln_func(pred_id, proc_id, pred_info, proc_info,
+		module_info, module_info, io__state, io__state).
+:- mode check_for_multisoln_func(in, in, in, in, in, out, di, uo) is det.
+
+check_for_multisoln_func(_PredId, _ProcId, PredInfo, ProcInfo,
+		ModuleInfo0, ModuleInfo) -->
+	{ proc_info_inferred_determinism(ProcInfo, InferredDetism) },
+
+	% Functions can only have more than one solution if it is a
+	% non-standard mode.  Otherwise, they would not be referentially
+	% transparent.  (Nondeterministic "functions" like C's `rand()'
+	% function are not allowed.)
+	(
+		% if it is a mode for a function...
+		{ pred_info_get_is_pred_or_func(PredInfo, function) },
+		% ... that can succeed more than once ...
+		{ determinism_components(InferredDetism, _CanFail, NumSolns) },
+		{ NumSolns \= at_most_zero },
+		{ NumSolns \= at_most_one },
+		% ... but for which all the arguments are input ...
+		{ proc_info_argmodes(ProcInfo, PredArgModes) },
+		{ pred_args_to_func_args(PredArgModes,
+			FuncArgModes, _FuncResultMode) },
+		{ \+ (
+			list__member(FuncArgMode, FuncArgModes),
+			\+ mode_is_fully_input(ModuleInfo0, FuncArgMode)
+		  )
+	 	} 
+	->
+		% ... then it is an error.
+		{ pred_info_name(PredInfo, PredName) },
+		{ proc_info_context(ProcInfo, FuncContext) },
+		prog_out__write_context(FuncContext),
+		io__write_string("Error: invalid determinism for function\n"),
+		prog_out__write_context(FuncContext),
+		io__write_string("  `"),
+		report_pred_name_mode(function, PredName, PredArgModes),
+		io__write_string("':\n"),
+		prog_out__write_context(FuncContext),
+		io__write_string(
+			"  the primary mode for a function cannot be `"),
+		mercury_output_det(InferredDetism),
+		io__write_string(
+			"'.\n"),
+		globals__io_lookup_bool_option(verbose_errors, VerboseErrors),
+		( { VerboseErrors = yes } ->
+			io__write_strings([
+"\tIn Mercury, a function is supposed to be a true mathematical function\n",
+"\tof its arguments; that is, the value of the function's result should\n",
+"\tbe determined only by the values of its arguments.\n",
+"\t(Allowing functions to have more than one result for the same\n",
+"\targuments would break referential transparency.)\n",
+"\tMost likely, this procedure should be a predicate, not a function.\n"
+			])
+		;
+			[]
+		),
+		{ module_info_incr_errors(ModuleInfo0, ModuleInfo) }
+	;
+		{ ModuleInfo = ModuleInfo0 }
+	).
 
 det_check_lambda(DeclaredDetism, InferredDetism, Goal, GoalInfo, DetInfo,
 		Msgs) :-
@@ -201,11 +303,6 @@ report_determinism_problem(PredId, ModeId, ModuleInfo, Message,
 
 %-----------------------------------------------------------------------------%
 
-:- type det_comparison	--->	tighter ; sameas ; looser.
-
-:- pred compare_determinisms(determinism, determinism, det_comparison).
-:- mode compare_determinisms(in, in, out) is det.
-
 compare_determinisms(DeclaredDetism, InferredDetism, CmpDetism) :-
 	determinism_components(DeclaredDetism, DeclaredCanFail, DeclaredSolns),
 	determinism_components(InferredDetism, InferredCanFail, InferredSolns),
@@ -261,7 +358,7 @@ compare_solncounts(at_most_many,    at_most_many,    sameas).
 	% The given goal should have determinism Desired, but doesn't.
 	% Find out what is wrong and print a report of the cause.
 
-:- pred det_diagnose_goal(hlds__goal, determinism, list(switch_context),
+:- pred det_diagnose_goal(hlds_goal, determinism, list(switch_context),
 	det_info, bool, io__state, io__state).
 :- mode det_diagnose_goal(in, in, in, in, out, di, uo) is det.
 
@@ -277,7 +374,7 @@ det_diagnose_goal(Goal - GoalInfo, Desired, SwitchContext, DetInfo,
 
 %-----------------------------------------------------------------------------%
 
-:- pred det_diagnose_goal_2(hlds__goal_expr, hlds__goal_info,
+:- pred det_diagnose_goal_2(hlds_goal_expr, hlds_goal_info,
 	determinism, determinism, list(switch_context), det_info, bool,
 	io__state, io__state).
 :- mode det_diagnose_goal_2(in, in, in, in, in, in, out, di, uo) is det.
@@ -326,19 +423,19 @@ det_diagnose_goal_2(switch(Var, SwitchCanFail, Cases, _), GoalInfo,
 			{ det_lookup_var_type(ModuleInfo, ProcInfo, Var,
 				TypeDefn) },
 			{ hlds_data__get_type_defn_body(TypeDefn, TypeBody) },
-			{ TypeBody = du_type(_, ConsTable, _) }
+			{ TypeBody = du_type(_, ConsTable, _, _) }
 		->
 			{ map__keys(ConsTable, ConsIds) },
 			{ det_diagnose_missing_consids(ConsIds, Cases,
 				Missing) },
 			io__write_string("  The switch on "),
-			mercury_output_var(Var, Varset),
+			mercury_output_var(Var, Varset, no),
 			io__write_string(" does not cover "),
 			det_output_consid_list(Missing, yes),
 			io__write_string(".\n")
 		;
 			io__write_string("  The switch on "),
-			mercury_output_var(Var, Varset),
+			mercury_output_var(Var, Varset, no),
 			io__write_string(" can fail.\n")
 		),
 		{ Diagnosed1 = yes }
@@ -352,110 +449,25 @@ det_diagnose_goal_2(switch(Var, SwitchCanFail, Cases, _), GoalInfo,
 det_diagnose_goal_2(call(PredId, ModeId, _, _, CallContext, _), GoalInfo,
 		Desired, Actual, _, DetInfo, yes) -->
 	{ goal_info_get_context(GoalInfo, Context) },
-	{ determinism_components(Desired, DesiredCanFail, DesiredSolns) },
-	{ determinism_components(Actual, ActualCanFail, ActualSolns) },
-	{ compare_canfails(DesiredCanFail, ActualCanFail, CmpCanFail) },
-	( { CmpCanFail = tighter } ->
+	det_diagnose_atomic_goal(Desired, Actual,
 		det_report_call_context(Context, CallContext, DetInfo,
 			PredId, ModeId),
-		io__write_string("can fail.\n"),
-		{ Diagnosed1 = yes }
-	;
-		{ Diagnosed1 = no }
-	),
-	{ compare_solncounts(DesiredSolns, ActualSolns, CmpSolns) },
-	( { CmpSolns = tighter } ->
-		det_report_call_context(Context, CallContext, DetInfo,
-			PredId, ModeId),
-		io__write_string("can succeed"),
-		( { DesiredSolns = at_most_one } ->
-			io__write_string(" more than once\n.")
-		;
-			io__write_string(".\n")
-		),
-		{ Diagnosed2 = yes }
-	;
-		{ Diagnosed2 = no }
-	),
-	{ bool__or(Diagnosed1, Diagnosed2, Diagnosed) },
-	(
-		{ Diagnosed = yes }
-	;
-		{ Diagnosed = no },
-		det_report_call_context(Context, CallContext, DetInfo,
-			PredId, ModeId),
-		io__write_string("has unknown determinism problem;\n"),
-		prog_out__write_context(Context),
-		io__write_string("  desired determinism is "),
-		hlds_out__write_determinism(Desired),
-		io__write_string(", while actual determinism is "),
-		hlds_out__write_determinism(Actual),
-		io__write_string(".\n")
-	).
+		Context).
 
-det_diagnose_goal_2(higher_order_call(_, _, _, _, _), GoalInfo,
-		Desired, Actual, _, _MiscInfo, yes) -->
+det_diagnose_goal_2(higher_order_call(_, _, _, _, _, _), GoalInfo,
+		Desired, Actual, _, _DetInfo, yes) -->
 	{ goal_info_get_context(GoalInfo, Context) },
-	{ determinism_components(Desired, DesiredCanFail, DesiredSolns) },
-	{ determinism_components(Actual, ActualCanFail, ActualSolns) },
-	{ compare_canfails(DesiredCanFail, ActualCanFail, CmpCanFail) },
-	( { CmpCanFail = tighter } ->
-		prog_out__write_context(Context),
-		io__write_string("  Higher-order predicate call can fail.\n"),
-		{ Diagnosed1 = yes }
-	;
-		{ Diagnosed1 = no }
-	),
-	{ compare_solncounts(DesiredSolns, ActualSolns, CmpSolns) },
-	( { CmpSolns = tighter } ->
-		prog_out__write_context(Context),
-		io__write_string("  Higher-order predicate call can succeed"),
-		( { DesiredSolns = at_most_one } ->
-			io__write_string(" more than once\n.")
-		;
-			io__write_string(".\n")
-		),
-		{ Diagnosed2 = yes }
-	;
-		{ Diagnosed2 = no }
-	),
-	{ bool__or(Diagnosed1, Diagnosed2, Diagnosed) },
-	(
-		{ Diagnosed = yes }
-	;
-		{ Diagnosed = no },
-		prog_out__write_context(Context),
-		io__write_string("  Higher-order predicate call has unknown determinism problem;\n"),
-		prog_out__write_context(Context),
-		io__write_string("  desired determinism is "),
-		hlds_out__write_determinism(Desired),
-		io__write_string(", while actual determinism is "),
-		hlds_out__write_determinism(Actual),
-		io__write_string(".\n")
-	).
+	prog_out__write_context(Context),
+	det_diagnose_atomic_goal(Desired, Actual,
+		report_higher_order_call_context(Context), Context).
 
 det_diagnose_goal_2(unify(LT, RT, _, _, UnifyContext), GoalInfo,
 		Desired, Actual, _, DetInfo, yes) -->
 	{ goal_info_get_context(GoalInfo, Context) },
-	{ determinism_components(Desired, DesiredCanFail, _DesiredSolns) },
-	{ determinism_components(Actual, ActualCanFail, _ActualSolns) },
 	{ First = yes, Last = yes },
-	det_report_unify_context(First, Last, Context, UnifyContext, DetInfo,
-				LT, RT),
-	(
-		{ DesiredCanFail = cannot_fail },
-		{ ActualCanFail = can_fail }
-	->
-		io__write_string(" can fail.\n")
-	;
-		io__write_string(" has unknown determinism problem;\n"),
-		prog_out__write_context(Context),
-		io__write_string("  desired determinism is "),
-		hlds_out__write_determinism(Desired),
-		io__write_string(", while actual determinism is "),
-		hlds_out__write_determinism(Actual),
-		io__write_string(".\n")
-	).
+	det_diagnose_atomic_goal(Desired, Actual,
+		det_report_unify_context(First, Last, Context, UnifyContext,
+			DetInfo, LT, RT), Context).
 
 det_diagnose_goal_2(if_then_else(_Vars, Cond, Then, Else, _), _GoalInfo,
 		Desired, _Actual, SwitchContext, DetInfo, Diagnosed) -->
@@ -516,7 +528,7 @@ det_diagnose_goal_2(some(_Vars, Goal), _, Desired, Actual,
 	det_diagnose_goal(Goal, InternalDesired, SwitchContext, DetInfo,
 		Diagnosed).
 
-det_diagnose_goal_2(pragma_c_code(_, _, _, _, _, _), GoalInfo, Desired, 
+det_diagnose_goal_2(pragma_c_code(_, _, _, _, _, _, _, _), GoalInfo, Desired, 
 		_, _, _, yes) -->
 	{ goal_info_get_context(GoalInfo, Context) },
 	prog_out__write_context(Context),
@@ -529,10 +541,64 @@ det_diagnose_goal_2(pragma_c_code(_, _, _, _, _, _), GoalInfo, Desired,
 	io__write_string("  pragma c_code declarations only allowed\n"),
 	prog_out__write_context(Context),
 	io__write_string("  for modes which don't succeed more than once.\n").
+	% XXX
 
 %-----------------------------------------------------------------------------%
 
-:- pred det_diagnose_conj(list(hlds__goal), determinism,
+:- pred report_higher_order_call_context(term__context::in,
+		io__state::di, io__state::uo) is det.
+report_higher_order_call_context(Context) -->
+	prog_out__write_context(Context),
+	io__write_string("  Higher-order call").
+
+%-----------------------------------------------------------------------------%
+
+:- pred det_diagnose_atomic_goal(determinism, determinism, 
+		pred(io__state, io__state), term__context,
+		io__state, io__state).
+:- mode det_diagnose_atomic_goal(in, in, pred(di, uo) is det, in,
+		di, uo) is det.
+
+det_diagnose_atomic_goal(Desired, Actual, WriteContext, Context) -->
+	{ determinism_components(Desired, DesiredCanFail, DesiredSolns) },
+	{ determinism_components(Actual, ActualCanFail, ActualSolns) },
+	{ compare_canfails(DesiredCanFail, ActualCanFail, CmpCanFail) },
+	( { CmpCanFail = tighter } ->
+		call(WriteContext),
+		io__write_string(" can fail.\n"),
+		{ Diagnosed1 = yes }
+	;
+		{ Diagnosed1 = no }
+	),
+	{ compare_solncounts(DesiredSolns, ActualSolns, CmpSolns) },
+	( { CmpSolns = tighter } ->
+		call(WriteContext),
+		io__write_string(" can succeed"),
+		( { DesiredSolns = at_most_one } ->
+			io__write_string(" more than once.\n")
+		;
+			io__write_string(".\n")
+		),
+		{ Diagnosed2 = yes }
+	;
+		{ Diagnosed2 = no }
+	),
+	{ bool__or(Diagnosed1, Diagnosed2, Diagnosed) },
+	(
+		{ Diagnosed = yes }
+	;
+		{ Diagnosed = no },
+		call(WriteContext),
+		io__write_string(" has unknown determinism problem;\n"),
+		prog_out__write_context(Context),
+		io__write_string("  desired determinism is "),
+		hlds_out__write_determinism(Desired),
+		io__write_string(", while actual determinism is "),
+		hlds_out__write_determinism(Actual),
+		io__write_string(".\n")
+	).
+
+:- pred det_diagnose_conj(list(hlds_goal), determinism,
 	list(switch_context), det_info, bool, io__state, io__state).
 :- mode det_diagnose_conj(in, in, in, in, out, di, uo) is det.
 
@@ -543,7 +609,7 @@ det_diagnose_conj([Goal | Goals], Desired, SwitchContext, DetInfo,
 	det_diagnose_conj(Goals, Desired, SwitchContext, DetInfo, Diagnosed2),
 	{ bool__or(Diagnosed1, Diagnosed2, Diagnosed) }.
 
-:- pred det_diagnose_disj(list(hlds__goal), determinism, determinism,
+:- pred det_diagnose_disj(list(hlds_goal), determinism, determinism,
 	list(switch_context), det_info, int, int, bool, io__state, io__state).
 :- mode det_diagnose_disj(in, in, in, in, in, in, out, out, di, uo) is det.
 
@@ -612,7 +678,7 @@ det_output_consid_list([ConsId | ConsIds], First) -->
 	( { First = yes } ->
 		[]
 	; { ConsIds = [] } ->
-		io__write_string(" and ")
+		io__write_string(" and/or ")
 	;
 		io__write_string(", ")
 	),
@@ -637,7 +703,7 @@ det_diagnose_write_switch_context(Context, [SwitchContext | SwitchContexts],
 	io__write_string("  Inside the case "),
 	hlds_out__write_cons_id(ConsId),
 	io__write_string(" of the switch on "),
-	mercury_output_var(Var, Varset),
+	mercury_output_var(Var, Varset, no),
 	io__write_string(":\n"),
 	det_diagnose_write_switch_context(Context, SwitchContexts, DetInfo).
 
@@ -665,15 +731,14 @@ det_report_call_context(Context, CallUnifyContext, DetInfo, PredId, ModeId) -->
 					call_unify_context(LT, RT, UC)) },
 			{ First = yes, Last = yes },
 			det_report_unify_context(First, Last,
-				Context, UC, DetInfo, LT, RT),
-			io__write_string(" ")
+				Context, UC, DetInfo, LT, RT)
 		;
 			% this shouldn't happen; every call to __Unify__
 			% should have a unify_context
 			{ CallUnifyContext = no },
 			prog_out__write_context(Context),
 			io__write_string(
-	"  Some wierd unification (or explicit call to `__Unify__'?) ")
+	"  Some weird unification (or explicit call to `__Unify__'?)")
 		)
 	;
 		(
@@ -688,11 +753,11 @@ det_report_call_context(Context, CallUnifyContext, DetInfo, PredId, ModeId) -->
 		),
 		{ pred_info_procedures(PredInfo, ProcTable) },
 		{ map__lookup(ProcTable, ModeId, ProcInfo) },
-		{ proc_info_argmodes(ProcInfo, ArgModes) },
+		{ proc_info_declared_argmodes(ProcInfo, ArgModes) },
 		prog_out__write_context(Context),
 		io__write_string("  call to `"),
 		report_pred_name_mode(PredOrFunc, PredName, ArgModes),
-		io__write_string("' ")
+		io__write_string("'")
 	).
 
 %-----------------------------------------------------------------------------%
@@ -736,16 +801,16 @@ det_report_unify_context(First0, Last, Context, UnifyContext, DetInfo, LT, RT)
 			\+ { varset__search_name(Varset, RV, _) }
 		->
 			io__write_string("with `"),
-			mercury_output_var(LT, Varset)
+			mercury_output_var(LT, Varset, no)
 		;
 			io__write_string("of `"),
-			mercury_output_var(LT, Varset),
+			mercury_output_var(LT, Varset, no),
 			io__write_string("' and `"),
-			hlds_out__write_unify_rhs(RT, ModuleInfo, Varset, 3)
+			hlds_out__write_unify_rhs(RT, ModuleInfo, Varset, no, 3)
 		)
 	;
 		io__write_string("with `"),
-		hlds_out__write_unify_rhs(RT, ModuleInfo, Varset, 3)
+		hlds_out__write_unify_rhs(RT, ModuleInfo, Varset, no, 3)
 	),
 	io__write_string("'").
 
@@ -769,6 +834,7 @@ det_report_and_handle_msgs(Msgs, ModuleInfo0, ModuleInfo) -->
 				HaltAtWarn = yes
 			}
 		->
+			io__set_exit_status(1),
 			{ module_info_incr_errors(ModuleInfo0, ModuleInfo) }
 		;
 			{ ModuleInfo = ModuleInfo0 }
@@ -776,96 +842,233 @@ det_report_and_handle_msgs(Msgs, ModuleInfo0, ModuleInfo) -->
 	).
 
 det_report_msgs(Msgs, ModuleInfo, WarnCnt, ErrCnt) -->
-	det_report_msgs_2(Msgs, ModuleInfo, 0, WarnCnt, 0, ErrCnt).
+	globals__io_lookup_bool_option(warn_simple_code, Warn),
+	det_report_msgs_2(Msgs, Warn, ModuleInfo, 0, WarnCnt, 0, ErrCnt).
 
-:- pred det_report_msgs_2(list(det_msg), module_info, int, int, int, int,
-	io__state, io__state).
-:- mode det_report_msgs_2(in, in, in, out, in, out, di, uo) is det.
+:- pred det_report_msgs_2(list(det_msg), bool,  module_info, int, int,
+	int, int, io__state, io__state).
+:- mode det_report_msgs_2(in, in, in, in, out, in, out, di, uo) is det.
 
-det_report_msgs_2([], _ModuleInfo, WarnCnt, WarnCnt, ErrCnt, ErrCnt) --> [].
-det_report_msgs_2([Msg | Msgs], ModuleInfo,
+det_report_msgs_2([], _, _ModuleInfo, WarnCnt, WarnCnt, ErrCnt, ErrCnt) --> [].
+det_report_msgs_2([Msg | Msgs], Warn, ModuleInfo,
 		WarnCnt0, WarnCnt, ErrCnt0, ErrCnt) -->
-	det_report_msg(Msg, ModuleInfo, Type),
-	(
-		{ Type = warning },
-		{ WarnCnt1 is WarnCnt0 + 1 },
+	{ det_msg_get_type(Msg, MsgType) },
+	( { Warn = no, MsgType = warning } ->
+		{ WarnCnt1 = WarnCnt0 },
 		{ ErrCnt1 = ErrCnt0 }
 	;
-		{ Type = error },
-		{ ErrCnt1 is ErrCnt0 + 1 },
-		{ WarnCnt1 = WarnCnt0 }
+		det_report_msg(Msg, ModuleInfo),
+		(
+			{ MsgType = warning },
+			{ WarnCnt1 is WarnCnt0 + 1 },
+			{ ErrCnt1 = ErrCnt0 }
+		;
+			{ MsgType = error },
+			{ ErrCnt1 is ErrCnt0 + 1 },
+			{ WarnCnt1 = WarnCnt0 }
+		)
 	),
-	det_report_msgs_2(Msgs, ModuleInfo,
+	det_report_msgs_2(Msgs, Warn, ModuleInfo,
 		WarnCnt1, WarnCnt, ErrCnt1, ErrCnt).
 
-:- pred det_report_msg(det_msg, module_info, det_msg_type,
-	io__state, io__state).
-:- mode det_report_msg(in, in, out, di, uo) is det.
+:- pred det_msg_get_type(det_msg, det_msg_type).
+:- mode det_msg_get_type(in, out) is det.
 
-det_report_msg(multidet_disj(GoalInfo, Disjuncts0), _, warning) -->
-	{ goal_info_get_context(GoalInfo, Context) },
+det_msg_get_type(multidet_disj(_, _), warning).
+det_msg_get_type(det_disj(_, _), warning).
+det_msg_get_type(semidet_disj(_, _), warning).
+det_msg_get_type(zero_soln_disj(_, _), warning).
+det_msg_get_type(zero_soln_disjunct(_), warning).
+det_msg_get_type(ite_cond_cannot_fail(_), warning).
+det_msg_get_type(ite_cond_cannot_succeed(_), warning).
+det_msg_get_type(negated_goal_cannot_fail(_), warning).
+det_msg_get_type(negated_goal_cannot_succeed(_), warning).
+det_msg_get_type(warn_obsolete(_, _), warning).
+det_msg_get_type(warn_infinite_recursion(_), warning).
+det_msg_get_type(duplicate_call(_, _, _), warning).
+det_msg_get_type(cc_unify_can_fail(_, _, _, _, _), error).
+det_msg_get_type(cc_unify_in_wrong_context(_, _, _, _, _), error).
+det_msg_get_type(cc_pred_in_wrong_context(_, _, _, _), error).
+det_msg_get_type(higher_order_cc_pred_in_wrong_context(_, _), error).
+det_msg_get_type(error_in_lambda(_, _, _, _, _, _), error).
+
+:- pred det_report_msg(det_msg, module_info, io__state, io__state).
+:- mode det_report_msg(in, in, di, uo) is det.
+
+det_report_msg(multidet_disj(Context, DisjunctContexts), _) -->
 	prog_out__write_context(Context),
 	io__write_string("Warning: the disjunction with arms on lines "),
-	{ det_report_sort_context_lines(Disjuncts0, Disjuncts) },
-	det_report_context_lines(Disjuncts, yes),
+	det_report_context_lines(DisjunctContexts, yes),
 	io__write_string("\n"),
 	prog_out__write_context(Context),
 	io__write_string("  has no outputs, but can succeed more than once.\n").
-det_report_msg(det_disj(GoalInfo, Disjuncts0), _, warning) -->
-	{ goal_info_get_context(GoalInfo, Context) },
+det_report_msg(det_disj(Context, DisjunctContexts), _) -->
 	prog_out__write_context(Context),
 	io__write_string("Warning: the disjunction with arms on lines "),
-	{ det_report_sort_context_lines(Disjuncts0, Disjuncts) },
-	det_report_context_lines(Disjuncts, yes),
+	det_report_context_lines(DisjunctContexts, yes),
 	io__write_string("\n"),
 	prog_out__write_context(Context),
 	io__write_string("  will succeed exactly once.\n").
-det_report_msg(semidet_disj(GoalInfo, Disjuncts0), _, warning) -->
-	{ goal_info_get_context(GoalInfo, Context) },
+det_report_msg(semidet_disj(Context, DisjunctContexts), _) -->
 	prog_out__write_context(Context),
 	io__write_string("Warning: the disjunction with arms on lines "),
-	{ det_report_sort_context_lines(Disjuncts0, Disjuncts) },
-	det_report_context_lines(Disjuncts, yes),
+	det_report_context_lines(DisjunctContexts, yes),
 	io__write_string("\n"),
 	prog_out__write_context(Context),
 	io__write_string("  is semidet, yet it has an output.\n").
-det_report_msg(zero_soln_disj(GoalInfo, Disjuncts0), _, warning) -->
-	{ goal_info_get_context(GoalInfo, Context) },
+det_report_msg(zero_soln_disj(Context, DisjunctContexts), _) -->
 	prog_out__write_context(Context),
 	io__write_string("Warning: the disjunction with arms on lines "),
-	{ det_report_sort_context_lines(Disjuncts0, Disjuncts) },
-	det_report_context_lines(Disjuncts, yes),
+	det_report_context_lines(DisjunctContexts, yes),
 	io__write_string("\n"),
 	prog_out__write_context(Context),
 	io__write_string("  cannot succeed.\n").
-det_report_msg(zero_soln_disjunct(GoalInfo), _, warning) -->
-	{ goal_info_get_context(GoalInfo, Context) },
+det_report_msg(zero_soln_disjunct(Context), _) -->
 	prog_out__write_context(Context),
 	io__write_string("Warning: this disjunct will never have any solutions.\n").
-det_report_msg(ite_cond_cannot_fail(GoalInfo), _, warning) -->
-	{ goal_info_get_context(GoalInfo, Context) },
+det_report_msg(ite_cond_cannot_fail(Context), _) -->
 	prog_out__write_context(Context),
 	io__write_string("Warning: the condition of this if-then-else cannot fail.\n").
-det_report_msg(ite_cond_cannot_succeed(GoalInfo), _, warning) -->
-	{ goal_info_get_context(GoalInfo, Context) },
+det_report_msg(ite_cond_cannot_succeed(Context), _) -->
 	prog_out__write_context(Context),
 	io__write_string("Warning: the condition of this if-then-else cannot succeed.\n").
-det_report_msg(negated_goal_cannot_fail(GoalInfo), _, warning) -->
-	{ goal_info_get_context(GoalInfo, Context) },
+det_report_msg(negated_goal_cannot_fail(Context), _) -->
 	prog_out__write_context(Context),
 	io__write_string("Warning: the negated goal cannot fail.\n").
-det_report_msg(negated_goal_cannot_succeed(GoalInfo), _, warning) -->
-	{ goal_info_get_context(GoalInfo, Context) },
+det_report_msg(negated_goal_cannot_succeed(Context), _) -->
 	prog_out__write_context(Context),
 	io__write_string("Warning: the negated goal cannot succeed.\n").
-det_report_msg(warn_obsolete(PredId, GoalInfo), ModuleInfo, warning) -->
-	{ goal_info_get_context(GoalInfo, Context) },
+det_report_msg(warn_obsolete(PredId, Context), ModuleInfo) -->
 	prog_out__write_context(Context),
 	io__write_string("Warning: call to obsolete "),
 	hlds_out__write_pred_id(ModuleInfo, PredId),
 	io__write_string(".\n").
+det_report_msg(warn_infinite_recursion(Context), _ModuleInfo) -->
+/*
+% it would be better if we supplied more information
+% than just the line number.
+	prog_out__write_context(Context),
+	io__write_string("In "),
+	hlds_out__write_pred_id(ModuleInfo, PredId),
+	io__write_string(":\n"),
+*/
+	prog_out__write_context(Context),
+	io__write_string(
+		"Warning: recursive call will lead to infinite recursion.\n"),
+	globals__io_lookup_bool_option(verbose_errors, VerboseErrors),
+	( { VerboseErrors = yes } ->
+		io__write_string(
+"\tIf this recursive call is executed, the procedure will call itself
+\twith exactly the same input arguments, leading to infinite recursion.\n")
+	;
+		[]
+	).
+det_report_msg(duplicate_call(SeenCall, PrevContext, Context), ModuleInfo) -->
+	prog_out__write_context(Context),
+	io__write_string("Warning: redundant "),
+	det_report_seen_call_id(SeenCall, ModuleInfo),
+	io__write_string(".\n"),
+	prog_out__write_context(PrevContext),
+	io__write_string("Here is the previous "),
+	det_report_seen_call_id(SeenCall, ModuleInfo),
+	io__write_string(".\n").
+det_report_msg(cc_unify_can_fail(GoalInfo, Var, Type, VarSet, GoalContext),
+		_ModuleInfo) -->
+	{ goal_info_get_context(GoalInfo, Context) },
+	{ First0 = yes },
+	( { GoalContext = switch },
+		prog_out__write_context(Context),
+		io__write_string("In switch on variable `"),
+		mercury_output_var(Var, VarSet, no),
+		io__write_string("':\n"),
+		{ First = no }
+	; { GoalContext = unify(UnifyContext) },
+		hlds_out__write_unify_context(First0, UnifyContext, Context,
+			First)
+	),
+	prog_out__write_context(Context),
+	( { First = yes } ->
+		io__write_string("Error: ")
+	;
+		io__write_string("  error: ")
+	),
+	io__write_string("unification for non-canonical type\n"),
+	prog_out__write_context(Context),
+	io__write_string("  `"),
+	( { type_to_type_id(Type, TypeId, _TypeArgs) } ->
+		hlds_out__write_type_id(TypeId)
+	;
+		{ error("det_report_message: type_to_type_id failed") }
+	),
+	io__write_string("'\n"),
+	prog_out__write_context(Context),
+	io__write_string("  is not guaranteed to succeed.\n"),
+	globals__io_lookup_bool_option(verbose_errors, VerboseErrors),
+	( { VerboseErrors = yes } ->
+		io__write_strings([
+"	Since the type has a user-defined equality predicate, I must\n",
+"	presume that there is more than one possible concrete\n",
+"	representation for each abstract value of this type.  The success\n",
+"	of this unification might depend on the choice of concrete\n",
+"	representation.  Figuring out whether there is a solution to\n",
+"	this unification would require backtracking over all possible\n",
+"	representations, but I'm not going to do that implicitly.\n",
+"	(If that's really what you want, you must do it explicitly.)\n"
+		])
+	;
+		[]
+	),
+	io__set_exit_status(1).
+det_report_msg(cc_unify_in_wrong_context(GoalInfo, Var, Type, VarSet,
+		GoalContext), _ModuleInfo) -->
+	{ goal_info_get_context(GoalInfo, Context) },
+	{ First0 = yes },
+	( { GoalContext = switch },
+		prog_out__write_context(Context),
+		io__write_string("In switch on variable `"),
+		mercury_output_var(Var, VarSet, no),
+		io__write_string("':\n"),
+		{ First = no }
+	; { GoalContext = unify(UnifyContext) },
+		hlds_out__write_unify_context(First0, UnifyContext, Context,
+			First)
+	),
+	prog_out__write_context(Context),
+	( { First = yes } ->
+		io__write_string("Error: ")
+	;
+		io__write_string("  error: ")
+	),
+	io__write_string("unification for non-canonical type\n"),
+	prog_out__write_context(Context),
+	io__write_string("  `"),
+	( { type_to_type_id(Type, TypeId, _TypeArgs) } ->
+		hlds_out__write_type_id(TypeId)
+	;
+		{ error("det_report_message: type_to_type_id failed") }
+	),
+	io__write_string("'\n"),
+	prog_out__write_context(Context),
+	io__write_string(
+		"  occurs in a context which requires all solutions.\n"),
+	globals__io_lookup_bool_option(verbose_errors, VerboseErrors),
+	( { VerboseErrors = yes } ->
+		io__write_strings([
+"	Since the type has a user-defined equality predicate, I must\n",
+"	presume that there is more than one possible concrete\n",
+"	representation for each abstract value of this type.  The results\n",
+"	of this unification might depend on the choice of concrete\n",
+"	representation.  Finding all possible solutions to this\n",
+"	unification would require backtracking over all possible\n",
+"	representations, but I'm not going to do that implicitly.\n",
+"	(If that's really what you want, you must do it explicitly.)\n"
+		])
+	;
+		[]
+	),
+	io__set_exit_status(1).
 det_report_msg(cc_pred_in_wrong_context(GoalInfo, Detism, PredId, _ModeId),
-		ModuleInfo, error) -->
+		ModuleInfo) -->
 	{ goal_info_get_context(GoalInfo, Context) },
 	prog_out__write_context(Context),
 	io__write_string("Error: call to "),
@@ -877,7 +1080,7 @@ det_report_msg(cc_pred_in_wrong_context(GoalInfo, Detism, PredId, _ModeId),
 	io__write_string("  occurs in a context which requires all solutions.\n"),
 	io__set_exit_status(1).
 det_report_msg(higher_order_cc_pred_in_wrong_context(GoalInfo, Detism),
-		_ModuleInfo, error) -->
+		_ModuleInfo) -->
 	{ goal_info_get_context(GoalInfo, Context) },
 	prog_out__write_context(Context),
 	io__write_string("Error: higher-order call to predicate with determinism `"),
@@ -887,7 +1090,7 @@ det_report_msg(higher_order_cc_pred_in_wrong_context(GoalInfo, Detism),
 	io__write_string("  occurs in a context which requires all solutions.\n"),
 	io__set_exit_status(1).
 det_report_msg(error_in_lambda(DeclaredDetism, InferredDetism, Goal, GoalInfo,
-			PredId, ProcId), ModuleInfo, error) -->
+			PredId, ProcId), ModuleInfo) -->
 	report_pred_proc_id(ModuleInfo, PredId, ProcId, no, _ProcContext),
 	{ goal_info_get_context(GoalInfo, Context) },
 	prog_out__write_context(Context),
@@ -905,52 +1108,35 @@ det_report_msg(error_in_lambda(DeclaredDetism, InferredDetism, Goal, GoalInfo,
 
 %-----------------------------------------------------------------------------%
 
-	% Insertion sort is good enough.
-
-:- pred det_report_sort_context_lines(list(hlds__goal), list(hlds__goal)).
-:- mode det_report_sort_context_lines(in, out) is det.
-
-det_report_sort_context_lines([], []).
-det_report_sort_context_lines([Goal0 | Goals0], Goals) :-
-	det_report_sort_context_lines(Goals0, Goals1),
-	det_report_insert_context_line(Goals1, Goal0, Goals).
-
-:- pred det_report_insert_context_line(list(hlds__goal), hlds__goal,
-	list(hlds__goal)).
-:- mode det_report_insert_context_line(in, in, out) is det.
-
-det_report_insert_context_line([], Goal, [Goal]).
-det_report_insert_context_line([Goal0 | Goals0], Goal, Goals) :-
-	Goal0 = _ - GoalInfo0,
-	goal_info_get_context(GoalInfo0, Context0),
-	term__context_line(Context0, Line0),
-	Goal = _ - GoalInfo,
-	goal_info_get_context(GoalInfo, Context),
-	term__context_line(Context, Line),
-	( Line < Line0 ->
-		Goals = [Goal, Goal0 | Goals0]
+:- pred det_report_seen_call_id(seen_call_id::in, module_info::in,
+		io__state::di, io__state::uo) is det.
+	
+det_report_seen_call_id(SeenCall, ModuleInfo) -->
+	(
+		{ SeenCall = seen_call(PredId, _) },
+		io__write_string("call to "),
+		hlds_out__write_pred_id(ModuleInfo, PredId)
 	;
-		det_report_insert_context_line(Goals0, Goal, Goals1),
-		Goals = [Goal0 | Goals1]
+		{ SeenCall = higher_order_call },
+		io__write_string("higher-order call")
 	).
-
 %-----------------------------------------------------------------------------%
 
-:- pred det_report_context_lines(list(hlds__goal), bool, io__state, io__state).
+:- pred det_report_context_lines(list(term__context), bool, 
+		io__state, io__state).
 :- mode det_report_context_lines(in, in, di, uo) is det.
 
 det_report_context_lines([], _) --> [].
-det_report_context_lines([_ - GoalInfo | Goals], First) -->
-	{ goal_info_get_context(GoalInfo, Context) },
+det_report_context_lines([Context | Contexts], First) -->
 	{ term__context_line(Context, Line) },
 	( { First = yes } ->
 		[]
-	; { Goals = [] } ->
+	; { Contexts = [] } ->
 		io__write_string(" and ")
 	;
 		io__write_string(", ")
 	),
 	io__write_int(Line),
-	det_report_context_lines(Goals, no).
+	det_report_context_lines(Contexts, no).
 
 %-----------------------------------------------------------------------------%
