@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1995 University of Melbourne.
+% Copyright (C) 1996-1997 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -59,7 +59,7 @@
 
 :- import_module hlds_goal, hlds_data, prog_data, instmap.
 :- import_module excess, make_hlds, mode_util.
-:- import_module require, map, list, string, int, bool, std_util.
+:- import_module require, map, list, string, int, bool, std_util, term, varset.
 
 	% Traverse the module structure.
 
@@ -114,7 +114,7 @@ dnf__transform_pred(PredId, MaybeNonAtomic, ModuleInfo0, ModuleInfo,
 	list__delete_all(Markers1, request(dnf), Markers2),
 	pred_info_set_marker_list(PredInfo1, [done(dnf) | Markers2], PredInfo),
 
-	map__set(PredTable1, PredId, PredInfo, PredTable),
+	map__det_update(PredTable1, PredId, PredInfo, PredTable),
 	module_info_set_preds(ModuleInfo1, PredTable, ModuleInfo).
 
 :- pred dnf__transform_procs(list(proc_id)::in, pred_id::in,
@@ -133,12 +133,12 @@ dnf__transform_procs([ProcId | ProcIds], PredId, MaybeNonAtomic,
 	dnf__transform_proc(ProcInfo1, PredInfo0, MaybeNonAtomic,
 		ModuleInfo0, ModuleInfo1, ProcInfo, NewPredIds0, NewPredIds1),
 
-	map__set(ProcTable0, ProcId, ProcInfo, ProcTable),
+	map__det_update(ProcTable0, ProcId, ProcInfo, ProcTable),
 	pred_info_set_procedures(PredInfo0, ProcTable, PredInfo),
 	% We must look up the pred table again
 	% since dnf__transform_proc may have added new predicates
 	module_info_preds(ModuleInfo1, PredTable1),
-	map__set(PredTable1, PredId, PredInfo, PredTable),
+	map__det_update(PredTable1, PredId, PredInfo, PredTable),
 	module_info_set_preds(ModuleInfo1, PredTable, ModuleInfo2),
 
 	dnf__transform_procs(ProcIds, PredId, MaybeNonAtomic,
@@ -169,9 +169,9 @@ dnf__transform_proc(ProcInfo0, PredInfo0, MaybeNonAtomic,
 				list(marker_status)
 			).
 
-:- pred dnf__transform_goal(hlds__goal::in, instmap::in,
+:- pred dnf__transform_goal(hlds_goal::in, instmap::in,
 	maybe(set(pred_proc_id))::in, module_info::in, module_info::out,
-	string::in, dnf_info::in, hlds__goal::out,
+	string::in, dnf_info::in, hlds_goal::out,
 	list(pred_id)::in, list(pred_id)::out) is det.
 
 dnf__transform_goal(Goal0, InstMap0, MaybeNonAtomic, ModuleInfo0, ModuleInfo,
@@ -215,7 +215,7 @@ dnf__transform_goal(Goal0, InstMap0, MaybeNonAtomic, ModuleInfo0, ModuleInfo,
 			DnfInfo, Cond, Then, Else, NewPredIds0, NewPredIds),
 		Goal = if_then_else(Vars, Cond, Then, Else, SM) - GoalInfo
 	;
-		GoalExpr0 = higher_order_call(_, _, _, _, _),
+		GoalExpr0 = higher_order_call(_, _, _, _, _, _),
 		ModuleInfo = ModuleInfo0,
 		NewPredIds = NewPredIds0,
 		Goal = Goal0
@@ -230,7 +230,7 @@ dnf__transform_goal(Goal0, InstMap0, MaybeNonAtomic, ModuleInfo0, ModuleInfo,
 		NewPredIds = NewPredIds0,
 		Goal = Goal0
 	;
-		GoalExpr0 = pragma_c_code(_, _, _, _, _, _),
+		GoalExpr0 = pragma_c_code(_, _, _, _, _, _, _, _),
 		ModuleInfo = ModuleInfo0,
 		NewPredIds = NewPredIds0,
 		Goal = Goal0
@@ -238,9 +238,9 @@ dnf__transform_goal(Goal0, InstMap0, MaybeNonAtomic, ModuleInfo0, ModuleInfo,
 
 %-----------------------------------------------------------------------------%
 
-:- pred dnf__transform_disj(list(hlds__goal)::in, instmap::in,
+:- pred dnf__transform_disj(list(hlds_goal)::in, instmap::in,
 	maybe(set(pred_proc_id))::in, module_info::in, module_info::out,
-	string::in, int::in, dnf_info::in, list(hlds__goal)::out,
+	string::in, int::in, dnf_info::in, list(hlds_goal)::out,
 	list(pred_id)::in, list(pred_id)::out) is det.
 
 dnf__transform_disj([], _, _, ModuleInfo, ModuleInfo, _, _, _, [],
@@ -281,10 +281,10 @@ dnf__transform_switch([Case0 | Cases0], InstMap0, MaybeNonAtomic,
 		ModuleInfo1, ModuleInfo, Base, Counter1, DnfInfo,
 		Cases, NewPredIds1, NewPredIds).
 
-:- pred dnf__transform_ite(hlds__goal::in, hlds__goal::in, hlds__goal::in,
+:- pred dnf__transform_ite(hlds_goal::in, hlds_goal::in, hlds_goal::in,
 	instmap::in, maybe(set(pred_proc_id))::in, module_info::in,
 	module_info::out, string::in, int::in, dnf_info::in,
-	hlds__goal::out, hlds__goal::out, hlds__goal::out,
+	hlds_goal::out, hlds_goal::out, hlds_goal::out,
 	list(pred_id)::in, list(pred_id)::out) is det.
 
 dnf__transform_ite(Cond0, Then0, Else0, InstMap0, MaybeNonAtomic,
@@ -311,9 +311,9 @@ dnf__transform_ite(Cond0, Then0, Else0, InstMap0, MaybeNonAtomic,
 
 %-----------------------------------------------------------------------------%
 
-:- pred dnf__transform_conj(list(hlds__goal)::in, instmap::in,
+:- pred dnf__transform_conj(list(hlds_goal)::in, instmap::in,
 	maybe(set(pred_proc_id))::in, module_info::in, module_info::out,
-	string::in, int::in, int::out, dnf_info::in, list(hlds__goal)::out,
+	string::in, int::in, int::out, dnf_info::in, list(hlds_goal)::out,
 	list(pred_id)::in, list(pred_id)::out) is det.
 
 dnf__transform_conj([], _, _, ModuleInfo, ModuleInfo, _, Counter, Counter,
@@ -331,9 +331,9 @@ dnf__transform_conj([Goal0 | Goals0], InstMap0, MaybeNonAtomic,
 		ModuleInfo1, ModuleInfo, Base, Counter1, Counter, DnfInfo,
 		Goals, NewPredIds1, NewPredIds).
 
-:- pred dnf__make_goal_literal(hlds__goal::in, instmap::in,
+:- pred dnf__make_goal_literal(hlds_goal::in, instmap::in,
 	maybe(set(pred_proc_id))::in, module_info::in, module_info::out,
-	string::in, int::in, int::out, dnf_info::in, hlds__goal::out,
+	string::in, int::in, int::out, dnf_info::in, hlds_goal::out,
 	list(pred_id)::in, list(pred_id)::out) is det.
 
 dnf__make_goal_literal(Goal0, InstMap0, MaybeNonAtomic,
@@ -367,7 +367,7 @@ dnf__get_new_pred_name(PredTable, Base, Name, N0, N) :-
 		N = N1
 	).
 
-:- pred dnf__define_new_pred(hlds__goal::in, hlds__goal::out, instmap::in,
+:- pred dnf__define_new_pred(hlds_goal::in, hlds_goal::out, instmap::in,
 	string::in, dnf_info::in, module_info::in, module_info::out,
 	pred_id::out) is det.
 
@@ -399,8 +399,7 @@ dnf__define_new_pred(Goal0, Goal, InstMap0, PredName, DnfInfo,
 	module_info_set_predicate_table(ModuleInfo0, PredTable,
 		ModuleInfo),
 
-	hlds__is_builtin_make_builtin(no, no, IsBuiltin),
-	GoalExpr = call(PredId, ProcId, ArgVars, IsBuiltin, no, SymName),
+	GoalExpr = call(PredId, ProcId, ArgVars, not_builtin, no, SymName),
 	Goal = GoalExpr - GoalInfo.
 
 :- pred dnf__compute_arg_types_modes(list(var)::in, map(var, type)::in,
@@ -418,7 +417,7 @@ dnf__compute_arg_types_modes([Var | Vars], VarTypes, InstMap0, InstMap,
 
 %-----------------------------------------------------------------------------%
 
-:- pred dnf__is_considered_literal_goal(hlds__goal::in,
+:- pred dnf__is_considered_literal_goal(hlds_goal::in,
 	maybe(set(pred_proc_id))::in) is semidet.
 
 dnf__is_considered_literal_goal(GoalExpr - _, MaybeNonAtomic) :-
@@ -428,7 +427,7 @@ dnf__is_considered_literal_goal(GoalExpr - _, MaybeNonAtomic) :-
 		dnf__is_considered_atomic_expr(GoalExpr, MaybeNonAtomic)
 	).
 
-:- pred dnf__is_considered_atomic_expr(hlds__goal_expr::in,
+:- pred dnf__is_considered_atomic_expr(hlds_goal_expr::in,
 	maybe(set(pred_proc_id))::in) is semidet.
 
 dnf__is_considered_atomic_expr(GoalExpr, MaybeNonAtomic) :-
@@ -439,10 +438,10 @@ dnf__is_considered_atomic_expr(GoalExpr, MaybeNonAtomic) :-
 		dnf__expr_free_of_nonatomic(GoalExpr, NonAtomic)
 	).
 
-:- pred dnf__is_atomic_expr(hlds__goal_expr::in, bool::out) is det.
+:- pred dnf__is_atomic_expr(hlds_goal_expr::in, bool::out) is det.
 
 dnf__is_atomic_expr(conj(_), no).
-dnf__is_atomic_expr(higher_order_call(_, _, _, _, _), yes).
+dnf__is_atomic_expr(higher_order_call(_, _, _, _, _, _), yes).
 dnf__is_atomic_expr(call(_, _, _, _, _, _), yes).
 dnf__is_atomic_expr(switch(_, _, _, _), no).
 dnf__is_atomic_expr(unify(_, _, _, _, _), yes).
@@ -451,9 +450,9 @@ dnf__is_atomic_expr(not(_), no).
 dnf__is_atomic_expr(some(_, GoalExpr - _), IsAtomic) :-
 	dnf__is_atomic_expr(GoalExpr, IsAtomic).
 dnf__is_atomic_expr(if_then_else(_, _, _, _, _), no).
-dnf__is_atomic_expr(pragma_c_code(_, _, _, _, _, _), yes).
+dnf__is_atomic_expr(pragma_c_code(_, _, _, _, _, _, _, _), yes).
 
-:- pred dnf__expr_free_of_nonatomic(hlds__goal_expr::in,
+:- pred dnf__expr_free_of_nonatomic(hlds_goal_expr::in,
 	set(pred_proc_id)::in) is semidet.
 
 dnf__expr_free_of_nonatomic(conj(Goals), NonAtomic) :-
@@ -473,15 +472,15 @@ dnf__expr_free_of_nonatomic(if_then_else(_, Cond, Then, Else, _), NonAtomic) :-
 	dnf__goal_free_of_nonatomic(Cond, NonAtomic),
 	dnf__goal_free_of_nonatomic(Then, NonAtomic),
 	dnf__goal_free_of_nonatomic(Else, NonAtomic).
-dnf__expr_free_of_nonatomic(pragma_c_code(_, _, _, _, _, _), _NonAtomic).
+dnf__expr_free_of_nonatomic(pragma_c_code(_, _, _, _, _, _, _, _), _NonAtomic).
 
-:- pred dnf__goal_free_of_nonatomic(hlds__goal::in,
+:- pred dnf__goal_free_of_nonatomic(hlds_goal::in,
 	set(pred_proc_id)::in) is semidet.
 
 dnf__goal_free_of_nonatomic(GoalExpr - _, NonAtomic) :-
 	dnf__expr_free_of_nonatomic(GoalExpr, NonAtomic).
 
-:- pred dnf__goals_free_of_nonatomic(list(hlds__goal)::in,
+:- pred dnf__goals_free_of_nonatomic(list(hlds_goal)::in,
 	set(pred_proc_id)::in) is semidet.
 
 dnf__goals_free_of_nonatomic([], _NonAtomic).
