@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1995 University of Melbourne.
+% Copyright (C) 1995-1997 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -31,15 +31,15 @@
 
 	% goal_util__rename_vars_in_goals(GoalList, MustRename, Substitution,
 	%	NewGoalList).
-:- pred goal_util__rename_vars_in_goals(list(hlds__goal), bool, map(var, var),
-	list(hlds__goal)).
+:- pred goal_util__rename_vars_in_goals(list(hlds_goal), bool, map(var, var),
+	list(hlds_goal)).
 :- mode goal_util__rename_vars_in_goals(in, in, in, out) is det.
 
-:- pred goal_util__rename_vars_in_goal(hlds__goal, map(var, var), hlds__goal).
+:- pred goal_util__rename_vars_in_goal(hlds_goal, map(var, var), hlds_goal).
 :- mode goal_util__rename_vars_in_goal(in, in, out) is det.
 
-:- pred goal_util__must_rename_vars_in_goal(hlds__goal,
-					map(var, var), hlds__goal).
+:- pred goal_util__must_rename_vars_in_goal(hlds_goal,
+					map(var, var), hlds_goal).
 :- mode goal_util__must_rename_vars_in_goal(in, in, out) is det.
 
 :- pred goal_util__rename_var_list(list(var), bool, map(var, var), list(var)).
@@ -58,19 +58,19 @@
 	% Return all the variables in the goal.
 	% Unlike quantification:goal_vars, this predicate returns
 	% even the explicitly quantified variables.
-:- pred goal_util__goal_vars(hlds__goal, set(var)).
+:- pred goal_util__goal_vars(hlds_goal, set(var)).
 :- mode goal_util__goal_vars(in, out) is det.
 
 	% See whether the goal is a branched structure.
-:- pred goal_util__goal_is_branched(hlds__goal_expr).
+:- pred goal_util__goal_is_branched(hlds_goal_expr).
 :- mode goal_util__goal_is_branched(in) is semidet.
 
 	% Return an indication of the size of the goal.
-:- pred goal_size(hlds__goal, int).
+:- pred goal_size(hlds_goal, int).
 :- mode goal_size(in, out) is det.
 
 	% Test whether the goal calls the given procedure.
-:- pred goal_calls(hlds__goal, pred_proc_id).
+:- pred goal_calls(hlds_goal, pred_proc_id).
 :- mode goal_calls(in, in) is semidet.
 
 %-----------------------------------------------------------------------------%
@@ -85,7 +85,7 @@
 
 goal_util__create_variables([], Varset, VarTypes, Subn, _OldVarTypes,
 		_OldVarNames, Varset, VarTypes, Subn).
-goal_util__create_variables([V|Vs], Varset0, VarTypes0, Subn0, OldVarTypes,
+goal_util__create_variables([V | Vs], Varset0, VarTypes0, Subn0, OldVarTypes,
 					OldVarNames, Varset, VarTypes, Subn) :-
 	(
 		map__contains(Subn0, V)
@@ -102,7 +102,7 @@ goal_util__create_variables([V|Vs], Varset0, VarTypes0, Subn0, OldVarTypes,
 		;
 			Varset2 = Varset1
 		),
-		map__set(Subn0, V, NV, Subn1),
+		map__det_insert(Subn0, V, NV, Subn1),
 		(
 			map__search(OldVarTypes, V, VT)
 		->
@@ -116,18 +116,28 @@ goal_util__create_variables([V|Vs], Varset0, VarTypes0, Subn0, OldVarTypes,
 
 %-----------------------------------------------------------------------------%
 
-:- pred goal_util__init_subn(assoc_list(var, var), map(var, var), map(var, var)).
+:- pred goal_util__init_subn(assoc_list(var, var),
+	map(var, var), map(var, var)).
 :- mode goal_util__init_subn(in, in, out) is det.
 
 goal_util__init_subn([], Subn, Subn).
-goal_util__init_subn([A-H|Vs], Subn0, Subn) :-
+goal_util__init_subn([A - H | Vs], Subn0, Subn) :-
 	map__set(Subn0, H, A, Subn1),
 	goal_util__init_subn(Vs, Subn1, Subn).
 
 %-----------------------------------------------------------------------------%
 
+:- pred goal_util__rename_var_pair_list(assoc_list(var, T), bool,
+	map(var, var), list(pair(var, T))).
+:- mode goal_util__rename_var_pair_list(in, in, in, out) is det.
+
+goal_util__rename_var_pair_list([], _Must, _Subn, []).
+goal_util__rename_var_pair_list([V - D | VDs], Must, Subn, [N - D | NDs]) :-
+	goal_util__rename_var(V, Must, Subn, N),
+	goal_util__rename_var_pair_list(VDs, Must, Subn, NDs).
+
 goal_util__rename_var_list([], _Must, _Subn, []).
-goal_util__rename_var_list([V|Vs], Must, Subn, [N|Ns]) :-
+goal_util__rename_var_list([V | Vs], Must, Subn, [N | Ns]) :-
 	goal_util__rename_var(V, Must, Subn, N),
 	goal_util__rename_var_list(Vs, Must, Subn, Ns).
 
@@ -140,8 +150,13 @@ goal_util__rename_var(V, Must, Subn, N) :-
 	->
 		N = N0
 	;
-		( Must = no, N = V
-		; Must = yes, error("goal_util__rename_var: no substitute") )
+		(
+			Must = no,
+			N = V
+		;
+			Must = yes,
+			error("goal_util__rename_var: no substitute")
+		)
 	).
 
 %-----------------------------------------------------------------------------%
@@ -159,8 +174,8 @@ goal_util__rename_vars_in_goals([Goal0 | Goals0], Must, Subn, [Goal | Goals]) :-
 	goal_util__rename_vars_in_goal(Goal0, Must, Subn, Goal),
 	goal_util__rename_vars_in_goals(Goals0, Must, Subn, Goals).
 
-:- pred goal_util__rename_vars_in_goal(hlds__goal, bool, map(var, var),
-		hlds__goal).
+:- pred goal_util__rename_vars_in_goal(hlds_goal, bool, map(var, var),
+		hlds_goal).
 :- mode goal_util__rename_vars_in_goal(in, in, in, out) is det.
 
 goal_util__rename_vars_in_goal(Goal0 - GoalInfo0, Must, Subn, Goal - GoalInfo) :-
@@ -169,8 +184,8 @@ goal_util__rename_vars_in_goal(Goal0 - GoalInfo0, Must, Subn, Goal - GoalInfo) :
 
 %-----------------------------------------------------------------------------%
 
-:- pred goal_util__name_apart_2(hlds__goal_expr, bool, map(var, var),
-		hlds__goal_expr).
+:- pred goal_util__name_apart_2(hlds_goal_expr, bool, map(var, var),
+		hlds_goal_expr).
 :- mode goal_util__name_apart_2(in, in, in, out) is det.
 
 goal_util__name_apart_2(conj(Goals0), Must, Subn, conj(Goals)) :-
@@ -202,9 +217,11 @@ goal_util__name_apart_2(some(Vars0, Goal0), Must, Subn, some(Vars, Goal)) :-
 	goal_util__rename_vars_in_goal(Goal0, Must, Subn, Goal).
 
 goal_util__name_apart_2(
-		higher_order_call(PredVar0, Args0, Types, Modes, Det),
+		higher_order_call(PredVar0, Args0, Types, Modes, Det,
+			IsPredOrFunc),
 		Must, Subn,
-		higher_order_call(PredVar, Args, Types, Modes, Det)) :-
+		higher_order_call(PredVar, Args, Types, Modes, Det,
+			IsPredOrFunc)) :-
 	goal_util__rename_var(PredVar0, Must, Subn, PredVar),
 	goal_util__rename_var_list(Args0, Must, Subn, Args).
 
@@ -220,18 +237,27 @@ goal_util__name_apart_2(unify(TermL0,TermR0,Mode,Unify0,Context), Must, Subn,
 	goal_util__rename_unify_rhs(TermR0, Must, Subn, TermR),
 	goal_util__rename_unify(Unify0, Must, Subn, Unify).
 
-goal_util__name_apart_2(pragma_c_code(A,B,C,D,Vars0,E), Must, Subn,
-		pragma_c_code(A,B,C,D,Vars,E)) :-
-	goal_util__rename_var_list(Vars0, Must, Subn, Vars).
+goal_util__name_apart_2(pragma_c_code(A,B,C,D,Vars0,F,G,Extra0), Must, Subn,
+		pragma_c_code(A,B,C,D,Vars,F,G,Extra)) :-
+	goal_util__rename_var_list(Vars0, Must, Subn, Vars),
+	(
+		Extra0 = none,
+		Extra = none
+	;
+		Extra0 = extra_pragma_info(SavedVars0, LabelNames),
+		goal_util__rename_var_pair_list(SavedVars0, Must, Subn,
+			SavedVars),
+		Extra = extra_pragma_info(SavedVars, LabelNames)
+	).
 
 %-----------------------------------------------------------------------------%
 
-:- pred goal_util__name_apart_list(list(hlds__goal), bool, map(var, var),
-							list(hlds__goal)).
+:- pred goal_util__name_apart_list(list(hlds_goal), bool, map(var, var),
+							list(hlds_goal)).
 :- mode goal_util__name_apart_list(in, in, in, out) is det.
 
 goal_util__name_apart_list([], _Must, _Subn, []).
-goal_util__name_apart_list([G0|Gs0], Must, Subn, [G|Gs]) :-
+goal_util__name_apart_list([G0 | Gs0], Must, Subn, [G | Gs]) :-
 	goal_util__rename_vars_in_goal(G0, Must, Subn, G),
 	goal_util__name_apart_list(Gs0, Must, Subn, Gs).
 
@@ -242,7 +268,8 @@ goal_util__name_apart_list([G0|Gs0], Must, Subn, [G|Gs]) :-
 :- mode goal_util__name_apart_cases(in, in, in, out) is det.
 
 goal_util__name_apart_cases([], _Must, _Subn, []).
-goal_util__name_apart_cases([case(Cons, G0)|Gs0], Must, Subn, [case(Cons, G)|Gs]) :-
+goal_util__name_apart_cases([case(Cons, G0) | Gs0], Must, Subn,
+		[case(Cons, G) | Gs]) :-
 	goal_util__rename_vars_in_goal(G0, Must, Subn, G),
 	goal_util__name_apart_cases(Gs0, Must, Subn, Gs).
 
@@ -254,7 +281,7 @@ goal_util__name_apart_cases([case(Cons, G0)|Gs0], Must, Subn, [case(Cons, G)|Gs]
 :- mode goal_util__rename_args(in, in, in, out) is det.
 
 goal_util__rename_args([], _Must, _Subn, []).
-goal_util__rename_args([T0|Ts0], Must, Subn, [T|Ts]) :-
+goal_util__rename_args([T0 | Ts0], Must, Subn, [T | Ts]) :-
 	goal_util__rename_term(T0, Must, Subn, T),
 	goal_util__rename_args(Ts0, Must, Subn, Ts).
 
@@ -324,8 +351,8 @@ goal_util__rename_var_maps_2([V - L | Vs], Must, Subn, [N - L | Ns]) :-
 
 %-----------------------------------------------------------------------------%
 
-:- pred goal_util__name_apart_goalinfo(hlds__goal_info,
-					bool, map(var, var), hlds__goal_info).
+:- pred goal_util__name_apart_goalinfo(hlds_goal_info,
+					bool, map(var, var), hlds_goal_info).
 :- mode goal_util__name_apart_goalinfo(in, in, in, out) is det.
 
 goal_util__name_apart_goalinfo(GoalInfo0, Must, Subn, GoalInfo) :-
@@ -380,14 +407,14 @@ goal_util__goal_vars(Goal - _GoalInfo, Set) :-
 	set__init(Set0),
 	goal_util__goal_vars_2(Goal, Set0, Set).
 
-:- pred goal_util__goal_vars_2(hlds__goal_expr, set(var), set(var)).
+:- pred goal_util__goal_vars_2(hlds_goal_expr, set(var), set(var)).
 :- mode goal_util__goal_vars_2(in, in, out) is det.
 
 goal_util__goal_vars_2(unify(Var, RHS, _, _, _), Set0, Set) :-
 	set__insert(Set0, Var, Set1),
 	goal_util__rhs_goal_vars(RHS, Set1, Set).
 
-goal_util__goal_vars_2(higher_order_call(PredVar, ArgVars, _, _, _),
+goal_util__goal_vars_2(higher_order_call(PredVar, ArgVars, _, _, _, _),
 		Set0, Set) :-
 	set__insert_list(Set0, [PredVar | ArgVars], Set).
 
@@ -417,10 +444,19 @@ goal_util__goal_vars_2(if_then_else(Vars, A - _, B - _, C - _, _), Set0, Set) :-
 	goal_util__goal_vars_2(B, Set2, Set3),
 	goal_util__goal_vars_2(C, Set3, Set).
 
-goal_util__goal_vars_2(pragma_c_code(_, _, _, _, ArgVars, _), Set0, Set) :-
-	set__insert_list(Set0, ArgVars, Set).
+goal_util__goal_vars_2(pragma_c_code(_, _, _, _, ArgVars, _, _, Extra),
+		Set0, Set) :-
+	set__insert_list(Set0, ArgVars, Set1),
+	(
+		Extra = none,
+		Set = Set1
+	;
+		Extra = extra_pragma_info(SavedVarNames, _),
+		assoc_list__keys(SavedVarNames, SavedVars),
+		set__insert_list(Set1, SavedVars, Set)
+	).
 
-:- pred goal_util__goals_goal_vars(list(hlds__goal), set(var), set(var)).
+:- pred goal_util__goals_goal_vars(list(hlds_goal), set(var), set(var)).
 :- mode goal_util__goals_goal_vars(in, in, out) is det.
 
 goal_util__goals_goal_vars([], Set, Set).
@@ -459,7 +495,7 @@ goal_util__goal_is_branched(disj(_, _)).
 goal_size(GoalExpr - _, Size) :-
 	goal_expr_size(GoalExpr, Size).
 
-:- pred goals_size(list(hlds__goal), int).
+:- pred goals_size(list(hlds_goal), int).
 :- mode goals_size(in, out) is det.
 
 goals_size([], 0).
@@ -477,7 +513,7 @@ cases_size([case(_, Goal) | Cases], Size) :-
 	cases_size(Cases, Size2),
 	Size is Size1 + Size2.
 
-:- pred goal_expr_size(hlds__goal_expr, int).
+:- pred goal_expr_size(hlds_goal_expr, int).
 :- mode goal_expr_size(in, out) is det.
 
 goal_expr_size(conj(Goals), Size) :-
@@ -500,16 +536,16 @@ goal_expr_size(some(_, Goal), Size) :-
 	goal_size(Goal, Size1),
 	Size is Size1 + 1.
 goal_expr_size(call(_, _, _, _, _, _), 1).
-goal_expr_size(higher_order_call(_, _, _, _, _), 1).
+goal_expr_size(higher_order_call(_, _, _, _, _, _), 1).
 goal_expr_size(unify(_, _, _, _, _), 1).
-goal_expr_size(pragma_c_code(_, _, _, _, _, _), 1).
+goal_expr_size(pragma_c_code(_, _, _, _, _, _, _, _), 1).
 
 %-----------------------------------------------------------------------------%
 
 goal_calls(GoalExpr - _, PredProcId) :-
 	goal_expr_calls(GoalExpr, PredProcId).
 
-:- pred goals_calls(list(hlds__goal), pred_proc_id).
+:- pred goals_calls(list(hlds_goal), pred_proc_id).
 :- mode goals_calls(in, in) is semidet.
 
 goals_calls([Goal | Goals], PredProcId) :-
@@ -529,7 +565,7 @@ cases_calls([case(_, Goal) | Cases], PredProcId) :-
 		cases_calls(Cases, PredProcId)
 	).
 
-:- pred goal_expr_calls(hlds__goal_expr, pred_proc_id).
+:- pred goal_expr_calls(hlds_goal_expr, pred_proc_id).
 :- mode goal_expr_calls(in, in) is semidet.
 
 goal_expr_calls(conj(Goals), PredProcId) :-
