@@ -1,5 +1,5 @@
-%------------------------------------------------------------------------------
-% Copyright (C) 1995 University of Melbourne.
+%-----------------------------------------------------------------------------
+% Copyright (C) 1996-1997 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -30,12 +30,12 @@
 
 :- implementation.
 
-:- import_module hlds_pred, hlds_goal, hlds_data, instmap.
+:- import_module hlds_pred, hlds_goal, hlds_data, instmap, (inst).
 :- import_module code_util, globals, make_hlds, mode_util, goal_util.
-:- import_module type_util, options, prog_data, quantification.
+:- import_module type_util, options, prog_data, quantification, (lambda).
 :- import_module mercury_to_mercury.
 
-:- import_module assoc_list, bool, int, list, map, require, set.
+:- import_module assoc_list, bool, char, int, list, map, require, set.
 :- import_module std_util, string, varset, term.
 
 	% Iterate collecting requests and processing them until there
@@ -53,7 +53,7 @@ specialize_higher_order(ModuleInfo0, ModuleInfo) -->
 
 process_requests(Requests0, GoalSizes0, NextHOid0, NextHOid,
 			NewPreds0, NewPreds, ModuleInfo1, ModuleInfo) -->
-	{ filter_requests(Requests0, GoalSizes0, Requests) },
+	{ filter_requests(ModuleInfo1, Requests0, GoalSizes0, Requests) },
 	(
 		{ Requests = [] }
 	->
@@ -69,13 +69,26 @@ process_requests(Requests0, GoalSizes0, NextHOid0, NextHOid,
 		{ map__keys(NewPredsForThisPass, SpecializedPreds) },
 		{ map__merge(NewPreds0, NewPredsForThisPass, NewPreds1) },
 		{ set__to_sorted_list(PredProcsToFix, PredProcs) },
-		{ fixup_preds(PredProcs, NewPreds1, ModuleInfo2, ModuleInfo3) },
 		{ set__init(NewRequests0) },
 		{ create_specialized_versions(SpecializedPreds, NewPreds1,
 				NewRequests0, NewRequests, GoalSizes0,
-				GoalSizes, ModuleInfo3, ModuleInfo4) },
+				GoalSizes, ModuleInfo2, ModuleInfo3) },
+			% The dependencies have changed, so the
+			% dependency graph needs to rebuilt for
+			% inlining to work properly.
+			% XXX Is it worth rebuilding the dependency graph
+			% here completely rebuilding it later? (it's only
+			% necessary in profiling grades, since otherwise
+			% the dependency graph isn't built before here). 
+		{ fixup_preds(PredProcs, NewPreds1, ModuleInfo3, ModuleInfo4) },
+		{ SpecializedPreds = [] ->
+			module_info_clobber_dependency_info(ModuleInfo4,
+				ModuleInfo5)
+		;
+			ModuleInfo5 = ModuleInfo4
+		},
 		process_requests(NewRequests, GoalSizes, NextHOid1,
-			NextHOid, NewPreds1, NewPreds, ModuleInfo4, ModuleInfo)
+			NextHOid, NewPreds1, NewPreds, ModuleInfo5, ModuleInfo)
 	).
 
 
@@ -193,17 +206,11 @@ get_specialization_requests_2([PredId | PredIds], Requests0, Requests,
 		traverse_goal(Goal0, Goal1, PredProcId, Changed,
 				GoalSize, Info0, info(_, Requests1,_,_)),
 		map__set(GoalSizes0, PredId, GoalSize, GoalSizes1),
+		proc_info_set_goal(ProcInfo0, Goal1, ProcInfo1),
 		(
 			Changed = changed
 		->
-			proc_info_vartypes(ProcInfo0, VarTypes0),
-			proc_info_headvars(ProcInfo0, HeadVars),
-			proc_info_variables(ProcInfo0, Varset0),
-			implicitly_quantify_clause_body(HeadVars, Goal1,
-				Varset0, VarTypes0, Goal, Varset, VarTypes, _),
-			proc_info_set_goal(ProcInfo0, Goal, ProcInfo1),
-			proc_info_set_variables(ProcInfo1, Varset, ProcInfo2),
-			proc_info_set_vartypes(ProcInfo2, VarTypes, ProcInfo),
+			requantify_proc(ProcInfo1, ProcInfo),
 			map__det_update(Procs0, ProcId, ProcInfo, Procs1)
 		;
 			Procs1 = Procs0
@@ -218,7 +225,7 @@ get_specialization_requests_2([PredId | PredIds], Requests0, Requests,
 			module_info_set_preds(ModuleInfo0, Preds, ModuleInfo1)
 		;
 			ModuleInfo1 = ModuleInfo0,
-			Requests2 = Requests1	
+			Requests2 = Requests1
 		)
 	),
 	get_specialization_requests_2(PredIds, Requests2, Requests,
@@ -264,7 +271,7 @@ traverse_other_procs(PredId, [ProcId | ProcIds], ModuleInfo, Requests0,
 	% The first time through the only predicate we can specialize
 	% is call/N. The pred_proc_id is that of the current procedure,
 	% used to find out which procedures need fixing up later.
-:- pred traverse_goal(hlds__goal::in, hlds__goal::out, pred_proc_id::in,
+:- pred traverse_goal(hlds_goal::in, hlds_goal::out, pred_proc_id::in,
 	changed::out, int::out, higher_order_info::in,
 	higher_order_info::out) is det.
 
@@ -285,7 +292,7 @@ traverse_goal(switch(Var, CanFail, Cases0, SM) - Info,
 
 		% check whether this call could be specialized
 traverse_goal(Goal0, Goal, PredProcId, Changed, 1) -->
-	{ Goal0 = higher_order_call(_,_,_,_,_) - _ }, 
+	{ Goal0 = higher_order_call(_,_,_,_,_,_) - _ }, 
 	maybe_specialize_higher_order_call(Goal0, Goal, PredProcId, Changed).
 
 		% check whether this call could be specialized
@@ -317,14 +324,14 @@ traverse_goal(some(Vars, Goal0) - Info, some(Vars, Goal) - Info,
 	traverse_goal(Goal0, Goal, PredProcId, Changed, GoalSize).
 
 traverse_goal(Goal, Goal, _, unchanged, 1) -->
-	{ Goal = pragma_c_code(_, _, _, _, _, _) - _ }.
+	{ Goal = pragma_c_code(_, _, _, _, _, _, _, _) - _ }.
 
 traverse_goal(Goal, Goal, _, unchanged, 1) -->
 	{ Goal = unify(_, _, _, Unify, _) - _ }, 
 	check_unify(Unify).
 
 
-:- pred traverse_conj(hlds__goals::in, hlds__goals::out, pred_proc_id::in,
+:- pred traverse_conj(hlds_goals::in, hlds_goals::out, pred_proc_id::in,
 	changed::in, changed::out, int::in, int::out, higher_order_info::in,
 	higher_order_info::out) is det.
 
@@ -341,7 +348,7 @@ traverse_conj([Goal0 | Goals0], [Goal | Goals],
 		% specialization information before the goal, then merge the
 		% results to give the specialization information after the
 		% disjunction.
-:- pred traverse_disj(hlds__goals::in, hlds__goals::out, pred_proc_id::in,
+:- pred traverse_disj(hlds_goals::in, hlds_goals::out, pred_proc_id::in,
 		changed::out, int::out, higher_order_info::in,
 		higher_order_info::out) is det.
 
@@ -354,7 +361,7 @@ traverse_disj([Goal0 | Goals0], [Goal | Goals], PredProcId,
 			Changed0, Changed, GoalSize0, GoalSize, Info0).
 
 
-:- pred traverse_disj_2(hlds__goals::in, hlds__goals::out, pred_proc_id::in,
+:- pred traverse_disj_2(hlds_goals::in, hlds_goals::out, pred_proc_id::in,
 	changed::in, changed::out, int::in, int::out, higher_order_info::in,
 	higher_order_info::in, higher_order_info::out) is det.
 
@@ -370,6 +377,12 @@ traverse_disj_2([Goal0 | Goals0], [Goal | Goals], PredProcId, Changed0, Changed,
 				GoalSize1, GoalSize, InitialInfo, Info1, Info).
 
 
+				% The dependencies have changed, so the
+				% dependency graph needs to rebuilt for
+				% inlining to work properly.
+				% XXX Fix the graph here instead of 
+				% completely rebuilding it later (only
+				% a problem in profiling grades).
 		% Switches are treated in exactly the same way as disjunctions.
 :- pred traverse_cases(list(case)::in, list(case)::out, pred_proc_id::in,
 		changed::out, int::out, higher_order_info::in,
@@ -499,8 +512,8 @@ check_unify(construct(LVar, ConsId, Args, _Modes), Info0, Info) :-
 				PredVars = PredVars0
 			)
 		;
-			map__set(PredVars0, LVar, yes(PredId, ProcId, Args),
-								PredVars)
+			map__det_insert(PredVars0, LVar,
+				yes(PredId, ProcId, Args), PredVars)
 		)
 	;
 		PredVars = PredVars0	
@@ -512,16 +525,20 @@ check_unify(complicated_unify(_, _)) -->
 
 		% Process a higher-order call to see if it could possibly
 		% be specialized.
-:- pred maybe_specialize_higher_order_call( hlds__goal::in, hlds__goal::out,
+:- pred maybe_specialize_higher_order_call( hlds_goal::in, hlds_goal::out,
 		pred_proc_id::in, changed::out, higher_order_info::in,
 		higher_order_info::out) is det.
 
 maybe_specialize_higher_order_call(Goal0 - GoalInfo, Goal - GoalInfo,
 		PredProcId, Changed, Info0, Info) :-
 	Info0 = info(PredVars, Requests0, NewPreds, Module),
-	( Goal0 = higher_order_call(PredVar0, Args0, _Types, _Modes, _Det) ->
+	(
+		Goal0 = higher_order_call(PredVar0, Args0, _Types, Modes0,
+				_Det, _IsPredOrFunc)
+	->
 		PredVar = PredVar0,
-		Args = Args0
+		Args = Args0,
+		Modes = Modes0
 	;
 		error("higher_order.m: higher_order_call expected")
 	),
@@ -534,12 +551,17 @@ maybe_specialize_higher_order_call(Goal0 - GoalInfo, Goal - GoalInfo,
 		module_info_pred_info(Module, PredId, PredInfo),
 		pred_info_module(PredInfo, ModuleName),
 		pred_info_name(PredInfo, PredName),
-		code_util__is_builtin(Module, PredId,
-					ProcId, IsBuiltin),
-		list__append(CurriedArgs, Args, AllArgs),
+		code_util__builtin_state(Module, PredId, ProcId, Builtin),
+
+		% We need to permute the arguments so that inputs come before
+		% outputs, since lambda.m will have done that to the arguments
+		% of the closure.
+		lambda__permute_argvars(Args, Modes, Module, PermutedArgs, _),
+
+		list__append(CurriedArgs, PermutedArgs, AllArgs),
 		MaybeContext = no,
 		Goal1 = call(PredId, ProcId, AllArgs,
-			IsBuiltin, MaybeContext,
+			Builtin, MaybeContext,
 			qualified(ModuleName, PredName)),
 		maybe_specialize_call(Goal1 - GoalInfo,
 			Goal - _, PredProcId, _, Info0,
@@ -554,7 +576,7 @@ maybe_specialize_higher_order_call(Goal0 - GoalInfo, Goal - GoalInfo,
 	Info = info(PredVars, Requests, NewPreds, Module).
 
 		% Process a call to see if it could possibly be specialized.
-:- pred maybe_specialize_call( hlds__goal::in, hlds__goal::out,
+:- pred maybe_specialize_call( hlds_goal::in, hlds_goal::out,
 		pred_proc_id::in, changed::out, higher_order_info::in,
 		higher_order_info::out) is det.
 
@@ -569,69 +591,93 @@ maybe_specialize_call(Goal0 - GoalInfo, Goal - GoalInfo, PredProcId,
 	;
 		error("higher_order.m: call expected")
 	),
-	find_higher_order_args(Args0, PredVars, 1,
-			[], HigherOrderArgs, Args0, Args1),
+	module_info_pred_info(Module, CalledPred, PredInfo),
 	(
-		HigherOrderArgs = []
+		pred_info_is_imported(PredInfo)
 	->
 		Requests = Requests0,
 		Changed = unchanged,
 		Goal = Goal0
-	;
-		% Check to see if any of the specialized
-		% versions of the called pred apply here.
-		map__search(NewPreds,
-			proc(CalledPred, CalledProc),
-			NewPredSet),
-		set__to_sorted_list(NewPredSet, NewPredList),	% NOP
-		list__filter(lambda([X::in] is semidet, (
-			X = new_pred(_,_,_, HigherOrderArgs)
-			)), NewPredList, Matches),
+	;	
+		pred_info_arg_types(PredInfo, _, ArgTypes),
+		find_higher_order_args(Module, Args0, ArgTypes, PredVars, 1,
+				[], HigherOrderArgs, Args0, Args1),
 		(
-			Matches = [Match],
-			Match = new_pred(NewCalledPred,
-					NewCalledProc, NewName,
-					_HOArgs)
+			HigherOrderArgs = [] 
+		->
+			Requests = Requests0,
+			Changed = unchanged,
+			Goal = Goal0
 		;
-			Matches = [_,_|_],
-			error("multiple specializations")
-		)
-	->
-		remove_listof_higher_order_args(Args1, 1,
-				HigherOrderArgs, Args2),
-		Goal = call(NewCalledPred, NewCalledProc,
-				Args2, IsBuiltin, MaybeContext, NewName),
-		Changed = changed,
-		Requests = Requests0
-	;
-		% There is a known higher order variable in the
-		% call, so we put in a request for a specialized
-		% version of the pred.
-		Goal = Goal0,
-		Request = request(PredProcId,
+			% Check to see if any of the specialized
+			% versions of the called pred apply here.
+			map__search(NewPreds,
 				proc(CalledPred, CalledProc),
-				HigherOrderArgs), 
-		set__insert(Requests0, Request, Requests),
-		Changed = request
+				NewPredSet),
+			set__to_sorted_list(NewPredSet, NewPredList),	% NOP
+			list__filter(lambda([X::in] is semidet, (
+				X = new_pred(_,_,_, HigherOrderArgs)
+				)), NewPredList, Matches),
+			(
+				Matches = [Match],
+				Match = new_pred(NewCalledPred,
+						NewCalledProc, NewName,
+						_HOArgs)
+			;
+				Matches = [_,_|_],
+				error("multiple specializations")
+			)
+		->
+			remove_listof_higher_order_args(Args1, 1,
+					HigherOrderArgs, Args2),
+			Goal = call(NewCalledPred, NewCalledProc,
+				Args2, IsBuiltin, MaybeContext, NewName),
+			Changed = changed,
+			Requests = Requests0
+		;
+			% There is a known higher order variable in the
+			% call, so we put in a request for a specialized
+			% version of the pred.
+			Goal = Goal0,
+			Request = request(PredProcId,
+					proc(CalledPred, CalledProc),
+					HigherOrderArgs), 
+			set__insert(Requests0, Request, Requests),
+			Changed = request
+		)
 	),
 	Info = info(PredVars, Requests, NewPreds, Module).
 
 	% Returns a list of the higher-order arguments in a call that have
 	% a known value. Also update the argument list to now include
 	% curried arguments that need to be explicitly passed.
-:- pred find_higher_order_args(list(var)::in, pred_vars::in, int::in,
-		list(higher_order_arg)::in, list(higher_order_arg)::out,
+	% The order of the argument list must match that generated
+	% by construct_higher_order terms.
+:- pred find_higher_order_args(module_info::in, list(var)::in, list(type)::in,
+		pred_vars::in, int::in, list(higher_order_arg)::in,
+		list(higher_order_arg)::out,
 		list(var)::in, list(var)::out) is det.
 
-find_higher_order_args([], _, _, HOArgs, HOArgs, NewArgs, NewArgs).
-find_higher_order_args([Arg | Args], PredVars, ArgNo,
-			HOArgs0, HOArgs, NewArgs0, NewArgs) :-
+find_higher_order_args(_, [], _, _, _, HOArgs, HOArgs, NewArgs, NewArgs).
+find_higher_order_args(_, [_|_], [], _, _, _, _, _, _) :-
+	error("find_higher_order_args: length mismatch").
+find_higher_order_args(ModuleInfo, [Arg | Args], [ArgType | ArgTypes],
+		PredVars, ArgNo, HOArgs0, HOArgs, NewArgs0, NewArgs) :-
 	NextArg is ArgNo + 1,
 	(
+		% We don't specialize arguments whose declared type is
+		% polymorphic. The closure they pass cannot possibly
+		% be called within the called predicate, since that predicate 
+		% doesn't know it's a closure (without some dodgy use of
+		% type_to_univ and univ_to_type).
+		type_is_higher_order(ArgType, _, _),
 		map__search(PredVars, Arg, yes(PredId, ProcId, CurriedArgs))
 	->
-		find_higher_order_args(CurriedArgs, PredVars, 1,
-			[], HOCurriedArgs, CurriedArgs, NewExtraArgs0),
+		module_info_pred_info(ModuleInfo, PredId, PredInfo),
+		pred_info_arg_types(PredInfo, _, CurriedArgTypes),
+		find_higher_order_args(ModuleInfo, CurriedArgs,
+			CurriedArgTypes, PredVars, 1, [], HOCurriedArgs,
+			CurriedArgs, NewExtraArgs0),
 		list__length(CurriedArgs, NumArgs),
 		remove_listof_higher_order_args(NewExtraArgs0, 1, HOCurriedArgs,
 								NewExtraArgs),
@@ -642,7 +688,7 @@ find_higher_order_args([Arg | Args], PredVars, ArgNo,
 		HOArgs1 = HOArgs0,
 		NewArgs1 = NewArgs0
 	),
-	find_higher_order_args(Args, PredVars, NextArg,
+	find_higher_order_args(ModuleInfo, Args, ArgTypes, PredVars, NextArg,
 			HOArgs1, HOArgs, NewArgs1, NewArgs).
 		
 		% if the right argument of an assignment is a higher order
@@ -681,17 +727,32 @@ update_changed_status(unchanged, Changed, Changed).
 		% specialized, as with inlining?
 		% Nonlocal predicates are filtered out here, since they
 		% will not have an entry in the goal_sizes.
-:- pred filter_requests(set(request)::in, goal_sizes::in,
+		% Don't create specialized versions of specialized
+		% versions, since for some fairly contrived examples 
+		% involving recursively building up lambda expressions
+		% this can create ridiculous numbers of versions.
+:- pred filter_requests(module_info::in, set(request)::in, goal_sizes::in,
 						list(request)::out) is det.
 
-filter_requests(Requests0, GoalSizes, Requests) :-
+filter_requests(ModuleInfo, Requests0, GoalSizes, Requests) :-
 	set__to_sorted_list(Requests0, Requests1),
 	list__filter(lambda([X::in] is semidet, (
 			X = request(_, CalledPredProcId, _),
 			CalledPredProcId = proc(CalledPredId, _),
 			map__search(GoalSizes, CalledPredId, GoalSize),
 			max_specialized_goal_size(MaxSize),
-			GoalSize =< MaxSize)),
+			GoalSize =< MaxSize,
+			predicate_name(ModuleInfo, CalledPredId, PredName),
+			\+ (
+				% There are probably cleaner ways to check 
+				% if this is a specialised version.
+				string__sub_string_search(PredName, 
+					"__ho", Index),
+				NumIndex is Index + 4,
+				string__index(PredName, NumIndex, Digit),
+				char__is_digit(Digit)
+			)
+		)),
 		Requests1, Requests).
 
 :- pred create_new_preds(list(request)::in, new_preds::in, new_preds::out,
@@ -733,7 +794,8 @@ create_new_preds([Request | Requests], NewPreds0, NewPreds, PredsToFix0,
 		create_new_pred(Request, NewPred, NextHOid0, NextHOid1,
 						Module0, Module1, IO0, IO1),
 		set__singleton_set(SpecVersions0, NewPred),
-		map__set(NewPreds0, CalledPredProcId, SpecVersions0, NewPreds1)
+		map__det_insert(NewPreds0, CalledPredProcId, SpecVersions0,
+			NewPreds1)
 	),
 	create_new_preds(Requests, NewPreds1, NewPreds, PredsToFix1, PredsToFix,
 			NextHOid1, NextHOid, Module1, Module, IO1, IO).
@@ -747,14 +809,12 @@ create_new_pred(request(_CallingPredProc, CalledPredProc, HOArgs),
 		new_pred(NewPredId, NewProcId, Name, HOArgs), NextHOid0,
 		NextHOid, ModuleInfo0, ModuleInfo, IOState0, IOState) :- 
 	CalledPredProc = proc(CalledPred, _),
-	module_info_get_predicate_table(ModuleInfo0,
-					 PredTable0),
+	module_info_get_predicate_table(ModuleInfo0, PredTable0),
 	predicate_table_get_preds(PredTable0, Preds0),
 	map__lookup(Preds0, CalledPred, PredInfo0),
 	pred_info_name(PredInfo0, Name0),
 	pred_info_arity(PredInfo0, Arity),
 	pred_info_get_is_pred_or_func(PredInfo0, PredOrFunc),
-	module_info_name(ModuleInfo0, ModuleName),
 	pred_info_module(PredInfo0, PredModule),
 	globals__io_lookup_bool_option(very_verbose, VeryVerbose,
 							IOState0, IOState1),
@@ -775,29 +835,22 @@ create_new_pred(request(_CallingPredProc, CalledPredProc, HOArgs),
        	),
 	string__int_to_string(NextHOid0, IdStr),
 	NextHOid is NextHOid0 + 1,
-	( ModuleName = PredModule ->
-		NamePrefix = ""
-	;
-		string__append(PredModule, "__", NamePrefix)
-	),
-	string__append_list([NamePrefix, Name0, "__ho", IdStr], PredName),
+	string__append_list([Name0, "__ho", IdStr], PredName),
 	pred_info_typevarset(PredInfo0, TypeVars),
 	remove_listof_higher_order_args(Types0, 1, HOArgs, Types),
 	pred_info_context(PredInfo0, Context),
-	pred_info_clauses_info(PredInfo0, ClausesInfo),
-	(
-		pred_info_is_inlined(PredInfo0)
-	->
-		Inline = yes
-	;
-		Inline = no
-	),
+	pred_info_get_marker_list(PredInfo0, MarkerList),
 	pred_info_get_goal_type(PredInfo0, GoalType),
-		% *** This will need to be fixed when the condition
-		%       field of the pred_info becomes used
-	Name = qualified(ModuleName, PredName),
-	pred_info_init(ModuleName, Name, Arity, Tvars,
-		Types, true, Context, ClausesInfo, local, Inline, GoalType,
+	Name = qualified(PredModule, PredName),
+	varset__init(EmptyVarSet),
+	map__init(EmptyVarTypes),
+	
+	% This isn't looked at after here, and just clutters up
+	% hlds dumps if it's filled in.
+	ClausesInfo = clauses_info(EmptyVarSet, EmptyVarTypes,
+		EmptyVarTypes, [], []),
+	pred_info_init(PredModule, Name, Arity, Tvars,
+		Types, true, Context, ClausesInfo, local, MarkerList, GoalType,
 		PredOrFunc, PredInfo1),
 	pred_info_set_typevarset(PredInfo1, TypeVars, PredInfo2),
 	pred_info_procedures(PredInfo2, Procs0),
@@ -950,12 +1003,11 @@ create_specialized_versions_2([NewPred | NewPreds], NewPredMap, NewProcInfo0,
 	apply_substitution_to_type_map(VarTypes0, Substitution, VarTypes1),
 	map__apply_to_list(HeadVars, VarTypes1, ArgTypes0),
 	term__vars_list(ArgTypes0, TypeVars),
-	varset__init(ArgTVarset0),
+	varset__init(VarSet0),
 	map__init(DummyVarTypes), % type vars don't have a type
 	map__init(Renaming0),
-	varset__init(OldVarNames),
-	goal_util__create_variables(TypeVars, ArgTVarset0, DummyVarTypes,
-		Renaming0, DummyVarTypes, OldVarNames, ArgTVarset, _, Renaming),
+	goal_util__create_variables(TypeVars, VarSet0, DummyVarTypes,
+		Renaming0, DummyVarTypes, VarSet0, ArgTVarset, _, Renaming),
 	term__apply_variable_renaming_to_list(ArgTypes0, Renaming, ArgTypes),
 	pred_info_set_arg_types(NewPredInfo1, ArgTVarset, ArgTypes,
 							 NewPredInfo2),
@@ -968,7 +1020,9 @@ create_specialized_versions_2([NewPred | NewPreds], NewPredMap, NewProcInfo0,
 					
 	implicitly_quantify_clause_body(HeadVars, Goal2, Varset0, VarTypes1,
 					Goal3, Varset, VarTypes, _),
-	recompute_instmap_delta(Goal3, Goal4, ModuleInfo0, ModuleInfo1),
+	proc_info_get_initial_instmap(NewProcInfo1, ModuleInfo0, InstMap0),
+	recompute_instmap_delta(no, Goal3, Goal4, InstMap0,
+		ModuleInfo0, ModuleInfo1),
 	proc_info_set_goal(NewProcInfo1, Goal4, NewProcInfo1a),
 	proc_info_set_variables(NewProcInfo1a, Varset, NewProcInfo2),
 	proc_info_set_vartypes(NewProcInfo2, VarTypes, NewProcInfo3),
@@ -984,7 +1038,7 @@ create_specialized_versions_2([NewPred | NewPreds], NewPredMap, NewProcInfo0,
 			ModuleInfo2, ModuleInfo).
 
 	
-		% Returns a list of hlds__goals which construct the list of
+		% Returns a list of hlds_goals which construct the list of
 		% higher order arguments which have been specialized. Traverse
 		% goal will then recognize these as having a unique possible
 		% value and will specialize any calls involving them.
@@ -1005,7 +1059,7 @@ create_specialized_versions_2([NewPred | NewPreds], NewPredMap, NewProcInfo0,
 		list(var)::out, list(mode)::in, list(mode)::out,
 		list(higher_order_arg)::in, proc_info::in, proc_info::out,
 		pred_info::in, pred_info::out, tsubst::in,
-		tsubst::out, list(hlds__goal)::out) is det.
+		tsubst::out, list(hlds_goal)::out) is det.
 
 construct_higher_order_terms(_, HeadVars, HeadVars, ArgModes, ArgModes,
 	[], ProcInfo, ProcInfo, PredInfo, PredInfo, Subst, Subst, []).
@@ -1095,18 +1149,17 @@ construct_higher_order_terms(ModuleInfo, HeadVars0, HeadVars, ArgModes0,
 					ModuleInfo, UniModes),
 	Unify = construct(LVar, pred_const(PredId, ProcId),
 					NewHeadVars0, UniModes),
-	proc_info_inferred_determinism(ProcInfo3, Detism),
-	pred_info_get_is_pred_or_func(PredInfo2, PredOrFunc),
+	proc_info_interface_determinism(CalledProcInfo, Detism),
+	pred_info_get_is_pred_or_func(CalledPredInfo, PredOrFunc),
 	PredInstInfo = pred_inst_info(PredOrFunc, UnCurriedArgModes, Detism),
 	Inst = ground(shared, yes(PredInstInfo)),
 	Unimode = (free -> Inst) - (Inst -> Inst),
 	Goal = unify(LVar, Rhs, Unimode, Unify, Context),
 
-	goal_info_init(Info0),
-	goal_info_set_determinism(Info0, Detism, Info1),
+	set__list_to_set([LVar | NewHeadVars0], NonLocals),
 	instmap_delta_init_reachable(InstmapDelta0),
 	instmap_delta_insert(InstmapDelta0, LVar, Inst, InstmapDelta),
-	goal_info_set_instmap_delta(Info1, InstmapDelta, Info),
+	goal_info_init(NonLocals, InstmapDelta, det, Info),
 	construct_higher_order_terms(ModuleInfo, HeadVars1, HeadVars, ArgModes1,
 		ArgModes, HOArgs, ProcInfo3, ProcInfo, PredInfo2, PredInfo,
 		Substitution2, Substitution, Goals1),
