@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1995 University of Melbourne.
+% Copyright (C) 1996-1997 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -13,7 +13,7 @@
 
 :- interface.
 
-:- import_module hlds_pred, llds, prog_data.
+:- import_module hlds_pred, llds, prog_data, (inst).
 :- import_module bool, list, map, varset.
 
 %-----------------------------------------------------------------------------%
@@ -23,7 +23,7 @@
 	% This table is used by the type-checker to look
 	% up the type of functors/constants.
 
-:- type cons_table	==	map(cons_id, list(hlds__cons_defn)).
+:- type cons_table	==	map(cons_id, list(hlds_cons_defn)).
 
 :- type cons_id		--->	cons(sym_name, arity)	% name, arity
 			;	int_const(int)
@@ -40,8 +40,8 @@
 	% A cons_defn is the definition of a constructor (i.e. a constant
 	% or a functor) for a particular type.
 
-:- type hlds__cons_defn
-	--->	hlds__cons_defn(
+:- type hlds_cons_defn
+	--->	hlds_cons_defn(
 			% maybe add tvarset?
 			list(type),		% The types of the arguments
 						% of this functor (if any)
@@ -57,13 +57,18 @@
 
 	% Various predicates for accessing the cons_id type.
 
-	% Given a cons_id, convert it into a const (from term.m) and
-	% an integer arity.  Fails if the cons_id is not representable
-	% as a const (for example, if it is a higher-order pred constant
-	% or an address constant or has a module qualifier).
+	% Given a cons_id and a list of argument terms, convert it into a
+	% term. Fails if the cons_id is a pred_const, code_addr_const or
+	% base_type_info_const.
 
-:- pred cons_id_to_const(cons_id, const, arity).
-:- mode cons_id_to_const(in, out, out) is semidet.
+:- pred cons_id_and_args_to_term(cons_id, list(term), term).
+:- mode cons_id_and_args_to_term(in, in, out) is semidet.
+
+	% Get the arity of a cons_id, aborting on pred_const, code_addr_const
+	% and base_type_info_const.
+
+:- pred cons_id_arity(cons_id, arity).
+:- mode cons_id_arity(in, out) is det.
 
 	% The reverse conversion - make a cons_id for a functor.
 	% Given a const and an arity for the functor, create a cons_id.
@@ -82,20 +87,51 @@
 
 :- implementation.
 
-:- import_module require.
+:- import_module prog_util.
+:- import_module require, std_util, term.
 
-	% Module qualified cons_ids can't be converted to consts.
-cons_id_to_const(cons(unqualified(Name), Arity), term__atom(Name), Arity).
-cons_id_to_const(int_const(Int), term__integer(Int), 0).
-cons_id_to_const(string_const(String), term__string(String), 0).
-cons_id_to_const(float_const(Float), term__float(Float), 0).
+cons_id_and_args_to_term(int_const(Int), [], Term) :-
+	term__context_init(Context),
+	Term = term__functor(term__integer(Int), [], Context).
+cons_id_and_args_to_term(float_const(Float), [], Term) :-
+	term__context_init(Context),
+	Term = term__functor(term__float(Float), [], Context).
+cons_id_and_args_to_term(string_const(String), [], Term) :-
+	term__context_init(Context),
+	Term = term__functor(term__string(String), [], Context).
+cons_id_and_args_to_term(cons(SymName, _Arity), Args, Term) :-
+	construct_qualified_term(SymName, Args, Term).
+
+cons_id_arity(cons(_, Arity), Arity).
+cons_id_arity(int_const(_), 0).
+cons_id_arity(string_const(_), 0).
+cons_id_arity(float_const(_), 0).
+cons_id_arity(pred_const(_, _), _) :-
+	error("cons_id_arity: can't get arity of pred_const").
+cons_id_arity(code_addr_const(_, _), _) :-
+	error("cons_id_arity: can't get arity of code_addr_const").
+cons_id_arity(base_type_info_const(_, _, _), _) :-
+	error("cons_id_arity: can't get arity of base_type_info_const").
 
 make_functor_cons_id(term__atom(Name), Arity, cons(unqualified(Name), Arity)).
 make_functor_cons_id(term__integer(Int), _, int_const(Int)).
 make_functor_cons_id(term__string(String), _, string_const(String)).
 make_functor_cons_id(term__float(Float), _, float_const(Float)).
 
-make_cons_id(SymName, Args, _TypeId, cons(SymName, Arity)) :-
+make_cons_id(SymName0, Args, TypeId, cons(SymName, Arity)) :-
+	(
+		SymName0 = unqualified(ConsName),
+		(
+			TypeId = unqualified(_) - _,
+			SymName = SymName0
+		;
+			TypeId = qualified(TypeModule, _) - _,
+			SymName = qualified(TypeModule, ConsName)
+		)
+	;
+		SymName0 = qualified(_, _),
+		SymName = SymName0
+	),
 	list__length(Args, Arity).
 
 %-----------------------------------------------------------------------------%
@@ -108,50 +144,51 @@ make_cons_id(SymName, Args, _TypeId, cons(SymName, Arity)) :-
 :- type type_id		== 	pair(sym_name, arity).
 				% name, arity
 
-:- type type_table	==	map(type_id, hlds__type_defn).
+:- type type_table	==	map(type_id, hlds_type_defn).
 
 	% This is how type, modes and constructors are represented.
 	% The parts that are not defined here (i.e. type_param, constructor,
 	% type, inst, mode, condition) are represented in the same way as
 	% in prog_io.m, and are defined there.
 
-	% An hlds__type_defn holds the information about a type definition.
+	% An hlds_type_defn holds the information about a type definition.
 
-:- type hlds__type_defn.
+:- type hlds_type_defn.
 
 :- pred hlds_data__set_type_defn(tvarset, list(type_param),
-	hlds__type_body, import_status, term__context, hlds__type_defn).
+	hlds_type_body, import_status, term__context, hlds_type_defn).
 :- mode hlds_data__set_type_defn(in, in, in, in, in, out) is det.
 
-:- pred hlds_data__get_type_defn_tvarset(hlds__type_defn, tvarset).
+:- pred hlds_data__get_type_defn_tvarset(hlds_type_defn, tvarset).
 :- mode hlds_data__get_type_defn_tvarset(in, out) is det.
 
-:- pred hlds_data__get_type_defn_tparams(hlds__type_defn, list(type_param)).
+:- pred hlds_data__get_type_defn_tparams(hlds_type_defn, list(type_param)).
 :- mode hlds_data__get_type_defn_tparams(in, out) is det.
 
-:- pred hlds_data__get_type_defn_body(hlds__type_defn, hlds__type_body).
+:- pred hlds_data__get_type_defn_body(hlds_type_defn, hlds_type_body).
 :- mode hlds_data__get_type_defn_body(in, out) is det.
 
-:- pred hlds_data__get_type_defn_status(hlds__type_defn, import_status).
+:- pred hlds_data__get_type_defn_status(hlds_type_defn, import_status).
 :- mode hlds_data__get_type_defn_status(in, out) is det.
 
-:- pred hlds_data__get_type_defn_context(hlds__type_defn, term__context).
+:- pred hlds_data__get_type_defn_context(hlds_type_defn, term__context).
 :- mode hlds_data__get_type_defn_context(in, out) is det.
 
-:- pred hlds_data__set_type_defn_status(hlds__type_defn, import_status,
-			hlds__type_defn).
+:- pred hlds_data__set_type_defn_status(hlds_type_defn, import_status,
+			hlds_type_defn).
 :- mode hlds_data__set_type_defn_status(in, in, out) is det.
 
-	% An `hlds__type_body' holds the body of a type definition:
+	% An `hlds_type_body' holds the body of a type definition:
 	% du = discriminated union, uu = undiscriminated union,
 	% eqv_type = equivalence type (a type defined to be equivalen
 	% to some other type)
 
-:- type hlds__type_body
+:- type hlds_type_body
 	--->	du_type(
 			list(constructor), 	% the ctors for this type
 			cons_tag_values,	% their tag values
-			bool		% is this type an enumeration?
+			bool,		% is this type an enumeration?
+			maybe(sym_name) % user-defined equality pred
 		)
 	;	uu_type(list(type))	% not yet implemented!
 	;	eqv_type(type)
@@ -237,12 +274,12 @@ make_cons_id(SymName, Args, _TypeId, cons(SymName, Arity)) :-
 
 :- implementation.
 
-:- type hlds__type_defn
-	--->	hlds__type_defn(
+:- type hlds_type_defn
+	--->	hlds_type_defn(
 			tvarset,		% Names of type vars (empty
 						% except for polymorphic types)
 			list(type_param),	% Formal type parameters
-			hlds__type_body,	% The definition of the type
+			hlds_type_body,	% The definition of the type
 
 			import_status,		% Is the type defined in this
 						% module, and if yes, is it
@@ -262,16 +299,16 @@ make_cons_id(SymName, Args, _TypeId, cons(SymName, Arity)) :-
 		).
 
 hlds_data__set_type_defn(Tvarset, Params, Body, Status, Context, Defn) :-
-	Defn = hlds__type_defn(Tvarset, Params, Body, Status, Context).
+	Defn = hlds_type_defn(Tvarset, Params, Body, Status, Context).
 
-hlds_data__get_type_defn_tvarset(hlds__type_defn(Tvarset, _, _, _, _), Tvarset).
-hlds_data__get_type_defn_tparams(hlds__type_defn(_, Params, _, _, _), Params).
-hlds_data__get_type_defn_body(hlds__type_defn(_, _, Body, _, _), Body).
-hlds_data__get_type_defn_status(hlds__type_defn(_, _, _, Status, _), Status).
-hlds_data__get_type_defn_context(hlds__type_defn(_, _, _, _, Context), Context).
+hlds_data__get_type_defn_tvarset(hlds_type_defn(Tvarset, _, _, _, _), Tvarset).
+hlds_data__get_type_defn_tparams(hlds_type_defn(_, Params, _, _, _), Params).
+hlds_data__get_type_defn_body(hlds_type_defn(_, _, Body, _, _), Body).
+hlds_data__get_type_defn_status(hlds_type_defn(_, _, _, Status, _), Status).
+hlds_data__get_type_defn_context(hlds_type_defn(_, _, _, _, Context), Context).
 
-hlds_data__set_type_defn_status(hlds__type_defn(A, B, C, _, E), Status, 
-				hlds__type_defn(A, B, C, Status, E)).
+hlds_data__set_type_defn_status(hlds_type_defn(A, B, C, _, E), Status, 
+				hlds_type_defn(A, B, C, Status, E)).
 
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
@@ -286,7 +323,7 @@ hlds_data__set_type_defn_status(hlds__type_defn(A, B, C, _, E), Status,
 :- type inst_table.
 
 :- type user_inst_table.
-:- type user_inst_defns ==	map(inst_id, hlds__inst_defn).
+:- type user_inst_defns ==	map(inst_id, hlds_inst_defn).
 
 :- type unify_inst_table ==	map(inst_name, maybe_inst_det).
 
@@ -295,7 +332,9 @@ hlds_data__set_type_defn_status(hlds__type_defn(A, B, C, _, E), Status,
 
 :- type merge_inst_table ==	map(pair(inst), maybe_inst).
 
-:- type ground_inst_table == 	map(inst_name, maybe_inst).
+:- type ground_inst_table == 	map(inst_name, maybe_inst_det).
+
+:- type any_inst_table == 	map(inst_name, maybe_inst_det).
 
 :- type shared_inst_table == 	map(inst_name, maybe_inst).
 
@@ -307,17 +346,17 @@ hlds_data__set_type_defn_status(hlds__type_defn(A, B, C, _, E), Status,
 :- type maybe_inst_det	--->	unknown
 			;	known(inst, determinism).
 
-	% An `hlds__inst_defn' holds the information we need to store
+	% An `hlds_inst_defn' holds the information we need to store
 	% about inst definitions such as
 	%	:- inst list_skel(I) = bound([] ; [I | list_skel(I)].
 
-:- type hlds__inst_defn
-	--->	hlds__inst_defn(
+:- type hlds_inst_defn
+	--->	hlds_inst_defn(
 			varset,			% The names of the inst
 						% parameters (if any).
 			list(inst_param),	% The inst parameters (if any).
 						% ([I] in the above example.)
-			hlds__inst_body,	% The definition of this inst.
+			hlds_inst_body,	% The definition of this inst.
 			condition,		% Unused (reserved for
 						% holding a user-defined 
 						% invariant).
@@ -328,7 +367,7 @@ hlds_data__set_type_defn_status(hlds__type_defn(A, B, C, _, E), Status,
 						% whether to output this inst.
 		).
 
-:- type hlds__inst_body
+:- type hlds_inst_body
 	--->	eqv_inst(inst)			% This inst is equivalent to
 						% some other inst.
 	;	abstract_inst.			% This inst is just a forward
@@ -354,6 +393,9 @@ hlds_data__set_type_defn_status(hlds__type_defn(A, B, C, _, E), Status,
 :- pred inst_table_get_ground_insts(inst_table, ground_inst_table).
 :- mode inst_table_get_ground_insts(in, out) is det.
 
+:- pred inst_table_get_any_insts(inst_table, any_inst_table).
+:- mode inst_table_get_any_insts(in, out) is det.
+
 :- pred inst_table_get_shared_insts(inst_table, shared_inst_table).
 :- mode inst_table_get_shared_insts(in, out) is det.
 
@@ -372,6 +414,9 @@ hlds_data__set_type_defn_status(hlds__type_defn(A, B, C, _, E), Status,
 :- pred inst_table_set_ground_insts(inst_table, ground_inst_table, inst_table).
 :- mode inst_table_set_ground_insts(in, in, out) is det.
 
+:- pred inst_table_set_any_insts(inst_table, any_inst_table, inst_table).
+:- mode inst_table_set_any_insts(in, in, out) is det.
+
 :- pred inst_table_set_shared_insts(inst_table, shared_inst_table, inst_table).
 :- mode inst_table_set_shared_insts(in, in, out) is det.
 
@@ -385,7 +430,7 @@ hlds_data__set_type_defn_status(hlds__type_defn(A, B, C, _, E), Status,
 :- pred user_inst_table_get_inst_ids(user_inst_table, list(inst_id)).
 :- mode user_inst_table_get_inst_ids(in, out) is det.
 
-:- pred user_inst_table_insert(user_inst_table, inst_id, hlds__inst_defn,
+:- pred user_inst_table_insert(user_inst_table, inst_id, hlds_inst_defn,
 					user_inst_table).
 :- mode user_inst_table_insert(in, in, in, out) is semidet.
 
@@ -402,6 +447,7 @@ hlds_data__set_type_defn_status(hlds__type_defn(A, B, C, _, E), Status,
 			unify_inst_table,
 			merge_inst_table,
 			ground_inst_table,
+			any_inst_table,
 			shared_inst_table,
 			mostly_uniq_inst_table
 		).
@@ -416,47 +462,56 @@ hlds_data__set_type_defn_status(hlds__type_defn(A, B, C, _, E), Status,
 		).
 
 inst_table_init(inst_table(UserInsts, UnifyInsts, MergeInsts, GroundInsts,
-				SharedInsts, NondetLiveInsts)) :-
+			AnyInsts, SharedInsts, NondetLiveInsts)) :-
 	map__init(UserInstDefns),
 	UserInsts = user_inst_table(UserInstDefns, []),
 	map__init(UnifyInsts),
 	map__init(MergeInsts),
 	map__init(GroundInsts),
 	map__init(SharedInsts),
+	map__init(AnyInsts),
 	map__init(NondetLiveInsts).
 
-inst_table_get_user_insts(inst_table(UserInsts, _, _, _, _, _), UserInsts).
+inst_table_get_user_insts(inst_table(UserInsts, _, _, _, _, _, _), UserInsts).
 
-inst_table_get_unify_insts(inst_table(_, UnifyInsts, _, _, _, _), UnifyInsts).
+inst_table_get_unify_insts(inst_table(_, UnifyInsts, _, _, _, _, _),
+			UnifyInsts).
 
-inst_table_get_merge_insts(inst_table(_, _, MergeInsts, _, _, _), MergeInsts).
+inst_table_get_merge_insts(inst_table(_, _, MergeInsts, _, _, _, _),
+			MergeInsts).
 
-inst_table_get_ground_insts(inst_table(_, _, _, GroundInsts, _, _),
+inst_table_get_ground_insts(inst_table(_, _, _, GroundInsts, _, _, _),
 			GroundInsts).
 
-inst_table_get_shared_insts(inst_table(_, _, _, _, SharedInsts, _),
+inst_table_get_any_insts(inst_table(_, _, _, _, AnyInsts, _, _), AnyInsts).
+
+inst_table_get_shared_insts(inst_table(_, _, _, _, _, SharedInsts, _),
 			SharedInsts).
 
-inst_table_get_mostly_uniq_insts(inst_table(_, _, _, _, _, NondetLiveInsts),
+inst_table_get_mostly_uniq_insts(inst_table(_, _, _, _, _, _, NondetLiveInsts),
 			NondetLiveInsts).
 
-inst_table_set_user_insts(inst_table(_, B, C, D, E, F), UserInsts,
-			inst_table(UserInsts, B, C, D, E, F)).
+inst_table_set_user_insts(inst_table(_, B, C, D, E, F, G), UserInsts,
+			inst_table(UserInsts, B, C, D, E, F, G)).
 
-inst_table_set_unify_insts(inst_table(A, _, C, D, E, F), UnifyInsts,
-			inst_table(A, UnifyInsts, C, D, E, F)).
+inst_table_set_unify_insts(inst_table(A, _, C, D, E, F, G), UnifyInsts,
+			inst_table(A, UnifyInsts, C, D, E, F, G)).
 
-inst_table_set_merge_insts(inst_table(A, B, _, D, E, F), MergeInsts,
-			inst_table(A, B, MergeInsts, D, E, F)).
+inst_table_set_merge_insts(inst_table(A, B, _, D, E, F, G), MergeInsts,
+			inst_table(A, B, MergeInsts, D, E, F, G)).
 
-inst_table_set_ground_insts(inst_table(A, B, C, _, E, F), GroundInsts,
-			inst_table(A, B, C, GroundInsts, E, F)).
+inst_table_set_ground_insts(inst_table(A, B, C, _, E, F, G), GroundInsts,
+			inst_table(A, B, C, GroundInsts, E, F, G)).
 
-inst_table_set_shared_insts(inst_table(A, B, C, D, _, F), SharedInsts,
-			inst_table(A, B, C, D, SharedInsts, F)).
+inst_table_set_any_insts(inst_table(A, B, C, D, _, F, G), AnyInsts,
+			inst_table(A, B, C, D, AnyInsts, F, G)).
 
-inst_table_set_mostly_uniq_insts(inst_table(A, B, C, D, E, _), NondetLiveInsts,
-			inst_table(A, B, C, D, E, NondetLiveInsts)).
+inst_table_set_shared_insts(inst_table(A, B, C, D, E, _, G), SharedInsts,
+			inst_table(A, B, C, D, E, SharedInsts, G)).
+
+inst_table_set_mostly_uniq_insts(inst_table(A, B, C, D, E, F, _),
+			NondetLiveInsts,
+			inst_table(A, B, C, D, E, F, NondetLiveInsts)).
 
 user_inst_table_get_inst_defns(user_inst_table(InstDefns, _), InstDefns).
 
@@ -482,9 +537,9 @@ user_inst_table_optimize(user_inst_table(InstDefns0, InstIds0),
 				% name, arity
 
 :- type mode_table.
-:- type mode_defns	 ==	map(mode_id, hlds__mode_defn).
+:- type mode_defns	 ==	map(mode_id, hlds_mode_defn).
 
-	% A hlds__mode_defn stores the information about a mode
+	% A hlds_mode_defn stores the information about a mode
 	% definition such as
 	%	:- mode out :: free -> ground.
 	% or
@@ -492,15 +547,15 @@ user_inst_table_optimize(user_inst_table(InstDefns0, InstIds0),
 	% or
 	%	:- mode in_list_skel :: in(list_skel).
 
-:- type hlds__mode_defn
-	--->	hlds__mode_defn(
+:- type hlds_mode_defn
+	--->	hlds_mode_defn(
 			varset,			% The names of the inst
 						% parameters (if any).
 			list(inst_param),	% The list of the inst
 						% parameters (if any).
 						% (e.g. [I] for the second
 						% example above.)
-			hlds__mode_body,	% The definition of this mode.
+			hlds_mode_body,	% The definition of this mode.
 			condition,		% Unused (reserved for
 						% holding a user-defined
 						% invariant).
@@ -514,11 +569,11 @@ user_inst_table_optimize(user_inst_table(InstDefns0, InstIds0),
 
 	% The only sort of mode definitions allowed are equivalence modes.
 
-:- type hlds__mode_body
+:- type hlds_mode_body
 	--->	eqv_mode(mode).		% This mode is equivalent to some
 					% other mode.
 
-	% Given a mode table get the mode_id - hlds__mode_defn map.
+	% Given a mode table get the mode_id - hlds_mode_defn map.
 :- pred mode_table_get_mode_defns(mode_table, mode_defns).
 :- mode mode_table_get_mode_defns(in, out) is det.
 
@@ -526,9 +581,9 @@ user_inst_table_optimize(user_inst_table(InstDefns0, InstIds0),
 :- pred mode_table_get_mode_ids(mode_table, list(mode_id)).
 :- mode mode_table_get_mode_ids(in, out) is det.
 
-	% Insert a mode_id and corresponding hlds__mode_defn into the
+	% Insert a mode_id and corresponding hlds_mode_defn into the
 	% mode_table. Fail if the mode_id is already present in the table.
-:- pred mode_table_insert(mode_table, mode_id, hlds__mode_defn, mode_table).
+:- pred mode_table_insert(mode_table, mode_id, hlds_mode_defn, mode_table).
 :- mode mode_table_insert(in, in, in, out) is semidet.
 
 :- pred mode_table_init(mode_table).
