@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1995 University of Melbourne.
+% Copyright (C) 1996-1997 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -13,8 +13,9 @@
 
 :- interface.
 
-:- import_module hlds_data, hlds_goal, hlds_module, prog_data, instmap.
-:- import_module bool, list, map, std_util, varset.
+:- import_module hlds_data, hlds_goal, hlds_module, llds, prog_data, instmap.
+:- import_module bool, list, map, std_util, term, varset.
+:- import_module term_util.
 
 :- implementation.
 
@@ -25,12 +26,49 @@
 
 :- interface.
 
-	% A proc_id is a mode number within a particular predicate -
+	% A proc_id is the name of a mode within a particular predicate -
 	% not to be confused with a mode_id, which is the name of a
 	% user-defined mode.
 
-:- type proc_id		==	int.
-:- type pred_id		==	int.
+:- type pred_id.
+:- type proc_id.
+
+:- pred hlds_pred__initial_pred_id(pred_id).
+:- mode hlds_pred__initial_pred_id(out) is det.
+
+:- pred hlds_pred__initial_proc_id(proc_id).
+:- mode hlds_pred__initial_proc_id(out) is det.
+
+:- pred hlds_pred__next_pred_id(pred_id, pred_id).
+:- mode hlds_pred__next_pred_id(in, out) is det.
+
+:- pred hlds_pred__next_proc_id(proc_id, proc_id).
+:- mode hlds_pred__next_proc_id(in, out) is det.
+
+:- pred pred_id_to_int(pred_id, int).
+:- mode pred_id_to_int(in, out) is det.
+:- mode pred_id_to_int(out, in) is det.
+
+:- pred proc_id_to_int(proc_id, int).
+:- mode proc_id_to_int(in, out) is det.
+:- mode proc_id_to_int(out, in) is det.
+
+	% For semidet complicated unifications with mode (in, in),
+	% these are defined to have the same proc_id (0).  This
+	% returns that proc_id.
+
+:- pred hlds_pred__in_in_unification_proc_id(proc_id).
+:- mode hlds_pred__in_in_unification_proc_id(out) is det.
+
+        % Return an invalid pred_id. Used to initialize the pred_id
+        % in call(...) goals before we do typechecking or when type-checking
+        % finds that there was no predicate which matched the call.
+
+:- pred invalid_pred_id(pred_id).
+:- mode invalid_pred_id(out) is det.
+
+:- pred invalid_proc_id(proc_id).
+:- mode invalid_proc_id(out) is det.
 
 :- type pred_info.
 :- type proc_info.
@@ -42,6 +80,13 @@
 :- type pred_proc_id	--->	proc(pred_id, proc_id).
 :- type pred_proc_list	==	list(pred_proc_id).
 
+	% The clauses_info structure contains the clauses for a predicate
+	% after conversion from the item_list by make_hlds.m.
+	% Typechecking is performed on the clauses info, then the clauses
+	% are copied to create the proc_info for each procedure.
+	% After mode analysis the clauses and the procedure goals are not
+	% guaranteed to be the same, and the clauses are only kept so that
+	% the optimized goal can be compared with the original in HLDS dumps.
 :- type clauses_info	--->	clauses_info(
 					varset,		% variable names
 					map(var, type), % variable types from
@@ -55,7 +100,7 @@
 :- type clause		--->	clause(
 					list(proc_id),	% modes for which
 							% this clause applies
-					hlds__goal,	% Body
+					hlds_goal,	% Body
 					term__context
 				).
 
@@ -88,17 +133,12 @@
 	% The type `import_status' describes whether an entity (a predicate,
 	% type, inst, or mode) is local to the current module, exported from
 	% the current module, or imported from some other module.
-	% Only predicates can have status opt_decl, pseudo_exported
-	% or pseudo_imported.
+	% Only predicates can have status pseudo_exported or pseudo_imported.
 	% Only types can have status abstract_exported or abstract_imported.
 
 :- type import_status
 	--->	imported	% defined in the interface of some other module
 				% or `external' (in some other language)
-	;	opt_decl	% predicate declared in an optimization
-				% interface -  this is needed to identify 
-				% predicates that must be ignored when
-				% processing local predicates
 	;	opt_imported	% defined in the optimization 
 				% interface of another module
 	;	abstract_imported % describes a type with only an abstract
@@ -160,16 +200,47 @@
 				% Since the transformation affects *other*
 				% predicates, the `done' status is not
 				% meaningful
+	;	no_inline	% Requests that this be predicate not be 
+				% inlined.
+				% Used for pragma(no_inline).
+				% Conflicts with `inline' marker.
 	;	dnf		% Requests that this predicate be transformed
 				% into disjunctive normal form.
 				% Used for pragma(memo).
 	;	magic		% Requests that this predicate be transformed
 				% using the magic set transformation
 				% Used for pragma(memo).
-	;	memo.		% Requests that this predicate be evaluated
+	;	memo		% Requests that this predicate be evaluated
 				% using memoing.
 				% Used for pragma(memo).
 
+				% The terminates and does_not_terminate
+				% pragmas are kept as markers to ensure
+				% that conflicting declarations are not
+				% made by the user.  Otherwise, the
+				% information could be added to the
+				% ProcInfos directly.
+	;	terminates	% The user guarantees that this predicate
+				% will terminate for all (finite?) input
+				% The `done' status could be meaningful,
+				% but it is currently not used.
+	;	does_not_terminate
+				% States that this predicate does not
+				% terminate.  This is useful for pragma c
+				% code, which the compiler assumes to be
+				% terminating.
+				% The `done' status could be meaningful,
+				% but it is currently not used.
+	;	check_termination
+				% The user requires the compiler to guarantee
+				% the termination of this predicate.
+				% If the compiler cannot guarantee termination
+				% then it must give an error message.
+				% The `done' status could be meaningful,
+				% but it is currently not used.
+	.
+	
+
 :- type marker_status
 	--->	request(marker)
 	;	done(marker).
@@ -179,7 +250,7 @@
 
 :- pred pred_info_init(module_name, sym_name, arity, tvarset, list(type),
 	condition, term__context, clauses_info, import_status,
-	bool, goal_type, pred_or_func, pred_info).
+	list(marker_status), goal_type, pred_or_func, pred_info).
 :- mode pred_info_init(in, in, in, in, in, in, in, in, in, in, in, in, out)
 	is det.
 
@@ -201,6 +272,8 @@
 :- pred pred_info_name(pred_info, string).
 :- mode pred_info_name(in, out) is det.
 
+	% pred_info_arity returns the arity of the predicate
+	% *not* counting inserted type_info arguments for polymorphic preds.
 :- pred pred_info_arity(pred_info, arity).
 :- mode pred_info_arity(in, out) is det.
 
@@ -275,8 +348,14 @@
 	% to inline a predicate even if there was no pragma(inline, ...)
 	% declaration for that predicate.
 
-:- pred pred_info_is_inlined(pred_info).
-:- mode pred_info_is_inlined(in) is semidet.
+:- pred pred_info_requested_inlining(pred_info).
+:- mode pred_info_requested_inlining(in) is semidet.
+
+	% Succeeds if there was a `:- pragma(no_inline, ...)' declaration
+	% for this predicate.
+
+:- pred pred_info_requested_no_inlining(pred_info).
+:- mode pred_info_requested_no_inlining(in) is semidet.
 
 :- pred pred_info_get_marker_list(pred_info, list(marker_status)).
 :- mode pred_info_get_marker_list(in, out) is det.
@@ -291,9 +370,35 @@
 
 :- implementation.
 
+:- type pred_id		==	int.
+:- type proc_id		==	int.
+
+hlds_pred__initial_pred_id(0).
+
+hlds_pred__initial_proc_id(0).
+
+hlds_pred__next_pred_id(PredId, NextPredId) :-
+	NextPredId is PredId + 1.
+
+hlds_pred__next_proc_id(ProcId, NextProcId) :-
+	NextProcId is ProcId + 1.
+
+pred_id_to_int(PredId, PredId).
+
+proc_id_to_int(ProcId, ProcId).
+
+hlds_pred__in_in_unification_proc_id(0).
+
+invalid_pred_id(-1).
+
+invalid_proc_id(-1).
+
 	% The information specific to a predicate, as opposed to a procedure.
+	%
 	% Any changes in this type definition will almost certainly require
 	% corresponding changes in define.m.
+	% XXX: This comment is either ancient, or prophesy. It seems
+	% define.m doesn't exist yet.
 
 :- type pred_info
 	--->	predicate(
@@ -313,6 +418,8 @@
 			module_name,	% module in which pred occurs
 			string,		% predicate name
 			arity,		% the arity of the pred
+					% (*not* counting any inserted
+					% type_info arguments)
 			import_status,
 			tvarset,	% names of type vars
 					% in the predicate's type decl
@@ -328,15 +435,10 @@
 		).
 
 pred_info_init(ModuleName, SymName, Arity, TypeVarSet, Types, Cond, Context,
-		ClausesInfo, Status, Inline, GoalType, PredOrFunc, PredInfo) :-
+		ClausesInfo, Status, Markers, GoalType, PredOrFunc, PredInfo) :-
 	map__init(Procs),
 	unqualify_name(SymName, PredName),
 	sym_name_get_module_name(SymName, ModuleName, PredModuleName),
-	( Inline = yes ->
-		Markers = [request(inline)]
-	;
-		Markers = []
-	),
 	PredInfo = predicate(TypeVarSet, Types, Cond, ClausesInfo, Procs,
 		Context, PredModuleName, PredName, Arity, Status, TypeVarSet, 
 		GoalType, Markers, PredOrFunc).
@@ -347,7 +449,7 @@ pred_info_create(ModuleName, SymName, TypeVarSet, Types, Cond, Context,
 	map__init(Procs0),
 	proc_info_declared_determinism(ProcInfo, MaybeDetism),
 	next_mode_id(Procs0, MaybeDetism, ProcId),
-	map__set(Procs0, ProcId, ProcInfo, Procs),
+	map__det_insert(Procs0, ProcId, ProcInfo, Procs),
 	list__length(Types, Arity),
 	proc_info_variables(ProcInfo, VarSet),
 	proc_info_vartypes(ProcInfo, VarTypes),
@@ -372,7 +474,7 @@ pred_info_procids(PredInfo, ProcIds) :-
 
 pred_info_non_imported_procids(PredInfo, ProcIds) :-
 	pred_info_import_status(PredInfo, ImportStatus),
-	( ( ImportStatus = imported ; ImportStatus = opt_decl ) ->
+	( ImportStatus = imported ->
 		ProcIds = []
 	; ImportStatus = pseudo_imported ->
 		pred_info_procids(PredInfo, ProcIds0),
@@ -432,14 +534,7 @@ pred_info_import_status(PredInfo, ImportStatus) :-
 				_).
 
 pred_info_is_imported(PredInfo) :-
-	pred_info_import_status(PredInfo, ImportStatus),
-	(
-		ImportStatus = imported
-	;
-		% opt_decl is equivalent to imported, except only 
-		% opt_imported preds can call an opt_decl pred.
-		ImportStatus = opt_decl
-	).
+	pred_info_import_status(PredInfo, imported).
 
 pred_info_is_pseudo_imported(PredInfo) :-
 	pred_info_import_status(PredInfo, ImportStatus),
@@ -476,10 +571,14 @@ pred_info_set_goal_type(PredInfo0, GoalType, PredInfo) :-
 	PredInfo0 = predicate(A, B, C, D, E, F, G, H, I, J, K, _, M, N),
 	PredInfo  = predicate(A, B, C, D, E, F, G, H, I, J, K, GoalType, M, N).
 
-pred_info_is_inlined(PredInfo0) :-
+pred_info_requested_inlining(PredInfo0) :-
 	pred_info_get_marker_list(PredInfo0, Markers),
 	list__member(request(inline), Markers).
 
+pred_info_requested_no_inlining(PredInfo0) :-
+	pred_info_get_marker_list(PredInfo0, Markers),
+	list__member(request(no_inline), Markers).
+
 pred_info_get_marker_list(PredInfo, Markers) :-
 	PredInfo = predicate(_, _, _, _, _, _, _, _, _, _, _, _, Markers, _).
 
@@ -498,23 +597,23 @@ pred_info_get_is_pred_or_func(PredInfo, IsPredOrFunc) :-
 
 :- interface.
 
-:- pred proc_info_init(arity, list(mode), maybe(list(is_live)),
-		maybe(determinism), term__context, proc_info).
-:- mode proc_info_init(in, in, in, in, in, out) is det.
+:- pred proc_info_init(arity, list(mode), maybe(list(mode)),
+	maybe(list(is_live)), maybe(determinism), term__context, proc_info).
+:- mode proc_info_init(in, in, in, in, in, in, out) is det.
 
 :- pred proc_info_set(maybe(determinism), varset, map(var, type), list(var),
-	list(mode), maybe(list(is_live)), hlds__goal, term__context,
+	list(mode), maybe(list(is_live)), hlds_goal, term__context,
 	stack_slots, determinism, bool, list(arg_info), liveness_info,
-	map(tvar, var), proc_info).
+	map(tvar, var), termination, proc_info).
 :- mode proc_info_set(in, in, in, in, in, in, in, in, in, in, in, in, in, in,
-	out) is det.
+	in, out) is det.
 
 :- pred proc_info_create(varset, map(var, type), list(var), list(mode),
-	determinism, hlds__goal, term__context, map(tvar, var), proc_info).
+	determinism, hlds_goal, term__context, map(tvar, var), proc_info).
 :- mode proc_info_create(in, in, in, in, in, in, in, in, out) is det.
 
 :- pred proc_info_set_body(proc_info, varset, map(var, type), list(var),
-	hlds__goal, proc_info).
+	hlds_goal, proc_info).
 :- mode proc_info_set_body(in, in, in, in, in, out) is det.
 
 :- pred proc_info_declared_determinism(proc_info, maybe(determinism)).
@@ -572,7 +671,7 @@ pred_info_get_is_pred_or_func(PredInfo, IsPredOrFunc) :-
 					proc_info).
 :- mode proc_info_set_maybe_arglives(in, in, out) is det.
 
-:- pred proc_info_goal(proc_info, hlds__goal).
+:- pred proc_info_goal(proc_info, hlds_goal).
 :- mode proc_info_goal(in, out) is det.
 
 :- pred proc_info_context(proc_info, term__context).
@@ -590,7 +689,7 @@ pred_info_get_is_pred_or_func(PredInfo, IsPredOrFunc) :-
 :- pred proc_info_set_inferred_determinism(proc_info, determinism, proc_info).
 :- mode proc_info_set_inferred_determinism(in, in, out) is det.
 
-:- pred proc_info_set_goal(proc_info, hlds__goal, proc_info).
+:- pred proc_info_set_goal(proc_info, hlds_goal, proc_info).
 :- mode proc_info_set_goal(in, in, out) is det.
 
 :- pred proc_info_arg_info(proc_info, list(arg_info)).
@@ -608,6 +707,12 @@ pred_info_get_is_pred_or_func(PredInfo, IsPredOrFunc) :-
 :- pred proc_info_set_stack_slots(proc_info, stack_slots, proc_info).
 :- mode proc_info_set_stack_slots(in, in, out) is det.
 
+:- pred proc_info_termination(proc_info, termination).
+:- mode proc_info_termination(in, out) is det.
+
+:- pred proc_info_set_termination(proc_info, termination, proc_info).
+:- mode proc_info_set_termination(in, in, out) is det.
+
 :- pred proc_info_set_can_process(proc_info, bool, proc_info).
 :- mode proc_info_set_can_process(in, in, out) is det.
 
@@ -617,6 +722,12 @@ pred_info_get_is_pred_or_func(PredInfo, IsPredOrFunc) :-
 :- pred proc_info_set_typeinfo_varmap(proc_info, map(tvar, var), proc_info).
 :- mode proc_info_set_typeinfo_varmap(in, in, out) is det.
 
+:- pred proc_info_maybe_declared_argmodes(proc_info, maybe(list(mode))).
+:- mode proc_info_maybe_declared_argmodes(in, out) is det.
+
+:- pred proc_info_declared_argmodes(proc_info, list(mode)).
+:- mode proc_info_declared_argmodes(in, out) is det.
+
 	% For a set of variables V, find all the type variables in the types 
 	% of the variables in V, and return set of typeinfo variables for 
 	% those type variables. (find all typeinfos for variables in V).
@@ -625,8 +736,8 @@ pred_info_get_is_pred_or_func(PredInfo, IsPredOrFunc) :-
 	% for accurate garbage collection - live variables need to have 
 	% their typeinfos stay live too.
 
-:- pred proc_info_get_used_typeinfos_setwise(proc_info, set(var), set(var)).
-:- mode proc_info_get_used_typeinfos_setwise(in, in, out) is det.
+:- pred proc_info_get_typeinfo_vars_setwise(proc_info, set(var), set(var)).
+:- mode proc_info_get_typeinfo_vars_setwise(in, in, out) is det.
 
 :- pred proc_info_ensure_unique_names(proc_info, proc_info).
 :- mode proc_info_ensure_unique_names(in, out) is det.
@@ -645,12 +756,12 @@ pred_info_get_is_pred_or_func(PredInfo, IsPredOrFunc) :-
 			maybe(list(is_live)),
 					% liveness (in the mode analysis sense)
 					% of the arguments
-			hlds__goal,	% Body
+			hlds_goal,	% Body
 			term__context,	% The context of the `:- mode' decl
 					% (or the context of the first clause,
 					% if there was no mode declaration).
 			stack_slots,	% stack allocations
-			determinism,	% _inferred_ det'ism
+			determinism,	% _inferred_ determinism
 			bool,		% no if we must not process this
 					% procedure yet (used to delay
 					% mode checking etc. for complicated
@@ -664,8 +775,13 @@ pred_info_get_is_pred_or_func(PredInfo, IsPredOrFunc) :-
 					% should be passed.
 			liveness_info,	% the initial liveness,
 					% for code generation
-			map(tvar, var)	% typeinfo vars for
+			map(tvar, var),	% typeinfo vars for
 					% type parameters
+			termination,	% The termination properties of the
+					% procedure.  Initially 'not_set'.
+					% Final value inferred by termination.m
+			maybe(list(mode))
+					% declared modes of arguments.
 		).
 
 
@@ -676,7 +792,8 @@ pred_info_get_is_pred_or_func(PredInfo, IsPredOrFunc) :-
 	% This is what `det_analysis.m' wants. det_analysis.m
 	% will later provide the correct inferred determinism for it.
 
-proc_info_init(Arity, Modes, MaybeArgLives, MaybeDet, MContext, NewProc) :-
+proc_info_init(Arity, Modes, DeclaredModes, MaybeArgLives,
+		MaybeDet, MContext, NewProc) :-
 	map__init(BodyTypes),
 	goal_info_init(GoalInfo),
 	varset__init(BodyVarSet0),
@@ -689,35 +806,37 @@ proc_info_init(Arity, Modes, MaybeArgLives, MaybeDet, MContext, NewProc) :-
 	ClauseBody = conj([]) - GoalInfo,
 	CanProcess = yes,
 	map__init(TVarsMap),
+	term_util__init(Termination),
 	NewProc = procedure(
 		MaybeDet, BodyVarSet, BodyTypes, HeadVars, Modes, MaybeArgLives,
 		ClauseBody, MContext, StackSlots, InferredDet, CanProcess,
-		ArgInfo, InitialLiveness, TVarsMap
+		ArgInfo, InitialLiveness, TVarsMap, Termination, DeclaredModes
 	).
 
 proc_info_set(DeclaredDetism, BodyVarSet, BodyTypes, HeadVars, HeadModes,
 		HeadLives, Goal,
 		Context, StackSlots, InferredDetism, CanProcess,
-		ArgInfo, Liveness, TVarMap, ProcInfo) :-
+		ArgInfo, Liveness, TVarMap, Termination, ProcInfo) :-
 	ProcInfo = procedure(
 		DeclaredDetism, BodyVarSet, BodyTypes, HeadVars, HeadModes,
 		HeadLives, Goal, Context, StackSlots, InferredDetism,
-		CanProcess, ArgInfo, Liveness, TVarMap).
+		CanProcess, ArgInfo, Liveness, TVarMap, Termination, no).
 
 proc_info_create(VarSet, VarTypes, HeadVars, HeadModes, Detism, Goal,
 		Context, TVarMap, ProcInfo) :-
 	map__init(StackSlots),
 	set__init(Liveness),
+	term_util__init(Termination),
 	MaybeHeadLives = no,
 	ProcInfo = procedure(yes(Detism), VarSet, VarTypes, HeadVars, HeadModes,
 		MaybeHeadLives, Goal, Context, StackSlots, Detism, yes, [],
-		Liveness, TVarMap).
+		Liveness, TVarMap, Termination, no).
 
 proc_info_set_body(ProcInfo0, VarSet, VarTypes, HeadVars, Goal, ProcInfo) :-
 	ProcInfo0 = procedure(A, _, _, _, E, F, _,
-		H, I, J, K, L, M, N),
+		H, I, J, K, L, M, N, O, P),
 	ProcInfo = procedure(A, VarSet, VarTypes, HeadVars, E, F, Goal,
-		H, I, J, K, L, M, N).
+		H, I, J, K, L, M, N, O, P).
 
 proc_info_interface_determinism(ProcInfo, Determinism) :-
 	proc_info_declared_determinism(ProcInfo, MaybeDeterminism),
@@ -762,43 +881,58 @@ proc_info_get_initial_instmap(ProcInfo, ModuleInfo, InstMap) :-
 	proc_info_headvars(ProcInfo, HeadVars),
 	proc_info_argmodes(ProcInfo, ArgModes),
 	mode_list_get_initial_insts(ArgModes, ModuleInfo, InitialInsts),
-/***********
-		% propagate type information into the modes
-	proc_info_vartypes(ProcInfo, VarTypes),
-	propagate_type_info_inst_list(VarTypes, ModuleInfo, InitialInsts0,
-					InitialInsts),
-***********/
 	assoc_list__from_corresponding_lists(HeadVars, InitialInsts, InstAL),
 	instmap__from_assoc_list(InstAL, InstMap).
 
 proc_info_declared_determinism(ProcInfo, Detism) :-
-	ProcInfo = procedure(Detism, _, _, _, _, _, _, _, _, _, _, _, _, _).
+    ProcInfo = procedure(Detism, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _).
 proc_info_variables(ProcInfo, VarSet) :-
-	ProcInfo = procedure(_, VarSet, _, _, _, _, _, _, _, _, _, _, _, _).
+    ProcInfo = procedure(_, VarSet, _, _, _, _, _, _, _, _, _, _, _, _, _, _).
 proc_info_vartypes(ProcInfo, VarTypes) :-
-	ProcInfo = procedure(_, _, VarTypes, _, _, _, _, _, _, _, _, _, _, _).
+    ProcInfo = procedure(_, _, VarTypes, _, _, _, _, _, 
+    		_, _, _, _, _, _, _, _).
 proc_info_headvars(ProcInfo, HeadVars) :-
-	ProcInfo = procedure(_, _, _, HeadVars, _, _, _, _, _, _, _, _, _, _).
+    ProcInfo = procedure(_, _, _, HeadVars, _, _, _, _, 
+    		_, _, _, _, _, _, _, _).
 proc_info_argmodes(ProcInfo, Modes) :-
-	ProcInfo = procedure(_, _, _, _, Modes, _, _, _, _, _, _, _, _, _).
+    ProcInfo = procedure(_, _, _, _, Modes, _, _, _, _, _, _, _, _, _, _, _).
 proc_info_maybe_arglives(ProcInfo, ArgLives) :-
-	ProcInfo = procedure(_, _, _, _, _, ArgLives, _, _, _, _, _, _, _, _).
+    ProcInfo = procedure(_, _, _, _, _, ArgLives, 
+    		_, _, _, _, _, _, _, _, _, _).
 proc_info_goal(ProcInfo, Goal) :-
-	ProcInfo = procedure(_, _, _, _, _, _, Goal, _, _, _, _, _, _, _).
+    ProcInfo = procedure(_, _, _, _, _, _, Goal, _, _, _, _, _, _, _, _, _).
 proc_info_context(ProcInfo, Context) :-
-	ProcInfo = procedure(_, _, _, _, _, _, _, Context, _, _, _, _, _, _).
+    ProcInfo = procedure(_, _, _, _, _, _, _, Context, _, _, _, _, _, _, _, _).
 proc_info_stack_slots(ProcInfo, StackSlots) :-
-	ProcInfo = procedure(_, _, _, _, _, _, _, _, StackSlots, _, _, _, _, _).
+    ProcInfo = procedure(_, _, _, _, _, _, _, _, StackSlots, 
+    		_, _, _, _, _, _, _).
 proc_info_inferred_determinism(ProcInfo, Detism) :-
-	ProcInfo = procedure(_, _, _, _, _, _, _, _, _, Detism, _, _, _, _).
+    ProcInfo = procedure(_, _, _, _, _, _, _, _, _, Detism, _, _, _, _, _, _).
 proc_info_can_process(ProcInfo, CanProcess) :-
- 	ProcInfo = procedure(_, _, _, _, _, _, _, _, _, _, CanProcess, _, _, _).
-proc_info_arg_info(ProcInfo, ArgInfo) :-
-	ProcInfo = procedure(_, _, _, _, _, _, _, _, _, _, _, ArgInfo, _, _).
+    ProcInfo = procedure(_, _, _, _, _, _, _, _, _, _, CanProcess, 
+    		_, _, _, _, _).
+proc_info_arg_info(ProcInfo, ArgInfo) :- 
+    ProcInfo = procedure(_, _, _, _, _, _, _, _, _, _, _, ArgInfo, 
+    		_, _, _, _).
 proc_info_liveness_info(ProcInfo, Liveness) :-
-	ProcInfo = procedure(_, _, _, _, _, _, _, _, _, _, _, _, Liveness, _).
+    ProcInfo = procedure(_, _, _, _, _, _, _, _, _, _, _, _, Liveness, 
+    		_, _, _).
 proc_info_typeinfo_varmap(ProcInfo, TVarMap) :-
-	ProcInfo = procedure(_, _, _, _, _, _, _, _, _, _, _, _, _, TVarMap).
+    ProcInfo = procedure(_, _, _, _, _, _, _, _, _, _, _, _, _, TVarMap, _, _).
+proc_info_termination(ProcInfo, Termination) :-
+    ProcInfo = procedure(_, _, _, _, _, _, _, _, _, _, _, _, _, _, 
+    		Termination, _).
+proc_info_maybe_declared_argmodes(ProcInfo, MaybeArgModes) :-
+    ProcInfo = procedure(_, _, _, _, _, _, _,
+		_, _, _, _, _, _, _, _, MaybeArgModes).
+
+proc_info_declared_argmodes(ProcInfo, ArgModes) :-
+	proc_info_maybe_declared_argmodes(ProcInfo, MaybeArgModes),
+	( MaybeArgModes = yes(ArgModes1) ->
+		ArgModes = ArgModes1
+	;
+		proc_info_argmodes(ProcInfo, ArgModes)
+	).
 
 % :- type proc_info	--->	procedure(
 % 				A	maybe(determinism),% _declared_ detism
@@ -807,7 +941,7 @@ proc_info_typeinfo_varmap(ProcInfo, TVarMap) :-
 % 				D	list(var),	% head vars
 % 				E	list(mode), 	% modes of args
 % 				F	list(is_live), 	% liveness of args
-% 				G	hlds__goal,	% Body
+% 				G	hlds_goal,	% Body
 % 				H	term__context,	% The context of
 % 							% the :- mode decl,
 % 							% not the clause.
@@ -821,71 +955,84 @@ proc_info_typeinfo_varmap(ProcInfo, TVarMap) :-
 % 				M	liveness_info	% the initial liveness
 %				N	map(tvar, var)  % typeinfo vars to
 %							% vars.
+%				O	termination	% Termination analys
+%				P	maybe(list(mode)) % declared modes
+%							% of args
 % 				).
 
 proc_info_set_varset(ProcInfo0, VarSet, ProcInfo) :-
-	ProcInfo0 = procedure(A, _, C, D, E, F, G, H, I, J, K, L, M, N),
-	ProcInfo = procedure(A, VarSet, C, D, E, F, G, H, I, J, K, L, M, N).
+    ProcInfo0 = procedure(A, _, C, D, E, F, G, H, I, J, K, L, M, N, O, P),
+    ProcInfo = procedure(A, VarSet, C, D, E, F, G, H, I, J, K, L, M, N, O, P).
 
 proc_info_set_variables(ProcInfo0, Vars, ProcInfo) :-
-	ProcInfo0 = procedure(A, _, C, D, E, F, G, H, I, J, K, L, M, N),
-	ProcInfo = procedure(A, Vars, C, D, E, F, G, H, I, J, K, L, M, N).
+    ProcInfo0 = procedure(A, _, C, D, E, F, G, H, I, J, K, L, M, N, O, P),
+    ProcInfo = procedure(A, Vars, C, D, E, F, G, H, I, J, K, L, M, N, O, P).
 
 proc_info_set_vartypes(ProcInfo0, Vars, ProcInfo) :-
-	ProcInfo0 = procedure(A, B, _, D, E, F, G, H, I, J, K, L, M, N),
-	ProcInfo = procedure(A, B, Vars, D, E, F, G, H, I, J, K, L, M, N).
+    ProcInfo0 = procedure(A, B, _, D, E, F, G, H, I, J, K, L, M, N, O, P),
+    ProcInfo = procedure(A, B, Vars, D, E, F, G, H, I, J, K, L, M, N, O, P).
 
-proc_info_set_headvars(ProcInfo0, HeadVars, ProcInfo) :-
-	ProcInfo0 = procedure(A, B, C, _, E, F, G, H, I, J, K, L, M, N),
-	ProcInfo = procedure(A, B, C, HeadVars, E, F, G, H, I, J, K, L, M, N).
+proc_info_set_headvars(ProcInfo0, HdVars, ProcInfo) :-
+    ProcInfo0 = procedure(A, B, C, _, E, F, G, H, I, J, K, L, M, N, O, P),
+    ProcInfo = procedure(A, B, C, HdVars, E, F, G, H, I, J, K, L, M, N, O, P).
 
 proc_info_set_argmodes(ProcInfo0, ArgModes, ProcInfo) :-
-	ProcInfo0 = procedure(A, B, C, D, _, F, G, H, I, J, K, L, M, N),
-	ProcInfo = procedure(A, B, C, D, ArgModes, F, G, H, I, J, K, L, M, N).
+    ProcInfo0 = procedure(A, B, C, D, _, F, G, H, I, J, K, L, M, N, O, P),
+    ProcInfo = procedure(A, B, C, D, ArgModes, F, G, H, I, 
+    		J, K, L, M, N, O, P).
 
 proc_info_set_maybe_arglives(ProcInfo0, ArgLives, ProcInfo) :-
-	ProcInfo0 = procedure(A, B, C, D, E, _, G, H, I, J, K, L, M, N),
-	ProcInfo = procedure(A, B, C, D, E, ArgLives, G, H, I, J, K, L, M, N).
+    ProcInfo0 = procedure(A, B, C, D, E, _, G, H, I, J, K, L, M, N, O, P),
+    ProcInfo = procedure(A, B, C, D, E, ArgLives, G, H, I, 
+    		J, K, L, M, N, O, P).
 
 proc_info_set_inferred_determinism(ProcInfo0, Detism, ProcInfo) :-
-	ProcInfo0 = procedure(A, B, C, D, E, F, G, H, I, _, K, L, M, N),
-	ProcInfo = procedure(A, B, C, D, E, F, G, H, I, Detism, K, L, M, N).
+    ProcInfo0 = procedure(A, B, C, D, E, F, G, H, I, _, K, L, M, N, O, P),
+    ProcInfo = procedure(A, B, C, D, E, F, G, H, I, Detism, K, L, M, N, O, P).
 
 proc_info_set_can_process(ProcInfo0, CanProcess, ProcInfo) :-
- 	ProcInfo0 = procedure(A, B, C, D, E, F, G, H, I, J, _, L, M, N),
- 	ProcInfo = procedure(A, B, C, D, E, F, G, H, I, J, CanProcess, L, M, N).
+    ProcInfo0 = procedure(A, B, C, D, E, F, G, H, I, J, _, L, M, N, O, P),
+    ProcInfo = procedure(A, B, C, D, E, F, G, H, I, J, CanProcess, 
+    		L, M, N, O, P).
 
 proc_info_set_goal(ProcInfo0, Goal, ProcInfo) :-
-	ProcInfo0 = procedure(A, B, C, D, E, F, _, H, I, J, K, L, M, N),
-	ProcInfo = procedure(A, B, C, D, E, F, Goal, H, I, J, K, L, M, N).
+    ProcInfo0 = procedure(A, B, C, D, E, F, _, H, I, J, K, L, M, N, O, P),
+    ProcInfo = procedure(A, B, C, D, E, F, Goal, H, I, J, K, L, M, N, O, P).
 
 proc_info_set_stack_slots(ProcInfo0, StackSlots, ProcInfo) :-
-	ProcInfo0 = procedure(A, B, C, D, E, F, G, H, _, J, K, L, M, N),
-	ProcInfo = procedure(A, B, C, D, E, F, G, H, StackSlots, J, K, L, M, N).
+    ProcInfo0 = procedure(A, B, C, D, E, F, G, H, _, J, K, L, M, N, O, P),
+    ProcInfo = procedure(A, B, C, D, E, F, G, H, StackSlots, J, K, 
+    		L, M, N, O, P).
 
 proc_info_set_arg_info(ProcInfo0, ArgInfo, ProcInfo) :-
-	ProcInfo0 = procedure(A, B, C, D, E, F, G, H, I, J, K, _, M, N),
-	ProcInfo = procedure(A, B, C, D, E, F, G, H, I, J, K, ArgInfo, M, N).
+    ProcInfo0 = procedure(A, B, C, D, E, F, G, H, I, J, K, _, M, N, O, P),
+    ProcInfo = procedure(A, B, C, D, E, F, G, H, I, J, K, ArgInfo, M, N, O, P).
 
 proc_info_set_liveness_info(ProcInfo0, Liveness, ProcInfo) :-
-	ProcInfo0 = procedure(A, B, C, D, E, F, G, H, I, J, K, L, _, N),
-	ProcInfo = procedure(A, B, C, D, E, F, G, H, I, J, K, L, Liveness, N).
+    ProcInfo0 = procedure(A, B, C, D, E, F, G, H, I, J, K, L, _, N, O, P),
+    ProcInfo = procedure(A, B, C, D, E, F, G, H, I, J, K, L, Liveness, 
+    		N, O, P).
 
 proc_info_set_typeinfo_varmap(ProcInfo0, TVarMap, ProcInfo) :-
-	ProcInfo0 = procedure(A, B, C, D, E, F, G, H, I, J, K, L, M, _),
-	ProcInfo = procedure(A, B, C, D, E, F, G, H, I, J, K, L, M, TVarMap).
+    ProcInfo0 = procedure(A, B, C, D, E, F, G, H, I, J, K, L, M, _, O, P),
+    ProcInfo = procedure(A, B, C, D, E, F, G, H, I, J, K, L, M, TVarMap, O, P).
 
-proc_info_get_used_typeinfos_setwise(ProcInfo, Vars, TypeInfoVars) :-
+proc_info_set_termination(ProcInfo0, Terminat, ProcInfo) :-
+    ProcInfo0 = procedure(A, B, C, D, E, F, G, H, I, J, K, L, M, N, _, P),
+    ProcInfo = procedure(A, B, C, D, E, F, G, H, I, J, K, L, 
+    		M, N, Terminat, P).
+
+proc_info_get_typeinfo_vars_setwise(ProcInfo, Vars, TypeInfoVars) :-
 	set__to_sorted_list(Vars, VarList),
-	proc_info_get_used_typeinfos_2(ProcInfo, VarList, TypeInfoVarList),
+	proc_info_get_typeinfo_vars_2(ProcInfo, VarList, TypeInfoVarList),
 	set__list_to_set(TypeInfoVarList, TypeInfoVars).
 
 	% auxiliary predicate - traverses variables and builds a list of
 	% variables that store typeinfos for these variables. 
-:- pred proc_info_get_used_typeinfos_2(proc_info, list(var), list(var)).
-:- mode proc_info_get_used_typeinfos_2(in, in, out) is det.
-proc_info_get_used_typeinfos_2(_, [], []).
-proc_info_get_used_typeinfos_2(ProcInfo, [Var | Vars1], TypeInfoVars) :-
+:- pred proc_info_get_typeinfo_vars_2(proc_info, list(var), list(var)).
+:- mode proc_info_get_typeinfo_vars_2(in, in, out) is det.
+proc_info_get_typeinfo_vars_2(_, [], []).
+proc_info_get_typeinfo_vars_2(ProcInfo, [Var | Vars1], TypeInfoVars) :-
 	proc_info_vartypes(ProcInfo, VarTypeMap),
 	( 
 		map__search(VarTypeMap, Var, Type)
@@ -895,20 +1042,20 @@ proc_info_get_used_typeinfos_2(ProcInfo, [Var | Vars1], TypeInfoVars) :-
 			% Optimize common case
 			TypeVars = []
 		->
-			proc_info_get_used_typeinfos_2(ProcInfo, Vars1, 
+			proc_info_get_typeinfo_vars_2(ProcInfo, Vars1, 
 				TypeInfoVars)
 		;
-			% It's possible there are some complications with
+			% XXX It's possible there are some complications with
 			% higher order pred types here -- if so, maybe
 			% treat them specially.
 			proc_info_typeinfo_varmap(ProcInfo, TVarMap),
 			map__apply_to_list(TypeVars, TVarMap, TypeInfoVars0),
-			proc_info_get_used_typeinfos_2(ProcInfo, Vars1,
+			proc_info_get_typeinfo_vars_2(ProcInfo, Vars1,
 				TypeInfoVars1),
 			list__append(TypeInfoVars0, TypeInfoVars1, TypeInfoVars)
 		)
 	;
-		error("proc_info_get_used_typeinfos_2: var not found in typemap")
+		error("proc_info_get_typeinfo_vars_2: var not found in typemap")
 	).
 
 proc_info_ensure_unique_names(ProcInfo0, ProcInfo) :-
