@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1995 University of Melbourne.
+% Copyright (C) 1994-1997 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -27,6 +27,9 @@
 	%
 	% 	* procedures which have a `:- pragma inline(name/arity).'
 	%
+	% It will not inline procedures which have a 
+	% 	`:- pragma no_inline(name/arity).'
+	%
 	% If inlining a procedure takes the total number of variables over
 	% a given threshold (from a command-line option), then the procedure
 	% is not inlined - note that this means that some calls to a
@@ -36,7 +39,7 @@
 	% the call-graph towards the top, first perfoming inlining on a
 	% procedure then deciding if calls to it (higher in the call-graph)
 	% should be inlined. SCCs get flattend and processed in the order
-	% returned by hlds__dependency_info_get_dependency_ordering.
+	% returned by hlds_dependency_info_get_dependency_ordering.
 	%
 	% There are a couple of classes of procedure that we clearly want
 	% to inline because doing so *reduces* the size of the generated
@@ -58,9 +61,9 @@
 	%	- arithmetic predicates where the as above, the cost of the
 	%	  call will often outweigh the cost of the arithmetic.
 	%
-	%	- pragma C code, where often the C operation is very small,
-	%	  inlining avoids a call and allows the C compiler to do a
-	%	  better job of optimizing it.
+	%	- det or semi pragma C code, where often the C operation is
+	%	  very small, inlining avoids a call and allows the C compiler
+	%	  to do a better job of optimizing it.
 	%
 	% The threshold on the size of simple goals (which covers both of the
 	% first two cases above), is to prevent the inlining of large goals
@@ -71,17 +74,20 @@
 	% the problem of inlining lots of calls and having a resulting
 	% procedure with so many variables that the back end of the compiler
 	% gets bogged down (for example in the pseudoknot benchmark).
+	%
+	% Due to the way in which we generate code for model_non pragma_c_code,
+	% procedures whose body is such a pragma_c_code must NOT be inlined.
 
 %-----------------------------------------------------------------------------%
 
 :- interface.
-:- import_module hlds_module, llds.
+:- import_module hlds_module.
 :- import_module io.
 
 :- pred inlining(module_info, module_info, io__state, io__state).
 :- mode inlining(in, out, di, uo) is det.
 
-:- pred inlining__is_simple_goal(hlds__goal, int).
+:- pred inlining__is_simple_goal(hlds_goal, int).
 :- mode inlining__is_simple_goal(in, in) is semidet.
 
 %-----------------------------------------------------------------------------%
@@ -89,9 +95,9 @@
 
 :- implementation.
 
-:- import_module hlds_pred, hlds_goal, globals, options.
+:- import_module hlds_pred, hlds_goal, globals, options, llds.
 :- import_module dead_proc_elim, type_util, mode_util, goal_util.
-:- import_module passes_aux, code_aux.
+:- import_module passes_aux, code_aux, quantification, det_analysis, prog_data.
 
 :- import_module bool, int, list, assoc_list, map, set, std_util.
 :- import_module term, varset, require, hlds_data, dependency_graph.
@@ -155,7 +161,7 @@ inlining(ModuleInfo0, ModuleInfo) -->
 		% that yet.
 	{ module_info_ensure_dependency_info(ModuleInfo0, ModuleInfo1) },
 	{ module_info_dependency_info(ModuleInfo1, DepInfo) },
-	{ hlds__dependency_info_get_dependency_ordering(DepInfo, SCCs) },
+	{ hlds_dependency_info_get_dependency_ordering(DepInfo, SCCs) },
 	{ list__condense(SCCs, PredProcs) },
 	{ set__init(InlinedProcs0) },
 	inlining__do_inlining(PredProcs, NeededMap, Params, InlinedProcs0,
@@ -169,7 +175,7 @@ inlining(ModuleInfo0, ModuleInfo) -->
 inlining__do_inlining([], _Needed, _Params, _Inlined, Module, Module) --> [].
 inlining__do_inlining([PPId|PPIds], Needed, Params, Inlined0,
 		Module0, Module) -->
-	{ inlining__in_predproc(PPId, Inlined0, Params, Module0, Module1) },
+	inlining__in_predproc(PPId, Inlined0, Params, Module0, Module1),
 	inlining__mark_predproc(PPId, Needed, Params, Module1,
 		Inlined0, Inlined1),
 	inlining__do_inlining(PPIds, Needed, Params, Inlined1, Module1, Module).
@@ -211,7 +217,21 @@ inlining__mark_predproc(PredProcId, NeededMap, Params, ModuleInfo,
 			{ NumUses = 1 }
 		),
 		% Don't inline recursive predicates
-		{ \+ goal_calls(CalledGoal, PredProcId) }
+		{ \+ goal_calls(CalledGoal, PredProcId) },
+
+		% Don't inline model_non pragma c that doesn't have an
+		% `extra_pragma_info'.  
+		%
+		% XXX  model_non pragma c without `extra_pragma_info' should
+		% not be accepted by the compiler, but at the moment it's
+		% the only way to get model_non pragma c (the ``correct''
+		% way of doing it hasn't been implemented yet).  We just
+		% have to make sure it doesn't get inlined because that stops
+		% it from working.
+		\+ {
+			CalledGoal = pragma_c_code(_,_,_,_,_,_,_,none) - _,
+			proc_info_interface_code_model(ProcInfo, model_non)
+		}
 	->
 		inlining__mark_proc_as_inlined(PredProcId, ModuleInfo,
 			InlinedProcs0, InlinedProcs)
@@ -236,7 +256,7 @@ inlining__is_simple_goal(CalledGoal, SimpleThreshold) :-
 		inlining__is_flat_simple_goal(CalledGoal)
 	).
 
-:- pred inlining__is_flat_simple_goal(hlds__goal::in) is semidet.
+:- pred inlining__is_flat_simple_goal(hlds_goal::in) is semidet.
 
 inlining__is_flat_simple_goal(conj(Goals) - _) :-
 	inlining__is_flat_simple_goal_list(Goals).
@@ -244,11 +264,11 @@ inlining__is_flat_simple_goal(not(Goal) - _) :-
 	inlining__is_flat_simple_goal(Goal).
 inlining__is_flat_simple_goal(some(_, Goal) - _) :-
 	inlining__is_flat_simple_goal(Goal).
-inlining__is_flat_simple_goal(call(_, _, _, Builtin, _, _) - _) :-
-	hlds__is_builtin_is_inline(Builtin).
+inlining__is_flat_simple_goal(call(_, _, _, BuiltinState, _, _) - _) :-
+	BuiltinState = inline_builtin.
 inlining__is_flat_simple_goal(unify(_, _, _, _, _) - _).
 
-:- pred inlining__is_flat_simple_goal_list(hlds__goals::in) is semidet.
+:- pred inlining__is_flat_simple_goal_list(hlds_goals::in) is semidet.
 
 inlining__is_flat_simple_goal_list([]).
 inlining__is_flat_simple_goal_list([Goal | Goals]) :-
@@ -263,7 +283,7 @@ inlining__mark_proc_as_inlined(proc(PredId, ProcId), ModuleInfo,
 		InlinedProcs0, InlinedProcs) -->
 	{ set__insert(InlinedProcs0, proc(PredId, ProcId), InlinedProcs) },
 	{ module_info_pred_info(ModuleInfo, PredId, PredInfo) },
-	( { pred_info_is_inlined(PredInfo) } ->
+	( { pred_info_requested_inlining(PredInfo) } ->
 		[]
 	;
 		write_proc_progress_message("% Inlining ", PredId, ProcId,
@@ -272,12 +292,40 @@ inlining__mark_proc_as_inlined(proc(PredId, ProcId), ModuleInfo,
 
 %-----------------------------------------------------------------------------%
 
-:- pred inlining__in_predproc(pred_proc_id, set(pred_proc_id),
-		inline_params, module_info, module_info).
-:- mode inlining__in_predproc(in, in, in, in, out) is det.
+		% inline_info contains the information that is changed
+		% as a result of inlining. It is threaded through the
+		% inlining process, and when finished, contains the
+		% updated information associated with the new goal.
+		%
+		% It also stores some necessary information that is not
+		% updated.
+
+:- type inline_info	--->
+
+	inline_info(
+		int,			% variable threshold for inlining
+		set(pred_proc_id),	% inlined procs
+		module_info,		% module_info
+
+			% the following fields are updated as a result
+			% of inlining
+		varset,			% varset
+		map(var, type),		% variable types
+		tvarset,		% type variables
+		map(tvar, var),		% type_info varset, a mapping from 
+					% type variables to variables
+					% where their type_info is
+					% stored.
+		bool			% Did we change the determinism
+					% of any subgoal?
+		).
+
+:- pred inlining__in_predproc(pred_proc_id, set(pred_proc_id), inline_params,
+		module_info, module_info, io__state, io__state).
+:- mode inlining__in_predproc(in, in, in, in, out, di, uo) is det.
 
 inlining__in_predproc(PredProcId, InlinedProcs, Params,
-		ModuleInfo0, ModuleInfo) :-
+		ModuleInfo0, ModuleInfo, IoState0, IoState) :-
 	Params = params(_Simple, _SingleUse, _CompoundThreshold,
 			_SimpleThreshold, VarThresh),
 	PredProcId = proc(PredId, ProcId),
@@ -287,89 +335,85 @@ inlining__in_predproc(PredProcId, InlinedProcs, Params,
 	pred_info_procedures(PredInfo0, ProcTable0),
 	map__lookup(ProcTable0, ProcId, ProcInfo0),
 
-	pred_info_typevarset(PredInfo0, TypeVarSet),
+	pred_info_typevarset(PredInfo0, TypeVarSet0),
+
 	proc_info_goal(ProcInfo0, Goal0),
-	proc_info_variables(ProcInfo0, Varset0),
+	proc_info_variables(ProcInfo0, VarSet0),
 	proc_info_vartypes(ProcInfo0, VarTypes0),
+	proc_info_typeinfo_varmap(ProcInfo0, TypeInfoVarMap0),
 
-	inlining__inlining_in_goal(Goal0, Varset0, VarTypes0, TypeVarSet,
-		ModuleInfo0, InlinedProcs, VarThresh, Goal, Varset, VarTypes),
+	DetChanged0 = no,
 
-	proc_info_set_variables(ProcInfo0, Varset, ProcInfo1),
+	InlineInfo0 = inline_info(VarThresh, InlinedProcs, ModuleInfo0,
+		VarSet0, VarTypes0, TypeVarSet0, TypeInfoVarMap0, DetChanged0),
+
+	inlining__inlining_in_goal(Goal0, Goal, InlineInfo0, InlineInfo),
+
+	InlineInfo = inline_info(_, _, _, VarSet, VarTypes, TypeVarSet, 
+		TypeInfoVarMap, DetChanged),
+
+	pred_info_set_typevarset(PredInfo0, TypeVarSet, PredInfo1),
+
+	proc_info_set_variables(ProcInfo0, VarSet, ProcInfo1),
 	proc_info_set_vartypes(ProcInfo1, VarTypes, ProcInfo2),
-	proc_info_set_goal(ProcInfo2, Goal, ProcInfo),
+	proc_info_set_typeinfo_varmap(ProcInfo2, TypeInfoVarMap, ProcInfo3),
+	proc_info_set_goal(ProcInfo3, Goal, ProcInfo),
 
-	map__set(ProcTable0, ProcId, ProcInfo, ProcTable),
-	pred_info_set_procedures(PredInfo0, ProcTable, PredInfo),
-	map__set(PredTable0, PredId, PredInfo, PredTable),
-	module_info_set_preds(ModuleInfo0, PredTable, ModuleInfo).
+	map__det_update(ProcTable0, ProcId, ProcInfo, ProcTable),
+	pred_info_set_procedures(PredInfo1, ProcTable, PredInfo),
+	map__det_update(PredTable0, PredId, PredInfo, PredTable),
+	module_info_set_preds(ModuleInfo0, PredTable, ModuleInfo1),
+
+		% If the determinism of some sub-goals has changed,
+		% then we re-run determinism analysis, because
+		% propagating the determinism information through
+		% the procedure may lead to more efficient code.
+	( DetChanged = yes,	
+		globals__io_get_globals(Globals, IoState0, IoState),
+		det_infer_proc(PredId, ProcId, ModuleInfo1, ModuleInfo,
+			Globals, _, _, _)
+	; DetChanged = no,
+		ModuleInfo = ModuleInfo1,
+		IoState = IoState0
+	).
 
 %-----------------------------------------------------------------------------%
 
-:- pred inlining__inlining_in_goal(hlds__goal, varset, map(var, type),
-	tvarset, module_info, set(pred_proc_id), int, hlds__goal,
-	varset, map(var, type)).
-:- mode inlining__inlining_in_goal(in, in, in, in, in, in, in, out, out, out)
-	is det.
+:- pred inlining__inlining_in_goal(hlds_goal, hlds_goal, inline_info,
+		inline_info).
+:- mode inlining__inlining_in_goal(in, out, in, out) is det.
 
-inlining__inlining_in_goal(Goal0 - GoalInfo, Varset0, VarTypes0, TypeVarSet,
-		ModuleInfo, InlinedProcs, VarThresh,
-			Goal - GoalInfo, Varset, VarTypes) :-
-	inlining__inlining_in_goal_2(Goal0, Varset0, VarTypes0, TypeVarSet,
-		ModuleInfo, InlinedProcs, VarThresh, Goal, Varset, VarTypes).
+inlining__inlining_in_goal(conj(Goals0) - GoalInfo, conj(Goals) - GoalInfo) -->
+	inlining__inlining_in_conj(Goals0, Goals).
 
-%-----------------------------------------------------------------------------%
+inlining__inlining_in_goal(disj(Goals0, SM) - GoalInfo,
+		disj(Goals, SM) - GoalInfo) -->
+	inlining__inlining_in_disj(Goals0, Goals).
 
-:- pred inlining__inlining_in_goal_2(hlds__goal_expr, varset, map(var, type),
-	tvarset, module_info, set(pred_proc_id), int, hlds__goal_expr,
-	varset, map(var, type)).
-:- mode inlining__inlining_in_goal_2(in, in, in, in, in, in, in, out, out, out)
-	is det.
+inlining__inlining_in_goal(switch(Var, Det, Cases0, SM) - GoalInfo,
+		switch(Var, Det, Cases, SM) - GoalInfo) -->
+	inlining__inlining_in_cases(Cases0, Cases).
 
-inlining__inlining_in_goal_2(conj(Goals0), Varset0, VarTypes0, TVarSet,
-		ModuleInfo, InlinedProcs, Thresh,
-		conj(Goals), Varset, VarTypes) :-
-	inlining__inlining_in_conj(Goals0, Varset0, VarTypes0, TVarSet,
-		ModuleInfo, InlinedProcs, Thresh, Goals, Varset, VarTypes).
+inlining__inlining_in_goal(
+		if_then_else(Vars, Cond0, Then0, Else0, SM) - GoalInfo,
+		if_then_else(Vars, Cond, Then, Else, SM) - GoalInfo) -->
+	inlining__inlining_in_goal(Cond0, Cond),
+	inlining__inlining_in_goal(Then0, Then),
+	inlining__inlining_in_goal(Else0, Else).
 
-inlining__inlining_in_goal_2(disj(Goals0, SM), Varset0, VarTypes0, TVarset,
-		ModuleInfo, InlinedProcs, Thresh,
-		disj(Goals, SM), Varset, VarTypes) :-
-	inlining__inlining_in_disj(Goals0, Varset0, VarTypes0, TVarset,
-		ModuleInfo, InlinedProcs, Thresh, Goals, Varset, VarTypes).
+inlining__inlining_in_goal(not(Goal0) - GoalInfo, not(Goal) - GoalInfo) -->
+	inlining__inlining_in_goal(Goal0, Goal).
 
-inlining__inlining_in_goal_2(switch(Var, Det, Cases0, SM),
-		Varset0, VarTypes0, TVarset, ModuleInfo, InlinedProcs, Thresh,
-		switch(Var, Det, Cases, SM), Varset, VarTypes) :-
-	inlining__inlining_in_cases(Cases0, Varset0, VarTypes0, TVarset,
-		ModuleInfo, InlinedProcs, Thresh, Cases, Varset, VarTypes).
+inlining__inlining_in_goal(some(Vars, Goal0) - GoalInfo,
+		some(Vars, Goal) - GoalInfo) -->
+	inlining__inlining_in_goal(Goal0, Goal).
 
-inlining__inlining_in_goal_2(if_then_else(Vars, Cond0, Then0, Else0, SM),
-		Varset0, VarTypes0, TVarSet, ModuleInfo, InlinedProcs, Thresh,
-		if_then_else(Vars, Cond, Then, Else, SM), Varset, VarTypes) :-
-	inlining__inlining_in_goal(Cond0, Varset0, VarTypes0, TVarSet,
-		ModuleInfo, InlinedProcs, Thresh, Cond, Varset1, VarTypes1),
-	inlining__inlining_in_goal(Then0, Varset1, VarTypes1, TVarSet,
-		ModuleInfo, InlinedProcs, Thresh, Then, Varset2, VarTypes2),
-	inlining__inlining_in_goal(Else0, Varset2, VarTypes2, TVarSet,
-		ModuleInfo, InlinedProcs, Thresh, Else, Varset, VarTypes).
+inlining__inlining_in_goal(call(PredId, ProcId, ArgVars, Builtin, Context,
+		Sym) - GoalInfo0, Goal - GoalInfo, InlineInfo0, InlineInfo) :-
 
-inlining__inlining_in_goal_2(not(Goal0), Varset0, VarTypes0, TVarSet,
-		ModuleInfo, InlinedProcs, Thresh,
-		not(Goal), Varset, VarTypes) :-
-	inlining__inlining_in_goal(Goal0, Varset0, VarTypes0, TVarSet,
-		ModuleInfo, InlinedProcs, Thresh, Goal, Varset, VarTypes).
-
-inlining__inlining_in_goal_2(some(Vars, Goal0), Varset0, VarTypes0, TVarSet,
-		ModuleInfo, InlinedProcs, Thresh,
-		some(Vars, Goal), Varset, VarTypes) :-
-	inlining__inlining_in_goal(Goal0, Varset0, VarTypes0, TVarSet,
-		ModuleInfo, InlinedProcs, Thresh, Goal, Varset, VarTypes).
-
-inlining__inlining_in_goal_2(
-		call(PredId, ProcId, ArgVars, Builtin, Context, Sym),
-		Varset0, VarTypes0, TypeVarSet,
-		ModuleInfo, InlinedProcs, Thresh, Goal, Varset, VarTypes) :-
+	InlineInfo0 = inline_info(VarThresh, InlinedProcs, ModuleInfo,
+		VarSet0, VarTypes0, TypeVarSet0, TypeInfoVarMap0,
+		DetChanged0),
 
 	% should we inline this call?
 	(
@@ -377,7 +421,7 @@ inlining__inlining_in_goal_2(
 				InlinedProcs, ModuleInfo),
 			% okay, but will we exceed the number-of-variables
 			% threshold?
-		varset__vars(Varset0, ListOfVars),
+		varset__vars(VarSet0, ListOfVars),
 		list__length(ListOfVars, ThisMany),
 			% We need to find out how many variables the
 			% Callee has
@@ -387,7 +431,7 @@ inlining__inlining_in_goal_2(
 		varset__vars(CalleeVarset, CalleeListOfVars),
 		list__length(CalleeListOfVars, CalleeThisMany),
 		TotalVars is ThisMany + CalleeThisMany,
-		TotalVars =< Thresh
+		TotalVars =< VarThresh
 	->
 		% Yes.  So look up the rest of the info for the
 		% called procedure.
@@ -396,6 +440,7 @@ inlining__inlining_in_goal_2(
 		proc_info_headvars(ProcInfo, HeadVars),
 		proc_info_goal(ProcInfo, CalledGoal),
 		proc_info_vartypes(ProcInfo, CalleeVarTypes0),
+		proc_info_typeinfo_varmap(ProcInfo, CalledTypeInfoVarMap0),
 
 		% Substitute the appropriate types into the type
 		% mapping of the called procedure.  For example, if we
@@ -404,11 +449,13 @@ inlining__inlining_in_goal_2(
 		% type `T' with type `int' when we inline it.
 
 		% first, rename apart the type variables in the callee.
-		% (we can throw away the new typevarset, since
-		% we are about to substitute away any new type variables)
+		% (we can almost throw away the new typevarset, since we
+		% are about to substitute away any new type variables,
+		% but any unbound type variables in the callee will not
+		% be substituted away)
 
-		varset__merge_subst(TypeVarSet, CalleeTypeVarSet,
-			_NewTypeVarSet, TypeRenaming),
+		varset__merge_subst(TypeVarSet0, CalleeTypeVarSet,
+			TypeVarSet, TypeRenaming),
 		apply_substitution_to_type_map(CalleeVarTypes0, TypeRenaming,
 			CalleeVarTypes1),
 
@@ -436,86 +483,88 @@ inlining__inlining_in_goal_2(
 
 		% Now rename apart the variables in the called goal.
 
-		varset__vars(CalleeVarset, CalleeVars0),
 		map__from_corresponding_lists(HeadVars, ArgVars, Subn0),
-		goal_util__create_variables(CalleeVars0, Varset0,
+		goal_util__create_variables(CalleeListOfVars, VarSet0,
 			VarTypes0, Subn0, CalleeVarTypes, CalleeVarset,
-				Varset, VarTypes, Subn),
+				VarSet, VarTypes, Subn),
 		goal_util__must_rename_vars_in_goal(CalledGoal, Subn,
-			Goal - _GInfo)
+			Goal - GoalInfo),
+
+			% If the inferred determinism of the called
+			% goal differs from the declared determinism,
+			% flag that we should re-run determinism analysis
+			% on this proc.
+		goal_info_get_determinism(GoalInfo0, Determinism0),
+		goal_info_get_determinism(GoalInfo, Determinism),
+		( Determinism0 = Determinism ->
+			DetChanged = DetChanged0
+		;
+			DetChanged = yes
+		),
+
+		apply_substitutions_to_var_map(CalledTypeInfoVarMap0, 
+			TypeRenaming, Subn, CalledTypeInfoVarMap1),
+		map__merge(TypeInfoVarMap0, CalledTypeInfoVarMap1,
+			TypeInfoVarMap)
 	;
 		Goal = call(PredId, ProcId, ArgVars, Builtin, Context, Sym),
-		Varset = Varset0,
-		VarTypes = VarTypes0
-	).
+		GoalInfo = GoalInfo0,
+		VarSet = VarSet0,
+		VarTypes = VarTypes0,
+		TypeVarSet = TypeVarSet0,
+		TypeInfoVarMap = TypeInfoVarMap0,
+		DetChanged = DetChanged0
+	),
+	InlineInfo = inline_info(VarThresh, InlinedProcs, ModuleInfo,
+		VarSet, VarTypes, TypeVarSet, TypeInfoVarMap, DetChanged).
 
-inlining__inlining_in_goal_2(higher_order_call(A, B, C, D, E),
-		Varset, VarTypes, _, _, _, _,
-		higher_order_call(A, B, C, D, E), Varset, VarTypes).
+inlining__inlining_in_goal(higher_order_call(A, B, C, D, E, F) - GoalInfo,
+		higher_order_call(A, B, C, D, E, F) - GoalInfo) --> [].
 
-inlining__inlining_in_goal_2(unify(A, B, C, D, E), Varset, VarTypes,
-		_, _, _, _, unify(A, B, C, D, E), Varset, VarTypes).
+inlining__inlining_in_goal(unify(A, B, C, D, E) - GoalInfo,
+		unify(A, B, C, D, E) - GoalInfo) --> [].
 
-inlining__inlining_in_goal_2(pragma_c_code(A, B, C, D, E, F), Varset, VarTypes,
-		_, _, _, _, pragma_c_code(A, B, C, D, E, F), Varset, VarTypes).
+inlining__inlining_in_goal(pragma_c_code(A, B, C, D, E, F, G, H) - GoalInfo,
+		pragma_c_code(A, B, C, D, E, F, G, H) - GoalInfo) --> [].
 
 %-----------------------------------------------------------------------------%
 
-:- pred inlining__inlining_in_disj(list(hlds__goal), varset, map(var, type),
-	tvarset, module_info, set(pred_proc_id), int, list(hlds__goal),
-	varset, map(var, type)).
-:- mode inlining__inlining_in_disj(in, in, in, in, in, in, in, out, out, out)
-	is det.
+:- pred inlining__inlining_in_disj(list(hlds_goal), list(hlds_goal), 
+		inline_info, inline_info).
+:- mode inlining__inlining_in_disj(in, out, in, out) is det.
 
-inlining__inlining_in_disj([], Varset, VarTypes, _, _, _, _,
-		[], Varset, VarTypes).
-inlining__inlining_in_disj([Goal0 | Goals0], Varset0, VarTypes0, TVarSet,
-		ModuleInfo, InlinedProcs, Threshold, [Goal | Goals],
-		Varset, VarTypes) :-
-	inlining__inlining_in_goal(Goal0, Varset0, VarTypes0, TVarSet,
-		ModuleInfo, InlinedProcs, Threshold, Goal, Varset1, VarTypes1),
-	inlining__inlining_in_disj(Goals0, Varset1, VarTypes1, TVarSet,
-		ModuleInfo, InlinedProcs, Threshold, Goals, Varset, VarTypes).
+inlining__inlining_in_disj([], []) --> [].
+inlining__inlining_in_disj([Goal0 | Goals0], [Goal | Goals]) -->
+	inlining__inlining_in_goal(Goal0, Goal),
+	inlining__inlining_in_disj(Goals0, Goals).
 
 %-----------------------------------------------------------------------------%
 
-:- pred inlining__inlining_in_cases(list(case), varset, map(var, type),
-	tvarset, module_info, set(pred_proc_id), int, list(case),
-	varset, map(var, type)).
-:- mode inlining__inlining_in_cases(in, in, in, in, in, in, in, out, out, out)
-	is det.
+:- pred inlining__inlining_in_cases(list(case), list(case), inline_info, 
+		inline_info).
+:- mode inlining__inlining_in_cases(in, out, in, out) is det.
 
-inlining__inlining_in_cases([], Varset, VarTypes, _, _, _, _,
-				[], Varset, VarTypes).
+inlining__inlining_in_cases([], []) --> [].
 inlining__inlining_in_cases([case(Cons, Goal0) | Goals0],
-		Varset0, VarTypes0, TVarSet, ModuleInfo, InlinedProcs,
-		Threshold, [case(Cons, Goal) | Goals], Varset, VarTypes) :-
-	inlining__inlining_in_goal(Goal0, Varset0, VarTypes0, TVarSet,
-		ModuleInfo, InlinedProcs, Threshold, Goal, Varset1, VarTypes1),
-	inlining__inlining_in_cases(Goals0, Varset1, VarTypes1, TVarSet,
-		ModuleInfo, InlinedProcs, Threshold, Goals, Varset, VarTypes).
+		[case(Cons, Goal) | Goals]) -->
+	inlining__inlining_in_goal(Goal0, Goal),
+	inlining__inlining_in_cases(Goals0, Goals).
 
 %-----------------------------------------------------------------------------%
 
-:- pred inlining__inlining_in_conj(list(hlds__goal), varset, map(var, type),
-	tvarset, module_info, set(pred_proc_id), int, list(hlds__goal),
-	varset, map(var, type)).
-:- mode inlining__inlining_in_conj(in, in, in, in, in, in, in, out, out, out)
-	is det.
+:- pred inlining__inlining_in_conj(list(hlds_goal), list(hlds_goal),
+		inline_info, inline_info).
+:- mode inlining__inlining_in_conj(in, out, in, out) is det.
 
 	% Since a single goal may become a conjunction,
 	% we flatten the conjunction as we go.
 
-inlining__inlining_in_conj([], Varset, VarTypes, _, _, _, _,
-		[], Varset, VarTypes).
-inlining__inlining_in_conj([Goal0 | Goals0], Varset0, VarTypes0, TVarSet,
-		ModuleInfo, InlinedProcs, Threshold, Goals, Varset, VarTypes) :-
-	inlining__inlining_in_goal(Goal0, Varset0, VarTypes0, TVarSet,
-		ModuleInfo, InlinedProcs, Threshold, Goal1, Varset1, VarTypes1),
-	goal_to_conj_list(Goal1, Goal1List),
-	inlining__inlining_in_conj(Goals0, Varset1, VarTypes1, TVarSet,
-		ModuleInfo, InlinedProcs, Threshold, Goals1, Varset, VarTypes),
-	list__append(Goal1List, Goals1, Goals).
+inlining__inlining_in_conj([], []) --> [].
+inlining__inlining_in_conj([Goal0 | Goals0], Goals) -->
+	inlining__inlining_in_goal(Goal0, Goal1),
+	{ goal_to_conj_list(Goal1, Goal1List) },
+	inlining__inlining_in_conj(Goals0, Goals1),
+	{ list__append(Goal1List, Goals1, Goals) }.
 
 %-----------------------------------------------------------------------------%
 
@@ -527,16 +576,16 @@ inlining__inlining_in_conj([Goal0 | Goals0], Varset0, VarTypes0, TVarSet,
 	% indicating that it should be inlined, or if the goal
 	% is a conjunction of builtins.
 
-:- pred inlining__should_inline_proc(pred_id, proc_id, is_builtin,
+:- pred inlining__should_inline_proc(pred_id, proc_id, builtin_state,
 	set(pred_proc_id), module_info).
 :- mode inlining__should_inline_proc(in, in, in, in, in) is semidet.
 
-inlining__should_inline_proc(PredId, ProcId, Builtin, InlinedProcs,
+inlining__should_inline_proc(PredId, ProcId, BuiltinState, InlinedProcs,
 		ModuleInfo) :-
 
 	% don't inline builtins, the code generator will handle them
 
-	\+ hlds__is_builtin_is_internal(Builtin),
+	BuiltinState = not_builtin,
 
 	% don't try to inline imported predicates, since we don't
 	% have the code for them.
@@ -545,11 +594,20 @@ inlining__should_inline_proc(PredId, ProcId, Builtin, InlinedProcs,
 	\+ pred_info_is_imported(PredInfo),
 		% this next line catches the case of locally defined
 		% unification predicates for imported types.
-	\+ (pred_info_is_pseudo_imported(PredInfo), ProcId = 0),
+	\+ (
+		pred_info_is_pseudo_imported(PredInfo),
+		hlds_pred__in_in_unification_proc_id(ProcId)
+	),
+
+	% don't inlining anything we have been specifically requested
+	% not to inline.
+
+	\+ pred_info_requested_no_inlining(PredInfo),
 
 	% OK, we could inline it - but should we?  Apply our heuristic.
+
 	(
-		pred_info_is_inlined(PredInfo)
+		pred_info_requested_inlining(PredInfo)
 	;
 		set__member(proc(PredId, ProcId), InlinedProcs)
 	).
