@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1995 University of Melbourne.
+% Copyright (C) 1996-1997 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -18,8 +18,8 @@
 :- module instmap.
 
 :- interface.
-:- import_module hlds_module, prog_data, mode_info.
-:- import_module set.
+:- import_module hlds_module, prog_data, mode_info, (inst).
+:- import_module set, term.
 
 :- type instmap.
 :- type instmap_delta.
@@ -78,6 +78,10 @@
 :- pred instmap_delta_from_assoc_list(assoc_list(var, inst), instmap_delta).
 :- mode instmap_delta_from_assoc_list(in, out) is det.
 
+:- pred instmap_delta_from_mode_list(list(var), list(mode),
+		module_info, instmap_delta).
+:- mode instmap_delta_from_mode_list(in, in, in, out) is det.
+
 %-----------------------------------------------------------------------------%
 
 	% Return the set of variables in an instmap.
@@ -105,11 +109,11 @@
 :- pred instmap__lookup_var(instmap, var, inst).
 :- mode instmap__lookup_var(in, in, out) is det.
 
-	% Given an instmap_delta and a variable, determine the inst
-	% of that variable.
+	% Given an instmap_delta and a variable, determine the new inst
+	% of that variable (if any).
 	%
-:- pred instmap_delta_lookup_var(instmap_delta, var, inst).
-:- mode instmap_delta_lookup_var(in, in, out) is det.
+:- pred instmap_delta_search_var(instmap_delta, var, inst).
+:- mode instmap_delta_search_var(in, in, out) is semidet.
 
 	% Given an instmap and a list of variables, return a list
 	% containing the insts of those variable.
@@ -117,11 +121,6 @@
 :- pred instmap__lookup_vars(list(var), instmap, list(inst)).
 :- mode instmap__lookup_vars(in, in, out) is det.
 
-	% Set an entry in an instmap.
-	%
-:- pred instmap__set(instmap, var, inst, instmap).
-:- mode instmap__set(in, in, in, out) is det.
-
 	% Insert an entry into an instmap_delta.  Note that you
 	% cannot call instmap_delta_insert for a variable already
 	% present.
@@ -129,6 +128,30 @@
 :- pred instmap_delta_insert(instmap_delta, var, inst, instmap_delta).
 :- mode instmap_delta_insert(in, in, in, out) is det.
 
+	% Set an entry in an instmap.
+	%
+:- pred instmap__set(instmap, var, inst, instmap).
+:- mode instmap__set(in, in, in, out) is det.
+
+:- pred instmap_delta_set(instmap_delta, var, inst, instmap_delta).
+:- mode instmap_delta_set(in, in, in, out) is det.
+
+	% Bind a variable in an instmap to a functor at the beginning
+	% of a case in a switch. Aborts on compiler generated cons_ids.
+:- pred instmap_delta_bind_var_to_functor(var, cons_id, instmap,
+		instmap_delta, instmap_delta, module_info, module_info).
+:- mode instmap_delta_bind_var_to_functor(in, in, in, in, out, in, out) is det.
+
+:- pred instmap__bind_var_to_functor(var, cons_id,
+		instmap, instmap, module_info, module_info).
+:- mode instmap__bind_var_to_functor(in, in, in, out, in, out) is det.
+
+	% Update the given instmap to include the initial insts of the
+	% lambda variables.
+:- pred instmap__pre_lambda_update(module_info, list(var), list(mode),
+		instmap, instmap).
+:- mode instmap__pre_lambda_update(in, in, in, in, out) is det.
+
 %-----------------------------------------------------------------------------%
 
 	% Given two instmaps and a set of variables, compute an instmap delta
@@ -189,10 +212,22 @@
 :- pred instmap__no_output_vars(instmap, instmap_delta, set(var), module_info).
 :- mode instmap__no_output_vars(in, in, in, in) is semidet.
 
-	% merge_instmap_delta
-:- pred merge_instmap_delta(instmap_delta, instmap_delta, instmap_delta,
-				module_info, module_info).
-:- mode merge_instmap_delta(in, in, out, in, out) is det.
+	% merge_instmap_delta(InitialInstMap, NonLocals,
+	%	InstMapDeltaA, InstMapDeltaB, ModuleInfo0, ModuleInfo)
+	% Merge the instmap_deltas of different branches of an ite, disj
+	% or switch.
+:- pred merge_instmap_delta(instmap, set(var), instmap_delta, instmap_delta,
+		instmap_delta, module_info, module_info).
+:- mode merge_instmap_delta(in, in, in, in, out, in, out) is det.
+
+	% merge_instmap_deltas(Vars, InstMapDeltas,
+	%	MergedInstMapDelta, ModuleInfo0, ModuleInfo)
+	% takes a list of instmap deltas from the branches of an if-then-else,
+	% switch, or disj and merges them. This is used in situations
+	% where the bindings are known to be compatible.
+:- pred merge_instmap_deltas(instmap, set(var), list(instmap_delta),
+		instmap_delta, module_info, module_info).
+:- mode merge_instmap_deltas(in, in, in, out, in, out) is det.
 
 %-----------------------------------------------------------------------------%
 
@@ -206,6 +241,9 @@
 		instmap_delta).
 :- mode instmap_delta_apply_sub(in, in, in, out) is det.
 
+:- pred instmap__apply_sub(instmap, bool, map(var, var), instmap).
+:- mode instmap__apply_sub(in, in, in, out) is det.
+
 %-----------------------------------------------------------------------------%
 
 :- pred instmap__to_assoc_list(instmap, assoc_list(var, inst)).
@@ -220,7 +258,8 @@
 :- implementation.
 
 :- import_module mode_util, inst_match, prog_data, mode_errors, goal_util.
-:- import_module std_util, bool, map, set, assoc_list, require.
+:- import_module hlds_data, inst_util.
+:- import_module list, std_util, bool, map, set, assoc_list, require.
 
 :- type instmap_delta	==	instmap.
 
@@ -263,6 +302,34 @@ instmap_delta_from_assoc_list(AL, reachable(Instmapping)) :-
 
 %-----------------------------------------------------------------------------%
 
+instmap_delta_from_mode_list(Var, Modes, ModuleInfo, InstMapDelta) :-
+	instmap_delta_init_reachable(InstMapDelta0),
+	instmap_delta_from_mode_list_2(Var, Modes, ModuleInfo,
+		InstMapDelta0, InstMapDelta).
+	
+:- pred instmap_delta_from_mode_list_2(list(var), list(mode),
+		module_info, instmap_delta, instmap_delta).
+:- mode instmap_delta_from_mode_list_2(in, in, in, in, out) is det.
+
+instmap_delta_from_mode_list_2([], [], _, InstMapDelta, InstMapDelta).
+instmap_delta_from_mode_list_2([], [_|_], _, _, _) :-
+	error("instmap_delta_from_mode_list_2").
+instmap_delta_from_mode_list_2([_|_], [], _, _, _) :-
+	error("instmap_delta_from_mode_list_2").
+instmap_delta_from_mode_list_2([Var | Vars], [Mode | Modes], ModuleInfo,
+		InstMapDelta0, InstMapDelta) :-
+	mode_get_insts(ModuleInfo, Mode, Inst1, Inst2),
+	( Inst1 = Inst2 ->
+		instmap_delta_from_mode_list_2(Vars, Modes, ModuleInfo,
+			InstMapDelta0, InstMapDelta)
+	;
+		instmap_delta_set(InstMapDelta0, Var, Inst2, InstMapDelta1),
+		instmap_delta_from_mode_list_2(Vars, Modes, ModuleInfo,
+			InstMapDelta1, InstMapDelta)
+	).
+
+%-----------------------------------------------------------------------------%
+
 instmap__vars(Instmap, Vars) :-
 	instmap__vars_list(Instmap, VarsList),
 	set__list_to_set(VarsList, Vars).
@@ -286,9 +353,6 @@ instmap__lookup_var(unreachable, _Var, not_reached).
 instmap__lookup_var(reachable(InstMap), Var, Inst) :-
 	instmapping_lookup_var(InstMap, Var, Inst).
 
-instmap_delta_lookup_var(InstmapDelta, Var, Inst) :-
-	instmap__lookup_var(InstmapDelta, Var, Inst).
-
 :- pred instmapping_lookup_var(instmapping, var, inst).
 :- mode instmapping_lookup_var(in, in, out) is det.
 
@@ -299,6 +363,10 @@ instmapping_lookup_var(InstMap, Var, Inst) :-
 		Inst = free
 	).
 
+instmap_delta_search_var(unreachable, _, not_reached).
+instmap_delta_search_var(reachable(InstMap), Var, Inst) :-
+	map__search(InstMap, Var, Inst).
+
 instmap__lookup_vars([], _InstMap, []).
 instmap__lookup_vars([Arg|Args], InstMap, [Inst|Insts]) :-
 	instmap__lookup_var(InstMap, Arg, Inst),
@@ -309,10 +377,96 @@ instmap__set(reachable(InstMapping0), Var, Inst,
 		reachable(InstMapping)) :-
 	map__set(InstMapping0, Var, Inst, InstMapping).
 
+instmap_delta_set(unreachable, _Var, _Inst, unreachable).
+instmap_delta_set(reachable(InstMapping0), Var, Inst, Instmap) :-
+	( Inst = not_reached ->
+		Instmap = unreachable
+	;
+		map__set(InstMapping0, Var, Inst, InstMapping),
+		Instmap = reachable(InstMapping)
+	).
+
 instmap_delta_insert(unreachable, _Var, _Inst, unreachable).
-instmap_delta_insert(reachable(InstMapping0), Var, Inst,
-		reachable(InstMapping)) :-
-	map__det_insert(InstMapping0, Var, Inst, InstMapping).
+instmap_delta_insert(reachable(InstMapping0), Var, Inst, Instmap) :-
+	( Inst = not_reached ->
+		Instmap = unreachable
+	;
+		map__det_insert(InstMapping0, Var, Inst, InstMapping),
+		Instmap = reachable(InstMapping)
+	).
+
+%-----------------------------------------------------------------------------%
+
+instmap_delta_bind_var_to_functor(Var, ConsId, InstMap,
+		InstmapDelta0, InstmapDelta, ModuleInfo0, ModuleInfo) :-
+	( 
+		InstmapDelta0 = unreachable,
+		InstmapDelta = unreachable,
+		ModuleInfo = ModuleInfo0
+	;
+		InstmapDelta0 = reachable(InstmappingDelta0),
+
+		%
+		% Get the initial inst from the InstMap
+		%
+		instmap__lookup_var(InstMap, Var, OldInst),
+
+		%
+		% Compute the new inst by taking the old inst,
+		% applying the instmap delta to it,
+		% and then unifying with bound(ConsId, ...)
+		%
+		( map__search(InstmappingDelta0, Var, NewInst0) ->
+			NewInst1 = NewInst0
+		;
+			NewInst1 = OldInst
+		),
+		bind_inst_to_functor(NewInst1, ConsId, NewInst,
+			ModuleInfo0, ModuleInfo),
+
+		%
+		% add `Var :: OldInst -> NewInst' to the instmap delta
+		%
+		( NewInst \= OldInst ->
+			instmap_delta_set(InstmapDelta0, Var, NewInst,
+				InstmapDelta)
+		;
+			InstmapDelta = InstmapDelta0
+		)
+	).
+
+instmap__bind_var_to_functor(Var, ConsId, InstMap0, InstMap,
+		ModuleInfo0, ModuleInfo) :-
+	instmap__lookup_var(InstMap0, Var, Inst0),
+	bind_inst_to_functor(Inst0, ConsId, Inst, ModuleInfo0, ModuleInfo),
+	instmap__set(InstMap0, Var, Inst, InstMap).
+
+:- pred bind_inst_to_functor((inst), cons_id, (inst),
+		module_info, module_info). 
+:- mode bind_inst_to_functor(in, in, out, in, out) is det.
+
+bind_inst_to_functor(Inst0, ConsId, Inst, ModuleInfo0, ModuleInfo) :-
+	cons_id_arity(ConsId, Arity),
+	list__duplicate(Arity, dead, ArgLives),
+	list__duplicate(Arity, free, ArgInsts),
+	(
+		abstractly_unify_inst_functor(dead, Inst0, ConsId, ArgInsts, 
+			ArgLives, real_unify, ModuleInfo0, Inst1, _Det,
+			ModuleInfo1)
+	->
+		ModuleInfo = ModuleInfo1,
+		Inst = Inst1
+	;
+		error("bind_inst_to_functor: mode error")
+	).
+
+%-----------------------------------------------------------------------------%
+
+instmap__pre_lambda_update(ModuleInfo, Vars, Modes, InstMap0, InstMap) :-
+	mode_list_get_initial_insts(Modes, ModuleInfo, Insts),
+	assoc_list__from_corresponding_lists(Vars, Insts, VarInsts),
+	instmap_delta_from_assoc_list(VarInsts, InstMapDelta),
+	instmap__apply_instmap_delta(InstMap0, InstMapDelta, InstMap).
 
 %-----------------------------------------------------------------------------%
 
@@ -454,6 +608,32 @@ instmap__merge_var([InstMap | InstMaps], Var, ModuleInfo0,
 		Inst = not_reached
 	).
 
+%-----------------------------------------------------------------------------%
+
+merge_instmap_deltas(InstMap, NonLocals, InstMapDeltaList, MergedDelta,
+		ModuleInfo0, ModuleInfo) :-
+	(
+		InstMapDeltaList = [],
+		error("merge_instmap_deltas: empty instmap_delta list.")
+	;
+		InstMapDeltaList = [Delta|Deltas],
+		merge_instmap_deltas(InstMap, NonLocals, Delta, Deltas,
+			MergedDelta, ModuleInfo0, ModuleInfo)
+	).
+
+:- pred merge_instmap_deltas(instmap, set(var), instmap_delta,
+		list(instmap_delta), instmap_delta, module_info, module_info).
+:- mode merge_instmap_deltas(in, in, in, in, out, in, out) is det.
+
+merge_instmap_deltas(_InstMap, _NonLocals, MergedDelta, [], MergedDelta,
+		ModuleInfo, ModuleInfo).
+merge_instmap_deltas(InstMap, NonLocals, MergedDelta0, [Delta|Deltas],
+		MergedDelta, ModuleInfo0, ModuleInfo) :-
+	merge_instmap_delta(InstMap, NonLocals, MergedDelta0, Delta,
+		MergedDelta1, ModuleInfo0, ModuleInfo1),
+	merge_instmap_deltas(InstMap, NonLocals, MergedDelta1, Deltas,
+		MergedDelta, ModuleInfo1, ModuleInfo).
+
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
 
@@ -517,51 +697,61 @@ instmap__no_output_vars_2([Var | Vars], InstMap0, InstMapDelta, ModuleInfo) :-
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
 
-	% Given two instmap deltas, merge them to produce a new instmap.
+	% Given two instmap deltas, merge them to produce a new instmap_delta.
 
-merge_instmap_delta(unreachable, InstMap, InstMap) --> [].
-merge_instmap_delta(reachable(InstMapping), unreachable,
+merge_instmap_delta(_, _, unreachable, InstMapDelta, InstMapDelta) --> [].
+merge_instmap_delta(_, _, reachable(InstMapping), unreachable,
 				reachable(InstMapping)) --> [].
-merge_instmap_delta(reachable(InstMappingA), reachable(InstMappingB),
-			reachable(InstMapping)) -->
-	merge_instmapping_delta(InstMappingA, InstMappingB, InstMapping).
+merge_instmap_delta(InstMap, NonLocals, reachable(InstMappingA),
+		reachable(InstMappingB), reachable(InstMapping)) -->
+	merge_instmapping_delta(InstMap, NonLocals, InstMappingA,
+		InstMappingB, InstMapping).
 
-:- pred merge_instmapping_delta(instmapping, instmapping, instmapping,
-				module_info, module_info).
-:- mode merge_instmapping_delta(in, in, out, in, out) is det.
+:- pred merge_instmapping_delta(instmap, set(var), instmapping, instmapping,
+		instmapping, module_info, module_info).
+:- mode merge_instmapping_delta(in, in, in, in, out, in, out) is det.
 
-merge_instmapping_delta(InstMappingA, InstMappingB, InstMapping) -->
+merge_instmapping_delta(InstMap, NonLocals, InstMappingA,
+		InstMappingB, InstMapping) -->
 	{ map__keys(InstMappingA, VarsInA) },
-	merge_instmapping_delta_2(VarsInA, InstMappingA, InstMappingB,
-		InstMapping).
+	{ map__keys(InstMappingB, VarsInB) },
+	{ set__sorted_list_to_set(VarsInA, SetofVarsInA) },
+	{ set__insert_list(SetofVarsInA, VarsInB, SetofVars0) },
+	{ set__intersect(SetofVars0, NonLocals, SetofVars) },
+	{ map__init(InstMapping0) },
+	{ set__to_sorted_list(SetofVars, ListofVars) },
+	merge_instmapping_delta_2(ListofVars, InstMap, InstMappingA,
+		InstMappingB, InstMapping0, InstMapping).
 
-:- pred merge_instmapping_delta_2(list(var), instmapping, instmapping,
-				instmapping, module_info, module_info).
-:- mode merge_instmapping_delta_2(in, in, in, out, in, out) is det.
+:- pred merge_instmapping_delta_2(list(var), instmap, instmapping, instmapping,
+			instmapping, instmapping, module_info, module_info).
+:- mode merge_instmapping_delta_2(in, in, in, in, in, out, in, out) is det.
 
-merge_instmapping_delta_2([], _, InstMapping, InstMapping, ModInfo, ModInfo).
-merge_instmapping_delta_2([Var | Vars], MergeInstMapping, InstMapping0,
-			InstMapping, ModuleInfo0, ModuleInfo) :-
-	map__lookup(MergeInstMapping, Var, MergeInst),
-	( map__search(InstMapping0, Var, Inst0) ->
-	    ( inst_merge(Inst0, MergeInst, ModuleInfo0, Inst, ModuleInfoPrime) ->
-		ModuleInfo1 = ModuleInfoPrime,
-		map__det_update(InstMapping0, Var, Inst, InstMapping1)
-	    ;
-		error("merge_instmapping_delta_2: unexpected mode error")
-	    )
+merge_instmapping_delta_2([], _, _, _, InstMapping, InstMapping,
+		ModInfo, ModInfo).
+merge_instmapping_delta_2([Var | Vars], InstMap, InstMappingA, InstMappingB,
+			InstMapping0, InstMapping, ModuleInfo0, ModuleInfo) :-
+	( map__search(InstMappingA, Var, InstInA) ->
+		InstA = InstInA
 	;
-	    % if a variable only occurs in one of the instmap deltas,
-	    % then mode correctness means that the delta must be adding
-	    % information only, not binding the variable any further;
-	    % since we don't know which path will actually get executed,
-	    % we should not add that information - the merged delta should
-	    % not have any entry for that variable.
-	    ModuleInfo1 = ModuleInfo0,
-	    InstMapping1 = InstMapping0
+		instmap__lookup_var(InstMap, Var, InstA)
 	),
-	merge_instmapping_delta_2(Vars, MergeInstMapping, InstMapping1,
-				InstMapping, ModuleInfo1, ModuleInfo).
+	( map__search(InstMappingB, Var, InstInB) ->
+		InstB = InstInB
+	;
+		instmap__lookup_var(InstMap, Var, InstB)
+	),
+	(
+		inst_merge(InstA, InstB, ModuleInfo0,
+			Inst, ModuleInfoPrime)
+	->
+		ModuleInfo1 = ModuleInfoPrime,
+		map__det_insert(InstMapping0, Var, Inst, InstMapping1)
+	;
+		error("merge_instmapping_delta_2: unexpected mode error")
+	),
+	merge_instmapping_delta_2(Vars, InstMap, InstMappingA, InstMappingB,
+		InstMapping1, InstMapping, ModuleInfo1, ModuleInfo).
 
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
@@ -574,6 +764,9 @@ instmap_delta_apply_sub(reachable(OldInstMapping), Must, Sub,
 	instmap_delta_apply_sub_2(InstMappingAL, Must, Sub,
 		InstMapping0, InstMapping).
 
+instmap__apply_sub(InstMap0, Must, Sub, InstMap) :-
+	instmap_delta_apply_sub(InstMap0, Must, Sub, InstMap).
+
 :- pred instmap_delta_apply_sub_2(assoc_list(var, inst), bool, map(var, var),
 		instmapping, instmapping).
 :- mode instmap_delta_apply_sub_2(in, in, in, in, out) is det.
