@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1996 University of Melbourne.
+% Copyright (C) 1996-1997 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -40,8 +40,9 @@
 :- import_module io.
 :- import_module hlds_module, modules.
 
-:- pred intermod__write_optfile(module_info, io__state, io__state).
-:- mode intermod__write_optfile(in, di, uo) is det.
+:- pred intermod__write_optfile(module_info, module_info,
+				io__state, io__state).
+:- mode intermod__write_optfile(in, out, di, uo) is det.
 
 	% Add the items from the .opt files of imported modules to
 	% the items for this module.
@@ -66,14 +67,18 @@
 :- import_module hlds_data, hlds_goal, hlds_pred, hlds_out, inlining, llds.
 :- import_module mercury_to_mercury, mode_util, modules.
 :- import_module options, passes_aux, prog_data, prog_io, prog_out, prog_util.
-:- import_module special_pred, typecheck, type_util.
+:- import_module special_pred, typecheck, type_util, instmap, (inst).
 
 %-----------------------------------------------------------------------------%
 
-% Open the file "<module-name>.opt", and write out the
+% Open the file "<module-name>.opt.tmp", and write out the
 % declarations and clauses for intermodule optimization.
+% Note that update_interface and touch_interface_datestamp
+% are called from mercury_compile.m since they must be called
+% after unused_args.m appends its information to the .opt.tmp
+% file.
 
-intermod__write_optfile(ModuleInfo0) -->
+intermod__write_optfile(ModuleInfo0, ModuleInfo) -->
 	{ module_info_name(ModuleInfo0, ModuleName) },
 	{ string__append(ModuleName, ".opt.tmp", TmpName) },
 	io__tell(TmpName, Result2),
@@ -82,7 +87,8 @@ intermod__write_optfile(ModuleInfo0) -->
 		{ io__error_message(Err2, Msg2) },
 		io__stderr_stream(ErrStream2),
 		io__write_string(ErrStream2, Msg2),
-		io__set_exit_status(1)
+		io__set_exit_status(1),
+		{ ModuleInfo = ModuleInfo0 }
 	;
 		{ Result2 = ok },
 		{ module_info_predids(ModuleInfo0, PredIds) },
@@ -95,25 +101,28 @@ intermod__write_optfile(ModuleInfo0) -->
 				IntermodInfo2) },
 		{ intermod_info_get_pred_decls(PredDeclsSet,
 				IntermodInfo2, IntermodInfo3) },
-		{ intermod_info_get_module_info(ModuleInfo,
-				IntermodInfo3, _) },
-		{ module_info_insts(ModuleInfo, Insts) },
+		{ intermod_info_get_module_info(ModuleInfo1,
+				IntermodInfo3, IntermodInfo4) },
+		{ module_info_insts(ModuleInfo1, Insts) },
 		{ inst_table_get_user_insts(Insts, UserInsts) },
 		{ user_inst_table_get_inst_defns(UserInsts, InstDefns) },
-		{ module_info_modes(ModuleInfo, Modes) },
+		{ module_info_modes(ModuleInfo1, Modes) },
 		{ mode_table_get_mode_defns(Modes, ModeDefns) },
 		{ set__to_sorted_list(PredDeclsSet, PredDecls) },
-		{ intermod__gather_modes(ModuleInfo, ModeDefns, InstDefns,
-				PredDecls, IntermodInfo2, IntermodInfo) },
+		{ intermod__gather_modes(ModuleInfo1, ModeDefns, InstDefns,
+				PredDecls, IntermodInfo4, IntermodInfo) },
 		intermod__write_intermod_info(IntermodInfo),
 		io__told,
-		{ string__append(ModuleName, ".opt", OptName) },
-		update_interface(OptName),
-		touch_interface_datestamp(ModuleName, ".optdate")
+		globals__io_lookup_bool_option(intermod_unused_args,
+			UnusedArgs),
+		( { UnusedArgs = yes } ->
+			{ do_adjust_pred_import_status(IntermodInfo,
+				ModuleInfo1, ModuleInfo) }
+		;
+			{ ModuleInfo = ModuleInfo1 }
+		)
 	).
 
-
-
 	% a collection of stuff to go in the .opt file
 :- type intermod_info
 		 ---> info(
@@ -166,13 +175,16 @@ intermod__gather_preds([PredId | PredIds], CollectTypes, InlineThreshold) -->
 			% Don't export builtins since they will be
 			% recreated in the importing module anyway.
 			{ \+ code_util__compiler_generated(PredInfo0) },
-			{ \+ code_util__predinfo_is_builtin(ModuleInfo0,
-					PredInfo0) },
+			{ \+ code_util__predinfo_is_builtin(PredInfo0) },
 			(
 				{ inlining__is_simple_goal(Goal,
-						InlineThreshold) }
+						InlineThreshold) },
+				{ pred_info_get_marker_list(PredInfo0, 
+					Markers) },
+				{ \+ list__member(request(no_inline), 
+					Markers) }
 			;
-				{ pred_info_is_inlined(PredInfo0) }
+				{ pred_info_requested_inlining(PredInfo0) }
 			;
 				{ has_ho_input(ModuleInfo0, ProcInfo) }
 			)
@@ -328,7 +340,7 @@ intermod__gather_abstract_exported_types -->
 
 	% Go over the goal of an exported proc looking for proc decls, types,
 	% insts and modes that we need to write to the optfile.
-:- pred intermod__traverse_goal(hlds__goal::in, hlds__goal::out, bool::out,
+:- pred intermod__traverse_goal(hlds_goal::in, hlds_goal::out, bool::out,
 			intermod_info::in, intermod_info::out) is det.
 
 intermod__traverse_goal(conj(Goals0) - Info, conj(Goals) - Info, DoWrite) -->
@@ -347,11 +359,20 @@ intermod__traverse_goal(
 	intermod_info_get_tvarset(TVarSet),
 	( { invalid_pred_id(PredId0) } ->
 		{ typecheck__resolve_pred_overloading(ModuleInfo, Args,
-			VarTypes, TVarSet, PredName0, PredName, PredId) }
+			VarTypes, TVarSet, PredName0, PredName1, PredId) }
 	;
 		{ PredId = PredId0 },
-		{ PredName = PredName0 }
+		{ PredName1 = PredName0 }
 	),	
+	(
+		{ PredName1 = qualified(_, _) },
+		{ PredName = PredName1 }
+	;
+		{ PredName1 = unqualified(Name) },
+		{ module_info_pred_info(ModuleInfo, PredId, PredInfo) },
+		{ pred_info_module(PredInfo, PredModule) },
+		{ PredName = qualified(PredModule, Name) }
+	),
 	(
 		% We don't need to export complicated unification
 		% pred declarations, since they will be recreated when 
@@ -359,18 +380,12 @@ intermod__traverse_goal(
 		{ MaybeUnifyContext = no }
 	->
 		intermod_info_add_proc(PredId, DoWrite)
-	;
-		intermod_info_get_module_info(ModuleInfo),
-		{ module_info_pred_info(ModuleInfo, PredId, PredInfo) },
-		{ \+ code_util__compiler_generated(PredInfo) }
-	->
-		intermod_info_add_proc(PredId, DoWrite)
 	;
 		{ DoWrite = yes }
 	).
 
-intermod__traverse_goal(higher_order_call(A,B,C,D,E) - Info,
-			higher_order_call(A,B,C,D,E) - Info, yes) --> [].
+intermod__traverse_goal(higher_order_call(A,B,C,D,E,F) - Info,
+			higher_order_call(A,B,C,D,E,F) - Info, yes) --> [].
 
 intermod__traverse_goal(switch(A, B, Cases0, D) - Info,
 		switch(A, B, Cases, D) - Info, DoWrite) -->
@@ -380,7 +395,7 @@ intermod__traverse_goal(switch(A, B, Cases0, D) - Info,
 	% or function calls.
 intermod__traverse_goal(unify(LVar, RHS0, C, D, E) - Info,
 			unify(LVar, RHS, C, D, E) - Info, DoWrite) -->
-	fix_function_or_higher_order_in_unify_rhs(LVar, RHS0, RHS, DoWrite).
+	intermod__module_qualify_unify_rhs(LVar, RHS0, RHS, DoWrite).
 
 intermod__traverse_goal(not(Goal0) - Info, not(Goal) - Info, DoWrite) -->
 	intermod__traverse_goal(Goal0, Goal, DoWrite).
@@ -398,11 +413,11 @@ intermod__traverse_goal(if_then_else(Vars, Cond0, Then0, Else0, SM) - Info,
 
 	% Inlineable exported pragma_c_code goals can't use any
 	% non-exported types, so we just write out the clauses. 
-intermod__traverse_goal(pragma_c_code(A,B,C,D,E,F) - Info,
-			pragma_c_code(A,B,C,D,E,F) - Info, yes) --> [].
+intermod__traverse_goal(pragma_c_code(A,B,C,D,E,F,G,H) - Info,
+			pragma_c_code(A,B,C,D,E,F,G,H) - Info, yes) --> [].
 
 
-:- pred intermod__traverse_list_of_goals(hlds__goals::in, hlds__goals::out,
+:- pred intermod__traverse_list_of_goals(hlds_goals::in, hlds_goals::out,
 		bool::out, intermod_info::in, intermod_info::out) is det.
 
 intermod__traverse_list_of_goals([], [], yes) --> [].
@@ -439,7 +454,8 @@ intermod_info_add_proc(PredId, DoWrite) -->
 	{ module_info_pred_info(ModuleInfo, PredId, PredInfo) },
 	{ pred_info_import_status(PredInfo, Status) },
 	{ pred_info_procids(PredInfo, ProcIds) },
-	( { ProcIds = [] } ->
+	{ pred_info_get_marker_list(PredInfo, Markers) },
+	( { list__member(request(infer_modes), Markers) } ->
 		% Don't write this pred if it calls preds without mode decls.
 		{ DoWrite = no }
 	; 
@@ -480,17 +496,13 @@ intermod_info_add_proc(PredId, DoWrite) -->
 		{ DoWrite = yes }
 	).
 
-	% Resolve overloading and module qualify function calls and
-	% higher-order predicate constants in a unify_rhs.
-	% This has to wait until I implement module qualification of
-	% functions.
-:- pred fix_function_or_higher_order_in_unify_rhs(var::in, unify_rhs::in,
+	% Resolve overloading and module qualify everything in a unify_rhs.
+:- pred intermod__module_qualify_unify_rhs(var::in, unify_rhs::in,
 		unify_rhs::out, bool::out, intermod_info::in,
 		intermod_info::out) is det.
 
-fix_function_or_higher_order_in_unify_rhs(_, var(Var), var(Var), yes) --> [].
-fix_function_or_higher_order_in_unify_rhs(_LVar,
-		lambda_goal(A,B,Modes,D,Goal0),
+intermod__module_qualify_unify_rhs(_, var(Var), var(Var), yes) --> [].
+intermod__module_qualify_unify_rhs(_LVar, lambda_goal(A,B,Modes,D,Goal0),
 		lambda_goal(A,B,Modes,D,Goal), DoWrite) -->
 	intermod__traverse_goal(Goal0, Goal, DoWrite),
 	intermod_info_get_module_info(ModuleInfo),
@@ -502,43 +514,75 @@ fix_function_or_higher_order_in_unify_rhs(_LVar,
 	intermod__gather_proc_modes(ModuleInfo, ModeDefns,
 				UserInstDefns, Modes).
 
-	% Check if the functor is actually a function call or a higher-order
-	% pred constant. If so, module qualify.
-fix_function_or_higher_order_in_unify_rhs(LVar, functor(Functor0, Vars),
+	% Check if the functor is a function call, a higher-order
+	% term, or an unqualified symbol. If so, module qualify.
+	% For function calls and higher-order terms, call intermod__add_proc
+	% so that the predicate or function will be exported if necessary.
+intermod__module_qualify_unify_rhs(LVar, functor(Functor0, Vars),
 				functor(Functor, Vars), DoWrite) -->
 	intermod_info_get_module_info(ModuleInfo),
 	{ module_info_get_predicate_table(ModuleInfo, PredTable) },
 	intermod_info_get_tvarset(TVarSet),
 	intermod_info_get_var_types(VarTypes),
 	(
-		{
-		Functor0 = cons(unqualified(FuncName), Arity),
-		predicate_table_search_func_name_arity(PredTable, FuncName,
-				Arity, PredIds),
-		list__append(Vars, [LVar], FuncArgs),
-		map__apply_to_list(FuncArgs, VarTypes, ArgTypes),
-		typecheck__find_matching_pred_id(PredIds, ModuleInfo,
-				TVarSet, ArgTypes, PredId, QualifiedFuncName)
-		}
+		%
+		% Is it a function call?
+		%
+		{ Functor0 = cons(FuncName, Arity) },
+		{ predicate_table_search_func_sym_arity(PredTable,
+				FuncName, Arity, PredIds) },
+		{ list__append(Vars, [LVar], FuncArgs) },
+		{ map__apply_to_list(FuncArgs, VarTypes, FuncArgTypes) },
+		{ typecheck__find_matching_pred_id(PredIds, ModuleInfo,
+			TVarSet, FuncArgTypes, PredId, QualifiedFuncName) }
 	->
-			% The unification is really a function call
+		%
+		% Yes, it is a function call.
+		% Module-qualify it.
+		% Make sure that the called function will be exported.
+		%
 		{ Functor = cons(QualifiedFuncName, Arity) },
 		intermod_info_add_proc(PredId, DoWrite)
 	;
+		%
+		% Is this a higher-order predicate or higher-order function
+		% term?
+		%
+		{ Functor0 = cons(PredName, Arity) },
 		intermod_info_get_var_types(VarTypes),
-		{
-		Functor0 = cons(unqualified(PredName), Arity),
-		map__lookup(VarTypes, LVar, LVarType),
-		type_is_higher_order(LVarType, PredOrFunc, PredArgTypes),
-		get_pred_id_and_proc_id(PredName, Arity, PredOrFunc,
-			PredArgTypes, ModuleInfo, PredId, _ProcId)
-		}
+		{ map__lookup(VarTypes, LVar, LVarType) },
+		{ type_is_higher_order(LVarType, PredOrFunc, PredArgTypes) }
 	->
-			% The unification creates a higher-order pred constant.
-		{ module_info_pred_info(ModuleInfo, PredId, PredInfo) },
-		{ pred_info_module(PredInfo, Module) },
-		{ Functor = cons(qualified(Module, PredName), Arity) },
-		intermod_info_add_proc(PredId, DoWrite)
+		%
+		% Yes, the unification creates a higher-order term.
+		% Make sure that the predicate/function is exported.
+		%
+		{ map__apply_to_list(Vars, VarTypes, Types) },
+		{ list__append(Types, PredArgTypes, ArgTypes) },
+		{ get_pred_id_and_proc_id(PredName, PredOrFunc,
+			TVarSet, ArgTypes, ModuleInfo, PredId, _ProcId) },
+		intermod_info_add_proc(PredId, DoWrite),
+		%
+		% Module-qualify it, if necessary.
+		%
+		{ PredName = unqualified(UnqualPredName) ->
+			predicate_module(ModuleInfo, PredId, Module),
+			QualifiedPredName = qualified(Module, UnqualPredName),
+			Functor = cons(QualifiedPredName, Arity)
+		;
+			Functor = Functor0
+		}
+	;
+		%
+		% Is it an unqualified functor symbol?
+		%
+		{ Functor0 = cons(unqualified(ConsName), ConsArity) },
+		{ map__lookup(VarTypes, LVar, VarType) },
+		{ type_to_type_id(VarType, TypeId, _) },
+		{ TypeId = qualified(TypeModule, _) - _ }
+	->
+		{ Functor = cons(qualified(TypeModule, ConsName), ConsArity) },
+		{ DoWrite = yes }
 	;
 		{ Functor = Functor0 },
 		{ DoWrite = yes }
@@ -568,7 +612,7 @@ intermod__gather_pred_modes(_, _, _, _, []) --> [].
 intermod__gather_pred_modes(ModuleInfo, Modes, Insts, Procs, [ProcId | ProcIds])
 		-->
 	{ map__lookup(Procs, ProcId, ProcInfo) }, 
-	{ proc_info_argmodes(ProcInfo, ArgModes) },
+	{ proc_info_declared_argmodes(ProcInfo, ArgModes) },
 	intermod__gather_proc_modes(ModuleInfo, Modes, Insts, ArgModes),
 	intermod__gather_pred_modes(ModuleInfo, Modes, Insts, Procs, ProcIds).
 
@@ -587,7 +631,7 @@ intermod__gather_proc_modes(ModuleInfo, ModeTable,
 		{ list__length(Args, Arity) },
 		{ ModeId = Name - Arity },
 		{ map__lookup(ModeTable, ModeId, ModeDefn) },
-		{ ModeDefn = hlds__mode_defn(_,_,_,_,_, Status) },
+		{ ModeDefn = hlds_mode_defn(_,_,_,_,_, Status) },
 		( { Status = local } ->
 			intermod_info_get_modes(ModesToExport0),
 			{ set__insert(ModesToExport0, ModeId,
@@ -631,7 +675,7 @@ intermod__add_inst(UserInstTable, Inst) -->
 		{ list__length(InstArgs, Arity) },
 		{ InstId = Name - Arity },
 		{ map__lookup(UserInstTable, InstId, InstDefn) },
-		{ InstDefn = hlds__inst_defn(_,_,_,_,_, Status) },
+		{ InstDefn = hlds_inst_defn(_,_,_,_,_, Status) },
 		( { Status = local } ->
 			intermod_info_get_insts(InstsToExport0),
 			{ set__insert(InstsToExport0, InstId,
@@ -672,14 +716,14 @@ intermod__write_intermod_info(IntermodInfo) -->
 	mercury_output_bracketed_constant(term__atom(ModName)),
 	io__write_string(".\n"),
 	( { Modules \= [] } ->
-		io__write_string(":- import_module "),
+		io__write_string(":- use_module "),
 		intermod__write_modules(Modules)
 	;
 		[]
 	),
 	% Disable verbose dumping of clauses.
-	globals__io_lookup_option(verbose_dump_hlds, VerboseDump),
-	globals__io_set_option(verbose_dump_hlds, bool(no)),
+	globals__io_lookup_string_option(verbose_dump_hlds, VerboseDump),
+	globals__io_set_option(verbose_dump_hlds, string("")),
 	( { WriteHeader = yes } ->
 		{ module_info_get_c_header(ModuleInfo, CHeader) },
 		intermod__write_c_header(CHeader)
@@ -697,7 +741,7 @@ intermod__write_intermod_info(IntermodInfo) -->
 	intermod__write_insts(ModuleInfo, InstDefns, Insts),
 	intermod__write_pred_decls(ModuleInfo, PredDecls),
 	intermod__write_preds(ModuleInfo, Preds),
-	globals__io_set_option(verbose_dump_hlds, VerboseDump).
+	globals__io_set_option(verbose_dump_hlds, string(VerboseDump)).
 
 
 :- pred intermod__write_modules(list(module_name)::in,
@@ -735,8 +779,9 @@ intermod__write_types(ModuleInfo, TypeTable, [TypeId | TypeIds]) -->
 	{ hlds_data__get_type_defn_body(TypeDefn, Body) },
 	{ hlds_data__get_type_defn_context(TypeDefn, Context) },
 	(
-		{ Body = du_type(Ctors, _, _) },
-		mercury_output_type_defn(VarSet, du_type(Name, Args, Ctors),
+		{ Body = du_type(Ctors, _, _, MaybeEqualityPred) },
+		mercury_output_type_defn(VarSet,
+				du_type(Name, Args, Ctors, MaybeEqualityPred),
 				Context)
 	;
 		{ Body = uu_type(_) },
@@ -759,7 +804,7 @@ intermod__write_modes(_, _, []) --> [].
 intermod__write_modes(ModuleInfo, ModeTable, [ModeId | Modes]) -->
 	{ ModeId = SymName - _Arity },
 	{ map__lookup(ModeTable, ModeId, ModeDefn) },
-	{ ModeDefn = hlds__mode_defn(Varset, Args, eqv_mode(Mode),
+	{ ModeDefn = hlds_mode_defn(Varset, Args, eqv_mode(Mode),
 							_, Context, _) },
 	mercury_output_mode_defn(
 			Varset,
@@ -773,10 +818,9 @@ intermod__write_modes(ModuleInfo, ModeTable, [ModeId | Modes]) -->
 
 intermod__write_insts(_, _, []) --> [].
 intermod__write_insts(ModuleInfo, UserInstTable, [Inst | Insts]) -->
-	io__write_char(':'),
 	{ Inst = SymName - _Arity },
 	{ map__lookup(UserInstTable, Inst, InstDefn) },
-	{ InstDefn = hlds__inst_defn(Varset, Args, Body, _, Context, _) },
+	{ InstDefn = hlds_inst_defn(Varset, Args, Body, _, Context, _) },
 	(
 		{ Body = eqv_inst(Inst2) },
 		mercury_output_inst_defn(
@@ -792,7 +836,6 @@ intermod__write_insts(ModuleInfo, UserInstTable, [Inst | Insts]) -->
 				Context
 		)
 	),
-	io__write_string(".\n"),
 	intermod__write_insts(ModuleInfo, UserInstTable, Insts).
 
 	% We need to write all the declarations for local predicates so
@@ -826,8 +869,10 @@ intermod__write_pred_decls(ModuleInfo, [PredId | PredIds]) -->
 		% same proc_id in the importing modules.
 	{ CompareProcId =
 		 lambda([ProcId1::in, ProcId2::in, Result::out] is det, (
-			ActualProcId1 is ProcId1 mod 10000,
-			ActualProcId2 is ProcId2 mod 10000,
+			proc_id_to_int(ProcId1, ProcInt1),
+			ActualProcId1 is ProcInt1 mod 10000,
+			proc_id_to_int(ProcId2, ProcInt2),
+			ActualProcId2 is ProcInt2 mod 10000,
 			compare(Result, ActualProcId1, ActualProcId2)
 		)) },
 	{ list__sort(CompareProcId, ProcIds, SortedProcIds) },
@@ -842,12 +887,13 @@ intermod__write_pred_decls(ModuleInfo, [PredId | PredIds]) -->
 intermod__write_pred_modes(_, _, _, []) --> [].
 intermod__write_pred_modes(Procs, SymName, PredOrFunc, [ProcId | ProcIds]) -->
 	{ map__lookup(Procs, ProcId, ProcInfo) },
-	{ proc_info_argmodes(ProcInfo, ArgModes) },
+	{ proc_info_maybe_declared_argmodes(ProcInfo, MaybeArgModes) },
 	{ proc_info_declared_determinism(ProcInfo, MaybeDetism) },
-	{ MaybeDetism = yes(Detism0) ->
+	{ MaybeArgModes = yes(ArgModes0), MaybeDetism = yes(Detism0) ->
+		ArgModes = ArgModes0,
 		Detism = Detism0
 	;
-		error("Attempt to write pred mode decl without det decl")
+		error("intermod__write_pred_modes: attempt to write undeclared mode")
 	},
 	{ proc_info_context(ProcInfo, Context) },
 	{ varset__init(Varset) },
@@ -886,10 +932,10 @@ intermod__write_preds(ModuleInfo, [PredId | PredIds]) -->
 		intermod__write_c_code(SymName, PredOrFunc, HeadVars, Varset,
 						Clauses, Procs)
 	;
-		{ pred_info_typevarset(PredInfo, _TVarSet) },
-		hlds_out__write_clauses(1, ModuleInfo, PredId, Varset,
-				HeadVars, PredOrFunc, Clauses, no)
-		%		HeadVars, Clauses, yes(TVarSet, VarTypes))
+		% { pred_info_typevarset(PredInfo, TVarSet) },
+		hlds_out__write_clauses(1, ModuleInfo, PredId, Varset, no,
+			HeadVars, PredOrFunc, Clauses, no)
+		%	HeadVars, Clauses, yes(TVarSet, VarTypes))
 	),
 	intermod__write_preds(ModuleInfo, PredIds).
 
@@ -911,13 +957,8 @@ intermod__write_pragmas(SymName, Arity, [MarkerStatus | Markers]) -->
 		 ;	{ Marker = infer_modes }
 		)
 	->
-		io__write_string(":- pragma "),
-		hlds_out__write_marker(Marker),
-		io__write_string("("),
-		mercury_output_bracketed_sym_name(SymName),
-		io__write_char(('/')),
-		io__write_int(Arity),
-		io__write_string(").\n")
+		{ hlds_out__marker_name(Marker, Name) },
+		mercury_output_pragma_decl(SymName, Arity, Name)
 	;
 		[]
 	),
@@ -938,18 +979,18 @@ intermod__write_c_code(SymName, PredOrFunc, HeadVars, Varset,
 			{ Goal = conj(Goals) - _ },
 			{ list__filter(
 				lambda([X::in] is semidet, (
-					X = pragma_c_code(_, _, _, _, _, _) - _
+					X = pragma_c_code(_,_,_,_,_,_,_,_) - _
 				)),
 				Goals, [CCodeGoal]) },
-			{ CCodeGoal = pragma_c_code(CCode, IsRec,
-						_, _, Vars, _) - _ }
+			{ CCodeGoal = pragma_c_code(CCode, MayCallMercury,
+						_, _, Vars, _, _, _) - _ }
 		;
-			{ Goal = pragma_c_code(CCode, IsRec,
-						_, _, Vars, _) - _ }
+			{ Goal = pragma_c_code(CCode, MayCallMercury,
+						_, _, Vars, _, _, _) - _ }
 		)
 	->	
 		intermod__write_c_clauses(Procs, ProcIds, PredOrFunc, CCode, 
-					IsRec, Vars, Varset, SymName)
+					MayCallMercury, Vars, Varset, SymName)
 	;
 		{ error("intermod__write_c_code called with non c_code goal") }
 	),
@@ -957,19 +998,25 @@ intermod__write_c_code(SymName, PredOrFunc, HeadVars, Varset,
 				Clauses, Procs).
 
 :- pred intermod__write_c_clauses(proc_table::in, list(proc_id)::in, 
-		pred_or_func::in, string::in, c_is_recursive::in, list(var)::in,
-		varset::in, sym_name::in, io__state::di, io__state::uo) is det.
+		pred_or_func::in, string::in, may_call_mercury::in,
+		list(var)::in, varset::in, sym_name::in,
+		io__state::di, io__state::uo) is det.
 
 intermod__write_c_clauses(_, [], _, _, _, _, _, _) --> [].
 intermod__write_c_clauses(Procs, [ProcId | ProcIds], PredOrFunc,
-			CCode, IsRec, Vars, Varset, SymName) -->
+			CCode, MayCallMercury, Vars, Varset, SymName) -->
 	{ map__lookup(Procs, ProcId, ProcInfo) },
-	{ proc_info_argmodes(ProcInfo, ArgModes) },
-	{ get_pragma_c_code_vars(Vars, Varset, ArgModes, PragmaVars) },
-	mercury_output_pragma_c_code(IsRec, SymName, PredOrFunc, PragmaVars,
-						Varset, CCode),
-	intermod__write_c_clauses(Procs, ProcIds, PredOrFunc, CCode, IsRec, 
-					Vars, Varset, SymName).
+	{ proc_info_maybe_declared_argmodes(ProcInfo, MaybeArgModes) },
+	( { MaybeArgModes = yes(ArgModes) } ->
+		{ get_pragma_c_code_vars(Vars, Varset, ArgModes, PragmaVars) },
+		% XXX will need modification for nondet pragma C code
+		mercury_output_pragma_c_code(MayCallMercury, SymName,
+			PredOrFunc, PragmaVars, no, Varset, CCode),
+		intermod__write_c_clauses(Procs, ProcIds, PredOrFunc, CCode,
+			MayCallMercury, Vars, Varset, SymName)
+	;
+		{ error("intermod__write_c_clauses: no mode declaration") }
+	).
 
 :- pred get_pragma_c_code_vars(list(var)::in, varset::in,
 		list(mode)::in, list(pragma_var)::out) is det.
@@ -1084,20 +1131,22 @@ intermod_info_set_tvarset(TVarSet, info(A,B,C,D,E,F,G,H,I,_),
 %-----------------------------------------------------------------------------%
 
 	% Make sure the labels of local preds needed by predicates in 
-	% the .opt file are exported.
+	% the .opt file are exported, and inhibit dead proc elimination
+	% on those preds.
 intermod__adjust_pred_import_status(Module0, Module) :-
 	init_intermod_info(Module0, Info0),
 	module_info_predids(Module0, PredIds),
-
-	% The .opt file may have been written with a higher 
-	% --inline-simple-threshold value, so to be safe all
-	% preds called by exported preds need to be exported too.
-
 	module_info_globals(Module0, Globals),
 	globals__lookup_int_option(Globals, intermod_inline_simple_threshold, 
 			Threshold),
 	intermod__gather_preds(PredIds, yes, Threshold, Info0, Info1),
-	intermod_info_get_pred_decls(PredDecls0, Info1, Info),
+	do_adjust_pred_import_status(Info1, Module0, Module).
+
+:- pred do_adjust_pred_import_status(intermod_info::in,
+		module_info::in, module_info::out) is det.
+
+do_adjust_pred_import_status(Info, Module0, Module) :-
+	intermod_info_get_pred_decls(PredDecls0, Info, _),
 	intermod_info_get_types(TypeIds0, Info, _),
 	set__to_sorted_list(PredDecls0, PredDecls),
 	set__to_sorted_list(TypeIds0, TypeIds),
@@ -1166,31 +1215,51 @@ intermod__grab_optfiles(Module0, Module, FoundError) -->
 	{ list__sort_and_remove_dups(DirectImports0, DirectImports1) },
 	read_optimization_interfaces(DirectImports1, [],
 		OptItems, no, OptError),
-	{ get_dependencies(OptItems, NewDeps0) },
+	{ get_dependencies(OptItems, NewImportDeps0, NewUseDeps0) },
+	{ list__append(NewImportDeps0, NewUseDeps0, NewDeps0) },
 	{ set__list_to_set(NewDeps0, NewDepsSet0) },
 	{ set__delete_list(NewDepsSet0, [ModuleName | DirectImports1],
 						NewDepsSet) },
 	{ set__to_sorted_list(NewDepsSet, NewDeps) },
-	globals__io_lookup_option(inhibit_warnings, NoWarn),
-		% stop a spurious warning
-	globals__io_set_option(inhibit_warnings, bool(yes)),
 	{ Module1 = module_imports(ModuleName, DirectImports1,
 					IndirectImports0, [], no) },
 	process_module_interfaces(NewDeps, [], Module1, Module2),
-	globals__io_set_option(inhibit_warnings, NoWarn),
 	{ Module2 = module_imports(_, DirectImports, IndirectImports,
 		InterfaceItems, IntError) },
-	{ ( IntError \= no ; OptError = yes ) ->
+	{ list__append(OptItems, InterfaceItems, NewItems) },
+	{ term__context_init("blah", 0, Context) },
+		% Let make_hlds know the opt_imported stuff is coming.
+	globals__io_lookup_bool_option(intermod_unused_args, UnusedArgs),
+	{ varset__init(Varset) },
+
+		%
+		% Get the :- pragma unused_args(...) declarations created
+		% when writing the .opt file for the current module. These
+		% are needed because we can probably remove more arguments
+		% with intermod_unused_args, but the interface for other
+		% modules must remain the same.
+		%
+
+	( { UnusedArgs = yes } ->
+		read_optimization_interfaces([ModuleName], [],
+				LocalItems, no, UAError),
+		{ IsPragmaUnusedArgs = lambda([Item::in] is semidet, (
+					Item = pragma(PragmaType) - _,
+					PragmaType = unused_args(_,_,_,_,_)
+				)) },
+		{ list__filter(IsPragmaUnusedArgs, LocalItems, PragmaItems) }
+	;
+		{ PragmaItems = [] },
+		{ UAError = no }
+	),
+	{ ( IntError \= no ; OptError = yes ; UAError = yes) ->
 		FoundError = yes
 	;
 		FoundError = no
 	},
-	{ list__append(OptItems, InterfaceItems, NewItems) },
-	{ term__context_init(Context) },
-		% Let make_hlds know the opt_imported stuff is coming.
-	{ varset__init(Varset) },
 	{ OptDefn = module_defn(Varset, opt_imported) - Context },
-	{ list__append(Items0, [OptDefn | NewItems], Items) }, 
+	{ list__append([OptDefn | PragmaItems], NewItems, NewItems2) }, 
+	{ list__append(Items0, NewItems2, Items) },
 	{ Module = module_imports(ModuleName, DirectImports,
 				IndirectImports, Items, no) }.
 
@@ -1212,10 +1281,16 @@ read_optimization_interfaces([Import | Imports],
 	maybe_write_string(VeryVerbose, "% done.\n"),
 
 	{ string__append(Import, ".opt", FileName) },
-        io__gc_call(
-		prog_io__read_module(FileName, Import, yes,
-				ModuleError, Messages, Items1)
-        ),
+	prog_io__read_module(FileName, Import, yes,
+			ModuleError, Messages, Items1),
+	update_error_status(FileName, ModuleError, Messages, Error0, Error1),
+	{ list__append(Items0, Items1, Items2) },
+	read_optimization_interfaces(Imports, Items2, Items, Error1, Error).
+
+:- pred update_error_status(string::in, module_error::in, message_list::in, 
+		bool::in, bool::out, io__state::di, io__state::uo) is det.
+
+update_error_status(FileName, ModuleError, Messages, Error0, Error1) -->
 	(
 		{ ModuleError = no },
 		{ Error1 = Error0 }
@@ -1240,8 +1315,6 @@ read_optimization_interfaces([Import | Imports],
 		;
 			{ Error1 = Error0 }	
 		)
-	),
-	{ list__append(Items0, Items1, Items2) },
-	read_optimization_interfaces(Imports, Items2, Items, Error1, Error).
+	).
 
 %-----------------------------------------------------------------------------%
