@@ -1,12 +1,20 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1995 University of Melbourne.
+% Copyright (C) 1994-1997 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
-
+%
 % jumpopt.m - optimize jumps to jumps.
-
+%
 % Author: zs.
+%
+% We first build up a bunch of tables giving information about labels.
+% This information includes the first instruction following each label,
+% the block following each label, whether a label is part of a procedure
+% epilog of various types (det, semidet or nondet).
+%
+% We then traverse the instruction list, using the information in the
+% tables to short-circuit jumps.
 
 %-----------------------------------------------------------------------------%
 
@@ -17,9 +25,6 @@
 :- import_module list, bool.
 :- import_module llds.
 
-	% Build up a table showing the first instruction following each label.
-	% Then traverse the instruction list, short-circuiting jump sequences.
-
 :- pred jumpopt__main(list(instruction), bool, bool, list(instruction), bool).
 :- mode jumpopt__main(in, in, in, out, out) is det.
 
@@ -131,7 +136,7 @@ jumpopt__build_forkmap([Instr - _Comment|Instrs], Sdprocmap,
 		Instr = label(Label),
 		opt_util__is_forkproceed_next(Instrs, Sdprocmap, Between)
 	->
-		map__set(Forkmap0, Label, Between, Forkmap1)
+		map__det_insert(Forkmap0, Label, Between, Forkmap1)
 	;
 		Forkmap1 = Forkmap0
 	),
@@ -157,6 +162,9 @@ jumpopt__build_forkmap([Instr - _Comment|Instrs], Sdprocmap,
 	%
 	% We handle if-vals by trying to turn them into the assignment
 	% of a boolean value to r1, or by short-circuiting the target label.
+	% We also try to eliminate a goto following an if-val, if we can
+	% do so by negating the condition and possibly also deleting a label
+	% between the if-val and the goto.
 
 :- pred jumpopt__instr_list(list(instruction), instr, instrmap, tailmap,
 	lvalmap, tailmap, tailmap, tailmap, tailmap, list(instruction), bool).
@@ -173,24 +181,31 @@ jumpopt__instr_list([Instr0 | Instrs0], PrevInstr, Instrmap, Blockmap,
 		Uinstr0 = call(Proc, label(RetLabel), GC, CallModel)
 	->
 		(
-			CallModel = det,
+			% Look for det style tailcalls. We look for this
+			% even if the call is semidet because one of the
+			% optimizations below turns a pair of semidet epilogs
+			% into a det epilog.
+			( CallModel = det ; CallModel = semidet ),
 			map__search(Procmap, RetLabel, Between0),
 			PrevInstr = livevals(Livevals) 
 		->
 			opt_util__filter_out_livevals(Between0, Between1),
 			list__append(Between1, [livevals(Livevals) - "",
 				goto(Proc) - Redirect], NewInstrs),
+			RemainInstrs = Instrs0,
 			Mod0 = yes
 		;
+			% Look for semidet style tailcalls.
 			CallModel = semidet,
-			map__search(Forkmap, RetLabel, Between0),
+			map__search(Forkmap, RetLabel, Between),
 			PrevInstr = livevals(Livevals) 
 		->
-			opt_util__filter_out_livevals(Between0, Between1),
-			list__append(Between1, [livevals(Livevals) - "",
+			list__append(Between, [livevals(Livevals) - "",
 				goto(Proc) - Redirect], NewInstrs),
+			RemainInstrs = Instrs0,
 			Mod0 = yes
 		;
+			% Look for nondet style tailcalls.
 			CallModel = nondet(yes),
 			map__search(Succmap, RetLabel, BetweenIncl),
 			BetweenIncl = [livevals(_) - _, goto(_) - _],
@@ -206,32 +221,38 @@ jumpopt__instr_list([Instr0 | Instrs0], PrevInstr, Instrmap, Blockmap,
 				livevals(Livevals) - "",
 				goto(Proc) - Redirect
 			],
+			RemainInstrs = Instrs0,
 			Mod0 = yes
 		;
+			% Short circuit the return label if possible.
 			map__search(Instrmap, RetLabel, RetInstr)
 		->
 			jumpopt__final_dest(RetLabel, RetInstr, Instrmap,
 				DestLabel, _DestInstr),
 			( RetLabel = DestLabel ->
 				NewInstrs = [Instr0],
+				RemainInstrs = Instrs0,
 				Mod0 = no
 			;
 				NewInstrs = [call(Proc, label(DestLabel),
 					GC, CallModel) - Redirect],
+				RemainInstrs = Instrs0,
 				Mod0 = yes
 			)
 		;
 			NewInstrs = [Instr0],
+			RemainInstrs = Instrs0,
 			Mod0 = no
 		)
 	;
 		Uinstr0 = goto(label(TargetLabel))
 	->
 		(
+			% Eliminate the goto if possible.
 			opt_util__is_this_label_next(TargetLabel, Instrs0, _)
 		->
-			% Eliminating is better than shortcircuiting.
 			NewInstrs = [],
+			RemainInstrs = Instrs0,
 			Mod0 = yes
 		;
 			PrevInstr = if_val(_, label(IfTargetLabel)),
@@ -244,38 +265,70 @@ jumpopt__instr_list([Instr0 | Instrs0], PrevInstr, Instrmap, Blockmap,
 			% We cannot eliminate the instruction here because
 			% that would require altering the if_val instruction.
 			NewInstrs = [Instr0],
+			RemainInstrs = Instrs0,
 			Mod0 = no
 		;
+			% Replace a jump to a det epilog with the epilog.
 			map__search(Procmap, TargetLabel, Between0)
 		->
 			jumpopt__adjust_livevals(PrevInstr, Between0, Between),
 			list__append(Between, [goto(succip) - "shortcircuit"],
 				NewInstrs),
+			RemainInstrs = Instrs0,
 			Mod0 = yes
 		;
+			% Replace a jump to a semidet epilog with the epilog.
 			map__search(Sdprocmap, TargetLabel, Between0)
 		->
 			jumpopt__adjust_livevals(PrevInstr, Between0, Between),
 			list__append(Between, [goto(succip) - "shortcircuit"],
 				NewInstrs),
+			RemainInstrs = Instrs0,
 			Mod0 = yes
 		;
+			% Replace a jump to a nondet epilog with the epilog.
 			map__search(Succmap, TargetLabel, BetweenIncl0)
 		->
 			jumpopt__adjust_livevals(PrevInstr, BetweenIncl0,
 				NewInstrs),
+			RemainInstrs = Instrs0,
 			Mod0 = yes
 		;
+			% Replace a jump to a non-epilog block with the
+			% block itself. These jumps are treated separately
+			% from jumps to epilog blocks, for two reasons.
+			% First, epilog blocks are always short, so we
+			% always want to replace jumps to them, whereas
+			% other blocks may be long, so we want to replace
+			% jumps to them only if the fulljumps option
+			% was given (if it wasn't, Blockmap will be empty).
+			% Second, non-epilog blocks may contain branches to
+			% other labels in this procedure, and we want to
+			% make sure that these are short-circuited.
+			% This short-circuiting is necessary because
+			% another optimization below eliminates labels,
+			% which is correct only if jumps to those labels
+			% are short-circuited everywhere.
 			map__search(Instrmap, TargetLabel, TargetInstr),
 			jumpopt__final_dest(TargetLabel, TargetInstr,
 				Instrmap, DestLabel, _DestInstr),
 			map__search(Blockmap, DestLabel, Block)
 		->
-			opt_util__filter_out_labels(Block, NewInstrs0),
-			jumpopt__adjust_livevals(PrevInstr, NewInstrs0,
-				NewInstrs),
+			opt_util__filter_out_labels(Block, FilteredBlock),
+			jumpopt__adjust_livevals(PrevInstr, FilteredBlock,
+				AdjustedBlock),
+			% Block may end with a goto to DestLabel. We avoid
+			% infinite expansion in such cases by removing
+			% DestLabel from Blockmap, though only while
+			% processing AdjustedBlock.
+			map__delete(Blockmap, DestLabel, CrippledBlockmap),
+			jumpopt__instr_list(AdjustedBlock, comment(""),
+				Instrmap, CrippledBlockmap, Lvalmap, Procmap,
+				Sdprocmap, Forkmap, Succmap, NewInstrs, _),
+			RemainInstrs = Instrs0,
 			Mod0 = yes
 		;
+			% Short-circuit the goto.
 			map__search(Instrmap, TargetLabel, TargetInstr)
 		->
 			jumpopt__final_dest(TargetLabel, TargetInstr,
@@ -287,14 +340,17 @@ jumpopt__instr_list([Instr0 | Instrs0], PrevInstr, Instrmap, Blockmap,
 				Canfallthrough),
 			( Canfallthrough = no ->
 				NewInstrs0 = [UdestInstr - Shorted],
+				RemainInstrs = Instrs0,
 				Mod0 = yes
 			;
 				( TargetLabel = DestLabel ->
 					NewInstrs0 = [Instr0],
+					RemainInstrs = Instrs0,
 					Mod0 = no
 				;
 					NewInstrs0 = [goto(label(DestLabel))
 						- Shorted],
+					RemainInstrs = Instrs0,
 					Mod0 = yes
 				)
 			),
@@ -305,12 +361,14 @@ jumpopt__instr_list([Instr0 | Instrs0], PrevInstr, Instrmap, Blockmap,
 				NewInstrs = NewInstrs0
 			)
 		;
-			% error("target label not in instrmap")
 			NewInstrs = [Instr0],
+			RemainInstrs = Instrs0,
 			Mod0 = no
 		)
 	; Uinstr0 = computed_goto(Index, LabelList0) ->
+		% Short-circuit all the destination labels.
 		jumpopt__short_labels(LabelList0, Instrmap, LabelList, Mod0),
+		RemainInstrs = Instrs0,
 		( Mod0 = yes ->
 			string__append(Comment0, " (some shortcircuits)",
 				Shorted),
@@ -320,6 +378,70 @@ jumpopt__instr_list([Instr0 | Instrs0], PrevInstr, Instrmap, Blockmap,
 		)
 	; Uinstr0 = if_val(Cond, label(TargetLabel)) ->
 		(
+			% Attempt to transform code such as
+			%
+			%	if (Cond) L2
+			% L1:
+			%	goto L3
+			% L2:	...
+			%
+			% into
+			%
+			%	if (! Cond) L3
+			% L2:	...
+			%
+			% The label L1 may be present or not. If it is present,
+			% we are eliminating it, which is possible only because
+			% we can short-circuit jumps to it (make them jump
+			% directly to L3). This may not be possible if L3 is
+			% a non-label code address; e.g. we cannot jump to
+			% non-label code addresses from computed gotos.
+
+			opt_util__skip_comments(Instrs0, Instrs1),
+			Instrs1 = [Instr1 | Instrs2],
+			( Instr1 = label(_) - _ ->
+				opt_util__skip_comments(Instrs2, Instrs3),
+				Instrs3 = [GotoInstr | AfterGoto],
+				HaveLabel = yes
+			;
+				Instr1 = GotoInstr,
+				AfterGoto = Instrs2,
+				HaveLabel = no
+			),
+			GotoInstr = goto(GotoTarget) - GotoComment,
+			( HaveLabel = no ; GotoTarget = label(_) ),
+			opt_util__skip_comments(AfterGoto, AfterGotoComments),
+			AfterGotoComments = [LabelInstr | _],
+			LabelInstr = label(TargetLabel) - _
+		->
+			code_util__neg_rval(Cond, NotCond),
+			NewInstr = if_val(NotCond, GotoTarget) - GotoComment,
+			NewInstrs = [],
+			% The transformed code may fit the pattern again,
+			% so make sure that we look for the pattern again
+			% by giving all of the transformed instructions to
+			% the recursive call. We can't go into an infinite
+			% loop because each application of the transformation
+			% strictly reduces the size of the code.
+			RemainInstrs = [NewInstr | AfterGoto],
+			Mod0 = yes
+		;
+			% Attempt to transform code such as
+			%
+			%	if (Cond) L2
+			%	r1 = TRUE
+			% 	<epilog>
+			%	...
+			% L2:
+			%	r1 = FALSE
+			%	<epilog>
+			%
+			% into
+			%
+			%	r1 = Cond
+			%	<epilog>
+			%
+
 			map__search(Instrmap, TargetLabel, TargetInstr)
 		->
 			jumpopt__final_dest(TargetLabel, TargetInstr,
@@ -341,43 +463,52 @@ jumpopt__instr_list([Instr0 | Instrs0], PrevInstr, Instrmap, Blockmap,
 					SuccessBR = true,
 					NewCond = Cond
 				),
-				\+ needs_workaround(reg(r(1)), NewCond)
+				\+ needs_workaround(reg(r, 1), NewCond)
 			->
-				NewAssign = assign(reg(r(1)), NewCond)
-					- "shortcircuited bool computation",
+				( NewCond = lval(reg(r, 1)) ->
+					NewAssign = comment("r1 = old r1") - ""
+				;
+					NewAssign = assign(reg(r, 1), NewCond) -
+						"shortcircuit bool computation"
+				),
 				Proceed = goto(succip) - "shortcircuit",
 				list__append([NewAssign | Between], [Proceed],
 					NewInstrs),
+				RemainInstrs = Instrs0,
 				Mod0 = yes
 			;
-				TargetLabel = DestLabel
+				% Short-circuit the destination.
+				TargetLabel \= DestLabel
 			->
-				NewInstrs = [Instr0],
-				Mod0 = no
-			;
 				string__append("shortcircuited jump: ",
 					Comment0, Shorted),
 				NewInstrs = [if_val(Cond, label(DestLabel))
 					- Shorted],
+				RemainInstrs = Instrs0,
 				Mod0 = yes
+			;
+				NewInstrs = [Instr0],
+				RemainInstrs = Instrs0,
+				Mod0 = no
 			)
 		;
-			% error("target label not in instrmap")
 			NewInstrs = [Instr0],
+			RemainInstrs = Instrs0,
 			Mod0 = no
 		)
 	;
 		NewInstrs = [Instr0],
+		RemainInstrs = Instrs0,
 		Mod0 = no
 	),
-	( Uinstr0 = comment(_) ->
+	( ( Uinstr0 = comment(_) ; NewInstrs = [] ) ->
 		NewPrevInstr = PrevInstr
 	;
 		NewPrevInstr = Uinstr0
 	),
-	jumpopt__instr_list(Instrs0, NewPrevInstr, Instrmap, Blockmap, Lvalmap,
-		Procmap, Sdprocmap, Forkmap, Succmap, Instrs1, Mod1),
-	list__append(NewInstrs, Instrs1, Instrs),
+	jumpopt__instr_list(RemainInstrs, NewPrevInstr, Instrmap, Blockmap,
+		Lvalmap, Procmap, Sdprocmap, Forkmap, Succmap, Instrs9, Mod1),
+	list__append(NewInstrs, Instrs9, Instrs),
 	( Mod0 = no, Mod1 = no ->
 		Mod = no
 	;
