@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1995 University of Melbourne.
+% Copyright (C) 1994-1997 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -63,16 +63,17 @@ allocate_stack_slots_in_proc(ProcInfo0, ModuleInfo, ProcInfo) :-
 % of the goal. The liveness information is computed from the liveness
 % delta annotations.
 
-:- pred build_live_sets_in_goal(hlds__goal, set(var), set(var), set(set(var)),
+:- pred build_live_sets_in_goal(hlds_goal, set(var), set(var), set(set(var)),
 	module_info, proc_info, set(var), set(var), set(set(var))).
 :- mode build_live_sets_in_goal(in, in, in, in, in, in, out, out, out) is det.
 
 build_live_sets_in_goal(Goal0 - GoalInfo, Liveness0, ResumeVars0, LiveSets0,
 		ModuleInfo, ProcInfo, Liveness, ResumeVars, LiveSets) :-
-	goal_info_get_pre_births(GoalInfo, PreBirths),
+	% note: we must be careful to apply deaths before births
 	goal_info_get_pre_deaths(GoalInfo, PreDeaths),
-	goal_info_get_post_births(GoalInfo, PostBirths),
+	goal_info_get_pre_births(GoalInfo, PreBirths),
 	goal_info_get_post_deaths(GoalInfo, PostDeaths),
+	goal_info_get_post_births(GoalInfo, PostBirths),
 
 	set__difference(Liveness0, PreDeaths, Liveness1),
 	set__union(Liveness1, PreBirths, Liveness2),
@@ -132,6 +133,7 @@ build_live_sets_in_goal(Goal0 - GoalInfo, Liveness0, ResumeVars0, LiveSets0,
 		% Add extra interference for variables that become live
 		% and variables that be come dead in this goal.
 	(
+/*******
 		% goal_is_atomic(Goal0)
 		fail
 		% NB: `fail' is a conservative approximation
@@ -141,6 +143,7 @@ build_live_sets_in_goal(Goal0 - GoalInfo, Liveness0, ResumeVars0, LiveSets0,
 		set__union(PreBirths, PostDeaths, ExtraInterference),
 		set__insert(LiveSets2, ExtraInterference, LiveSets)
 	;
+*******/
 		LiveSets = LiveSets2
 	).
 
@@ -156,8 +159,8 @@ build_live_sets_in_goal(Goal0 - GoalInfo, Liveness0, ResumeVars0, LiveSets0,
 	% `LiveSets' is the interference graph, i.e. the set of sets
 	% of variables which need to be on the stack at the same time.
 
-:- pred build_live_sets_in_goal_2(hlds__goal_expr, set(var), set(var),
-	set(set(var)), hlds__goal_info, module_info, proc_info,
+:- pred build_live_sets_in_goal_2(hlds_goal_expr, set(var), set(var),
+	set(set(var)), hlds_goal_info, module_info, proc_info,
 	set(var), set(var), set(set(var))).
 :- mode build_live_sets_in_goal_2(in, in, in, in, in, in, in, out, out, out)
 	is det.
@@ -213,7 +216,8 @@ build_live_sets_in_goal_2(some(_Vs, Goal0), Liveness0, ResumeVars0, LiveSets0,
 		ResumeVars = ResumeVars0
 	).
 
-build_live_sets_in_goal_2(higher_order_call(_, ArgVars, Types, Modes, Det),
+build_live_sets_in_goal_2(higher_order_call(_PredVar, ArgVars,
+				Types, Modes, Det, _IsPredOrFunc),
 		Liveness, ResumeVars0, LiveSets0,
 		GoalInfo, ModuleInfo, ProcInfo,
 		Liveness, ResumeVars, LiveSets) :-
@@ -250,12 +254,12 @@ build_live_sets_in_goal_2(higher_order_call(_, ArgVars, Types, Modes, Det),
 		ResumeVars = ResumeVars0
 	).
 
-build_live_sets_in_goal_2(call(PredId, ProcId, ArgVars, Builtin, _, _),
+build_live_sets_in_goal_2(call(PredId, ProcId, ArgVars, BuiltinState, _, _),
 		Liveness, ResumeVars0, LiveSets0,
 		GoalInfo, ModuleInfo, ProcInfo,
 		Liveness, ResumeVars, LiveSets) :-
 	(
-		hlds__is_builtin_is_inline(Builtin)
+		BuiltinState = inline_builtin
 	->
 		ResumeVars = ResumeVars0,
 		LiveSets = LiveSets0
@@ -301,15 +305,15 @@ build_live_sets_in_goal_2(unify(_,_,_,D,_), Liveness, ResumeVars0, LiveSets0,
 		LiveSets = LiveSets0
 	).
 
-build_live_sets_in_goal_2(pragma_c_code(_, IsRec, PredId, ProcId, Args, _),
-		Liveness, ResumeVars0, LiveSets0,
+build_live_sets_in_goal_2(pragma_c_code(_, MayCallMercury, PredId, ProcId,
+		Args, _, _, Extra), Liveness, ResumeVars0, LiveSets0,
 		GoalInfo, ModuleInfo, ProcInfo,
 		Liveness, ResumeVars, LiveSets) :-
 
 	goal_info_get_code_model(GoalInfo, CodeModel),
 	(
 		CodeModel \= model_non,
-		IsRec = non_recursive
+		MayCallMercury = will_not_call_mercury
 	->
 		% We don't need to save any variables onto the stack
 		% before a pragma_c_code if we know that it can't succeed
@@ -321,18 +325,28 @@ build_live_sets_in_goal_2(pragma_c_code(_, IsRec, PredId, ProcId, Args, _),
 	;
 		% The variables which need to be saved onto the stack
 		% before the call are all the variables that are live
-		% after the call, except for the output arguments produced
-		% by the call, plus all the variables that may be needed
-		% at an enclosing resumption point.
+		% after the call (except for the output arguments produced
+		% by the call), plus any variables needed by a nondet
+		% pragma to communication between incarnations, plus
+		% all the variables that may be needed at an enclosing
+		% resumption point.
 
 		find_output_vars(PredId, ProcId, Args, ModuleInfo, OutVars),
 		set__difference(Liveness, OutVars, InputLiveness),
-		set__union(InputLiveness, ResumeVars0, StackVars0),
+		(
+			Extra = none,
+			StackVars0 = InputLiveness
+		;
+			Extra = extra_pragma_info(SavedVarNames, _),
+			assoc_list__keys(SavedVarNames, SavedVars),
+			set__insert_list(InputLiveness, SavedVars, StackVars0)
+		),
+		set__union(StackVars0, ResumeVars0, StackVars1),
 
 		% Might need to add more live variables with accurate GC.
 
 		maybe_add_accurate_gc_typeinfos(ModuleInfo,
-			ProcInfo, OutVars, StackVars0, StackVars),
+			ProcInfo, OutVars, StackVars1, StackVars),
 
 		set__insert(LiveSets0, StackVars, LiveSets),
 
@@ -345,7 +359,7 @@ build_live_sets_in_goal_2(pragma_c_code(_, IsRec, PredId, ProcId, Args, _),
 
 %-----------------------------------------------------------------------------%
 
-:- pred build_live_sets_in_conj(list(hlds__goal), set(var), set(var),
+:- pred build_live_sets_in_conj(list(hlds_goal), set(var), set(var),
 	set(set(var)), module_info, proc_info, set(var), set(var),
 	set(set(var))).
 :- mode build_live_sets_in_conj(in, in, in, in, in, in, out, out, out) is det.
@@ -373,8 +387,8 @@ build_live_sets_in_conj([Goal0 | Goals0], Liveness0, ResumeVars0, LiveSets0,
 
 %-----------------------------------------------------------------------------%
 
-:- pred build_live_sets_in_disj(list(hlds__goal), set(var), set(var),
-	set(set(var)), hlds__goal_info, module_info, proc_info,
+:- pred build_live_sets_in_disj(list(hlds_goal), set(var), set(var),
+	set(set(var)), hlds_goal_info, module_info, proc_info,
 	set(var), set(var), set(set(var))).
 :- mode build_live_sets_in_disj(in, in, in, in, in, in, in, out, out, out)
 	is det.
@@ -455,9 +469,9 @@ maybe_add_accurate_gc_typeinfos(ModuleInfo, ProcInfo, OutVars,
 	(
 		GC_Method = accurate
 	->
-		proc_info_get_used_typeinfos_setwise(ProcInfo, LiveVars1,
+		proc_info_get_typeinfo_vars_setwise(ProcInfo, LiveVars1,
 			TypeInfoVarsLive),
-		proc_info_get_used_typeinfos_setwise(ProcInfo, OutVars,
+		proc_info_get_typeinfo_vars_setwise(ProcInfo, OutVars,
 			TypeInfoVarsOut),
 		set__union(LiveVars1, TypeInfoVarsOut, LiveVars2),
 		set__union(LiveVars2, TypeInfoVarsLive, LiveVars)
@@ -542,7 +556,7 @@ allocate_stack_slots_2([Vars | VarSets], N0, CodeModel,
 
 allocate_same_stack_slot([], _Slot, StackSlots, StackSlots).
 allocate_same_stack_slot([Var | Vars], Slot, StackSlots0, StackSlots) :-
-	map__set(StackSlots0, Var, Slot, StackSlots1),
+	map__det_insert(StackSlots0, Var, Slot, StackSlots1),
 	allocate_same_stack_slot(Vars, Slot, StackSlots1, StackSlots).
 
 %-----------------------------------------------------------------------------%
