@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1995 University of Melbourne.
+% Copyright (C) 1993-1997 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -16,8 +16,8 @@
 
 :- interface.
 
-:- import_module tree, shapes.
-:- import_module bool, list, set, term, std_util.
+:- import_module hlds_pred, tree, prog_data, (inst).
+:- import_module assoc_list, bool, list, set, term, std_util.
 
 %-----------------------------------------------------------------------------%
 
@@ -52,9 +52,8 @@
 			data_name,		% A representation of the name
 						% of the variable; it will be
 						% qualified with the basename.
-			bool,			% Does it have to use Word *
-						% instead of Word?
-			bool,			% Should this item be exported?
+			bool,			% Should this item be exported
+						% from this Mercury module?
 			list(maybe(rval)),	% The arguments of the create.
 			list(pred_proc_id)	% The procedures referenced.
 						% Used by dead_proc_elim.
@@ -75,11 +74,11 @@
 	--->	c_procedure(
 			string,			% predicate name
 			int,			% arity
-			llds__proc_id,		% mode number
+			llds_proc_id,		% mode number
 			list(instruction)	% the code for this procedure
 		).
 
-:- type llds__proc_id	==	int.
+:- type llds_proc_id	==	int.
 
 	% the code for `pragma export' is generated directly as strings
 	% by export.m.
@@ -121,10 +120,11 @@
 	;	call(code_addr, code_addr, list(liveinfo), call_model)
 			% call(Target, Continuation, _, _) is the same as
 			% succip = Continuation; goto(Target).
-			% The third argument is the shape table for the
+			% The third argument is the live value info for the
 			% values live on return. The last gives the model
 			% of the called procedure, and if it is nondet,
-			% says whether tail recursion is applicable to the call.
+			% says whether tail recursion elimination is
+			% potentially applicable to the call.
 
 	;	mkframe(string, int, code_addr)
 			% mkframe(Comment, SlotCount, FailureContinuation)
@@ -141,8 +141,9 @@
 	;	goto(code_addr)
 			% goto(Target)
 			% Branch to the specified address.
-			% Note that jumps to do_fail, etc., can get
-			% optimized into the invocations of macros fail(), etc..
+			% Note that jumps to do_fail, do_redo, etc., can get
+			% optimized into the invocations of macros
+			% fail(), redo(), etc..
 
 	;	computed_goto(rval, list(label))
 			% Evaluate rval, which should be an integer,
@@ -174,19 +175,41 @@
 			% was allocated since that call to mark_hp.
 
 	;	store_ticket(lval)
-			% Get a ticket from the constraint solver,
-			% push it onto the ticket stack,
-			% and store its address in the lval.
+			% Allocate a new "ticket" and store it in the lval.
 
-	;	restore_ticket(rval)
-			% Restore the the constraint solver to the state
-			% it was in when the ticket pointed to by the
-			% specified rval was obtained with store_ticket().
-			% Reset the ticket stack so that the specified
-			% rval is now at the top of the ticket stack.
+	;	reset_ticket(rval, reset_trail_reason)
+			% The rval must specify a ticket allocated with
+			% `store_ticket' and not yet invalidated or
+			% deallocated.
+			% If undo_reason is `undo' or `exception', restore
+			% any mutable global state to the state it was in when
+			% the ticket was obtained with store_ticket();
+			% invalidates any tickets allocated after this one.
+			% If undo_reason is `commit' or `solve', leave the state
+			% unchanged, just check that it is safe to commit
+			% to this solution (i.e. that there are no outstanding
+			% delayed goals -- this is the "floundering" check).
+			% Note that we do not discard trail entries after
+			% commits, because that would in general be unsafe.
+			%
+			% Any invalidated ticket is useless and should
+			% be deallocated with either `discard_ticket'
+			% or `discard_tickets_to'.
 
 	;	discard_ticket
-			% Pop the top ticket off the ticket stack.
+			% Deallocates the most-recently allocated ticket.
+
+	;	mark_ticket_stack(lval)
+			% Tell the trail sub-system to store a ticket counter
+			% (for later use in discard_tickets_upto)
+			% in the specified lval.
+
+	;	discard_tickets_to(rval)
+			% The rval must be a ticket counter obtained via
+			% `mark_ticket_stack' and not yet invalidated.
+			% Deallocates any trail tickets allocated after
+			% the corresponding call to mark_ticket_stack.
+			% Invalidates any later ticket counters.
 
 	;	incr_sp(int, string)
 			% Increment the det stack pointer. The string is
@@ -197,12 +220,20 @@
 	;	decr_sp(int)
 			% Decrement the det stack pointer.
 
-	;	pragma_c(list(pragma_c_decl), list(pragma_c_input), string,
-			list(pragma_c_output)).
-			% The local variable decs, placing the inputs in the
-			% variables, the c code, and where to
-			% find the outputs for pragma(c_code, ... ) decs.
+	;	pragma_c(list(pragma_c_decl), list(pragma_c_input),
+			string, list(pragma_c_output), term__context).
+			% The local variable declarations, the info required
+			% for placing the inputs in the variables, the c code,
+			% the info required for picking up the outputs, and
+			% the context of the original appearance of the C code
+			% in the Mercury source.
 
+%	;	frame_pragma_c(list(pragma_c_decl), list(pragma_c_input),
+%			string, list(pragma_c_output), list(label), term__context).
+%			% The same as above, plus the list of labels to use
+%			% in LABEL_1 and DEFINE_LABEL_1 style macros.
+%			% For use in model_non pragma_c_codes, where it
+%			% should be preceded by a mkframe.
 
 	% pragma_c_decl holds the information needed for a variable
 	% declaration for a pragma_c instruction.
@@ -223,32 +254,56 @@
 				% where to put the output val, type and name
 				% of variable containing the output val
 
-	% Each call instruction has a list of liveinfo,
-	% which stores information about which variables
-	% are live at the point of that call.  The information
-	% is intended for use by the non-conservative garbage collector.
+	% see runtime/mercury_trail.h
+:- type reset_trail_reason
+	--->	undo
+	;	commit
+	;	solve
+	;	exception
+	;	gc
+	.
+
+	% Each call instruction has a list of liveinfo, which stores
+	% information about which variables are live after the call
+	% (that is, on return).  The information is intended for use by
+	% the non-conservative garbage collector.
 :- type liveinfo
 	--->	live_lvalue(
 			lval,
 				% What stackslot/reg does
 				% this lifeinfo structure
 				% refer to?
-			shape_num,
-				% What is the shape of this (bound) variable?
-			maybe(list(lval))
+			live_value_type,
+				% What is the type of this live value?
+			assoc_list(tvar, lval)
 				% Where are the typeinfos that determine the
 				% types of the actual parameters of the type
-				% parameters of this shape (if it is
-				% polymorphic), in the order of the arguments.
+				% parameters of this type (if it is
+				% polymorphic), and the type variable
+				% for each one.
 		).
 
+	% live_value_type describes the different sorts of data that
+	% can be considered live.
+:- type live_value_type 
+	--->	succip		% a stored succip
+	;	curfr		% a stored curfr
+	;	maxfr		% a stored maxfr
+	;	redoip
+	;	hp
+	;	var(type, inst)	% a variable
+	;	unwanted.	% something we don't need, or used as
+				% a placeholder for non-accurate gc.
+	
+
 	% An lval represents a data location or register that can be used
 	% as the target of an assignment.
 :- type lval --->
 
 	/* virtual machine registers */
 
-		reg(reg)	% one of the general-purpose virtual machine
+		reg(reg_type, int)
+				% one of the general-purpose virtual machine
 				% registers (either an int or float reg)
 	;	succip		% virtual machine register holding the
 				% return address for det/semidet code
@@ -260,7 +315,8 @@
 				% pointer
 	;	sp		% virtual machine register point to the
 				% top of det stack
-	;	temp(temp_reg)	% a local temporary register
+	;	temp(reg_type, int)
+				% a local temporary register
 				% These temporary registers are actually
 				% local variables declared in `block'
 				% instructions.  They may only be
@@ -302,6 +358,12 @@
 				% field(Tag, Address, FieldNum)
 				% selects a field of a compound term
 
+	/* values somewhere in memory */
+
+	;	mem_ref(rval)	% a word in the heap, in the det stack or
+				% in the nondet stack. The rval should have
+				% originally come from a mem_addr rval.
+
 	/* pseudo-values */
 
 	;	lvar(var).	% the location of the specified variable
@@ -344,7 +406,17 @@
 		% mkword returns a tagged pointer
 	;	const(rval_const)
 	;	unop(unary_op, rval)
-	;	binop(binary_op, rval, rval).
+	;	binop(binary_op, rval, rval)
+	;	mem_addr(mem_ref).
+		% The addess of a word in the heap, the det stack or
+		% the nondet stack.
+
+:- type mem_ref
+	--->	stackvar_ref(int)		% stack slot number
+	;	framevar_ref(int)		% stack slot number
+	;	heap_ref(rval, int, int).	% the cell pointer,
+						% the tag to subtract,
+						% and the field number
 
 :- type rval_const
 	--->	true
@@ -356,16 +428,21 @@
 	;	data_addr_const(data_addr).
 
 :- type data_addr
-	--->	data_addr(string, data_name, bool).
-			% module name; which var; does it have any
-			% addresses inside it (i.e. Word or Word *)?
+	--->	data_addr(string, data_name).
+			% module name; which var
 
 :- type data_name
 	--->	common(int)
-	;	base_type_info(string, arity)
-			% type name, type arity
-	;	base_type_layout(string, arity).
-			% type name, type arity
+	;	base_type(base_data, string, arity).
+			% base_data, type name, type arity
+
+:- type base_data
+	--->	info
+			% basic information, including special preds
+	;	layout
+			% layout information
+	;	functors.
+			% information on functors
 
 :- type unary_op
 	--->	mktag
@@ -383,8 +460,10 @@
 	--->	(+)	% integer arithmetic
 	;	(-)
 	;	(*)
-	;	(/)
-	;	(mod)
+	;	(/)	% integer division
+			% assumed to truncate toward zero
+	;	(mod)	% remainder (w.r.t. truncating integer division)
+			% XXX `mod' should be renamed `rem'
 	;	(<<)	% left shift
 	;	(>>)	% right shift
 	;	(&)	% bitwise and
@@ -416,12 +495,9 @@
 	;	float_le
 	;	float_ge.
 
-	% one of the general-purpose virtual machine registers
-:- type reg	
-	--->	r(int)		% integer regs
-	;	f(int).		% floating point regs
-
-:- type temp_reg == reg.
+:- type reg_type	
+	--->	r		% general-purpose (integer) regs
+	;	f.		% floating point regs
 
 :- type label
 	--->	local(proc_label, int)	% internal to procedure
@@ -441,14 +517,22 @@
 	;	do_fail
 	;	do_det_closure
 	;	do_semidet_closure
-	;	do_nondet_closure.
+	;	do_nondet_closure
+	;	do_not_reached.		% we should never jump to this address
 
-	% A proc_label is a label used for the entry point to a procedure
+	% A proc_label is a label used for the entry point to a procedure.
+	% The defining module is the module that provides the code for the
+	% predicate, the declaring module contains the `:- pred' declaration.
+	% When these are different, as for specialised versions of predicates
+	% from `.opt' files, the defining module's name is added as a
+	% qualifier to the label.
 :- type proc_label
-	--->	proc(string, pred_or_func, string, int, int)
-		%	 module, predicate/function, name, arity, mode #
-	;	special_proc(string, string, sym_name, int, int).
-		%	module, pred name, type name, type arity, mode #
+	--->	proc(string, pred_or_func, string, string, int, proc_id)
+			% defining module, predicate/function,
+			% declaring module, name, arity, mode #
+	;	special_proc(string, string, string, string, int, proc_id).
+			% defining module, pred name, type module,
+			% type name, type arity, mode #
 
 	% A tag (used in mkword, create and field expressions
 	% and in incr_hp instructions) is a small integer.
@@ -459,12 +543,26 @@
 	% choosing the right sort of register for a given value
 	% to avoid unnecessary boxing/unboxing of floats.
 :- type llds_type
-	--->	bool
-	;	integer
-	;	unsigned
-	;	float
-	;	word.	% anything whose size is a word
-			% may really be integer, unsigned, or pointer
+	--->	bool		% a boolean value
+				% represented using the C type `Integer'
+	;	integer		% a Mercury `int', represented in C as a
+				% value of type `Integer' (which is
+				% a signed integral type of the same
+				% size as a pointer)
+	;	unsigned	% something whose C type is `Unsigned'
+				% (the unsigned equivalent of `Integer')
+	;	float		% a Mercury `float', represented in C as a
+				% value of type `Float' (which may be either
+				% `float' or `double', but is usually `double')
+	;	data_ptr	% a pointer to data; represented in C
+				% as a value of C type `Word *'
+	;	code_ptr	% a pointer to code; represented in C
+				% as a value of C type `Code *'
+	;	word.		% something that can be assigned to a value
+				% of C type `Word', i.e., something whose
+				% size is a word but which may be either
+				% signed or unsigned
+				% (used for registers, stack slots, etc.)
 
 	% given a non-var rval, figure out its type
 :- pred llds__rval_type(rval::in, llds_type::out) is det.
@@ -478,55 +576,71 @@
 	% given a unary operator, figure out its return type
 :- pred llds__unop_return_type(unary_op::in, llds_type::out) is det.
 
+	% given a unary operator, figure out the type of its argument
+:- pred llds__unop_arg_type(unary_op::in, llds_type::out) is det.
+
 	% given a binary operator, figure out its return type
 :- pred llds__binop_return_type(binary_op::in, llds_type::out) is det.
 
 	% given a register, figure out its type
-:- pred llds__register_type(reg::in, llds_type::out) is det.
+:- pred llds__register_type(reg_type::in, llds_type::out) is det.
 
 :- implementation.
 :- import_module require.
 
-llds__lval_type(reg(Reg), Type) :-
-	llds__register_type(Reg, Type).
-llds__lval_type(succip, word).		% really `Code*'
-llds__lval_type(maxfr, word).		% really `Word*'
-llds__lval_type(curfr, word).		% really `Word*'
-llds__lval_type(hp, word).		% really `Word*'
-llds__lval_type(sp, word).		% really `Word*'
-llds__lval_type(temp(TempReg), Type) :-
-	llds__register_type(TempReg, Type).
+llds__lval_type(reg(RegType, _), Type) :-
+	llds__register_type(RegType, Type).
+llds__lval_type(succip, code_ptr).
+llds__lval_type(maxfr, data_ptr).
+llds__lval_type(curfr, data_ptr).
+llds__lval_type(hp, data_ptr).
+llds__lval_type(sp, data_ptr).
+llds__lval_type(temp(RegType, _), Type) :-
+	llds__register_type(RegType, Type).
 llds__lval_type(stackvar(_), word).
 llds__lval_type(framevar(_), word).
-llds__lval_type(succip(_), word).	% really `Code*'
-llds__lval_type(redoip(_), word).	% really `Code*'
-llds__lval_type(succfr(_), word).	% really `Word*'
-llds__lval_type(prevfr(_), word).	% really `Word*'
+llds__lval_type(succip(_), code_ptr).
+llds__lval_type(redoip(_), code_ptr).
+llds__lval_type(succfr(_), data_ptr).
+llds__lval_type(prevfr(_), data_ptr).
 llds__lval_type(field(_, _, _), word).
 llds__lval_type(lvar(_), _) :-
 	error("lvar unexpected in llds__lval_type").
+llds__lval_type(mem_ref(_), word).
 
 llds__rval_type(lval(Lval), Type) :-
 	llds__lval_type(Lval, Type).
 llds__rval_type(var(_), _) :-
 	error("var unexpected in llds__rval_type").
-llds__rval_type(create(_, _, _, _), word). % the create() macro calls mkword().
-llds__rval_type(mkword(_, _), word).
+llds__rval_type(create(_, _, _, _), data_ptr).
+	%
+	% Note that create and mkword must both be of type data_ptr,
+	% not of type word, to ensure that static consts containing
+	% them get type `const Word *', not type `Word'; this is
+	% necessary because casts from pointer to int must not be used
+	% in the initializers for constant expressions -- if they are,
+	% then lcc barfs, and gcc generates bogus code on some systems,
+	% (e.g. IRIX with shared libs).  If the second argument to mkword
+	% is an integer, not a pointer, then we will end up casting it
+	% to a pointer, but casts from integer to pointer are OK, it's
+	% only the reverse direction we need to avoid.
+	%
+llds__rval_type(mkword(_, _), data_ptr).
 llds__rval_type(const(Const), Type) :-
 	llds__const_type(Const, Type).
 llds__rval_type(unop(UnOp, _), Type) :-
 	llds__unop_return_type(UnOp, Type).
 llds__rval_type(binop(BinOp, _, _), Type) :-
 	llds__binop_return_type(BinOp, Type).
+llds__rval_type(mem_addr(_), data_ptr).
 
 llds__const_type(true, bool).
 llds__const_type(false, bool).
 llds__const_type(int_const(_), integer).
 llds__const_type(float_const(_), float).
-llds__const_type(string_const(_), word).
-	% the string_const() macro returns a value cast to type Word
-llds__const_type(code_addr_const(_), word).
-llds__const_type(data_addr_const(_), word).
+llds__const_type(string_const(_), data_ptr).
+llds__const_type(code_addr_const(_), code_ptr).
+llds__const_type(data_addr_const(_), data_ptr).
 
 llds__unop_return_type(mktag, word).
 llds__unop_return_type(tag, word).
@@ -539,6 +653,17 @@ llds__unop_return_type(hash_string, integer).
 llds__unop_return_type(bitwise_complement, integer).
 llds__unop_return_type(not, bool).
 
+llds__unop_arg_type(mktag, word).
+llds__unop_arg_type(tag, word).
+llds__unop_arg_type(unmktag, word).
+llds__unop_arg_type(mkbody, word).
+llds__unop_arg_type(unmkbody, word).
+llds__unop_arg_type(body, word).
+llds__unop_arg_type(cast_to_unsigned, word).
+llds__unop_arg_type(hash_string, word).
+llds__unop_arg_type(bitwise_complement, integer).
+llds__unop_arg_type(not, bool).
+
 llds__binop_return_type((+), integer).
 llds__binop_return_type((-), integer).
 llds__binop_return_type((*), integer).
@@ -575,6 +700,5 @@ llds__binop_return_type(float_gt, bool).
 llds__binop_return_type(float_le, bool).
 llds__binop_return_type(float_ge, bool).
 
-llds__register_type(r(_), word).
-llds__register_type(f(_), float).
-
+llds__register_type(r, word).
+llds__register_type(f, float).
