@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1995 University of Melbourne.
+% Copyright (C) 1996-1997 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -42,13 +42,13 @@
 :- interface.
 
 :- import_module hlds_goal, hlds_data, llds, switch_gen, code_info.
-:- import_module list.
+:- import_module list, term.
 
 :- type case_consts == list(pair(int, list(rval))).
 
 :- type rval_map == map(var, list(pair(int, rval))).
 
-:- pred lookup_switch__is_lookup_switch(var, cases_list, hlds__goal_info,
+:- pred lookup_switch__is_lookup_switch(var, cases_list, hlds_goal_info,
 		can_fail, int, code_model, int, int, can_fail, can_fail,
 		list(var), case_consts, maybe(set(var)),
 		code_info, code_info).
@@ -111,16 +111,17 @@ lookup_switch__is_lookup_switch(CaseVar, TaggedCases, GoalInfo,
 	(
 		{ NumCases = Range }
 	->
-		{ NeedBitVecTest = cannot_fail }
+		{ NeedBitVecTest0 = cannot_fail }
 	;
-		{ NeedBitVecTest = can_fail }
+		{ NeedBitVecTest0 = can_fail }
 	),
 	( { CanFail0 = can_fail } ->
 		% For semidet switches, we normally need to check that
 		% the variable is in range before we index into the jump table.
 		% However, if the range of the type is sufficiently small,
 		% we can make the jump table large enough to hold all
-		% of the values for the type.
+		% of the values for the type, but then we will need to do the
+		% bitvector test.
 		code_info__variable_type(CaseVar, Type),
 		code_info__get_module_info(ModuleInfo),
 		{ classify_type(Type, ModuleInfo, TypeCategory) },
@@ -130,15 +131,18 @@ lookup_switch__is_lookup_switch(CaseVar, TaggedCases, GoalInfo,
 			{ DetDensity > ReqDensity }
 		->
 			{ NeedRangeCheck = cannot_fail },
+			{ NeedBitVecTest = can_fail },
 			{ FirstVal = 0 },
 			{ LastVal is TypeRange - 1 }
 		;
 			{ NeedRangeCheck = CanFail0 },
+			{ NeedBitVecTest = NeedBitVecTest0 },
 			{ FirstVal = FirstCaseVal },
 			{ LastVal = LastCaseVal }
 		)
 	;
 		{ NeedRangeCheck = CanFail0 },
+		{ NeedBitVecTest = NeedBitVecTest0 },
 		{ FirstVal = FirstCaseVal },
 		{ LastVal = LastCaseVal }
 	),
@@ -148,7 +152,7 @@ lookup_switch__is_lookup_switch(CaseVar, TaggedCases, GoalInfo,
 
 %---------------------------------------------------------------------------%
 
-:- pred lookup_switch__figure_out_output_vars(hlds__goal_info, list(var),
+:- pred lookup_switch__figure_out_output_vars(hlds_goal_info, list(var),
 		code_info, code_info).
 :- mode lookup_switch__figure_out_output_vars(in, out, in, out) is det.
 
@@ -197,8 +201,8 @@ lookup_switch__generate_constants([Case|Cases], Vars, CodeModel,
 	{ Case = case(_, int_constant(CaseTag), _, Goal) },
 	code_info__grab_code_info(CodeInfo),
 	code_gen__generate_goal(CodeModel, Goal, Code),
-	code_info__get_liveness_info(Liveness),
-	{ lookup_switch__code_is_empty(Code) },
+	code_info__get_forward_live_vars(Liveness),
+	{ tree__is_empty(Code) },
 	lookup_switch__get_case_rvals(Vars, CaseRvals),
 	{ CaseVal = CaseTag - CaseRvals },
 	code_info__slap_code_info(CodeInfo),
@@ -206,16 +210,6 @@ lookup_switch__generate_constants([Case|Cases], Vars, CodeModel,
 
 %---------------------------------------------------------------------------%
 
-:- pred lookup_switch__code_is_empty(code_tree).
-:- mode lookup_switch__code_is_empty(in) is semidet.
-
-lookup_switch__code_is_empty(empty).
-lookup_switch__code_is_empty(tree(L, R)) :-
-	lookup_switch__code_is_empty(L),
-	lookup_switch__code_is_empty(R).
-
-%---------------------------------------------------------------------------%
-
 :- pred lookup_switch__get_case_rvals(list(var), list(rval),
 		code_info, code_info).
 :- mode lookup_switch__get_case_rvals(in, out, in, out) is semidet.
@@ -223,7 +217,7 @@ lookup_switch__code_is_empty(tree(L, R)) :-
 lookup_switch__get_case_rvals([], []) --> [].
 lookup_switch__get_case_rvals([Var|Vars], [Rval|Rvals]) -->
 	code_info__produce_variable(Var, Code, Rval),
-	{ lookup_switch__code_is_empty(Code) },
+	{ tree__is_empty(Code) },
 	code_info__get_globals(Globals),
 	{ globals__get_options(Globals, Options) },
 	{ exprn_aux__init_exprn_opts(Options, ExprnOpts) },
@@ -284,7 +278,7 @@ lookup_switch__generate(Var, OutVars, CaseValues,
 	(
 		{ NeedRangeCheck = can_fail },
 		{ Difference is EndVal - StartVal },
-		code_info__generate_test_and_fail(
+		code_info__fail_if_rval_is_false(
 			binop(<=, unop(cast_to_unsigned, Index),
 				const(int_const(Difference))), RangeCheck)
 	;
@@ -307,7 +301,7 @@ lookup_switch__generate(Var, OutVars, CaseValues,
 		% wrong liveness.
 	(
 		{ MLiveness = yes(Liveness) },
-		code_info__set_liveness_info(Liveness)
+		code_info__set_forward_live_vars(Liveness)
 	;
 		{ MLiveness = no },
 		{ error("lookup_switch__generate: no liveness!") }
@@ -336,31 +330,59 @@ lookup_switch__generate(Var, OutVars, CaseValues,
 	% iff we have a case for that tag value.
 lookup_switch__generate_bitvec_test(Index, CaseVals, Start, _End,
 		CheckCode) -->
-	generate_bit_vec(CaseVals, Start, BitVec),
+	lookup_switch__get_word_bits(WordBits),
+	generate_bit_vec(CaseVals, Start, WordBits, BitVec),
+
+	{ UIndex = unop(cast_to_unsigned, Index) },
+		%
+		% Optimize the single-word case:
+		% if all the cases fit into a single word, then
+		% the word to use is always that word, and the index
+		% specifies which bit.  Otherwise, the high bits
+		% of the index specify which word to use and the
+		% low bits specify which bit.
+		%
+	{
+		BitVec = create(_, [yes(SingleWord)], _, _)
+	->
+		Word = SingleWord,
+		BitNum = UIndex
+	;
+		WordNum = binop(/, UIndex, const(int_const(WordBits))),
+		Word = lval(field(0, BitVec, WordNum)),
+		BitNum = binop(mod, UIndex, const(int_const(WordBits)))
+	},
+	{ HasBit = binop((&),
+			binop((<<), const(int_const(1)), BitNum),
+			Word) },
+	code_info__fail_if_rval_is_false(HasBit, CheckCode).
+
+
+:- pred lookup_switch__get_word_bits(int, code_info, code_info).
+:- mode lookup_switch__get_word_bits(out, in, out) is det.
+
+	% Prevent cross-compilation errors by making sure that
+	% the bitvector uses a number of bits that will fit both
+	% on this machine (so that we can correctly generate it),
+	% and on the target machine (so that it can be executed
+	% correctly)
+
+lookup_switch__get_word_bits(WordBits) -->
+	{ int__bits_per_int(HostWordBits) },
 	code_info__get_globals(Globals),
 	{ globals__get_options(Globals, Options) },
-	{ getopt__lookup_int_option(Options, bits_per_word, WordBits) },
-	{ UIndex = unop(cast_to_unsigned, Index) },
-	{ Word = lval(field(0, BitVec,
-			binop(/, UIndex,const(int_const(WordBits)))))},
-	{ HasBit = binop((&), binop((<<),
-			const(int_const(1)),
-			binop(mod, UIndex, const(int_const(WordBits)))),Word) },
-	code_info__generate_test_and_fail(HasBit, CheckCode).
+	{ getopt__lookup_int_option(Options, bits_per_word, TargetWordBits) },
+	{ int__min(HostWordBits, TargetWordBits, WordBits) }.
 
-:- pred generate_bit_vec(case_consts, int, rval,
-		code_info, code_info).
-:- mode generate_bit_vec(in, in, out, in, out) is det.
+:- pred generate_bit_vec(case_consts, int, int, rval, code_info, code_info).
+:- mode generate_bit_vec(in, in, in, out, in, out) is det.
 
 	% we generate the bitvector by iterating through the cases
 	% marking the bit for each case. (We represent the bitvector
 	% here as a map from the word number in the vector to the bits
 	% for that word.
-generate_bit_vec(CaseVals, Start, BitVec) -->
+generate_bit_vec(CaseVals, Start, WordBits, BitVec) -->
 	{ map__init(Empty) },
-	code_info__get_globals(Globals),
-	{ globals__get_options(Globals, Options) },
-	{ getopt__lookup_int_option(Options, bits_per_word, WordBits) },
 	{ generate_bit_vec_2(CaseVals, Start, WordBits, Empty, BitMap) },
 	{ map__to_assoc_list(BitMap, WordVals) },
 	{ generate_bit_vec_args(WordVals, 0, Args) },
@@ -368,7 +390,7 @@ generate_bit_vec(CaseVals, Start, BitVec) -->
 	{ BitVec = create(0, Args, no, CellNo) }.
 
 :- pred generate_bit_vec_2(case_consts, int, int,
-		map(int, int), map(int, int)).
+			map(int, int), map(int, int)).
 :- mode generate_bit_vec_2(in, in, in, in, out) is det.
 
 generate_bit_vec_2([], _, _, Bits, Bits).
