@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1995 University of Melbourne.
+% Copyright (C) 1995-1997 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -38,7 +38,7 @@
 :- interface.
 
 :- import_module hlds_module.
-:- import_module int, io.
+:- import_module io.
 
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
@@ -58,7 +58,7 @@
 :- import_module prog_util, mode_util, hlds_out, stack, quantification.
 :- import_module globals, options.
 :- import_module string, map, list, require, std_util, term, term_io, getopt.
-:- import_module bool, set, varset.
+:- import_module bool, set, varset, int.
 
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
@@ -177,12 +177,12 @@ c_gen_pred(Indent, ModuleInfo, PredId, PredInfo) -->
 		{ ClausesInfo = clauses_info(VarSet, _VarTypes, _, HeadVars,
 			Clauses) },
 
-		globals__io_lookup_bool_option(verbose_dump_hlds, Verbose),
-		globals__io_set_option(verbose_dump_hlds, bool(no)),
+		globals__io_lookup_string_option(verbose_dump_hlds, Verbose),
+		globals__io_set_option(verbose_dump_hlds, string("")),
 		{ pred_info_get_is_pred_or_func(PredInfo, PredOrFunc) },
-		hlds_out__write_clauses(Indent, ModuleInfo, PredId, VarSet,
+		hlds_out__write_clauses(Indent, ModuleInfo, PredId, VarSet, no,
 			HeadVars, PredOrFunc, Clauses, no),
-		globals__io_set_option(verbose_dump_hlds, bool(Verbose)),
+		globals__io_set_option(verbose_dump_hlds, string(Verbose)),
 
 		io__write_string("****/\n"),
 
@@ -289,7 +289,7 @@ c_gen_proc(Indent, ModuleInfo, PredId, ProcId, Pred, Proc) -->
 
 %-----------------------------------------------------------------------------%
 
-:- pred c_gen_predeclare_labels(hlds__goal, c_gen_info, io__state, io__state).
+:- pred c_gen_predeclare_labels(hlds_goal, c_gen_info, io__state, io__state).
 :- mode c_gen_predeclare_labels(in, in, di, uo) is det.
 
 	% XXX this should traverse the goal and count how many function labels
@@ -396,7 +396,8 @@ c_gen_proc_name(ModuleInfo, PredId, ProcId) -->
 	io__write_string("_"),
 	io__write_int(Arity),
 	io__write_string("_"),
-        { ModeNum is ProcId mod 10000 },      % strip off the priority
+	{ proc_id_to_int(ProcId, ProcInt) },
+	{ ModeNum is ProcInt mod 10000 },	% strip off the priority
 	io__write_int(ModeNum).
 
 :- pred c_gen_select_output_vars(module_info, list(var), list(mode), list(var)).
@@ -459,15 +460,15 @@ c_gen_arg_decl(ModuleInfo, Var, Type, Mode, VarSet) -->
 	;
 		io__write_string(" ")
 	),
-	mercury_output_var(Var, VarSet).
+	mercury_output_var(Var, VarSet, no).
 
-:- pred c_gen_local_var_decls(int, hlds__goal, varset, map(var, type),
+:- pred c_gen_local_var_decls(int, hlds_goal, varset, map(var, type),
 				list(var),
 				io__state, io__state).
 :- mode c_gen_local_var_decls(in, in, in, in, in, di, uo) is det.
 
 c_gen_local_var_decls(Indent, Goal, VarSet, VarTypes, HeadVars) -->
-	{ goal_vars(Goal, Vars0) },
+	{ quantification__goal_vars(Goal, Vars0) },
 	{ set__to_sorted_list(Vars0, Vars) },
 	{ list__delete_elems(Vars, HeadVars, LocalVars) },
 	c_gen_local_var_decls_2(LocalVars, VarSet, VarTypes, Indent).
@@ -482,13 +483,13 @@ c_gen_local_var_decls_2([Var|Vars], VarSet, VarTypes, Indent) -->
 	{ map__lookup(VarTypes, Var, Type) },
 	c_gen_type(Type),
 	io__write_string(" "),
-	mercury_output_var(Var, VarSet),
+	mercury_output_var(Var, VarSet, no),
 	io__write_string(";\n"),
 	c_gen_local_var_decls_2(Vars, VarSet, VarTypes, Indent).
 
 %-----------------------------------------------------------------------------%
 
-:- pred c_gen_goal(hlds__goal, int, c_gen_info, c_gen_info,
+:- pred c_gen_goal(hlds_goal, int, c_gen_info, c_gen_info,
 				io__state, io__state).
 :- mode c_gen_goal(in, in, in, out, di, uo) is det.
 
@@ -512,7 +513,7 @@ c_gen_goal(Goal - GoalInfo, Indent, CGenInfo0, CGenInfo) -->
 	),
 	c_gen_goal_2(Goal, Indent, CGenInfo0, CGenInfo).
 
-:- pred c_gen_goal_2(hlds__goal_expr, int, c_gen_info, c_gen_info,
+:- pred c_gen_goal_2(hlds_goal_expr, int, c_gen_info, c_gen_info,
 					io__state, io__state).
 :- mode c_gen_goal_2(in, in, in, out, di, uo) is det.
 
@@ -522,7 +523,7 @@ c_gen_goal_2(switch(Var, _CanFail, CasesList, _), Indent,
 	io__write_string("/* "),
 	% c_gen_can_fail(CanFail), 
 	io__write_string(" switch on `"),
-	mercury_output_var(Var, _VarSet),
+	mercury_output_var(Var, _VarSet, no),
 	io__write_string("' */"),
 	{ Indent1 is Indent + 1 },
 	mercury_output_newline(Indent1),
@@ -539,7 +540,7 @@ c_gen_goal_2(switch(Var, _CanFail, CasesList, _), Indent,
 c_gen_goal_2(some(Vars, Goal), Indent, CGenInfo0, CGenInfo) -->
 	{ sorry(8) },
 	io__write_string("some ["),
-	mercury_output_vars(Vars, _VarSet),
+	mercury_output_vars(Vars, _VarSet, no),
 	io__write_string("] ("),
 	{ Indent1 is Indent + 1 },
 	mercury_output_newline(Indent1),
@@ -624,7 +625,7 @@ c_gen_goal_2(disj(List, _), Indent, CGenInfo0, CGenInfo) -->
 		c_gen_failure(Indent, CGenInfo0, CGenInfo)
 	).
 
-c_gen_goal_2(higher_order_call(_, _, _, _, _), _, _, _) -->
+c_gen_goal_2(higher_order_call(_, _, _, _, _, _), _, _, _) -->
 	{ error("mercury_to_c: higher_order_call not implemented") }.
 c_gen_goal_2(call(PredId, ProcId, ArgVars, _, _, _PredName),
 					Indent, CGenInfo0, CGenInfo) -->
@@ -669,7 +670,7 @@ c_gen_goal_2(call(PredId, ProcId, ArgVars, _, _, _PredName),
 c_gen_goal_2(unify(_A, _B, _, Unification, _), Indent, CGenInfo0, CGenInfo) -->
 	c_gen_unification(Unification, Indent, CGenInfo0, CGenInfo).
 
-c_gen_goal_2(pragma_c_code(C_Code, _, _, _, _, ArgNames), _, _, _) -->
+c_gen_goal_2(pragma_c_code(C_Code, _, _, _, _, ArgNames, _, _), _, _, _) -->
 	{ sorry(4) },
 	{ get_pragma_c_var_names(ArgNames, Names) },
 	io__write_string("$pragma(c_code, ["),
@@ -742,7 +743,7 @@ c_gen_arg_list(Vars, Modes, CGenInfo0, CGenInfo) -->
 		;
 			[]
 		),
-		mercury_output_var(Var, VarSet),	% XXX name mangling
+		mercury_output_var(Var, VarSet, no),	% XXX name mangling
 		( { Vars1 = [] } ->
 			[]
 		;
@@ -764,7 +765,7 @@ c_gen_var(Var, CGenInfo, CGenInfo) -->
 	;
 		[]
 	),
-	mercury_output_var(Var, VarSet).	% XXX name mangling
+	mercury_output_var(Var, VarSet, no).	% XXX name mangling
 
 :- pred sorry(int::in) is erroneous.
 
@@ -778,7 +779,7 @@ sorry(N) :-
 
 c_gen_var_modes([], [], _) --> [].
 c_gen_var_modes([Var|Vars], [Mode|Modes], VarSet) -->
-	mercury_output_var(Var, VarSet),
+	mercury_output_var(Var, VarSet, no),
 	io__write_string("::"),
 	mercury_output_mode(Mode, VarSet),
 	( { Vars \= [] } ->
@@ -792,7 +793,7 @@ c_gen_var_modes([], [_|_], _) -->
 c_gen_var_modes([_|_], [], _) -->
 	{ error("c_gen_var_modes: length mis-match") }.
 
-:- pred c_gen_conj(list(hlds__goal), int, c_gen_info, c_gen_info,
+:- pred c_gen_conj(list(hlds_goal), int, c_gen_info, c_gen_info,
 				io__state, io__state).
 :- mode c_gen_conj(in, in, in, out, di, uo) is det.
 
@@ -801,7 +802,7 @@ c_gen_conj([Goal | Goals], Indent, CGenInfo0, CGenInfo) -->
 	c_gen_goal(Goal, Indent, CGenInfo0, CGenInfo1),
 	c_gen_conj(Goals, Indent, CGenInfo1, CGenInfo).
 
-:- pred c_gen_disj(list(hlds__goal), c_label_func, int, c_gen_info, c_gen_info,
+:- pred c_gen_disj(list(hlds_goal), c_label_func, int, c_gen_info, c_gen_info,
 				io__state, io__state).
 :- mode c_gen_disj(in, in, in, in, out, di, uo) is det.
 
@@ -819,7 +820,7 @@ c_gen_disj([Goal|Goals], Label, Indent, CGenInfo0, CGenInfo) -->
 
 c_gen_case(case(_ConsId, Goal), Var, Indent, CGenInfo0, CGenInfo) -->
 	{ sorry(10) },
-	mercury_output_var(Var, _VarSet),
+	mercury_output_var(Var, _VarSet, no),
 	io__write_string(" has functor "),
 	% c_gen_cons_id(ConsId),
 	io__write_string(","),
@@ -864,9 +865,9 @@ c_gen_var_types_2([Var | Vars], Indent, VarSet, VarTypes, TypeVarSet)
 	c_gen_indent(Indent),
 	c_gen_type(Type),
 	io__write_string(" "),
-	mercury_output_var(Var, VarSet),
+	mercury_output_var(Var, VarSet, no),
 	io__write_string(";\t/* "),
-	mercury_output_term(Type, TypeVarSet),
+	mercury_output_term(Type, TypeVarSet, no),
 	io__write_string(" */\n"),
 	c_gen_var_types_2(Vars, Indent, VarSet, VarTypes, TypeVarSet).
 
