@@ -1,5 +1,5 @@
 %----------------------------------------------------------------------------%
-% Copyright (C) 1995 University of Melbourne.
+% Copyright (C) 1994-1997 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -30,7 +30,7 @@
 :- module mercury_to_goedel.
 :- interface.
 
-:- import_module string, list, io.
+:- import_module list, io.
 :- import_module prog_data.
 
 :- pred convert_to_goedel(string, list(item_and_context), io__state, io__state).
@@ -39,7 +39,7 @@
 %-----------------------------------------------------------------------------%
 
 :- implementation.
-:- import_module bool, int, char, std_util, varset, term, require.
+:- import_module bool, int, char, std_util, varset, term, require, string.
 :- import_module prog_io, prog_out, prog_util, equiv_type.
 :- import_module globals, options.
 %-----------------------------------------------------------------------------%
@@ -234,8 +234,17 @@ goedel_output_type_defn_2(eqv_type(_Name, _Args, _Body), _VarSet, Context) -->
 	io__write_string("equivalence type unexpected.\n"),
 	io__set_output_stream(OldStream, _).
 
-goedel_output_type_defn_2(du_type(Name, Args, Ctors), VarSet, Context) -->
+goedel_output_type_defn_2(du_type(Name, Args, Ctors, EqualityPred),
+		VarSet, Context) -->
 	{ unqualify_name(Name, Name2) },
+	( { EqualityPred \= no } ->
+		io__write_string("% WARNING: ignored user-defined equality "),
+		io__write_string(" for type `"),
+		io__write_string(Name2),
+		io__write_string("'\n")
+	;
+		[]
+	),
 	{ convert_functor_name(Name2, Name3) },
 	( { option_handle_functor_overloading(Name2) } ->
 		{ string__append("Type__", Name3, TypeModule) },
@@ -854,7 +863,7 @@ goedel_quote_string(S0) -->
 		[]
 	).
 
-:- pred goedel_quote_char(character, character).
+:- pred goedel_quote_char(char, char).
 :- mode goedel_quote_char(in, out) is semidet.
 
 goedel_quote_char('"', '"').
