@@ -1,5 +1,5 @@
 %---------------------------------------------------------------------------%
-% Copyright (C) 1995 University of Melbourne.
+% Copyright (C) 1994-1997 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %---------------------------------------------------------------------------%
@@ -16,7 +16,7 @@
 
 :- import_module hlds_goal, llds, code_info.
 
-:- pred middle_rec__match_and_generate(hlds__goal, code_tree,
+:- pred middle_rec__match_and_generate(hlds_goal, code_tree,
 	code_info, code_info).
 :- mode middle_rec__match_and_generate(in, out, in, out) is semidet.
 
@@ -28,6 +28,7 @@
 :- import_module hlds_module, hlds_data.
 :- import_module code_gen, unify_gen, code_util, code_aux, opt_util.
 :- import_module bool, set, int, std_util, tree, list, assoc_list, require.
+:- import_module string, term.
 
 %---------------------------------------------------------------------------%
 
@@ -85,8 +86,8 @@ middle_rec__match_and_generate(Goal, Instrs, CodeInfo0, CodeInfo) :-
 
 %---------------------------------------------------------------------------%
 
-:- pred middle_rec__generate_ite(list(var), hlds__goal, hlds__goal,
-	hlds__goal, ite_rec, store_map, hlds__goal_info,
+:- pred middle_rec__generate_ite(list(var), hlds_goal, hlds_goal,
+	hlds_goal, ite_rec, store_map, hlds_goal_info,
 	code_tree, code_info, code_info).
 :- mode middle_rec__generate_ite(in, in, in, in, in, in, in, out, in, out)
 	is det.
@@ -101,8 +102,8 @@ middle_rec__generate_ite(_Vars, _Cond, _Then, _Else, _Rec, _IteGoalInfo, _SM,
 
 %---------------------------------------------------------------------------%
 
-:- pred middle_rec__generate_switch(var, cons_id, hlds__goal, hlds__goal,
-	store_map, hlds__goal_info, code_tree, code_info, code_info).
+:- pred middle_rec__generate_switch(var, cons_id, hlds_goal, hlds_goal,
+	store_map, hlds_goal_info, code_tree, code_info, code_info).
 :- mode middle_rec__generate_switch(in, in, in, in, in, in, out, in, out)
 	is semidet.
 
@@ -114,11 +115,10 @@ middle_rec__generate_switch(Var, BaseConsId, Base, Recursive, StoreMap,
 	code_info__get_module_info(ModuleInfo),
 	code_info__get_pred_id(PredId),
 	code_info__get_proc_id(ProcId),
-	{ predicate_name(ModuleInfo, PredId, PredName) },
 	{ code_util__make_local_entry_label(ModuleInfo, PredId, ProcId, no,
 		EntryLabel) },
 
-	code_aux__pre_goal_update(SwitchGoalInfo, no),
+	code_info__pre_goal_update(SwitchGoalInfo, no),
 	unify_gen__generate_tag_test(Var, BaseConsId, branch_on_success,
 		BaseLabel, EntryTestCode),
 	{ tree__flatten(EntryTestCode, EntryTestListList) },
@@ -131,7 +131,7 @@ middle_rec__generate_switch(Var, BaseConsId, Base, Recursive, StoreMap,
 	code_gen__generate_goal(model_det, Recursive, RecGoalCode),
 	code_info__generate_branch_end(model_det, StoreMap, RecSaveCode),
 
-	code_aux__post_goal_update(SwitchGoalInfo),
+	code_info__post_goal_update(SwitchGoalInfo),
 	code_info__remake_with_store_map(StoreMap),
 
 	code_info__get_arginfo(ArgModes),
@@ -182,7 +182,10 @@ middle_rec__generate_switch(Var, BaseConsId, Base, Recursive, StoreMap,
 				label(Loop2Label))
 				- "test on upward loop"]
 	;
-		MaybeIncrSp = [incr_sp(FrameSize, PredName) - ""],
+		predicate_module(ModuleInfo, PredId, ModuleName),
+		predicate_name(ModuleInfo, PredId, PredName),
+		string__append_list([ModuleName, ":", PredName], PushMsg),
+		MaybeIncrSp = [incr_sp(FrameSize, PushMsg) - ""],
 		MaybeDecrSp = [decr_sp(FrameSize) - ""],
 		InitAuxReg =  [assign(AuxReg, lval(sp))
 				- "initialize counter register"],
@@ -354,10 +357,10 @@ middle_rec__find_unused_register(Instrs, UnusedReg) :-
 :- pred middle_rec__find_unused_register_2(list(int), int, lval).
 :- mode middle_rec__find_unused_register_2(in, in, out) is det.
 
-middle_rec__find_unused_register_2([], N, reg(r(N))).
+middle_rec__find_unused_register_2([], N, reg(r, N)).
 middle_rec__find_unused_register_2([H | T], N, Reg) :-
 	( N < H ->
-		Reg = reg(r(N))
+		Reg = reg(r, N)
 	;
 		N1 is N + 1,
 		middle_rec__find_unused_register_2(T, N1, Reg)
@@ -402,12 +405,17 @@ middle_rec__find_used_registers_instr(restore_hp(Rval), Used0, Used) :-
 	middle_rec__find_used_registers_rval(Rval, Used0, Used).
 middle_rec__find_used_registers_instr(store_ticket(Lval), Used0, Used) :-
 	middle_rec__find_used_registers_lval(Lval, Used0, Used).
-middle_rec__find_used_registers_instr(restore_ticket(Rval), Used0, Used) :-
+middle_rec__find_used_registers_instr(reset_ticket(Rval, _Rsn), Used0, Used) :-
 	middle_rec__find_used_registers_rval(Rval, Used0, Used).
 middle_rec__find_used_registers_instr(discard_ticket, Used, Used).
+middle_rec__find_used_registers_instr(mark_ticket_stack(Lval), Used0, Used) :-
+	middle_rec__find_used_registers_lval(Lval, Used0, Used).
+middle_rec__find_used_registers_instr(discard_tickets_to(Rval), Used0, Used) :-
+	middle_rec__find_used_registers_rval(Rval, Used0, Used).
 middle_rec__find_used_registers_instr(incr_sp(_, _), Used, Used).
 middle_rec__find_used_registers_instr(decr_sp(_), Used, Used).
-middle_rec__find_used_registers_instr(pragma_c(_, Ins, _, Outs), Used0, Used) :-
+middle_rec__find_used_registers_instr(pragma_c(_, Ins, _, Outs, _),
+		Used0, Used) :-
 	insert_pragma_c_input_registers(Ins, Used0, Used1),
 	insert_pragma_c_output_registers(Outs, Used1, Used).
 
@@ -423,7 +431,7 @@ middle_rec__find_used_registers_lvals([Lval | Lvals], Used0, Used) :-
 :- mode middle_rec__find_used_registers_lval(in, di, uo) is det.
 
 middle_rec__find_used_registers_lval(Lval, Used0, Used) :-
-	( Lval = reg(r(N)) ->
+	( Lval = reg(r, N) ->
 		copy(N, N1),
 		set__insert(Used0, N1, Used)
 	; Lval = field(_, Rval, FieldNum) ->
@@ -462,8 +470,19 @@ middle_rec__find_used_registers_rval(Rval, Used0, Used) :-
 		Rval = binop(_, Rval1, Rval2),
 		middle_rec__find_used_registers_rval(Rval1, Used0, Used1),
 		middle_rec__find_used_registers_rval(Rval2, Used1, Used)
+	;
+		Rval = mem_addr(MemRef),
+		middle_rec__find_used_registers_mem_ref(MemRef, Used0, Used)
 	).
 
+:- pred middle_rec__find_used_registers_mem_ref(mem_ref, set(int), set(int)).
+:- mode middle_rec__find_used_registers_mem_ref(in, di, uo) is det.
+
+middle_rec__find_used_registers_mem_ref(stackvar_ref(_), Used, Used).
+middle_rec__find_used_registers_mem_ref(framevar_ref(_), Used, Used).
+middle_rec__find_used_registers_mem_ref(heap_ref(Rval, _, _), Used0, Used) :-
+	middle_rec__find_used_registers_rval(Rval, Used0, Used).
+
 :- pred middle_rec__find_used_registers_maybe_rvals(list(maybe(rval)),
 	set(int), set(int)).
 :- mode middle_rec__find_used_registers_maybe_rvals(in, di, uo) is det.
