@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1995 University of Melbourne.
+% Copyright (C) 1994-1997 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -17,7 +17,7 @@
 
 :- interface.
 
-:- import_module prog_data, mode_info.
+:- import_module hlds_data, prog_data, mode_info, (inst).
 :- import_module set, assoc_list.
 
 %-----------------------------------------------------------------------------%
@@ -32,7 +32,7 @@
 	--->	delayed_goal(
 			set(var),	% The vars it's waiting on
 			mode_error_info,% The reason it can't be scheduled
-			hlds__goal	% The goal itself
+			hlds_goal	% The goal itself
 		).
 
 :- type mode_error
@@ -70,7 +70,7 @@
 			% a negated context
 	;	mode_error_unify_var_var(var, var, inst, inst)
 			% attempt to unify two free variables
-	;	mode_error_unify_var_functor(var, const, list(var),
+	;	mode_error_unify_var_functor(var, cons_id, list(var),
 							inst, list(inst))
 			% attempt to unify a free var with a functor containing
 			% free arguments
@@ -95,7 +95,7 @@
 
 :- type mode_error_unify_rhs
 	--->	error_at_var(var)
-	;	error_at_functor(const, list(var))
+	;	error_at_functor(cons_id, list(var))
 	;	error_at_lambda(list(var), list(mode)).
 
 :- type mode_error_info
@@ -265,7 +265,7 @@ report_mode_error_conj_2([delayed_goal(Vars, Error, Goal) | Rest],
 		prog_out__write_context(Context),
 		io__write_string("Floundered goal, waiting on { "),
 		{ set__to_sorted_list(Vars, VarList) },
-		mercury_output_vars(VarList, VarSet),
+		mercury_output_vars(VarList, VarSet, no),
 		io__write_string(" } :\n")
 	;
 		[]
@@ -274,7 +274,7 @@ report_mode_error_conj_2([delayed_goal(Vars, Error, Goal) | Rest],
 	( { VeryVerbose = yes } ->
 		io__write_string("\t\t"),
 		{ mode_info_get_module_info(ModeInfo, ModuleInfo) },
-		hlds_out__write_goal(Goal, ModuleInfo, VarSet, 2, ".")
+		hlds_out__write_goal(Goal, ModuleInfo, VarSet, no, 2, ".")
 	;
 		[]
 	),
@@ -309,9 +309,9 @@ write_merge_error_list([Var - Insts | Errors], ModeInfo) -->
 	{ mode_info_get_instvarset(ModeInfo, InstVarSet) },
 	prog_out__write_context(Context),
 	io__write_string("  `"),
-	mercury_output_var(Var, VarSet),
+	mercury_output_var(Var, VarSet, no),
 	io__write_string("' :: "),
-	mercury_output_inst_list(Insts, InstVarSet),
+	output_inst_list(Insts, InstVarSet),
 	io__write_string(".\n"),
 	write_merge_error_list(Errors, ModeInfo).
 
@@ -339,13 +339,13 @@ report_mode_error_bind_var(ModeInfo, Var, VarInst, Inst) -->
 		"  scope error: attempt to bind variable inside a negation.\n"),
 	prog_out__write_context(Context),
 	io__write_string("  Variable `"),
-	mercury_output_var(Var, VarSet),
+	mercury_output_var(Var, VarSet, no),
 	io__write_string("' has instantiatedness `"),
-	mercury_output_inst(VarInst, InstVarSet),
+	output_inst(VarInst, InstVarSet),
 	io__write_string("',\n"),
 	prog_out__write_context(Context),
 	io__write_string("  expected instantiatedness was `"),
-	mercury_output_inst(Inst, InstVarSet),
+	output_inst(Inst, InstVarSet),
 	io__write_string("'.\n"),
 	globals__io_lookup_bool_option(verbose_errors, VerboseErrors),
 	( { VerboseErrors = yes } ->
@@ -373,11 +373,11 @@ report_mode_error_no_matching_mode(ModeInfo, Vars, Insts) -->
 	mode_info_write_context(ModeInfo),
 	prog_out__write_context(Context),
 	io__write_string("  mode error: arguments `"),
-	mercury_output_vars(Vars, VarSet),
+	mercury_output_vars(Vars, VarSet, no),
 	io__write_string("'\n"),
 	prog_out__write_context(Context),
 	io__write_string("  have insts `"),
-	mercury_output_inst_list(Insts, InstVarSet),
+	output_inst_list(Insts, InstVarSet),
 	io__write_string("',\n"),
 	prog_out__write_context(Context),
 	io__write_string("  which does not match any of the modes for "),
@@ -403,9 +403,9 @@ report_mode_error_higher_order_pred_var(ModeInfo, PredOrFunc, Var, VarInst,
 	mode_info_write_context(ModeInfo),
 	prog_out__write_context(Context),
 	io__write_string("  mode error: variable `"),
-	mercury_output_var(Var, VarSet),
+	mercury_output_var(Var, VarSet, no),
 	io__write_string("' has instantiatedness `"),
-	mercury_output_inst(VarInst, InstVarSet),
+	output_inst(VarInst, InstVarSet),
 	io__write_string("',\n"),
 	prog_out__write_context(Context),
 	(	{ PredOrFunc = predicate },
@@ -429,7 +429,7 @@ report_mode_error_var_is_live(ModeInfo, Var) -->
 	io__write_string("  unique-mode error: the called procedure would clobber\n"),
 	prog_out__write_context(Context),
 	io__write_string("  its argument, but variable `"),
-	mercury_output_var(Var, VarSet),
+	mercury_output_var(Var, VarSet, no),
 	io__write_string("' is still live.\n").
 
 :- pred report_mode_error_var_has_inst(mode_info, var, inst, inst,
@@ -443,13 +443,13 @@ report_mode_error_var_has_inst(ModeInfo, Var, VarInst, Inst) -->
 	mode_info_write_context(ModeInfo),
 	prog_out__write_context(Context),
 	io__write_string("  mode error: variable `"),
-	mercury_output_var(Var, VarSet),
+	mercury_output_var(Var, VarSet, no),
 	io__write_string("' has instantiatedness `"),
-	mercury_output_inst(VarInst, InstVarSet),
+	output_inst(VarInst, InstVarSet),
 	io__write_string("',\n"),
 	prog_out__write_context(Context),
 	io__write_string("  expected instantiatedness was `"),
-	mercury_output_inst(Inst, InstVarSet),
+	output_inst(Inst, InstVarSet),
 	io__write_string("'.\n").
 
 :- pred report_mode_error_implied_mode(mode_info, var, inst, inst,
@@ -470,13 +470,13 @@ report_mode_error_implied_mode(ModeInfo, Var, VarInst, Inst) -->
 		io__write_string("  sorry, implied modes not implemented.\n"),
 		prog_out__write_context(Context),
 		io__write_string("  Variable `"),
-		mercury_output_var(Var, VarSet),
+		mercury_output_var(Var, VarSet, no),
 		io__write_string("' has instantiatedness `"),
-		mercury_output_inst(VarInst, InstVarSet),
+		output_inst(VarInst, InstVarSet),
 		io__write_string("',\n"),
 		prog_out__write_context(Context),
 		io__write_string("  expected instantiatedness was `"),
-		mercury_output_inst(Inst, InstVarSet),
+		output_inst(Inst, InstVarSet),
 		io__write_string("'.\n")
 	;
 		[]
@@ -503,18 +503,18 @@ report_mode_error_unify_pred(ModeInfo, X, RHS, Type, PredOrFunc) -->
 	mode_info_write_context(ModeInfo),
 	prog_out__write_context(Context),
 	io__write_string("  In unification of `"),
-	mercury_output_var(X, VarSet),
+	mercury_output_var(X, VarSet, no),
 	io__write_string("' with `"),
 	(
 		{ RHS = error_at_var(Y) },
-		mercury_output_var(Y, VarSet)
+		mercury_output_var(Y, VarSet, no)
 	;
-		{ RHS = error_at_functor(Const, ArgVars) },
-		hlds_out__write_functor(Const, ArgVars, VarSet)
+		{ RHS = error_at_functor(ConsId, ArgVars) },
+		hlds_out__write_functor_cons_id(ConsId, ArgVars, VarSet, no)
 	;
 		{ RHS = error_at_lambda(ArgVars, ArgModes) },
 		io__write_string("lambda(["),
-		hlds_out__write_var_modes(ArgVars, ArgModes, VarSet),
+		hlds_out__write_var_modes(ArgVars, ArgModes, VarSet, no),
 		io__write_string("] ... )")
 	),
 	io__write_string("':\n"),
@@ -523,7 +523,7 @@ report_mode_error_unify_pred(ModeInfo, X, RHS, Type, PredOrFunc) -->
 	prog_out__write_context(Context),
 	io__write_string("  Cannot unify two terms of type `"),
 	{ varset__init(TypeVarSet) },
-	mercury_output_term(Type, TypeVarSet),
+	mercury_output_term(Type, TypeVarSet, no),
 	io__write_string("'.\n"),
 	globals__io_lookup_bool_option(verbose_errors, VerboseErrors),
 	( { VerboseErrors = yes } ->
@@ -557,21 +557,21 @@ report_mode_error_unify_var_var(ModeInfo, X, Y, InstX, InstY) -->
 	mode_info_write_context(ModeInfo),
 	prog_out__write_context(Context),
 	io__write_string("  mode error in unification of `"),
-	mercury_output_var(X, VarSet),
+	mercury_output_var(X, VarSet, no),
 	io__write_string("' and `"),
-	mercury_output_var(Y, VarSet),
+	mercury_output_var(Y, VarSet, no),
 	io__write_string("'.\n"),
 	prog_out__write_context(Context),
 	io__write_string("  Variable `"),
-	mercury_output_var(X, VarSet),
+	mercury_output_var(X, VarSet, no),
 	io__write_string("' has instantiatedness `"),
-	mercury_output_inst(InstX, InstVarSet),
+	output_inst(InstX, InstVarSet),
 	io__write_string("',\n"),
 	prog_out__write_context(Context),
 	io__write_string("  variable `"),
-	mercury_output_var(Y, VarSet),
+	mercury_output_var(Y, VarSet, no),
 	io__write_string("' has instantiatedness `"),
-	mercury_output_inst(InstY, InstVarSet),
+	output_inst(InstY, InstVarSet),
 	io__write_string("'.\n").
 
 %-----------------------------------------------------------------------------%
@@ -588,27 +588,27 @@ report_mode_error_unify_var_lambda(ModeInfo, X, InstX, InstY) -->
 	mode_info_write_context(ModeInfo),
 	prog_out__write_context(Context),
 	io__write_string("  mode error in unification of `"),
-	mercury_output_var(X, VarSet),
+	mercury_output_var(X, VarSet, no),
 	io__write_string("' and lambda expression.\n"),
 	prog_out__write_context(Context),
 	io__write_string("  Variable `"),
-	mercury_output_var(X, VarSet),
+	mercury_output_var(X, VarSet, no),
 	io__write_string("' has instantiatedness `"),
-	mercury_output_inst(InstX, InstVarSet),
+	output_inst(InstX, InstVarSet),
 	io__write_string("',\n"),
 	prog_out__write_context(Context),
 	io__write_string("  lambda expression has instantiatedness `"),
-	mercury_output_inst(InstY, InstVarSet),
+	output_inst(InstY, InstVarSet),
 	io__write_string("'.\n").
 
 %-----------------------------------------------------------------------------%
 
-:- pred report_mode_error_unify_var_functor(mode_info, var, const, list(var),
+:- pred report_mode_error_unify_var_functor(mode_info, var, cons_id, list(var),
 					inst, list(inst), io__state, io__state).
 :- mode report_mode_error_unify_var_functor(mode_info_ui, in, in, in, in, in,
 			di, uo) is det.
 
-report_mode_error_unify_var_functor(ModeInfo, X, Name, Args, InstX, ArgInsts)
+report_mode_error_unify_var_functor(ModeInfo, X, ConsId, Args, InstX, ArgInsts)
 		-->
 	{ mode_info_get_context(ModeInfo, Context) },
 	{ mode_info_get_varset(ModeInfo, VarSet) },
@@ -616,30 +616,30 @@ report_mode_error_unify_var_functor(ModeInfo, X, Name, Args, InstX, ArgInsts)
 	mode_info_write_context(ModeInfo),
 	prog_out__write_context(Context),
 	io__write_string("  mode error in unification of `"),
-	mercury_output_var(X, VarSet),
+	mercury_output_var(X, VarSet, no),
 	io__write_string("' and `"),
-	hlds_out__write_functor(Name, Args, VarSet),
+	hlds_out__write_functor_cons_id(ConsId, Args, VarSet, no),
 	io__write_string("'.\n"),
 	prog_out__write_context(Context),
 	io__write_string("  Variable `"),
-	mercury_output_var(X, VarSet),
+	mercury_output_var(X, VarSet, no),
 	io__write_string("' has instantiatedness `"),
-	mercury_output_inst(InstX, InstVarSet),
+	output_inst(InstX, InstVarSet),
 	io__write_string("',\n"),
 	prog_out__write_context(Context),
 	io__write_string("  term `"),
-	hlds_out__write_functor(Name, Args, VarSet),
+	hlds_out__write_functor_cons_id(ConsId, Args, VarSet, no),
 	( { Args \= [] } ->
 		io__write_string("'\n"),
 		prog_out__write_context(Context),
 		io__write_string("  has instantiatedness `"),
-		term_io__write_constant(Name),
+		mercury_output_cons_id(ConsId, no),
 		io__write_string("("),
-		mercury_output_inst_list(ArgInsts, InstVarSet),
+		output_inst_list(ArgInsts, InstVarSet),
 		io__write_string(")")
 	;
 		io__write_string("' has instantiatedness `"),
-		term_io__write_constant(Name)
+		mercury_output_cons_id(ConsId, no)
 	),
 	io__write_string("'.\n").
 
@@ -658,7 +658,7 @@ mode_info_write_context(ModeInfo) -->
 	{ pred_info_procedures(PredInfo, Procs) },
 	{ pred_info_get_is_pred_or_func(PredInfo, PredOrFunc) },
 	{ map__lookup(Procs, ProcId, ProcInfo) },
-	{ proc_info_argmodes(ProcInfo, Modes0) },
+	{ proc_info_declared_argmodes(ProcInfo, Modes0) },
 	{ strip_builtin_qualifiers_from_mode_list(Modes0, Modes) },
 	{ pred_info_name(PredInfo, Name0) },
 	{ Name = unqualified(Name0) },
@@ -706,14 +706,14 @@ report_mode_error_final_inst(ModeInfo, ArgNum, Var, VarInst, Inst, Reason) -->
 
 	prog_out__write_context(Context),
 	io__write_string("  Final instantiatedness of `"),
-	mercury_output_var(Var, VarSet),
+	mercury_output_var(Var, VarSet, no),
 	io__write_string("' was `"),
-	mercury_output_inst(VarInst, InstVarSet),
+	output_inst(VarInst, InstVarSet),
 	io__write_string("',\n"),
 
 	prog_out__write_context(Context),
 	io__write_string("  expected final instantiatedness was `"),
-	mercury_output_inst(Inst, InstVarSet),
+	output_inst(Inst, InstVarSet),
 	io__write_string("'.\n").
 
 
@@ -895,5 +895,21 @@ report_mode_errors(ModeInfo0, ModeInfo) :-
                 ModeInfo = ModeInfo0
         ).
 
+%-----------------------------------------------------------------------------%
+
+:- pred output_inst((inst), varset, io__state, io__state).
+:- mode output_inst(in, in, di, uo) is det.
+
+output_inst(Inst0, VarSet) -->
+	{ strip_builtin_qualifiers_from_inst(Inst0, Inst) },
+	mercury_output_inst(Inst, VarSet).
+
+:- pred output_inst_list(list(inst), varset, io__state, io__state).
+:- mode output_inst_list(in, in, di, uo) is det.
+
+output_inst_list(Insts0, VarSet) -->
+	{ strip_builtin_qualifiers_from_inst_list(Insts0, Insts) },
+	mercury_output_inst_list(Insts, VarSet).
+
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
