@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1995 University of Melbourne.
+% Copyright (C) 1996-1997 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -21,10 +21,11 @@
 :- interface.
 
 :- import_module hlds_goal, mode_info, modes.
+:- import_module term.
 
 	% Modecheck a unification
 :- pred modecheck_unification( var, unify_rhs, unification, unify_context,
-			hlds__goal_info, how_to_check_goal, hlds__goal_expr,
+			hlds_goal_info, how_to_check_goal, hlds_goal_expr,
 			mode_info, mode_info).
 :- mode modecheck_unification(in, in, in, in, in, in, out,
 			mode_info_di, mode_info_uo) is det.
@@ -32,19 +33,27 @@
 	% Work out what kind of unification a var-var unification is.
 :- pred categorize_unify_var_var(mode, mode, is_live, is_live, var, var,
 			determinism, unify_context, map(var, type), mode_info,
-			hlds__goal_expr, mode_info).
+			hlds_goal_expr, mode_info).
 :- mode categorize_unify_var_var(in, in, in, in, in, in, in, in, in,
 			mode_info_di, out, mode_info_uo) is det.
 
+	% Given a unification goal, which must be something that can be
+	% returned from categorize_unify_var_var (i.e.  a unification
+	% of two vars, a unification of a var and a functor, or a
+	% conjunction thereof), return a list of all the variables in the
+	% goal.
+:- pred unify_vars(hlds_goal_expr, list(var)).
+:- mode unify_vars(in, out) is det.
+
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
 
 :- implementation.
-:- import_module llds, prog_data, type_util, module_qual, instmap.
+:- import_module llds, prog_data, prog_util, type_util, module_qual, instmap.
 :- import_module hlds_module, hlds_goal, hlds_pred, hlds_data, hlds_out.
 :- import_module mode_debug, mode_util, mode_info, modes, mode_errors.
-:- import_module inst_match, unify_proc, code_util, unique_modes.
-:- import_module typecheck, modecheck_call.
+:- import_module inst_match, inst_util, unify_proc, code_util, unique_modes.
+:- import_module typecheck, modecheck_call, (inst), quantification.
 :- import_module bool, list, std_util, int, map, set, require, varset.
 :- import_module string, assoc_list.
 
@@ -97,7 +106,7 @@ modecheck_unification(X, var(Y), _Unification0, UnifyContext, _GoalInfo, _,
 		Unify = unify(X, var(Y), Modes, Unification, UnifyContext)
 	).
 
-modecheck_unification(X0, functor(Name, ArgVars0), Unification0,
+modecheck_unification(X0, functor(ConsId0, ArgVars0), Unification0,
 			UnifyContext, GoalInfo0, HowToCheckGoal,
 			Goal, ModeInfo0, ModeInfo) :-
 	mode_info_get_module_info(ModeInfo0, ModuleInfo0),
@@ -116,9 +125,9 @@ modecheck_unification(X0, functor(Name, ArgVars0), Unification0,
 		% been expanded.)
 		%
 		HowToCheckGoal \= check_unique_modes,
-		Name = cons(unqualified(ApplyName), _),
+		ConsId0 = cons(unqualified(ApplyName), _),
 		( ApplyName = "apply" ; ApplyName = "" ),
-		Arity >= 2,
+		Arity >= 1,
 		ArgVars0 = [FuncVar | FuncArgVars]
 	->
 		%
@@ -143,7 +152,23 @@ modecheck_unification(X0, functor(Name, ArgVars0), Unification0,
 
 		% Find the set of candidate predicates which have the
 		% specified name and arity (and module, if module-qualified)
-		Name = cons(PredName, _),
+		ConsId0 = cons(PredName, _),
+		module_info_pred_info(ModuleInfo0, ThisPredId, PredInfo),
+
+		%
+		% We don't do this for compiler-generated predicates;
+		% they are assumed to have been generated with all
+		% functions already expanded.
+		% If we did this check for compiler-generated
+		% predicates, it would cause the wrong behaviour
+		% in the case where there is a user-defined function
+		% whose type is exactly the same as the type of
+		% a constructor.  (Normally that would cause
+		% a type ambiguity error, but compiler-generated
+		% predicates are not type-checked.)
+		%
+
+		\+ code_util__compiler_generated(PredInfo),
 		(
 			PredName = unqualified(UnqualPName),
 			predicate_table_search_func_name_arity(PredTable,
@@ -153,27 +178,12 @@ modecheck_unification(X0, functor(Name, ArgVars0), Unification0,
 			% have argument/return types which subsume the actual
 			% argument/return types of this function call
   
-			module_info_pred_info(ModuleInfo0,
-					ThisPredId, PredInfo),
 			pred_info_typevarset(PredInfo, TVarSet),
 			map__apply_to_list(ArgVars0, VarTypes0, ArgTypes0),
 			list__append(ArgTypes0, [TypeOfX], ArgTypes),
 			typecheck__find_matching_pred_id(PredIds, ModuleInfo0,
-				TVarSet, ArgTypes, PredId, QualifiedFuncName),
+				TVarSet, ArgTypes, PredId, QualifiedFuncName)
 
-			%
-			% We don't do this for compiler-generated predicates;
-			% they are assumed to have been generated with all
-			% functions already expanded.
-			% If we did this check for compiler-generated
-			% predicates, it would cause the wrong behaviour
-			% in the case where there is a user-defined function
-			% whose type is exactly the same as the type of
-			% a constructor.  (Normally that would cause
-			% a type ambiguity error, but compiler-generated
-			% predicates are not type-checked.)
-			%
-			\+ code_util__compiler_generated(PredInfo)
 		;
 			PredName = qualified(FuncModule, UnqualName),
 			predicate_table_search_func_m_n_a(PredTable,
@@ -186,12 +196,11 @@ modecheck_unification(X0, functor(Name, ArgVars0), Unification0,
 		% replace `X = f(A, B, C)'
 		% with `f(A, B, C, X)'
 		%
-		ProcId = 0,
+		invalid_proc_id(ProcId),
 		list__append(ArgVars0, [X0], ArgVars),
-		hlds__is_builtin_make_builtin(no, no, Builtin),
 		FuncCallUnifyContext = call_unify_context(X0,
-					functor(Name, ArgVars0), UnifyContext),
-		FuncCall = call(PredId, ProcId, ArgVars, Builtin,
+			functor(ConsId0, ArgVars0), UnifyContext),
+		FuncCall = call(PredId, ProcId, ArgVars, not_builtin,
 			yes(FuncCallUnifyContext), QualifiedFuncName),
 		%
 		% now modecheck it
@@ -222,9 +231,9 @@ modecheck_unification(X0, functor(Name, ArgVars0), Unification0,
 	% back into a higher-order pred constant again.
 	%
 
-		% check if variable has a higher-order pred type
+		% check if variable has a higher-order type
 		type_is_higher_order(TypeOfX, PredOrFunc, PredArgTypes),
-		Name = cons(PName, _),
+		ConsId0 = cons(PName, _),
 		% but in case we are redoing mode analysis, make sure
 		% we don't mess with the address constants for type_info
 		% fields created by polymorphism.m
@@ -242,45 +251,26 @@ modecheck_unification(X0, functor(Name, ArgVars0), Unification0,
 		mode_info_set_var_types(VarTypes, ModeInfo1, ModeInfo2),
 
 		%
-		% Build up the hlds__goal_expr for the call that will form
+		% Build up the hlds_goal_expr for the call that will form
 		% the lambda goal
 		%
-		(
-			PName = unqualified(UnqualPName),
-			get_pred_id_and_proc_id(UnqualPName, Arity, PredOrFunc,
-				PredArgTypes, ModuleInfo0, PredId, ProcId),
-			module_info_pred_info(ModuleInfo0, PredId, PredInfo),
-			pred_info_module(PredInfo, PredModule),
-			QualifiedPName = qualified(PredModule, UnqualPName)
-		;
-			PName = qualified(PredModule, UnqualName),
-			QualifiedPName = PName,
-			(
-				predicate_table_search_sym(PredTable,
-					PName, [PredId0]),
-				module_info_pred_info(ModuleInfo0, PredId0,
-					PredInfo),
-				pred_info_procids(PredInfo, [ProcId0])
-			->
-				ProcId = ProcId0,
-				PredId = PredId0
-			;
-				string__int_to_string(Arity, ArStr),
-				string__append_list([
-					"sorry, not implemented: ",
-					"taking address of ", "\n`",
-					PredModule, ":", UnqualName, "/",
-					ArStr, "' with multiple modes.\n",
-					"(use an explicit lambda expression instead)"],
-					Message),
-				error(Message)
-			)
-		),
 
-		hlds__is_builtin_make_builtin(no, no, Builtin),
+		module_info_pred_info(ModuleInfo0, ThisPredId, ThisPredInfo),
+		pred_info_typevarset(ThisPredInfo, TVarSet),
+		map__apply_to_list(Args, VarTypes, ArgTypes),
+		get_pred_id_and_proc_id(PName, PredOrFunc, TVarSet, ArgTypes,
+			ModuleInfo0, PredId, ProcId),
+		module_info_pred_proc_info(ModuleInfo0, PredId, ProcId,
+					PredInfo, ProcInfo),
+
+		% module-qualify the pred name (is this necessary?)
+		pred_info_module(PredInfo, PredModule),
+		unqualify_name(PName, UnqualPName),
+		QualifiedPName = qualified(PredModule, UnqualPName),
+
 		CallUnifyContext = call_unify_context(X0,
-					functor(Name, ArgVars0), UnifyContext),
-		LambdaGoalExpr = call(PredId, ProcId, Args, Builtin,
+				functor(ConsId0, ArgVars0), UnifyContext),
+		LambdaGoalExpr = call(PredId, ProcId, Args, not_builtin,
 				yes(CallUnifyContext), QualifiedPName),
 
 		%
@@ -292,7 +282,10 @@ modecheck_unification(X0, functor(Name, ArgVars0), Unification0,
 		set__list_to_set(Args, InsideVars),
 		set__intersect(OutsideVars, InsideVars, LambdaNonLocals),
 		goal_info_init(LambdaGoalInfo0),
-		goal_info_set_nonlocals(LambdaGoalInfo0, LambdaNonLocals,
+		mode_info_get_context(ModeInfo2, Context),
+		goal_info_set_context(LambdaGoalInfo0, Context,
+				LambdaGoalInfo1),
+		goal_info_set_nonlocals(LambdaGoalInfo1, LambdaNonLocals,
 				LambdaGoalInfo),
 		LambdaGoal = LambdaGoalExpr - LambdaGoalInfo,
 
@@ -300,8 +293,6 @@ modecheck_unification(X0, functor(Name, ArgVars0), Unification0,
 		% work out the modes of the introduced lambda variables
 		% and the determinism of the lambda goal
 		%
-		module_info_pred_proc_info(ModuleInfo0, PredId, ProcId,
-					_PredInfo, ProcInfo),
 		proc_info_argmodes(ProcInfo, ArgModes),
 		( list__drop(Arity, ArgModes, LambdaModes0) ->
 			LambdaModes = LambdaModes0
@@ -329,54 +320,13 @@ modecheck_unification(X0, functor(Name, ArgVars0), Unification0,
 		% It's not a higher-order pred unification - just
 		% call modecheck_unify_functor to do the ordinary thing.
 		%
-		( cons_id_to_const(Name, Const0, _) ->
-			Const = Const0
-		;
-			% This should be caught by typecheck.
-			error("sorry, not implemented: module qualified constructors")
-		),
-		mode_info_get_instmap(ModeInfo0, InstMap0),
-		modecheck_unify_functor(X0, Const, ArgVars0, Unification0,
-				ExtraGoals, Mode, ArgVars, Unification,
-				ModeInfo0, ModeInfo),
-		%
-		% Optimize away construction of unused terms by
-		% replace the unification with `true'.
-		%
-		(
-			Unification = construct(ConstructTarget, _, _, _),
-			mode_info_var_is_live(ModeInfo, ConstructTarget, dead)
-		->
-			Goal = conj([])
-		;
-			Functor = functor(Name, ArgVars),
-			Unify = unify(X, Functor, Mode, Unification,
-				UnifyContext),
-			X = X0,
-			%
-			% modecheck_unification sometimes needs to introduce
-			% new goals to handle complicated sub-unifications
-			% in deconstructions.  But this should never happen
-			% during unique mode analysis.
-			% (If it did, the code would be wrong since it
-			% wouldn't have the correct determinism annotations.)
-			%
-			(
-				HowToCheckGoal = check_unique_modes,
-				ExtraGoals \= [] - []
-			->
-				error("unique_modes.m: re-modecheck of unification encountered complicated sub-unifies")
-			;
-				true
-			),
-
-			handle_extra_goals(Unify, ExtraGoals, GoalInfo0,
-						[X0|ArgVars0], [X|ArgVars],
-						InstMap0, ModeInfo, Goal)
-		)
+		modecheck_unify_functor(X0, TypeOfX,
+			ConsId0, ArgVars0, Unification0, UnifyContext,
+			HowToCheckGoal, GoalInfo0,
+			Goal, ModeInfo0, ModeInfo)
 	).
 
-modecheck_unification(X, lambda_goal(PredOrFunc, Vars, Modes, Det, Goal0),
+modecheck_unification(X, lambda_goal(PredOrFunc, Vars, Modes0, Det, Goal0),
 			Unification0, UnifyContext, _GoalInfo, HowToCheckGoal,
 			unify(X, RHS, Mode, Unification, UnifyContext),
 			ModeInfo0, ModeInfo) :-
@@ -404,6 +354,15 @@ modecheck_unification(X, lambda_goal(PredOrFunc, Vars, Modes, Det, Goal0),
 	%
 
 	mode_info_get_module_info(ModeInfo0, ModuleInfo0),
+
+	( HowToCheckGoal = check_modes ->
+		% This only needs to be done once.
+		mode_info_get_types_of_vars(ModeInfo0, Vars, VarTypes),
+		propagate_types_into_mode_list(VarTypes, ModuleInfo0,
+			Modes0, Modes)
+ 	;
+		Modes = Modes0
+	),
  
 	% initialize the initial insts of the lambda variables
 	mode_list_get_initial_insts(Modes, ModuleInfo0, VarInitialInsts),
@@ -504,24 +463,32 @@ modecheck_unify_lambda(X, PredOrFunc, ArgVars, LambdaModes, LambdaDet,
 		Unification = Unification0
 	).
 
-:- pred modecheck_unify_functor(var, const, list(var), unification,
-			pair(list(hlds__goal)), pair(mode), list(var),
-			unification,
-			mode_info, mode_info).
-:- mode modecheck_unify_functor(in, in, in, in, out, out, out, out,
-			mode_info_di, mode_info_uo) is det.
+:- pred modecheck_unify_functor(var, (type), cons_id, list(var), unification,
+			unify_context, how_to_check_goal, hlds_goal_info,
+			hlds_goal_expr, mode_info, mode_info).
+:- mode modecheck_unify_functor(in, in, in, in, in, in, in, in,
+			out, mode_info_di, mode_info_uo) is det.
 
-modecheck_unify_functor(X, Name, ArgVars0, Unification0,
-			ExtraGoals, Mode, ArgVars, Unification,
-			ModeInfo0, ModeInfo) :-
+modecheck_unify_functor(X, TypeOfX, ConsId0, ArgVars0, Unification0,
+			UnifyContext, HowToCheckGoal, GoalInfo0,
+			Goal, ModeInfo0, ModeInfo) :-
 	mode_info_get_module_info(ModeInfo0, ModuleInfo0),
+	list__length(ArgVars0, Arity),
+	(
+		% module qualify cons_ids
+		ConsId0 = cons(unqualified(Name), _),
+		type_to_type_id(TypeOfX, TypeId, _),
+		TypeId = qualified(TypeModule, _) - _
+	->
+		ConsId = cons(qualified(TypeModule, Name), Arity)
+	;
+		ConsId = ConsId0
+	),
 	mode_info_get_instmap(ModeInfo0, InstMap0),
 	instmap__lookup_var(InstMap0, X, InstOfX),
 	instmap__lookup_vars(ArgVars0, InstMap0, InstArgs),
 	mode_info_var_is_live(ModeInfo0, X, LiveX),
 	mode_info_var_list_is_live(ArgVars0, ModeInfo0, LiveArgs),
-	list__length(ArgVars0, Arity),
-	make_functor_cons_id(Name, Arity, ConsId),
 	InstOfY = bound(unique, [functor(ConsId, InstArgs)]),
 	(
 		% The occur check: X = f(X) is considered a mode error
@@ -535,11 +502,12 @@ modecheck_unify_functor(X, Name, ArgVars0, Unification0,
 	->
 		set__list_to_set([X], WaitingVars),
 		mode_info_error(WaitingVars,
-			mode_error_unify_var_functor(X, Name, ArgVars0,
+			mode_error_unify_var_functor(X, ConsId, ArgVars0,
 							InstOfX, InstArgs),
 			ModeInfo0, ModeInfo1
 		),
 		Inst = not_reached,
+		Det = erroneous,
 			% If we get an error, set the inst to not_reached
 			% to avoid cascading errors
 			% But don't call categorize_unification, because
@@ -557,13 +525,14 @@ modecheck_unify_functor(X, Name, ArgVars0, Unification0,
 			% return any old garbage
 		Unification = Unification0,
 		ArgVars = ArgVars0,
-		ExtraGoals = [] - []
+		ExtraGoals = no_extra_goals
 	;
-		abstractly_unify_inst_functor(LiveX, InstOfX, Name,
+		abstractly_unify_inst_functor(LiveX, InstOfX, ConsId,
 			InstArgs, LiveArgs, real_unify, ModuleInfo0,
-			UnifyInst, ModuleInfo1)
+			UnifyInst, Det1, ModuleInfo1)
 	->
 		Inst = UnifyInst,
+		Det = Det1,
 		mode_info_set_module_info(ModeInfo0, ModuleInfo1, ModeInfo1),
 		ModeOfX = (InstOfX -> Inst),
 		ModeOfY = (InstOfY -> Inst),
@@ -584,12 +553,12 @@ modecheck_unify_functor(X, Name, ArgVars0, Unification0,
 		),
 		mode_info_get_var_types(ModeInfo1, VarTypes),
 		categorize_unify_var_functor(ModeOfX, ModeOfXArgs, ModeArgs,
-				X, Name, ArgVars0, VarTypes,
+				X, ConsId, ArgVars0, VarTypes,
 				Unification0, ModeInfo1,
 				Unification1, ModeInfo2),
 		split_complicated_subunifies(Unification1, ArgVars0,
-					Unification, ArgVars, ExtraGoals,
-					ModeInfo2, ModeInfo3),
+				Unification, ArgVars, ExtraGoals,
+				ModeInfo2, ModeInfo3),
 		modecheck_set_var_inst(X, Inst, ModeInfo3, ModeInfo4),
 		( bind_args(Inst, ArgVars, ModeInfo4, ModeInfo5) ->
 			ModeInfo = ModeInfo5
@@ -599,7 +568,7 @@ modecheck_unify_functor(X, Name, ArgVars0, Unification0,
 	;
 		set__list_to_set([X | ArgVars0], WaitingVars), % conservative
 		mode_info_error(WaitingVars,
-			mode_error_unify_var_functor(X, Name, ArgVars0,
+			mode_error_unify_var_functor(X, ConsId, ArgVars0,
 							InstOfX, InstArgs),
 			ModeInfo0, ModeInfo1
 		),
@@ -609,6 +578,7 @@ modecheck_unify_functor(X, Name, ArgVars0, Unification0,
 			% that could cause an invalid call to
 			% `unify_proc__request_unify'
 		Inst = not_reached,
+		Det = erroneous,
 		ModeOfX = (InstOfX -> Inst),
 		ModeOfY = (InstOfY -> Inst),
 		Mode = ModeOfX - ModeOfY,
@@ -621,35 +591,57 @@ modecheck_unify_functor(X, Name, ArgVars0, Unification0,
 			% return any old garbage
 		Unification = Unification0,
 		ArgVars = ArgVars0,
-		ExtraGoals = [] - []
+		ExtraGoals = no_extra_goals
+	),
+
+	%
+	% Optimize away construction of unused terms by
+	% replacing the unification with `true'.  Optimize
+	% away unifications which always fail by replacing
+	% them with `fail'.
+	%
+	(
+		Unification = construct(ConstructTarget, _, _, _),
+		mode_info_var_is_live(ModeInfo, ConstructTarget, dead)
+	->
+		Goal = conj([])
+	;
+		Det = failure
+	->
+		% This optimisation is safe because the only way that
+		% we can analyse a unification as having no solutions
+		% is that the unification always fails.
+		%
+		% Unifying two preds is not erroneous as far as the
+		% mode checker is concerned, but a mode _error_.
+		map__init(Empty),
+		Goal = disj([], Empty)
+	;
+		Functor = functor(ConsId, ArgVars),
+		Unify = unify(X, Functor, Mode, Unification,
+			UnifyContext),
+		X = X0,
+		%
+		% modecheck_unification sometimes needs to introduce
+		% new goals to handle complicated sub-unifications
+		% in deconstructions.  But this should never happen
+		% during unique mode analysis.
+		% (If it did, the code would be wrong since it
+		% wouldn't have the correct determinism annotations.)
+		%
+		(
+			HowToCheckGoal = check_unique_modes,
+			ExtraGoals \= no_extra_goals
+		->
+			error("unique_modes.m: re-modecheck of unification encountered complicated sub-unifies")
+		;
+			true
+		),
+		handle_extra_goals(Unify, ExtraGoals, GoalInfo0,
+					[X0|ArgVars0], [X|ArgVars],
+					InstMap0, ModeInfo, Goal)
 	).
 
-%-----------------------------------------------------------------------------%
-
-:- pred modecheck_higher_order_func_call(var, list(var), var, hlds__goal_info,
-		hlds__goal_expr, mode_info, mode_info).
-:- mode modecheck_higher_order_func_call(in, in, in, in, out,
-		mode_info_di, mode_info_uo) is det.
-
-modecheck_higher_order_func_call(FuncVar, Args0, RetVar, GoalInfo0, Goal) -->
-	mode_checkpoint(enter, "higher-order function call"),
-	mode_info_set_call_context(higher_order_call(function)),
-
-	=(ModeInfo0),
-	{ mode_info_get_instmap(ModeInfo0, InstMap0) },
-
-	{ list__append(Args0, [RetVar], Args1) },
-	modecheck_higher_order_call(function, FuncVar, Args1,
-			Types, Modes, Det, Args, ExtraGoals),
-
-	=(ModeInfo),
-	{ Call = higher_order_call(FuncVar, Args, Types, Modes, Det) },
-	{ handle_extra_goals(Call, ExtraGoals, GoalInfo0, Args1, Args,
-				InstMap0, ModeInfo, Goal) },
-
-	mode_info_unset_call_context,
-	mode_checkpoint(exit, "higher-order function call").
-
 %-----------------------------------------------------------------------------%
 
 	% The argument unifications in a construction or deconstruction
@@ -658,7 +650,7 @@ modecheck_higher_order_func_call(FuncVar, Args0, RetVar, GoalInfo0, Goal) -->
 	% into separate unifications by introducing fresh variables here.
 
 :- pred split_complicated_subunifies(unification, list(var),
-			unification, list(var), pair(list(hlds__goal)),
+			unification, list(var), extra_goals,
 			mode_info, mode_info).
 :- mode split_complicated_subunifies(in, in, out, out, out,
 			mode_info_di, mode_info_uo) is det.
@@ -683,16 +675,16 @@ split_complicated_subunifies(Unification0, ArgVars0,
 	;
 		{ Unification = Unification0 },
 		{ ArgVars = ArgVars0 },
-		{ ExtraGoals = [] - [] }
+		{ ExtraGoals = no_extra_goals }
 	).
 
 :- pred split_complicated_subunifies_2(list(var), list(uni_mode),
-			list(var), list(uni_mode), pair(list(hlds__goal)),
+			list(var), list(uni_mode), extra_goals,
 			mode_info, mode_info).
 :- mode split_complicated_subunifies_2(in, in, out, out, out,
-		mode_info_di, mode_info_uo) is semidet.
+			mode_info_di, mode_info_uo) is semidet.
 
-split_complicated_subunifies_2([], [], [], [], [] - []) --> [].
+split_complicated_subunifies_2([], [], [], [], no_extra_goals) --> [].
 split_complicated_subunifies_2([Var0 | Vars0], [UniMode0 | UniModes0],
 			Vars, UniModes, ExtraGoals, ModeInfo0, ModeInfo) :-
 	mode_info_get_module_info(ModeInfo0, ModuleInfo),
@@ -705,23 +697,22 @@ split_complicated_subunifies_2([Var0 | Vars0], [UniMode0 | UniModes0],
 		mode_to_arg_mode(ModuleInfo, ModeX, VarType, top_in),
 		mode_to_arg_mode(ModuleInfo, ModeY, VarType, top_in)
 	->
-		split_complicated_subunifies_2(Vars0, UniModes0,
-				Vars1, UniModes1, ExtraGoals0,
-				ModeInfo0, ModeInfo1),
-		ExtraGoals0 = BeforeGoals - AfterGoals0,
-
 		% introduce a new variable `Var'
-		mode_info_get_varset(ModeInfo1, VarSet1),
-		mode_info_get_var_types(ModeInfo1, VarTypes1),
-		varset__new_var(VarSet1, Var, VarSet),
-		map__set(VarTypes1, Var, VarType, VarTypes),
-		mode_info_set_varset(VarSet, ModeInfo1, ModeInfo2),
-		mode_info_set_var_types(VarTypes, ModeInfo2, ModeInfo3),
+		mode_info_get_varset(ModeInfo0, VarSet0),
+		mode_info_get_var_types(ModeInfo0, VarTypes0),
+		varset__new_var(VarSet0, Var, VarSet),
+		map__set(VarTypes0, Var, VarType, VarTypes),
+		mode_info_set_varset(VarSet, ModeInfo0, ModeInfo1),
+		mode_info_set_var_types(VarTypes, ModeInfo1, ModeInfo2),
 
 		% change the main unification to use `Var' instead of Var0
 		UniMode = (InitialInstX - free -> InitialInstX - InitialInstX),
-		UniModes = [UniMode | UniModes1],
-		Vars = [Var | Vars1],
+
+		% Compute the instmap that results after the main unification.
+		% We just need to set the inst of `Var'.
+		mode_info_get_instmap(ModeInfo2, OrigInstMap),
+		instmap__set(OrigInstMap, Var, InitialInstX, InstMapAfterMain),
+		mode_info_set_instmap(InstMapAfterMain, ModeInfo2, ModeInfo3),
 
 		% create code to do a unification between `Var' and `Var0'
 		ModeVar0 = (InitialInstY -> FinalInstY),
@@ -747,32 +738,43 @@ split_complicated_subunifies_2([Var0 | Vars0], [UniMode0 | UniModes0],
 						UnifyContext),
 		categorize_unify_var_var(ModeVar0, ModeVar,
 			live, dead, Var0, Var, Det, UnifyContext,
-			VarTypes, ModeInfo4, AfterGoal, ModeInfo),
+			VarTypes, ModeInfo4, NewUnifyGoal, ModeInfo5),
+		unify_vars(NewUnifyGoal, NewUnifyGoalVars),
 
 		% compute the goal_info nonlocal vars & instmap delta
 		% for the newly created goal
-		set__list_to_set([Var0, Var], NonLocals),
+		% N.B. This may overestimate the set of non-locals,
+		% but that shouldn't cause any problems.
+		set__list_to_set(NewUnifyGoalVars, NonLocals),
 		( InitialInstY = FinalInstY ->
-			InstMapAL0 = []
+			InstMapDeltaAL0 = []
 		;
-			InstMapAL0 = [Var0 - FinalInstY]
+			InstMapDeltaAL0 = [Var0 - FinalInstY]
 		),
-		InstMapAL = [Var - FinalInstX | InstMapAL0],
-		instmap_delta_from_assoc_list(InstMapAL, InstMapDelta),
+		InstMapDeltaAL = [Var - FinalInstX | InstMapDeltaAL0],
+		instmap_delta_from_assoc_list(InstMapDeltaAL, InstMapDelta),
 		goal_info_init(GoalInfo0),
-		goal_info_set_nonlocals(GoalInfo0, NonLocals,
-			GoalInfo1),
+		goal_info_set_nonlocals(GoalInfo0, NonLocals, GoalInfo1),
 		goal_info_set_instmap_delta(GoalInfo1, InstMapDelta, GoalInfo),
 
-		% insert the unification between `Var' and `Var0' at
-		% the start of the AfterGoals
-		AfterGoals = [AfterGoal - GoalInfo | AfterGoals0],
-		ExtraGoals = BeforeGoals - AfterGoals
-	;
-		Vars = [Var0 | Vars1],
-		UniModes = [UniMode0 | UniModes1],
+		% insert the new unification at
+		% the start of the extra goals
+		ExtraGoals0 = extra_goals(InstMapAfterMain,
+					[NewUnifyGoal - GoalInfo]),
+
+		% recursive call to handle the remaining variables...
 		split_complicated_subunifies_2(Vars0, UniModes0,
-			Vars1, UniModes1, ExtraGoals, ModeInfo0, ModeInfo)
+				Vars1, UniModes1, ExtraGoals1,
+				ModeInfo5, ModeInfo),
+		Vars = [Var | Vars1],
+		UniModes = [UniMode | UniModes1],
+		append_extra_goals(ExtraGoals0, ExtraGoals1, ExtraGoals)
+	;
+		split_complicated_subunifies_2(Vars0, UniModes0,
+				Vars1, UniModes1, ExtraGoals,
+				ModeInfo0, ModeInfo),
+		Vars = [Var0 | Vars1],
+		UniModes = [UniMode0 | UniModes1]
 	).
 
 %-----------------------------------------------------------------------------%
@@ -844,8 +846,9 @@ categorize_unify_var_var(ModeOfX, ModeOfY, LiveX, LiveY, X, Y, Det,
 			;
 				type_to_type_id(Type, TypeId, _)
 			->
-				unify_proc__request_unify(TypeId - UniMode, Det,
-					ModuleInfo0, ModuleInfo),
+				mode_info_get_context(ModeInfo0, Context),
+				unify_proc__request_unify(TypeId - UniMode,
+					Det, Context, ModuleInfo0, ModuleInfo),
 				mode_info_set_module_info(ModeInfo0, ModuleInfo,
 					ModeInfo)
 			;
@@ -861,6 +864,8 @@ categorize_unify_var_var(ModeOfX, ModeOfY, LiveX, LiveY, X, Y, Det,
 	% because otherwise determinism analysis assumes they can fail.
 	% The optimization of assignments to dead variables may be
 	% necessary to stop the code generator from getting confused.)
+	% Optimize away unifications which always fail by replacing
+	% them with `fail'.
 	%
 	(
 		Unification = assign(AssignTarget, _),
@@ -872,6 +877,17 @@ categorize_unify_var_var(ModeOfX, ModeOfY, LiveX, LiveY, X, Y, Det,
 		Det = det
 	->
 		Unify = conj([])
+	;
+		Det = failure
+	->
+		% This optimisation is safe because the only way that
+		% we can analyse a unification as having no solutions
+		% is that the unification always fails.
+		%
+		% Unifying two preds is not erroneous as far as the
+		% mode checker is concerned, but a mode _error_.
+		map__init(Empty),
+		Unify = disj([], Empty)
 	;
 		Unify = unify(X, var(Y), ModeOfX - ModeOfY, Unification,
 				UnifyContext)
@@ -928,17 +944,16 @@ categorize_unify_var_lambda(ModeOfX, ArgModes0, X, ArgVars, PredOrFunc,
 % unification or a deconstruction.  It also works out whether it will
 % be deterministic or semideterministic.
 
-:- pred categorize_unify_var_functor(mode, list(mode), list(mode), var, const,
-			list(var), map(var, type), unification, mode_info,
-			unification, mode_info).
+:- pred categorize_unify_var_functor(mode, list(mode), list(mode), var,
+		cons_id, list(var), map(var, type),
+		unification, mode_info, unification, mode_info).
 :- mode categorize_unify_var_functor(in, in, in, in, in, in, in, in,
 			mode_info_di, out, mode_info_uo) is det.
 
 categorize_unify_var_functor(ModeOfX, ModeOfXArgs, ArgModes0,
-		X, Name, ArgVars, VarTypes, Unification0, ModeInfo0,
-		Unification, ModeInfo) :-
+		X, NewConsId, ArgVars, VarTypes,
+		Unification0, ModeInfo0, Unification, ModeInfo) :-
 	mode_info_get_module_info(ModeInfo0, ModuleInfo),
-	list__length(ArgVars, Arity),
 	map__lookup(VarTypes, X, TypeOfX),
 	% if we are re-doing mode analysis, preserve the existing cons_id
 	( Unification0 = construct(_, ConsId0, _, _) ->
@@ -946,7 +961,7 @@ categorize_unify_var_functor(ModeOfX, ModeOfXArgs, ArgModes0,
 	; Unification0 = deconstruct(_, ConsId1, _, _, _) ->
 		ConsId = ConsId1
 	;
-		make_functor_cons_id(Name, Arity, ConsId)
+		ConsId = NewConsId
 	),
 	mode_util__modes_to_uni_modes(ModeOfXArgs, ArgModes0,
 						ModuleInfo, ArgModes),
@@ -989,7 +1004,7 @@ categorize_unify_var_functor(ModeOfX, ModeOfXArgs, ArgModes0,
 				set__init(WaitingVars),
 				mode_info_error(WaitingVars,
 			mode_error_unify_pred(X,
-					error_at_functor(Name, ArgVars),
+					error_at_functor(ConsId, ArgVars),
 					TypeOfX, PredOrFunc),
 					ModeInfo0, ModeInfo)
 			;
@@ -1037,41 +1052,6 @@ ground_args(Uniq, [Arg | Args]) -->
 
 %-----------------------------------------------------------------------------%
 
-:- pred get_arg_insts(inst, cons_id, arity, list(inst)).
-:- mode get_arg_insts(in, in, in, out) is semidet.
-
-get_arg_insts(not_reached, _ConsId, Arity, ArgInsts) :-
-	list__duplicate(Arity, not_reached, ArgInsts).
-get_arg_insts(ground(Uniq, _PredInst), _ConsId, Arity, ArgInsts) :-
-	list__duplicate(Arity, ground(Uniq, no), ArgInsts).
-get_arg_insts(bound(_Uniq, List), ConsId, Arity, ArgInsts) :-
-	( get_arg_insts_2(List, ConsId, ArgInsts0) ->
-		ArgInsts = ArgInsts0
-	;
-		% the code is unreachable
-		list__duplicate(Arity, not_reached, ArgInsts)
-	).
-get_arg_insts(free, _ConsId, Arity, ArgInsts) :-
-	list__duplicate(Arity, free, ArgInsts).
-get_arg_insts(free(_Type), _ConsId, Arity, ArgInsts) :-
-	list__duplicate(Arity, free, ArgInsts).
-get_arg_insts(any(Uniq), _ConsId, Arity, ArgInsts) :-
-	list__duplicate(Arity, any(Uniq), ArgInsts).
-
-:- pred get_arg_insts_2(list(bound_inst), cons_id, list(inst)).
-:- mode get_arg_insts_2(in, in, out) is semidet.
-
-get_arg_insts_2([BoundInst | BoundInsts], ConsId, ArgInsts) :-
-	(
-		BoundInst = functor(ConsId, ArgInsts0)
-	->
-		ArgInsts = ArgInsts0
-	;
-		get_arg_insts_2(BoundInsts, ConsId, ArgInsts)
-	).
-
-%-----------------------------------------------------------------------------%
-
 % get_mode_of_args(FinalInst, InitialArgInsts, ArgModes):
 %       for a var-functor unification,
 %       given the final inst of the var
@@ -1125,5 +1105,35 @@ make_fresh_vars([Type|Types], VarSet0, VarTypes0,
 	map__det_insert(VarTypes0, Var, Type, VarTypes1),
 	make_fresh_vars(Types, VarSet1, VarTypes1, Vars, VarSet, VarTypes).
 
+%-----------------------------------------------------------------------------%
+
+	% Given a unification goal, which must be something that can be
+	% returned from categorize_unify_var_var (i.e. a unification
+	% of two vars, a unification of a var and a functor,
+	% an empty disjunction (fail), or a conjunction thereof),
+	% return a list of all the variables in the goal.
+unify_vars(Unify, UnifyVars) :-
+	( Unify = unify(LHS, RHS, _, _, _), RHS = var(RHS_Var) ->
+		UnifyVars = [LHS, RHS_Var]
+	; Unify = unify(LHS, RHS, _, _, _), RHS = functor(_, RHS_ArgVars) ->
+		UnifyVars = [LHS | RHS_ArgVars]
+	; Unify = conj(List) ->
+		unify_conj_vars(List, UnifyVars)
+	; Unify = disj([], _StoreMap) ->
+		UnifyVars = []
+	;
+		error("modecheck_unify: unify_vars")
+	).
+
+:- pred unify_conj_vars(list(hlds_goal), list(var)).
+:- mode unify_conj_vars(in, out) is det.
+
+unify_conj_vars([], []).
+unify_conj_vars([Goal | Goals], Vars) :-
+	quantification__goal_vars(Goal, Vars1Set),
+	set__to_sorted_list(Vars1Set, Vars1),
+	unify_conj_vars(Goals, Vars2),
+	list__append(Vars1, Vars2, Vars).
+
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
