@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1995 University of Melbourne.
+% Copyright (C) 1994-1997 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -81,15 +81,16 @@ If goal is
 	(d) a unification
 		Check that the unification doesn't attempt to unify
 		two free variables (or in general two free sub-terms)
-		unless one of them is dead. (Also we ought to split
-		unifications up if necessary to avoid complicated
-		sub-unifications.)
+		unless one of them is dead.  Split unifications
+		up if necessary to avoid complicated sub-unifications.
+		We also figure out at this point whether or not each
+		unification can fail.
 	(e) a predicate call
 		Check that there is a mode declaration for the
 		predicate which matches the current instantiation of
 		the arguments.  (Also handle calls to implied modes.)
 		If the called predicate is one for which we must infer
-		the modes, then a new mode for the called predicate
+		the modes, then create a new mode for the called predicate
 		whose initial insts are the result of normalising
 		the current inst of the arguments.
 	(f) an if-then-else
@@ -114,11 +115,14 @@ a variable live if its value will be used later on in the computation.
 
 ******************************************/
 
-% XXX we need to allow unification of free with free even when both
+% XXX we ought to allow unification of free with free even when both
 %     *variables* are live, if one of the particular *sub-nodes* is 
-%     dead (causes problems handling e.g. `same_length').
-% XXX break unifications into "micro-unifications"
-% XXX would even the above fixes be enough?
+%     dead (causes problems handling e.g. `list__same_length').
+
+% XXX we ought to break unifications into "micro-unifications", because
+%     some code can't be scheduled without splitting up unifications.
+%     For example, `p(X) :- X = f(A, B), B is A + 1.', where
+%     p is declared as `:- mode p(bound(f(ground,free))->ground).'.
 
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
@@ -127,7 +131,7 @@ a variable live if its value will be used later on in the computation.
 
 :- interface.
 
-:- import_module hlds_module, hlds_pred, instmap.
+:- import_module hlds_module, hlds_pred, (inst), instmap.
 :- import_module bool, io.
 
 	% modecheck(HLDS0, HLDS, UnsafeToContinue):
@@ -152,11 +156,15 @@ a variable live if its value will be used later on in the computation.
 			io__state, io__state).
 :- mode modecheck_proc(in, in, in, out, out, di, uo) is det.
 
+:- pred modecheck_proc_info(proc_id, pred_id, module_info, proc_info,
+		module_info, proc_info, int, io__state, io__state).
+:- mode modecheck_proc_info(in, in, in, in, out, out, out, di, uo) is det.
+
 %-----------------------------------------------------------------------------%
 
 % The following predicates are used by unique_modes.m.
 
-:- import_module mode_info, mode_errors, clause_to_proc.
+:- import_module mode_info.
 
 	% Modecheck a unification.
 
@@ -200,7 +208,7 @@ a variable live if its value will be used later on in the computation.
 :- mode modecheck_set_var_inst(in, in, mode_info_di, mode_info_uo) is det.
 
 :- pred modecheck_set_var_inst_list(list(var), list(inst), list(inst),
-					list(var), pair(list(hlds__goal)),
+					list(var), extra_goals,
 					mode_info, mode_info).
 :- mode modecheck_set_var_inst_list(in, in, in, out, out,
 					mode_info_di, mode_info_uo) is det.
@@ -211,10 +219,10 @@ a variable live if its value will be used later on in the computation.
 :- pred modecheck_final_insts(list(var), list(inst), mode_info, mode_info).
 :- mode modecheck_final_insts(in, in, mode_info_di, mode_info_uo) is det.
 
-:- pred mode_info_add_goals_live_vars(list(hlds__goal), mode_info, mode_info).
+:- pred mode_info_add_goals_live_vars(list(hlds_goal), mode_info, mode_info).
 :- mode mode_info_add_goals_live_vars(in, mode_info_di, mode_info_uo) is det.
 
-:- pred mode_info_remove_goals_live_vars(list(hlds__goal), mode_info,
+:- pred mode_info_remove_goals_live_vars(list(hlds_goal), mode_info,
 					mode_info).
 :- mode mode_info_remove_goals_live_vars(in, mode_info_di, mode_info_uo) is det.
 
@@ -222,21 +230,36 @@ a variable live if its value will be used later on in the computation.
 
 % The following predicates are used by modecheck_unify.m.
 
-:- pred modecheck_goal(hlds__goal, hlds__goal, mode_info, mode_info).
+:- pred modecheck_goal(hlds_goal, hlds_goal, mode_info, mode_info).
 :- mode modecheck_goal(in, out, mode_info_di, mode_info_uo) is det.
 
 	% Mode-check a single goal-expression.
 
-:- pred modecheck_goal_expr(hlds__goal_expr, hlds__goal_info, hlds__goal_expr,
+:- pred modecheck_goal_expr(hlds_goal_expr, hlds_goal_info, hlds_goal_expr,
 			mode_info, mode_info).
 :- mode modecheck_goal_expr(in, in, out, mode_info_di, mode_info_uo) is det.
 
-	% handle_extra_goals combines MainGoal and ExtraGoals into a single
-	% hlds__goal_expr.
+
+:- type extra_goals
+	--->	no_extra_goals
+	;	extra_goals(
+			instmap,		% instmap at end of main goal
+			list(hlds_goal)		% goals to append after
+						% the main goal
+		).
+
+	% append_extra_goals inserts adds some goals to the
+	% list of goals to insert before/after the main goal.
 	%
-:- pred handle_extra_goals(hlds__goal_expr, pair(list(hlds__goal)),
-		hlds__goal_info, list(var), list(var),
-		instmap, mode_info, hlds__goal_expr).
+:- pred append_extra_goals(extra_goals, extra_goals, extra_goals).
+:- mode append_extra_goals(in, in, out) is det.
+
+	% handle_extra_goals combines MainGoal and ExtraGoals into a single
+	% hlds_goal_expr.
+	%
+:- pred handle_extra_goals(hlds_goal_expr, extra_goals,
+		hlds_goal_info, list(var), list(var),
+		instmap, mode_info, hlds_goal_expr).
 :- mode handle_extra_goals(in, in, in, in, in, in, mode_info_ui, out)
 	is det.
 
@@ -252,8 +275,8 @@ a variable live if its value will be used later on in the computation.
 :- import_module mode_info, delay_info, mode_errors, inst_match, instmap.
 :- import_module type_util, mode_util, code_util, prog_data, unify_proc.
 :- import_module globals, options, mercury_to_mercury, hlds_out, int, set.
-:- import_module passes_aux, typecheck, module_qual.
-:- import_module modecheck_unify, modecheck_call.
+:- import_module passes_aux, typecheck, module_qual, clause_to_proc.
+:- import_module modecheck_unify, modecheck_call, inst_util.
 :- import_module list, map, varset, term, prog_out, string, require, std_util.
 :- import_module assoc_list.
 
@@ -455,6 +478,11 @@ modecheck_proc_2(ProcId, PredId, ModuleInfo0, Changed0,
 		{ module_info_set_preds(ModuleInfo1, Preds, ModuleInfo) }
 	).
 
+modecheck_proc_info(ProcId, PredId, ModuleInfo0, ProcInfo0,
+		ModuleInfo, ProcInfo, NumErrors) -->
+	modecheck_proc_3(ProcId, PredId, ModuleInfo0, ProcInfo0, no,
+		ModuleInfo, ProcInfo, _Changed, NumErrors).
+
 :- pred modecheck_proc_3(proc_id, pred_id, module_info, proc_info, bool,
 			module_info, proc_info, bool, int,
 			io__state, io__state).
@@ -481,23 +509,16 @@ modecheck_proc_3(ProcId, PredId, ModuleInfo0, ProcInfo0, Changed0,
 	;
 		proc_info_context(ProcInfo0, Context)
 	),
-/**************
-		% extract the predicate's type from the pred_info
-		% and propagate the type information into the modes
-	pred_info_arg_types(PredInfo, _TypeVars, ArgTypes),
-	propagate_type_info_mode_list(ArgTypes, ModuleInfo0, ArgModes0,
-			ArgModes1),
-**************/
-	ArgModes1 = ArgModes0,
+
 		% modecheck the clause - first set the initial instantiation
 		% of the head arguments, mode-check the body, and
 		% then check that the final instantiation matches that in
 		% the mode declaration
-	mode_list_get_initial_insts(ArgModes1, ModuleInfo0, ArgInitialInsts),
+	mode_list_get_initial_insts(ArgModes0, ModuleInfo0, ArgInitialInsts),
 	assoc_list__from_corresponding_lists(HeadVars, ArgInitialInsts, InstAL),
 	instmap__from_assoc_list(InstAL, InstMap0),
 		% initially, only the non-clobbered head variables are live
-	mode_list_get_final_insts(ArgModes1, ModuleInfo0, ArgFinalInsts0),
+	mode_list_get_final_insts(ArgModes0, ModuleInfo0, ArgFinalInsts0),
 	get_live_vars(HeadVars, ArgLives0, LiveVarsList),
 	set__list_to_set(LiveVarsList, LiveVars),
 	mode_info_init(IOState0, ModuleInfo0, PredId, ProcId,
@@ -713,7 +734,6 @@ modecheck_goal_expr(if_then_else(Vs, A0, B0, C0, SM), GoalInfo0, Goal) -->
 	modecheck_goal(A0, A),
 	mode_info_remove_live_vars(B_Vars),
 	mode_info_unlock_vars(NonLocals),
-	mode_info_dcg_get_instmap(InstMapA),
 	modecheck_goal(B0, B),
 	mode_info_dcg_get_instmap(InstMapB),
 	mode_info_set_instmap(InstMap0),
@@ -721,29 +741,7 @@ modecheck_goal_expr(if_then_else(Vs, A0, B0, C0, SM), GoalInfo0, Goal) -->
 	mode_info_dcg_get_instmap(InstMapC),
 	mode_info_set_instmap(InstMap0),
 	instmap__merge(NonLocals, [InstMapB, InstMapC], if_then_else),
-	( { instmap__is_unreachable(InstMapA) } ->
-		% if the condition can never succeed, we delete the
-		% unreachable `then' part by replacing
-		%	if some [Vs] A then B else C
-		% with
-		%	(not some [Vs] A), C
-		{ goal_get_nonlocals(A, A_Vars) },
-		{ goal_get_nonlocals(C, C_Vars) },
-		{ set__union(NonLocals, C_Vars, OutsideVars) },
-		{ set__delete_list(OutsideVars, Vs, OutsideVars1) },
-		{ set__intersect(OutsideVars1, A_Vars, SomeA_NonLocals) },
-		{ goal_info_init(EmptyGoalInfo) },
-		{ goal_info_set_nonlocals(EmptyGoalInfo, SomeA_NonLocals,
-			SomeA_GoalInfo) },
-		{ instmap_delta_init_reachable(EmptyInstmapDelta) },
-		{ goal_info_set_instmap_delta(SomeA_GoalInfo,
-			EmptyInstmapDelta, NotSomeA_GoalInfo) },
-		
-		{ Goal = conj([not(some(Vs, A) - SomeA_GoalInfo) -
-				NotSomeA_GoalInfo, C]) }
-	;
-		{ Goal = if_then_else(Vs, A, B, C, SM) }
-	),
+	{ Goal = if_then_else(Vs, A, B, C, SM) },
 	mode_checkpoint(exit, "if-then-else").
 
 modecheck_goal_expr(not(A0), GoalInfo0, not(A)) -->
@@ -776,7 +774,7 @@ modecheck_goal_expr(call(PredId0, _, Args0, _, Context, PredName0),
 
 	=(ModeInfo),
 	{ mode_info_get_module_info(ModeInfo, ModuleInfo) },
-	{ code_util__is_builtin(ModuleInfo, PredId, Mode, Builtin) },
+	{ code_util__builtin_state(ModuleInfo, PredId, Mode, Builtin) },
 	{ Call = call(PredId, Mode, Args, Builtin, Context, PredName) },
 	{ handle_extra_goals(Call, ExtraGoals, GoalInfo0, Args0, Args,
 				InstMap0, ModeInfo, Goal) },
@@ -784,9 +782,10 @@ modecheck_goal_expr(call(PredId0, _, Args0, _, Context, PredName0),
 	mode_info_unset_call_context,
 	mode_checkpoint(exit, "call").
 
-modecheck_goal_expr(higher_order_call(PredVar, Args0, _, _, _),
+modecheck_goal_expr(higher_order_call(PredVar, Args0, _, _, _, PredOrFunc),
 		GoalInfo0, Goal) -->
-	modecheck_higher_order_pred_call(PredVar, Args0, GoalInfo0, Goal).
+	modecheck_higher_order_pred_call(PredVar, Args0, PredOrFunc, GoalInfo0,
+		Goal).
 
 modecheck_goal_expr(unify(A0, B0, _, UnifyInfo0, UnifyContext), GoalInfo0, Goal)
 		-->
@@ -814,7 +813,7 @@ modecheck_goal_expr(switch(Var, CanFail, Cases0, SM), GoalInfo0,
 	% to modecheck a pragma_c_code, we just modecheck the proc for 
 	% which it is the goal.
 modecheck_goal_expr(pragma_c_code(IsRecursive, C_Code, PredId, _ProcId0, Args0,
-			ArgNameMap), GoalInfo, Goal) -->
+		ArgNameMap, OrigArgTypes, ExtraPragmaInfo), GoalInfo, Goal) -->
 	mode_checkpoint(enter, "pragma_c_code"),
 	mode_info_set_call_context(call(PredId)),
 
@@ -824,7 +823,7 @@ modecheck_goal_expr(pragma_c_code(IsRecursive, C_Code, PredId, _ProcId0, Args0,
 
 	=(ModeInfo),
 	{ Pragma = pragma_c_code(IsRecursive, C_Code, PredId, ProcId, Args0,
-			ArgNameMap) },
+			ArgNameMap, OrigArgTypes, ExtraPragmaInfo) },
 	{ handle_extra_goals(Pragma, ExtraGoals, GoalInfo, Args0, Args,
 				InstMap0, ModeInfo, Goal) },
 
@@ -842,12 +841,36 @@ unify_rhs_vars(lambda_goal(_PredOrFunc, LambdaVars, _Modes, _Det,
 	set__delete_list(NonLocals0, LambdaVars, NonLocals),
 	set__to_sorted_list(NonLocals, Vars).
 
+append_extra_goals(no_extra_goals, ExtraGoals, ExtraGoals).
+append_extra_goals(extra_goals(InstMap, AfterGoals),
+		no_extra_goals, extra_goals(InstMap, AfterGoals)).
+append_extra_goals(extra_goals(InstMap0, AfterGoals0),
+			extra_goals(_InstMap1, AfterGoals1),
+			extra_goals(InstMap, AfterGoals)) :-
+	InstMap = InstMap0,
+	list__append(AfterGoals0, AfterGoals1, AfterGoals).
+
 handle_extra_goals(MainGoal, ExtraGoals, GoalInfo0, Args0, Args,
-		InstMap0, ModeInfo, Goal) :-
+		InstMapAtStart, _ModeInfo, Goal) :-
 	% did we introduced any extra variables (and code)?
-	( ExtraGoals = [] - [] ->
+	(
+		ExtraGoals = no_extra_goals,
 		Goal = MainGoal	% no
 	;
+		ExtraGoals = extra_goals(InstMapAfterMain, AfterGoals0),
+
+		%
+		% We need to be careful to update the delta-instmaps
+		% correctly, using the appropriate instmaps:
+		%
+		%		% InstMapAtStart is here
+		%	 main goal,
+		%		% InstMapAfterMain is here
+		%	 AfterGoals
+		%		% _InstMapAtEnd (= the instmap from _ModeInfo)
+		%		% is here, but as it happens we don't need it
+		%
+
 		% recompute the new set of non-local variables for the main goal
 		goal_info_get_nonlocals(GoalInfo0, NonLocals0),
 		set__list_to_set(Args0, OldArgVars),
@@ -858,23 +881,20 @@ handle_extra_goals(MainGoal, ExtraGoals, GoalInfo0, Args0, Args,
 		goal_info_set_nonlocals(GoalInfo0, NonLocals, GoalInfo1),
 
 		% compute the instmap delta for the main goal
-		mode_info_get_instmap(ModeInfo, InstMap),
-		compute_instmap_delta(InstMap0, InstMap, NonLocals,
-			DeltaInstMap),
+		compute_instmap_delta(InstMapAtStart, InstMapAfterMain,
+			NonLocals, DeltaInstMap),
 		goal_info_set_instmap_delta(GoalInfo1, DeltaInstMap, GoalInfo),
 
 		% combine the main goal and the extra goals into a conjunction
 		Goal0 = MainGoal - GoalInfo,
-		ExtraGoals = BeforeGoals0 - AfterGoals0,
 		goal_info_get_context(GoalInfo0, Context),
-		handle_extra_goals_contexts(BeforeGoals0, Context, BeforeGoals),
 		handle_extra_goals_contexts(AfterGoals0, Context, AfterGoals),
-		list__append(BeforeGoals, [Goal0 | AfterGoals], GoalList),
+		GoalList = [Goal0 | AfterGoals],
 		Goal = conj(GoalList)
 	).
 
-:- pred handle_extra_goals_contexts(list(hlds__goal), term__context,
-	list(hlds__goal)).
+:- pred handle_extra_goals_contexts(list(hlds_goal), term__context,
+	list(hlds_goal)).
 :- mode handle_extra_goals_contexts(in, in, out) is det.
 
 handle_extra_goals_contexts([], _Context, []).
@@ -884,7 +904,7 @@ handle_extra_goals_contexts([Goal0 | Goals0], Context, [Goal | Goals]) :-
 	goal_info_set_context(GoalInfo0, Context, GoalInfo),
 	handle_extra_goals_contexts(Goals0, Context, Goals).
 
-:- pred goal_get_nonlocals(hlds__goal, set(var)).
+:- pred goal_get_nonlocals(hlds_goal, set(var)).
 :- mode goal_get_nonlocals(in, out) is det.
 
 goal_get_nonlocals(_Goal - GoalInfo, NonLocals) :-
@@ -892,7 +912,7 @@ goal_get_nonlocals(_Goal - GoalInfo, NonLocals) :-
 
 %-----------------------------------------------------------------------------%
 
-:- pred modecheck_conj_list(list(hlds__goal), list(hlds__goal),
+:- pred modecheck_conj_list(list(hlds_goal), list(hlds_goal),
 				mode_info, mode_info).
 :- mode modecheck_conj_list(in, out, mode_info_di, mode_info_uo) is det.
 
@@ -938,7 +958,7 @@ mode_info_remove_goals_live_vars([Goal | Goals]) -->
 	mode_info_remove_live_vars(Vars),
 	mode_info_remove_goals_live_vars(Goals).
 
-:- pred modecheck_conj_list_2(list(hlds__goal), list(hlds__goal),
+:- pred modecheck_conj_list_2(list(hlds_goal), list(hlds_goal),
 				mode_info, mode_info).
 :- mode modecheck_conj_list_2(in, out, mode_info_di, mode_info_uo) is det.
 
@@ -1029,7 +1049,7 @@ get_all_waiting_vars_2([delayed_goal(Vars1, _, _) | Rest], Vars0, Vars) :-
 
 %-----------------------------------------------------------------------------%
 
-:- pred modecheck_disj_list(list(hlds__goal), list(hlds__goal), list(instmap),
+:- pred modecheck_disj_list(list(hlds_goal), list(hlds_goal), list(instmap),
 				mode_info, mode_info).
 :- mode modecheck_disj_list(in, out, out, mode_info_di, mode_info_uo) is det.
 
@@ -1055,20 +1075,15 @@ modecheck_case_list([Case0 | Cases0], Var,
 
 		% record the fact that Var was bound to ConsId in the
 		% instmap before processing this case
-	( { cons_id_to_const(ConsId, _Const, Arity) } ->
-		{ list__duplicate(Arity, free, ArgInsts) },
-		modecheck_set_var_inst(Var,
-			bound(unique, [functor(ConsId, ArgInsts)]))
-	;
-		% cons_id_to_const will fail for pred_consts and
-		% address_consts; we don't worry about them,
-		% since you can't have a switch on a higher-order
-		% pred term anyway.
-		[]
-	),
+	{ cons_id_arity(ConsId, Arity) },
+	{ list__duplicate(Arity, free, ArgInsts) },
+	modecheck_set_var_inst(Var,
+		bound(unique, [functor(ConsId, ArgInsts)])),
 
-	modecheck_goal(Goal0, Goal),
+	modecheck_goal(Goal0, Goal1),
 	mode_info_dcg_get_instmap(InstMap),
+	% Don't lose the information added by the functor test above.
+	{ fixup_switch_var(Var, InstMap0, InstMap, Goal1, Goal) },
 	mode_info_set_instmap(InstMap0),
 	modecheck_case_list(Cases0, Var, Cases, InstMaps).
 
@@ -1148,7 +1163,7 @@ modecheck_var_has_inst(VarId, Inst, ModeInfo0, ModeInfo) :-
 modecheck_set_var_inst_list(Vars0, InitialInsts, FinalInsts, Vars, Goals) -->
 	(
 		modecheck_set_var_inst_list_2(Vars0, InitialInsts, FinalInsts,
-			Vars1, Goals1)
+			no_extra_goals, Vars1, Goals1)
 	->
 		{ Vars = Vars1, Goals = Goals1 }
 	;
@@ -1156,30 +1171,27 @@ modecheck_set_var_inst_list(Vars0, InitialInsts, FinalInsts, Vars, Goals) -->
 	).
 
 :- pred modecheck_set_var_inst_list_2(list(var), list(inst), list(inst),
-					list(var), pair(list(hlds__goal)),
+					extra_goals, list(var), extra_goals,
 					mode_info, mode_info).
-:- mode modecheck_set_var_inst_list_2(in, in, in, out, out,
+:- mode modecheck_set_var_inst_list_2(in, in, in, in, out, out,
 					mode_info_di, mode_info_uo) is semidet.
 
-modecheck_set_var_inst_list_2([], [], [], [], [] - []) --> [].
+modecheck_set_var_inst_list_2([], [], [], ExtraGoals, [], ExtraGoals) --> [].
 modecheck_set_var_inst_list_2([Var0 | Vars0], [InitialInst | InitialInsts],
-			[FinalInst | FinalInsts], [Var | Vars], Goals) -->
+			[FinalInst | FinalInsts], ExtraGoals0,
+			[Var | Vars], ExtraGoals) -->
 	modecheck_set_var_inst(Var0, InitialInst, FinalInst,
-				Var, BeforeGoals0 - AfterGoals0),
+				Var, ExtraGoals0, ExtraGoals1),
 	modecheck_set_var_inst_list_2(Vars0, InitialInsts, FinalInsts,
-				Vars, BeforeGoals1 - AfterGoals1),
-	{ list__append(BeforeGoals0, BeforeGoals1, BeforeGoals) },
-	{ list__append(AfterGoals0, AfterGoals1, AfterGoals) },
-	{ Goals = BeforeGoals - AfterGoals }.
+				ExtraGoals1, Vars, ExtraGoals).
 
-% XXX this might need to be revisited to handle unique modes
-
-:- pred modecheck_set_var_inst(var, inst, inst, var, pair(list(hlds__goal)),
+:- pred modecheck_set_var_inst(var, inst, inst, var, extra_goals, extra_goals,
 				mode_info, mode_info).
-:- mode modecheck_set_var_inst(in, in, in, out, out,
+:- mode modecheck_set_var_inst(in, in, in, out, in, out,
 				mode_info_di, mode_info_uo) is det.
 
-modecheck_set_var_inst(Var0, InitialInst, FinalInst, Var, Goals,
+modecheck_set_var_inst(Var0, InitialInst, FinalInst, Var,
+			ExtraGoals0, ExtraGoals,
 			ModeInfo0, ModeInfo) :-
 	mode_info_get_instmap(ModeInfo0, InstMap0),
 	( instmap__is_reachable(InstMap0) ->
@@ -1201,12 +1213,12 @@ modecheck_set_var_inst(Var0, InitialInst, FinalInst, Var, Goals,
 		mode_info_set_module_info(ModeInfo0, ModuleInfo, ModeInfo1),
 		handle_implied_mode(Var0,
 			VarInst0, VarInst, InitialInst, FinalInst, Det,
-		 	Var, Goals, ModeInfo1, ModeInfo2),
+		 	Var, ExtraGoals0, ExtraGoals, ModeInfo1, ModeInfo2),
 		modecheck_set_var_inst(Var0, FinalInst, ModeInfo2, ModeInfo3),
 		modecheck_set_var_inst(Var, FinalInst, ModeInfo3, ModeInfo)
 	;
 		Var = Var0,
-		Goals = [] - [],
+		ExtraGoals = ExtraGoals0,
 		ModeInfo = ModeInfo0
 	).
 
@@ -1215,8 +1227,6 @@ modecheck_set_var_inst(Var0, InitialInst, FinalInst, Var, Goals,
 	% The former is used for predicate calls, where we may need
 	% to introduce unifications to handle calls to implied modes.
 
-% XXX this might need to be revisited to handle unique modes
-
 modecheck_set_var_inst(Var0, FinalInst, ModeInfo0, ModeInfo) :-
 	mode_info_get_instmap(ModeInfo0, InstMap0),
 	( instmap__is_reachable(InstMap0) ->
@@ -1290,13 +1300,13 @@ modecheck_set_var_inst(Var0, FinalInst, ModeInfo0, ModeInfo) :-
 % introduce a fresh variable.
 
 :- pred handle_implied_mode(var, inst, inst, inst, inst, determinism,
-				var, pair(list(hlds__goal)),
+				var, extra_goals, extra_goals,
 				mode_info, mode_info).
-:- mode handle_implied_mode(in, in, in, in, in, in, out, out,
+:- mode handle_implied_mode(in, in, in, in, in, in, out, in, out,
 				mode_info_di, mode_info_uo) is det.
 
 handle_implied_mode(Var0, VarInst0, VarInst, InitialInst, FinalInst, Det,
-		Var, Goals, ModeInfo0, ModeInfo) :-
+		Var, ExtraGoals0, ExtraGoals, ModeInfo0, ModeInfo) :-
 	mode_info_get_module_info(ModeInfo0, ModuleInfo0),
 	(
 		% If the initial inst of the variable matches_final
@@ -1306,17 +1316,18 @@ handle_implied_mode(Var0, VarInst0, VarInst, InitialInst, FinalInst, Det,
 		inst_matches_final(VarInst0, InitialInst, ModuleInfo0)
 	->
 		Var = Var0,
-		Goals = [] - [],
+		ExtraGoals = ExtraGoals0,
 		ModeInfo = ModeInfo0
 	;
 		% This is the implied mode case.
 		% We do not yet handle implied modes for partially
 		% instantiated vars, since that would require
-		% doing a deep copy, and we don't know how to do that yet.
+		% doing a partially instantiated deep copy, and we
+		% don't know how to do that yet.
 		( inst_is_bound(ModuleInfo0, InitialInst) ->
 			% This is the case we can't handle
 			Var = Var0,
-			Goals = [] - [],
+			ExtraGoals = ExtraGoals0,
 			set__singleton_set(WaitingVars, Var0),
 			mode_info_error(WaitingVars,
 				mode_error_implied_mode(Var0, VarInst0,
@@ -1336,34 +1347,52 @@ handle_implied_mode(Var0, VarInst0, VarInst, InitialInst, FinalInst, Det,
 			mode_info_set_varset(VarSet, ModeInfo0, ModeInfo1),
 			mode_info_set_var_types(VarTypes, ModeInfo1, ModeInfo2),
 
+			% Calculate the instmap after the main goal.
+			% We just need to set the inst of the newly
+			% introduced variable Var to the procedure
+			% argument's final inst.
+			mode_info_get_instmap(ModeInfo2, OrigInstMap),
+			instmap__set(OrigInstMap, Var, FinalInst,
+				InstMapAfterMain),
+			mode_info_set_instmap(InstMapAfterMain,
+				ModeInfo2, ModeInfo3),
+
 			% Construct the code to do the unification
 			ModeVar0 = (VarInst0 -> VarInst),
 			ModeVar = (FinalInst -> VarInst),
-			mode_info_get_mode_context(ModeInfo2, ModeContext),
-			mode_context_to_unify_context(ModeContext, ModeInfo2,
+			mode_info_get_mode_context(ModeInfo3, ModeContext),
+			mode_context_to_unify_context(ModeContext, ModeInfo3,
 				UnifyContext),
+
 			categorize_unify_var_var(ModeVar0, ModeVar,
 				live, dead, Var0, Var, Det, UnifyContext,
-				VarTypes, ModeInfo2,
-				AfterGoal, ModeInfo),
+				VarTypes, ModeInfo3, NewUnifyGoal, ModeInfo),
+			unify_vars(NewUnifyGoal, NewUnifyGoalVars),
 
 			% compute the goal_info nonlocal vars & instmap delta
-			set__list_to_set([Var0, Var], NonLocals),
+			% N.B.  This may overestimate the set of nonlocal vars,
+			% but that should not cause any problems.
+			set__list_to_set(NewUnifyGoalVars, NonLocals),
 			( VarInst = VarInst0 ->
 				InstMapDeltaAL0 = []
 			;
 				InstMapDeltaAL0 = [Var0 - VarInst]
 			),
-			
 			InstMapDeltaAL = [Var - VarInst | InstMapDeltaAL0],
+			instmap_delta_from_assoc_list(InstMapDeltaAL,
+				InstMapDelta),
 			goal_info_init(GoalInfo0),
 			goal_info_set_nonlocals(GoalInfo0, NonLocals,
 				GoalInfo1),
-			instmap_delta_from_assoc_list(InstMapDeltaAL,
-				InstMapDelta),
 			goal_info_set_instmap_delta(GoalInfo1, InstMapDelta,
 				GoalInfo),
-			Goals = [] - [AfterGoal - GoalInfo]
+
+			% append the goals together in the appropriate order:
+			% ExtraGoals0, then NewUnify
+			NewUnifyExtraGoal = extra_goals(InstMapAfterMain,
+						[NewUnifyGoal - GoalInfo]),
+			append_extra_goals(ExtraGoals0, NewUnifyExtraGoal,
+				ExtraGoals)
 		)
 	).
 
