@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1995 University of Melbourne.
+% Copyright (C) 1996-1997 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -32,38 +32,44 @@
 	% ReportUndefErrors should be no when module qualifying the
 	% short interface.
 :- pred module_qual__module_qualify_items(item_list, item_list,
-		string, bool, int, bool, bool, io__state, io__state).
+		string, bool, mq_info, int, bool, bool, io__state, io__state).
 :- mode module_qual__module_qualify_items(in, out, in, in,
-		out, out, out, di, uo) is det.
+		out, out, out, out, di, uo) is det.
 
 	% This is called from make_hlds.m to qualify the mode of a lambda
 	% expression.
-:- pred module_qual__qualify_mode_list(list(mode), list(mode),
+:- pred module_qual__qualify_lambda_mode_list(list(mode), list(mode),
 		term__context, mq_info, mq_info,
 		io__state, io__state) is det.
-:- mode module_qual__qualify_mode_list(in, out, in, in, out, di, uo) is det.
+:- mode module_qual__qualify_lambda_mode_list(in, out, 
+		in, in, out, di, uo) is det.
 
-:- pred module_qual__qualify_type(type, type, term__context,
+	% This is called from make_hlds.m to qualify an 
+	% explicit type qualification.
+:- pred module_qual__qualify_type_qualification(type, type, term__context,
 		mq_info, mq_info, io__state, io__state).
-:- mode module_qual__qualify_type(in, out, in, in, out, di, uo) is det.
+:- mode module_qual__qualify_type_qualification(in, out, in, in,
+		out, di, uo) is det.
 
 :- type mq_info.
 
-:- pred init_mq_info_module(module_info::in, mq_info::out) is det.
-
 :- pred mq_info_get_num_errors(mq_info::in, int::out) is det.
 :- pred mq_info_get_type_error_flag(mq_info::in, bool::out) is det.
 :- pred mq_info_get_mode_error_flag(mq_info::in, bool::out) is det.
+:- pred mq_info_set_need_qual_flag(mq_info::in, 
+		need_qualifier::in, mq_info::out) is det.
+:- pred mq_info_get_need_qual_flag(mq_info::in, need_qualifier::out) is det.
 
 %-----------------------------------------------------------------------------%
 :- implementation.
 
 :- import_module hlds_data, hlds_module, hlds_pred, type_util, prog_out.
 :- import_module prog_util, mercury_to_mercury, globals, options.
+:- import_module (inst), instmap.
 :- import_module int, list, map, require, set, std_util, string, term, varset.
 
 module_qual__module_qualify_items(Items0, Items, ModuleName, ReportErrors,
-			NumErrors, UndefTypes, UndefModes) -->
+			Info, NumErrors, UndefTypes, UndefModes) -->
 	{ init_mq_info(ReportErrors, Info0) },
 	{ collect_mq_info(Items0, Info0, Info1) },
 	do_module_qualify_items(Items0, Items, Info1, Info),
@@ -78,16 +84,14 @@ module_qual__module_qualify_items(Items0, Items, ModuleName, ReportErrors,
 	),
 	{ mq_info_get_num_errors(Info, NumErrors) }.
 
-
-module_qual__qualify_mode_list(Modes0, Modes, Context, Info0, Info) -->
+module_qual__qualify_lambda_mode_list(Modes0, Modes, Context, Info0, Info) -->
 	{ mq_info_set_error_context(Info0, lambda_expr - Context, Info1) },
 	qualify_mode_list(Modes0, Modes, Info1, Info).
 
-module_qual__qualify_type(Type0, Type, Context, Info0, Info) -->
+module_qual__qualify_type_qualification(Type0, Type, Context, Info0, Info) -->
 	{ mq_info_set_error_context(Info0, type_qual - Context, Info1) },
 	qualify_type(Type0, Type, Info1, Info).
 
-
 :- type mq_info
 	--->	mq_info(
 			type_id_set,	% Sets of all types, modes and
@@ -102,10 +106,10 @@ module_qual__qualify_type(Type0, Type, Context, Info0, Info) -->
 			bool,	% are there any undefined insts or modes.
 			bool, 	% do we want to report errors.
 			error_context,	% context of the current item.
-			unit	% junk slot
+			need_qualifier	% must uses of the current item be 
+				% explicitly module qualified.
 	).
 		
-
 	% Pass over the item list collecting all defined type, mode and
 	% inst ids and the names of all modules imported in the interface.
 :- pred collect_mq_info(item_list::in, mq_info::in, mq_info::out) is det.
@@ -134,25 +138,23 @@ collect_mq_info_2(func_mode(_,_,_,_,_,_), Info, Info).
 collect_mq_info_2(pragma(_), Info, Info).
 collect_mq_info_2(nothing, Info, Info).
 
-
-
 % Predicates to add the type, inst and mode ids visible
 % in this module to the mq_info.
 
 :- pred add_type_defn(type_defn::in, mq_info::in, mq_info::out) is det.
 
 add_type_defn(TypeDefn, Info0, Info) :-
-	(	TypeDefn = du_type(SymName, Params, _)
+	(	TypeDefn = du_type(SymName, Params, _, _EqualityPred)
 	;	TypeDefn = uu_type(SymName, Params, _)
 	;	TypeDefn = eqv_type(SymName, Params, _)
 	;	TypeDefn = abstract_type(SymName, Params)
 	),
 	list__length(Params, Arity),
 	mq_info_get_types(Info0, Types0),
-	id_set_insert(SymName - Arity, Types0, Types),
+	mq_info_get_need_qual_flag(Info0, NeedQualifier),
+	id_set_insert(NeedQualifier, SymName - Arity, Types0, Types),
 	mq_info_set_types(Info0, Types, Info).
 
-
 :- pred add_inst_defn(inst_defn::in, mq_info::in, mq_info::out) is det.
 
 add_inst_defn(InstDefn, Info0, Info) :-
@@ -161,16 +163,17 @@ add_inst_defn(InstDefn, Info0, Info) :-
 	),
 	list__length(Params, Arity),
 	mq_info_get_insts(Info0, Insts0),
-	id_set_insert(SymName - Arity, Insts0, Insts),
+	mq_info_get_need_qual_flag(Info0, NeedQualifier),
+	id_set_insert(NeedQualifier, SymName - Arity, Insts0, Insts),
 	mq_info_set_insts(Info0, Insts, Info).
 
-
 :- pred add_mode_defn(mode_defn::in, mq_info::in, mq_info::out) is det.
 
 add_mode_defn(eqv_mode(SymName, Params, _), Info0, Info) :-
 	list__length(Params, Arity),
 	mq_info_get_modes(Info0, Modes0),
-	id_set_insert(SymName - Arity, Modes0, Modes),
+	mq_info_get_need_qual_flag(Info0, NeedQualifier),
+	id_set_insert(NeedQualifier, SymName - Arity, Modes0, Modes),
 	mq_info_set_modes(Info0, Modes, Info).
 
 	% Update import status.
@@ -183,24 +186,31 @@ process_module_defn(interface, Info0, Info) :-
 process_module_defn(implementation, Info0, Info) :-
 	mq_info_set_import_status(Info0, local, Info).
 process_module_defn(imported, Info0, Info) :-
-	mq_info_set_import_status(Info0, imported, Info).
+	mq_info_set_import_status(Info0, imported, Info1),
+	mq_info_set_need_qual_flag(Info1, may_be_unqualified, Info).
+process_module_defn(used, Info0, Info) :-
+	mq_info_set_import_status(Info0, imported, Info1),
+	mq_info_set_need_qual_flag(Info1, must_be_qualified, Info).
 process_module_defn(opt_imported, Info0, Info) :-
-	mq_info_set_import_status(Info0, opt_imported, Info).
+	mq_info_set_import_status(Info0, opt_imported, Info1),
+	mq_info_set_need_qual_flag(Info1, must_be_qualified, Info).
 process_module_defn(external(_), Info, Info).
 process_module_defn(end_module(_), Info, Info).
 process_module_defn(export(_), Info, Info).
 process_module_defn(import(Imports), Info0, Info) :-
+	add_interface_imports(Imports, Info0, Info).
+process_module_defn(use(Imports), Info0, Info) :-
+	add_interface_imports(Imports, Info0, Info).
+
+:- pred add_interface_imports(sym_list::in,
+		mq_info::in, mq_info::out) is det.
+
+add_interface_imports(Imports, Info0, Info) :-
 	( Imports = module(ImportedModules) ->
-		( mq_info_get_import_status(Info0, exported) ->
-			mq_info_add_interface_modules(Info0,
-						ImportedModules, Info)
-		;
-			Info = Info0
-		)
+		mq_info_add_interface_modules(Info0, ImportedModules, Info)
 	;
 		Info = Info0
 	).
-process_module_defn(use(_), Info, Info).	% not implemented 
 
 %------------------------------------------------------------------------------
 
@@ -284,40 +294,45 @@ module_qualify_item(func_mode(A, SymName, Modes0, Mode0, C, D) - Context,
 
 module_qualify_item(pragma(Pragma0) - Context, pragma(Pragma) - Context,
 						Info0, Info, yes) -->
-	qualify_pragma(Pragma0, Pragma, Info0, Info).
+	{ mq_info_set_error_context(Info0, (pragma) - Context, Info1) },
+	qualify_pragma(Pragma0, Pragma, Info1, Info).
 module_qualify_item(nothing - Context, nothing - Context,
 						Info, Info, yes) --> [].
 
-
 :- pred update_import_status(module_defn::in, mq_info::in, mq_info::out,
 							bool::out) is det.
 
-update_import_status(opt_imported, Info0, Info, no) :-
-	mq_info_set_import_status(Info0, opt_imported, Info).
+update_import_status(opt_imported, Info, Info, no).
 update_import_status(module(_), Info, Info, yes).
 update_import_status(interface, Info0, Info, yes) :-
 	mq_info_set_import_status(Info0, exported, Info).
 update_import_status(implementation, Info0, Info, yes) :-
 	mq_info_set_import_status(Info0, local, Info).
-update_import_status(imported, Info0, Info, no) :-
-	mq_info_set_import_status(Info0, imported, Info).
+update_import_status(imported, Info, Info, no).
+update_import_status(used, Info, Info, no).
 update_import_status(external(_), Info, Info, yes).
 update_import_status(end_module(_), Info, Info, yes).
 update_import_status(export(_), Info, Info, yes).
 update_import_status(import(_), Info, Info, yes).
 update_import_status(use(_), Info, Info, yes).
 
-
 	% Qualify the constructors or other types in a type definition.	
 :- pred qualify_type_defn(type_defn::in, type_defn::out, mq_info::in,
 	mq_info::out, term__context::in, io__state::di, io__state::uo) is det.
 
-qualify_type_defn(du_type(SymName, Params, Ctors0),
-		du_type(SymName, Params, Ctors), Info0, Info, Context) -->
+qualify_type_defn(du_type(SymName, Params, Ctors0, MaybeEqualityPred0),
+		du_type(SymName, Params, Ctors, MaybeEqualityPred),
+		Info0, Info, Context) -->
 	{ list__length(Params, Arity) },
 	{ mq_info_set_error_context(Info0, type(SymName - Arity) - Context,
 								Info1) },
-	qualify_constructors(Ctors0, Ctors, Info1, Info).
+	qualify_constructors(Ctors0, Ctors, Info1, Info),
+
+	% User-defined equality pred names will be converted into
+	% predicate calls and then module-qualified after type analysis
+	% (during mode analysis).  That way they get full type overloading
+	% resolution, etc.  Thus we don't module-qualify them here.
+	{ MaybeEqualityPred = MaybeEqualityPred0 }.
 
 qualify_type_defn(uu_type(SymName, Params, Types0),
 		uu_type(SymName, Params, Types), Info0, Info, Context) -->
@@ -346,7 +361,6 @@ qualify_constructors([SymName - Args0 | Ctors0], [SymName - Args | Ctors],
 	qualify_constructor_arg_list(Args0, Args, Info0, Info1),
 	qualify_constructors(Ctors0, Ctors, Info1, Info).
 
-
 	% Qualify the inst parameters of an inst definition.
 :- pred qualify_inst_defn(inst_defn::in, inst_defn::out, mq_info::in,
 	mq_info::out, term__context::in, io__state::di, io__state::uo) is det.
@@ -372,7 +386,6 @@ qualify_mode_defn(eqv_mode(SymName, Params, Mode0),
 								Info1) },
 	qualify_mode(Mode0, Mode, Info1, Info).	
 
-
 	% Qualify a list of items of the form Type::Mode, as in a
 	% predicate declaration.
 :- pred qualify_types_and_modes(list(type_and_mode)::in,
@@ -385,7 +398,6 @@ qualify_types_and_modes([TypeAndMode0 | TypesAndModes0],
 	qualify_type_and_mode(TypeAndMode0, TypeAndMode, Info0, Info1),
 	qualify_types_and_modes(TypesAndModes0, TypesAndModes, Info1, Info).
 
-
 :- pred qualify_type_and_mode(type_and_mode::in, type_and_mode::out,
 	mq_info::in, mq_info::out, io__state::di, io__state::uo) is det.
 
@@ -397,7 +409,6 @@ qualify_type_and_mode(type_and_mode(Type0, Mode0), type_and_mode(Type, Mode),
 	qualify_type(Type0, Type, Info0, Info1),
 	qualify_mode(Mode0, Mode, Info1, Info).
 
-
 :- pred qualify_mode_list(list(mode)::in, list(mode)::out, mq_info::in,
 		mq_info::out, io__state::di, io__state::uo) is det.
 
@@ -421,7 +432,6 @@ qualify_mode(user_defined_mode(SymName0, Insts0),
 	find_unique_match(SymName0 - Arity, SymName - _, Modes,
 						mode_id, Info1, Info).
 	
-
 :- pred qualify_inst_list(list(inst)::in, list(inst)::out, mq_info::in,
 		mq_info::out, io__state::di, io__state::uo) is det.
 
@@ -430,7 +440,6 @@ qualify_inst_list([Inst0 | Insts0], [Inst | Insts], Info0, Info) -->
 	qualify_inst(Inst0, Inst, Info0, Info1),
 	qualify_inst_list(Insts0, Insts, Info1, Info).
 
-
 	% Qualify a single inst.
 :- pred qualify_inst((inst)::in, (inst)::out, mq_info::in, mq_info::out,
 			io__state::di, io__state::uo) is det.
@@ -461,7 +470,6 @@ qualify_inst(abstract_inst(Name, Args0), abstract_inst(Name, Args),
 				Info0, Info) -->
 	qualify_inst_list(Args0, Args, Info0, Info).
 
-
 	% Find the unique inst_id that matches this inst, and qualify
 	% the argument insts.
 :- pred qualify_inst_name(inst_name::in, inst_name::out, mq_info::in,
@@ -480,6 +488,8 @@ qualify_inst_name(unify_inst(_, _, _, _), _, _, _) -->
 	{ error("compiler generated inst unexpected") }.
 qualify_inst_name(ground_inst(_, _, _, _), _, _, _) -->
 	{ error("compiler generated inst unexpected") }.
+qualify_inst_name(any_inst(_, _, _, _), _, _, _) -->
+	{ error("compiler generated inst unexpected") }.
 qualify_inst_name(shared_inst(_), _, _, _) -->
 	{ error("compiler generated inst unexpected") }.
 qualify_inst_name(mostly_uniq_inst(_), _, _, _) -->
@@ -489,7 +499,6 @@ qualify_inst_name(typed_ground(_, _), _, _, _) -->
 qualify_inst_name(typed_inst(_, _), _, _, _) -->
 	{ error("compiler generated inst unexpected") }.
 
-
 	% Qualify an inst of the form bound(functor(...)).
 :- pred qualify_bound_inst_list(list(bound_inst)::in, list(bound_inst)::out,
 	mq_info::in, mq_info::out, io__state::di, io__state::uo) is det.
@@ -500,7 +509,6 @@ qualify_bound_inst_list([functor(ConsId, Insts0) | BoundInsts0],
 	qualify_inst_list(Insts0, Insts, Info0, Info1),
 	qualify_bound_inst_list(BoundInsts0, BoundInsts, Info1, Info).
 
-
 :- pred qualify_constructor_arg_list(list(constructor_arg)::in,
 	list(constructor_arg)::out, mq_info::in, mq_info::out,
 	io__state::di, io__state::uo) is det.
@@ -528,7 +536,7 @@ qualify_type(Type0, Type, Info0, Info) -->
 	{ Type0 = term__functor(F, As, _) },
 	( { is_builtin_func_type(F, As, ArgTypes0, RetType0) } ->
 		qualify_type_list(ArgTypes0, ArgTypes, Info0, Info1),
-		qualify_type(RetType0, RetType, Info1, Info),
+		qualify_type(RetType0, RetType, Info1, Info2),
 		{ term__context_init(Context) },
 		{ Type = term__functor(term__atom("="),
 				[term__functor(term__atom("func"),
@@ -545,16 +553,33 @@ qualify_type(Type0, Type, Info0, Info) -->
 			find_unique_match(TypeId0, TypeId, Types,
 						type_id, Info0, Info1)
 		),
-		qualify_type_list(Args0, Args, Info1, Info),
+		qualify_type_list(Args0, Args, Info1, Info2),
 		{ TypeId = SymName - _ },
 		{ construct_qualified_term(SymName, Args, Type) }	
 	;
 		{ mq_info_get_error_context(Info0, ErrorContext) },
 		report_invalid_type(Type0, ErrorContext),
 		{ Type = Type0 },
-		{ Info = Info0 }
-	).
-
+		{ Info2 = Info0 }
+	),
+	%
+	% The types `int', `float', and `string' are builtin types,
+	% defined by the compiler, but arguably they ought to be
+	% defined in int.m, float.m, and string.m, and so if someone
+	% uses the type `int' in the interface, then we don't want
+	% to warn about `import_module int' in the interface.
+	%
+	{
+		Type = term__functor(term__atom(Typename), [], _),
+		( Typename = "int"
+		; Typename = "string"
+		; Typename = "float"
+		)
+	->
+		mq_info_set_module_used(Info2, Typename, Info)
+	;
+		Info = Info2
+	}.
 
 	% Qualify the modes in a pragma(c_code, ...) decl.
 :- pred qualify_pragma((pragma_type)::in, (pragma_type)::out,
@@ -567,14 +592,31 @@ qualify_pragma(c_code(Rec, SymName, PredOrFunc, PragmaVars0, Varset, CCode),
 	c_code(Rec, SymName, PredOrFunc, PragmaVars, Varset, CCode), 
 		Info0, Info) -->
 	qualify_pragma_vars(PragmaVars0, PragmaVars, Info0, Info).
+qualify_pragma(c_code(Rec, SymName, PredOrFunc, PragmaVars0,
+		SavedVars, LabelCount, Varset, CCode),
+	c_code(Rec, SymName, PredOrFunc, PragmaVars,
+		SavedVars, LabelCount, Varset, CCode), Info0, Info) -->
+	qualify_pragma_vars(PragmaVars0, PragmaVars, Info0, Info).
 qualify_pragma(memo(A, B), memo(A, B), Info, Info) --> [].
 qualify_pragma(inline(A, B), inline(A, B), Info, Info) --> [].
+qualify_pragma(no_inline(A, B), no_inline(A, B), Info, Info) --> [].
 qualify_pragma(obsolete(A, B), obsolete(A, B), Info, Info) --> [].
-qualify_pragma(export(Name, Modes0, CFunc), export(Name, Modes, CFunc),
-				Info0, Info) -->
+qualify_pragma(export(Name, PredOrFunc, Modes0, CFunc),
+		export(Name, PredOrFunc, Modes, CFunc), Info0, Info) -->
 	qualify_mode_list(Modes0, Modes, Info0, Info).
+qualify_pragma(unused_args(A, B, C, D, E), unused_args(A, B, C, D, E),
+				Info, Info) --> [].
 qualify_pragma(fact_table(SymName, Arity, FileName),
 	fact_table(SymName, Arity, FileName), Info, Info) --> [].
+qualify_pragma(termination_info(PredOrFunc, SymName, ModeList0, Termination), 
+		termination_info(PredOrFunc, SymName, ModeList, Termination), 
+		Info0, Info) --> 
+	qualify_mode_list(ModeList0, ModeList, Info0, Info).
+qualify_pragma(terminates(A, B), terminates(A, B), Info, Info) --> [].
+qualify_pragma(does_not_terminate(A, B), does_not_terminate(A, B), 
+		Info, Info) --> [].
+qualify_pragma(check_termination(A, B), check_termination(A, B), Info, 
+		Info) --> [].
 
 :- pred qualify_pragma_vars(list(pragma_var)::in, list(pragma_var)::out,
 		mq_info::in, mq_info::out, io__state::di, io__state::uo) is det.
@@ -585,8 +627,6 @@ qualify_pragma_vars([pragma_var(Var, Name, Mode0) | PragmaVars0],
 	qualify_mode(Mode0, Mode, Info0, Info1),
 	qualify_pragma_vars(PragmaVars0, PragmaVars, Info1, Info).
 
-
-
 	% Find the unique match in the current name space for a given id
 	% from a list of ids. If none exists, either because no match was
 	% found or mulitiple matches were found, report an error.
@@ -600,11 +640,7 @@ find_unique_match(Id0, Id, Ids, TypeOfId, Info0, Info) -->
 		{ Id0 = qualified(Module, Name) - Arity },
 		{ Id = Id0 },
 		( { id_set_search_m_n_a(Ids, Module, Name, Arity) } ->
-			( { mq_info_get_import_status(Info0, exported) } ->
-				{ mq_info_set_module_used(Info0, Module, Info) }
-			;
-				{ Info = Info0 }
-			)
+			{ mq_info_set_module_used(Info0, Module, Info) }
 		;
 			( { mq_info_get_report_error_flag(Info0, yes) } ->
 				{ mq_info_get_error_context(Info0,
@@ -641,12 +677,7 @@ find_unique_match(Id0, Id, Ids, TypeOfId, Info0, Info) -->
 		; { Modules = [Module] } ->
 			% A unique match for this ID.
 			{ Id = qualified(Module, IdName) - Arity },
-			( { mq_info_get_import_status(Info0, exported) } ->
-				{ mq_info_set_module_used(Info0, Module,
-								Info) }
-			;
-				{ Info = Info0 }
-			)
+			{ mq_info_set_module_used(Info0, Module, Info) }
 		;
 			% There are multiple matches.
 			{ Id = Id0 },
@@ -683,6 +714,7 @@ find_unique_match(Id0, Id, Ids, TypeOfId, Info0, Info) -->
 	;	func(id)
 	; 	pred_mode(id)
 	;	func_mode(id)
+	;	(pragma)
 	;	lambda_expr
 	;	type_qual.
 
@@ -717,17 +749,17 @@ report_multiply_defined(ErrorContext - Context, Id, IdType, Modules) -->
 	write_error_context2(ErrorContext),
 	io__write_string("\n"),
 	prog_out__write_context(Context),
-	io__write_string("  error: multiple possible matches for "),
+	io__write_string("  ambiguity error: multiple possible matches for "),
 	{ id_type_to_string(IdType, IdStr) },
 	io__write_string(IdStr),
 	io__write_string(" "),
 	write_id(Id),
-	io__write_string("\n"),
+	io__write_string(".\n"),
 	prog_out__write_context(Context),
-	io__write_string("  The possible matches are in modules:\n"),
+	io__write_string("  The possible matches are in modules\n"),
 	prog_out__write_context(Context),
 	io__write_string("  "),
-	write_module_list(Modules),
+	prog_out__write_module_list(Modules),
 	io__write_string(".\n"),
 	globals__io_lookup_bool_option(verbose_errors, Verbose),
 	( { Verbose = yes } ->
@@ -764,6 +796,8 @@ write_error_context2(func_mode(Id)) -->
 	write_id(Id).
 write_error_context2(lambda_expr) -->
 	io__write_string("mode declaration for lambda expression").
+write_error_context2(pragma) -->
+	io__write_string("pragma").
 write_error_context2(type_qual) -->
 	io__write_string("explicit type qualification").
 
@@ -782,7 +816,6 @@ write_id(SymName - Arity) -->
 	io__write_string("'/"),
 	io__write_int(Arity).
 
-
 	% Warn about modules imported in the interface when they do not
 	% need to be.
 :- pred maybe_warn_unused_interface_imports(module_name::in,
@@ -806,7 +839,7 @@ maybe_warn_unused_interface_imports(ModuleName, UnusedImports) -->
 		;
 			io__write_string("modules ")
 		),
-		write_module_list(UnusedImports),
+		prog_out__write_module_list(UnusedImports),
 		io__write_string("\n"),
 		prog_out__write_context(Context),
 		{ is_or_are(UnusedImports, IsOrAre) },
@@ -825,22 +858,6 @@ is_or_are([], "") :- error("module_qual:is_or_are").
 is_or_are([_], "is").
 is_or_are([_, _ | _], "are").
 
-
-:- pred write_module_list(list(module_name)::in, io__state::di,
-					io__state::uo) is det.
-
-write_module_list([Import1, Import2, Import3 | Imports]) --> 
-	io__write_string(Import1),
-	io__write_string(".m, "),
-	write_module_list([Import2, Import3 | Imports]).
-write_module_list([Import1, Import2]) -->
-	io__write_strings([Import1, ".m and ", Import2 ,".m"]).
-write_module_list([Import]) -->
-	io__write_string(Import).
-write_module_list([]) -->
-	{ error("module_qual:write_module_list") }.
-
-
 	% Output an error message about an ill-formed type.
 :- pred report_invalid_type(term, error_context, io__state, io__state).
 :- mode report_invalid_type(in, in, di, uo) is det.
@@ -854,7 +871,7 @@ report_invalid_type(Type, ErrorContext - Context) -->
 	prog_out__write_context(Context),
 	io__write_string("  error: ill-formed type `"),
 	{ varset__init(VarSet) },
-	mercury_output_term(Type, VarSet),
+	mercury_output_term(Type, VarSet, no),
 	io__write_string("'.\n").
 
 %-----------------------------------------------------------------------------%
@@ -899,27 +916,9 @@ init_mq_info(ReportErrors, Info0) :-
 	term__context_init(Context),
 	ErrorContext = type(unqualified("") - 0) - Context,
 	set__init(InterfaceModules0),
-	id_set_init(Types0),
-	id_set_init(Insts0),
-	id_set_init(Modes0),
-	Info0 = mq_info(Types0, Insts0, Modes0, InterfaceModules0,
-			local, 0, no, no, ReportErrors, ErrorContext, unit).
-
-init_mq_info_module(ModuleInfo, Info0) :-
-	module_info_typeids(ModuleInfo, TypeIds),
-	id_set_init(Types0),
-	list__foldl(id_set_insert, TypeIds, Types0, Types),
-	module_info_instids(ModuleInfo, InstIds),
-	id_set_init(Insts0),
-	list__foldl(id_set_insert, InstIds, Insts0, Insts),
-	module_info_modeids(ModuleInfo, ModeIds),
-	id_set_init(Modes0),
-	list__foldl(id_set_insert, ModeIds, Modes0, Modes),
-	term__context_init(Context),
-	ErrorContext = type(unqualified("") - 0) - Context,
-	set__init(InterfaceModules0),
-	Info0 = mq_info(Types, Insts, Modes, InterfaceModules0,
-		local, 0, no, no, yes, ErrorContext, unit).
+	id_set_init(Empty),
+	Info0 = mq_info(Empty, Empty, Empty, InterfaceModules0, local, 0,
+		no, no, ReportErrors, ErrorContext, may_be_unqualified).
 
 :- pred mq_info_get_types(mq_info::in, type_id_set::out) is det.
 :- pred mq_info_get_insts(mq_info::in, inst_id_set::out) is det.
@@ -932,7 +931,6 @@ init_mq_info_module(ModuleInfo, Info0) :-
 % :- pred mq_info_get_mode_error_flag(mq_info::in, bool::out) is det.
 :- pred mq_info_get_report_error_flag(mq_info::in, bool::out) is det.
 :- pred mq_info_get_error_context(mq_info::in, error_context::out) is det.
-:- pred mq_info_get_junk(mq_info::in, unit::out) is det.
 
 mq_info_get_types(mq_info(Types, _,_,_,_,_,_,_,_,_,_), Types).
 mq_info_get_insts(mq_info(_, Insts, _,_,_,_,_,_,_,_,_), Insts).
@@ -945,18 +943,19 @@ mq_info_get_mode_error_flag(mq_info(_,_,_,_,_,_,_, ModeError, _,_,_),
 						ModeError).
 mq_info_get_report_error_flag(mq_info(_,_,_,_,_,_,_,_, Report,_,_), Report).
 mq_info_get_error_context(mq_info(_,_,_,_,_,_,_,_,_, Context,_), Context).
-mq_info_get_junk(mq_info(_,_,_,_,_,_,_,_,_,_,Junk), Junk).
+mq_info_get_need_qual_flag(mq_info(_,_,_,_,_,_,_,_,_,_,UseModule), UseModule).
 
 :- pred mq_info_set_types(mq_info::in, type_id_set::in, mq_info::out) is det.
 :- pred mq_info_set_insts(mq_info::in, inst_id_set::in, mq_info::out) is det.
 :- pred mq_info_set_modes(mq_info::in, mode_id_set::in, mq_info::out) is det.
+:- pred mq_info_set_interface_modules(mq_info::in, set(module_name)::in,
+						mq_info::out) is det.
 :- pred mq_info_set_import_status(mq_info::in, import_status::in,
 						mq_info::out) is det.
 :- pred mq_info_set_type_error_flag(mq_info::in, mq_info::out) is det.
 :- pred mq_info_set_mode_error_flag(mq_info::in, mq_info::out) is det.
 :- pred mq_info_set_error_context(mq_info::in, error_context::in,
 						mq_info::out) is det.
-:- pred mq_info_set_junk(mq_info::in, unit::in, mq_info::out) is det.
 
 mq_info_set_types(mq_info(_, B,C,D,E,F,G,H,I,J,K), Types,
 		mq_info(Types, B,C,D,E,F,G,H,I,J,K)).
@@ -964,6 +963,8 @@ mq_info_set_insts(mq_info(A,_,C,D,E,F,G,H,I,J,K), Insts,
 		mq_info(A, Insts, C,D,E,F,G,H,I,J,K)).
 mq_info_set_modes(mq_info(A,B,_,D,E,F,G,H,I,J,K), Modes,
 		mq_info(A,B, Modes, D,E,F,G,H,I,J,K)).
+mq_info_set_interface_modules(mq_info(A,B,C,_,E,F,G,H,I,J,K), Modules,
+		mq_info(A,B,C, Modules, E,F,G,H,I,J,K)).
 mq_info_set_import_status(mq_info(A,B,C,D,_,F,G,H,I,J,K), Status,
 		mq_info(A,B,C,D, Status, F,G,H,I,J,K)).
 mq_info_set_type_error_flag(mq_info(A,B,C,D,E,F, _, H,I,J,K),
@@ -972,8 +973,8 @@ mq_info_set_mode_error_flag(mq_info(A,B,C,D,E,F,G,_,I,J,K),
 		mq_info(A,B,C,D,E,F,G, yes, I,J,K)).
 mq_info_set_error_context(mq_info(A,B,C,D,E,F,G,H,I,_,K), Context,
 		mq_info(A,B,C,D,E,F,G,H,I, Context,K)).
-mq_info_set_junk(mq_info(A,B,C,D,E,F,G,H,I,J,_), Junk,
-		mq_info(A,B,C,D,E,F,G,H,I,J, Junk)).
+mq_info_set_need_qual_flag(mq_info(A,B,C,D,E,F,G,H,I,J,_), Flag,
+		mq_info(A,B,C,D,E,F,G,H,I,J, Flag)).
 
 :- pred mq_info_incr_errors(mq_info::in, mq_info::out) is det.
 
@@ -981,7 +982,6 @@ mq_info_incr_errors(mq_info(A,B,C,D,E, NumErrors0, G,H,I,J,K),
 		mq_info(A,B,C,D,E, NumErrors, G,H,I,J,K)) :-
 	NumErrors is NumErrors0 + 1.
 
-
 :- pred mq_info_set_error_flag(mq_info::in, id_type::in, mq_info::out) is det.
 
 mq_info_set_error_flag(Info0, type_id, Info) :-
@@ -991,25 +991,44 @@ mq_info_set_error_flag(Info0, mode_id, Info) :-
 mq_info_set_error_flag(Info0, inst_id, Info) :-
 	mq_info_set_mode_error_flag(Info0, Info).
 
-
+	% If the current item is in the interface, remove its module 
+	% name from the list of modules not used in the interface.
 :- pred mq_info_set_module_used(mq_info::in, module_name::in,
 						mq_info::out) is det.
+
+mq_info_set_module_used(Info0, Module, Info) :-
+	( mq_info_get_import_status(Info0, exported) ->
+		mq_info_get_interface_modules(Info0, Modules0),
+		set__delete(Modules0, Module, Modules),
+		mq_info_set_interface_modules(Info0, Modules, Info)
+	;
+		Info = Info0
+	).
+
+	% Add to the list of modules imported in the interface.
 :- pred mq_info_add_interface_modules(mq_info::in, list(module_name)::in,
 						mq_info::out) is det.
 
-mq_info_set_module_used(mq_info(A,B,C,Modules0,E,F,G,H,I,J,K), Module,
-		mq_info(A,B,C, Modules, E,F,G,H,I,J,K)) :-
-	set__delete(Modules0, Module, Modules).
-mq_info_add_interface_modules(mq_info(A,B,C,Modules0, E,F,G,H,I,J,K),
-		NewModules, mq_info(A,B,C, Modules, E,F,G,H,I,J,K)) :-
-	set__insert_list(Modules0, NewModules, Modules).
+mq_info_add_interface_modules(Info0, NewModules, Info) :-
+	( mq_info_get_import_status(Info0, exported) ->
+		mq_info_get_interface_modules(Info0, Modules0),
+		set__insert_list(Modules0, NewModules, Modules),
+		mq_info_set_interface_modules(Info0, Modules, Info)
+	;
+		Info = Info0
+	).
 
 %----------------------------------------------------------------------------%
 % Define a type for representing sets of ids during module qualification
 % to allow efficient retrieval of all the modules which define an id
 % with a certain name and arity.
 
-:- type id_set == map(pair(string, arity), set(module_name)).
+% The first set of module_names can be used without module qualifiers,
+% items from the second set can only be used with module qualifiers.
+% Items from modules imported with a :- use_module declaration and from `.opt'
+% files should go into the second set.
+:- type id_set == map(pair(string, arity), pair(set(module_name))).
+
 :- type type_id_set == id_set.
 :- type mode_id_set == id_set.
 :- type inst_id_set == id_set.
@@ -1021,25 +1040,36 @@ id_set_init(IdSet) :-
 
 	% Insert an id into an id_set, aborting with an error if the
 	% id is not module qualified.
-:- pred id_set_insert(id::in, id_set::in, id_set::out) is det.
+:- pred id_set_insert(need_qualifier::in, id::in, 
+		id_set::in, id_set::out) is det.
 
-id_set_insert(unqualified(_) - _, _, _) :-
+id_set_insert(_, unqualified(_) - _, _, _) :-
 	error("module_qual:id_set_insert - unqualified id").
-id_set_insert(qualified(Module, Name) - Arity, IdSet0, IdSet) :-
-	( map__search(IdSet0, Name - Arity, Modules0) ->
-		Modules1 = Modules0 
+id_set_insert(NeedQualifier, qualified(Module, Name) - Arity, IdSet0, IdSet) :-
+	( map__search(IdSet0, Name - Arity, ImportModules0 - UseModules0) ->
+		ImportModules1 = ImportModules0,
+		UseModules1 = UseModules0
 	;
-		set__init(Modules1)
+		set__init(ImportModules1),
+		set__init(UseModules1)
 	),
-	set__insert(Modules1, Module, Modules),
-	map__set(IdSet0, Name - Arity, Modules, IdSet).
+	(
+		NeedQualifier = must_be_qualified,
+		set__insert(UseModules1, Module, UseModules),
+		ImportModules = ImportModules1
+	;
+		NeedQualifier = may_be_unqualified,
+		set__insert(ImportModules1, Module, ImportModules),
+		UseModules = UseModules1
+	),
+	map__set(IdSet0, Name - Arity, ImportModules - UseModules, IdSet).
 
 :- pred id_set_search_name_arity(id_set::in, string::in, int::in,
 				list(module_name)::out) is det.
 
 id_set_search_name_arity(IdSet0, Name, Arity, Modules) :-
-	( map__search(IdSet0, Name - Arity, ModuleSet) ->
-		set__to_sorted_list(ModuleSet, Modules)
+	( map__search(IdSet0, Name - Arity, ImportModules - _) ->
+		set__to_sorted_list(ImportModules, Modules)
 	;
 		Modules = []
 	).
@@ -1048,7 +1078,11 @@ id_set_search_name_arity(IdSet0, Name, Arity, Modules) :-
 			 	string::in, int::in) is semidet.
 
 id_set_search_m_n_a(IdSet0, Module, Name, Arity) :-
-	map__search(IdSet0, Name - Arity, ModuleSet),
-	set__member(Module, ModuleSet).
+	map__search(IdSet0, Name - Arity, ImportModules - UseModules),
+	( 
+		set__member(Module, ImportModules)
+	;
+		set__member(Module, UseModules)
+	).
 
 %----------------------------------------------------------------------------%
