@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1995 University of Melbourne.
+% Copyright (C) 1996-1997 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -52,7 +52,6 @@
 		module_error, io__state, io__state).
 :- mode read_mod_ignore_errors(in, in, in, in, out, out, di, uo) is det.
 
-
 	% make_interface(ModuleName, Items):
 	%	Given a module name and the list of items in that module,
 	%	output the long (`.int') and short (`.int2') interface files
@@ -101,19 +100,21 @@
 	% generate_dependencies(ModuleName):
 	%	Generate the per-program makefile dependencies (`.dep') file
 	%	for a program whose top-level module is `ModuleName'.
-	%	This involes first transitively reading in all imported
+	%	This involves first transitively reading in all imported
 	%	modules.  While we're at it, we also save the per-module
 	%	makefile dependency (`.d') files for all those modules.
 	%
 :- pred generate_dependencies(string, io__state, io__state).
 :- mode generate_dependencies(in, di, uo) is det.
 
-	% Given a module (well, a list of items),
-	% determine all the modules that it depends upon
-	% (both interface dependencies and also implementation dependencies).
-
-:- pred get_dependencies(item_list, list(string)).
-:- mode get_dependencies(in, out) is det.
+	% get_dependencies(Items, ImportDeps, UseDeps).
+	%	Get the list of modules that a list of items depends on.
+	%	ImportDeps is the list of modules imported using
+	% 	`:- import_module', UseDeps is the list of modules imported
+	%	using `:- use_module'.
+	%
+:- pred get_dependencies(item_list, list(string), list(string)).
+:- mode get_dependencies(in, out, out) is det.
 
 	% Do the importing of the interface files of imported modules.
 
@@ -144,9 +145,9 @@
 
 :- implementation.
 :- import_module passes_aux, prog_out, mercury_to_mercury.
-:- import_module globals, options, intermod, module_qual.
+:- import_module prog_io_util, globals, options, intermod, module_qual.
 :- import_module bool, string, set, map, term, varset, dir, std_util, library.
-
+:- import_module assoc_list, relation.
 
 	% Read in the .int3 files that the current module depends on,
 	% and use these to qualify all items in the interface as much as
@@ -156,30 +157,42 @@ make_interface(ModuleName, Items0) -->
 	{ get_interface(Items0, yes, InterfaceItems0) },
 	{ term__context_init(Context) },
 	{ varset__init(Varset) },
-	{ get_dependencies(InterfaceItems0, InterfaceDeps) },
+	{ get_dependencies(InterfaceItems0, 
+		InterfaceImportDeps, InterfaceUseDeps) },
 	{ list__append(InterfaceItems0,
 			[module_defn(Varset, imported) - Context],
-			InterfaceItems0a) },
-	{ Module0 = module_imports(ModuleName, [], [], InterfaceItems0a, no) },
+			InterfaceItems1) },
+	{ Module1 = module_imports(ModuleName, [], [], InterfaceItems1, no) },
 		% Get the .int3s that the current .int depends on.
-	process_module_short_interfaces(["mercury_builtin" | InterfaceDeps],
-					".int3", Module0, Module),
-	{ Module = module_imports(_, _, _, InterfaceItems1, Error) },
+	process_module_short_interfaces(
+			["mercury_builtin" | InterfaceImportDeps],
+			".int3", Module1, Module2),
+	{ Module2 = module_imports(_, Direct2, Indirect2,
+			InterfaceItems2, Error2) },
+	{ list__append(InterfaceItems2,
+			[module_defn(Varset, used) - Context],
+			InterfaceItems3) },
+	{ Module3 = module_imports(ModuleName, Direct2, Indirect2,
+			InterfaceItems3, Error2) },
+	process_module_short_interfaces(InterfaceUseDeps,
+			".int3", Module3, Module4),
+
+	{ Module4 = module_imports(_, _, _, InterfaceItems4, Error) },
 	( { Error = yes } ->
 		io__write_strings(["Error reading short interface files.\n",
 				ModuleName, ".int and ",
 				ModuleName, ".int2 not written.\n"])
 	;
 			% Qualify all items.
-		module_qual__module_qualify_items(InterfaceItems1,
-				InterfaceItems2, ModuleName, yes, _, _, _),
+		module_qual__module_qualify_items(InterfaceItems4,
+				InterfaceItems5, ModuleName, yes, _, _, _, _),
 		io__get_exit_status(Status),
 		( { Status \= 0 } ->
 			io__write_strings([ModuleName, ".int not written.\n"])
 		;
-			{ strip_imported_items(InterfaceItems2, [],
-							InterfaceItems3) },
-			check_for_clauses_in_interface(InterfaceItems3,
+			{ strip_imported_items(InterfaceItems5, [],
+							InterfaceItems6) },
+			check_for_clauses_in_interface(InterfaceItems6,
 							InterfaceItems),
 			check_for_no_exports(InterfaceItems, ModuleName),
 			write_interface_file(ModuleName, ".int",
@@ -199,7 +212,7 @@ make_short_interface(ModuleName, Items0) -->
 	check_for_clauses_in_interface(InterfaceItems0, InterfaceItems),
 	{ get_short_interface(InterfaceItems, ShortInterfaceItems0) },
 	module_qual__module_qualify_items(ShortInterfaceItems0,
-			ShortInterfaceItems, ModuleName, no, _, _, _),
+			ShortInterfaceItems, ModuleName, no, _, _, _, _),
 	write_interface_file(ModuleName, ".int3", ShortInterfaceItems),
 	touch_interface_datestamp(ModuleName, ".date3").
 
@@ -210,9 +223,9 @@ make_short_interface(ModuleName, Items0) -->
 strip_imported_items([], Items0, Items) :-
 	list__reverse(Items0, Items). 
 strip_imported_items([Item - Context | Rest], Items0, Items) :-
-	(
-		Item = module_defn(_, imported)
-	->
+	( Item = module_defn(_, imported) ->
+		list__reverse(Items0, Items)
+	; Item = module_defn(_, used) ->
 		list__reverse(Items0, Items)
 	;
 		strip_imported_items(Rest, [Item - Context | Items0], Items)
@@ -224,17 +237,34 @@ strip_imported_items([Item - Context | Rest], Items0, Items) :-
 :- mode check_for_clauses_in_interface(in, out, di, uo) is det.
 
 check_for_clauses_in_interface([], []) --> [].
-check_for_clauses_in_interface([Item0 | Items0], Items) -->
+check_for_clauses_in_interface([ItemAndContext0 | Items0], Items) -->
+	{ ItemAndContext0 = Item0 - Context },
 	(
-		( { Item0 = pred_clause(_,_,_,_) - Context }
-		; { Item0 = func_clause(_,_,_,_,_) - Context }
+		( { Item0 = pred_clause(_,_,_,_) }
+		; { Item0 = func_clause(_,_,_,_,_) }
 		)
 	->
 		prog_out__write_context(Context),
-		io__write_string("Warning: clause in module interface.\n"),
+		report_warning("Warning: clause in module interface.\n"),
 		check_for_clauses_in_interface(Items0, Items)
 	;
-		{ Items = [Item0 | Items1] },
+		% pragma `obsolete', `terminates', `does_not_terminate' 
+		% `termination_info' and `check_termination' declarations
+		% are supposed to go in the interface, but all other pragma
+		% declarations should go in the implementation.
+		{ Item0 = pragma(Pragma) },
+		{ Pragma \= obsolete(_, _) },
+		{ Pragma \= terminates(_, _) },
+		{ Pragma \= does_not_terminate(_, _) },
+		{ Pragma \= check_termination(_, _) },
+		{ Pragma \= termination_info(_, _, _, _) }
+	->
+		prog_out__write_context(Context),
+		report_warning("Warning: pragma in module interface.\n"),
+		check_for_clauses_in_interface(Items0, Items)
+
+	;
+		{ Items = [ItemAndContext0 | Items1] },
 		check_for_clauses_in_interface(Items0, Items1)
 	).
 
@@ -353,20 +383,72 @@ touch_interface_datestamp(ModuleName, Ext) -->
 %-----------------------------------------------------------------------------%
 
 grab_imported_modules(ModuleName, Items0, Module, FactDeps, Error) -->
-	{ get_dependencies(Items0, ImportedModules) },
+	{ get_dependencies(Items0, ImportedModules, UsedModules0) },
+
+	{ set__list_to_set(ImportedModules, ImportedSet) },
+	{ set__list_to_set(UsedModules0, UsedSet) },
+
+	{ set__intersect(ImportedSet, UsedSet, BothSet) },
+
+	% Report errors for modules imported using both :- use_module
+	% and :- import_module. Remove the import_module declaration.
+	{ string__append(ModuleName, ".m", FileName) },
+	{ term__context_init(FileName, 0, Context) },
+	( { set__empty(BothSet) } ->
+		{ UsedModules = UsedModules0 }
+	;
+		prog_out__write_context(Context),
+		io__write_string("Warning:"),
+		{ set__to_sorted_list(BothSet, BothList) },
+		( { BothList = [_] } ->
+			io__write_string(" module "),
+			prog_out__write_module_list(BothList),
+			io__write_string(" is ")
+		;
+			io__write_string(" modules "),
+			prog_out__write_module_list(BothList),
+			io__write_string(" are ")
+		),
+		io__write_string("imported using both\n"),
+		prog_out__write_context(Context),
+		io__write_string("  `:- import_module' and `:- use_module' declarations.\n"),
+
+		% Treat the modules with both types of import as if they 
+		% were imported using :- import_module.
+		{ list__delete_elems(UsedModules0, BothList,
+			UsedModules) },
+		globals__io_lookup_bool_option(halt_at_warn, Halt),
+		( { Halt = yes } ->
+			io__set_exit_status(1)
+		;
+			[]
+		)
+	),
+
 	{ get_fact_table_dependencies(Items0, FactDeps) },
 
-		% we add a pseudo-declaration `:- imported' at the end
-		% of the item list, so that make_hlds knows which items
-		% are imported and which are defined in the main module
+		% we add a pseudo-declarations `:- imported' at the end
+		% of the item list. Uses of the items with declarations 
+		% following this do not need module qualifiers.
 	{ varset__init(VarSet) },
-	{ term__context_init(ModuleName, 0, Context) },
 	{ list__append(Items0,
 		[module_defn(VarSet, imported) - Context], Items1) },
 	{ dir__basename(ModuleName, BaseModuleName) },
-	{ Module0 = module_imports(BaseModuleName, [], [], Items1, no) },
+	{ Module1 = module_imports(BaseModuleName, [], [], Items1, no) },
+
 	process_module_interfaces(["mercury_builtin" | ImportedModules], 
-		[], Module0, Module),
+		[], Module1, Module2),
+	{ Module2 = module_imports(_, Direct2, Indirect2, Items2, Error2) },
+
+		% we add a pseudo-declarations `:- used' at the end
+		% of the item list. Uses of the items with declarations 
+		% following this must be module qualified.
+	{ list__append(Items2,
+		[module_defn(VarSet, used) - Context], Items3) },
+	{ Module3 = module_imports(BaseModuleName, Direct2, Indirect2, 
+		Items3, Error2) },
+	process_module_interfaces(UsedModules, [], Module3, Module),
+
 	{ Module = module_imports(_, _, _, _, Error) }.
 
 %-----------------------------------------------------------------------------%
@@ -387,6 +469,39 @@ write_dependency_file(ModuleName, LongDeps0, ShortDeps0, FactDeps0) -->
 		{ list__delete_all(ShortDeps2, ModuleName, ShortDeps) },
 		{ list__sort_and_remove_dups(FactDeps0, FactDeps) },
 
+		( { FactDeps \= [] } ->
+			io__write_strings(DepStream, 
+				[ModuleName, ".fact_tables = "]),
+			write_dependencies_list(FactDeps, "", DepStream),
+			io__nl(DepStream),
+			globals__io_lookup_bool_option(assume_gmake,
+				AssumeGmake),
+			( { AssumeGmake = no} ->
+				io__write_strings(DepStream,
+					[ModuleName, ".fact_tables.os = "]),
+				write_dependencies_list(FactDeps, ".o",
+					DepStream),
+				io__write_strings(DepStream, [
+					"\n\n", 
+					ModuleName, ".fact_tables.cs = "]),
+				write_dependencies_list(FactDeps, ".c",
+					DepStream),
+				io__nl(DepStream)
+			;
+				io__write_strings(DepStream, [
+					"\n\n", ModuleName,
+					".fact_tables.os = $(", ModuleName,
+					".fact_tables:%=%.o)\n\n",
+					ModuleName,
+					".fact_tables.cs = $(", ModuleName,
+					".fact_tables:%=%.c)\n\n"
+				])
+			)
+		;
+			[]
+		),
+
+
 		io__write_strings(DepStream, [
 			ModuleName, ".optdate ",
 			ModuleName, ".c ",
@@ -396,7 +511,19 @@ write_dependency_file(ModuleName, LongDeps0, ShortDeps0, FactDeps0) -->
 		] ),
 		write_dependencies_list(LongDeps, ".int", DepStream),
 		write_dependencies_list(ShortDeps, ".int2", DepStream),
-		write_dependencies_list(FactDeps, "", DepStream),
+
+		( { FactDeps \= [] } ->
+			io__write_strings(DepStream, [
+				" \\\n\t$(", ModuleName, ".fact_tables)\n\n",
+				"$(", ModuleName, ".fact_tables.os) : $(",
+				ModuleName, ".fact_tables) ",
+				ModuleName, ".m\n\n",
+				"$(", ModuleName, ".fact_tables.cs) : ",
+				ModuleName, ".o\n"
+			] )
+		;
+			[]
+		),
 
 		globals__io_lookup_bool_option(intermodule_optimization,
 							Intermod),
@@ -414,7 +541,9 @@ write_dependency_file(ModuleName, LongDeps0, ShortDeps0, FactDeps0) -->
 			% is to make sure the module gets type-checked without
 			% having the definitions of abstract types from other
 			% modules.
-			get_curr_dir_deps(LongDeps, [], OptDeps),
+			globals__io_lookup_accumulating_option(
+				intermod_directories, IntermodDirs),
+			get_opt_deps(LongDeps, IntermodDirs, [], OptDeps),
 			write_dependencies_list([ModuleName | OptDeps],
 				".opt", DepStream)
 		;
@@ -444,21 +573,29 @@ write_dependency_file(ModuleName, LongDeps0, ShortDeps0, FactDeps0) -->
 		report_error(Message)
 	).
 
-	% Filter out the dependencies that are in the current directory.
-:- pred get_curr_dir_deps(list(string)::in, list(string)::in, 
+	% For each dependency, search intermod_directories for a .opt
+	% file or a .m file, filtering out those for which the search fails.
+:- pred get_opt_deps(list(string)::in, list(string)::in, list(string)::in, 
 		list(string)::out, io__state::di, io__state::uo) is det.
 
-get_curr_dir_deps([], CurrDirDeps, CurrDirDeps) --> [].
-get_curr_dir_deps([Dep | Deps], CurrDirDeps0, CurrDirDeps) -->
+get_opt_deps([], _, OptDeps, OptDeps) --> [].
+get_opt_deps([Dep | Deps], IntermodDirs, OptDeps0, OptDeps) -->
 	{ string__append(Dep, ".m", DepName) },
-	io__see(DepName, Res),
-	( { Res = ok } ->
-		{ CurrDirDeps1 = [Dep | CurrDirDeps0] },
+	search_for_file(IntermodDirs, DepName, Result1),
+	( { Result1 = yes } ->
+		{ OptDeps1 = [Dep | OptDeps0] },
 		io__seen
 	;
-		{ CurrDirDeps1 = CurrDirDeps0 }
+		{ string__append(Dep, ".opt", OptName) },
+		search_for_file(IntermodDirs, OptName, Result2),
+		( { Result2 = yes } ->
+			{ OptDeps1 = [Dep | OptDeps0] },
+			io__seen
+		;
+			{ OptDeps1 = OptDeps0 }
+		)
 	),
-	get_curr_dir_deps(Deps, CurrDirDeps1, CurrDirDeps).
+	get_opt_deps(Deps, IntermodDirs, OptDeps1, OptDeps).
 
 %-----------------------------------------------------------------------------%
 
@@ -473,6 +610,7 @@ generate_dependencies(Module) -->
 	% check whether we couldn't read the main `.m' file
 	%
 	{ map__lookup(DepsMap, Module, deps(_, Error, _, _, _)) },
+	globals__io_lookup_bool_option(verbose, Verbose),
 	( { Error = fatal } ->
 	    { string__append_list(["fatal error reading module `",
 				Module, "'."], Message) },
@@ -482,22 +620,54 @@ generate_dependencies(Module) -->
 	    % now, write the `.dep' file
 	    %
 	    { string__append(Module, ".dep", DepFileName) },
-	    globals__io_lookup_bool_option(verbose, Verbose),
 	    maybe_write_string(Verbose, "% Creating auto-dependency file `"),
 	    maybe_write_string(Verbose, DepFileName),
 	    maybe_write_string(Verbose, "'...\n"),
-	    io__open_output(DepFileName, Result),
-	    ( { Result = ok(DepStream) } ->
+	    io__open_output(DepFileName, DepResult),
+	    ( { DepResult = ok(DepStream) } ->
 		generate_dep_file(Module, DepsMap, DepStream),
 		io__close_output(DepStream),
 		maybe_write_string(Verbose, "% done\n")
 	    ;
 		{ string__append_list(["can't open file `", DepFileName,
-				"' for output."], Message) },
-		report_error(Message)
+				"' for output."], DepMessage) },
+		report_error(DepMessage)
+	    ),
+	    globals__io_lookup_bool_option(generate_module_order, Order),
+	    ( { Order = yes } ->
+	  	{ string__append(Module, ".order", OrdFileName) },
+		maybe_write_string(Verbose, "% Creating module order file `"),
+		maybe_write_string(Verbose, OrdFileName),
+		maybe_write_string(Verbose, "'...\n"),
+		io__open_output(OrdFileName, OrdResult),
+		( { OrdResult = ok(OrdStream) } ->
+		    { relation__init(DepsRel0) },
+		    { map__to_assoc_list(DepsMap, DepsList) },
+		    { deps_map_to_deps_rel(DepsList, DepsMap, 
+				DepsRel0, DepsRel) },
+		    { relation__atsort(DepsRel, DepsOrdering) },
+		    io__write_list(OrdStream, DepsOrdering, "\n\n", 
+		    		write_module_scc(OrdStream)),
+		    io__close_output(OrdStream),
+		    maybe_write_string(Verbose, "% done\n")
+		;
+		    { string__append_list(["can't open file `", 
+				OrdFileName, "' for output."], OrdMessage) },
+		    report_error(OrdMessage)
+		)
+	    ;
+		[]
 	    )
 	).
 
+:- pred write_module_scc(io__output_stream::in, set(module_name)::in,
+		io__state::di, io__state::uo) is det.
+
+write_module_scc(Stream, SCC0) -->
+	{ set__to_sorted_list(SCC0, SCC) },
+	io__write_list(Stream, SCC, "\n", io__write_string).
+
+
 % This is the data structure we use to record the dependencies.
 % We keep a map from module name to information about the module.
 
@@ -511,6 +681,9 @@ generate_dependencies(Module) -->
 		list(string)	% fact table dependencies
 	).
 
+	% (Module1 deps_rel Module2) means Module1 is imported by Module2.
+:- type deps_rel == relation(string).
+
 % This is the predicate which creates the above data structure.
 
 :- pred generate_deps_map(list(string), deps_map, deps_map,
@@ -521,8 +694,8 @@ generate_deps_map([], DepsMap, DepsMap) --> [].
 generate_deps_map([Module | Modules], DepsMap0, DepsMap) -->
 		% Look up the module's dependencies, and determine whether
 		% it has been processed yet.
-	lookup_dependencies(Module, DepsMap0, no, Done, Error, ImplDeps, 
-				IntDeps, FactDeps, DepsMap1),
+	lookup_dependencies(Module, DepsMap0, no, Done, Error, IntDeps, 
+				ImplDeps, FactDeps, DepsMap1),
 		% If the module hadn't been processed yet, compute its
 		% transitive dependencies (we already know its primary ones),
 		% (1) output this module's dependencies to its `.d' file
@@ -550,6 +723,30 @@ generate_deps_map([Module | Modules], DepsMap0, DepsMap) -->
 	generate_deps_map(Modules2, DepsMap3, DepsMap).
 
 
+	% Construct a dependency relation of all the modules in the program.
+:- pred deps_map_to_deps_rel(assoc_list(string, deps), deps_map,
+		deps_rel, deps_rel).
+:- mode deps_map_to_deps_rel(in, in, in, out) is det.
+
+deps_map_to_deps_rel([], _, Rel, Rel).
+deps_map_to_deps_rel([Module - Deps | DepsList], DepsMap, Rel0, Rel) :-
+	Deps = deps(_, ModuleError, IntDeps, ImplDeps, _),
+	( ModuleError \= fatal ->
+		relation__add_element(Rel0, Module, ModuleRelKey, Rel1),
+		AddDeps =
+		    lambda([Dep::in, Relation0::in, Relation::out] is det, (
+			relation__add_element(Relation0, Dep,
+				DepRelKey, Relation1),
+			relation__add(Relation1, DepRelKey,
+				ModuleRelKey, Relation)
+		    )),
+		list__foldl(AddDeps, IntDeps, Rel1, Rel2),
+		list__foldl(AddDeps, ImplDeps, Rel2, Rel3)
+	;
+		Rel3 = Rel0
+	),
+	deps_map_to_deps_rel(DepsList, DepsMap, Rel3, Rel).
+
 % Write out the `.dep' file, using the information collected in the
 % deps_map data structure.
 
@@ -591,136 +788,100 @@ generate_dep_file(ModuleName, DepsMap, DepStream) -->
 	io__write_string(DepStream, ModuleName),
 	io__write_string(DepStream, ".nos = "),
 	write_compact_dependencies_list(Modules, ".no", Basis, DepStream),
-	io__write_string(DepStream, "\n\n"),
+	io__write_string(DepStream, "\n"),
 
 	io__write_string(DepStream, ModuleName),
 	io__write_string(DepStream, ".qls = "),
 	write_compact_dependencies_list(Modules, ".ql", Basis, DepStream),
-	io__write_string(DepStream, "\n\n"),
+	io__write_string(DepStream, "\n"),
 
 	io__write_string(DepStream, ModuleName),
 	io__write_string(DepStream, ".cs = "),
 	write_compact_dependencies_list(Modules, ".c", Basis, DepStream),
-	io__write_string(DepStream, "\n\n"),
+	write_dependencies_list(ExtraLinkObjs, ".c", DepStream),
+	io__write_string(DepStream, "\n"),
 
 	io__write_string(DepStream, ModuleName),
 	io__write_string(DepStream, ".os = "),
 	write_compact_dependencies_list(Modules, ".o", Basis, DepStream),
 	write_dependencies_list(ExtraLinkObjs, ".o", DepStream),
-	io__write_string(DepStream, "\n\n"),
+	io__write_string(DepStream, "\n"),
 
 	io__write_string(DepStream, ModuleName),
 	io__write_string(DepStream, ".pic_os = "),
 	write_compact_dependencies_list(Modules, ".$(EXT_FOR_PIC_OBJECTS)",
 		Basis, DepStream),
-	io__write_string(DepStream, "\n\n"),
+	io__write_string(DepStream, "\n"),
 
 	io__write_string(DepStream, ModuleName),
 	io__write_string(DepStream, ".dirs = "),
 	write_compact_dependencies_list(Modules, ".dir", Basis, DepStream),
-	io__write_string(DepStream, "\n\n"),
+	io__write_string(DepStream, "\n"),
 
 	io__write_string(DepStream, ModuleName),
 	io__write_string(DepStream, ".dir_os = "),
 	write_compact_dependencies_list(Modules, ".dir/*.o", Basis, DepStream),
-	io__write_string(DepStream, "\n\n"),
+	io__write_string(DepStream, "\n"),
 
 	io__write_string(DepStream, ModuleName),
 	io__write_string(DepStream, ".ss = "),
 	write_compact_dependencies_list(Modules, ".s", Basis, DepStream),
-	io__write_string(DepStream, "\n\n"),
+	io__write_string(DepStream, "\n"),
 
 	io__write_string(DepStream, ModuleName),
 	io__write_string(DepStream, ".errs = "),
 	write_compact_dependencies_list(Modules, ".err", Basis, DepStream),
-	io__write_string(DepStream, "\n\n"),
+	io__write_string(DepStream, "\n"),
 
 	io__write_string(DepStream, ModuleName),
 	io__write_string(DepStream, ".dates = "),
 	write_compact_dependencies_list(Modules, ".date", Basis, DepStream),
-	io__write_string(DepStream, "\n\n"),
+	io__write_string(DepStream, "\n"),
 
 	io__write_string(DepStream, ModuleName),
 	io__write_string(DepStream, ".date3s = "),
 	write_compact_dependencies_list(Modules, ".date3", Basis, DepStream),
-	io__write_string(DepStream, "\n\n"),
+	io__write_string(DepStream, "\n"),
 
 	io__write_string(DepStream, ModuleName),
 	io__write_string(DepStream, ".optdates = "),
 	write_compact_dependencies_list(Modules, ".optdate", Basis, DepStream),
-	io__write_string(DepStream, "\n\n"),
+	io__write_string(DepStream, "\n"),
 
 	io__write_string(DepStream, ModuleName),
 	io__write_string(DepStream, ".ds = "),
 	write_compact_dependencies_list(Modules, ".d", Basis, DepStream),
-	io__write_string(DepStream, "\n\n"),
+	io__write_string(DepStream, "\n"),
 
 	io__write_string(DepStream, ModuleName),
 	io__write_string(DepStream, ".hs = "),
 	write_compact_dependencies_list(Modules, ".h", Basis, DepStream),
-	io__write_string(DepStream, "\n\n"),
+	io__write_string(DepStream, "\n"),
 
 	io__write_string(DepStream, ModuleName),
 	io__write_string(DepStream, ".ints = "),
 	write_compact_dependencies_list(Modules, ".int", Basis, DepStream),
 	write_compact_dependencies_separator(Basis, DepStream),
 	write_compact_dependencies_list(Modules, ".int2", Basis, DepStream),
-	io__write_string(DepStream, "\n\n"),
+	io__write_string(DepStream, "\n"),
 
 	io__write_string(DepStream, ModuleName),
 	io__write_string(DepStream, ".int3s = "),
 	write_compact_dependencies_list(Modules, ".int3", Basis, DepStream),
-	io__write_string(DepStream, "\n\n"),
+	io__write_string(DepStream, "\n"),
 
 	io__write_string(DepStream, ModuleName),
 	io__write_string(DepStream, ".opts = "),
 	write_compact_dependencies_list(Modules, ".opt", Basis, DepStream),
 	io__write_string(DepStream, "\n\n"),
 
-	globals__io_lookup_string_option(gc, GC_Opt),
-	( { GC_Opt = "accurate" } ->
-		{ map__init(AllMap) },
-		trans_impl_dependencies([ModuleName], AllMap, yes,
-			AllDeps, _),
-		io__write_string(DepStream, ModuleName),
-		io__write_string(DepStream, ".garbs = "),
-		write_compact_dependencies_list(AllDeps, ".garb",
-			Basis, DepStream),
-		io__write_string(DepStream, "\n\n")
-	;
-		[]
-	),
-
-	( { GC_Opt = "accurate" } ->
-		io__write_strings(DepStream, [
-			ModuleName, "_garb.c : $(",
-			ModuleName, ".garbs)\n",
-                        "\t$(MLINK) $(MLINKFLAGS) -o ", ModuleName,
-                        "_garb.c $(", ModuleName, ".garbs)\n\n"
-		])
-	;
-		[]
-	),
-
-	( { GC_Opt = "accurate" } ->
-		io__write_strings(DepStream, [
-			ModuleName, " : $(", ModuleName, ".os) ",
-			ModuleName, "_init.o ",
-			ModuleName, "_garb.o\n",
+	io__write_strings(DepStream, [
+		ModuleName, " : $(", ModuleName, ".os) ",
+		ModuleName, "_init.o\n",
 		"\t$(ML) -s $(GRADE) $(MLFLAGS) -o ", ModuleName, " ",
-			ModuleName, "_garb.o ",
-			ModuleName, "_init.o \\\n",
-			"\t$(", ModuleName, ".os) $(MLLIBS)\n\n"
-		])
-	;
-		io__write_strings(DepStream, [
-			ModuleName, " : $(", ModuleName, ".os) ",
-			ModuleName, "_init.o\n",
-			"\t$(ML) -s $(GRADE) $(MLFLAGS) -o ", ModuleName, " ",
-			ModuleName, "_init.o \\\n",
-			"\t$(", ModuleName, ".os) $(MLLIBS)\n\n"
-		])
-	),
+		ModuleName, "_init.o \\\n",
+		"\t$(", ModuleName, ".os) $(MLLIBS)\n\n"
+	]),
 
 	io__write_strings(DepStream, [
 		ModuleName, ".split : ", ModuleName, ".split.a ",
@@ -733,32 +894,43 @@ generate_dep_file(ModuleName, DepsMap, DepStream) -->
 	io__write_strings(DepStream, [
 		ModuleName, ".split.a : $(", ModuleName, ".dir_os)\n",
 		"\trm -f ", ModuleName, ".split.a\n",
-		"\tar cr ", ModuleName, ".split.a\n",
+		"\t$(AR) $(ARFLAGS) ", ModuleName, ".split.a\n",
 		"\tfor dir in $(", ModuleName, ".dirs); do \\\n",
-		"\t	ar q ", ModuleName, ".split.a $$dir/*.o; \\\n",
+		"\t	$(AR) q ", ModuleName, ".split.a $$dir/*.o; \\\n",
 		"\tdone\n",
-		"\tranlib ", ModuleName, ".split.a\n\n"
-	]),
-
-/************
-% I decided to leave the rules for `foo.so' and `foo.a' out,
-% mainly because the rule for `foo.so' is hard to make portable.
-% The rules here would conflict with the one in library/Mmake.
-
-	io__write_strings(DepStream, [
-		ModuleName, ".so : $(", ModuleName, ".pic_os)\n",
-		"\t$(LINK_SHARED_LIB) -o ", ModuleName, ".so ",
-			"$(", ModuleName, ".pic_os)\n\n"
+		"\t$(RANLIB) $(RANLIBFLAGS) ", ModuleName, ".split.a\n\n"
 	]),
 
 	io__write_strings(DepStream, [
-		ModuleName, ".a : $(", ModuleName, ".os)\n",
+		"lib", ModuleName, " : ",
+		"lib", ModuleName, ".a ",
+		"lib", ModuleName, ".$(EXT_FOR_SHARED_LIB) \\\n",
+		"\t\t$(", ModuleName, ".ints) ",
+		"$(", ModuleName, ".opts) ",
+		ModuleName, ".init\n\n"
+	]),
+
+	io__write_strings(DepStream, [
+		"lib", ModuleName, ".so : $(", ModuleName, ".pic_os)\n",
+		"\t$(ML) --make-shared-lib --grade $(GRADE) $(MLFLAGS) -o ",
+			"lib", ModuleName, ".so \\\n",
+		"\t\t$(", ModuleName, ".pic_os) $(MLLIBS)\n\n"
+	]),
+
+	io__write_strings(DepStream, [
+		"lib", ModuleName, ".a : $(", ModuleName, ".os)\n",
 		"\trm -f ", ModuleName, ".a\n",
-		"\tar cr ", ModuleName, ".a ",
+		"\t$(AR) $(ARFLAGS) lib", ModuleName, ".a ",
 			"$(", ModuleName, ".os)\n",
-		"\tranlib ", ModuleName, ".a\n\n"
+		"\t$(RANLIB) $(RANLIBFLAGS) lib", ModuleName, ".a\n\n"
+	]),
+
+	io__write_strings(DepStream, [
+		ModuleName, ".init : ", ModuleName, ".dep\n",
+		"\tfor file in $(", ModuleName, ".ms); do \\\n",
+		"\t\techo ""INIT mercury__`basename $$file .m`__init""; \\\n",
+		"\tdone > ", ModuleName, ".init\n\n"
 	]),
-************/
 
 	io__write_strings(DepStream, [
 		ModuleName, "_init.c :\n",
@@ -809,15 +981,7 @@ generate_dep_file(ModuleName, DepsMap, DepStream) -->
 		"\t-rm -f $(", ModuleName, ".errs)\n"
 	]),
 
-	( { GC_Opt = "accurate" } ->
-		io__write_strings(DepStream, [
-			"\t-rm -f $(", ModuleName, ".garbs)\n",
-			"\t-rm -f ", ModuleName, "_garb.o ", ModuleName, 
-				"_garb.c ", ModuleName, "_garb.s\n\n"
-		])
-	;
-		io__write_string(DepStream, "\n")
-	),
+	io__write_string(DepStream, "\n"),
 
 	io__write_strings(DepStream, [
 		ModuleName, ".change_clean :\n",
@@ -826,8 +990,14 @@ generate_dep_file(ModuleName, DepsMap, DepStream) -->
 		"\t-rm -f $(", ModuleName, ".os) ", ModuleName, "_init.o\n",
 		"\t-rm -f $(", ModuleName, ".hs)\n",
 		"\t-rm -f $(", ModuleName, ".ds)\n",
-		"\t-rm -f ", ModuleName, ".dep ", ModuleName, "\n",
-		"\t-rm -f ", ModuleName, ".split ", ModuleName,".split.a\n\n"
+		"\t-rm -f ",
+			ModuleName, " ",
+			ModuleName, ".split ",
+			ModuleName, ".split.a ",
+			ModuleName, ".init ",
+			"lib", ModuleName, ".a ",
+			"lib", ModuleName, ".so ",
+			ModuleName, ".dep\n\n"
 	]),
 
 	io__write_strings(DepStream, [
@@ -850,6 +1020,9 @@ generate_dep_file(ModuleName, DepsMap, DepStream) -->
 			ModuleName, " ",
 			ModuleName, ".split ",
 			ModuleName, ".split.a ",
+			ModuleName, ".init ",
+			"lib", ModuleName, ".a ",
+			"lib", ModuleName, ".so ",
 			ModuleName, ".nu ",
 			ModuleName, ".nu.save ",
 			ModuleName, ".nu.debug.save ",
@@ -1045,9 +1218,9 @@ lookup_dependencies(Module, DepsMap0, Search, Done, Error, IntDeps,
 		{ FactDeps = FactDeps0 },
 		{ DepsMap = DepsMap0 }
 	;
-		read_dependencies(Module, Search, ImplDeps, IntDeps, FactDeps, 
+		read_dependencies(Module, Search, IntDeps, ImplDeps, FactDeps, 
 				Error),
-		{ map__set(DepsMap0, Module, 
+		{ map__det_insert(DepsMap0, Module, 
 		    deps(no, Error, IntDeps, ImplDeps, FactDeps), DepsMap) },
 		{ Done = no }
 	).
@@ -1060,18 +1233,23 @@ lookup_dependencies(Module, DepsMap0, Search, Done, Error, IntDeps,
 
 read_dependencies(Module, Search, InterfaceDeps, ImplementationDeps, 
 		FactTableDeps, Error) -->
-	io__gc_call(read_mod_ignore_errors(Module, ".m",
-			"Getting dependencies for module", Search, Items0, Error)),
+	read_mod_ignore_errors(Module, ".m",
+		"Getting dependencies for module", Search, Items0, Error),
 	( { Items0 = [], Error = fatal } ->
-		io__gc_call(read_mod_ignore_errors(Module, ".int", 
+		read_mod_ignore_errors(Module, ".int", 
 		    "Getting dependencies for module interface", Search, 
-		    Items, _Error))
+		    Items, _Error)
 	;
 		{ Items = Items0 }
 	),
-	{ get_dependencies(Items, ImplementationDeps0) },
+	{ get_dependencies(Items, ImplImportDeps, ImplUseDeps) },
+	{ list__append(ImplImportDeps, ImplUseDeps, ImplementationDeps0) },
 	{ get_interface(Items, no, InterfaceItems) },
-	{ get_dependencies(InterfaceItems, InterfaceDeps) },
+	{ get_dependencies(InterfaceItems, InterfaceImportDeps,
+		InterfaceUseDeps) },
+	{ list__append(InterfaceImportDeps, InterfaceUseDeps, 
+		InterfaceDeps0) },
+	{ InterfaceDeps = ["mercury_builtin" | InterfaceDeps0] },
 	{ ImplementationDeps = ["mercury_builtin" | ImplementationDeps0] },
 	{ get_fact_table_dependencies(Items, FactTableDeps) }.
 
@@ -1151,9 +1329,9 @@ process_module_interfaces([Import | Imports], IndirectImports0, Module0, Module)
 		( { ModuleName = "mercury_builtin" } ->
 			[]
 		;
-			globals__io_lookup_bool_option(inhibit_warnings,
-							NoWarn),
-			( { NoWarn = no } ->
+			globals__io_lookup_bool_option(warn_simple_code,
+							Warn),
+			( { Warn = yes } ->
 				{ term__context_init(ModuleName, 1, Context) },
 				prog_out__write_context(Context),
 				io__write_string(
@@ -1170,11 +1348,9 @@ process_module_interfaces([Import | Imports], IndirectImports0, Module0, Module)
 		process_module_interfaces(Imports, IndirectImports0,
 					Module0, Module)
 	;
-		io__gc_call(
-			read_mod_interface(Import,
-				"Reading interface for module", yes, 
-				LongIntItems1, Error1)
-		),
+		read_mod_interface(Import,
+			"Reading interface for module", yes, 
+			LongIntItems1, Error1),
 		% strip off the `:- interface' declaration at the start, if any
 		{
 			LongIntItems1 = [ FirstItem | LongIntItems2 ],
@@ -1193,7 +1369,7 @@ process_module_interfaces([Import | Imports], IndirectImports0, Module0, Module)
 		globals__io_lookup_bool_option(statistics, Statistics),
 		maybe_report_stats(Statistics),
 
-		{ get_dependencies(Items1, IndirectImports1) },
+		{ get_dependencies(Items1, IndirectImports1, IndirectUses1) },
 		( { Error1 = fatal } ->
 			{ DirectImports1 = DirectImports0 }
 		;
@@ -1201,10 +1377,12 @@ process_module_interfaces([Import | Imports], IndirectImports0, Module0, Module)
 		),
 		{ list__append(IndirectImports0, IndirectImports1,
 			IndirectImports2) },
+		{ list__append(IndirectImports2, IndirectUses1,
+			IndirectImports3) },
 		{ list__append(Items0, Items1, Items2) },
 		{ Module1 = module_imports(ModuleName, DirectImports1, 
 					OldIndirectImports, Items2, Error2) },
-		process_module_interfaces(Imports, IndirectImports2,
+		process_module_interfaces(Imports, IndirectImports3,
 				Module1, Module)
 	).
 
@@ -1215,7 +1393,21 @@ process_module_interfaces([Import | Imports], IndirectImports0, Module0, Module)
 :- mode process_module_short_interfaces(in, in, out, di, uo) is det.
 
 process_module_short_interfaces(Imports, Module0, Module) -->
-	process_module_short_interfaces(Imports, ".int2", Module0, Module).
+	{ Module0 = module_imports(ModuleName, DirectImports0,
+		IndirectImports0, Items0, Error0) },
+
+		% Treat indirectly imported items as if they were imported 
+		% using `:- use_module', since all uses of them in the `.int'
+		% files must be module qualified.
+	{ varset__init(Varset) },
+	{ term__context_init(Context) },
+	{ list__append(Items0, [module_defn(Varset, used) - Context], 
+		Items1) },
+
+	{ Module1 = module_imports(ModuleName, DirectImports0,
+		IndirectImports0, Items1, Error0) },
+
+	process_module_short_interfaces(Imports, ".int2", Module1, Module).
 
 
 :- pred process_module_short_interfaces(list(string), string, 
@@ -1235,11 +1427,9 @@ process_module_short_interfaces([Import | Imports], Ext, Module0, Module) -->
 	->
 		process_module_short_interfaces(Imports, Ext, Module0, Module)
 	;
-		io__gc_call(
-			read_mod_short_interface(Import, Ext,
+		read_mod_short_interface(Import, Ext,
 				"Reading short interface for module", yes,
-					ShortIntItems1, Error1)
-		),
+				ShortIntItems1, Error1),
 		% strip off the `:- interface' declaration at the start, if any
 		{
 			ShortIntItems1 = [ FirstItem | ShortIntItems2 ],
@@ -1258,7 +1448,7 @@ process_module_short_interfaces([Import | Imports], Ext, Module0, Module) -->
 		globals__io_lookup_bool_option(statistics, Statistics),
 		maybe_report_stats(Statistics),
 
-		{ get_dependencies(Items1, Imports1) },
+		{ get_dependencies(Items1, Imports1, _Uses1) },
 		{ list__append(Imports, Imports1, Imports2) },
 		{ list__append(Items0, Items1, Items2) },
 		{ IndirectImports1 = [Import | IndirectImports0] },
@@ -1269,22 +1459,31 @@ process_module_short_interfaces([Import | Imports], Ext, Module0, Module) -->
 
 %-----------------------------------------------------------------------------%
 
-get_dependencies(Items, Deps) :-
-	get_dependencies_2(Items, [], Deps).
+get_dependencies(Items, ImportDeps, UseDeps) :-
+	get_dependencies_2(Items, [], ImportDeps, [], UseDeps).
 
-:- pred get_dependencies_2(item_list, list(string), list(string)).
-:- mode get_dependencies_2(in, in, out) is det.
+:- pred get_dependencies_2(item_list, list(string), list(string), 
+		list(string), list(string)).
+:- mode get_dependencies_2(in, in, out, in, out) is det.
 
-get_dependencies_2([], Deps, Deps).
-get_dependencies_2([Item - _Context | Items], Deps0, Deps) :-
+get_dependencies_2([], ImportDeps, ImportDeps, UseDeps, UseDeps).
+get_dependencies_2([Item - _Context | Items], ImportDeps0, ImportDeps,
+		UseDeps0, UseDeps) :-
 	( 
 		Item = module_defn(_VarSet, import(module(Modules)))
 	->
-		list__append(Deps0, Modules, Deps1)
+		list__append(ImportDeps0, Modules, ImportDeps1),
+		UseDeps1 = UseDeps0
 	;
-		Deps1 = Deps0
+		Item = module_defn(_VarSet, use(module(Modules)))
+	->
+		list__append(UseDeps0, Modules, UseDeps1),
+		ImportDeps1 = ImportDeps0
+	;
+		ImportDeps1 = ImportDeps0,
+		UseDeps1 = UseDeps0
 	),
-	get_dependencies_2(Items, Deps1, Deps).
+	get_dependencies_2(Items, ImportDeps1, ImportDeps, UseDeps1, UseDeps).
 
 %-----------------------------------------------------------------------------%
 	% get the fact table dependencies for a module
@@ -1332,16 +1531,22 @@ get_interface_2([Item - Context | Rest], InInterface0,
 	( Item = module_defn(_, interface) ->
 		Items1 = Items0,
 		InInterface1 = yes
-	; Item = module_defn(_, imported) ->
+	; 
+		Item = module_defn(_, Defn),
+		( Defn = imported
+		; Defn = used
+		)
+	->
 		% module_qual.m needs the :- imported declaration.
-		( IncludeImported = yes ->
-			InInterface1 = yes,
+		( IncludeImported = yes, InInterface0 = yes ->
 			Items1 = [Item - Context | Items0]
 		;
-			InInterface1 = no,
 			Items1 = Items0
-		)
-	; Item = module_defn(_, implementation) ->
+		),
+		InInterface1 = InInterface0
+	;
+		Item = module_defn(_, implementation) 
+	->
 		Items1 = Items0,
 		InInterface1 = no
 	;
@@ -1393,6 +1598,10 @@ get_short_interface_2([ItemAndContext | Rest], Items0, Imports0, NeedsImports0,
 		Items1 = Items0,
 		Imports1 = [ItemAndContext | Imports0],
 		NeedsImports1 = NeedsImports0
+	; Item0 = module_defn(_, use(_)) ->
+		Items1 = Items0,
+		Imports1 = [ItemAndContext | Imports0],
+		NeedsImports1 = NeedsImports0
 	; make_abstract_type_defn(Item0, Item1) ->
 		Imports1 = Imports0,
 		Items1 = [Item1 - Context | Items0],
@@ -1420,7 +1629,7 @@ include_in_short_interface(module_defn(_, _)).
 :- pred make_abstract_type_defn(item, item).
 :- mode make_abstract_type_defn(in, out) is semidet.
 
-make_abstract_type_defn(type_defn(VarSet, du_type(Name, Args, _Ctors), Cond),
+make_abstract_type_defn(type_defn(VarSet, du_type(Name, Args, _, _), Cond),
 			type_defn(VarSet, abstract_type(Name, Args), Cond)).
 make_abstract_type_defn(type_defn(VarSet, abstract_type(Name, Args), Cond),
 			type_defn(VarSet, abstract_type(Name, Args), Cond)).
