@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1995 University of Melbourne.
+% Copyright (C) 1994-1997 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -112,8 +112,8 @@
 :- pred opt_debug__dump_vnlvals(list(vnlval), string).
 :- mode opt_debug__dump_vnlvals(in, out) is det.
 
-:- pred opt_debug__dump_reg(reg, string).
-:- mode opt_debug__dump_reg(in, out) is det.
+:- pred opt_debug__dump_reg(reg_type, int, string).
+:- mode opt_debug__dump_reg(in, in, out) is det.
 
 :- pred opt_debug__dump_vnlval(vnlval, string).
 :- mode opt_debug__dump_vnlval(in, out) is det.
@@ -130,6 +130,9 @@
 :- pred opt_debug__dump_rvals(list(rval), string).
 :- mode opt_debug__dump_rvals(in, out) is det.
 
+:- pred opt_debug__dump_mem_ref(mem_ref, string).
+:- mode opt_debug__dump_mem_ref(in, out) is det.
+
 :- pred opt_debug__dump_const(rval_const, string).
 :- mode opt_debug__dump_const(in, out) is det.
 
@@ -182,7 +185,7 @@
 
 :- implementation.
 
-:- import_module llds_out, opt_util, vn_util.
+:- import_module llds_out, opt_util, vn_util, hlds_pred.
 :- import_module int, set, map, string.
 
 opt_debug__msg(OptDebug, Msg) -->
@@ -315,10 +318,16 @@ opt_debug__dump_vninstr(vn_restore_hp(Vn), Str) :-
 opt_debug__dump_vninstr(vn_store_ticket(Vnlval), Str) :-
 	opt_debug__dump_vnlval(Vnlval, V_str),
 	string__append_list(["store_ticket(", V_str, ")"], Str).
-opt_debug__dump_vninstr(vn_restore_ticket(Vn), Str) :-
+opt_debug__dump_vninstr(vn_reset_ticket(Vn, _Reason), Str) :-
 	opt_debug__dump_vn(Vn, Vn_str),
-	string__append_list(["restore_ticket(", Vn_str, ")"], Str).
+	string__append_list(["reset_ticket(", Vn_str, ", _)"], Str).
 opt_debug__dump_vninstr(vn_discard_ticket, "discard_ticket").
+opt_debug__dump_vninstr(vn_mark_ticket_stack(Vnlval), Str) :-
+	opt_debug__dump_vnlval(Vnlval, V_str),
+	string__append_list(["mark_ticket_stack(", V_str, ")"], Str).
+opt_debug__dump_vninstr(vn_discard_tickets_to(Vn), Str) :-
+	opt_debug__dump_vn(Vn, Vn_str),
+	string__append_list(["discard_tickets_to(", Vn_str, ", _)"], Str).
 opt_debug__dump_vninstr(vn_incr_sp(N, _), Str) :-
 	string__int_to_string(N, N_str),
 	string__append_list(["incr_sp(", N_str, ")"], Str).
@@ -460,16 +469,19 @@ opt_debug__dump_vnlvals([Lval | Lvals], Str) :-
 	opt_debug__dump_vnlvals(Lvals, Tail_str),
 	string__append_list([" ", Lval_str, Tail_str], Str).
 
-opt_debug__dump_reg(r(N), Str) :-
+opt_debug__dump_reg(r, N, Str) :-
 	string__int_to_string(N, N_str),
 	string__append_list(["r(", N_str, ")"], Str).
-opt_debug__dump_reg(f(N), Str) :-
+opt_debug__dump_reg(f, N, Str) :-
 	string__int_to_string(N, N_str),
 	string__append_list(["f(", N_str, ")"], Str).
 
-opt_debug__dump_vnlval(vn_reg(R), Str) :-
-	opt_debug__dump_reg(R, R_str),
+opt_debug__dump_vnlval(vn_reg(Type, Num), Str) :-
+	opt_debug__dump_reg(Type, Num, R_str),
 	string__append_list(["vn_reg(", R_str, ")"], Str).
+opt_debug__dump_vnlval(vn_temp(Type, Num), Str) :-
+	opt_debug__dump_reg(Type, Num, R_str),
+	string__append_list(["vn_temp(", R_str, ")"], Str).
 opt_debug__dump_vnlval(vn_stackvar(N), Str) :-
 	string__int_to_string(N, N_str),
 	string__append_list(["vn_stackvar(", N_str, ")"], Str).
@@ -504,9 +516,9 @@ opt_debug__dump_vnlval(vn_field(T, N, F), Str) :-
 	string__int_to_string(F, F_str),
 	string__append_list(["vn_field(", T_str, ", ", N_str, ", ",
 		F_str, ")"], Str).
-opt_debug__dump_vnlval(vn_temp(R), Str) :-
-	opt_debug__dump_reg(R, R_str),
-	string__append_list(["vn_temp(", R_str, ")"], Str).
+opt_debug__dump_vnlval(vn_mem_ref(N), Str) :-
+	string__int_to_string(N, N_str),
+	string__append_list(["vn_mem_ref(", N_str, ")"], Str).
 
 opt_debug__dump_vnrval(vn_origlval(Vnlval), Str) :-
 	opt_debug__dump_vnlval(Vnlval, Lval_str),
@@ -534,9 +546,21 @@ opt_debug__dump_vnrval(vn_binop(O, N1, N2), Str) :-
 	string__int_to_string(N2, N2_str),
 	string__append_list(["vn_binop(", O_str, ", ", N1_str, ", ",
 		N2_str, ")"], Str).
+opt_debug__dump_vnrval(vn_stackvar_addr(N), Str) :-
+	string__int_to_string(N, N_str),
+	string__append_list(["vn_stackvar_addr(", N_str, ")"], Str).
+opt_debug__dump_vnrval(vn_framevar_addr(N), Str) :-
+	string__int_to_string(N, N_str),
+	string__append_list(["vn_framevar_addr(", N_str, ")"], Str).
+opt_debug__dump_vnrval(vn_heap_addr(B, T, N), Str) :-
+	string__int_to_string(B, B_str),
+	string__int_to_string(T, T_str),
+	string__int_to_string(N, N_str),
+	string__append_list(["vn_heap_addr(", B_str, ", ", T_str, ", ",
+		N_str, ")"], Str).
 
-opt_debug__dump_lval(reg(R), Str) :-
-	opt_debug__dump_reg(R, R_str),
+opt_debug__dump_lval(reg(Type, Num), Str) :-
+	opt_debug__dump_reg(Type, Num, R_str),
 	string__append_list(["reg(", R_str, ")"], Str).
 opt_debug__dump_lval(stackvar(N), Str) :-
 	string__int_to_string(N, N_str),
@@ -574,9 +598,12 @@ opt_debug__dump_lval(field(T, N, F), Str) :-
 		F_str, ")"], Str).
 opt_debug__dump_lval(lvar(_), Str) :-
 	string__append_list(["lvar(_)"], Str).
-opt_debug__dump_lval(temp(R), Str) :-
-	opt_debug__dump_reg(R, R_str),
+opt_debug__dump_lval(temp(Type, Num), Str) :-
+	opt_debug__dump_reg(Type, Num, R_str),
 	string__append_list(["temp(", R_str, ")"], Str).
+opt_debug__dump_lval(mem_ref(R), Str) :-
+	opt_debug__dump_rval(R, R_str),
+	string__append_list(["mem_ref(", R_str, ")"], Str).
 
 opt_debug__dump_rval(lval(Lval), Str) :-
 	opt_debug__dump_lval(Lval, Lval_str),
@@ -613,6 +640,9 @@ opt_debug__dump_rval(binop(O, N1, N2), Str) :-
 	opt_debug__dump_rval(N2, N2_str),
 	string__append_list(["binop(", O_str, ", ", N1_str, ", ",
 		N2_str, ")"], Str).
+opt_debug__dump_rval(mem_addr(M), Str) :-
+	opt_debug__dump_mem_ref(M, M_str),
+	string__append_list(["mem_addr(", M_str, ")"], Str).
 
 opt_debug__dump_rvals([], "").
 opt_debug__dump_rvals([Rval | Rvals], Str) :-
@@ -620,6 +650,19 @@ opt_debug__dump_rvals([Rval | Rvals], Str) :-
 	opt_debug__dump_rvals(Rvals, S_str),
 	string__append_list([R_str, ", ", S_str], Str).
 
+opt_debug__dump_mem_ref(stackvar_ref(N), Str) :-
+	string__int_to_string(N, N_str),
+	string__append_list(["stackvar_ref(", N_str, ")"], Str).
+opt_debug__dump_mem_ref(framevar_ref(N), Str) :-
+	string__int_to_string(N, N_str),
+	string__append_list(["framevar_ref(", N_str, ")"], Str).
+opt_debug__dump_mem_ref(heap_ref(R, T, N), Str) :-
+	opt_debug__dump_rval(R, R_str),
+	string__int_to_string(T, T_str),
+	string__int_to_string(N, N_str),
+	string__append_list(["heap_ref(", R_str, ", ", T_str, ", ",
+		N_str, ")"], Str).
+
 opt_debug__dump_const(true, "true").
 opt_debug__dump_const(false, "false").
 opt_debug__dump_const(int_const(I), Str) :-
@@ -631,7 +674,7 @@ opt_debug__dump_const(string_const(I), Str) :-
 opt_debug__dump_const(code_addr_const(CodeAddr), Str) :-
 	opt_debug__dump_code_addr(CodeAddr, C_str),
 	string__append_list(["code_addr_const(", C_str, ")"], Str).
-opt_debug__dump_const(data_addr_const(data_addr(BaseName, VarName, _)), Str) :-
+opt_debug__dump_const(data_addr_const(data_addr(BaseName, VarName)), Str) :-
 	opt_debug__dump_data_name(VarName, N_str),
 	string__append_list(["data_addr_const(", BaseName, ", ", N_str, ")"],
 		Str).
@@ -639,12 +682,8 @@ opt_debug__dump_const(data_addr_const(data_addr(BaseName, VarName, _)), Str) :-
 opt_debug__dump_data_name(common(N), Str) :-
 	string__int_to_string(N, N_str),
 	string__append("common", N_str, Str).
-opt_debug__dump_data_name(base_type_info(TypeName, TypeArity), Str) :-
-	string__int_to_string(TypeArity, A_str),
-	string__append_list(["base_type_info_", TypeName, "_", A_str], Str).
-opt_debug__dump_data_name(base_type_layout(TypeName, TypeArity), Str) :-
-	string__int_to_string(TypeArity, A_str),
-	string__append_list(["base_type_layout_", TypeName, "_", A_str], Str).
+opt_debug__dump_data_name(base_type(BaseData, TypeName, TypeArity), Str) :-
+	llds_out__make_base_type_name(BaseData, TypeName, TypeArity, Str).
 
 opt_debug__dump_unop(mktag, "mktag").
 opt_debug__dump_unop(tag, "tag").
@@ -693,6 +732,7 @@ opt_debug__dump_code_addr(do_fail, "do_fail").
 opt_debug__dump_code_addr(do_det_closure, "do_det_closure").
 opt_debug__dump_code_addr(do_semidet_closure, "do_semidet_closure").
 opt_debug__dump_code_addr(do_nondet_closure, "do_nondet_closure").
+opt_debug__dump_code_addr(do_not_reached, "do_not_reached").
 
 opt_debug__dump_code_addrs([], "").
 opt_debug__dump_code_addrs([Addr | Addrs], Str) :-
@@ -724,16 +764,26 @@ opt_debug__dump_label_pairs([L1 - L2 | Labels], Str) :-
 	opt_debug__dump_label_pairs(Labels, L_str),
 	string__append_list([" ", L1_str, "-", L2_str, L_str], Str).
 
-opt_debug__dump_proclabel(proc(Module, _PredOrFunc, Name, Arity, Mode), Str) :-
+opt_debug__dump_proclabel(proc(Module, _PredOrFunc, PredModule,
+		PredName, Arity, ProcId), Str) :-
 	string__int_to_string(Arity, A_str),
+	proc_id_to_int(ProcId, Mode),
 	string__int_to_string(Mode, M_str),
-	string__append_list([Module, "_", Name, "_", A_str, "_", M_str], Str).
-opt_debug__dump_proclabel(special_proc(Module, Pred, Type, Arity, Mode), Str) :-
+	( Module = PredModule ->
+		ExtraModule = ""
+	;
+		string__append(Module, "_", ExtraModule)
+	),
+	string__append_list([ExtraModule, Module, "_", PredName,
+		"_", A_str, "_", M_str], Str).
+opt_debug__dump_proclabel(special_proc(Module, Pred, TypeModule,
+		Type, Arity, ProcId), Str) :-
 	string__int_to_string(Arity, A_str),
+	proc_id_to_int(ProcId, Mode),
 	string__int_to_string(Mode, M_str),
-	llds_out__sym_name_mangle(Type, TypeStr),
-	string__append_list(
-		[Module, "_", Pred, "_", TypeStr, "_", A_str, "_", M_str], Str).
+	llds_out__qualify_name(TypeModule, Type, TypeName),
+	string__append_list([Module, "_", Pred, "_",
+		TypeName, "_", A_str, "_", M_str], Str).
 
 opt_debug__dump_bool(yes, "yes").
 opt_debug__dump_bool(no, "no").
@@ -804,22 +854,26 @@ opt_debug__dump_instr(restore_hp(Rval), Str) :-
 opt_debug__dump_instr(store_ticket(Lval), Str) :-
 	opt_debug__dump_lval(Lval, L_str),
 	string__append_list(["store_ticket(", L_str, ")"], Str).
-opt_debug__dump_instr(restore_ticket(Rval), Str) :-
+opt_debug__dump_instr(reset_ticket(Rval, _Reason), Str) :-
 	opt_debug__dump_rval(Rval, R_str),
-	string__append_list(["restore_ticket(", R_str, ")"], Str).
+	string__append_list(["reset_ticket(", R_str, ", _)"], Str).
 opt_debug__dump_instr(discard_ticket, "discard_ticket").
+opt_debug__dump_instr(mark_ticket_stack(Lval), Str) :-
+	opt_debug__dump_lval(Lval, L_str),
+	string__append_list(["mark_ticket_stack(", L_str, ")"], Str).
+opt_debug__dump_instr(discard_tickets_to(Rval), Str) :-
+	opt_debug__dump_rval(Rval, R_str),
+	string__append_list(["discard_tickets_to(", R_str, ")"], Str).
 opt_debug__dump_instr(incr_sp(Size, _), Str) :-
 	string__int_to_string(Size, S_str),
 	string__append_list(["incr_sp(", S_str, ")"], Str).
 opt_debug__dump_instr(decr_sp(Size), Str) :-
 	string__int_to_string(Size, S_str),
 	string__append_list(["decr_sp(", S_str, ")"], Str).
-
 % XXX  should probably give more info than this
-opt_debug__dump_instr(pragma_c(_, _, Code, _), Str) :-
+opt_debug__dump_instr(pragma_c(_, _, Code, _, _), Str) :-
 	string__append_list(["pragma_c(", Code, ")"], Str).
 
-
 opt_debug__dump_fullinstr(Uinstr - Comment, Str) :-
 	opt_debug__dump_instr(Uinstr, U_str),
 	string__append_list([U_str, " - ", Comment, "\n"], Str).
