@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1995 University of Melbourne.
+% Copyright (C) 1994-1997 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -21,9 +21,9 @@
 :- type tailmap == map(label, list(instruction)).
 :- type succmap == map(label, bool).
 
-:- pred opt_util__get_prologue(list(instruction), proc_label,
+:- pred opt_util__get_prologue(list(instruction), proc_label, instruction,
 	list(instruction), list(instruction)).
-:- mode opt_util__get_prologue(in, out, out, out) is det.
+:- mode opt_util__get_prologue(in, out, out, out, out) is det.
 
 :- pred opt_util__gather_comments(list(instruction),
 	list(instruction), list(instruction)).
@@ -69,9 +69,6 @@
 :- pred opt_util__find_first_label(list(instruction), label).
 :- mode opt_util__find_first_label(in, out) is det.
 
-	% Check whether the named label follows without any intervening code.
-	% If yes, return the instructions after the label.
-
 	% Skip to the next label, returning the code before the label,
 	% and the label together with the code after the label.
 
@@ -118,7 +115,11 @@
 :- mode opt_util__is_succeed_next(in, out) is semidet.
 
  	% Is the following code a test of r1, followed in both continuations
-	% by a semidet proceed with the same value of r1?
+	% by a semidet proceed? Is the code in both continuations the same,
+	% modulo livevals annotations and the value assigned to r1? Is TRUE
+	% assigned to r1 in the success continuation and FALSE in the failure
+	% continuation? If the answer is yes to all these questions, return
+	% the code shared by the two continuations.
 
 :- pred opt_util__is_forkproceed_next(list(instruction), tailmap,
 	list(instruction)).
@@ -176,18 +177,6 @@
 :- pred opt_util__is_const_condition(rval, bool).
 :- mode opt_util__is_const_condition(in, out) is semidet.
 
-	% See if an instruction sequence contains incr_sp, and if yes,
-	% what is the increment.
-
-% :- pred opt_util__has_incr_sp(list(instruction), int).
-% :- mode opt_util__has_incr_sp(in, out) is semidet.
-
-	% See if an instruction sequence contains decr_sp, and if yes,
-	% what is the decrement.
-
-% :- pred opt_util__has_decr_sp(list(instruction), int).
-% :- mode opt_util__has_decr_sp(in, out) is semidet.
-
 	% Check whether an instruction can possibly branch away.
 
 :- pred opt_util__can_instr_branch_away(instr, bool).
@@ -320,15 +309,16 @@
 
 :- implementation.
 
-:- import_module exprn_aux, llds_out.
+:- import_module exprn_aux, llds_out, hlds_pred.
 :- import_module int, string, set, map, require.
 
-opt_util__get_prologue(Instrs0, ProcLabel, Comments, Instrs) :-
+opt_util__get_prologue(Instrs0, ProcLabel, LabelInstr, Comments, Instrs) :-
 	opt_util__gather_comments(Instrs0, Comments1, Instrs1),
 	(
 		Instrs1 = [Instr1 | Instrs2],
 		Instr1 = label(FirstLabel) - _
 	->
+		LabelInstr = Instr1,
 		( FirstLabel = exported(ProcLabelPrime) ->
 			ProcLabel = ProcLabelPrime
 		; FirstLabel = local(ProcLabelPrime) ->
@@ -336,9 +326,8 @@ opt_util__get_prologue(Instrs0, ProcLabel, Comments, Instrs) :-
 		;
 			error("procedure begins with bad label type")
 		),
-		opt_util__gather_comments(Instrs2, Comments2, Instrs3),
-		list__append(Comments1, Comments2, Comments),
-		Instrs = [Instr1 | Instrs3]
+		opt_util__gather_comments(Instrs2, Comments2, Instrs),
+		list__append(Comments1, Comments2, Comments)
 	;
 		error("procedure does not begin with label")
 	).
@@ -534,7 +523,7 @@ opt_util__is_sdproceed_next_sf(Instrs0, InstrsBetween, Success) :-
 		Instrs5 = Instrs3
 	),
 	Instrs5 = [Instr5 | Instrs6],
-	Instr5 = assign(reg(r(1)), const(R1val)) - _,
+	Instr5 = assign(reg(r, 1), const(R1val)) - _,
 	(
 		R1val = true,
 		Success = yes
@@ -562,16 +551,20 @@ opt_util__is_succeed_next(Instrs0, InstrsBetweenIncl) :-
 opt_util__is_forkproceed_next(Instrs0, Sdprocmap, Between) :-
 	opt_util__skip_comments_labels(Instrs0, Instrs1),
 	Instrs1 = [Instr1 | Instrs2],
-	( Instr1 = if_val(lval(reg(r(1))), label(BranchLabel)) - _ ->
-		map__search(Sdprocmap, BranchLabel, BetweenBranch),
-		opt_util__filter_out_r1(BetweenBranch, yes(true), Between),
+	( Instr1 = if_val(lval(reg(r, 1)), label(JumpLabel)) - _ ->
+		map__search(Sdprocmap, JumpLabel, BetweenJump),
 		opt_util__is_sdproceed_next(Instrs2, BetweenFall),
-		opt_util__filter_out_r1(BetweenFall, yes(false), Between)
-	; Instr1 = if_val(unop(not, lval(reg(r(1)))), label(BranchLabel)) - _ ->
-		map__search(Sdprocmap, BranchLabel, BetweenBranch),
-		opt_util__filter_out_r1(BetweenBranch, yes(false), Between),
+		opt_util__filter_out_r1(BetweenJump, yes(true), BetweenTrue0),
+		opt_util__filter_out_livevals(BetweenTrue0, Between),
+		opt_util__filter_out_r1(BetweenFall, yes(false), BetweenFalse0),
+		opt_util__filter_out_livevals(BetweenFalse0, Between)
+	; Instr1 = if_val(unop(not, lval(reg(r, 1))), label(JumpLabel)) - _ ->
+		map__search(Sdprocmap, JumpLabel, BetweenJump),
 		opt_util__is_sdproceed_next(Instrs2, BetweenFall),
-		opt_util__filter_out_r1(BetweenFall, yes(true), Between)
+		opt_util__filter_out_r1(BetweenJump, yes(false), BetweenFalse0),
+		opt_util__filter_out_livevals(BetweenFalse0, Between),
+		opt_util__filter_out_r1(BetweenFall, yes(true), BetweenTrue0),
+		opt_util__filter_out_livevals(BetweenTrue0, Between)
 	;
 		fail
 	).
@@ -579,7 +572,7 @@ opt_util__is_forkproceed_next(Instrs0, Sdprocmap, Between) :-
 opt_util__filter_out_r1([], no, []).
 opt_util__filter_out_r1([Instr0 | Instrs0], Success, Instrs) :-
 	opt_util__filter_out_r1(Instrs0, Success0, Instrs1),
-	( Instr0 = assign(reg(r(1)), const(Success1)) - _ ->
+	( Instr0 = assign(reg(r, 1), const(Success1)) - _ ->
 		Instrs = Instrs1,
 		Success = yes(Success1)
 	;
@@ -647,7 +640,7 @@ opt_util__no_stack_straight_line_2([Instr0 | Instrs0], After0, After, Instrs) :-
 		Instrs = [Instr0 | Instrs0]
 	).
 
-opt_util__lval_refers_stackvars(reg(_), no).
+opt_util__lval_refers_stackvars(reg(_, _), no).
 opt_util__lval_refers_stackvars(stackvar(_), yes).
 opt_util__lval_refers_stackvars(framevar(_), yes).
 opt_util__lval_refers_stackvars(succip, no).
@@ -677,7 +670,17 @@ opt_util__lval_refers_stackvars(field(_, Rval, FieldNum), Refers) :-
 	bool__or(Refers1, Refers2, Refers).
 opt_util__lval_refers_stackvars(lvar(_), _) :-
 	error("found lvar in lval_refers_stackvars").
-opt_util__lval_refers_stackvars(temp(_), no).
+opt_util__lval_refers_stackvars(temp(_, _), no).
+opt_util__lval_refers_stackvars(mem_ref(Rval), Refers) :-
+	opt_util__rval_refers_stackvars(Rval, Refers).
+
+:- pred opt_util__mem_ref_refers_stackvars(mem_ref, bool).
+:- mode opt_util__mem_ref_refers_stackvars(in, out) is det.
+
+opt_util__mem_ref_refers_stackvars(stackvar_ref(_), yes).
+opt_util__mem_ref_refers_stackvars(framevar_ref(_), yes).
+opt_util__mem_ref_refers_stackvars(heap_ref(Rval, _, _), Refers) :-
+	opt_util__rval_refers_stackvars(Rval, Refers).
 
 opt_util__rval_refers_stackvars(lval(Lval), Refers) :-
 	opt_util__lval_refers_stackvars(Lval, Refers).
@@ -685,15 +688,17 @@ opt_util__rval_refers_stackvars(var(_), _) :-
 	error("found var in rval_refers_stackvars").
 opt_util__rval_refers_stackvars(create(_, Rvals, _, _), Refers) :-
 	opt_util__rvals_refer_stackvars(Rvals, Refers).
-opt_util__rval_refers_stackvars(mkword(_, Baserval), Refers) :-
-	opt_util__rval_refers_stackvars(Baserval, Refers).
+opt_util__rval_refers_stackvars(mkword(_, Rval), Refers) :-
+	opt_util__rval_refers_stackvars(Rval, Refers).
 opt_util__rval_refers_stackvars(const(_), no).
-opt_util__rval_refers_stackvars(unop(_, Baserval), Refers) :-
-	opt_util__rval_refers_stackvars(Baserval, Refers).
-opt_util__rval_refers_stackvars(binop(_, Baserval1, Baserval2), Refers) :-
-	opt_util__rval_refers_stackvars(Baserval1, Refers1),
-	opt_util__rval_refers_stackvars(Baserval2, Refers2),
+opt_util__rval_refers_stackvars(unop(_, Rval), Refers) :-
+	opt_util__rval_refers_stackvars(Rval, Refers).
+opt_util__rval_refers_stackvars(binop(_, Rval1, Rval2), Refers) :-
+	opt_util__rval_refers_stackvars(Rval1, Refers1),
+	opt_util__rval_refers_stackvars(Rval2, Refers2),
 	bool__or(Refers1, Refers2, Refers).
+opt_util__rval_refers_stackvars(mem_addr(MemRef), Refers) :-
+	opt_util__mem_ref_refers_stackvars(MemRef, Refers).
 
 opt_util__rvals_refer_stackvars([], no).
 opt_util__rvals_refer_stackvars([MaybeRval | Tail], Refers) :-
@@ -849,7 +854,7 @@ opt_util__block_refers_stackvars([Uinstr0 - _ | Instrs0], Need) :-
 			opt_util__block_refers_stackvars(Instrs0, Need)
 		)
 	;
-		Uinstr0 = restore_ticket(Rval),
+		Uinstr0 = reset_ticket(Rval, _Reason),
 		opt_util__rval_refers_stackvars(Rval, Use),
 		( Use = yes ->
 			Need = yes
@@ -859,6 +864,22 @@ opt_util__block_refers_stackvars([Uinstr0 - _ | Instrs0], Need) :-
 	;
 		Uinstr0 = discard_ticket,
 		opt_util__block_refers_stackvars(Instrs0, Need)
+	;
+		Uinstr0 = mark_ticket_stack(Lval),
+		opt_util__lval_refers_stackvars(Lval, Use),
+		( Use = yes ->
+			Need = yes
+		;
+			opt_util__block_refers_stackvars(Instrs0, Need)
+		)
+	;
+		Uinstr0 = discard_tickets_to(Rval),
+		opt_util__rval_refers_stackvars(Rval, Use),
+		( Use = yes ->
+			Need = yes
+		;
+			opt_util__block_refers_stackvars(Instrs0, Need)
+		)
 	;
 		% handled specially
 		Uinstr0 = incr_sp(_, _),
@@ -868,7 +889,7 @@ opt_util__block_refers_stackvars([Uinstr0 - _ | Instrs0], Need) :-
 		Uinstr0 = decr_sp(_),
 		Need = no
 	;
-		Uinstr0 = pragma_c(_, _, _, _),
+		Uinstr0 = pragma_c(_, _, _, _, _),
 		Need = no
 	).
 
@@ -934,20 +955,6 @@ opt_util__is_const_condition(binop(Op, Rval1, Rval2), Taken) :-
 	Rval1 = Rval2,
 	Taken = yes.
 
-% opt_util__has_incr_sp([Instr0 | Instrs0], Inc) :-
-% 	( Instr0 = incr_sp(N) - _ ->
-% 		Inc = N
-% 	;
-% 		opt_util__has_incr_sp(Instrs0, Inc)
-% 	).
-
-% opt_util__has_decr_sp([Instr0 | Instrs0], Dec) :-
-% 	( Instr0 = decr_sp(N) - _ ->
-% 		Dec = N
-% 	;
-% 		opt_util__has_decr_sp(Instrs0, Dec)
-% 	).
-
 opt_util__new_label_no([], N, N).
 opt_util__new_label_no([Instr0 | Instrs0], N0, N) :-
 	(
@@ -976,11 +983,13 @@ opt_util__can_instr_branch_away(incr_hp(_, _, _), no).
 opt_util__can_instr_branch_away(mark_hp(_), no).
 opt_util__can_instr_branch_away(restore_hp(_), no).
 opt_util__can_instr_branch_away(store_ticket(_), no).
-opt_util__can_instr_branch_away(restore_ticket(_), no).
+opt_util__can_instr_branch_away(reset_ticket(_, _), no).
 opt_util__can_instr_branch_away(discard_ticket, no).
+opt_util__can_instr_branch_away(mark_ticket_stack(_), no).
+opt_util__can_instr_branch_away(discard_tickets_to(_), no).
 opt_util__can_instr_branch_away(incr_sp(_, _), no).
 opt_util__can_instr_branch_away(decr_sp(_), no).
-opt_util__can_instr_branch_away(pragma_c(_, _, _, _), no).
+opt_util__can_instr_branch_away(pragma_c(_, _, _, _, _), no).
 
 opt_util__can_instr_fall_through(comment(_), yes).
 opt_util__can_instr_fall_through(livevals(_), yes).
@@ -999,11 +1008,13 @@ opt_util__can_instr_fall_through(incr_hp(_, _, _), yes).
 opt_util__can_instr_fall_through(mark_hp(_), yes).
 opt_util__can_instr_fall_through(restore_hp(_), yes).
 opt_util__can_instr_fall_through(store_ticket(_), yes).
-opt_util__can_instr_fall_through(restore_ticket(_), yes).
+opt_util__can_instr_fall_through(reset_ticket(_, _), yes).
 opt_util__can_instr_fall_through(discard_ticket, yes).
+opt_util__can_instr_fall_through(mark_ticket_stack(_), yes).
+opt_util__can_instr_fall_through(discard_tickets_to(_), yes).
 opt_util__can_instr_fall_through(incr_sp(_, _), yes).
 opt_util__can_instr_fall_through(decr_sp(_), yes).
-opt_util__can_instr_fall_through(pragma_c(_, _, _, _), yes).
+opt_util__can_instr_fall_through(pragma_c(_, _, _, _, _), yes).
 
 	% Check whether an instruction sequence can possibly fall through
 	% to the next instruction without using its label.
@@ -1038,11 +1049,13 @@ opt_util__can_use_livevals(incr_hp(_, _, _), no).
 opt_util__can_use_livevals(mark_hp(_), no).
 opt_util__can_use_livevals(restore_hp(_), no).
 opt_util__can_use_livevals(store_ticket(_), no).
-opt_util__can_use_livevals(restore_ticket(_), no).
+opt_util__can_use_livevals(reset_ticket(_, _), no).
 opt_util__can_use_livevals(discard_ticket, no).
+opt_util__can_use_livevals(mark_ticket_stack(_), no).
+opt_util__can_use_livevals(discard_tickets_to(_), no).
 opt_util__can_use_livevals(incr_sp(_, _), no).
 opt_util__can_use_livevals(decr_sp(_), no).
-opt_util__can_use_livevals(pragma_c(_, _, _, _), no).
+opt_util__can_use_livevals(pragma_c(_, _, _, _, _), no).
 
 % determine all the labels and code_addresses that are referenced by Instr
 
@@ -1094,11 +1107,13 @@ opt_util__instr_labels_2(incr_hp(_, _, _), [], []).
 opt_util__instr_labels_2(mark_hp(_), [], []).
 opt_util__instr_labels_2(restore_hp(_), [], []).
 opt_util__instr_labels_2(store_ticket(_), [], []).
-opt_util__instr_labels_2(restore_ticket(_), [], []).
+opt_util__instr_labels_2(reset_ticket(_, _), [], []).
 opt_util__instr_labels_2(discard_ticket, [], []).
+opt_util__instr_labels_2(mark_ticket_stack(_), [], []).
+opt_util__instr_labels_2(discard_tickets_to(_), [], []).
 opt_util__instr_labels_2(incr_sp(_, _), [], []).
 opt_util__instr_labels_2(decr_sp(_), [], []).
-opt_util__instr_labels_2(pragma_c(_, _, _, _), [], []).
+opt_util__instr_labels_2(pragma_c(_, _, _, _, _), [], []).
 
 :- pred opt_util__instr_rvals_and_lvals(instr, list(rval), list(lval)).
 :- mode opt_util__instr_rvals_and_lvals(in, out, out) is det.
@@ -1122,11 +1137,13 @@ opt_util__instr_rvals_and_lvals(incr_hp(Lval, _, Rval), [Rval], [Lval]).
 opt_util__instr_rvals_and_lvals(mark_hp(Lval), [], [Lval]).
 opt_util__instr_rvals_and_lvals(restore_hp(Rval), [Rval], []).
 opt_util__instr_rvals_and_lvals(store_ticket(Lval), [], [Lval]).
-opt_util__instr_rvals_and_lvals(restore_ticket(Rval), [Rval], []).
+opt_util__instr_rvals_and_lvals(reset_ticket(Rval, _Reason), [Rval], []).
 opt_util__instr_rvals_and_lvals(discard_ticket, [], []).
+opt_util__instr_rvals_and_lvals(mark_ticket_stack(Lval), [], [Lval]).
+opt_util__instr_rvals_and_lvals(discard_tickets_to(Rval), [Rval], []).
 opt_util__instr_rvals_and_lvals(incr_sp(_, _), [], []).
 opt_util__instr_rvals_and_lvals(decr_sp(_), [], []).
-opt_util__instr_rvals_and_lvals(pragma_c(_, In, _, Out), Rvals, Lvals) :-
+opt_util__instr_rvals_and_lvals(pragma_c(_, In, _, Out, _), Rvals, Lvals) :-
 	pragma_c_inputs_get_rvals(In, Rvals),
 	pragma_c_outputs_get_lvals(Out, Lvals).
 
@@ -1182,6 +1199,7 @@ opt_util__livevals_addr(do_fail, no).
 opt_util__livevals_addr(do_det_closure, yes).
 opt_util__livevals_addr(do_semidet_closure, yes).
 opt_util__livevals_addr(do_nondet_closure, yes).
+opt_util__livevals_addr(do_not_reached, no).
 
 opt_util__count_temps_instr_list([], R, R, F, F).
 opt_util__count_temps_instr_list([Uinstr - _Comment | Instrs], R0, R, F0, F) :-
@@ -1213,24 +1231,28 @@ opt_util__count_temps_instr(restore_hp(Rval), R0, R, F0, F) :-
 	opt_util__count_temps_rval(Rval, R0, R, F0, F).
 opt_util__count_temps_instr(store_ticket(Lval), R0, R, F0, F) :-
 	opt_util__count_temps_lval(Lval, R0, R, F0, F).
-opt_util__count_temps_instr(restore_ticket(Rval), R0, R, F0, F) :-
+opt_util__count_temps_instr(reset_ticket(Rval, _Reason), R0, R, F0, F) :-
 	opt_util__count_temps_rval(Rval, R0, R, F0, F).
 opt_util__count_temps_instr(discard_ticket, R, R, F, F).
+opt_util__count_temps_instr(mark_ticket_stack(Lval), R0, R, F0, F) :-
+	opt_util__count_temps_lval(Lval, R0, R, F0, F).
+opt_util__count_temps_instr(discard_tickets_to(Rval), R0, R, F0, F) :-
+	opt_util__count_temps_rval(Rval, R0, R, F0, F).
 opt_util__count_temps_instr(incr_sp(_, _), R, R, F, F).
 opt_util__count_temps_instr(decr_sp(_), R, R, F, F).
-opt_util__count_temps_instr(pragma_c(_, _, _, _), R, R, F, F).
+opt_util__count_temps_instr(pragma_c(_, _, _, _, _), R, R, F, F).
 
 :- pred opt_util__count_temps_lval(lval, int, int, int, int).
 :- mode opt_util__count_temps_lval(in, in, out, in, out) is det.
 
 opt_util__count_temps_lval(Lval, R0, R, F0, F) :-
-	( Lval = temp(Temp) ->
+	( Lval = temp(Type, N) ->
 		(
-			Temp = r(N),
+			Type = r,
 			int__max(R0, N, R),
 			F = F0
 		;
-			Temp = f(N),
+			Type = f,
 			int__max(F0, N, F),
 			R = R0
 		)
@@ -1261,16 +1283,17 @@ opt_util__format_label(exported(ProcLabel), Str) :-
 :- pred opt_util__format_proclabel(proc_label, string).
 :- mode opt_util__format_proclabel(in, out) is det.
 
-opt_util__format_proclabel(proc(_Module, _PredOrFunc, Name, Arity, Mode), Str)
-		:-
-	string__int_to_string(Arity, ArityStr),
-	string__int_to_string(Mode, ModeStr),
-	string__append_list([Name, "/", ArityStr, " mode ", ModeStr], Str).
-opt_util__format_proclabel(special_proc(_Module, Pred, Type0, Arity, Mode),
+opt_util__format_proclabel(proc(_Module, _PredOrFunc, _, Name, Arity, ProcId),
 		Str) :-
 	string__int_to_string(Arity, ArityStr),
+	proc_id_to_int(ProcId, Mode),
+	string__int_to_string(Mode, ModeStr),
+	string__append_list([Name, "/", ArityStr, " mode ", ModeStr], Str).
+opt_util__format_proclabel(special_proc(_Module, Pred, _, Type, Arity, ProcId),
+		Str) :-
+	string__int_to_string(Arity, ArityStr),
+	proc_id_to_int(ProcId, Mode),
 	string__int_to_string(Mode, ModeStr),
-	llds_out__sym_name_mangle(Type0, Type),
 	string__append_list(
 		[Pred, "_", Type, "/", ArityStr, " mode ", ModeStr], Str).
 
@@ -1331,7 +1354,7 @@ opt_util__touches_nondet_ctrl_instr(Uinstr, Touch) :-
 :- pred opt_util__touches_nondet_ctrl_lval(lval, bool).
 :- mode opt_util__touches_nondet_ctrl_lval(in, out) is det.
 
-opt_util__touches_nondet_ctrl_lval(reg(_), no).
+opt_util__touches_nondet_ctrl_lval(reg(_, _), no).
 opt_util__touches_nondet_ctrl_lval(stackvar(_), no).
 opt_util__touches_nondet_ctrl_lval(framevar(_), no).
 opt_util__touches_nondet_ctrl_lval(succip, no).
@@ -1348,7 +1371,9 @@ opt_util__touches_nondet_ctrl_lval(field(_, Rval1, Rval2), Touch) :-
 	opt_util__touches_nondet_ctrl_rval(Rval2, Touch2),
 	bool__or(Touch1, Touch2, Touch).
 opt_util__touches_nondet_ctrl_lval(lvar(_), no).
-opt_util__touches_nondet_ctrl_lval(temp(_), no).
+opt_util__touches_nondet_ctrl_lval(temp(_, _), no).
+opt_util__touches_nondet_ctrl_lval(mem_ref(Rval), Touch) :-
+	opt_util__touches_nondet_ctrl_rval(Rval, Touch).
 
 :- pred opt_util__touches_nondet_ctrl_rval(rval, bool).
 :- mode opt_util__touches_nondet_ctrl_rval(in, out) is det.
@@ -1366,10 +1391,20 @@ opt_util__touches_nondet_ctrl_rval(binop(_, Rval1, Rval2), Touch) :-
 	opt_util__touches_nondet_ctrl_rval(Rval1, Touch1),
 	opt_util__touches_nondet_ctrl_rval(Rval2, Touch2),
 	bool__or(Touch1, Touch2, Touch).
+opt_util__touches_nondet_ctrl_rval(mem_addr(MemRef), Touch) :-
+	opt_util__touches_nondet_ctrl_mem_ref(MemRef, Touch).
+
+:- pred opt_util__touches_nondet_ctrl_mem_ref(mem_ref, bool).
+:- mode opt_util__touches_nondet_ctrl_mem_ref(in, out) is det.
+
+opt_util__touches_nondet_ctrl_mem_ref(stackvar_ref(_), no).
+opt_util__touches_nondet_ctrl_mem_ref(framevar_ref(_), no).
+opt_util__touches_nondet_ctrl_mem_ref(heap_ref(Rval, _, _), Touch) :-
+	opt_util__touches_nondet_ctrl_rval(Rval, Touch).
 
 %-----------------------------------------------------------------------------%
 
-opt_util__lval_access_rvals(reg(_), []).
+opt_util__lval_access_rvals(reg(_, _), []).
 opt_util__lval_access_rvals(stackvar(_), []).
 opt_util__lval_access_rvals(framevar(_), []).
 opt_util__lval_access_rvals(succip, []).
@@ -1382,9 +1417,10 @@ opt_util__lval_access_rvals(succfr(Rval), [Rval]).
 opt_util__lval_access_rvals(hp, []).
 opt_util__lval_access_rvals(sp, []).
 opt_util__lval_access_rvals(field(_, Rval1, Rval2), [Rval1, Rval2]).
-opt_util__lval_access_rvals(temp(_), []).
+opt_util__lval_access_rvals(temp(_, _), []).
 opt_util__lval_access_rvals(lvar(_), _) :-
 	error("lvar detected in opt_util__lval_access_rvals").
+opt_util__lval_access_rvals(mem_ref(Rval), [Rval]).
 
 %-----------------------------------------------------------------------------%
 
@@ -1411,7 +1447,7 @@ opt_util__rval_free_of_lval(binop(_, Rval1, Rval2), Forbidden) :-
 
 %-----------------------------------------------------------------------------%
 
-opt_util__lvals_in_lval(reg(_), []).
+opt_util__lvals_in_lval(reg(_, _), []).
 opt_util__lvals_in_lval(stackvar(_), []).
 opt_util__lvals_in_lval(framevar(_), []).
 opt_util__lvals_in_lval(succip, []).
@@ -1432,7 +1468,9 @@ opt_util__lvals_in_lval(field(_, Rval1, Rval2), Lvals) :-
 	opt_util__lvals_in_rval(Rval2, Lvals2),
 	list__append(Lvals1, Lvals2, Lvals).
 opt_util__lvals_in_lval(lvar(_), []).
-opt_util__lvals_in_lval(temp(_), []).
+opt_util__lvals_in_lval(temp(_, _), []).
+opt_util__lvals_in_lval(mem_ref(Rval), Lvals) :-
+	opt_util__lvals_in_rval(Rval, Lvals).
 
 opt_util__lvals_in_rval(lval(Lval), [Lval | Lvals]) :-
 	opt_util__lvals_in_lval(Lval, Lvals).
@@ -1448,6 +1486,17 @@ opt_util__lvals_in_rval(binop(_, Rval1, Rval2), Lvals) :-
 	opt_util__lvals_in_rval(Rval1, Lvals1),
 	opt_util__lvals_in_rval(Rval2, Lvals2),
 	list__append(Lvals1, Lvals2, Lvals).
+opt_util__lvals_in_rval(mem_addr(MemRef), Lvals) :-
+	opt_util__lvals_in_mem_ref(MemRef, Lvals).
+
+	% XXX
+:- pred opt_util__lvals_in_mem_ref(mem_ref, list(lval)).
+:- mode opt_util__lvals_in_mem_ref(in, out) is det.
+
+opt_util__lvals_in_mem_ref(stackvar_ref(_), []).
+opt_util__lvals_in_mem_ref(framevar_ref(_), []).
+opt_util__lvals_in_mem_ref(heap_ref(Rval, _, _), Lvals) :-
+	opt_util__lvals_in_rval(Rval, Lvals).
 
 %-----------------------------------------------------------------------------%
 
