@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1995 University of Melbourne.
+% Copyright (C) 1994-1997 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -34,19 +34,22 @@
 
 :- interface.
 
-:- import_module hlds_goal.
-:- import_module list, set, term.
+:- import_module hlds_goal, hlds_pred, prog_data.
+:- import_module list, set, term, varset.
 
 :- pred implicitly_quantify_clause_body(list(var),
-		hlds__goal, varset, map(var, type),
-		hlds__goal, varset, map(var, type), list(quant_warning)).
+		hlds_goal, varset, map(var, type),
+		hlds_goal, varset, map(var, type), list(quant_warning)).
 :- mode implicitly_quantify_clause_body(in, in, in, in, out, out, out, out)
 	is det.
 
-:- pred implicitly_quantify_goal(hlds__goal, varset, map(var, type), set(var),
-		hlds__goal, varset, map(var, type), list(quant_warning)).
+:- pred implicitly_quantify_goal(hlds_goal, varset, map(var, type), set(var),
+		hlds_goal, varset, map(var, type), list(quant_warning)).
 :- mode implicitly_quantify_goal(in, in, in, in, out, out, out, out) is det.
 
+:- pred requantify_proc(proc_info, proc_info) is det.
+:- mode requantify_proc(in, out) is det.
+
 	% We return a list of warnings back to make_hlds.m.
 	% Currently the only thing we warn about is variables with
 	% overlapping scopes.
@@ -54,11 +57,11 @@
 :- type quant_warning
 	--->	warn_overlap(list(var), term__context).
 
-	% goal_vars(Goal, Vars):
+	% quantification__goal_vars(Goal, Vars):
 	%	Vars is the set of variables that are free (unquantified)
 	%	in Goal.
-:- pred goal_vars(hlds__goal, set(var)).
-:- mode goal_vars(in, out) is det.
+:- pred quantification__goal_vars(hlds_goal, set(var)).
+:- mode quantification__goal_vars(in, out) is det.
 
 %-----------------------------------------------------------------------------%
 
@@ -91,10 +94,11 @@
 	% `QuantVars' are the variables not in `OutsideVars' 
 	% that have been explicitly existentially quantified over a scope
 	% which includes the current goal in a positive (non-negated) context.
-	% `OutsideLambdaVars' are the variables
-	% that have occurred free in a lambda
-	% expression outside this goal, not counting occurrences in
-	% parallel goals.
+	% `OutsideLambdaVars' are the variables that have occurred free in
+	% a lambda expression outside this goal, not counting occurrences in
+	% parallel goals (and if this goal is itself inside a lambda
+	% expression, not counting occurrences outside that lambda
+	% expression).
 	%
 	% For example, for
 	%
@@ -117,6 +121,17 @@ implicitly_quantify_clause_body(HeadVars, Goal0, Varset0, VarTypes0,
 	implicitly_quantify_goal(Goal0, Varset0, VarTypes0,
 			OutsideVars, Goal, Varset, VarTypes, Warnings).
 
+requantify_proc(ProcInfo0, ProcInfo) :-
+	proc_info_variables(ProcInfo0, Varset0),
+	proc_info_vartypes(ProcInfo0, VarTypes0),
+	proc_info_headvars(ProcInfo0, HeadVars),
+	proc_info_goal(ProcInfo0, Goal0),
+	implicitly_quantify_clause_body(HeadVars, Goal0, Varset0, VarTypes0,
+		Goal, Varset, VarTypes, _),
+	proc_info_set_variables(ProcInfo0, Varset, ProcInfo1),
+	proc_info_set_vartypes(ProcInfo1, VarTypes, ProcInfo2),
+	proc_info_set_goal(ProcInfo2, Goal, ProcInfo).
+
 implicitly_quantify_goal(Goal0, Varset0, VarTypes0, OutsideVars,
 				Goal, Varset, VarTypes, Warnings) :-
 	quantification__init(OutsideVars, Varset0, VarTypes0, QuantInfo0),
@@ -126,7 +141,7 @@ implicitly_quantify_goal(Goal0, Varset0, VarTypes0, OutsideVars,
 	quantification__get_warnings(Warnings0, QuantInfo, _),
 	list__reverse(Warnings0, Warnings).
 
-:- pred implicitly_quantify_goal(hlds__goal, hlds__goal,
+:- pred implicitly_quantify_goal(hlds_goal, hlds_goal,
 					quant_info, quant_info).
 :- mode implicitly_quantify_goal(in, out, in, out) is det.
 
@@ -139,7 +154,7 @@ implicitly_quantify_goal(Goal0 - GoalInfo0, Goal - GoalInfo) -->
 		% If there are any variables that are local to the goal
 		% which we have come across before, then we rename them
 		% apart.
-		{ goal_vars(Goal0 - GoalInfo0, GoalVars0) },
+		{ quantification__goal_vars(Goal0 - GoalInfo0, GoalVars0) },
 		{ set__difference(GoalVars0, NonLocalVars, LocalVars) },
 		{ set__intersect(SeenVars, LocalVars, RenameVars) },
 		{ \+ set__empty(RenameVars) }
@@ -152,8 +167,8 @@ implicitly_quantify_goal(Goal0 - GoalInfo0, Goal - GoalInfo) -->
 	),
 	{ goal_info_set_nonlocals(GoalInfo1, NonLocalVars, GoalInfo) }.
 
-:- pred implicitly_quantify_goal_2(hlds__goal_expr, term__context,
-				hlds__goal_expr, quant_info, quant_info).
+:- pred implicitly_quantify_goal_2(hlds_goal_expr, term__context,
+				hlds_goal_expr, quant_info, quant_info).
 :- mode implicitly_quantify_goal_2(in, in, out, in, out) is det.
 
 	% we retain explicit existential quantifiers in the source code,
@@ -249,7 +264,7 @@ implicitly_quantify_goal_2(if_then_else(Vars0, Cond0, Then0, Else0, SM),
 		{ goal_util__rename_var_list(Vars0, no, RenameMap, Vars) }
 	),
 	{ set__insert_list(QuantVars, Vars, QuantVars1) },
-	{ goal_vars(Then1, VarsThen, LambdaVarsThen) },
+	{ quantification__goal_vars(Then1, VarsThen, LambdaVarsThen) },
 	{ set__union(OutsideVars, VarsThen, OutsideVars1) },
 	{ set__union(LambdaOutsideVars, LambdaVarsThen, LambdaOutsideVars1) },
 	quantification__set_quant_vars(QuantVars1),
@@ -269,32 +284,37 @@ implicitly_quantify_goal_2(if_then_else(Vars0, Cond0, Then0, Else0, SM),
 	quantification__get_nonlocals(NonLocalsElse),
 	{ set__union(NonLocalsCond, NonLocalsThen, NonLocalsIfThen) },
 	{ set__union(NonLocalsIfThen, NonLocalsElse, NonLocalsIfThenElse) },
-	{ set__intersect(NonLocalsIfThenElse, OutsideVars, NonLocals) },
+	{ set__intersect(NonLocalsIfThenElse, OutsideVars, NonLocalsO) },
+	{ set__intersect(NonLocalsIfThenElse, LambdaOutsideVars, NonLocalsL) },
+	{ set__union(NonLocalsO, NonLocalsL, NonLocals) },
 	quantification__set_nonlocals(NonLocals).
 
 implicitly_quantify_goal_2(call(A, B, HeadVars, D, E, F), _,
 		call(A, B, HeadVars, D, E, F)) -->
 	implicitly_quantify_atomic_goal(HeadVars).
 
-implicitly_quantify_goal_2(higher_order_call(PredVar, ArgVars, C, D, E), _,
-		higher_order_call(PredVar, ArgVars, C, D, E)) -->
+implicitly_quantify_goal_2(higher_order_call(PredVar, ArgVars, C, D, E, F), _,
+		higher_order_call(PredVar, ArgVars, C, D, E, F)) -->
 	implicitly_quantify_atomic_goal([PredVar|ArgVars]).
 
-implicitly_quantify_goal_2(unify(A, B0, X, Y, Z), Context,
-		unify(A, B, X, Y, Z)) -->
+implicitly_quantify_goal_2(
+		unify(Var, UnifyRHS0, Mode, Unification0, UnifyContext),
+		Context,
+		unify(Var, UnifyRHS, Mode, Unification, UnifyContext)) -->
 	quantification__get_outside(OutsideVars),
 	quantification__get_lambda_outside(LambdaOutsideVars),
-	implicitly_quantify_unify_rhs(B0, Context, B),
-	quantification__get_nonlocals(VarsB),
-	{ set__insert(VarsB, A, GoalVars) },
+	implicitly_quantify_unify_rhs(UnifyRHS0, Unification0, Context,
+		UnifyRHS, Unification),
+	quantification__get_nonlocals(VarsUnifyRHS),
+	{ set__insert(VarsUnifyRHS, Var, GoalVars) },
 	quantification__update_seen_vars(GoalVars),
 	{ set__intersect(GoalVars, OutsideVars, NonLocalVars1) },
 	{ set__intersect(GoalVars, LambdaOutsideVars, NonLocalVars2) },
 	{ set__union(NonLocalVars1, NonLocalVars2, NonLocalVars) },
 	quantification__set_nonlocals(NonLocalVars).
 
-implicitly_quantify_goal_2(pragma_c_code(A,B,C,D,Vars,F), _,
-		pragma_c_code(A,B,C,D,Vars,F)) --> 
+implicitly_quantify_goal_2(pragma_c_code(A,B,C,D,Vars,F,G,H), _,
+		pragma_c_code(A,B,C,D,Vars,F,G,H)) --> 
 	implicitly_quantify_atomic_goal(Vars).
 
 :- pred implicitly_quantify_atomic_goal(list(var), quant_info, quant_info).
@@ -310,21 +330,25 @@ implicitly_quantify_atomic_goal(HeadVars) -->
 	{ set__union(NonLocals1, NonLocals2, NonLocals) },
 	quantification__set_nonlocals(NonLocals).
 
-:- pred implicitly_quantify_unify_rhs(unify_rhs, term__context, unify_rhs,
+:- pred implicitly_quantify_unify_rhs(unify_rhs, unification, term__context,
+					unify_rhs, unification,
 					quant_info, quant_info).
-:- mode implicitly_quantify_unify_rhs(in, in, out, in, out) is det.
+:- mode implicitly_quantify_unify_rhs(in, in, in, out, out, in, out) is det.
 
-implicitly_quantify_unify_rhs(var(X), _, var(X)) -->
+implicitly_quantify_unify_rhs(var(X), Unification, _, var(X), Unification) -->
 	{ set__singleton_set(Vars, X) },
 	quantification__set_nonlocals(Vars).
-implicitly_quantify_unify_rhs(functor(Functor, ArgVars), _,
-				functor(Functor, ArgVars)) -->
+implicitly_quantify_unify_rhs(functor(Functor, ArgVars), Unification, _,
+				functor(Functor, ArgVars), Unification) -->
 	{ set__list_to_set(ArgVars, Vars) },
 	quantification__set_nonlocals(Vars).
 implicitly_quantify_unify_rhs(
 		lambda_goal(PredOrFunc, LambdaVars0, Modes, Det, Goal0),
+		Unification0,
 		Context,
-		lambda_goal(PredOrFunc, LambdaVars, Modes, Det, Goal)) -->
+		lambda_goal(PredOrFunc, LambdaVars, Modes, Det, Goal),
+		Unification
+		) -->
 
 	quantification__get_outside(OutsideVars0),
 	{ set__list_to_set(LambdaVars0, QVars) },
@@ -353,21 +377,53 @@ implicitly_quantify_unify_rhs(
 		% and initialize the new quantified vars set to be empty.
 	quantification__get_quant_vars(QuantVars0),
 	{ set__union(OutsideVars0, QuantVars0, OutsideVars1) },
+	{ set__init(QuantVars) },
+	quantification__set_quant_vars(QuantVars),
 		% Add the lambda vars as outside vars, since they are
 		% outside of the lambda goal
 	{ set__insert_list(OutsideVars1, LambdaVars, OutsideVars) },
-	{ set__init(QuantVars) },
 	quantification__set_outside(OutsideVars),
-	quantification__set_quant_vars(QuantVars),
+		% Set the LambdaOutsideVars set to empty, because
+		% variables that occur outside this lambda expression
+		% only in other lambda expressions should not be
+		% considered non-local.
+	quantification__get_lambda_outside(LambdaOutsideVars0),
+	{ set__init(LambdaOutsideVars) },
+	quantification__set_lambda_outside(LambdaOutsideVars),
 	implicitly_quantify_goal(Goal1, Goal),
 	quantification__get_nonlocals(NonLocals0),
 		% lambda-quantified variables are local
 	{ set__delete_list(NonLocals0, LambdaVars, NonLocals) },
 	quantification__set_quant_vars(QuantVars0),
 	quantification__set_outside(OutsideVars0),
-	quantification__set_nonlocals(NonLocals).
+	quantification__set_lambda_outside(LambdaOutsideVars0),
+	quantification__set_nonlocals(NonLocals),
 
-:- pred implicitly_quantify_conj(list(hlds__goal), list(hlds__goal), 
+	%
+	% For a unification that constructs a lambda expression,
+	% the argument variables of the construction are the non-local
+	% variables of the lambda expression.  So if we recompute the
+	% non-locals, we need to recompute the argument variables of
+	% the construction, and hence we also need to recompute their modes.
+	% The non-locals set must only ever decrease, not increase,
+	% so we can just use the old modes.
+	%
+	{
+		Unification0 = construct(ConstructVar, ConsId, Args0, ArgModes0)
+	->
+		map__from_corresponding_lists(Args0, ArgModes0, ArgModesMap),
+		set__to_sorted_list(NonLocals, Args),
+		map__apply_to_list(Args, ArgModesMap, ArgModes),
+		Unification = construct(ConstructVar, ConsId, Args, ArgModes)
+	;
+		% after mode analysis, unifications with lambda variables
+		% should always be construction unifications, but
+		% quantification gets invoked before mode analysis,
+		% so we need to allow this case...
+		Unification = Unification0
+	}.
+
+:- pred implicitly_quantify_conj(list(hlds_goal), list(hlds_goal), 
 					quant_info, quant_info).
 :- mode implicitly_quantify_conj(in, out, in, out) is det.
 
@@ -375,8 +431,8 @@ implicitly_quantify_conj(Goals0, Goals) -->
 	{ get_vars(Goals0, FollowingVarsList) },
 	implicitly_quantify_conj_2(Goals0, FollowingVarsList, Goals).
 
-:- pred implicitly_quantify_conj_2(list(hlds__goal), list(pair(set(var))),
-			list(hlds__goal), quant_info, quant_info).
+:- pred implicitly_quantify_conj_2(list(hlds_goal), list(pair(set(var))),
+			list(hlds_goal), quant_info, quant_info).
 :- mode implicitly_quantify_conj_2(in, in, out, in, out) is det.
 
 implicitly_quantify_conj_2([], _, []) -->
@@ -403,11 +459,13 @@ implicitly_quantify_conj_2([Goal0 | Goals0],
 				Goals),
 	quantification__get_nonlocals(NonLocalVars2),
 	{ set__union(NonLocalVars1, NonLocalVars2, NonLocalVarsConj) },
-	{ set__intersect(NonLocalVarsConj, OutsideVars, NonLocalVars) },
+	{ set__intersect(NonLocalVarsConj, OutsideVars, NonLocalVarsO) },
+	{ set__intersect(NonLocalVarsConj, LambdaOutsideVars, NonLocalVarsL) },
+	{ set__union(NonLocalVarsO, NonLocalVarsL, NonLocalVars) },
 	quantification__set_outside(OutsideVars),
 	quantification__set_nonlocals(NonLocalVars).
 
-:- pred implicitly_quantify_disj(list(hlds__goal), list(hlds__goal), 
+:- pred implicitly_quantify_disj(list(hlds_goal), list(hlds_goal), 
 					quant_info, quant_info).
 :- mode implicitly_quantify_disj(in, out, in, out) is det.
 
@@ -461,14 +519,14 @@ quantification__update_seen_vars(NewVars) -->
 	% and the second contains following variables that
 	% occur in lambda goals.
 
-:- pred get_vars(list(hlds__goal), list(pair(set(var)))).
+:- pred get_vars(list(hlds_goal), list(pair(set(var)))).
 :- mode get_vars(in, out) is det.
 
 get_vars([], []).
 get_vars([_Goal | Goals], [Set - LambdaSet | SetPairs]) :-
 	get_vars_2(Goals, Set, LambdaSet, SetPairs).
 
-:- pred get_vars_2(list(hlds__goal), set(var), set(var), list(pair(set(var)))).
+:- pred get_vars_2(list(hlds_goal), set(var), set(var), list(pair(set(var)))).
 :- mode get_vars_2(in, out, out, out) is det.
 
 get_vars_2([], Set, LambdaSet, []) :-
@@ -476,18 +534,18 @@ get_vars_2([], Set, LambdaSet, []) :-
 	set__init(LambdaSet).
 get_vars_2([Goal | Goals], Set, LambdaSet, SetPairList) :-
 	get_vars_2(Goals, Set0, LambdaSet0, SetPairList0),
-	goal_vars(Goal, Set1, LambdaSet1),
+	quantification__goal_vars(Goal, Set1, LambdaSet1),
 	set__union(Set0, Set1, Set),
 	set__union(LambdaSet0, LambdaSet1, LambdaSet),
 	SetPairList = [Set0 - LambdaSet0 | SetPairList0].
 
-:- pred goal_list_vars_2(list(hlds__goal), set(var), set(var),
+:- pred goal_list_vars_2(list(hlds_goal), set(var), set(var),
 			set(var), set(var)).
 :- mode goal_list_vars_2(in, in, in, out, out) is det.
 
 goal_list_vars_2([], Set, LambdaSet, Set, LambdaSet).
 goal_list_vars_2([Goal - _GoalInfo| Goals], Set0, LambdaSet0, Set, LambdaSet) :-
-	goal_vars_2(Goal, Set0, LambdaSet0, Set1, LambdaSet1),
+	quantification__goal_vars_2(Goal, Set0, LambdaSet0, Set1, LambdaSet1),
 	goal_list_vars_2(Goals, Set1, LambdaSet1, Set, LambdaSet).
 
 :- pred case_list_vars_2(list(case), set(var), set(var), set(var), set(var)).
@@ -496,81 +554,88 @@ goal_list_vars_2([Goal - _GoalInfo| Goals], Set0, LambdaSet0, Set, LambdaSet) :-
 case_list_vars_2([], Set, LambdaSet, Set, LambdaSet).
 case_list_vars_2([case(_Cons, Goal - _GoalInfo)| Cases], Set0, LambdaSet0,
 			Set, LambdaSet) :-
-	goal_vars_2(Goal, Set0, LambdaSet0, Set1, LambdaSet1),
+	quantification__goal_vars_2(Goal, Set0, LambdaSet0, Set1, LambdaSet1),
 	case_list_vars_2(Cases, Set1, LambdaSet1, Set, LambdaSet).
 
-	% goal_vars(Goal, Vars):
+	% quantification__goal_vars(Goal, Vars):
 	%	Vars is the set of variables that occur free (unquantified)
 	%	in Goal.
-goal_vars(Goal, BothSet) :-
-	goal_vars(Goal, NonLambdaSet, LambdaSet),
+quantification__goal_vars(Goal, BothSet) :-
+	quantification__goal_vars(Goal, NonLambdaSet, LambdaSet),
 	set__union(NonLambdaSet, LambdaSet, BothSet).
 
-	% goal_vars(Goal, NonLambdaSet, LambdaSet):
+	% quantification__goal_vars(Goal, NonLambdaSet, LambdaSet):
 	%	Set is the set of variables that occur free (unquantified)
 	%	in Goal, not counting occurrences in lambda expressions.
 	%	LambdaSet is the set of variables that occur free (unquantified)
 	%	in lambda expressions in Goal.
-:- pred goal_vars(hlds__goal, set(var), set(var)).
-:- mode goal_vars(in, out, out) is det.
+:- pred quantification__goal_vars(hlds_goal, set(var), set(var)).
+:- mode quantification__goal_vars(in, out, out) is det.
 
-goal_vars(Goal - _GoalInfo, Set, LambdaSet) :-
+quantification__goal_vars(Goal - _GoalInfo, Set, LambdaSet) :-
 	set__init(Set0),
 	set__init(LambdaSet0),
-	goal_vars_2(Goal, Set0, LambdaSet0, Set, LambdaSet).
+	quantification__goal_vars_2(Goal, Set0, LambdaSet0, Set, LambdaSet).
 
-:- pred goal_vars_2(hlds__goal_expr, set(var), set(var), set(var), set(var)).
-:- mode goal_vars_2(in, in, in, out, out) is det.
+:- pred quantification__goal_vars_2(hlds_goal_expr, set(var), set(var),
+		set(var), set(var)).
+:- mode quantification__goal_vars_2(in, in, in, out, out) is det.
 
-goal_vars_2(unify(A, B, _, _, _), Set0, LambdaSet0, Set, LambdaSet) :-
+quantification__goal_vars_2(unify(A, B, _, _, _), Set0, LambdaSet0,
+		Set, LambdaSet) :-
 	set__insert(Set0, A, Set1),
 	quantification__unify_rhs_vars(B, Set1, LambdaSet0, Set, LambdaSet).
 
-goal_vars_2(higher_order_call(PredVar, ArgVars, _, _, _), Set0, LambdaSet,
-			Set, LambdaSet) :-
+quantification__goal_vars_2(higher_order_call(PredVar, ArgVars, _, _, _, _),
+		Set0, LambdaSet, Set, LambdaSet) :-
 	set__insert_list(Set0, [PredVar | ArgVars], Set).
 
-goal_vars_2(call(_, _, ArgVars, _, _, _), Set0, LambdaSet, Set, LambdaSet) :-
+quantification__goal_vars_2(call(_, _, ArgVars, _, _, _), Set0, LambdaSet,
+		Set, LambdaSet) :-
 	set__insert_list(Set0, ArgVars, Set).
 
-goal_vars_2(conj(Goals), Set0, LambdaSet0, Set, LambdaSet) :-
+quantification__goal_vars_2(conj(Goals), Set0, LambdaSet0, Set, LambdaSet) :-
 	goal_list_vars_2(Goals, Set0, LambdaSet0, Set, LambdaSet).
 
-goal_vars_2(disj(Goals, _), Set0, LambdaSet0, Set, LambdaSet) :-
+quantification__goal_vars_2(disj(Goals, _), Set0, LambdaSet0, Set, LambdaSet) :-
 	goal_list_vars_2(Goals, Set0, LambdaSet0, Set, LambdaSet).
 
-goal_vars_2(switch(Var, _Det, Cases, _), Set0, LambdaSet0, Set, LambdaSet) :-
+quantification__goal_vars_2(switch(Var, _Det, Cases, _), Set0, LambdaSet0,
+		Set, LambdaSet) :-
 	set__insert(Set0, Var, Set1),
 	case_list_vars_2(Cases, Set1, LambdaSet0, Set, LambdaSet).
 
-goal_vars_2(some(Vars, Goal), Set0, LambdaSet0, Set, LambdaSet) :-
-	goal_vars(Goal, Set1, LambdaSet1),
+quantification__goal_vars_2(some(Vars, Goal), Set0, LambdaSet0,
+		Set, LambdaSet) :-
+	quantification__goal_vars(Goal, Set1, LambdaSet1),
 	set__delete_list(Set1, Vars, Set2),
 	set__delete_list(LambdaSet1, Vars, LambdaSet2),
 	set__union(Set0, Set2, Set),
 	set__union(LambdaSet0, LambdaSet2, LambdaSet).
 
-goal_vars_2(not(Goal - _GoalInfo), Set0, LambdaSet0, Set, LambdaSet) :-
-	goal_vars_2(Goal, Set0, LambdaSet0, Set, LambdaSet).
+quantification__goal_vars_2(not(Goal - _GoalInfo), Set0, LambdaSet0,
+		Set, LambdaSet) :-
+	quantification__goal_vars_2(Goal, Set0, LambdaSet0, Set, LambdaSet).
 
-goal_vars_2(if_then_else(Vars, A, B, C, _), Set0, LambdaSet0, Set, LambdaSet) :-
-		% This code does the following:
-		%     Set = Set0 + ( (vars(A) + vars(B)) \ Vars ) + vars(C)
-		% where `+' is set union and `\' is relative complement.
-	goal_vars(A, Set1, LambdaSet1),
-	goal_vars(B, Set2, LambdaSet2),
+quantification__goal_vars_2(if_then_else(Vars, A, B, C, _), Set0, LambdaSet0,
+		Set, LambdaSet) :-
+	% This code does the following:
+	%     Set = Set0 + ( (vars(A) + vars(B)) \ Vars ) + vars(C)
+	% where `+' is set union and `\' is relative complement.
+	quantification__goal_vars(A, Set1, LambdaSet1),
+	quantification__goal_vars(B, Set2, LambdaSet2),
 	set__union(Set1, Set2, Set3),
 	set__union(LambdaSet1, LambdaSet2, LambdaSet3),
 	set__delete_list(Set3, Vars, Set4),
 	set__delete_list(LambdaSet3, Vars, LambdaSet4),
 	set__union(Set0, Set4, Set5),
 	set__union(LambdaSet0, LambdaSet4, LambdaSet5),
-	goal_vars(C, Set6, LambdaSet6),
+	quantification__goal_vars(C, Set6, LambdaSet6),
 	set__union(Set5, Set6, Set),
 	set__union(LambdaSet5, LambdaSet6, LambdaSet).
 
-goal_vars_2(pragma_c_code(_, _, _, _, ArgVars, _), LambdaSet, Set0,
-		Set, LambdaSet) :-
+quantification__goal_vars_2(pragma_c_code(_, _, _, _, ArgVars, _, _, _),
+		Set0, LambdaSet, Set, LambdaSet) :-
 	set__insert_list(Set0, ArgVars, Set).
 
 :- pred quantification__unify_rhs_vars(unify_rhs, set(var), set(var),
@@ -584,7 +649,7 @@ quantification__unify_rhs_vars(functor(_Functor, ArgVars), Set0, LambdaSet,
 	set__insert_list(Set0, ArgVars, Set).
 quantification__unify_rhs_vars(lambda_goal(_PredOrFunc, LambdaVars, _Modes,
 		_Detism, Goal), Set, LambdaSet0, Set, LambdaSet) :-
-	goal_vars(Goal, GoalVars),
+	quantification__goal_vars(Goal, GoalVars),
 	set__delete_list(GoalVars, LambdaVars, GoalVars1),
 	set__union(LambdaSet0, GoalVars1, LambdaSet).
 
@@ -608,7 +673,7 @@ quantification__warn_overlapping_scope(OverlapVars, Context) -->
 %	Apply RenameMap to Goal0 giving Goal.
 
 :- pred quantification__rename_apart(set(var), map(var, var),
-				hlds__goal, hlds__goal, quant_info, quant_info).
+				hlds_goal, hlds_goal, quant_info, quant_info).
 :- mode quantification__rename_apart(in, out, in, out, in, out) is det.
 
 quantification__rename_apart(RenameSet, RenameMap, Goal0, Goal) -->
