@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1995 University of Melbourne.
+% Copyright (C) 1996-1997 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 %  Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -56,7 +56,7 @@
 :- import_module hlds_module, type_util, mode_util, prog_data, passes_aux.
 :- import_module prog_out, globals, options.
 
-:- import_module map, list, set, bool, std_util, relation, require.
+:- import_module assoc_list, map, list, set, bool, std_util, relation, require.
 
 
 
@@ -64,22 +64,22 @@ stratify__check_stratification(Module0, Module) -->
 	{ module_info_ensure_dependency_info(Module0, Module1) },
 	{ module_info_dependency_info(Module1, DepInfo) },
 	
-	{ hlds__dependency_info_get_dependency_graph(DepInfo, DepGraph0) },
+	{ hlds_dependency_info_get_dependency_graph(DepInfo, DepGraph0) },
 	{ relation__atsort(DepGraph0, FOSCCs1) },	
 	{ dep_sets_to_lists_and_sets(FOSCCs1, [], FOSCCs) },
 	globals__io_lookup_bool_option(warn_non_stratification, Warn),
 	{ module_info_stratified_preds(Module1, StratifiedPreds) },
 	first_order_check_sccs(FOSCCs, StratifiedPreds, Warn, Module1, Module).
 
-%%%%%%%%%%%
-%%% The following code was used for the second pass of this module but
-%%% as that pass is disabled so is this code. The higher order code
-%%% is disabled because it is currently unable to detect cases where a 
-%%% higher order proc is hidden in some complex data structure
-%%%	{ gen_conservative_graph(Module2, DepGraph0, DepGraph, HOInfo) },
-%%%	{ relation__atsort(DepGraph, HOSCCs1) },	
-%%%	{ dep_sets_to_lists_and_sets(HOSCCs1, [], HOSCCs) },
-%%%	higher_order_check_sccs(HOSCCs, HOInfo, Module2, Module).
+	% The following code was used for the second pass of this module but
+	% as that pass is disabled so is this code. The higher order code
+	% is disabled because it is currently unable to detect cases where a 
+	% higher order proc is hidden in some complex data structure
+	%	
+	%{ gen_conservative_graph(Module2, DepGraph0, DepGraph, HOInfo) },
+	%{ relation__atsort(DepGraph, HOSCCs1) },	
+	%{ dep_sets_to_lists_and_sets(HOSCCs1, [], HOSCCs) },
+	%higher_order_check_sccs(HOSCCs, HOInfo, Module2, Module).
 
 
 
@@ -152,7 +152,7 @@ first_order_check_scc_2([PredProcId|Remaining], WholeScc, Error,
 		PredProcId, Error, Module0, Module1),
 	first_order_check_scc_2(Remaining, WholeScc, Error, Module1, Module).
 
-:- pred first_order_check_goal(hlds__goal_expr, hlds__goal_info, bool,
+:- pred first_order_check_goal(hlds_goal_expr, hlds_goal_info, bool,
 		list(pred_proc_id), pred_proc_id, bool,
 		module_info, module_info, io__state, io__state).
 :- mode first_order_check_goal(in, in, in, in, in, in, in, out, di, uo) is det.
@@ -186,7 +186,7 @@ first_order_check_goal(not(Goal - GoalInfo), _GoalInfo, _Negated,
 		WholeScc, ThisPredProcId, Error, Module0, Module) -->
 	first_order_check_goal(Goal, GoalInfo, yes, WholeScc, ThisPredProcId,
 		Error, Module0, Module).
-first_order_check_goal(pragma_c_code(_, _IsRec, CPred, CProc, _, _), 
+first_order_check_goal(pragma_c_code(_, _IsRec, CPred, CProc, _, _, _, _), 
 		GoalInfo, Negated, WholeScc, ThisPredProcId, 
 		Error, Module0, Module) -->
 	(
@@ -202,7 +202,7 @@ first_order_check_goal(pragma_c_code(_, _IsRec, CPred, CProc, _, _),
 	).
 first_order_check_goal(unify(_Var, _RHS, _Mode, _Uni, _Context), _GoalInfo,
 		_Negated, _WholeScc, _ThisPredProcId, _, Module, Module) --> [].
-first_order_check_goal(call(CPred, CProc, _Args, Builtin, _Contex, _Sym), 
+first_order_check_goal(call(CPred, CProc, _Args, BuiltinState, _Contex, _Sym), 
 		GInfo, Negated, WholeScc, ThisPredProcId, 
 		Error, Module0, Module) -->
 	{ Callee = proc(CPred, CProc) },
@@ -215,7 +215,7 @@ first_order_check_goal(call(CPred, CProc, _Args, Builtin, _Contex, _Sym),
 			"call introduces a non-stratified loop", 
 			Error, Module0, Module)	
 	;
-		{ \+ hlds__is_builtin_is_internal(Builtin) },
+		{ BuiltinState = not_builtin },
 		{ \+ local_proc(Module0, Callee) }
 	->
 		{ goal_info_get_context(GInfo, Context) },
@@ -227,14 +227,15 @@ first_order_check_goal(call(CPred, CProc, _Args, Builtin, _Contex, _Sym),
 		{ Module = Module0 }
 	).
 first_order_check_goal(higher_order_call(_Var, _Vars, _Types, _Modes, 
-	_Det), GInfo, _Negated, _WholeScc, ThisPredProcId, Error,  
-	Module0, Module) --> 
+			_Det, _PredOrFunc),
+		GInfo, _Negated, _WholeScc, ThisPredProcId, Error,  
+		Module0, Module) --> 
 	{ goal_info_get_context(GInfo, Context) },
 	emit_message(ThisPredProcId, Context,
 		"higher order call may introduce a non-stratified loop",
 		Error, Module0, Module).
 
-:- pred first_order_check_goal_list(list(hlds__goal), bool, 
+:- pred first_order_check_goal_list(list(hlds_goal), bool, 
 	list(pred_proc_id), pred_proc_id, bool, module_info, 
 	module_info, io__state, io__state).
 :- mode first_order_check_goal_list(in, in, in, in, in, in, out, di, uo) is det.
@@ -271,12 +272,8 @@ local_proc(Module, proc(PredId, ProcId)) :-
 
 %-----------------------------------------------------------------------------%
 
-/*******************************
- ******************************* 
- ******************************* 
- ******************************* 
- * XXX : Currently we don't allow the higher order case so this code
- *  is commented out.
+ % XXX : Currently we don't allow the higher order case so this code
+ % is disabled.
 
 	% check the higher order SCCs for stratification		
 :- pred higher_order_check_sccs(list(pair(list(pred_proc_id), 
@@ -301,7 +298,7 @@ higher_order_check_scc([PredProcId|Remaining], WholeScc, HOInfo, Module0,
 	globals__io_lookup_bool_option(warn_non_stratification, Warn),
 	{ pred_info_get_marker_list(PredInfo, Markers) },
 	( 	
-		{ list__member(request(bottom_up), Markers) }
+		{ list__member(request(memo), Markers) }
 	->
 		{ Error = yes }
 	;
@@ -330,7 +327,7 @@ higher_order_check_scc([PredProcId|Remaining], WholeScc, HOInfo, Module0,
 	),
 	higher_order_check_scc(Remaining, WholeScc, HOInfo, Module1, Module).
 
-:- pred higher_order_check_goal(hlds__goal_expr, hlds__goal_info, bool,
+:- pred higher_order_check_goal(hlds_goal_expr, hlds_goal_info, bool,
 		set(pred_proc_id), pred_proc_id, bool, bool,
 		module_info, module_info, io__state, io__state).
 :- mode higher_order_check_goal(in, in, in, in, in, in, in, in, 
@@ -367,7 +364,7 @@ higher_order_check_goal(not(Goal - GoalInfo), _GoalInfo, _Negated, WholeScc,
 		ThisPredProcId, HighOrderLoops, Error, Module0, Module) -->
 	higher_order_check_goal(Goal, GoalInfo, yes, WholeScc, ThisPredProcId,
 		HighOrderLoops, Error, Module0, Module).
-higher_order_check_goal(pragma_c_code(_, _IsRec, _, _, _, _), _GoalInfo, 
+higher_order_check_goal(pragma_c_code(_, _IsRec, _, _, _, _, _, _), _GoalInfo, 
 	_Negated, _WholeScc, _ThisPredProcId, _HighOrderLoops, 
 	_, Module, Module) --> [].
 higher_order_check_goal(unify(_Var, _RHS, _Mode, _Uni, _Context), _GoalInfo,
@@ -394,7 +391,8 @@ higher_order_check_goal((call(_CPred, _CProc, _Args, _Builtin, _Contex, Sym)),
 		{ Module = Module0 }
 	).
 	
-higher_order_check_goal(higher_order_call(_Var, _Vars, _Types, _Modes, _Det), 
+higher_order_check_goal(higher_order_call(_Var, _Vars, _Types, _Modes, _Det,
+			_PredOrFunc), 
 		GoalInfo, Negated, _WholeScc, ThisPredProcId, HighOrderLoops, 
 		Error, Module0, Module) -->
 	(
@@ -409,7 +407,7 @@ higher_order_check_goal(higher_order_call(_Var, _Vars, _Types, _Modes, _Det),
 		{ Module = Module0 }
 	).
 	
-:- pred higher_order_check_goal_list(list(hlds__goal), bool, set(pred_proc_id),
+:- pred higher_order_check_goal_list(list(hlds_goal), bool, set(pred_proc_id),
 	pred_proc_id, bool, bool, module_info, module_info, 
 	io__state, io__state).
 :- mode higher_order_check_goal_list(in, in, in, in, in, in, in, out, 
@@ -773,7 +771,7 @@ higherorder_in_out1([Type|Types], [Mode|Modes], Module, HOIn0, HOIn,
 
 	% return the set of all procs called in and all addresses
 	% taken, in a given goal
-:- pred check_goal(hlds__goal_expr, set(pred_proc_id), set(pred_proc_id), 
+:- pred check_goal(hlds_goal_expr, set(pred_proc_id), set(pred_proc_id), 
 	bool). 
 :- mode check_goal(in, out, out, out) is det.
 
@@ -782,7 +780,7 @@ check_goal(Goal, Calls, TakenAddrs, CallsHO) :-
 	set__init(TakenAddrs0),
 	check_goal1(Goal, Calls0, Calls, TakenAddrs0, TakenAddrs, no, CallsHO).
 
-:- pred check_goal1(hlds__goal_expr, set(pred_proc_id), set(pred_proc_id), 
+:- pred check_goal1(hlds_goal_expr, set(pred_proc_id), set(pred_proc_id), 
 	set(pred_proc_id), set(pred_proc_id), bool, bool).
 :- mode check_goal1(in, in, out, in, out, in, out) is det.
 
@@ -827,8 +825,8 @@ check_goal1(call(CPred, CProc, _Args, _Builtin, _Contex, _Sym), Calls0, Calls,
 	set__insert(Calls0, proc(CPred, CProc), Calls).
 
 	% record that the higher order call was made
-check_goal1(higher_order_call(_Var, _Vars, _Types, _Modes, _Det), Calls, Calls, 
-		HasAT, HasAT, _, yes).
+check_goal1(higher_order_call(_Var, _Vars, _Types, _Modes, _Det, _PredOrFUnc),
+		Calls, Calls, HasAT, HasAT, _, yes).
 
 check_goal1(conj(Goals), Calls0, Calls, HasAT0, HasAT, CallsHO0, CallsHO) :-
 	check_goal_list(Goals, Calls0, Calls, HasAT0, HasAT, CallsHO0, CallsHO).
@@ -852,11 +850,11 @@ check_goal1(not(Goal - _GoalInfo), Calls0, Calls, HasAT0, HasAT, CallsHO0,
 		CallsHO) :- 
 	check_goal1(Goal, Calls0, Calls, HasAT0, HasAT, CallsHO0, CallsHO).
 
-check_goal1(pragma_c_code(_, _IsRec, _CPred, _CProc, _, _), Calls, Calls, 
+check_goal1(pragma_c_code(_, _IsRec, _CPred, _CProc, _, _, _, _), Calls, Calls, 
 		HasAT, HasAT, CallsHO, CallsHO).
 
 	
-:- pred check_goal_list(list(hlds__goal), set(pred_proc_id), set(pred_proc_id), 
+:- pred check_goal_list(list(hlds_goal), set(pred_proc_id), set(pred_proc_id), 
 	set(pred_proc_id), set(pred_proc_id), bool, bool). 
 :- mode check_goal_list(in, in, out, in, out, in, out) is det.
 
@@ -881,7 +879,7 @@ check_case_list([Case|Goals], Calls0, Calls, HasAT0, HasAT, CallsHO0,
 	% This pred returns a list of all the calls in a given set of
 	% goals including calls in unification lambda functions and
 	% pred_proc_id's in constructs 
-:- pred get_called_procs(hlds__goal_expr, list(pred_proc_id), 
+:- pred get_called_procs(hlds_goal_expr, list(pred_proc_id), 
 	list(pred_proc_id)). 
 :- mode get_called_procs(in, in, out) is det.
 
@@ -923,8 +921,8 @@ get_called_procs(call(CPred, CProc, _Args, _Builtin, _Contex, _Sym), Calls0,
 		Calls) :- 
 	Calls = [proc(CPred, CProc) | Calls0].
 
-get_called_procs(higher_order_call(_Var, _Vars, _Types, _Modes, _Det), Calls, 
-	Calls).
+get_called_procs(higher_order_call(_Var, _Vars, _Types, _Modes, _Det,
+		_PredOrFunc), Calls, Calls).
 
 
 get_called_procs(conj(Goals), Calls0, Calls) :-
@@ -942,10 +940,11 @@ get_called_procs(some(_Vars, Goal - _GoalInfo), Calls0, Calls) :-
 	get_called_procs(Goal, Calls0, Calls).
 get_called_procs(not(Goal - _GoalInfo), Calls0, Calls) :-
 	get_called_procs(Goal, Calls0, Calls).
-get_called_procs(pragma_c_code(_, _IsRec, _CPred, _CProc, _, _), Calls, Calls).
+get_called_procs(pragma_c_code(_, _IsRec, _CPred, _CProc, _, _, _, _),
+	Calls, Calls).
 
 
-:- pred check_goal_list(list(hlds__goal), list(pred_proc_id), 
+:- pred check_goal_list(list(hlds_goal), list(pred_proc_id), 
 	list(pred_proc_id)).
 :- mode check_goal_list(in, in, out) is det.
 
@@ -963,9 +962,6 @@ check_case_list([Case|Goals], Calls0, Calls) :-
 	get_called_procs(Goal, Calls0, Calls1), 
 	check_case_list(Goals, Calls1, Calls).
 
-***********************
-***********************/
-
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
 
