@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1995 University of Melbourne.
+% Copyright (C) 1994-1997 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -29,9 +29,10 @@
 
 :- implementation.
 
-:- import_module hlds_goal, hlds_data, prog_data, instmap.
+:- import_module hlds_goal, hlds_data, prog_data, instmap, inst_match.
 :- import_module modes, mode_util, type_util, det_util.
-:- import_module int, list, assoc_list, map, set, std_util, term, require.
+:- import_module passes_aux.
+:- import_module char, int, list, assoc_list, map, set, std_util, term, require.
 
 %-----------------------------------------------------------------------------%
 
@@ -57,10 +58,15 @@ detect_switches_in_preds([PredId | PredIds], ModuleInfo0, ModuleInfo) -->
 	io__state, io__state).
 :- mode detect_switches_in_pred(in, in, in, out, di, uo) is det.
 
-detect_switches_in_pred(PredId, PredInfo0, ModuleInfo0, ModuleInfo,
-		IOstate, IOstate) :-
-	pred_info_non_imported_procids(PredInfo0, ProcIds),
-	detect_switches_in_procs(ProcIds, PredId, ModuleInfo0, ModuleInfo).
+detect_switches_in_pred(PredId, PredInfo0, ModuleInfo0, ModuleInfo) -->
+	{ pred_info_non_imported_procids(PredInfo0, ProcIds) },
+	( { ProcIds \= [] } ->
+		write_pred_progress_message("% Detecting switches in ", PredId,
+			ModuleInfo0)
+	;
+		[]
+	),
+	{ detect_switches_in_procs(ProcIds, PredId, ModuleInfo0, ModuleInfo) }.
 
 :- pred detect_switches_in_procs(list(proc_id), pred_id,
 	module_info, module_info).
@@ -86,9 +92,9 @@ detect_switches_in_proc(ProcId, PredId, ModuleInfo0, ModuleInfo) :-
 	detect_switches_in_goal(Goal0, InstMap0, VarTypes, ModuleInfo0, Goal),
 
 	proc_info_set_goal(ProcInfo0, Goal, ProcInfo),
-	map__set(ProcTable0, ProcId, ProcInfo, ProcTable),
+	map__det_update(ProcTable0, ProcId, ProcInfo, ProcTable),
 	pred_info_set_procedures(PredInfo0, ProcTable, PredInfo),
-	map__set(PredTable0, PredId, PredInfo, PredTable),
+	map__det_update(PredTable0, PredId, PredInfo, PredTable),
 	module_info_set_preds(ModuleInfo0, PredTable, ModuleInfo).
 
 %-----------------------------------------------------------------------------%
@@ -96,8 +102,8 @@ detect_switches_in_proc(ProcId, PredId, ModuleInfo0, ModuleInfo) :-
 	% Given a goal, and the instmap on entry to that goal,
 	% replace disjunctions with switches whereever possible.
 
-:- pred detect_switches_in_goal(hlds__goal, instmap, map(var, type),
-	module_info, hlds__goal).
+:- pred detect_switches_in_goal(hlds_goal, instmap, map(var, type),
+	module_info, hlds_goal).
 :- mode detect_switches_in_goal(in, in, in, in, out) is det.
 
 detect_switches_in_goal(Goal0, InstMap0, VarTypes, ModuleInfo, Goal) :-
@@ -109,8 +115,8 @@ detect_switches_in_goal(Goal0, InstMap0, VarTypes, ModuleInfo, Goal) :-
 	% computed by applying the instmap delta specified in the
 	% goal's goalinfo.
 
-:- pred detect_switches_in_goal_1(hlds__goal, instmap, map(var, type),
-	module_info, hlds__goal, instmap).
+:- pred detect_switches_in_goal_1(hlds_goal, instmap, map(var, type),
+	module_info, hlds_goal, instmap).
 :- mode detect_switches_in_goal_1(in, in, in, in, out, out) is det.
 
 detect_switches_in_goal_1(Goal0 - GoalInfo, InstMap0, VarTypes, ModuleInfo,
@@ -121,8 +127,8 @@ detect_switches_in_goal_1(Goal0 - GoalInfo, InstMap0, VarTypes, ModuleInfo,
 
 	% Here we process each of the different sorts of goals.
 
-:- pred detect_switches_in_goal_2(hlds__goal_expr, hlds__goal_info, instmap,
-		map(var, type), module_info, hlds__goal_expr).
+:- pred detect_switches_in_goal_2(hlds_goal_expr, hlds_goal_info, instmap,
+		map(var, type), module_info, hlds_goal_expr).
 :- mode detect_switches_in_goal_2(in, in, in, in, in, out) is det.
 
 detect_switches_in_goal_2(disj(Goals0, SM), GoalInfo, InstMap0,
@@ -157,8 +163,8 @@ detect_switches_in_goal_2(some(Vars, Goal0), _GoalInfo, InstMap0,
 		VarTypes, ModuleInfo, some(Vars, Goal)) :-
 	detect_switches_in_goal(Goal0, InstMap0, VarTypes, ModuleInfo, Goal).
 
-detect_switches_in_goal_2(higher_order_call(A,B,C,D,E), _, _, _, _,
-		higher_order_call(A,B,C,D,E)).
+detect_switches_in_goal_2(higher_order_call(A,B,C,D,E,F), _, _, _, _,
+		higher_order_call(A,B,C,D,E,F)).
 
 detect_switches_in_goal_2(call(A,B,C,D,E,F), _, _, _, _,
 		call(A,B,C,D,E,F)).
@@ -168,10 +174,8 @@ detect_switches_in_goal_2(unify(A,RHS0,C,D,E), __GoalInfo, InstMap0,
 	( RHS0 = lambda_goal(PredOrFunc, Vars, Modes, Det, Goal0) ->
 		% we need to insert the initial insts for the lambda
 		% variables in the instmap before processing the lambda goal
-		mode_list_get_initial_insts(Modes, ModuleInfo, Insts),
-		assoc_list__from_corresponding_lists(Vars, Insts, VarInsts),
-		instmap_delta_from_assoc_list(VarInsts, InstmapDelta),
-		instmap__apply_instmap_delta(InstMap0, InstmapDelta, InstMap1),
+		instmap__pre_lambda_update(ModuleInfo, 
+			Vars, Modes, InstMap0, InstMap1),
 		detect_switches_in_goal(Goal0, InstMap1, VarTypes, ModuleInfo,
 			Goal),
 		RHS = lambda_goal(PredOrFunc, Vars, Modes, Det, Goal)
@@ -183,8 +187,8 @@ detect_switches_in_goal_2(switch(Var, CanFail, Cases0, SM), _, InstMap,
 		VarTypes, ModuleInfo, switch(Var, CanFail, Cases, SM)) :-
 	detect_switches_in_cases(Cases0, InstMap, VarTypes, ModuleInfo, Cases).
 
-detect_switches_in_goal_2(pragma_c_code(A,B,C,D,E,F), _, _, _, _,
-		pragma_c_code(A,B,C,D,E,F)).
+detect_switches_in_goal_2(pragma_c_code(A,B,C,D,E,F,G,H), _, _, _, _,
+		pragma_c_code(A,B,C,D,E,F,G,H)).
 
 %-----------------------------------------------------------------------------%
 
@@ -195,17 +199,17 @@ detect_switches_in_goal_2(pragma_c_code(A,B,C,D,E,F), _, _, _, _,
 	% each group of disjunctions can only succeed if the variable
 	% is bound to a different functor.
 
-:- type cases == map(cons_id, list(hlds__goal)).
+:- type cases == map(cons_id, list(hlds_goal)).
 
-:- type sorted_case_list == assoc_list(cons_id, list(hlds__goal)).
+:- type sorted_case_list == list(case).
 	% the sorted_case_list should always be sorted on cons_id -
 	% `delete_unreachable_cases' relies on this.
 
-:- type again ---> again(var, list(hlds__goal), sorted_case_list).
+:- type again ---> again(var, list(hlds_goal), sorted_case_list).
 
-:- pred detect_switches_in_disj(list(var), list(hlds__goal), hlds__goal_info,
+:- pred detect_switches_in_disj(list(var), list(hlds_goal), hlds_goal_info,
 	store_map, instmap, map(var, type), list(var), module_info,
-	list(again), hlds__goal_expr).
+	list(again), hlds_goal_expr).
 :- mode detect_switches_in_disj(in, in, in, in, in, in, in, in, in, out) is det.
 
 detect_switches_in_disj([Var | Vars], Goals0, GoalInfo, SM, InstMap,
@@ -214,18 +218,46 @@ detect_switches_in_disj([Var | Vars], Goals0, GoalInfo, SM, InstMap,
 	(
 		instmap__lookup_var(InstMap, Var, VarInst0),
 		inst_is_bound(ModuleInfo, VarInst0),
-		partition_disj(Goals0, Var, Left, CasesList)
+		partition_disj(Goals0, Var, GoalInfo, Left, CasesList)
 	->
-		% are there any disjuncts that are not part of the switch?
+		%
+		% A switch needs to have at least two cases.
+		%
+		% But, if there is a complete one-case switch
+		% for a goal, we must leave it as a disjunction
+		% rather than doing an incomplete switch on a
+		% different variable, because otherwise we might
+		% get determinism analysis wrong.  (The complete
+		% one-case switch may be decomposable into other
+		% complete sub-switches on the functor's arguments)
+		%
 		(
+			% are there any disjuncts that are not part of the
+			% switch?
 			Left = []
 		->
-			cases_to_switch(CasesList, Var, VarTypes, GoalInfo,
-				SM, InstMap, ModuleInfo, Goal)
+			( CasesList = [_, _ | _] ->
+				cases_to_switch(CasesList, Var, VarTypes,
+					GoalInfo, SM, InstMap, ModuleInfo,
+					Goal)
+			;
+				detect_sub_switches_in_disj(Goals0, InstMap,
+					VarTypes, ModuleInfo, Goals),
+				Goal = disj(Goals, SM)
+			)
 		;
+			% insert this switch into the list of incomplete
+			% switches only if it has at least two cases
+			%
+			( CasesList = [_, _ | _] ->
+				Again1 = [again(Var, Left, CasesList) | Again0]
+			;
+				Again1 = Again0
+			),
+			% try to find a switch
 			detect_switches_in_disj(Vars, Goals0, GoalInfo,
 				SM, InstMap, VarTypes, AllVars, ModuleInfo,
-				[again(Var, Left, CasesList) | Again0], Goal)
+				Again1, Goal)
 		)
 	;
 		detect_switches_in_disj(Vars, Goals0, GoalInfo, SM, InstMap,
@@ -267,8 +299,8 @@ select_best_switch([Again | AgainList], BestAgain0, BestAgain) :-
 	),
 	select_best_switch(AgainList, BestAgain1, BestAgain).
 
-:- pred detect_sub_switches_in_disj(list(hlds__goal), instmap, map(var, type),
-	module_info, list(hlds__goal)).
+:- pred detect_sub_switches_in_disj(list(hlds_goal), instmap, map(var, type),
+	module_info, list(hlds_goal)).
 :- mode detect_sub_switches_in_disj(in, in, in, in, out) is det.
 
 detect_sub_switches_in_disj([], _InstMap, _VarTypes, _ModuleInfo, []).
@@ -290,8 +322,8 @@ detect_switches_in_cases([Case0 | Cases0], InstMap, VarTypes, ModuleInfo,
 	Case = case(Functor, Goal),
 	detect_switches_in_cases(Cases0, InstMap, VarTypes, ModuleInfo, Cases).
 
-:- pred detect_switches_in_conj(list(hlds__goal), instmap, map(var, type),
-	module_info, list(hlds__goal)).
+:- pred detect_switches_in_conj(list(hlds_goal), instmap, map(var, type),
+	module_info, list(hlds_goal)).
 :- mode detect_switches_in_conj(in, in, in, in, out) is det.
 
 detect_switches_in_conj([], _InstMap, _VarTypes, _ModuleInfo, []).
@@ -316,18 +348,19 @@ detect_switches_in_conj([Goal0 | Goals0], InstMap0, VarTypes, ModuleInfo,
 	% We partition the goals by abstractly interpreting the unifications
 	% at the start of each disjunction, to build up a substitution.
 
-:- pred partition_disj(list(hlds__goal), var, list(hlds__goal),
+:- pred partition_disj(list(hlds_goal), var, hlds_goal_info, list(hlds_goal),
 	sorted_case_list).
-:- mode partition_disj(in, in, out, out) is semidet.
+:- mode partition_disj(in, in, in, out, out) is semidet.
 
-partition_disj(Goals0, Var, Left, CasesAssocList) :-
+partition_disj(Goals0, Var, GoalInfo, Left, CasesList) :-
 	map__init(Cases0),
 	partition_disj_trial(Goals0, Var, [], Left, Cases0, Cases),
 	map__to_assoc_list(Cases, CasesAssocList),
-	CasesAssocList = [_, _ | _].	% there must be more than one case
+	CasesAssocList \= [], % there must be at least one case
+	fix_case_list(CasesAssocList, GoalInfo, CasesList).
 
-:- pred partition_disj_trial(list(hlds__goal), var,
-	list(hlds__goal), list(hlds__goal), cases, cases).
+:- pred partition_disj_trial(list(hlds_goal), var,
+	list(hlds_goal), list(hlds_goal), cases, cases).
 :- mode partition_disj_trial(in, in, in, out, in, out) is det.
 
 partition_disj_trial([], _Var, Left, Left, Cases, Cases).
@@ -342,11 +375,12 @@ partition_disj_trial([Goal0 | Goals], Var, Left0, Left, Cases0, Cases) :-
 		Left1 = Left0,
 		conj_list_to_goal(ConjList, GoalInfo, Goal),
 		( map__search(Cases0, Functor, DisjList0) ->
-			DisjList1 = [Goal | DisjList0]
+			DisjList1 = [Goal | DisjList0],
+			map__det_update(Cases0, Functor, DisjList1, Cases1)
 		;
-			DisjList1 = [Goal]
-		),
-		map__set(Cases0, Functor, DisjList1, Cases1)
+			DisjList1 = [Goal],
+			map__det_insert(Cases0, Functor, DisjList1, Cases1)
+		)
 	;
 		MaybeFunctor = no,
 		Left1 = [Goal0 | Left0],
@@ -365,8 +399,8 @@ partition_disj_trial([Goal0 | Goals], Var, Left0, Left, Cases0, Cases) :-
 	%	`MaybeFunctor' to `no', and computes `Subst' as the
 	%	resulting substitution from interpreting through the goal.
 
-:- pred find_bind_var_for_switch(list(hlds__goal), substitution, var,
-	list(hlds__goal), substitution, maybe(cons_id)).
+:- pred find_bind_var_for_switch(list(hlds_goal), substitution, var,
+	list(hlds_goal), substitution, maybe(cons_id)).
 :- mode find_bind_var_for_switch(in, in, in, out, out, out) is det.
 
 find_bind_var_for_switch([], Substitution, _Var, [], Substitution, no).
@@ -424,11 +458,11 @@ find_bind_var_for_switch([Goal0 - GoalInfo | Goals0], Substitution0, Var,
 		MaybeFunctor = no
 	).
 
-:- pred cases_to_switch(sorted_case_list, var, map(var, type), hlds__goal_info,
-	store_map, instmap, module_info, hlds__goal_expr).
+:- pred cases_to_switch(sorted_case_list, var, map(var, type), hlds_goal_info,
+	store_map, instmap, module_info, hlds_goal_expr).
 :- mode cases_to_switch(in, in, in, in, in, in, in, out) is det.
 
-cases_to_switch(CasesList, Var, VarTypes, GoalInfo, SM, InstMap, ModuleInfo,
+cases_to_switch(CasesList, Var, VarTypes, _GoalInfo, SM, InstMap, ModuleInfo,
 		Goal) :-
 	instmap__lookup_var(InstMap, Var, VarInst),
 	( inst_is_bound_to_functors(ModuleInfo, VarInst, Functors) ->
@@ -442,15 +476,15 @@ cases_to_switch(CasesList, Var, VarTypes, GoalInfo, SM, InstMap, ModuleInfo,
 		)
 	;
 		map__lookup(VarTypes, Var, Type),
-		( switch_covers_all_cases(CasesList, Type, ModuleInfo) ->
+		CasesList1 = CasesList,
+		( switch_covers_all_cases(CasesList1, Type, ModuleInfo) ->
 			CanFail = cannot_fail
 		;
 			CanFail = can_fail
-		),
-		CasesList1 = CasesList
+		)
 	),
-	fix_case_list(CasesList1, GoalInfo, Cases0),
-	detect_switches_in_cases(Cases0, InstMap, VarTypes, ModuleInfo, Cases),
+	detect_switches_in_cases(CasesList1, InstMap, VarTypes,
+		ModuleInfo, Cases),
 
 	% We turn switches with no arms into fail, since this avoids having
 	% the code generator flush the control variable of the switch.
@@ -467,65 +501,35 @@ cases_to_switch(CasesList, Var, VarTypes, GoalInfo, SM, InstMap, ModuleInfo,
 		Goal = switch(Var, CanFail, Cases, SM)
 	).
 
-:- pred delete_unreachable_cases(sorted_case_list, list(cons_id),
-	sorted_case_list).
-:- mode delete_unreachable_cases(in, in, out) is det.
-
-	% Given a list of cases, and a list of the possible cons_ids
-	% that the switch variable could be bound to, select out only
-	% those cases whose cons_id occurs in the list of cases
-	% We assume that the list of cases and the list of cons_ids
-	% are sorted, so that we can do this using a simple sorted merge.
-
-delete_unreachable_cases([], _, []).
-delete_unreachable_cases([_ | _], [], []).
-delete_unreachable_cases([Case | Cases0], [ConsId | ConsIds], Cases) :-
-	Case = CaseConsId - _DisjList,
-	( CaseConsId = ConsId ->
-		Cases = [Case | Cases1],
-		delete_unreachable_cases(Cases0, ConsIds, Cases1)
-	; compare(<, CaseConsId, ConsId) ->
-		delete_unreachable_cases(Cases0, [ConsId | ConsIds], Cases)
-	;
-		delete_unreachable_cases([Case | Cases0], ConsIds, Cases)
-	).
-
-	% Given a list of bound_insts, get the corresponding list of cons_ids
-	% 
-:- pred functors_to_cons_ids(list(bound_inst), list(cons_id)).
-:- mode functors_to_cons_ids(in, out) is det.
-
-functors_to_cons_ids([], []).
-functors_to_cons_ids([Functor | Functors], [ConsId | ConsIds]) :-
-	Functor = functor(ConsId, _ArgInsts),
-	functors_to_cons_ids(Functors, ConsIds).
-
 	% check whether a switch handles all the possible
 	% constants/functors for the type
 
-:- pred switch_covers_all_cases(assoc_list(cons_id, list(hlds__goal)),
-	type, module_info).
+:- pred switch_covers_all_cases(sorted_case_list, type, module_info).
 :- mode switch_covers_all_cases(in, in, in) is semidet.
 
 switch_covers_all_cases(CasesList, Type, _ModuleInfo) :-
 	Type = term__functor(term__atom("character"), [], _),
-	list__length(CasesList, 127).	% XXX should be 256
-		% NU-Prolog only allows chars '\001' .. '\0177'.
-		% Currently we assume the same set.
+	% XXX the following code uses the source machine's character size,
+	% not the target's, so it won't work if cross-compiling to a
+	% machine with a different size character.
+	char__max_char_value(MaxChar),
+	char__min_char_value(MinChar),
+	NumChars is MaxChar - MinChar + 1,
+	list__length(CasesList, NumChars).
 
 switch_covers_all_cases(CasesList, Type, ModuleInfo) :-
 	type_to_type_id(Type, TypeId, _),
 	module_info_types(ModuleInfo, TypeTable),
 	map__search(TypeTable, TypeId, TypeDefn),
 	hlds_data__get_type_defn_body(TypeDefn, TypeBody),
-	TypeBody = du_type(_, ConsTable, _),
+	TypeBody = du_type(_, ConsTable, _, _),
 	map__keys(ConsTable, Constructors),
 	list__same_length(CasesList, Constructors).
 
-	% convert the assoc_list(cons_id, list(hlds__goal) back into
+	% convert the assoc_list(cons_id, list(hlds_goal) back into
 	% a plain list(case).
 
-:- pred fix_case_list(assoc_list(cons_id, list(hlds__goal)), hlds__goal_info,
+:- pred fix_case_list(assoc_list(cons_id, list(hlds_goal)), hlds_goal_info,
 	list(case)).
 :- mode fix_case_list(in, in, out) is det.
 
