@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1995 University of Melbourne.
+% Copyright (C) 1994-1997 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -28,7 +28,7 @@
 :- pred type_is_atomic(type, module_info).
 :- mode type_is_atomic(in, in) is semidet.
 
-	% type_is_higher_order_type(Type, PredOrFunc, ArgTypes) succeeds iff
+	% type_is_higher_order(Type, PredOrFunc, ArgTypes) succeeds iff
 	% Type is a higher-order predicate or function type with the specified
 	% argument types (for functions, the return type is appended to the
 	% end of the argument types).
@@ -36,6 +36,12 @@
 :- pred type_is_higher_order(type, pred_or_func, list(type)).
 :- mode type_is_higher_order(in, out, out) is semidet.
 
+	% type_id_is_higher_order(TypeId, PredOrFunc) succeeds iff
+	% TypeId is a higher-order predicate or function type.
+
+:- pred type_id_is_higher_order(type_id, pred_or_func).
+:- mode type_id_is_higher_order(in, out) is semidet.
+
 	% Given a type, determine what sort of type it is.
 
 :- pred classify_type(type, module_info, builtin_type).
@@ -48,7 +54,7 @@
 			;	pred_type
 			;	enum_type
 			;	polymorphic_type
-			;	user_type(type).
+			;	user_type.
 
 	% Given a non-variable type, return its type-id and argument types.
 
@@ -59,6 +65,7 @@
 	
 :- pred type_util__var(type, var).
 :- mode type_util__var(in, out) is semidet.
+:- mode type_util__var(out, in) is det.
 
 	% Given a type_id, a list of argument types and maybe a context,
 	% construct a type.
@@ -86,11 +93,15 @@
 :- pred type_util__type_id_arity(module_info, type_id, arity).
 :- mode type_util__type_id_arity(in, in, out) is det.
 
-	% If the type is a du type, return the list of it's constructors.
+	% If the type is a du type, return the list of its constructors.
 
 :- pred type_constructors(type, module_info, list(constructor)).
 :- mode type_constructors(in, in, out) is semidet.
 
+	% Work out the types of the arguments of a functor.
+:- pred type_util__get_cons_id_arg_types(module_info::in, (type)::in,
+		cons_id::in, list(type)::out) is det.
+
 	% Given a list of constructors for a type,
 	% check whether that type is a no_tag type
 	% (i.e. one with only one constructor, and
@@ -138,13 +149,22 @@
 						 map(var, type)).
 :- mode apply_rec_substitution_to_type_map(in, in, out) is det.
 
+	% Update a map from tvar to var, using the type substititon to
+	% rename tvars and a variable substition to rename vars.
+	%
+	% If tvar maps to a another type variable, we keep the new
+	% variable, if it maps to a type, we remove it from the map.
+
+:- pred apply_substitutions_to_var_map(map(tvar, var), tsubst, map(var, var), 
+		map(tvar, var)).
+:- mode apply_substitutions_to_var_map(in, in, in, out) is det.
 
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
 
 :- implementation.
 :- import_module bool, list, term, require, map, std_util.
-:- import_module prog_io, prog_util.
+:- import_module prog_io, prog_io_goal, prog_util.
 
 type_util__type_id_module(_ModuleInfo, qualified(ModuleName, _) -_, ModuleName).
 type_util__type_id_module(_ModuleInfo, unqualified(_) - _, "").
@@ -158,7 +178,7 @@ type_is_atomic(Type, ModuleInfo) :-
 	classify_type(Type, ModuleInfo, BuiltinType),
 	BuiltinType \= polymorphic_type,
 	BuiltinType \= pred_type,
-	BuiltinType \= user_type(_).
+	BuiltinType \= user_type.
 
 type_util__var(term__variable(Var), Var).
 
@@ -196,7 +216,7 @@ classify_type(VarType, ModuleInfo, Type) :-
 	->
 		Type = enum_type
 	;
-		Type = user_type(VarType)
+		Type = user_type
 	).
 
 type_is_higher_order(Type, PredOrFunc, PredArgTypes) :-
@@ -213,6 +233,17 @@ type_is_higher_order(Type, PredOrFunc, PredArgTypes) :-
 		PredOrFunc = function
 	).
 
+type_id_is_higher_order(SymName - Arity, PredOrFunc) :-
+	unqualify_name(SymName, TypeName),
+	( 
+		TypeName = "pred",
+		PredOrFunc = predicate
+	; 
+		TypeName = "=",
+		Arity = 2,
+		PredOrFunc = function
+	).
+
 :- pred type_is_enumeration(type, module_info).
 :- mode type_is_enumeration(in, in) is semidet.
 
@@ -221,29 +252,49 @@ type_is_enumeration(Type, ModuleInfo) :-
 	module_info_types(ModuleInfo, TypeDefnTable),
 	map__search(TypeDefnTable, TypeId, TypeDefn),
 	hlds_data__get_type_defn_body(TypeDefn, TypeBody),
-	TypeBody = du_type(_, _, IsEnum),
+	TypeBody = du_type(_, _, IsEnum, _),
 	IsEnum = yes.
 
 type_to_type_id(Type, SymName - Arity, Args) :-
-	Type = term__functor(term__atom(Atom), Args0, _),
+	sym_name_and_args(Type, SymName, Args1),
+
+	% higher order types may have representations where
+	% their arguments don't directly correspond to the
+	% arguments of the term.
 	(
-		Atom = ":", Args0 = [ModuleTerm, TypeAndArgsTerm]
+		type_is_higher_order(Type, _, PredArgTypes) 
 	->
-		ModuleTerm = term__functor(term__atom(ModuleName), [], _),
-		TypeAndArgsTerm = term__functor(term__atom(TypeName), Args, _),
-		SymName = qualified(ModuleName, TypeName)
+		Args = PredArgTypes,
+		list__length(Args1, Arity)	% functions have arity 2, 
+						% (they are =/2)
 	;
-		SymName = unqualified(Atom),
-		Args = Args0
+		Args = Args1,
+		list__length(Args, Arity)
+	).
+
+construct_type(TypeId, Args, Type) :-
+	term__context_init(Context),
+	construct_type(TypeId, Args, Context, Type).
+
+construct_type(TypeId, Args, Context, Type) :-
+	(
+		type_id_is_higher_order(TypeId, PredOrFunc)
+	->
+		(
+			PredOrFunc = predicate,
+			NewArgs = Args
+		;
+			PredOrFunc = function,
+			pred_args_to_func_args(Args, FuncArgTypes, FuncRetType),
+			NewArgs = [term__functor(term__atom("func"),
+						FuncArgTypes, Context),
+					 FuncRetType]
+		)
+	;
+		NewArgs = Args
 	),
-	list__length(Args, Arity).
-
-
-construct_type(SymName - _, Args, Type) :-
-	construct_qualified_term(SymName, Args, Type).
-
-construct_type(SymName - _, Args, Context, Type) :-
-	construct_qualified_term(SymName, Args, Context, Type).
+	TypeId = SymName - _,
+	construct_qualified_term(SymName, NewArgs, Context, Type).
 
 %-----------------------------------------------------------------------------%
 
@@ -256,7 +307,7 @@ make_type_id(term__atom(Name), Arity, unqualified(Name) - Arity).
 
 %-----------------------------------------------------------------------------%
 
-	% If the type is a du type, return the list of it's constructors.
+	% If the type is a du type, return the list of its constructors.
 
 type_constructors(Type, ModuleInfo, Constructors) :-
 	type_to_type_id(Type, TypeId, TypeArgs),
@@ -264,10 +315,35 @@ type_constructors(Type, ModuleInfo, Constructors) :-
 	map__search(TypeTable, TypeId, TypeDefn),
 	hlds_data__get_type_defn_tparams(TypeDefn, TypeParams),
 	hlds_data__get_type_defn_body(TypeDefn, TypeBody),
-	TypeBody = du_type(Constructors0, _, _),
+	TypeBody = du_type(Constructors0, _, _, _),
 	substitute_type_args(TypeParams, TypeArgs, Constructors0,
 		Constructors).
 
+%-----------------------------------------------------------------------------%
+
+type_util__get_cons_id_arg_types(ModuleInfo, VarType, ConsId, ArgTypes) :-
+	(
+		type_to_type_id(VarType, TypeId, TypeArgs),
+		module_info_ctors(ModuleInfo, Ctors),
+		% will fail for builtin cons_ids.
+		map__search(Ctors, ConsId, ConsDefns),
+		CorrectCons = lambda([ConsDefn::in] is semidet, (
+				ConsDefn = hlds_cons_defn(_, TypeId, _)
+			)),
+		list__filter(CorrectCons, ConsDefns,
+			[hlds_cons_defn(ArgTypes0, _, _)]),
+		ArgTypes0 \= []
+	->
+		module_info_types(ModuleInfo, Types),
+		map__lookup(Types, TypeId, TypeDefn),
+		hlds_data__get_type_defn_tparams(TypeDefn, TypeDefnParams),
+		term__term_list_to_var_list(TypeDefnParams, TypeDefnVars),
+		term__substitute_corresponding_list(TypeDefnVars, TypeArgs,
+			ArgTypes0, ArgTypes)
+	;
+		ArgTypes = []
+	).
+
 %-----------------------------------------------------------------------------%
 
 	% the checks for type_info and base_type_info
@@ -298,29 +374,28 @@ substitute_type_args(TypeParams0, TypeArgs, Constructors0, Constructors) :-
 		Constructors = Constructors0
 	;
 		term__term_list_to_var_list(TypeParams0, TypeParams),
-		substitute_type_args_2(Constructors0, TypeParams, TypeArgs,
-			Constructors)
+		map__from_corresponding_lists(TypeParams, TypeArgs, Subst),
+		substitute_type_args_2(Constructors0, Subst, Constructors)
 	).
 
-:- pred substitute_type_args_2(list(constructor), list(var), list(type),
+:- pred substitute_type_args_2(list(constructor), substitution,
 				list(constructor)).
-:- mode substitute_type_args_2(in, in, in, out) is det.
+:- mode substitute_type_args_2(in, in, out) is det.
 
-substitute_type_args_2([], _TypeParams, _TypeArgs, []).
-substitute_type_args_2([Name - Args0 | Ctors0], TypeParams, TypeArgs,
+substitute_type_args_2([], _, []).
+substitute_type_args_2([Name - Args0 | Ctors0], Subst,
 		[Name - Args | Ctors]) :-
-	substitute_type_args_3(Args0, TypeParams, TypeArgs, Args),
-	substitute_type_args_2(Ctors0, TypeParams, TypeArgs, Ctors).
+	substitute_type_args_3(Args0, Subst, Args),
+	substitute_type_args_2(Ctors0, Subst, Ctors).
 
-:- pred substitute_type_args_3(list(constructor_arg), list(var), list(type),
+:- pred substitute_type_args_3(list(constructor_arg), substitution,
 				list(constructor_arg)).
-:- mode substitute_type_args_3(in, in, in, out) is det.
+:- mode substitute_type_args_3(in, in, out) is det.
 
-substitute_type_args_3([], _TypeParams, _TypeArgs, []).
-substitute_type_args_3([Name - Arg0 | Args0], TypeParams, TypeArgs,
-		[Name - Arg | Args]) :-
-	term__substitute_corresponding(TypeParams, TypeArgs, Arg0, Arg),
-	substitute_type_args_3(Args0, TypeParams, TypeArgs, Args).
+substitute_type_args_3([], _, []).
+substitute_type_args_3([Name - Arg0 | Args0], Subst, [Name - Arg | Args]) :-
+	term__apply_substitution(Arg0, Subst, Arg),
+	substitute_type_args_3(Args0, Subst, Args).
 
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
@@ -344,8 +419,6 @@ type_list_subsumes(TypesA, TypesB, TypeSubst) :-
 	% (the type variables that occur in the head of the clause),
 	% and because one day we might want to handle equivalent types.
 
-:- type_unify(X, Y, _, _, _) when X and Y.		% NU-Prolog indexing
-
 type_unify(term__variable(X), term__variable(Y), HeadTypeParams, Bindings0,
 		Bindings) :-
 	( list__member(Y, HeadTypeParams) ->
@@ -368,7 +441,7 @@ type_unify(term__variable(X), term__variable(Y), HeadTypeParams, Bindings0,
 				Bindings = Bindings0
 			;
 				\+ term__occurs(SubstBindingOfX, Y, Bindings0),
-				map__set(Bindings0, Y, SubstBindingOfX,
+				map__det_insert(Bindings0, Y, SubstBindingOfX,
 					Bindings)
 			)
 		)
@@ -381,7 +454,7 @@ type_unify(term__variable(X), term__variable(Y), HeadTypeParams, Bindings0,
 				Bindings = Bindings0
 			;
 				\+ term__occurs(SubstBindingOfY, X, Bindings0),
-				map__set(Bindings0, X, SubstBindingOfY,
+				map__det_insert(Bindings0, X, SubstBindingOfY,
 					Bindings)
 			)
 		;
@@ -390,7 +463,7 @@ type_unify(term__variable(X), term__variable(Y), HeadTypeParams, Bindings0,
 			( X = Y ->
 				Bindings = Bindings0
 			; 
-				map__set(Bindings0, X, term__variable(Y),
+				map__det_insert(Bindings0, X, term__variable(Y),
 					Bindings)
 			)
 		)
@@ -406,7 +479,7 @@ type_unify(term__variable(X), term__functor(F, As, C), HeadTypeParams,
 	;
 		\+ term__occurs_list(As, X, Bindings0),
 		\+ list__member(X, HeadTypeParams),
-		map__set(Bindings0, X, term__functor(F, As, C), Bindings)
+		map__det_insert(Bindings0, X, term__functor(F, As, C), Bindings)
 	).
 
 type_unify(term__functor(F, As, C), term__variable(X), HeadTypeParams,
@@ -419,7 +492,7 @@ type_unify(term__functor(F, As, C), term__variable(X), HeadTypeParams,
 	;
 		\+ term__occurs_list(As, X, Bindings0),
 		\+ list__member(X, HeadTypeParams),
-		map__set(Bindings0, X, term__functor(F, As, C), Bindings)
+		map__det_insert(Bindings0, X, term__functor(F, As, C), Bindings)
 	).
 
 type_unify(term__functor(FX, AsX, _CX), term__functor(FY, AsY, _CY),
@@ -467,7 +540,7 @@ replace_eqv_type(Functor, Arity, Args, EqvType) :-
 
 	make_type_id(Functor, Arity, TypeId),
 	map__search(TypeTable, TypeId, TypeDefn),
-	TypeDefn = hlds__type_defn(TypeVarSet, TypeParams0,
+	TypeDefn = hlds_type_defn(TypeVarSet, TypeParams0,
 			eqv_type(EqvType0), _Condition, Context, _Status),
 	varset__merge(TVarSet0, TypeVarSet, [EqvType0 | TypeParams0],
 			TVarSet, [EqvType1, TypeParams1]),
@@ -496,7 +569,7 @@ type_unify_head_type_param(Var, HeadVar, HeadTypeParams, Bindings0,
 			Bindings = Bindings0
 		;
 			\+ list__member(Var, HeadTypeParams),
-			map__set(Bindings0, Var, term__variable(HeadVar),
+			map__det_insert(Bindings0, Var, term__variable(HeadVar),
 				Bindings)
 		)
 	).
@@ -552,7 +625,53 @@ apply_rec_substitution_to_type_map_2([Var | Vars], VarTypes0, Subst,
 	map__det_update(VarTypes0, Var, VarType, VarTypes1),
 	apply_rec_substitution_to_type_map_2(Vars, VarTypes1, Subst, VarTypes).
 
+%-----------------------------------------------------------------------------%
 
+apply_substitutions_to_var_map(VarMap0, TSubst, Subst, VarMap) :-
+	% optimize the common case of empty substitutions
+	( map__is_empty(Subst), map__is_empty(TSubst) ->
+		VarMap = VarMap0
+	;
+		map__keys(VarMap0, TVars),
+		map__init(NewVarMap),
+		apply_substitutions_to_var_map_2(TVars, VarMap0, TSubst,
+			Subst, NewVarMap, VarMap)
+	).
+
+
+:- pred apply_substitutions_to_var_map_2(list(var)::in, map(tvar, var)::in,
+		tsubst::in, map(var, var)::in, map(tvar, var)::in, 
+		map(tvar, var)::out) is det.
+
+apply_substitutions_to_var_map_2([], _VarMap0, _, _, NewVarMap, NewVarMap).
+apply_substitutions_to_var_map_2([TVar | TVars], VarMap0, TSubst, Subst, 
+		NewVarMap0, NewVarMap) :-
+	map__lookup(VarMap0, TVar, Var),
+
+		% find the new tvar, if there is one, otherwise just
+		% create the old var as a type variable.
+	( map__search(TSubst, TVar, NewTerm0) ->
+		NewTerm = NewTerm0 
+	; 
+		type_util__var(NewTerm, TVar)
+	),
+
+		% find the new var, if there is one
+	( map__search(Subst, Var, NewVar0) ->
+		NewVar = NewVar0
+	;
+		NewVar = Var
+	),
+
+		% if the tvar is still a variable, insert it into the
+		% map with the new var.
+	( type_util__var(NewTerm, NewTVar) ->
+		map__det_insert(NewVarMap0, NewTVar, NewVar, NewVarMap1)
+	;
+		NewVarMap1 = NewVarMap0
+	),
+	apply_substitutions_to_var_map_2(TVars, VarMap0, TSubst, Subst, 
+		NewVarMap1, NewVarMap).
 
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
