@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1995 University of Melbourne.
+% Copyright (C) 1996-1997 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -20,6 +20,10 @@
 % called predicate (e.g. if a predicate is overloaded with both
 % `ui' and `in' modes)?
 
+% Variables can become nondet live in several places:
+% in negations, in the conditions of if-then-elses,
+% and in disjunctions, and at nondet calls.
+
 % XXX we currently make the conservative assumption that
 % any non-local variable in a disjunction or nondet call
 % is nondet-live - and stays nondet-live.
@@ -41,7 +45,7 @@
 :- mode unique_modes__check_proc(in, in, in, out, di, uo) is det.
 
 	% just check a single goal
-:- pred unique_modes__check_goal(hlds__goal, hlds__goal, mode_info, mode_info).
+:- pred unique_modes__check_goal(hlds_goal, hlds_goal, mode_info, mode_info).
 :- mode unique_modes__check_goal(in, out, mode_info_di, mode_info_uo) is det.
 
 %-----------------------------------------------------------------------------%
@@ -49,9 +53,10 @@
 
 :- implementation.
 
-:- import_module hlds_data, mode_debug, modecheck_unify, instmap.
+:- import_module hlds_data, mode_debug, modecheck_unify.
 :- import_module mode_util, prog_out, hlds_out, mercury_to_mercury, passes_aux.
-:- import_module modes, inst_match, prog_data, mode_errors, llds, unify_proc.
+:- import_module modes, prog_data, mode_errors, llds, unify_proc.
+:- import_module (inst), instmap, inst_match, inst_util.
 :- import_module bool, int, list, map, set, std_util, require, term, varset.
 :- import_module assoc_list.
 
@@ -130,24 +135,13 @@ unique_modes__check_proc_2(ProcInfo0, PredId, ProcId, ModuleInfo0,
 	proc_info_arglives(ProcInfo0, ModuleInfo0, ArgLives),
 	proc_info_goal(ProcInfo0, Goal0),
 
-/**************
-	%
-	% extract the predicate's type from the pred_info
-	% and propagate the type information into the modes
-	%
-	pred_info_arg_types(PredInfo, _TypeVars, ArgTypes),
-	propagate_type_info_mode_list(ArgTypes, ModuleInfo0, ArgModes0,
-			ArgModes),
-**************/
-
 	%
 	% Figure out the right context to use.
 	% We use the context of the first clause, unless
 	% there weren't any clauses at all, in which case
 	% we use the context of the mode declaration.
 	%
-	module_info_preds(ModuleInfo0, Preds),
-	map__lookup(Preds, PredId, PredInfo),
+	module_info_pred_info(ModuleInfo0, PredId, PredInfo),
 	pred_info_clauses_info(PredInfo, ClausesInfo),
 	ClausesInfo = clauses_info(_, _, _, _, ClauseList),
 	( ClauseList = [FirstClause | _] ->
@@ -326,7 +320,7 @@ select_changed_inst_vars([Var | Vars], DeltaInstMap, ModeInfo, ChangedVars) :-
 	instmap__lookup_var(InstMap0, Var, Inst0),
 	(
 		instmap_delta_is_reachable(DeltaInstMap),
-		instmap_delta_lookup_var(DeltaInstMap, Var, Inst),
+		instmap_delta_search_var(DeltaInstMap, Var, Inst),
 		\+ inst_matches_final(Inst, Inst0, ModuleInfo)
 	->
 		ChangedVars = [Var | ChangedVars1],
@@ -373,8 +367,8 @@ make_var_mostly_uniq(Var, ModeInfo0, ModeInfo) :-
 		ModeInfo = ModeInfo0
 	).
 
-:- pred unique_modes__check_goal_2(hlds__goal_expr, hlds__goal_info,
-		hlds__goal_expr, mode_info, mode_info).
+:- pred unique_modes__check_goal_2(hlds_goal_expr, hlds_goal_info,
+		hlds_goal_expr, mode_info, mode_info).
 :- mode unique_modes__check_goal_2(in, in, out, mode_info_di, mode_info_uo)
 		is det.
 
@@ -463,32 +457,6 @@ unique_modes__check_goal_2(if_then_else(Vs, A0, B0, C0, SM), GoalInfo0, Goal)
 	mode_info_dcg_get_instmap(InstMapC),
 	mode_info_set_instmap(InstMap0),
 	instmap__merge(NonLocals, [InstMapB, InstMapC], if_then_else),
-/*
-% this optimization from modes.m does not need to be repeated here
-	( { InstMapA = unreachable } ->
-		% if the condition can never succeed, we delete the
-		% unreachable `then' part by replacing
-		%	if some [Vs] A then B else C
-		% with
-		%	(not some [Vs] A), C
-		{ A = _A_Goal - A_GoalInfo },
-		{ goal_info_get_nonlocals(A_GoalInfo, A_Vars) },
-		{ set__union(NonLocals, C_Vars, OutsideVars) },
-		{ set__delete_list(OutsideVars, Vs, OutsideVars1) },
-		{ set__intersect(OutsideVars1, A_Vars, SomeA_NonLocals) },
-		{ goal_info_init(EmptyGoalInfo) },
-		{ goal_info_set_nonlocals(EmptyGoalInfo, SomeA_NonLocals,
-			SomeA_GoalInfo) },
-		{ map__init(EmptyInstmapDelta) },
-		{ goal_info_set_instmap_delta(SomeA_GoalInfo,
-			reachable(EmptyInstmapDelta), NotSomeA_GoalInfo) },
-		
-		{ Goal = conj([not(some(Vs, A) - SomeA_GoalInfo) -
-				NotSomeA_GoalInfo, C]) }
-	;
-		{ Goal = if_then_else(Vs, A, B, C) }
-	),
-*/
 	{ Goal = if_then_else(Vs, A, B, C, SM) },
 	mode_checkpoint(exit, "if-then-else").
 
@@ -511,10 +479,10 @@ unique_modes__check_goal_2(some(Vs, G0), _, some(Vs, G)) -->
 	unique_modes__check_goal(G0, G),
 	mode_checkpoint(exit, "some").
 
-unique_modes__check_goal_2(higher_order_call(PredVar, Args, Types, Modes, Det),
-		_GoalInfo0, Goal) -->
+unique_modes__check_goal_2(higher_order_call(PredVar, Args, Types, Modes, Det,
+		PredOrFunc), _GoalInfo0, Goal) -->
 	mode_checkpoint(enter, "higher-order call"),
-	mode_info_set_call_context(higher_order_call(predicate)),
+	mode_info_set_call_context(higher_order_call(PredOrFunc)),
 	{ determinism_components(Det, _, at_most_zero) ->
 		NeverSucceeds = yes
 	;
@@ -522,7 +490,8 @@ unique_modes__check_goal_2(higher_order_call(PredVar, Args, Types, Modes, Det),
 	},
 	{ determinism_to_code_model(Det, CodeModel) },
 	unique_modes__check_call_modes(Args, Modes, CodeModel, NeverSucceeds),
-	{ Goal = higher_order_call(PredVar, Args, Types, Modes, Det) },
+	{ Goal = higher_order_call(PredVar, Args, Types, Modes, Det,
+			PredOrFunc) },
 	mode_info_unset_call_context,
 	mode_checkpoint(exit, "higher-order call").
 
@@ -563,12 +532,13 @@ unique_modes__check_goal_2(switch(Var, CanFail, Cases0, SM), GoalInfo0,
 	% to modecheck a pragma_c_code, we just modecheck the proc for 
 	% which it is the goal.
 unique_modes__check_goal_2(pragma_c_code(IsRecursive, C_Code, PredId, ProcId,
-		Args, ArgNameMap), _GoalInfo, Goal) -->
+		Args, ArgNameMap, OrigArgTypes, ExtraPragmaInfo),
+		_GoalInfo, Goal) -->
 	mode_checkpoint(enter, "pragma_c_code"),
 	mode_info_set_call_context(call(PredId)),
 	unique_modes__check_call(PredId, ProcId, Args),
 	{ Goal = pragma_c_code(IsRecursive, C_Code, PredId, ProcId, Args,
-			ArgNameMap) },
+			ArgNameMap, OrigArgTypes, ExtraPragmaInfo) },
 	mode_info_unset_call_context,
 	mode_checkpoint(exit, "pragma_c_code").
 
@@ -596,16 +566,9 @@ unique_modes__check_call(PredId, ProcId, ArgVars,
 :- mode unique_modes__check_call_modes(in, in, in, in,
 			mode_info_di, mode_info_uo) is det.
 
-unique_modes__check_call_modes(ArgVars, ProcArgModes0, CodeModel, NeverSucceeds,
+unique_modes__check_call_modes(ArgVars, ProcArgModes, CodeModel, NeverSucceeds,
 			ModeInfo0, ModeInfo) :-
 	mode_info_get_module_info(ModeInfo0, ModuleInfo),
-/*********************
-	% propagate type info into modes
-	mode_info_get_types_of_vars(ModeInfo0, ArgVars, ArgTypes),
-	propagate_type_info_mode_list(ArgTypes, ModuleInfo,
-		ProcArgModes0, ProcArgModes),
-*********************/
-	ProcArgModes = ProcArgModes0,
 	mode_list_get_initial_insts(ProcArgModes, ModuleInfo,
 				InitialInsts),
 	modecheck_var_has_inst_list(ArgVars, InitialInsts, 0,
@@ -613,7 +576,7 @@ unique_modes__check_call_modes(ArgVars, ProcArgModes0, CodeModel, NeverSucceeds,
 	mode_list_get_final_insts(ProcArgModes, ModuleInfo, FinalInsts),
 	modecheck_set_var_inst_list(ArgVars, InitialInsts, FinalInsts,
 		NewArgVars, ExtraGoals, ModeInfo1, ModeInfo2),
-	( NewArgVars = ArgVars, ExtraGoals = [] - [] ->
+	( NewArgVars = ArgVars, ExtraGoals = no_extra_goals ->
 		true
 	;	
 		% this shouldn't happen, since modes.m should do
@@ -639,7 +602,7 @@ unique_modes__check_call_modes(ArgVars, ProcArgModes0, CodeModel, NeverSucceeds,
 
 %-----------------------------------------------------------------------------%
 
-:- pred unique_modes__check_conj(list(hlds__goal), list(hlds__goal),
+:- pred unique_modes__check_conj(list(hlds_goal), list(hlds_goal),
 		mode_info, mode_info).
 :- mode unique_modes__check_conj(in, out, mode_info_di, mode_info_uo) is det.
 
@@ -659,7 +622,7 @@ unique_modes__check_conj([Goal0 | Goals0], [Goal | Goals]) -->
 	% the original instmap before processing the next one.
 	% Collect up a list of the resulting instmaps.
 
-:- pred unique_modes__check_disj(list(hlds__goal), list(hlds__goal),
+:- pred unique_modes__check_disj(list(hlds_goal), list(hlds_goal),
 		list(instmap), mode_info, mode_info).
 :- mode unique_modes__check_disj(in, out, out, mode_info_di, mode_info_uo)
 		is det.
@@ -689,26 +652,20 @@ unique_modes__check_case_list([Case0 | Cases0], Var,
 
 		% record the fact that Var was bound to ConsId in the
 		% instmap before processing this case
-	( { cons_id_to_const(ConsId, _Const, Arity) } ->
-		{ list__duplicate(Arity, free, ArgInsts) },
-		modecheck_set_var_inst(Var,
-			bound(unique, [functor(ConsId, ArgInsts)]))
-	;
-		% cons_id_to_const will fail for pred_consts and
-		% address_consts; we don't worry about them,
-		% since you can't have a switch on a higher-order
-		% pred term anyway.
-		[]
-	),
+	{ cons_id_arity(ConsId, Arity) },
+	{ list__duplicate(Arity, free, ArgInsts) },
+	modecheck_set_var_inst(Var,
+		bound(unique, [functor(ConsId, ArgInsts)])),
 
-	unique_modes__check_goal(Goal0, Goal),
+	unique_modes__check_goal(Goal0, Goal1),
 	mode_info_dcg_get_instmap(InstMap),
+	{ fixup_switch_var(Var, InstMap0, InstMap, Goal1, Goal) },
 	mode_info_set_instmap(InstMap0),
 	unique_modes__check_case_list(Cases0, Var, Cases, InstMaps).
 
 %-----------------------------------------------------------------------------%
 
-:- pred unique_modes__goal_get_nonlocals(hlds__goal, set(var)).
+:- pred unique_modes__goal_get_nonlocals(hlds_goal, set(var)).
 :- mode unique_modes__goal_get_nonlocals(in, out) is det.
 
 unique_modes__goal_get_nonlocals(_Goal - GoalInfo, NonLocals) :-
