@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1995 University of Melbourne.
+% Copyright (C) 1994-1997 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -14,7 +14,7 @@
 
 :- interface.
 
-:- import_module bool, list, io.
+:- import_module list, io.
 :- import_module llds.
 
 	% Find straight-line code sequences and optimize them using
@@ -40,6 +40,7 @@
 :- implementation.
 
 :- import_module set, map, bimap, require, int, string, std_util, assoc_list.
+:- import_module bool.
 
 :- import_module vn_type, vn_table, vn_block, vn_order, vn_flush, vn_temploc. 
 :- import_module vn_cost, vn_debug, vn_util, vn_verify, vn_filter.
@@ -50,9 +51,10 @@
 	% so we don't optimize any predicates containing them.
 
 value_number__main(Instrs0, Instrs) -->
-	{ opt_util__get_prologue(Instrs0, ProcLabel, Comments, Instrs1) },
+	{ opt_util__get_prologue(Instrs0, ProcLabel,
+		LabelInstr, Comments, Instrs1) },
 	{ opt_util__new_label_no(Instrs1, 1000, N0) },
-	{ value_number__prepare_for_vn(Instrs1, ProcLabel,
+	{ value_number__prepare_for_vn([LabelInstr | Instrs1], ProcLabel,
 		no, AllocSet, BreakSet, N0, N, Instrs2) },
 	{ labelopt__build_useset(Instrs2, UseSet) },
 	{ livemap__build(Instrs2, MaybeLiveMap) },
@@ -406,8 +408,7 @@ value_number__optimize_fragment_2(Instrs0, LiveMap, Params, ParEntries,
 		{ value_number__push_livevals_back(Instrs3, Instrs4) },
 		{ value_number__convert_back_modframe(Instrs4, Instrs5) },
 		{ vn_filter__block(Instrs5, Instrs6) },
-		{ bimap__init(TeardownMap) },
-		{ peephole__optimize(Instrs6, Instrs7, TeardownMap, no, _) },
+		{ peephole__optimize(Instrs6, Instrs7, _) },
 
 		vn_debug__cost_header_msg("original code sequence"),
 		vn_cost__block_cost(Instrs0, Params, yes, OrigCost),
@@ -536,19 +537,33 @@ value_number__try_again([Instr0 | Instrs0], RevInstrs0, RestartLabel, LiveMap,
 			{ Instrs = [Instr0 | Instrs1] }
 		;
 			vn_debug__divide_msg(Instr0),
+			%
+			% we need to append a `goto' instruction at the
+			% end of the fragment so that optimize_fragment
+			% can know what is live at the end.
+			%
 			{ GotoInstr = goto(label(RestartLabel)) - "" },
 			{ list__reverse([GotoInstr | RevInstrs0],
 				FrontInstrs0) },
 			value_number__optimize_fragment(FrontInstrs0, LiveMap,
-				Params, [], LabelNo0, _, FrontInstrs),
+				Params, [], LabelNo0, _, FrontInstrs1),
 			value_number__optimize_fragment(Instrs0, LiveMap,
 				Params, [], LabelNo0, _, BackInstrs),
+			%
+			% we need to get rid of the introduced goto,
+			% which should still be at the end of FrontInstrs1,
+			% otherwise we would violate the invariant that
+			% labels in the middle of a block are not targets
+			% of branches.
+			%
+			{ list__reverse(FrontInstrs1, RevInstrs1) },
+			{ RevInstrs1 = [GotoInstr | RevInstrs2] ->
+				list__reverse(RevInstrs2, FrontInstrs)
+			;
+				error("value_number__try_again: lost goto")
+			},
 			{ list__append(FrontInstrs, [Instr0 | BackInstrs],
-				Instrs1) },
-			{ bimap__init(TeardownMap) },
-			% to get rid of the introduced goto
-			{ peephole__optimize(Instrs1, Instrs, TeardownMap,
-				no, _) }
+				Instrs) }
 		)
 	;
 		{ RevInstrs1 = [Instr0 | RevInstrs0] },
@@ -1047,6 +1062,11 @@ value_number__push_livevals_back_2([Instr0 | Instrs0], Livevals, Instrs) :-
 
 %-----------------------------------------------------------------------------%
 
+	% Return true if we must treat this instruction as a boundary
+	% instruction, i.e. an instruction across which it is neither safe
+	% nor necessary to push stack frame setup or teardown instructions
+	% (incr_sp, livevals and decr_sp).
+
 :- pred value_number__boundary_instr(instr, bool).
 :- mode value_number__boundary_instr(in, out) is det.
 
@@ -1065,12 +1085,14 @@ value_number__boundary_instr(if_val(_, _), yes).
 value_number__boundary_instr(incr_hp(_, _, _), no).
 value_number__boundary_instr(mark_hp(_), no).
 value_number__boundary_instr(restore_hp(_), no).
-value_number__boundary_instr(store_ticket(_), yes).
-value_number__boundary_instr(restore_ticket(_), yes).
-value_number__boundary_instr(discard_ticket, yes).
+value_number__boundary_instr(store_ticket(_), no).
+value_number__boundary_instr(reset_ticket(_, _), no).
+value_number__boundary_instr(discard_ticket, no).
+value_number__boundary_instr(mark_ticket_stack(_), no).
+value_number__boundary_instr(discard_tickets_to(_), no).
 value_number__boundary_instr(incr_sp(_, _), yes).
 value_number__boundary_instr(decr_sp(_), yes).
-value_number__boundary_instr(pragma_c(_, _, _, _), yes).
+value_number__boundary_instr(pragma_c(_, _, _, _, _), yes).
 
 %-----------------------------------------------------------------------------%
 
