@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1995 University of Melbourne.
+% Copyright (C) 1995-1997 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -203,7 +203,25 @@ vn_block__handle_instr(assign(Lval, Rval),
 	vn_util__rval_to_vn(Rval, Vn, VnTables0, VnTables1),
 	vn_util__lval_to_vnlval(Lval, Vnlval, VnTables1, VnTables2),
 	vn_table__set_desired_value(Vnlval, Vn, VnTables2, VnTables),
-	vn_util__find_specials(Vnlval, Specials),
+	vn_util__find_specials(Vnlval, LeftSpecials),
+	(
+		% Assignments of this form occur in the setup for commit.
+		% We must record the left hand side (which will be a stackvar)
+		% as a must flush location, because liveness will not force
+		% it to be flushed. The reason for this is that livemap.m
+		% may be confused by an incr_sp and assignment to a similarly
+		% numbered stackvar in later code into thinking that this
+		% assignment is redundant.
+		Rval = lval(SubLval),
+		( SubLval = curfr
+		; SubLval = maxfr
+		; SubLval = redoip(_)
+		)
+	->
+		Specials = [Vnlval | LeftSpecials]
+	;
+		Specials = LeftSpecials
+	),
 	set__insert_list(Liveset0, Specials, Liveset).
 vn_block__handle_instr(call(Proc, Return, Info, CallModel),
 		Livemap, Params, VnTables0, VnTables, Liveset0, Liveset,
@@ -296,11 +314,11 @@ vn_block__handle_instr(store_ticket(Lval),
 	vn_block__new_ctrl_node(vn_store_ticket(Vnlval), Livemap,
 		Params, VnTables1, VnTables,
 		Liveset0, Liveset, Tuple0, Tuple).
-vn_block__handle_instr(restore_ticket(Rval),
+vn_block__handle_instr(reset_ticket(Rval, Reason),
 		Livemap, Params, VnTables0, VnTables, Liveset0, Liveset,
 		SeenIncr, SeenIncr, Tuple0, Tuple) :-
 	vn_util__rval_to_vn(Rval, Vn, VnTables0, VnTables1),
-	vn_block__new_ctrl_node(vn_restore_ticket(Vn), Livemap,
+	vn_block__new_ctrl_node(vn_reset_ticket(Vn, Reason), Livemap,
 		Params, VnTables1, VnTables,
 		Liveset0, Liveset, Tuple0, Tuple).
 vn_block__handle_instr(discard_ticket,
@@ -309,6 +327,20 @@ vn_block__handle_instr(discard_ticket,
 	vn_block__new_ctrl_node(vn_discard_ticket, Livemap,
 		Params, VnTables0, VnTables,
 		Liveset0, Liveset, Tuple0, Tuple).
+vn_block__handle_instr(mark_ticket_stack(Lval),
+		Livemap, Params, VnTables0, VnTables, Liveset0, Liveset,
+		SeenIncr, SeenIncr, Tuple0, Tuple) :-
+	vn_util__lval_to_vnlval(Lval, Vnlval, VnTables0, VnTables1),
+	vn_block__new_ctrl_node(vn_mark_ticket_stack(Vnlval), Livemap,
+		Params, VnTables1, VnTables,
+		Liveset0, Liveset, Tuple0, Tuple).
+vn_block__handle_instr(discard_tickets_to(Rval),
+		Livemap, Params, VnTables0, VnTables, Liveset0, Liveset,
+		SeenIncr, SeenIncr, Tuple0, Tuple) :-
+	vn_util__rval_to_vn(Rval, Vn, VnTables0, VnTables1),
+	vn_block__new_ctrl_node(vn_discard_tickets_to(Vn), Livemap,
+		Params, VnTables1, VnTables,
+		Liveset0, Liveset, Tuple0, Tuple).
 vn_block__handle_instr(incr_sp(N, Msg),
 		Livemap, Params, VnTables0, VnTables, Liveset0, Liveset,
 		SeenIncr, SeenIncr, Tuple0, Tuple) :-
@@ -321,7 +353,7 @@ vn_block__handle_instr(decr_sp(N),
 	vn_block__new_ctrl_node(vn_decr_sp(N), Livemap,
 		Params, VnTables0, VnTables,
 		Liveset0, Liveset, Tuple0, Tuple).
-vn_block__handle_instr(pragma_c(_, _, _, _),
+vn_block__handle_instr(pragma_c(_, _, _, _, _),
 		_Livemap, _Params, VnTables, VnTables, Liveset, Liveset,
 		SeenIncr, SeenIncr, Tuple, Tuple) :-
 	error("value numbering not supported for pragma_c").
@@ -425,7 +457,7 @@ vn_block__new_ctrl_node(VnInstr, Livemap, Params,
 		LabelNo = LabelNo0,
 		Parallels = []
 	;
-		VnInstr = vn_restore_ticket(_),
+		VnInstr = vn_reset_ticket(_, _),
 		VnTables = VnTables0,
 		Liveset = Liveset0,
 		FlushEntry = FlushEntry0,
@@ -438,6 +470,27 @@ vn_block__new_ctrl_node(VnInstr, Livemap, Params,
 		FlushEntry = FlushEntry0,
 		LabelNo = LabelNo0,
 		Parallels = []
+	;
+		VnInstr = vn_mark_ticket_stack(Vnlval),
+		( vn_table__search_desired_value(Vnlval, Vn_prime, VnTables0) ->
+			Vn = Vn_prime,
+			VnTables1 = VnTables0
+		;
+			vn_table__record_first_vnlval(Vnlval, Vn, 
+				VnTables0, VnTables1)
+		),
+		vn_table__set_desired_value(Vnlval, Vn, VnTables1, VnTables),
+		set__insert(Liveset0, Vnlval, Liveset),
+		FlushEntry = FlushEntry0,
+		LabelNo = LabelNo0,
+		Parallels = []
+	;
+		VnInstr = vn_discard_tickets_to(_),
+		VnTables = VnTables0,
+		Liveset = Liveset0,
+		FlushEntry = FlushEntry0,
+		LabelNo = LabelNo0,
+		Parallels = []
 	;
 		VnInstr = vn_incr_sp(_, _),
 		VnTables = VnTables0,
@@ -779,7 +832,7 @@ vn_block__find_cheaper_copies_2([Vnlval - Vn | VnlvalVns], SeekVn, OldCost,
 			MaybeLval = yes(Lval),
 			vn_cost__lval_cost(Lval, Params, LvalCost),
 			LvalCost < OldCost,
-			Lval \= temp(_),
+			Lval \= temp(_, _),
 			Lval \= hp
 		->
 			Rvals = [lval(Lval) | Rvals0]
@@ -808,6 +861,11 @@ vn_block__split_at_next_ctrl_instr([Instr0 | Instrs0], Before, Instr, After) :-
 		Before = [Instr0 | Before0]
 	).
 
+	% Return true if we need to create a control node for this instruction.
+	% We do not create control nodes for instructions that either (a)
+	% can be optimized away by value numbering or (b) cannot happen in
+	% code sequences presented to value numbering.
+
 :- pred vn_block__is_ctrl_instr(instr, bool).
 :- mode vn_block__is_ctrl_instr(in, out) is det.
 
@@ -827,11 +885,13 @@ vn_block__is_ctrl_instr(incr_hp(_, _, _), no).
 vn_block__is_ctrl_instr(mark_hp(_), yes).
 vn_block__is_ctrl_instr(restore_hp(_), yes).
 vn_block__is_ctrl_instr(store_ticket(_), yes).
-vn_block__is_ctrl_instr(restore_ticket(_), yes).
+vn_block__is_ctrl_instr(reset_ticket(_, _), yes).
 vn_block__is_ctrl_instr(discard_ticket, yes).
+vn_block__is_ctrl_instr(mark_ticket_stack(_), yes).
+vn_block__is_ctrl_instr(discard_tickets_to(_), yes).
 vn_block__is_ctrl_instr(incr_sp(_, _), yes).
 vn_block__is_ctrl_instr(decr_sp(_), yes).
-vn_block__is_ctrl_instr(pragma_c(_, _, _, _), no).
+vn_block__is_ctrl_instr(pragma_c(_, _, _, _, _), no).
 
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
