@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1995 University of Melbourne.
+% Copyright (C) 1995-1997 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -250,16 +250,32 @@ vn_flush__ctrl_node(Vn_instr, N, VnTables0, VnTables, Templocs0, Templocs,
 		Instr = store_ticket(Lval) - "",
 		list__append(FlushInstrs, [Instr], Instrs)
 	;
-		Vn_instr = vn_restore_ticket(Vn),
+		Vn_instr = vn_reset_ticket(Vn, Reason),
 		vn_flush__vn(Vn, [src_ctrl(N)], [], Rval, VnTables0, VnTables,
 			Templocs0, Templocs, Params, FlushInstrs),
-		Instr = restore_ticket(Rval) - "",
+		Instr = reset_ticket(Rval, Reason) - "",
 		list__append(FlushInstrs, [Instr], Instrs)
 	;
 		Vn_instr = vn_discard_ticket,
 		VnTables = VnTables0,
 		Templocs = Templocs0,
 		Instrs = [discard_ticket - ""]
+	;
+		Vn_instr = vn_mark_ticket_stack(Vnlval),
+		vn_flush__access_path(Vnlval, [src_ctrl(N)], [], Lval,
+			VnTables0, VnTables1, Templocs0, Templocs, Params,
+			FlushInstrs),
+		vn_table__lookup_assigned_vn(vn_origlval(Vnlval), OldVn,
+			"vn_flush__ctrl_node", VnTables1),
+		vn_table__set_current_value(Vnlval, OldVn, VnTables1, VnTables),
+		Instr = mark_ticket_stack(Lval) - "",
+		list__append(FlushInstrs, [Instr], Instrs)
+	;
+		Vn_instr = vn_discard_tickets_to(Vn),
+		vn_flush__vn(Vn, [src_ctrl(N)], [], Rval, VnTables0, VnTables,
+			Templocs0, Templocs, Params, FlushInstrs),
+		Instr = discard_tickets_to(Rval) - "",
+		list__append(FlushInstrs, [Instr], Instrs)
 	;
 		Vn_instr = vn_incr_sp(Incr, Msg),
 		VnTables = VnTables0,
@@ -392,7 +408,7 @@ vn_flush__find_cheap_users_2([Src | Srcs], Vnlvals, VnTables) :-
 					vn_table__search_current_value(User,
 						UserVn, VnTables)
 				->
-					User = vn_reg(_),
+					User = vn_reg(_, _),
 					vn_table__lookup_uses(UserVn, [],
 						"vn_flush__find_cheap_users_2",
 						VnTables)
@@ -694,6 +710,24 @@ vn_flush__vn_value(Vn, Srcs, Forbidden, Rval, VnTables0, VnTables,
 			Instrs2),
 		Rval = binop(Binop, Rval1, Rval2),
 		list__append(Instrs1, Instrs2, Instrs)
+	;
+		Vnrval = vn_stackvar_addr(N),
+		Rval = mem_addr(stackvar_ref(N)),
+		VnTables = VnTables0,
+		Templocs = Templocs0,
+		Instrs = []
+	;
+		Vnrval = vn_framevar_addr(N),
+		Rval = mem_addr(framevar_ref(N)),
+		VnTables = VnTables0,
+		Templocs = Templocs0,
+		Instrs = []
+	;
+		Vnrval = vn_heap_addr(SubVn1, Tag, Field),
+		vn_flush__vn(SubVn1, [src_vn(Vn) | Srcs], Forbidden, Rval1,
+			VnTables0, VnTables, Templocs0, Templocs, Params,
+			Instrs),
+		Rval = mem_addr(heap_ref(Rval1, Tag, Field))
 	).
 
 %-----------------------------------------------------------------------------%
@@ -763,7 +797,7 @@ vn_flush__old_hp(Srcs0, Forbidden0, ReturnRval, VnTables0, VnTables,
 		; 
 			vn_flush__find_cheap_users(UserVn, UserLocs, VnTables3),
 			vn_util__choose_cheapest_loc(UserLocs, UserLoc),
-			UserLoc = vn_reg(_)
+			UserLoc = vn_reg(_, _)
 		->
 			Vnlval = UserLoc,
 			vn_util__no_access_vnlval_to_lval(Vnlval, MaybeLval),
@@ -897,6 +931,15 @@ vn_flush__hp_incr(Vn, Srcs, Forbidden, MaybeRval, VnTables0, VnTables,
 				MaybeRval2 = no,
 				error("two 'no's in flush_hp_incr")
 			)
+		;
+			Vnrval = vn_stackvar_addr(_),
+			error("stackvar_addr in calculation of new hp")
+		;
+			Vnrval = vn_framevar_addr(_),
+			error("framevar_addr in calculation of new hp")
+		;
+			Vnrval = vn_heap_addr(_, _, _),
+			error("heap_addr in calculation of new hp")
 		),
 		( Srcs = [SrcPrime | _] ->
 			Src = SrcPrime
@@ -945,8 +988,14 @@ vn_flush__rec_find_ref_vns_list([Vn | Vns], SubVns, VnTables) :-
 vn_flush__access_path(Vnlval, Srcs, Forbidden, Lval, VnTables0, VnTables,
 		Templocs0, Templocs, Params, AccessInstrs) :-
 	(
-		Vnlval = vn_reg(Reg),
-		Lval = reg(Reg),
+		Vnlval = vn_reg(Type, Num),
+		Lval = reg(Type, Num),
+		VnTables = VnTables0,
+		Templocs = Templocs0,
+		AccessInstrs = []
+	;
+		Vnlval = vn_temp(Type, Num),
+		Lval = temp(Type, Num),
 		VnTables = VnTables0,
 		Templocs = Templocs0,
 		AccessInstrs = []
@@ -1027,11 +1076,11 @@ vn_flush__access_path(Vnlval, Srcs, Forbidden, Lval, VnTables0, VnTables,
 		Lval = field(Tag, Rval1, Rval2),
 		list__append(AccessInstrs1, AccessInstrs2, AccessInstrs)
 	;
-		Vnlval = vn_temp(Num),
-		Lval = temp(Num),
-		VnTables = VnTables0,
-		Templocs = Templocs0,
-		AccessInstrs = []
+		Vnlval = vn_mem_ref(Vn),
+		vn_flush__vn(Vn, [src_access(Vnlval) | Srcs], Forbidden, Rval,
+			VnTables0, VnTables,
+			Templocs0, Templocs, Params, AccessInstrs),
+		Lval = mem_ref(Rval)
 	).
 
 %-----------------------------------------------------------------------------%
@@ -1077,11 +1126,16 @@ vn_flush__maybe_save_prev_value(Vnlval, OldVn, NewVn, Forbidden,
 				MaybePresumed),
 			(
 				MaybePresumed = yes(PresumedLval),
-				( list__member(PresumedLval, Forbidden) ->
+				(
+					list__member(PresumedLval, Forbidden)
+				->
 					vn_flush__choose_temp(OldVn,
 						VnTables0, Templocs0,
 						Templocs1, Chosen)
-				; RealUses = [_,_|_], \+ Presumed = vn_reg(_) ->
+				;
+					RealUses = [_,_|_],
+					\+ Presumed = vn_reg(_, _)
+				->
 					vn_flush__choose_temp(OldVn,
 						VnTables0, Templocs0,
 						Templocs1, Chosen)
