@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1995 University of Melbourne.
+% Copyright (C) 1995-1997 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -370,7 +370,7 @@ vn_order__vn_ctrl_order(Ctrl, Ctrlmap, VnTables0, VnTables,
 				VnTables0, VnTables1,
 				Succmap0, Succmap1, Predmap0, Predmap1)
 		;
-			Vn_instr = vn_restore_ticket(Vn),
+			Vn_instr = vn_reset_ticket(Vn, _Reason),
 			vn_order__find_links(Vn, node_ctrl(Ctrl),
 				VnTables0, VnTables1,
 				Succmap0, Succmap1, Predmap0, Predmap1)
@@ -379,6 +379,17 @@ vn_order__vn_ctrl_order(Ctrl, Ctrlmap, VnTables0, VnTables,
 			Succmap1 = Succmap0,
 			Predmap1 = Predmap0,
 			VnTables1 = VnTables0
+		;
+			Vn_instr = vn_mark_ticket_stack(Vnlval),
+			vn_util__vnlval_access_vns(Vnlval, Vns),
+			vn_order__find_all_links(Vns, node_ctrl(Ctrl),
+				VnTables0, VnTables1,
+				Succmap0, Succmap1, Predmap0, Predmap1)
+		;
+			Vn_instr = vn_discard_tickets_to(Vn),
+			vn_order__find_links(Vn, node_ctrl(Ctrl),
+				VnTables0, VnTables1,
+				Succmap0, Succmap1, Predmap0, Predmap1)
 		;
 			Vn_instr = vn_incr_sp(_, _),
 			Succmap1 = Succmap0,
@@ -461,12 +472,8 @@ vn_order__use_sink_before_redef(Sink, VnTables, Liveset,
 
 		vn_order__add_users(Users, Sink, Vnlval, VnTables, Liveset,
 			Succmap0, Succmap1, Predmap0, Predmap1),
-		(
-			{ vn_table__search_desired_value(Vnlval, Vn,
-				VnTables) },
-			{ map__search(Succmap1, node_shared(Vn), _) }
-		->
-			vn_order__add_users(Users, node_shared(Vn), Vnlval,
+		( { map__search(Succmap1, node_shared(NewVn), _) } ->
+			vn_order__add_users(Users, node_shared(NewVn), Vnlval,
 				VnTables, Liveset,
 				Succmap1, Succmap, Predmap1, Predmap)
 		;
@@ -642,6 +649,20 @@ vn_order__find_links(Vn, Sink, VnTables0, VnTables,
 				Succmap0, Succmap1, Predmap0, Predmap1),
 			vn_order__find_links(SubVn2, Sink, VnTables1, VnTables,
 				Succmap1, Succmap, Predmap1, Predmap)
+		;
+			Vnrval = vn_stackvar_addr(_),
+			Succmap = Succmap0,
+			Predmap = Predmap0,
+			VnTables = VnTables0
+		;
+			Vnrval = vn_framevar_addr(_),
+			Succmap = Succmap0,
+			Predmap = Predmap0,
+			VnTables = VnTables0
+		;
+			Vnrval = vn_heap_addr(SubVn, _, _),
+			vn_order__find_links(SubVn, Sink, VnTables0, VnTables,
+				Succmap0, Succmap, Predmap0, Predmap)
 		)
 	).
 
@@ -688,26 +709,26 @@ vn_order__add_link(Source, Sink, Succmap0, Succmap, Predmap0, Predmap) :-
 	;
 		( map__search(Succmap0, Source, Sinks0) ->
 			vn_order__insert_if_not_there(Sink, Sinks0, Sinks1),
-			map__set(Succmap0, Source, Sinks1, Succmap1)
+			map__det_update(Succmap0, Source, Sinks1, Succmap1)
 		;
-			map__set(Succmap0, Source, [Sink], Succmap1)
+			map__det_insert(Succmap0, Source, [Sink], Succmap1)
 		),
-		( map__search(Succmap0, Sink, _) ->
+		( map__search(Succmap1, Sink, _) ->
 			Succmap = Succmap1
 		;
-			map__set(Succmap1, Sink, [], Succmap)
+			map__det_insert(Succmap1, Sink, [], Succmap)
 		),
 		( map__search(Predmap0, Sink, Sources0) ->
 			vn_order__insert_if_not_there(Source,
 				Sources0, Sources1),
-			map__set(Predmap0, Sink, Sources1, Predmap1)
+			map__det_update(Predmap0, Sink, Sources1, Predmap1)
 		;
-			map__set(Predmap0, Sink, [Source], Predmap1)
+			map__det_insert(Predmap0, Sink, [Source], Predmap1)
 		),
-		( map__search(Predmap0, Source, _) ->
+		( map__search(Predmap1, Source, _) ->
 			Predmap = Predmap1
 		;
-			map__set(Predmap1, Source, [], Predmap)
+			map__det_insert(Predmap1, Source, [], Predmap)
 		)
 	).
 
@@ -733,12 +754,12 @@ vn_order__add_node(Node, Succmap0, Succmap, Predmap0, Predmap) :-
 	( map__search(Succmap0, Node, _) ->
 		Succmap = Succmap0
 	;
-		map__set(Succmap0, Node, [], Succmap)
+		map__det_insert(Succmap0, Node, [], Succmap)
 	),
 	( map__search(Predmap0, Node, _) ->
 		Predmap = Predmap0
 	;
-		map__set(Predmap0, Node, [], Predmap)
+		map__det_insert(Predmap0, Node, [], Predmap)
 	).
 
 %-----------------------------------------------------------------------------%
@@ -900,7 +921,7 @@ vn_order__find_regs([], Regs, []) :-
 	set__init(Regs).
 vn_order__find_regs([Node0 | Nodes0], Regs, Nodes) :-
 	vn_order__find_regs(Nodes0, Regs1, Nodes1),
-	( Node0 = node_lval(vn_reg(_)) ->
+	( Node0 = node_lval(vn_reg(_, _)) ->
 		set__insert(Regs1, Node0, Regs),
 		Nodes = Nodes1
 	;
