@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1995 University of Melbourne.
+% Copyright (C) 1995-1997 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -15,11 +15,11 @@
 :- interface.
 
 :- import_module vn_type.
-:- import_module list, map.
+:- import_module assoc_list, list, map.
 
 :- type vn_tables.
 
-% The definitions of the types are exported only for debugging.
+% The definitions of the following types are exported only for debugging.
 
 :- type lval_to_vn_table == map(vnlval, vn).
 :- type rval_to_vn_table == map(vnrval, vn).
@@ -107,6 +107,10 @@
 % :- mode vn_table__record_first_vnrval(in, out, di, uo) is det.
 :- mode vn_table__record_first_vnrval(in, out, in, out) is det.
 
+:- pred vn_table__record_new_vnrval(vnrval, vn, vn_tables, vn_tables).
+% :- mode vn_table__record_first_vnrval(in, out, di, uo) is det.
+:- mode vn_table__record_new_vnrval(in, out, in, out) is det.
+
 :- pred vn_table__record_first_vnlval(vnlval, vn, vn_tables, vn_tables).
 % :- mode vn_table__record_first_vnlval(in, out, di, uo) is det.
 :- mode vn_table__record_first_vnlval(in, out, in, out) is det.
@@ -294,32 +298,37 @@ vn_table__add_new_use(Vn, NewUse, VnTables0, VnTables) :-
 	;
 		error("cannot find old use set in add_new_use")
 	),
-	(
-		list__member(NewUse, Uses0),
-		\+ (
-			NewUse = src_access(_)
-		;
-			NewUse = src_vn(UserVn),
-			map__lookup(Vn_to_rval_table0, UserVn, VnRval),
-			VnRval = vn_binop(Operator, _, _),
-			( Operator = (+)
-			; Operator = (*)
-			; Operator = float_plus
-			; Operator = float_times
-			)
-		)
-	->
-		opt_debug__dump_tables(VnTables0, T_str),
-		opt_debug__dump_vn(Vn, V_str),
-		opt_debug__dump_use(NewUse, U_str),
-		string__append_list(["\n", T_str, "\n",
-			"new use for vn ", V_str, " = ", U_str, "\n",
-			"new use already known"], Msg),
-		error(Msg)
-	;
-		Uses1 = [NewUse | Uses0]
-	),
-	map__set(Vn_to_uses_table0, Vn, Uses1, Vn_to_uses_table1),
+%		Reenable the commented out stuff if you want the compiler
+%		to fail when an operator is used with two identical operands
+%		in a situation that might be optimizable.
+%	(
+%		list__member(NewUse, Uses0),
+%		\+ (
+%			NewUse = src_access(_)
+%		;
+%			NewUse = src_vn(UserVn),
+%			map__lookup(Vn_to_rval_table0, UserVn, VnRval),
+%			VnRval = vn_binop(Operator, _, _),
+%			( Operator = (+)
+%			; Operator = (*)
+%			; Operator = float_plus
+%			; Operator = float_times
+%			; Operator = (<<)
+%			; Operator = (>>)
+%			)
+%		)
+%	->
+%		opt_debug__dump_tables(VnTables0, T_str),
+%		opt_debug__dump_vn(Vn, V_str),
+%		opt_debug__dump_use(NewUse, U_str),
+%		string__append_list(["\n", T_str, "\n",
+%			"new use for vn ", V_str, " = ", U_str, "\n",
+%			"new use already known"], Msg),
+%		error(Msg)
+%	;
+		Uses1 = [NewUse | Uses0],
+%	),
+	map__det_update(Vn_to_uses_table0, Vn, Uses1, Vn_to_uses_table1),
 	VnTables = vn_tables(NextVn0,
 		Lval_to_vn_table0, Rval_to_vn_table0,
 		Vn_to_rval_table0, Vn_to_uses_table1,
@@ -344,7 +353,7 @@ vn_table__del_old_use(Vn, OldUse, VnTables0, VnTables) :-
 		% src_liveval for a shared vn
 		Uses1 = Uses0
 	),
-	map__set(Vn_to_uses_table0, Vn, Uses1, Vn_to_uses_table1),
+	map__det_update(Vn_to_uses_table0, Vn, Uses1, Vn_to_uses_table1),
 	VnTables = vn_tables(NextVn0,
 		Lval_to_vn_table0, Rval_to_vn_table0,
 		Vn_to_rval_table0, Vn_to_uses_table1,
@@ -371,6 +380,26 @@ vn_table__record_first_vnrval(Vnrval, Vn, VnTables0, VnTables) :-
 		Vn_to_rval_table1, Vn_to_uses_table1,
 		Vn_to_locs_table1, Loc_to_vn_table0).
 
+vn_table__record_new_vnrval(Vnrval, Vn, VnTables0, VnTables) :-
+	VnTables0 = vn_tables(NextVn0,
+		Lval_to_vn_table0, Rval_to_vn_table0,
+		Vn_to_rval_table0, Vn_to_uses_table0,
+		Vn_to_locs_table0, Loc_to_vn_table0),
+	Vn = NextVn0,
+	NextVn1 is NextVn0 + 1,
+	( map__insert(Rval_to_vn_table0, Vnrval, Vn, NewRval_to_vn_table) ->
+		Rval_to_vn_table1 = NewRval_to_vn_table
+	;
+		Rval_to_vn_table1 = Rval_to_vn_table0
+	),
+	map__det_insert(Vn_to_rval_table0, Vn, Vnrval, Vn_to_rval_table1),
+	map__det_insert(Vn_to_uses_table0, Vn, [], Vn_to_uses_table1),
+	map__det_insert(Vn_to_locs_table0, Vn, [], Vn_to_locs_table1),
+	VnTables = vn_tables(NextVn1,
+		Lval_to_vn_table0, Rval_to_vn_table1,
+		Vn_to_rval_table1, Vn_to_uses_table1,
+		Vn_to_locs_table1, Loc_to_vn_table0).
+
 vn_table__record_first_vnlval(Vnlval, Vn, VnTables0, VnTables) :-
 	VnTables0 = vn_tables(NextVn0,
 		Lval_to_vn_table0, Rval_to_vn_table0,
@@ -397,7 +426,8 @@ vn_table__set_desired_value(Vnlval, Vn, VnTables0, VnTables) :-
 		Vn_to_rval_table0, Vn_to_uses_table0,
 		Vn_to_locs_table0, Loc_to_vn_table0),
 	( map__search(Loc_to_vn_table0, Vnlval, _) ->
-		map__set(Lval_to_vn_table0, Vnlval, Vn, Lval_to_vn_table1),
+		map__det_update(Lval_to_vn_table0, Vnlval, Vn,
+			Lval_to_vn_table1),
 		VnTables = vn_tables(NextVn0,
 			Lval_to_vn_table1, Rval_to_vn_table0,
 			Vn_to_rval_table0, Vn_to_uses_table0,
@@ -429,27 +459,30 @@ vn_table__set_current_value(Vnlval, Vn, VnTables0, VnTables) :-
 	( map__search(Loc_to_vn_table0, Vnlval, OldVn) ->
 
 		% change the forward mapping
-		map__set(Loc_to_vn_table0, Vnlval, Vn, Loc_to_vn_table1),
+		map__det_update(Loc_to_vn_table0, Vnlval, Vn, Loc_to_vn_table1),
 
 		% change the reverse mapping, first for old vn, then the new
 		map__lookup(Vn_to_locs_table0, OldVn, OldLocs0),
 		list__delete_all(OldLocs0, Vnlval, OldLocs1),
-		map__set(Vn_to_locs_table0, OldVn, OldLocs1, Vn_to_locs_table1),
+		map__det_update(Vn_to_locs_table0, OldVn, OldLocs1,
+			Vn_to_locs_table1),
 
-		map__lookup(Vn_to_locs_table0, Vn, NewLocs0),
+		map__lookup(Vn_to_locs_table1, Vn, NewLocs0),
 		list__append(NewLocs0, [Vnlval], NewLocs1),
-		map__set(Vn_to_locs_table1, Vn, NewLocs1, Vn_to_locs_table2)
+		map__det_update(Vn_to_locs_table1, Vn, NewLocs1,
+			Vn_to_locs_table2)
 	;
 		% The search in the condition can fail for newly introduced
 		% templocs
 
 		% change the forward mapping
-		map__set(Loc_to_vn_table0, Vnlval, Vn, Loc_to_vn_table1),
+		map__det_insert(Loc_to_vn_table0, Vnlval, Vn, Loc_to_vn_table1),
 
 		% change the reverse mapping
 		map__lookup(Vn_to_locs_table0, Vn, NewLocs0),
 		list__append(NewLocs0, [Vnlval], NewLocs1),
-		map__set(Vn_to_locs_table0, Vn, NewLocs1, Vn_to_locs_table2)
+		map__det_update(Vn_to_locs_table0, Vn, NewLocs1,
+			Vn_to_locs_table2)
 	),
 	VnTables = vn_tables(NextVn0,
 		Lval_to_vn_table0, Rval_to_vn_table0,
