@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1995 University of Melbourne.
+% Copyright (C) 1995-1997 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -37,7 +37,8 @@
 
 :- import_module opt_util.
 
-vn_util__find_specials(vn_reg(_), []).
+vn_util__find_specials(vn_reg(_, _), []).
+vn_util__find_specials(vn_temp(_, _), []).
 vn_util__find_specials(vn_stackvar(_), []).
 vn_util__find_specials(vn_framevar(_), []).
 vn_util__find_specials(vn_succip, [vn_succip]).
@@ -50,7 +51,7 @@ vn_util__find_specials(vn_prevfr(Vn), [vn_prevfr(Vn)]).
 vn_util__find_specials(vn_hp, [vn_hp]).
 vn_util__find_specials(vn_sp, [vn_sp]).
 vn_util__find_specials(vn_field(_, _, _), []).
-vn_util__find_specials(vn_temp(_), []).
+vn_util__find_specials(vn_mem_ref(_), []).
 
 vn_util__convert_to_vnlval_and_insert([], Liveset, Liveset).
 vn_util__convert_to_vnlval_and_insert([Lval | Lvals], Liveset0, Liveset) :-
@@ -104,9 +105,6 @@ vn_util__convert_to_vnlval_and_insert([Lval | Lvals], Liveset0, Liveset) :-
 :- pred vn_util__find_sub_vns(vnrval, list(vn)).
 :- mode vn_util__find_sub_vns(in, out) is det.
 
-:- pred vn_util__find_sub_vns_vnlval(vnlval, list(vn)).
-:- mode vn_util__find_sub_vns_vnlval(in, out) is det.
-
 	% Find all lvals inside a given rval.
 
 :- pred vn_util__find_lvals_in_rval(rval, list(lval)).
@@ -172,6 +170,29 @@ vn_util__rval_to_vn(Rval, Vn, VnTables0, VnTables) :-
 		vn_util__rval_to_vn(Rval2, SubVn2, VnTables1, VnTables2),
 		vn_util__vnrval_to_vn(vn_binop(Binop, SubVn1, SubVn2), Vn,
 			VnTables2, VnTables)
+	;
+		Rval = mem_addr(MemRef),
+		vn_util__mem_ref_to_vn(MemRef, Vn, VnTables0, VnTables)
+	).
+
+:- pred vn_util__mem_ref_to_vn(mem_ref, vn, vn_tables, vn_tables).
+% :- mode vn_util__mem_ref_to_vn(in, out, di, uo) is det.
+:- mode vn_util__mem_ref_to_vn(in, out, in, out) is det.
+
+vn_util__mem_ref_to_vn(MemRef, Vn, VnTables0, VnTables) :-
+	(
+		MemRef = stackvar_ref(Slot), 
+		vn_util__vnrval_to_vn(vn_stackvar_addr(Slot), Vn,
+			VnTables0, VnTables)
+	;
+		MemRef = framevar_ref(Slot), 
+		vn_util__vnrval_to_vn(vn_framevar_addr(Slot), Vn,
+			VnTables0, VnTables)
+	;
+		MemRef = heap_ref(Rval, Tag, Field), 
+		vn_util__rval_to_vn(Rval, SubVn, VnTables0, VnTables1),
+		vn_util__vnrval_to_vn(vn_heap_addr(SubVn, Tag, Field), Vn,
+			VnTables1, VnTables)
 	).
 
 :- pred vn_util__vnrval_to_vn(vnrval, vn, vn_tables, vn_tables).
@@ -181,12 +202,29 @@ vn_util__rval_to_vn(Rval, Vn, VnTables0, VnTables) :-
 vn_util__vnrval_to_vn(Vnrval, Vn, VnTables0, VnTables) :-
 	vn_util__simplify_vnrval(Vnrval, Vnrval1, VnTables0, VnTables1),
 	( vn_table__search_assigned_vn(Vnrval1, Vn_prime, VnTables1) ->
-		Vn = Vn_prime,
-		VnTables = VnTables1
+		( vn_util__vnrval_may_share_vn(Vnrval1) ->
+			Vn = Vn_prime,
+			VnTables = VnTables1
+		;
+			vn_table__record_new_vnrval(Vnrval1, Vn,
+				VnTables1, VnTables)
+		)
 	;
 		vn_table__record_first_vnrval(Vnrval1, Vn, VnTables1, VnTables)
 	).
 
+:- pred vn_util__vnrval_may_share_vn(vnrval).
+:- mode vn_util__vnrval_may_share_vn(in) is semidet.
+
+vn_util__vnrval_may_share_vn(Vnrval) :-
+	(
+		Vnrval = vn_unop(cast_to_unsigned, _)
+	->
+		fail
+	;
+		true
+	).
+
 	% Simplify the vnrval by partially evaluating expressions involving
 	% constants. To make this simpler, swap the arguments of commutative
 	% expressions around to put the constants on the right side.
@@ -301,6 +339,12 @@ vn_util__simplify_vnrval_binop(Binop, Vn1, Vn2, Vnrval, VnTables0, VnTables) :-
 			vn_util__vnrval_to_vn(vn_binop((+), Vn11, Vn2), VnExpr,
 				VnTables0, VnTables),
 			Vnrval = vn_binop((+), VnExpr, Vn12)
+		;
+						% e+0 => e
+			Vnrval2 = vn_const(int_const(0))
+		->
+			Vnrval = Vnrval1,
+			VnTables = VnTables0
 		;
 			fail
 		)
@@ -373,6 +417,12 @@ vn_util__simplify_vnrval_binop(Binop, Vn1, Vn2, Vnrval, VnTables0, VnTables) :-
 			vn_util__vnrval_to_vn(vn_binop((float_plus), Vn11,
 				Vn2), VnExpr, VnTables0, VnTables),
 			Vnrval = vn_binop((float_plus), VnExpr, Vn12)
+		;
+						% e+0 => e
+			Vnrval2 = vn_const(float_const(0.0))
+		->
+			Vnrval = Vnrval1,
+			VnTables = VnTables0
 		;
 			fail
 		)
@@ -404,6 +454,12 @@ vn_util__simplify_vnrval_binop(Binop, Vn1, Vn2, Vnrval, VnTables0, VnTables) :-
 		->
 			Vnrval = vn_const(int_const(0)),
 			VnTables = VnTables0
+		;
+						% e-0 => e
+			Vnrval2 = vn_const(int_const(0))
+		->
+			Vnrval = Vnrval1,
+			VnTables = VnTables0
 		;
 			fail
 		)
@@ -435,6 +491,12 @@ vn_util__simplify_vnrval_binop(Binop, Vn1, Vn2, Vnrval, VnTables0, VnTables) :-
 		->
 			Vnrval = vn_const(int_const(0)),
 			VnTables = VnTables0
+		;
+						% e-0 => e
+			Vnrval2 = vn_const(float_const(0.0))
+		->
+			Vnrval = Vnrval1,
+			VnTables = VnTables0
 		;
 			fail
 		)
@@ -448,6 +510,18 @@ vn_util__simplify_vnrval_binop(Binop, Vn1, Vn2, Vnrval, VnTables0, VnTables) :-
 			I is I1 * I2,
 			Vnrval = vn_const(int_const(I)),
 			VnTables = VnTables0
+		;
+						% e*1 => e
+			Vnrval2 = vn_const(int_const(1))
+		->
+			Vnrval = Vnrval1,
+			VnTables = VnTables0
+		;
+						% 1*e => e
+			Vnrval1 = vn_const(int_const(1))
+		->
+			Vnrval = Vnrval2,
+			VnTables = VnTables0
 		;
 			fail
 		)
@@ -461,6 +535,18 @@ vn_util__simplify_vnrval_binop(Binop, Vn1, Vn2, Vnrval, VnTables0, VnTables) :-
 			F is F1 * F2,
 			Vnrval = vn_const(float_const(F)),
 			VnTables = VnTables0
+		;
+						% e*1 => e
+			Vnrval2 = vn_const(float_const(1.0))
+		->
+			Vnrval = Vnrval1,
+			VnTables = VnTables0
+		;
+						% 1*e => e
+			Vnrval1 = vn_const(float_const(1.0))
+		->
+			Vnrval = Vnrval2,
+			VnTables = VnTables0
 		;
 			fail
 		)
@@ -475,6 +561,12 @@ vn_util__simplify_vnrval_binop(Binop, Vn1, Vn2, Vnrval, VnTables0, VnTables) :-
 			I is I1 // I2,
 			Vnrval = vn_const(int_const(I)),
 			VnTables = VnTables0
+		;
+						% e/1 => e
+			Vnrval2 = vn_const(int_const(1))
+		->
+			Vnrval = Vnrval1,
+			VnTables = VnTables0
 		;
 			fail
 		)
@@ -489,17 +581,133 @@ vn_util__simplify_vnrval_binop(Binop, Vn1, Vn2, Vnrval, VnTables0, VnTables) :-
 			F is F1 / F2,
 			Vnrval = vn_const(float_const(F)),
 			VnTables = VnTables0
+		;
+						% e/1 => e
+			Vnrval2 = vn_const(float_const(1.0))
+		->
+			Vnrval = Vnrval1,
+			VnTables = VnTables0
+		;
+			fail
+		)
+	;	
+		Binop = ('|'),
+		(
+						% c1\/c2 => c
+			Vnrval1 = vn_const(int_const(I1)),
+			Vnrval2 = vn_const(int_const(I2))
+		->
+			I is I1 \/ I2,
+			Vnrval = vn_const(int_const(I)),
+			VnTables = VnTables0
+		;
+						% e\/0 => e
+			Vnrval2 = vn_const(int_const(0))
+		->
+			Vnrval = Vnrval1,
+			VnTables = VnTables0
+		;
+						% 0\/e => e
+			Vnrval1 = vn_const(int_const(0))
+		->
+			Vnrval = Vnrval2,
+			VnTables = VnTables0
+		;
+						% e\/e => e
+			Vn1 = Vn2
+		->
+			Vnrval = Vnrval1,
+			VnTables = VnTables0
+		;
+			fail
+		)
+	;	
+		Binop = (&),
+		(
+						% c1/\c2 => c
+			Vnrval1 = vn_const(int_const(I1)),
+			Vnrval2 = vn_const(int_const(I2))
+		->
+			I is I1 /\ I2,
+			Vnrval = vn_const(int_const(I)),
+			VnTables = VnTables0
+		;
+						% e/\0 => 0
+			Vnrval2 = vn_const(int_const(0))
+		->
+			Vnrval = vn_const(int_const(0)),
+			VnTables = VnTables0
+		;
+						% 0/\e => 0
+			Vnrval1 = vn_const(int_const(0))
+		->
+			Vnrval = vn_const(int_const(0)),
+			VnTables = VnTables0
+		;
+						% e/\e => e
+			Vn1 = Vn2
+		->
+			Vnrval = Vnrval1,
+			VnTables = VnTables0
+		;
+			fail
+		)
+	;	
+		Binop = (>>),
+		(
+						% c1>>c2 => c
+			Vnrval1 = vn_const(int_const(I1)),
+			Vnrval2 = vn_const(int_const(I2))
+		->
+			I is I1 >> I2,
+			Vnrval = vn_const(int_const(I)),
+			VnTables = VnTables0
+		;
+						% e>>0 => e
+			Vnrval2 = vn_const(int_const(0))
+		->
+			Vnrval = Vnrval1,
+			VnTables = VnTables0
+		;
+			fail
+		)
+	;	
+		Binop = (<<),
+		(
+						% c1<<c2 => c
+			Vnrval1 = vn_const(int_const(I1)),
+			Vnrval2 = vn_const(int_const(I2))
+		->
+			I is I1 << I2,
+			Vnrval = vn_const(int_const(I)),
+			VnTables = VnTables0
+		;
+						% e<<0 => e
+			Vnrval2 = vn_const(int_const(0))
+		->
+			Vnrval = Vnrval1,
+			VnTables = VnTables0
 		;
 			fail
 		)
 	;	
 		Binop = eq,
 		(
+						% e==e => true
 			Vn1 = Vn2
 		->
 			Vnrval = vn_const(true),
 			VnTables = VnTables0
 		;
+						% otherwise, c1==c2 => false
+						% (true case handled above)
+			Vnrval1 = vn_const(_C1),
+			Vnrval2 = vn_const(_C2)
+		->
+			Vnrval = vn_const(false),
+			VnTables = VnTables0
+		;
+						% tag(mktag(e))==e => true
 			Vnrval1 = vn_unop(tag, WordVn),
 			Vnrval2 = vn_unop(mktag, Tag),
 			vn_table__lookup_defn(WordVn, WordVnrval,
@@ -514,11 +722,21 @@ vn_util__simplify_vnrval_binop(Binop, Vn1, Vn2, Vnrval, VnTables0, VnTables) :-
 	;	
 		Binop = ne,
 		(
+						% e!=e => false
 			Vn1 = Vn2
 		->
 			Vnrval = vn_const(false),
 			VnTables = VnTables0
 		;
+						% otherwise, c1!=c2 => true
+						% (false case handled above)
+			Vnrval1 = vn_const(_C1),
+			Vnrval2 = vn_const(_C2)
+		->
+			Vnrval = vn_const(true),
+			VnTables = VnTables0
+		;
+						% tag(mktag(e))!=e => false
 			Vnrval1 = vn_unop(tag, WordVn),
 			Vnrval2 = vn_unop(mktag, Tag),
 			vn_table__lookup_defn(WordVn, WordVnrval,
@@ -532,43 +750,53 @@ vn_util__simplify_vnrval_binop(Binop, Vn1, Vn2, Vnrval, VnTables0, VnTables) :-
 		)
 	;
 		Binop = (>=),
-		vn_util__const_if_equal_vns(Vn1, Vn2, true, Vnrval),
+		vn_util__simplify_int_compare_op(>=, Vn1, Vnrval1, Vn2, Vnrval2,
+			true, Vnrval),
 		VnTables = VnTables0
 	;
 		Binop = (<=),
-		vn_util__const_if_equal_vns(Vn1, Vn2, true, Vnrval),
+		vn_util__simplify_int_compare_op(=<, Vn1, Vnrval1, Vn2, Vnrval2,
+			true, Vnrval),
 		VnTables = VnTables0
 	;
 		Binop = (>),
-		vn_util__const_if_equal_vns(Vn1, Vn2, false, Vnrval),
+		vn_util__simplify_int_compare_op(>, Vn1, Vnrval1, Vn2, Vnrval2,
+			false, Vnrval),
 		VnTables = VnTables0
 	;
 		Binop = (<),
-		vn_util__const_if_equal_vns(Vn1, Vn2, false, Vnrval),
+		vn_util__simplify_int_compare_op(<, Vn1, Vnrval1, Vn2, Vnrval2,
+			false, Vnrval),
 		VnTables = VnTables0
 	;
 		Binop = (float_eq),
-		vn_util__const_if_equal_vns(Vn1, Vn2, true, Vnrval),
+		vn_util__simplify_float_compare_op(float_eq, Vn1, Vnrval1,
+			Vn2, Vnrval2, true, Vnrval),
 		VnTables = VnTables0
 	;
 		Binop = (float_ge),
-		vn_util__const_if_equal_vns(Vn1, Vn2, true, Vnrval),
+		vn_util__simplify_float_compare_op(>=, Vn1, Vnrval1,
+			Vn2, Vnrval2, true, Vnrval),
 		VnTables = VnTables0
 	;
 		Binop = (float_le),
-		vn_util__const_if_equal_vns(Vn1, Vn2, true, Vnrval),
+		vn_util__simplify_float_compare_op(=<, Vn1, Vnrval1,
+			Vn2, Vnrval2, true, Vnrval),
 		VnTables = VnTables0
 	;
 		Binop = (float_ne),
-		vn_util__const_if_equal_vns(Vn1, Vn2, false, Vnrval),
+		vn_util__simplify_float_compare_op(float_ne, Vn1, Vnrval1,
+			Vn2, Vnrval2, false, Vnrval),
 		VnTables = VnTables0
 	;
 		Binop = (float_gt),
-		vn_util__const_if_equal_vns(Vn1, Vn2, false, Vnrval),
+		vn_util__simplify_float_compare_op(>=, Vn1, Vnrval1,
+			Vn2, Vnrval2, false, Vnrval),
 		VnTables = VnTables0
 	;
 		Binop = (float_lt),
-		vn_util__const_if_equal_vns(Vn1, Vn2, false, Vnrval),
+		vn_util__simplify_float_compare_op(=<, Vn1, Vnrval1,
+			Vn2, Vnrval2, false, Vnrval),
 		VnTables = VnTables0
 	;
 		Binop = (str_eq),
@@ -630,6 +858,56 @@ vn_util__simplify_vnrval_binop(Binop, Vn1, Vn2, Vnrval, VnTables0, VnTables) :-
 		)
 	).
 
+:- pred vn_util__simplify_int_compare_op(pred(int, int),
+		vn, vnrval, vn, vnrval, rval_const, vnrval).
+:- mode vn_util__simplify_int_compare_op(pred(in, in) is semidet,
+		in, in, in, in, in, out) is semidet.
+
+vn_util__simplify_int_compare_op(ComparePred, Vn1, Vnrval1, Vn2, Vnrval2,
+		ResultIfEqual, Vnrval) :-
+	(
+		Vnrval1 = vn_const(int_const(C1)),
+		Vnrval2 = vn_const(int_const(C2))
+	->
+		(
+			call(ComparePred, C1, C2)
+		->
+			Vnrval = vn_const(true)
+		;
+			Vnrval = vn_const(false)
+		)
+	;
+		vn_util__const_if_equal_vns(Vn1, Vn2, ResultIfEqual, Vnrval)
+	).
+
+:- pred vn_util__simplify_float_compare_op(pred(float, float),
+		vn, vnrval, vn, vnrval, rval_const, vnrval).
+:- mode vn_util__simplify_float_compare_op(pred(in, in) is semidet,
+		in, in, in, in, in, out) is semidet.
+
+vn_util__simplify_float_compare_op(ComparePred, Vn1, Vnrval1, Vn2, Vnrval2,
+		ResultIfEqual, Vnrval) :-
+	(
+		Vnrval1 = vn_const(float_const(C1)),
+		Vnrval2 = vn_const(float_const(C2))
+	->
+		(
+			call(ComparePred, C1, C2)
+		->
+			Vnrval = vn_const(true)
+		;
+			Vnrval = vn_const(false)
+		)
+	;
+		vn_util__const_if_equal_vns(Vn1, Vn2, ResultIfEqual, Vnrval)
+	).
+
+:- pred float_eq(float::in, float::in) is semidet.
+float_eq(X, X).
+
+:- pred float_ne(float::in, float::in) is semidet.
+float_ne(X, Y) :- X \= Y.
+
 :- pred vn_util__const_if_equal_vns(vn, vn, rval_const, vnrval).
 :- mode vn_util__const_if_equal_vns(in, in, in, out) is semidet.
 
@@ -658,6 +936,9 @@ vn_util__lval_to_vnlval(Lval, Vnlval, VnTables0, VnTables) :-
 		vn_util__rval_to_vn(Rval1, Vn1, VnTables0, VnTables1),
 		vn_util__rval_to_vn(Rval2, Vn2, VnTables1, VnTables),
 		Vnlval = vn_field(Tag, Vn1, Vn2)
+	; Lval = mem_ref(Rval1) ->
+		vn_util__rval_to_vn(Rval1, Vn1, VnTables0, VnTables),
+		Vnlval = vn_mem_ref(Vn1)
 	; Lval = succfr(Rval1) ->
 		vn_util__rval_to_vn(Rval1, Vn1, VnTables0, VnTables),
 		Vnlval = vn_succfr(Vn1)
@@ -676,8 +957,11 @@ vn_util__lval_to_vnlval(Lval, Vnlval, VnTables0, VnTables) :-
 
 % If you to add to this list to fix a determinism error,
 % check vn_util__lval_to_vnlval above as well.
+% vn_util__lval_to_vnlval should have code to handle every lval
+% that is mapped to "no" by vn_util__no_access_lval_to_vnlval.
 
-vn_util__no_access_lval_to_vnlval(reg(Reg),		yes(vn_reg(Reg))).
+vn_util__no_access_lval_to_vnlval(reg(T, N),		yes(vn_reg(T, N))).
+vn_util__no_access_lval_to_vnlval(temp(T, N),		yes(vn_temp(T, N))).
 vn_util__no_access_lval_to_vnlval(stackvar(N),		yes(vn_stackvar(N))).
 vn_util__no_access_lval_to_vnlval(framevar(N),		yes(vn_framevar(N))).
 vn_util__no_access_lval_to_vnlval(succip,		yes(vn_succip)).
@@ -689,12 +973,13 @@ vn_util__no_access_lval_to_vnlval(prevfr(_),		no).
 vn_util__no_access_lval_to_vnlval(succfr(_),		no).
 vn_util__no_access_lval_to_vnlval(hp,			yes(vn_hp)).
 vn_util__no_access_lval_to_vnlval(sp,			yes(vn_sp)).
+vn_util__no_access_lval_to_vnlval(mem_ref(_),		no).
 vn_util__no_access_lval_to_vnlval(field(_, _, _),	no).
-vn_util__no_access_lval_to_vnlval(temp(Reg),		yes(vn_temp(Reg))).
 vn_util__no_access_lval_to_vnlval(lvar(_Var), _) :-
 	error("lvar detected in value_number").
 
-vn_util__no_access_vnlval_to_lval(vn_reg(Reg),		yes(reg(Reg))).
+vn_util__no_access_vnlval_to_lval(vn_reg(T, N),		yes(reg(T, N))).
+vn_util__no_access_vnlval_to_lval(vn_temp(T, N),	yes(temp(T, N))).
 vn_util__no_access_vnlval_to_lval(vn_stackvar(N),	yes(stackvar(N))).
 vn_util__no_access_vnlval_to_lval(vn_framevar(N),	yes(framevar(N))).
 vn_util__no_access_vnlval_to_lval(vn_succip,		yes(succip)).
@@ -707,10 +992,11 @@ vn_util__no_access_vnlval_to_lval(vn_succip(_),		no).
 vn_util__no_access_vnlval_to_lval(vn_hp,		yes(hp)).
 vn_util__no_access_vnlval_to_lval(vn_sp,		yes(sp)).
 vn_util__no_access_vnlval_to_lval(vn_field(_, _, _),	no).
-vn_util__no_access_vnlval_to_lval(vn_temp(Reg),		yes(temp(Reg))).
+vn_util__no_access_vnlval_to_lval(vn_mem_ref(_),	no).
 
 /* one of these preds should be eliminated XXX */
-vn_util__vnlval_access_vns(vn_reg(_), []).
+vn_util__vnlval_access_vns(vn_reg(_, _), []).
+vn_util__vnlval_access_vns(vn_temp(_, _), []).
 vn_util__vnlval_access_vns(vn_stackvar(_), []).
 vn_util__vnlval_access_vns(vn_framevar(_), []).
 vn_util__vnlval_access_vns(vn_succip, []).
@@ -723,30 +1009,18 @@ vn_util__vnlval_access_vns(vn_succip(Vn), [Vn]).
 vn_util__vnlval_access_vns(vn_hp, []).
 vn_util__vnlval_access_vns(vn_sp, []).
 vn_util__vnlval_access_vns(vn_field(_, Vn1, Vn2), [Vn1, Vn2]).
-vn_util__vnlval_access_vns(vn_temp(_), []).
+vn_util__vnlval_access_vns(vn_mem_ref(Vn), [Vn]).
 
 vn_util__find_sub_vns(vn_origlval(Vnlval), SubVns) :-
-	vn_util__find_sub_vns_vnlval(Vnlval, SubVns).
+	vn_util__vnlval_access_vns(Vnlval, SubVns).
 vn_util__find_sub_vns(vn_mkword(_, SubVn), [SubVn]).
 vn_util__find_sub_vns(vn_const(_), []).
 vn_util__find_sub_vns(vn_create(_, _, _, _), []).
 vn_util__find_sub_vns(vn_unop(_, SubVn), [SubVn]).
 vn_util__find_sub_vns(vn_binop(_, SubVn1, SubVn2), [SubVn1, SubVn2]).
-
-vn_util__find_sub_vns_vnlval(vn_reg(_), []).
-vn_util__find_sub_vns_vnlval(vn_stackvar(_), []).
-vn_util__find_sub_vns_vnlval(vn_framevar(_), []).
-vn_util__find_sub_vns_vnlval(vn_succip, []).
-vn_util__find_sub_vns_vnlval(vn_maxfr, []).
-vn_util__find_sub_vns_vnlval(vn_curfr, []).
-vn_util__find_sub_vns_vnlval(vn_succfr(Vn), [Vn]).
-vn_util__find_sub_vns_vnlval(vn_prevfr(Vn), [Vn]).
-vn_util__find_sub_vns_vnlval(vn_redoip(Vn), [Vn]).
-vn_util__find_sub_vns_vnlval(vn_succip(Vn), [Vn]).
-vn_util__find_sub_vns_vnlval(vn_hp, []).
-vn_util__find_sub_vns_vnlval(vn_sp, []).
-vn_util__find_sub_vns_vnlval(vn_field(_, Vn1, Vn2), [Vn1, Vn2]).
-vn_util__find_sub_vns_vnlval(vn_temp(_), []).
+vn_util__find_sub_vns(vn_stackvar_addr(_), []).
+vn_util__find_sub_vns(vn_framevar_addr(_), []).
+vn_util__find_sub_vns(vn_heap_addr(SubVn, _, _), [SubVn]).
 
 vn_util__is_const_expr(Vn, IsConst, VnTables) :-
 	vn_table__lookup_defn(Vn, Vnrval, "vn_util__is_const_expr", VnTables),
@@ -770,6 +1044,15 @@ vn_util__is_const_expr(Vn, IsConst, VnTables) :-
 		vn_util__is_const_expr(Vn1, IsConst1, VnTables),
 		vn_util__is_const_expr(Vn2, IsConst2, VnTables),
 		bool__and(IsConst1, IsConst2, IsConst)
+	;
+		Vnrval = vn_stackvar_addr(_),
+		IsConst = no
+	;
+		Vnrval = vn_framevar_addr(_),
+		IsConst = no
+	;
+		Vnrval = vn_heap_addr(_, _, _),
+		IsConst = no
 	).
 
 vn_util__find_lvals_in_rval(Rval, Lvals) :-
@@ -798,6 +1081,9 @@ vn_util__find_lvals_in_rval(Rval, Lvals) :-
 		vn_util__find_lvals_in_rval(Rval1, Lvals1),
 		vn_util__find_lvals_in_rval(Rval2, Lvals2),
 		list__append(Lvals1, Lvals2, Lvals)
+	;
+		Rval = mem_addr(MemRef),
+		vn_util__find_lvals_in_mem_ref(MemRef, Lvals)
 	).
 
 vn_util__find_lvals_in_rvals([], []).
@@ -806,6 +1092,15 @@ vn_util__find_lvals_in_rvals([Rval | Rvals], Lvals) :-
 	vn_util__find_lvals_in_rvals(Rvals, Lvals2),
 	list__append(Lvals1, Lvals2, Lvals).
 
+:- pred vn_util__find_lvals_in_mem_ref(mem_ref, list(lval)).
+:- mode vn_util__find_lvals_in_mem_ref(in, out) is det.
+
+	% XXX
+vn_util__find_lvals_in_mem_ref(stackvar_ref(_), []).
+vn_util__find_lvals_in_mem_ref(framevar_ref(_), []).
+vn_util__find_lvals_in_mem_ref(heap_ref(Rval, _, _), Lvals) :-
+	vn_util__find_lvals_in_rval(Rval, Lvals).
+
 vn_util__is_vn_shared(Vn, Vnrval, Uses0, VnTables) :-
 	vn_util__is_const_expr(Vn, no, VnTables),
 	\+ Vnrval = vn_origlval(vn_hp),
@@ -877,7 +1172,8 @@ vn_util__choose_cheapest_loc_2([], Stack0, Heap0, BestLoc) :-
 		fail
 	).
 
-vn_util__classify_loc_cost(vn_reg(_), 0).
+vn_util__classify_loc_cost(vn_reg(_, _), 0).
+vn_util__classify_loc_cost(vn_temp(_, _), 0).
 vn_util__classify_loc_cost(vn_stackvar(_), 1).
 vn_util__classify_loc_cost(vn_framevar(_), 1).
 vn_util__classify_loc_cost(vn_succip, 0).
@@ -890,7 +1186,7 @@ vn_util__classify_loc_cost(vn_succip(_), 1).
 vn_util__classify_loc_cost(vn_hp, 0).
 vn_util__classify_loc_cost(vn_sp, 0).
 vn_util__classify_loc_cost(vn_field(_, _, _), 2).
-vn_util__classify_loc_cost(vn_temp(_), 0).
+vn_util__classify_loc_cost(vn_mem_ref(_), 2).
 
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
@@ -954,12 +1250,21 @@ vn_util__build_uses_from_ctrl(Ctrl, Ctrlmap, VnTables0, VnTables) :-
 			vn_util__record_use_list(Vns, src_ctrl(Ctrl),
 				VnTables0, VnTables1)
 		;
-			VnInstr = vn_restore_ticket(Vn),
+			VnInstr = vn_reset_ticket(Vn, _Reason),
 			vn_util__record_use(Vn, src_ctrl(Ctrl),
 				VnTables0, VnTables1)
 		;
 			VnInstr = vn_discard_ticket,
 			VnTables1 = VnTables0
+		;
+			VnInstr = vn_mark_ticket_stack(Vnlval),
+			vn_util__vnlval_access_vns(Vnlval, Vns),
+			vn_util__record_use_list(Vns, src_ctrl(Ctrl),
+				VnTables0, VnTables1)
+		;
+			VnInstr = vn_discard_tickets_to(Vn),
+			vn_util__record_use(Vn, src_ctrl(Ctrl),
+				VnTables0, VnTables1)
 		;
 			VnInstr = vn_incr_sp(_, _),
 			VnTables1 = VnTables0
@@ -1038,6 +1343,16 @@ vn_util__record_use(Vn, Src, VnTables0, VnTables) :-
 				VnTables1, VnTables2),
 			vn_util__record_use(SubVn2, src_vn(Vn),
 				VnTables2, VnTables)
+		;
+			Vnrval = vn_stackvar_addr(_),
+			VnTables = VnTables1
+		;
+			Vnrval = vn_framevar_addr(_),
+			VnTables = VnTables1
+		;
+			Vnrval = vn_heap_addr(SubVn, _, _),
+			vn_util__record_use(SubVn, src_vn(Vn),
+				VnTables1, VnTables)
 		)
 	;
 		VnTables = VnTables1
