@@ -1,5 +1,5 @@
 %-----------------------------------------------------------------------------%
-% Copyright (C) 1995 University of Melbourne.
+% Copyright (C) 1994-1997 The University of Melbourne.
 % This file may only be copied under the terms of the GNU General
 % Public License - see the file COPYING in the Mercury distribution.
 %-----------------------------------------------------------------------------%
@@ -24,20 +24,21 @@
 % (An option can have as many names as you like, long or short.)
 % You must provide a predicate `option_default(Option, OptionData)'
 % which specifies both the type and the default value for every option.
-% We support four different option types: bool, int, string, and
+% We support five different option types: bool, int, string, maybe_string
+% (which have either a value of `no' or `yes(string)'), and
 % "accumulating" (which accumulates a list of strings).
-% For the first three option types, if there are multiple occurrences
+% For the first four option types, if there are multiple occurrences
 % of the same option on the command-line, then the last (right-most)
 % occurrence will take precedence.  For "accumulating" options,
 % multiple occurrences will be appended together into a list.
-% Single-character boolean options can be negated by following them
-% with another `-', e.g. `-x-' will negate the `-x' option.
-% Long boolean options can be negated by preceding them with `--no-',
-% e.g. `--no-foo' will negate the `--foo' option.
+% Single-character boolean or maybe-string options can be negated by
+% following them with another `-', e.g. `-x-' will negate the `-x' option.
+% Long boolean or maybe-string options can be negated by preceding them with
+% `--no-', e.g. `--no-foo' will negate the `--foo' option.
 
 :- module getopt.
 :- interface.
-:- import_module int, string, bool, list, map, std_util.
+:- import_module bool, char, list, map, std_util.
 
 % getopt__process_options(OptionOps, Args, NonOptionArgs, Result)
 %
@@ -52,7 +53,7 @@
 %	predicates used to categorize a set of options. Their
 %	interfaces should be like these:
 %
-% :- pred short_option(character::in, option::out) is semidet.
+% :- pred short_option(char::in, option::out) is semidet.
 % 	True if the character names a valid single-character option.
 %
 % :- pred long_option(string::in, option::out) is semidet.
@@ -81,12 +82,12 @@
 
 :- type option_ops(OptionType)
 	--->	option_ops(
-			pred(character, OptionType),	% short_option
+			pred(char, OptionType),		% short_option
 			pred(string, OptionType),	% long_option
 			pred(OptionType, option_data)	% option_default
 		)
 	;	option_ops(
-			pred(character, OptionType),	% short_option
+			pred(char, OptionType),		% short_option
 			pred(string, OptionType),	% long_option
 			pred(OptionType, option_data),	% option_default
 			pred(OptionType, special_data,	% special option handler
@@ -113,6 +114,7 @@
 	--->	bool(bool)
 	;	int(int)
 	;	string(string)
+	;	maybe_string(maybe(string))
 	;	accumulating(list(string))
 	;	special
 	;	bool_special
@@ -145,10 +147,18 @@
 :- pred getopt__lookup_string_option(option_table(Option), Option, string).
 :- mode getopt__lookup_string_option(in, in, out) is det.
 
+:- pred getopt__lookup_maybe_string_option(option_table(Option), Option,
+		maybe(string)).
+:- mode getopt__lookup_maybe_string_option(in, in, out) is det.
+
+:- pred getopt__lookup_accumulating_option(option_table(Option), Option,
+		list(string)).
+:- mode getopt__lookup_accumulating_option(in, in, out) is det.
+
 %-----------------------------------------------------------------------------%
 
 :- implementation.
-:- import_module require.
+:- import_module require, string.
 
 getopt__process_options(OptionOps, Args0, Args, Result) :-
 	getopt__get_option_defaults(OptionOps, OptionDefaultsPred),
@@ -173,7 +183,7 @@ getopt__process_arguments([Option | Args0], Args, OptionOps,
 		getopt__get_long_options(OptionOps, LongOptionPred),
 		( call(LongOptionPred, LongOption, Flag) ->
 			string__append("--", LongOption, OptName),
-			process_negated_bool_option(OptName, Flag,
+			process_negated_option(OptName, Flag,
 				OptionOps, OptionTable0, Result1),
 			( Result1 = ok(OptionTable1) ->
 				getopt__process_arguments(Args0, Args,
@@ -219,7 +229,7 @@ getopt__process_arguments([Option | Args0], Args, OptionOps,
 			Result = error(ErrorMsg),
 			Args = Args0
 		)
-	; string__first_char(Option, '-', ShortOptions) ->
+	; string__first_char(Option, '-', ShortOptions), ShortOptions \= "" ->
 		string__to_char_list(ShortOptions, ShortOptionsList),
 		% Process a single negated option `-x-'.
 		( ShortOptionsList = [SingleShortOpt, '-'] ->
@@ -227,7 +237,7 @@ getopt__process_arguments([Option | Args0], Args, OptionOps,
 			( call(ShortOptionPred, SingleShortOpt, Flag) ->
 				string__from_char_list(['-', SingleShortOpt],
 					OptName),
-				process_negated_bool_option(OptName, Flag,
+				process_negated_option(OptName, Flag,
 					OptionOps, OptionTable0, Result1),
 				( Result1 = ok(OptionTable1) ->
 					getopt__process_arguments(Args0, Args,
@@ -310,7 +320,7 @@ getopt__handle_long_option(Option, Flag, OptionData, MaybeOptionArg0,
 		)
 	).
 
-:- pred getopt__handle_short_options(list(character)::in,
+:- pred getopt__handle_short_options(list(char)::in,
 	option_ops(OptionType)::in(option_ops), list(string)::in,
 	list(string)::out, option_table(OptionType)::in,
 	maybe_option_table(OptionType)::out) is det.
@@ -357,7 +367,7 @@ getopt__handle_short_options([Opt | Opts0], OptionOps, Args0, Args,
 		Args = Args0
 	).
 
-:- pred getopt__get_short_option_arg(list(character), string,
+:- pred getopt__get_short_option_arg(list(char), string,
 	list(string), list(string)).
 :- mode getopt__get_short_option_arg(in, out, in, out) is det.
 
@@ -407,6 +417,15 @@ getopt__process_option(string(_), _Option, Flag, MaybeArg, _OptionOps,
 	;
 		error("string argument expected in getopt__process_option")
 	).
+getopt__process_option(maybe_string(_), _Option, Flag, MaybeArg, _OptionOps,
+		OptionTable0, Result) :-
+	( MaybeArg = yes(Arg) ->
+		map__set(OptionTable0, Flag, maybe_string(yes(Arg)),
+				OptionTable),
+		Result = ok(OptionTable)
+	;
+		error("string argument expected in getopt__process_option")
+	).
 getopt__process_option(accumulating(List0), _Option, Flag, MaybeArg, _OptionOps,
 		OptionTable0, Result) :-
 	( MaybeArg = yes(Arg) ->
@@ -454,15 +473,19 @@ getopt__process_option(string_special, Option, Flag, MaybeArg, OptionOps,
 		error("string_special argument expected in getopt__process_option")
 	).
 
-:- pred process_negated_bool_option(string, OptionType, option_ops(OptionType),
+:- pred process_negated_option(string, OptionType, option_ops(OptionType),
 	option_table(OptionType), maybe_option_table(OptionType)).
-:- mode process_negated_bool_option(in, in, in(option_ops), in, out) is det.
+:- mode process_negated_option(in, in, in(option_ops), in, out) is det.
 
-process_negated_bool_option(Option, Flag, OptionOps, OptionTable0, Result) :-
+process_negated_option(Option, Flag, OptionOps, OptionTable0, Result) :-
 	( map__search(OptionTable0, Flag, OptionData) ->
 		( OptionData = bool(_) ->
 			map__set(OptionTable0, Flag, bool(no), OptionTable),
 			Result = ok(OptionTable)
+		; OptionData = maybe_string(_) ->
+			map__set(OptionTable0, Flag, maybe_string(no),
+					OptionTable),
+			Result = ok(OptionTable)
 		; OptionData = bool_special ->
 			getopt__process_special(Option, Flag, bool(no),
 				OptionOps, OptionTable0, Result)
@@ -509,6 +532,7 @@ getopt__process_special(Option, Flag, OptionData, OptionOps,
 getopt__need_arg(bool(_), no).
 getopt__need_arg(int(_), yes).
 getopt__need_arg(string(_), yes).
+getopt__need_arg(maybe_string(_), yes).
 getopt__need_arg(accumulating(_), yes).
 getopt__need_arg(special, no).
 getopt__need_arg(bool_special, no).
@@ -527,7 +551,7 @@ getopt__numeric_argument(Option, Arg, Result) :-
 %-----------------------------------------------------------------------------%
 
 :- pred getopt__get_short_options(option_ops(OptionType)::in(option_ops),
-	pred(character, OptionType)::out(pred(in, out) is semidet)) is det.
+	pred(char, OptionType)::out(pred(in, out) is semidet)) is det.
 
 getopt__get_short_options(option_ops(ShortOpt, _, _), ShortOpt).
 getopt__get_short_options(option_ops(ShortOpt, _, _, _), ShortOpt).
@@ -574,4 +598,18 @@ getopt__lookup_string_option(OptionTable, Opt, Val) :-
 		error("Expected string option and didn't get one.")
 	).
 
+getopt__lookup_maybe_string_option(OptionTable, Opt, Val) :-
+	( map__lookup(OptionTable, Opt, maybe_string(Val0)) ->
+		Val = Val0
+	;
+		error("Expected maybe_string option and didn't get one.")
+	).
+
+getopt__lookup_accumulating_option(OptionTable, Opt, Val) :-
+	( map__lookup(OptionTable, Opt, accumulating(Val0)) ->
+		Val = Val0
+	;
+		error("Expected accumulating option and didn't get one.")
+	).
+
 %-----------------------------------------------------------------------------%
