@@ -1,5 +1,5 @@
 %---------------------------------------------------------------------------%
-% Copyright (C) 1995 University of Melbourne.
+% Copyright (C) 1993-1997 The University of Melbourne.
 % This file may only be copied under the terms of the GNU Library General
 % Public License - see the file COPYING.LIB in the Mercury distribution.
 %---------------------------------------------------------------------------%
@@ -17,7 +17,7 @@
 
 :- module lexer.
 :- interface.
-:- import_module char, string, int, float, list, std_util, io.
+:- import_module io.
 
 :- type	token
 	--->	name(string)
@@ -35,7 +35,7 @@
 	;	ht_sep			% '|'
 	;	comma			% ','
 	;	end			% '.'
-	;	junk(character)		% junk character in the input stream
+	;	junk(char)		% junk character in the input stream
 	;	error(string)		% some other invalid token
 	;	io_error(io__error)	% error reading from the input stream
 	;	eof.			% end-of-file
@@ -45,7 +45,9 @@
 
 :- type token_context == int.	% line number
 
-:- type token_list == list(pair(token, token_context)).
+% This "fat list" representation is more efficient than a list of pairs.
+:- type token_list	--->	token_cons(token, token_context, token_list)
+			;	token_nil.
 
 :- pred lexer__get_token_list(token_list, io__state, io__state).
 :- mode lexer__get_token_list(out, di, uo) is det.
@@ -57,10 +59,20 @@
 :- mode lexer__token_to_string(in, out) is det.
 %	Convert a token to a human-readable string describing the token.
 
+%-----------------------------------------------------------------------------%
+:- implementation.
+%-----------------------------------------------------------------------------%
+:- interface.
+
+	% lexer__graphic_token_char(Char): true iff `Char'
+	% is "graphic token char" (ISO Prolog 6.4.2).
+	% This is exported for use by term_io__quote_atom.
+:- pred lexer__graphic_token_char(char).
+:- mode lexer__graphic_token_char(in) is semidet.
 %-----------------------------------------------------------------------------%
 
 :- implementation.
-:- import_module require.
+:- import_module list, require, char, string, int.
 
 lexer__token_to_string(name(Name), String) :-
 	string__append_list(["token '", Name, "'"], String).
@@ -108,11 +120,11 @@ lexer__token_to_string(error(Message), String) :-
 lexer__get_token_list(Tokens) -->
 	lexer__get_token(Token, Context),
 	( { Token = eof } ->
-		{ Tokens = [] }
+		{ Tokens = token_nil }
 	; { Token = end ; Token = error(_) ; Token = io_error(_) } ->
-		{ Tokens = [Token - Context] }
+		{ Tokens = token_cons(Token, Context, token_nil) }
 	;
-		{ Tokens = [Token - Context | Tokens1] },
+		{ Tokens = token_cons(Token, Context, Tokens1) },
 		lexer__get_token_list(Tokens1)
 	).
 
@@ -134,8 +146,6 @@ lexer__get_context(Context) -->
 :- pred lexer__get_token_1(token, io__state, io__state).
 :- mode lexer__get_token_1(out, di, uo) is det.
 
-	% We use a big switch on Char here for efficiency - obviously not
-	% for conciseness.
 lexer__get_token_1(Token) -->
 	io__read_char(Result),
 	( { Result = error(Error) }, !,
@@ -143,136 +153,36 @@ lexer__get_token_1(Token) -->
 	; { Result = eof }, !,
 		{ Token = eof }
 	; { Result = ok(Char) },
-		( { Char = '\001', Token = junk(Char) }
-		; { Char = '\002', Token = junk(Char) }
-		; { Char = '\003', Token = junk(Char) }
-		; { Char = '\004', Token = junk(Char) }
-		; { Char = '\005', Token = junk(Char) }
-		; { Char = '\006', Token = junk(Char) }
-		; { Char = '\007', Token = junk(Char) }
-		; { Char = '\010', Token = junk(Char) }
-% \t		; { Char = '\011', Token = junk(Char) }
-		; { Char = '\t' }, lexer__get_token_2(Token)
-% \n		; { Char = '\012', Token = junk(Char) }
-		; { Char = '\n' }, lexer__get_token_2(Token)
-		; { Char = '\013', Token = junk(Char) }
-% \f		; { Char = '\014', Token = junk(Char) }
-		; { Char = '\f' }, lexer__get_token_2(Token)
-		; { Char = '\015', Token = junk(Char) }
-		; { Char = '\016', Token = junk(Char) }
-		; { Char = '\017', Token = junk(Char) }
-		; { Char = '\020', Token = junk(Char) }
-		; { Char = '\021', Token = junk(Char) }
-		; { Char = '\022', Token = junk(Char) }
-		; { Char = '\023', Token = junk(Char) }
-		; { Char = '\024', Token = junk(Char) }
-		; { Char = '\025', Token = junk(Char) }
-		; { Char = '\026', Token = junk(Char) }
-		; { Char = '\027', Token = junk(Char) }
-		; { Char = '\030', Token = junk(Char) }
-		; { Char = '\031', Token = junk(Char) }
-		; { Char = '\032', Token = junk(Char) }
-		; { Char = '\033', Token = junk(Char) }
-		; { Char = '\034', Token = junk(Char) }
-		; { Char = '\035', Token = junk(Char) }
-		; { Char = '\036', Token = junk(Char) }
-		; { Char = '\037', Token = junk(Char) }
-		; { Char = ' ' }, lexer__get_token_2(Token)
-		; { Char = ('!') }, { Token = name("!") }
-		; { Char = '"' }, lexer__get_quoted_name(Char, [], Token)
-		; { Char = ('#') }, lexer__get_source_line_number([], Token)
-		; { Char = ('$') }, lexer__get_graphic([Char], Token)
-		; { Char = ('%') }, lexer__skip_to_eol(Token)
-		; { Char = ('&') }, lexer__get_graphic([Char], Token)
-		; { Char = '''' }, lexer__get_quoted_name(Char, [], Token)
-		; { Char = ('(') }, { Token = open_ct }
-		; { Char = (')') }, { Token = close }
-		; { Char = ('*') }, lexer__get_graphic([Char], Token)
-		; { Char = ('+') }, lexer__get_graphic([Char], Token)
-		; { Char = (',') }, { Token = comma }
-		; { Char = ('-') }, lexer__get_graphic([Char], Token)
-		; { Char = ('.') }, lexer__get_dot(Token)
-		; { Char = ('/') }, lexer__get_slash(Token)
-		; { Char = '0' }, lexer__get_zero(Token)
-		; { Char = '1' }, lexer__get_number([Char], Token)
-		; { Char = '2' }, lexer__get_number([Char], Token)
-		; { Char = '3' }, lexer__get_number([Char], Token)
-		; { Char = '4' }, lexer__get_number([Char], Token)
-		; { Char = '5' }, lexer__get_number([Char], Token)
-		; { Char = '6' }, lexer__get_number([Char], Token)
-		; { Char = '7' }, lexer__get_number([Char], Token)
-		; { Char = '8' }, lexer__get_number([Char], Token)
-		; { Char = '9' }, lexer__get_number([Char], Token)
-		; { Char = (':') }, lexer__get_graphic([Char], Token)
-		; { Char = (';') }, { Token = name(";") }
-		; { Char = ('<') }, lexer__get_graphic([Char], Token)
-		; { Char = ('=') }, lexer__get_graphic([Char], Token)
-		; { Char = ('>') }, lexer__get_graphic([Char], Token)
-		; { Char = ('?') }, lexer__get_graphic([Char], Token)
-		; { Char = ('@') }, lexer__get_graphic([Char], Token)
-		; { Char = 'A' }, lexer__get_variable([Char], Token)
-		; { Char = 'B' }, lexer__get_variable([Char], Token)
-		; { Char = 'C' }, lexer__get_variable([Char], Token)
-		; { Char = 'D' }, lexer__get_variable([Char], Token)
-		; { Char = 'E' }, lexer__get_variable([Char], Token)
-		; { Char = 'F' }, lexer__get_variable([Char], Token)
-		; { Char = 'G' }, lexer__get_variable([Char], Token)
-		; { Char = 'H' }, lexer__get_variable([Char], Token)
-		; { Char = 'I' }, lexer__get_variable([Char], Token)
-		; { Char = 'J' }, lexer__get_variable([Char], Token)
-		; { Char = 'K' }, lexer__get_variable([Char], Token)
-		; { Char = 'L' }, lexer__get_variable([Char], Token)
-		; { Char = 'M' }, lexer__get_variable([Char], Token)
-		; { Char = 'N' }, lexer__get_variable([Char], Token)
-		; { Char = 'O' }, lexer__get_variable([Char], Token)
-		; { Char = 'P' }, lexer__get_variable([Char], Token)
-		; { Char = 'Q' }, lexer__get_variable([Char], Token)
-		; { Char = 'R' }, lexer__get_variable([Char], Token)
-		; { Char = 'S' }, lexer__get_variable([Char], Token)
-		; { Char = 'T' }, lexer__get_variable([Char], Token)
-		; { Char = 'U' }, lexer__get_variable([Char], Token)
-		; { Char = 'V' }, lexer__get_variable([Char], Token)
-		; { Char = 'W' }, lexer__get_variable([Char], Token)
-		; { Char = 'X' }, lexer__get_variable([Char], Token)
-		; { Char = 'Y' }, lexer__get_variable([Char], Token)
-		; { Char = 'Z' }, lexer__get_variable([Char], Token)
-		; { Char = ('[') }, { Token = open_list }
-		; { Char = ('\\') }, lexer__get_graphic([Char], Token)
-		; { Char = (']') }, { Token = close_list }
-		; { Char = ('^') }, lexer__get_graphic([Char], Token)
-		; { Char = '_' }, lexer__get_variable([Char], Token)
-		; { Char = 'a' }, lexer__get_name([Char], Token)
-		; { Char = 'b' }, lexer__get_name([Char], Token)
-		; { Char = 'c' }, lexer__get_name([Char], Token)
-		; { Char = 'd' }, lexer__get_name([Char], Token)
-		; { Char = 'e' }, lexer__get_name([Char], Token)
-		; { Char = 'f' }, lexer__get_name([Char], Token)
-		; { Char = 'g' }, lexer__get_name([Char], Token)
-		; { Char = 'h' }, lexer__get_name([Char], Token)
-		; { Char = 'i' }, lexer__get_name([Char], Token)
-		; { Char = 'j' }, lexer__get_name([Char], Token)
-		; { Char = 'k' }, lexer__get_name([Char], Token)
-		; { Char = 'l' }, lexer__get_name([Char], Token)
-		; { Char = 'm' }, lexer__get_name([Char], Token)
-		; { Char = 'n' }, lexer__get_name([Char], Token)
-		; { Char = 'o' }, lexer__get_name([Char], Token)
-		; { Char = 'p' }, lexer__get_name([Char], Token)
-		; { Char = 'q' }, lexer__get_name([Char], Token)
-		; { Char = 'r' }, lexer__get_name([Char], Token)
-		; { Char = 's' }, lexer__get_name([Char], Token)
-		; { Char = 't' }, lexer__get_name([Char], Token)
-		; { Char = 'u' }, lexer__get_name([Char], Token)
-		; { Char = 'v' }, lexer__get_name([Char], Token)
-		; { Char = 'w' }, lexer__get_name([Char], Token)
-		; { Char = 'x' }, lexer__get_name([Char], Token)
-		; { Char = 'y' }, lexer__get_name([Char], Token)
-		; { Char = 'z' }, lexer__get_name([Char], Token)
-		; { Char = ('{') }, { Token = open_curly }
-		; { Char = ('}') }, { Token = close_curly }
-		; { Char = ('|') }, { Token = ht_sep }
-		; { Char = ('~') }, lexer__get_graphic([Char], Token)
-		; { Char = '\140', Token = junk(Char) } % `
-		; { Char = '\177', Token = junk(Char) }
+		( { Char = ' ' ; Char = '\t' ; Char = '\n' } ->
+			lexer__get_token_2(Token)
+		; { char__is_upper(Char) ; Char = '_' } ->
+			lexer__get_variable([Char], Token)
+		; { char__is_lower(Char) } ->
+			lexer__get_name([Char], Token)
+		; { Char = '0' } ->
+			lexer__get_zero(Token)
+		; { char__is_digit(Char) } ->
+			lexer__get_number([Char], Token)
+		; { lexer__special_token(Char, SpecialToken) } ->
+			{ SpecialToken = open ->
+				Token = open_ct
+			;
+				Token = SpecialToken
+			}
+		; { Char = ('.') } ->
+			lexer__get_dot(Token)
+		; { Char = ('%') } ->
+			lexer__skip_to_eol(Token)
+		; { Char = '"' ; Char = '''' } ->
+			lexer__get_quoted_name(Char, [], Token)
+		; { Char = ('/') } ->
+			lexer__get_slash(Token)
+		; { Char = ('#') } ->
+			lexer__get_source_line_number([], Token)
+		; { lexer__graphic_token_char(Char) } ->
+			lexer__get_graphic([Char], Token)
+		;
+			{ Token = junk(Char) }
 		)
 	).
 
@@ -282,8 +192,6 @@ lexer__get_token_1(Token) -->
 	% This is just like get_token_1, except that we have already
 	% scanned past some whitespace, so '(' gets scanned as `open'
 	% rather than `open_ct'.
-	% We use a big switch on Char here for efficiency - obviously not
-	% for consiseness.
 
 lexer__get_token_2(Token) -->
 	io__read_char(Result),
@@ -292,144 +200,38 @@ lexer__get_token_2(Token) -->
 	; { Result = eof }, !,
 		{ Token = eof }
 	; { Result = ok(Char) },
-		( { Char = '\001', Token = junk(Char) }
-		; { Char = '\002', Token = junk(Char) }
-		; { Char = '\003', Token = junk(Char) }
-		; { Char = '\004', Token = junk(Char) }
-		; { Char = '\005', Token = junk(Char) }
-		; { Char = '\006', Token = junk(Char) }
-		; { Char = '\007', Token = junk(Char) }
-		; { Char = '\010', Token = junk(Char) }
-% \t		; { Char = '\011', Token = junk(Char) }
-		; { Char = '\t' }, lexer__get_token_2(Token)
-% \n		; { Char = '\012', Token = junk(Char) }
-		; { Char = '\n' }, lexer__get_token_2(Token)
-		; { Char = '\013', Token = junk(Char) }
-% \f		; { Char = '\014', Token = junk(Char) }
-		; { Char = '\f' }, lexer__get_token_2(Token)
-		; { Char = '\015', Token = junk(Char) }
-		; { Char = '\016', Token = junk(Char) }
-		; { Char = '\017', Token = junk(Char) }
-		; { Char = '\020', Token = junk(Char) }
-		; { Char = '\021', Token = junk(Char) }
-		; { Char = '\022', Token = junk(Char) }
-		; { Char = '\023', Token = junk(Char) }
-		; { Char = '\024', Token = junk(Char) }
-		; { Char = '\025', Token = junk(Char) }
-		; { Char = '\026', Token = junk(Char) }
-		; { Char = '\027', Token = junk(Char) }
-		; { Char = '\030', Token = junk(Char) }
-		; { Char = '\031', Token = junk(Char) }
-		; { Char = '\032', Token = junk(Char) }
-		; { Char = '\033', Token = junk(Char) }
-		; { Char = '\034', Token = junk(Char) }
-		; { Char = '\035', Token = junk(Char) }
-		; { Char = '\036', Token = junk(Char) }
-		; { Char = '\037', Token = junk(Char) }
-		; { Char = ' ' }, lexer__get_token_2(Token)
-		; { Char = ('!'),  Token = name("!") }
-		; { Char = '"' },
+		( { Char = ' ' ; Char = '\t' ; Char = '\n' } ->
+			lexer__get_token_2(Token)
+		; { char__is_upper(Char) ; Char = '_' } ->
+			lexer__get_variable([Char], Token)
+		; { char__is_lower(Char) } ->
+			lexer__get_name([Char], Token)
+		; { Char = '0' } ->
+			lexer__get_zero(Token)
+		; { char__is_digit(Char) } ->
+			lexer__get_number([Char], Token)
+		; { lexer__special_token(Char, SpecialToken) } ->
+			{ Token = SpecialToken }
+		; { Char = ('.') } ->
+			lexer__get_dot(Token)
+		; { Char = ('%') } ->
+			lexer__skip_to_eol(Token)
+		; { Char = '"' ; Char = '''' } ->
 			lexer__get_quoted_name(Char, [], Token)
-		; { Char = ('#') }, lexer__get_source_line_number([], Token)
-		; { Char = ('$') }, lexer__get_graphic([Char], Token)
-		; { Char = ('%') }, lexer__skip_to_eol(Token)
-		; { Char = ('&') }, lexer__get_graphic([Char], Token)
-		; { Char = '''' },
-			lexer__get_quoted_name(Char, [], Token)
-		; { Char = '(',  Token = open }
-		; { Char = ')',  Token = close }
-		; { Char = ('*') }, lexer__get_graphic([Char], Token)
-		; { Char = ('+') }, lexer__get_graphic([Char], Token)
-		; { Char = (','),  Token = comma }
-		; { Char = ('-') }, lexer__get_graphic([Char], Token)
-		; { Char = ('.') }, lexer__get_dot(Token)
-		; { Char = ('/') }, lexer__get_slash(Token)
-		; { Char = '0' }, lexer__get_zero(Token)
-		; { Char = '1' }, lexer__get_number([Char], Token)
-		; { Char = '2' }, lexer__get_number([Char], Token)
-		; { Char = '3' }, lexer__get_number([Char], Token)
-		; { Char = '4' }, lexer__get_number([Char], Token)
-		; { Char = '5' }, lexer__get_number([Char], Token)
-		; { Char = '6' }, lexer__get_number([Char], Token)
-		; { Char = '7' }, lexer__get_number([Char], Token)
-		; { Char = '8' }, lexer__get_number([Char], Token)
-		; { Char = '9' }, lexer__get_number([Char], Token)
-		; { Char = (':') }, lexer__get_graphic([Char], Token)
-		; { Char = (';'),  Token = name(";") }
-		; { Char = ('<') }, lexer__get_graphic([Char], Token)
-		; { Char = ('=') }, lexer__get_graphic([Char], Token)
-		; { Char = ('>') }, lexer__get_graphic([Char], Token)
-		; { Char = ('?') }, lexer__get_graphic([Char], Token)
-		; { Char = ('@') }, lexer__get_graphic([Char], Token)
-		; { Char = 'A' }, lexer__get_variable([Char], Token)
-		; { Char = 'B' }, lexer__get_variable([Char], Token)
-		; { Char = 'C' }, lexer__get_variable([Char], Token)
-		; { Char = 'D' }, lexer__get_variable([Char], Token)
-		; { Char = 'E' }, lexer__get_variable([Char], Token)
-		; { Char = 'F' }, lexer__get_variable([Char], Token)
-		; { Char = 'G' }, lexer__get_variable([Char], Token)
-		; { Char = 'H' }, lexer__get_variable([Char], Token)
-		; { Char = 'I' }, lexer__get_variable([Char], Token)
-		; { Char = 'J' }, lexer__get_variable([Char], Token)
-		; { Char = 'K' }, lexer__get_variable([Char], Token)
-		; { Char = 'L' }, lexer__get_variable([Char], Token)
-		; { Char = 'M' }, lexer__get_variable([Char], Token)
-		; { Char = 'N' }, lexer__get_variable([Char], Token)
-		; { Char = 'O' }, lexer__get_variable([Char], Token)
-		; { Char = 'P' }, lexer__get_variable([Char], Token)
-		; { Char = 'Q' }, lexer__get_variable([Char], Token)
-		; { Char = 'R' }, lexer__get_variable([Char], Token)
-		; { Char = 'S' }, lexer__get_variable([Char], Token)
-		; { Char = 'T' }, lexer__get_variable([Char], Token)
-		; { Char = 'U' }, lexer__get_variable([Char], Token)
-		; { Char = 'V' }, lexer__get_variable([Char], Token)
-		; { Char = 'W' }, lexer__get_variable([Char], Token)
-		; { Char = 'X' }, lexer__get_variable([Char], Token)
-		; { Char = 'Y' }, lexer__get_variable([Char], Token)
-		; { Char = 'Z' }, lexer__get_variable([Char], Token)
-		; { Char = ('['),  Token = open_list }
-		; { Char = ('\\') }, lexer__get_graphic([Char], Token)
-		; { Char = (']'),  Token = close_list }
-		; { Char = ('^') }, lexer__get_graphic([Char], Token)
-		; { Char = '_' }, lexer__get_variable([Char], Token)
-		; { Char = '\140', Token = junk(Char) } % `
-		; { Char = 'a' }, lexer__get_name([Char], Token)
-		; { Char = 'b' }, lexer__get_name([Char], Token)
-		; { Char = 'c' }, lexer__get_name([Char], Token)
-		; { Char = 'd' }, lexer__get_name([Char], Token)
-		; { Char = 'e' }, lexer__get_name([Char], Token)
-		; { Char = 'f' }, lexer__get_name([Char], Token)
-		; { Char = 'g' }, lexer__get_name([Char], Token)
-		; { Char = 'h' }, lexer__get_name([Char], Token)
-		; { Char = 'i' }, lexer__get_name([Char], Token)
-		; { Char = 'j' }, lexer__get_name([Char], Token)
-		; { Char = 'k' }, lexer__get_name([Char], Token)
-		; { Char = 'l' }, lexer__get_name([Char], Token)
-		; { Char = 'm' }, lexer__get_name([Char], Token)
-		; { Char = 'n' }, lexer__get_name([Char], Token)
-		; { Char = 'o' }, lexer__get_name([Char], Token)
-		; { Char = 'p' }, lexer__get_name([Char], Token)
-		; { Char = 'q' }, lexer__get_name([Char], Token)
-		; { Char = 'r' }, lexer__get_name([Char], Token)
-		; { Char = 's' }, lexer__get_name([Char], Token)
-		; { Char = 't' }, lexer__get_name([Char], Token)
-		; { Char = 'u' }, lexer__get_name([Char], Token)
-		; { Char = 'v' }, lexer__get_name([Char], Token)
-		; { Char = 'w' }, lexer__get_name([Char], Token)
-		; { Char = 'x' }, lexer__get_name([Char], Token)
-		; { Char = 'y' }, lexer__get_name([Char], Token)
-		; { Char = 'z' }, lexer__get_name([Char], Token)
-		; { Char = ('{'),  Token = open_curly }
-		; { Char = ('|'),  Token = ht_sep }
-		; { Char = ('}'),  Token = close_curly }
-		; { Char = ('~') }, lexer__get_graphic([Char], Token)
-		; { Char = '\177', Token = junk(Char) }
+		; { Char = ('/') } ->
+			lexer__get_slash(Token)
+		; { Char = ('#') } ->
+			lexer__get_source_line_number([], Token)
+		; { lexer__graphic_token_char(Char) } ->
+			lexer__get_graphic([Char], Token)
+		;
+			{ Token = junk(Char) }
 		)
 	).
 
 %-----------------------------------------------------------------------------%
 
-:- pred lexer__special_token(character, token).
+:- pred lexer__special_token(char, token).
 :- mode lexer__special_token(in, out) is semidet.
 
 lexer__special_token('(', open).	% May get converted to open_ct
@@ -443,9 +245,6 @@ lexer__special_token(',', comma).
 lexer__special_token(';', name(";")).
 lexer__special_token('!', name("!")).
 
-:- pred lexer__graphic_token_char(character).
-:- mode lexer__graphic_token_char(in) is semidet.
-
 lexer__graphic_token_char('#').
 lexer__graphic_token_char('$').
 lexer__graphic_token_char('&').
@@ -487,7 +286,7 @@ lexer__get_dot(Token) -->
 		)
 	).
 
-:- pred lexer__whitespace_after_dot(character).
+:- pred lexer__whitespace_after_dot(char).
 :- mode lexer__whitespace_after_dot(in) is semidet.
 
 lexer__whitespace_after_dot(' ').
@@ -576,7 +375,7 @@ lexer__get_comment_2(Token) -->
 
 	% quoted names and quoted strings
 
-:- pred lexer__get_quoted_name(character, list(character), token,
+:- pred lexer__get_quoted_name(char, list(char), token,
 				io__state, io__state).
 :- mode lexer__get_quoted_name(in, in, out, di, uo) is det.
 
@@ -596,7 +395,7 @@ lexer__get_quoted_name(QuoteChar, Chars, Token) -->
 		)
 	).
 
-:- pred lexer__get_quoted_name_quote(character, list(character), token,
+:- pred lexer__get_quoted_name_quote(char, list(char), token,
 				io__state, io__state).
 :- mode lexer__get_quoted_name_quote(in, in, out, di, uo) is det.
 
@@ -615,7 +414,7 @@ lexer__get_quoted_name_quote(QuoteChar, Chars, Token) -->
 		)
 	).
 
-:- pred lexer__finish_quoted_name(character, list(character), token).
+:- pred lexer__finish_quoted_name(char, list(char), token).
 :- mode lexer__finish_quoted_name(in, in, out) is det.
 
 lexer__finish_quoted_name(QuoteChar, Chars, Token) :-
@@ -628,7 +427,7 @@ lexer__finish_quoted_name(QuoteChar, Chars, Token) :-
 		error("lexer.m: unknown quote character")
 	).
 
-:- pred lexer__get_quoted_name_escape(character, list(character), token,
+:- pred lexer__get_quoted_name_escape(char, list(char), token,
 					io__state, io__state).
 :- mode lexer__get_quoted_name_escape(in, in, out, di, uo) is det.
 
@@ -654,7 +453,7 @@ lexer__get_quoted_name_escape(QuoteChar, Chars, Token) -->
 		)
 	).
 
-:- pred lexer__escape_char(character, character).
+:- pred lexer__escape_char(char, char).
 :- mode lexer__escape_char(in, out) is semidet.
 
 lexer__escape_char('a', '\a').
@@ -669,7 +468,7 @@ lexer__escape_char('''', '''').
 lexer__escape_char('"', '"').
 lexer__escape_char('`', '`').
 
-:- pred lexer__get_hex_escape(character, list(character), list(character),
+:- pred lexer__get_hex_escape(char, list(char), list(char),
 				token, io__state, io__state).
 :- mode lexer__get_hex_escape(in, in, in, out, di, uo) is det.
 
@@ -691,7 +490,7 @@ lexer__get_hex_escape(QuoteChar, Chars, HexChars, Token) -->
 		)
 	).
 
-:- pred lexer__finish_hex_escape(character, list(character), list(character),
+:- pred lexer__finish_hex_escape(char, list(char), list(char),
 				token, io__state, io__state).
 :- mode lexer__finish_hex_escape(in, in, in, out, di, uo) is det.
 
@@ -710,7 +509,7 @@ lexer__finish_hex_escape(QuoteChar, Chars, HexChars, Token) -->
 		)
 	).
 
-:- pred lexer__get_octal_escape(character, list(character), list(character),
+:- pred lexer__get_octal_escape(char, list(char), list(char),
 				token, io__state, io__state).
 :- mode lexer__get_octal_escape(in, in, in, out, di, uo) is det.
 
@@ -740,7 +539,7 @@ lexer__get_octal_escape(QuoteChar, Chars, OctalChars, Token) -->
 		)
 	).
 
-:- pred lexer__finish_octal_escape(character, list(character), list(character),
+:- pred lexer__finish_octal_escape(char, list(char), list(char),
 				token, io__state, io__state).
 :- mode lexer__finish_octal_escape(in, in, in, out, di, uo) is det.
 
@@ -763,7 +562,7 @@ lexer__finish_octal_escape(QuoteChar, Chars, OctalChars, Token) -->
 
 	% names and variables
 
-:- pred lexer__get_name(list(character), token, io__state, io__state).
+:- pred lexer__get_name(list(char), token, io__state, io__state).
 :- mode lexer__get_name(in, out, di, uo) is det.
 
 lexer__get_name(Chars, Token) -->
@@ -832,7 +631,7 @@ lexer__get_source_line_number(Chars, Token) -->
 		)
 	).
 
-:- pred lexer__get_graphic(list(character), token, io__state, io__state).
+:- pred lexer__get_graphic(list(char), token, io__state, io__state).
 :- mode lexer__get_graphic(in, out, di, uo) is det.
 
 lexer__get_graphic(Chars, Token) -->
@@ -852,7 +651,7 @@ lexer__get_graphic(Chars, Token) -->
 		)
 	).
 
-:- pred lexer__get_variable(list(character), token, io__state, io__state).
+:- pred lexer__get_variable(list(char), token, io__state, io__state).
 :- mode lexer__get_variable(in, out, di, uo) is det.
 
 lexer__get_variable(Chars, Token) -->
@@ -938,7 +737,7 @@ lexer__get_binary(Token) -->
 		)
 	).
 
-:- pred lexer__get_binary_2(list(character), token, io__state, io__state).
+:- pred lexer__get_binary_2(list(char), token, io__state, io__state).
 :- mode lexer__get_binary_2(in, out, di, uo) is det.
 
 lexer__get_binary_2(Chars, Token) -->
@@ -974,7 +773,7 @@ lexer__get_octal(Token) -->
 		)
 	).
 
-:- pred lexer__get_octal_2(list(character), token, io__state, io__state).
+:- pred lexer__get_octal_2(list(char), token, io__state, io__state).
 :- mode lexer__get_octal_2(in, out, di, uo) is det.
 
 lexer__get_octal_2(Chars, Token) -->
@@ -1010,7 +809,7 @@ lexer__get_hex(Token) -->
 		)
 	).
 
-:- pred lexer__get_hex_2(list(character), token, io__state, io__state).
+:- pred lexer__get_hex_2(list(char), token, io__state, io__state).
 :- mode lexer__get_hex_2(in, out, di, uo) is det.
 
 lexer__get_hex_2(Chars, Token) -->
@@ -1028,7 +827,7 @@ lexer__get_hex_2(Chars, Token) -->
 		)
 	).
 
-:- pred lexer__get_number(list(character), token, io__state, io__state).
+:- pred lexer__get_number(list(char), token, io__state, io__state).
 :- mode lexer__get_number(in, out, di, uo) is det.
 
 lexer__get_number(Chars, Token) -->
@@ -1052,7 +851,7 @@ lexer__get_number(Chars, Token) -->
 
 	% XXX the float literal syntax doesn't match ISO Prolog
 
-:- pred lexer__get_int_dot(list(character), token, io__state, io__state).
+:- pred lexer__get_int_dot(list(char), token, io__state, io__state).
 :- mode lexer__get_int_dot(in, out, di, uo) is det.
 
 lexer__get_int_dot(Chars, Token) -->
@@ -1072,7 +871,7 @@ lexer__get_int_dot(Chars, Token) -->
 		)
 	).
 
-:- pred lexer__get_float_decimals(list(character), token, io__state, io__state).
+:- pred lexer__get_float_decimals(list(char), token, io__state, io__state).
 :- mode lexer__get_float_decimals(in, out, di, uo) is det.
 
 	% we've read past the decimal point, so now get the decimals
@@ -1094,7 +893,7 @@ lexer__get_float_decimals(Chars, Token) -->
 		)
 	).
 
-:- pred lexer__get_float_exponent(list(character), token, io__state, io__state).
+:- pred lexer__get_float_exponent(list(char), token, io__state, io__state).
 :- mode lexer__get_float_exponent(in, out, di, uo) is det.
 
 lexer__get_float_exponent(Chars, Token) -->
@@ -1115,7 +914,7 @@ lexer__get_float_exponent(Chars, Token) -->
 		)
 	).
 
-:- pred lexer__get_float_exponent_2(list(character), token,
+:- pred lexer__get_float_exponent_2(list(char), token,
 				io__state, io__state).
 :- mode lexer__get_float_exponent_2(in, out, di, uo) is det.
 
@@ -1139,7 +938,7 @@ lexer__get_float_exponent_2(Chars, Token) -->
 		)
 	).
 
-:- pred lexer__get_float_exponent_3(list(character), token,
+:- pred lexer__get_float_exponent_3(list(char), token,
 					io__state, io__state).
 :- mode lexer__get_float_exponent_3(in, out, di, uo) is det.
 
@@ -1165,7 +964,7 @@ lexer__get_float_exponent_3(Chars, Token) -->
 
 	% Utility routines
 
-:- pred lexer__rev_char_list_to_int(list(character), int, token).
+:- pred lexer__rev_char_list_to_int(list(char), int, token).
 :- mode lexer__rev_char_list_to_int(in, in, out) is det.
 
 lexer__rev_char_list_to_int(RevChars, Base, Token) :-
@@ -1176,7 +975,7 @@ lexer__rev_char_list_to_int(RevChars, Base, Token) :-
 		Token = error("invalid integer token")
 	).
 
-:- pred lexer__rev_char_list_to_float(list(character), token).
+:- pred lexer__rev_char_list_to_float(list(char), token).
 :- mode lexer__rev_char_list_to_float(in, out) is det.
 
 lexer__rev_char_list_to_float(RevChars, Token) :-
@@ -1187,55 +986,10 @@ lexer__rev_char_list_to_float(RevChars, Token) :-
 		Token = error("invalid float token")
 	).
 
-:- pred lexer__rev_char_list_to_string(list(character), string).
+:- pred lexer__rev_char_list_to_string(list(char), string).
 :- mode lexer__rev_char_list_to_string(in, out) is det.
 
-% lexer__rev_char_list_to_string was originally implemented like this:
-%
-%lexer__rev_char_list_to_string(RevChars, String) :-
-%       list__reverse(RevChars, Chars),
-%       string__from_char_list(Chars, String).
-%
-% The optimized implementation in C below is there for efficiency since
-% it improves the overall speed of parsing by about 7%.
-
-:- pragma(c_code, lexer__rev_char_list_to_string(Chars::in, Str::out), "
-{
-	Word list_ptr;
-	Word size, len;
-	Word str_ptr;
-/*
-** loop to calculate list length + sizeof(Word) in `size' using list in
-** `list_ptr' and separately count the length of the string
-*/
-	size = sizeof(Word);
-	len = 1;
-	list_ptr = Chars;
-	while (!list_is_empty(list_ptr)) {
-		size++;
-		len++;
-		list_ptr = list_tail(list_ptr);
-	}
-/*
-** allocate (length + 1) bytes of heap space for string
-** i.e. (length + 1 + sizeof(Word) - 1) / sizeof(Word) words
-*/
-	incr_hp_atomic(str_ptr, size / sizeof(Word));
-	Str = (char *) str_ptr;
-/*
-** set size to be the offset of the end of the string
-** (ie the \\0) and null terminate the string.
-*/
-	Str[--len] = '\\0';
-/*
-** loop to copy the characters from the list_ptr to the string
-** in reverse order.
-*/
-	list_ptr = Chars;
-	while (!list_is_empty(list_ptr)) {
-		Str[--len] = (char) list_head(list_ptr);
-		list_ptr = list_tail(list_ptr);
-	}
-}").
+lexer__rev_char_list_to_string(RevChars, String) :-
+       string__from_rev_char_list(RevChars, String).
 
 %-----------------------------------------------------------------------------%
