@@ -1,5 +1,5 @@
 %---------------------------------------------------------------------------%
-% Copyright (C) 1995 University of Melbourne.
+% Copyright (C) 1993-1997 The University of Melbourne.
 % This file may only be copied under the terms of the GNU Library General
 % Public License - see the file COPYING.LIB in the Mercury distribution.
 %---------------------------------------------------------------------------%
@@ -7,15 +7,14 @@
 % Module `list' - defines the list type, and various utility predicates
 % that operate on lists.
 %
-% Main author: fjh.
+% Authors: fjh, conway, trd, zs, philip, warwick, ...
 % Stability: medium to high.
 %
 %---------------------------------------------------------------------------%
 
 :- module list.
-:- import_module int.
-
 :- interface.
+:- import_module int.
 
 %-----------------------------------------------------------------------------%
 
@@ -25,11 +24,7 @@
 	%	of type `list(T)', denoted `[Head | Tail]'.
 	%	
 
-% :- type list(T) ---> [] ; [T | list(T)].
-
-	% The definition of `list(T)' is actually in mercury_builtin.m because
-	% the implementation of type_to_term/3 and term_to_type/3 means
-	% that it has to be.
+:- type list(T) ---> [] ; [T | list(T)].
 
 %-----------------------------------------------------------------------------%
 
@@ -53,7 +48,7 @@
 :- mode output_list_skel :: out_list_skel.
 :- mode list_skel_output :: list_skel_out.
 
-	% These modes are particularly useful for passing round lists
+	% These modes are particularly useful for passing around lists
 	% of higher order terms, since they have complicated insts
 	% which are not correctly approximated by "ground".
 :- mode list_skel_in(I) :: list_skel(I) -> list_skel(I).
@@ -74,7 +69,7 @@
 %	The following mode is semidet in the sense that it doesn't
 %	succeed more than once - but it does create a choice-point,
 %	which means it's inefficient and that the compiler can't deduce
-%	that it is semidet.  Use list__remove_prefix instead.
+%	that it is semidet.  Use list__remove_suffix instead.
 % :- mode list__append(out, in, in) is semidet.
 
 	% list__remove_suffix(List, Suffix, Prefix):
@@ -84,15 +79,16 @@
 :- mode list__remove_suffix(in, in, out) is semidet.
 
 	% list__merge(L1, L2, L):
-	%	L is the result of merging L1 and L2.
-	%	L1 and L2 must be sorted.
+	%	L is the result of merging the elements of L1 and L2,
+	%	in ascending order.  L1 and L2 must be sorted.
 :- pred list__merge(list(T), list(T), list(T)).
 :- mode list__merge(in, in, out) is det.
 
 	% list__merge_and_remove_dups(L1, L2, L):
-	%	L is the result of merging L1 and L2 and eliminating
-	%	any duplicates.
-	%	L1 and L2 must be sorted.
+	%	L is the result of merging the elements of L1 and L2,
+	%	in ascending order, and eliminating any duplicates.
+	%	L1 and L2 must be sorted and must each not contain any
+	%	duplicates.
 :- pred list__merge_and_remove_dups(list(T), list(T), list(T)).
 :- mode list__merge_and_remove_dups(in, in, out) is det.
 
@@ -128,7 +124,6 @@
 	%
 :- pred list__length(list(_T), int).
 :- mode list__length(in, out) is det.
-% :- mode list__length(in, in) is semidet.	% implied
 	% XXX The current mode checker can't handle this mode
 % :- mode list__length(input_list_skel, out) is det.
 
@@ -325,14 +320,10 @@
 % various list processing tasks. They implement pretty much standard
 % sorts of operations provided by standard libraries for functional languages.
 %
-% Most of this code was originally by philip, modified and reformatted
-% by conway.
-%
 %-----------------------------------------------------------------------------%
-:- import_module std_util.
 
 	% list__map(T, L, M) uses the closure T
-	% to transform the elements of L into the elements of L.
+	% to transform the elements of L into the elements of M.
 :- pred list__map(pred(X, Y), list(X), list(Y)).
 :- mode list__map(pred(in, out) is det, in, out) is det.
 :- mode list__map(pred(in, out) is semidet, in, out) is semidet.
@@ -356,6 +347,35 @@
 :- mode list__foldr(pred(in, in, out) is det, in, in, out) is det.
 :- mode list__foldr(pred(in, in, out) is semidet, in, in, out) is semidet.
 
+	% list__foldl2(Pred, List, Start, End, Start2, End2) 
+	% calls Pred with each element of List (working left-to-right),
+	% 2 accumulators (with the initial values of Start and Start2),
+	% and returns the final values in End and End2.
+	% (Although no more expressive than list__foldl, this is often
+	% a more convenient format, and a little more efficient).
+:- pred list__foldl2(pred(X, Y, Y, Z, Z), list(X), Y, Y, Z, Z).
+:- mode list__foldl2(pred(in, in, out, in, out) is det,
+		in, in, out, in, out) is det.
+:- mode list__foldl2(pred(in, in, out, mdi, muo) is det,
+		in, in, out, mdi, muo) is det.
+:- mode list__foldl2(pred(in, in, out, di, uo) is det,
+		in, in, out, di, uo) is det.
+:- mode list__foldl2(pred(in, di, uo, di, uo) is det,
+		in, di, uo, di, uo) is det.
+
+	% list__map_foldl(Pred, InList, OutList, Start, End) calls Pred
+	% with an accumulator (with the initial value of Start) on
+	% each element of InList (working left-to-right) to transform
+	% InList into OutList.  The final value of the acumulator is
+	% returned in End.
+:- pred list__map_foldl(pred(X, Y, Z, Z), list(X), list(Y), Z, Z).
+:- mode list__map_foldl(pred(in, out, di, uo) is det, in, out, di, uo)
+								is det.
+:- mode list__map_foldl(pred(in, out, in, out) is det, in, out, in, out)
+								is det.
+:- mode list__map_foldl(pred(in, out, in, out) is semidet, in, out, in, out)
+                                                                is semidet.
+
 	% list__filter(Pred, List, TrueList) takes a closure with one
 	% input argument and for each member of List `X', calls the closure.
 	% Iff call(Pred, X) is true, then X is included in TrueList.
@@ -376,22 +396,58 @@
 :- pred list__filter_map(pred(X, Y), list(X), list(Y)).
 :- mode list__filter_map(pred(in, out) is semidet, in, out) is det.
 
+	% list__filter_map(Transformer, List, TrueList, FalseList) takes
+	% a predicate with one input argument and one output argument.
+	% It is called with each element of List. If a call succeeds,
+	% then the output is included in TrueList; otherwise, the failing
+	% input is included in FalseList.
+:- pred list__filter_map(pred(X, Y), list(X), list(Y), list(X)).
+:- mode list__filter_map(pred(in, out) is semidet, in, out, out) is det.
+
+	% list__takewhile(Predicate, List, UptoList, AfterList) takes a
+	% closure with one input argument, and calls it on successive members
+	% of List as long as the calls succeed. The elements for which
+	% the call succeeds are placed in UptoList and the first element for
+	% which the call fails, and all the remaining elements of List are
+	% placed in AfterList.
+:- pred list__takewhile(pred(T), list(T), list(T), list(T)).
+:- mode list__takewhile(pred(in) is semidet, in, out, out) is det.
+
+%-----------------------------------------------------------------------------%
+
 	% list__sort(Compare, Unsorted, Sorted) is true iff Sorted is a
 	% list containing the same elements as Unsorted, where Sorted is
-	% a sorted list, wrt the ordering defined by the predicate term
-	% Compare.
-	% (implementation due to Philip Dart)
+	% a sorted list, with respect to the ordering defined by the predicate
+	% term Compare.
 :- pred list__sort(pred(X, X, comparison_result), list(X), list(X)).
 :- mode list__sort(pred(in, in, out) is det, in, out) is det.
 
+	% list__sort_and_remove_dups(Compare, Unsorted, Sorted) is true iff 
+	% Sorted is a list containing the same elements as Unsorted, but with
+	% any duplicates removed. Where Sorted is a sorted list, with respect  
+	% to the ordering defined by the predicate term Compare.
+:- pred list__sort_and_remove_dups(pred(X, X, comparison_result), list(X), 
+	list(X)).
+:- mode list__sort_and_remove_dups(pred(in, in, out) is det, in, out) is det.
+
 	% list__merge(Compare, As, Bs, Sorted) is true iff Sorted is a
 	% list containing the elements of As and Bs in the order implied
 	% by their sorted merge. The ordering of elements is defined by
 	% the higher order comparison predicate Compare.
-	% (implementation due to Philip Dart)
 :- pred list__merge(pred(X, X, comparison_result), list(X), list(X), list(X)).
 :- mode list__merge(pred(in, in, out) is det, in, in, out) is det.
 
+	% list__merge_and_remove_dups(P, As, Bs, Sorted) is true if and only if
+	% Sorted is a list containing the elements of As and Bs in the order 
+	% implied by their sorted merge. The ordering of elements is defined by
+	% the higher order comparison predicate P.
+	% As and Bs must be sorted.
+:- pred list__merge_and_remove_dups(pred(X, X, comparison_result),
+	list(X), list(X), list(X)).
+:- mode list__merge_and_remove_dups(pred(in, in, out) is det,
+	in, in, out) is det.
+
+
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
 
@@ -581,7 +637,6 @@ list__length(L, N) :-
 
 :- pred list__length_2(list(T), int, int).
 :- mode list__length_2(in, in, out) is det.
-% :- mode list__length_2(in, in, in) is semidet. % implied
 
 list__length_2([], N, N).
 list__length_2([_ | L1], N0, N) :-
@@ -849,6 +904,17 @@ list__foldl(P, [H|T], Acc0, Acc) :-
 	call(P, H, Acc0, Acc1),
 	list__foldl(P, T, Acc1, Acc).
 
+list__foldl2(_, [], FirstAcc, FirstAcc, SecAcc, SecAcc).
+list__foldl2(P, [H|T], FirstAcc0, FirstAcc, SecAcc0, SecAcc) :-
+	call(P, H, FirstAcc0, FirstAcc1, SecAcc0, SecAcc1),
+	list__foldl2(P, T, FirstAcc1, FirstAcc, SecAcc1, SecAcc).
+
+list__map_foldl(_, [],  []) -->
+        [].
+list__map_foldl(P, [H0|T0], [H|T]) -->
+        call(P, H0, H),
+        list__map_foldl(P, T0, T).
+
 list__foldr(_, [], Acc, Acc).
 list__foldr(P, [H|T], Acc0, Acc) :-
 	list__foldr(P, T, Acc0, Acc1),
@@ -877,6 +943,30 @@ list__filter_map(P, [H0|T0], L) :-
 	),
 	list__filter_map(P, T0, L1).
 
+list__filter_map(_, [], [], []).
+list__filter_map(P, [H0|T0], L, M) :-
+        ( call(P, H0, H) ->
+                L = [H|L1],
+		M = M1
+        ;
+                L = L1,
+		M = [H0|M1]
+        ),
+        list__filter_map(P, T0, L1, M1).
+
+list__takewhile(_, [], [], []).
+list__takewhile(P, [X|Xs], Ins, Outs) :-
+	( call(P, X) ->
+		Ins = [X|Ins0],
+		list__takewhile(P, Xs, Ins0, Outs)
+	;
+		Ins = [],
+		Outs = [X|Xs]
+	).
+
+list__sort_and_remove_dups(P, L0, L) :-
+	list__sort(P, L0, L1),
+	list__remove_adjacent_dups(L1, L).
 
 list__sort(P, L0, L) :-
         list__length(L0, N),
@@ -951,24 +1041,24 @@ list__merge(P, [H1|T1], [H2|T2], L) :-
 		list__merge(P, [H1|T1], T2, T)
 	).
 
-%-----------------------------------------------------------------------------%
+list__merge_and_remove_dups(_P, [], [], []).
+list__merge_and_remove_dups(_P, [], [Y|Ys], [Y|Ys]).
+list__merge_and_remove_dups(_P, [X|Xs], [], [X|Xs]).
+list__merge_and_remove_dups(P, [H1|T1], [H2|T2], L) :-
+	call(P, H1, H2, C),
+	(
+		C = (<),
+		L = [H1|T],   
+		list__merge(P, T1, [H2|T2], T)
+	;
+		C = (=),
+		L = [H1 | T],
+		list__merge(P, T1, T2, T)
+	;
+		C = (>),
+		L = [H2|T],   
+		list__merge(P, [H1|T1], T2, T)
+	).
 
-:- interface.
-
-	% list__apply(Cs, Bs) takes a list of closures with one
-	% output argument Cs, and calls the closures, returning
-	% the resulting bindings in Bs.
-:- pred list__apply(list(pred(T)), list(T)).
-:- mode list__apply(list_skel_in(pred(out) is det), out) is det.
-:- mode list__apply(list_skel_in(pred(out) is semidet), out) is semidet.
-:- mode list__apply(list_skel_in(pred(out) is multi), out) is multi.
-:- mode list__apply(list_skel_in(pred(out) is nondet), out) is nondet.
-
-:- implementation.
-
-list__apply([], []).
-list__apply([C|Cs], [B|Bs]) :-
-	call(C, B),
-	list__apply(Cs, Bs).
 
 %-----------------------------------------------------------------------------%
