@@ -1,5 +1,5 @@
 %---------------------------------------------------------------------------%
-% Copyright (C) 1995 University of Melbourne.
+% Copyright (C) 1994-1997 The University of Melbourne.
 % This file may only be copied under the terms of the GNU Library General
 % Public License - see the file COPYING.LIB in the Mercury distribution.
 %---------------------------------------------------------------------------%
@@ -25,6 +25,7 @@
 
 % The types `character', `int', `float', and `string',
 % and the types `pred', `pred(T)', `pred(T1, T2)', `pred(T1, T2, T3)', ...
+% and `func(T1) = T2', `func(T1, T2) = T3', `func(T1, T2, T3) = T4', ...
 % are builtin and are implemented using special code in the
 % type-checker.  (XXX TODO: report an error for attempts to redefine
 % these types.)
@@ -42,13 +43,17 @@
 % So are the standard unique insts `unique', `unique(...)',
 % `mostly_unique', `mostly_unique(...)', and `clobbered'.
 % The name `dead' is allowed as a synonym for `clobbered'.
+% Similarly `mostly_dead' is a synonym for `mostly_clobbered'.
 
 :- inst dead = clobbered.
+:- inst mostly_dead = mostly_clobbered.
 
 % The not yet properly supported `any' inst used for the
 % constraint solver interface is also builtin.
 
-% Higher-order predicate insts `pred(<modes>) is <detism>' are also builtin.
+% Higher-order predicate insts `pred(<modes>) is <detism>'
+% and higher-order functions insts `func(<modes>) = <mode> is det'
+% are also builtin.
 
 %-----------------------------------------------------------------------------%
 
@@ -66,6 +71,7 @@
 :- mode in(Inst) :: (Inst -> Inst).
 :- mode out(Inst) :: (free -> Inst).
 :- mode di(Inst) :: (Inst -> clobbered).
+:- mode mdi(Inst) :: (Inst -> mostly_clobbered).
 
 % Unique modes.  These are still not fully implemented.
 
@@ -96,15 +102,22 @@
 
 % PREDICATES.
 
-% copy/2 is used to make a `unique' copy of a data structure,
-% so that you can use destructive update.
-% At the moment it doesn't actually do any copying, since we
-% haven't implemented destructive update yet and so there is no need.
+% copy/2 makes a deep copy of a data structure.  The resulting copy is a
+% `unique' value, so you can use destructive update on it.
 
 :- pred copy(T, T).
 :- mode copy(ui, uo) is det.
 :- mode copy(in, uo) is det.
 
+% unsafe_promise_unique/2 is used to promise the compiler that you have a
+% `unique' copy of a data structure, so that you can use destructive update.
+% It is used to work around limitations in the current support for unique
+% modes.  `unsafe_promise_unique(X, Y)' is the same as `Y = X' except that
+% the compiler will assume that `Y' is unique.
+
+:- pred unsafe_promise_unique(T, T).
+:- mode unsafe_promise_unique(in, uo) is det.
+
 % We define !/0 (and !/2 for dcgs) to be equivalent to `true'.  This is for
 % backwards compatibility with Prolog systems.  But of course it only works
 % if all your cuts are green cuts.
@@ -151,16 +164,6 @@
 :- mode compare(uo, in, ui) is det.
 :- mode compare(uo, in, in) is det.
 
-	% The following three predicates can convert values of any
-	% type to the type `term' and back again.
-	% However, they are not yet implemented.
-
-:- pred term_to_type(term :: in, T :: out) is semidet.
-
-:- pred det_term_to_type(term :: in, T :: out) is det.
-
-:- pred type_to_term(T :: in, term :: out) is det.
-
 %-----------------------------------------------------------------------------%
 
 :- implementation.
@@ -184,36 +187,33 @@
 :- pred builtin_unify_int(int::in, int::in) is semidet.
 :- pred builtin_index_int(int::in, int::out) is det.
 :- pred builtin_compare_int(comparison_result::out, int::in, int::in) is det.
-:- pred builtin_term_to_type_int(term :: in, int :: out) is semidet.
-:- pred builtin_type_to_term_int(int :: in, term :: out) is det.
 
 :- pred builtin_unify_character(character::in, character::in) is semidet.
 :- pred builtin_index_character(character::in, int::out) is det.
 :- pred builtin_compare_character(comparison_result::out, character::in,
 	character::in) is det.
-:- pred builtin_term_to_type_character(term :: in, character :: out) is semidet.
-:- pred builtin_type_to_term_character(character :: in, term :: out) is det.
 
 :- pred builtin_unify_string(string::in, string::in) is semidet.
 :- pred builtin_index_string(string::in, int::out) is det.
 :- pred builtin_compare_string(comparison_result::out, string::in, string::in)
 	is det.
-:- pred builtin_term_to_type_string(term :: in, string :: out) is semidet.
-:- pred builtin_type_to_term_string(string :: in, term :: out) is det.
 
 :- pred builtin_unify_float(float::in, float::in) is semidet.
 :- pred builtin_index_float(float::in, int::out) is det.
 :- pred builtin_compare_float(comparison_result::out, float::in, float::in)
 	is det.
-:- pred builtin_term_to_type_float(term :: in, float :: out) is semidet.
-:- pred builtin_type_to_term_float(float :: in, term :: out) is det.
 
 :- pred builtin_unify_pred((pred)::in, (pred)::in) is semidet.
 :- pred builtin_index_pred((pred)::in, int::out) is det.
 :- pred builtin_compare_pred(comparison_result::out, (pred)::in, (pred)::in)
 	is det.
-:- pred builtin_term_to_type_pred(term::in, (pred)::out) is semidet.
-:- pred builtin_type_to_term_pred((pred)::in, term::out) is det.
+
+% The following two preds are used for index/1 or compare/3 on
+% non-canonical types (types for which there is a `where equality is ...'
+% declaration).
+:- pred builtin_index_non_canonical_type(T::in, int::out) is det.
+:- pred builtin_compare_non_canonical_type(comparison_result::out,
+		T::in, T::in) is det.
 
 :- pred unused is det.
 
@@ -245,67 +245,10 @@
 :- mode builtin_int_gt(in, in) is semidet.
 :- external(builtin_int_gt/2).
 
-% The types term, const, var and var_supply should be defined in
-% term.m, but we define them here since they're need for implementation
-% of term_to_type/2 and type_to_term/2.
-
-:- type term		--->	term__functor(const, list(term), term__context)
-			;	term__variable(var).
-:- type const		--->	term__atom(string)
-			;	term__integer(int)
-			;	term__string(string)
-			;	term__float(float).
-:- type var.
-:- type var_supply.
-
-% The type list should be defined in list.m, but we define it here since
-% it is needed for the definition of type term, which is needed for the
-% implementation of term_to_type/2 and type_to_term/2.
-
-:- type list(T) ---> [] ; [T | list(T)].
-
-        % At the moment, the only context we store is the line
-        % number.
-
-:- type term__context	--->	term__context(string, int).
-				% file name, line number.
-
-%-----------------------------------------------------------------------------%
-
-% The following three predicates should be defined in term.m, but
-% type var has to be defined here for term_to_type.
-
-% To manage a supply of variables, use the following 2 predicates.
-% (We might want to give these a unique mode later.)
-
-:- pred term__init_var_supply(var_supply).
-:- mode term__init_var_supply(out) is det.
-:- mode term__init_var_supply(in) is semidet. % implied
-%       term__init_var_supply(VarSupply) :
-%               returns a fresh var_supply for producing fresh variables.
-
-:- pred term__create_var(var_supply, var, var_supply).
-:- mode term__create_var(in, out, out) is det.
-%       term__create_var(VarSupply0, Variable, VarSupply) :
-%               create a fresh variable (var) and return the
-%               updated var_supply.
-
-:- pred term__var_to_int(var, int).
-:- mode term__var_to_int(in, out) is det.
-%               Convert a variable to an int.
-%               Different variables map to different ints.
-%               Other than that, the mapping is unspecified.
-
-%-----------------------------------------------------------------------------%
-
-
-:- pred term__context_init(term__context).
-:- mode term__context_init(out) is det.
-
 %-----------------------------------------------------------------------------%
 
 :- implementation.
-:- import_module require, std_util, int, float, char, string, list.
+:- import_module require, string, std_util, int, float, char, string, list.
 
 % Many of the predicates defined in this module are builtin -
 % the compiler generates code for them inline.
@@ -320,15 +263,8 @@
 :- external(unify/2).
 :- external(index/2).
 :- external(compare/3).
-:- external(term_to_type/2).
-:- external(type_to_term/2).
 
-det_term_to_type(Term, X) :-
-	( term_to_type(Term, X1) ->
-		X = X1
-	;
-		error("det_term_to_type failed as term doesn't represent a valid ground value of the appropriate type")
-	).
+%-----------------------------------------------------------------------------%
 
 builtin_unify_int(X, X).
 
@@ -343,12 +279,6 @@ builtin_compare_int(R, X, Y) :-
 		R = (>)
 	).
 
-builtin_term_to_type_int(term__functor(term__integer(Int), _TermList, _Context),
-									Int).
-
-builtin_type_to_term_int(Int, term__functor(term__integer(Int), [], Context)) :-
-	term__context_init(Context).
-
 builtin_unify_character(C, C).
 
 builtin_index_character(C, N) :-
@@ -365,15 +295,6 @@ builtin_compare_character(R, X, Y) :-
 		R = (>)
 	).
 
-builtin_term_to_type_character(
-		term__functor(term__atom(String), [], _Context), Character) :-
-	string__first_char(String, Character, "").
-
-builtin_type_to_term_character(
-		Character, term__functor(term__atom(String), [], Context)) :-
-	term__context_init(Context),
-	string__char_to_string(Character, String).
-
 builtin_unify_string(S, S).
 
 builtin_index_string(_, -1).
@@ -388,77 +309,44 @@ builtin_compare_string(R, S1, S2) :-
 		R = (>)
 	).
 
-builtin_term_to_type_string(
-		term__functor(term__string(String), [], _Context), String).
-
-builtin_type_to_term_string(
-		String, term__functor(term__string(String), [], Context)) :- 
-        term__context_init(Context).
-
 builtin_unify_float(F, F).
 
 builtin_index_float(_, -1).
 
 builtin_compare_float(R, F1, F2) :-
-	( builtin_float_lt(F1, F2) ->
+	( F1 < F2 ->
 		R = (<)
-	; builtin_float_gt(F1, F2) ->
+	; F1 > F2 ->
 		R = (>)
 	;
 		R = (=)
 	).
 
-builtin_term_to_type_float(term__functor(
-			term__float(Float), _TermList, _Context), Float).
-
-builtin_type_to_term_float(
-		Float, term__functor(term__float(Float), [], Context)) :-
-	term__context_init(Context).
-
 :- pred builtin_strcmp(int, string, string).
 :- mode builtin_strcmp(out, in, in) is det.
 
 :- pragma(c_code, builtin_strcmp(Res::out, S1::in, S2::in),
 	"Res = strcmp(S1, S2);").
 
-builtin_unify_pred(_Pred1, _Pred2) :-
+builtin_index_non_canonical_type(_, -1).
+
+builtin_compare_non_canonical_type(Res, X, _Y) :-
 	% suppress determinism warning
 	( semidet_succeed ->
-		error("attempted unification of higher-order predicate terms")
-	;
-		semidet_fail
-	).
-
-builtin_index_pred(_, -1).
-
-builtin_compare_pred(Res, _Pred1, _Pred2) :-
-	% suppress determinism warning
-	( semidet_succeed ->
-		error("attempted comparison of higher-order predicate terms")
+		string__append_list([
+			"call to compare/3 for non-canonical type `",
+			type_name(type_of(X)),
+			"'"],
+			Message),
+		error(Message)
 	;
 		% the following is never executed
 		Res = (<)
 	).
 
-builtin_term_to_type_pred(_Term, Pred) :-
-	% suppress determinism warning
-	semidet_succeed,
-	( semidet_succeed ->
-		error("attempted conversion of a term to a higher-order predicate")
-	;
-		% the following is never executed
-		Pred = semidet_succeed
-	).
-
-builtin_type_to_term_pred(_Pred, Term) :-
-	% suppress determinism warning
-	( semidet_succeed ->
-		error("attempted comparison of higher-order predicate terms")
-	;
-		% the following is never executed
-		term__context_init(Context),
-		Term = term__functor(term__atom(""), [], Context)
-	).
+:- external(builtin_unify_pred/2).
+:- external(builtin_index_pred/2).
+:- external(builtin_compare_pred/3).
 
 unused :-
 	( semidet_succeed ->
@@ -471,185 +359,312 @@ unused :-
 :- pragma(c_header_code, "#include ""type_info.h""").
 
 :- pragma(c_code, "
-#ifdef	SHARED_ONE_OR_TWO_CELL_TYPE_INFO
 
 
 #ifdef  USE_TYPE_LAYOUT
 
-Word * mercury_data___base_type_layout_int_0[] = {
+	/* base_type_layout definitions */ 
+
+	/* base_type_layout for `int' */
+
+const struct mercury_data___base_type_layout_int_0_struct {
+	TYPE_LAYOUT_FIELDS
+} mercury_data___base_type_layout_int_0 = {
 	make_typelayout_for_all_tags(TYPELAYOUT_CONST_TAG, 
 		mkbody(TYPELAYOUT_INT_VALUE))
 };
 
-Word * mercury_data___base_type_layout_character_0[] = {
+	/* base_type_layout for `character' */
+
+const struct mercury_data___base_type_layout_character_0_struct {
+	TYPE_LAYOUT_FIELDS
+} mercury_data___base_type_layout_character_0 = {
 	make_typelayout_for_all_tags(TYPELAYOUT_CONST_TAG, 
 		mkbody(TYPELAYOUT_CHARACTER_VALUE))
 };
 
-Word * mercury_data___base_type_layout_string_0[] = {
+	/* base_type_layout for `string' */
+
+const struct mercury_data___base_type_layout_string_0_struct {
+	TYPE_LAYOUT_FIELDS
+} mercury_data___base_type_layout_string_0 = {
 	make_typelayout_for_all_tags(TYPELAYOUT_CONST_TAG, 
 		mkbody(TYPELAYOUT_STRING_VALUE))
 };
 
-Word * mercury_data___base_type_layout_float_0[] = {
+	/* base_type_layout for `float' */
+
+const struct mercury_data___base_type_layout_float_0_struct {
+	TYPE_LAYOUT_FIELDS
+} mercury_data___base_type_layout_float_0 = {
 	make_typelayout_for_all_tags(TYPELAYOUT_CONST_TAG, 
 		mkbody(TYPELAYOUT_FLOAT_VALUE))
 };
 
-Word * mercury_data___base_type_layout_pred_0[] = {
+	/* base_type_layout for `void' */
+
+const struct mercury_data___base_type_layout_void_0_struct {
+	TYPE_LAYOUT_FIELDS
+} mercury_data___base_type_layout_void_0 = {
 	make_typelayout_for_all_tags(TYPELAYOUT_CONST_TAG, 
-		mkbody(TYPELAYOUT_PREDICATE_VALUE))
+		mkbody(TYPELAYOUT_VOID_VALUE))
 };
 
-#endif
+	/* base_type_functors definitions */
+
+	/* base_type_functors for `int' */
+
+const struct mercury_data___base_type_functors_int_0_struct {
+	Integer f1;
+} mercury_data___base_type_functors_int_0 = {
+	MR_TYPEFUNCTORS_SPECIAL
+};
+
+	/* base_type_functors for `character' */
+
+const struct mercury_data___base_type_functors_character_0_struct {
+	Integer f1;
+} mercury_data___base_type_functors_character_0 = {
+	MR_TYPEFUNCTORS_SPECIAL
+};
+
+	/* base_type_functors for `string' */
+
+const struct mercury_data___base_type_functors_string_0_struct {
+	Integer f1;
+} mercury_data___base_type_functors_string_0 = {
+	MR_TYPEFUNCTORS_SPECIAL
+};
+
+	/* base_type_functors for `float' */
+
+const struct mercury_data___base_type_functors_float_0_struct {
+	Integer f1;
+} mercury_data___base_type_functors_float_0 = {
+	MR_TYPEFUNCTORS_SPECIAL
+};
+
+	/* base_type_functors for `void' */
+
+const struct mercury_data___base_type_functors_void_0_struct {
+	Integer f1;
+} mercury_data___base_type_functors_void_0 = {
+	MR_TYPEFUNCTORS_SPECIAL
+};
+
+#endif /* USE_TYPE_LAYOUT */
+
+	/* base_type_infos definitions */
+
+	/* base_type_info for `int' */
 
 Declare_entry(mercury__builtin_unify_int_2_0);
 Declare_entry(mercury__builtin_index_int_2_0);
 Declare_entry(mercury__builtin_compare_int_3_0);
-Declare_entry(mercury__builtin_term_to_type_int_2_0);
-Declare_entry(mercury__builtin_type_to_term_int_2_0);
-Word * mercury_data___base_type_info_int_0[] = {
-	(Word *) ((Integer) 0),
-	(Word *) (Integer) ENTRY(mercury__builtin_unify_int_2_0),
-	(Word *) (Integer) ENTRY(mercury__builtin_index_int_2_0),
-	(Word *) (Integer) ENTRY(mercury__builtin_compare_int_3_0),
-#ifdef USE_TYPE_TO_TERM
-	(Word *) (Integer) ENTRY(mercury__builtin_term_to_type_int_2_0),
-	(Word *) (Integer) ENTRY(mercury__builtin_type_to_term_int_2_0),
+MR_STATIC_CODE_CONST struct mercury_data___base_type_info_int_0_struct {
+	Integer f1;
+	Code *f2;
+	Code *f3;
+	Code *f4;
+#ifdef USE_TYPE_LAYOUT
+	const Word *f5;
+	const Word *f6;
+	const Word *f7;
+	const Word *f8;
 #endif
+} mercury_data___base_type_info_int_0 = {
+	((Integer) 0),
+	MR_MAYBE_STATIC_CODE(ENTRY(mercury__builtin_unify_int_2_0)),
+	MR_MAYBE_STATIC_CODE(ENTRY(mercury__builtin_index_int_2_0)),
+	MR_MAYBE_STATIC_CODE(ENTRY(mercury__builtin_compare_int_3_0)),
 #ifdef  USE_TYPE_LAYOUT
-	(Word *) (Integer) mercury_data___base_type_layout_int_0
+	(const Word *) & mercury_data___base_type_layout_int_0,
+	(const Word *) & mercury_data___base_type_functors_int_0,
+	(const Word *) string_const(""mercury_builtin"", 15),
+	(const Word *) string_const(""int"", 3)
 #endif
 };
 
+	/* base_type_info for `character' */
 
 Declare_entry(mercury__builtin_unify_character_2_0);
 Declare_entry(mercury__builtin_index_character_2_0);
 Declare_entry(mercury__builtin_compare_character_3_0);
-Declare_entry(mercury__builtin_term_to_type_character_2_0);
-Declare_entry(mercury__builtin_type_to_term_character_2_0);
-Word * mercury_data___base_type_info_character_0[] = {
-	(Word *) ((Integer) 0),
-	(Word *) (Integer) ENTRY(mercury__builtin_unify_character_2_0),
-	(Word *) (Integer) ENTRY(mercury__builtin_index_character_2_0),
-	(Word *) (Integer) ENTRY(mercury__builtin_compare_character_3_0),
-#ifdef USE_TYPE_TO_TERM
-	(Word *) (Integer) ENTRY(mercury__builtin_term_to_type_character_2_0),
-	(Word *) (Integer) ENTRY(mercury__builtin_type_to_term_character_2_0),
+MR_STATIC_CODE_CONST struct 
+mercury_data___base_type_info_character_0_struct {
+	Integer f1;
+	Code *f2;
+	Code *f3;
+	Code *f4;
+#ifdef USE_TYPE_LAYOUT
+	const Word *f5;
+	const Word *f6;
+	const Word *f7;
+	const Word *f8;
 #endif
+} mercury_data___base_type_info_character_0 = {
+	((Integer) 0),
+	MR_MAYBE_STATIC_CODE(ENTRY(mercury__builtin_unify_character_2_0)),
+	MR_MAYBE_STATIC_CODE(ENTRY(mercury__builtin_index_character_2_0)),
+	MR_MAYBE_STATIC_CODE(ENTRY(mercury__builtin_compare_character_3_0)),
 #ifdef  USE_TYPE_LAYOUT
-	(Word *) (Integer) mercury_data___base_type_layout_character_0
+	(const Word *) & mercury_data___base_type_layout_character_0,
+	(const Word *) & mercury_data___base_type_functors_character_0,
+	(const Word *) string_const(""mercury_builtin"", 15),
+	(const Word *) string_const(""char"", 4)
 #endif
 };
 
+	/* base_type_info for `string' */
 
 Declare_entry(mercury__builtin_unify_string_2_0);
 Declare_entry(mercury__builtin_index_string_2_0);
 Declare_entry(mercury__builtin_compare_string_3_0);
-Declare_entry(mercury__builtin_term_to_type_string_2_0);
-Declare_entry(mercury__builtin_type_to_term_string_2_0);
-Word * mercury_data___base_type_info_string_0[] = {
-	(Word *) ((Integer) 0),
-	(Word *) (Integer) ENTRY(mercury__builtin_unify_string_2_0),
-	(Word *) (Integer) ENTRY(mercury__builtin_index_string_2_0),
-	(Word *) (Integer) ENTRY(mercury__builtin_compare_string_3_0),
-#ifdef USE_TYPE_TO_TERM
-	(Word *) (Integer) ENTRY(mercury__builtin_term_to_type_string_2_0),
-	(Word *) (Integer) ENTRY(mercury__builtin_type_to_term_string_2_0)
+MR_STATIC_CODE_CONST struct mercury_data___base_type_info_string_0_struct {
+	Integer f1;
+	Code *f2;
+	Code *f3;
+	Code *f4;
+#ifdef USE_TYPE_LAYOUT
+	const Word *f5;
+	const Word *f6;
+	const Word *f7;
+	const Word *f8;
 #endif
+} mercury_data___base_type_info_string_0 = {
+	((Integer) 0),
+	MR_MAYBE_STATIC_CODE(ENTRY(mercury__builtin_unify_string_2_0)),
+	MR_MAYBE_STATIC_CODE(ENTRY(mercury__builtin_index_string_2_0)),
+	MR_MAYBE_STATIC_CODE(ENTRY(mercury__builtin_compare_string_3_0)),
 #ifdef  USE_TYPE_LAYOUT
-	(Word *) (Integer) mercury_data___base_type_layout_string_0
+	(const Word *) & mercury_data___base_type_layout_string_0,
+	(const Word *) & mercury_data___base_type_functors_string_0,
+	(const Word *) string_const(""mercury_builtin"", 15),
+	(const Word *) string_const(""string"", 6)
 #endif
 };
 
+	/* base_type_info for `float' */
 
 Declare_entry(mercury__builtin_unify_float_2_0);
 Declare_entry(mercury__builtin_index_float_2_0);
 Declare_entry(mercury__builtin_compare_float_3_0);
-Declare_entry(mercury__builtin_term_to_type_float_2_0);
-Declare_entry(mercury__builtin_type_to_term_float_2_0);
-Word * mercury_data___base_type_info_float_0[] = {
-	(Word *) ((Integer) 0),
-	(Word *) (Integer) ENTRY(mercury__builtin_unify_float_2_0),
-	(Word *) (Integer) ENTRY(mercury__builtin_index_float_2_0),
-	(Word *) (Integer) ENTRY(mercury__builtin_compare_float_3_0),
-#ifdef USE_TYPE_TO_TERM
-	(Word *) (Integer) ENTRY(mercury__builtin_term_to_type_float_2_0),
-	(Word *) (Integer) ENTRY(mercury__builtin_type_to_term_float_2_0)
+MR_STATIC_CODE_CONST struct mercury_data___base_type_info_float_0_struct {
+	Integer f1;
+	Code *f2;
+	Code *f3;
+	Code *f4;
+#ifdef USE_TYPE_LAYOUT
+	const Word *f5;
+	const Word *f6;
+	const Word *f7;
+	const Word *f8;
 #endif
+} mercury_data___base_type_info_float_0 = {
+	((Integer) 0),
+	MR_MAYBE_STATIC_CODE(ENTRY(mercury__builtin_unify_float_2_0)),
+	MR_MAYBE_STATIC_CODE(ENTRY(mercury__builtin_index_float_2_0)),
+	MR_MAYBE_STATIC_CODE(ENTRY(mercury__builtin_compare_float_3_0)),
 #ifdef  USE_TYPE_LAYOUT
-	(Word *) (Integer) mercury_data___base_type_layout_float_0
+	(const Word *) & mercury_data___base_type_layout_float_0,
+	(const Word *) & mercury_data___base_type_functors_float_0,
+	(const Word *) string_const(""mercury_builtin"", 15),
+	(const Word *) string_const(""float"", 5)
 #endif
 };
 
+	/* base_type_info for `void' */
 
-Declare_entry(mercury__builtin_unify_pred_2_0);
-Declare_entry(mercury__builtin_index_pred_2_0);
-Declare_entry(mercury__builtin_compare_pred_3_0);
-Declare_entry(mercury__builtin_term_to_type_pred_2_0);
-Declare_entry(mercury__builtin_type_to_term_pred_2_0);
-Word * mercury_data___base_type_info_pred_0[] = {
-	(Word *) ((Integer) 0),
-	(Word *) (Integer) ENTRY(mercury__builtin_unify_pred_2_0),
-	(Word *) (Integer) ENTRY(mercury__builtin_index_pred_2_0),
-	(Word *) (Integer) ENTRY(mercury__builtin_compare_pred_3_0),
-#ifdef USE_TYPE_TO_TERM
-	(Word *) (Integer) ENTRY(mercury__builtin_term_to_type_pred_2_0),
-	(Word *) (Integer) ENTRY(mercury__builtin_type_to_term_pred_2_0)
+Declare_entry(mercury__unused_0_0);
+MR_STATIC_CODE_CONST struct mercury_data___base_type_info_void_0_struct {
+	Integer f1;
+	Code *f2;
+	Code *f3;
+	Code *f4;
+#ifdef USE_TYPE_LAYOUT
+	const Word *f5;
+	const Word *f6;
+	const Word *f7;
+	const Word *f8;
 #endif
+} mercury_data___base_type_info_void_0 = {
+	((Integer) 0),
+	MR_MAYBE_STATIC_CODE(ENTRY(mercury__unused_0_0)),
+	MR_MAYBE_STATIC_CODE(ENTRY(mercury__unused_0_0)),
+	MR_MAYBE_STATIC_CODE(ENTRY(mercury__unused_0_0)),
 #ifdef  USE_TYPE_LAYOUT
-	(Word *) (Integer) mercury_data___base_type_layout_pred_0
+	(const Word *) & mercury_data___base_type_layout_void_0,
+	(const Word *) & mercury_data___base_type_functors_void_0,
+	(const Word *) string_const(""mercury_builtin"", 15),
+	(const Word *) string_const(""void"", 4)
 #endif
 };
 
-#endif
+BEGIN_MODULE(builtin_types_module)
+
+BEGIN_CODE
+
+END_MODULE
+
+/*
+INIT sys_init_builtin_types_module
+*/
+void sys_init_builtin_types_module(void);
+void sys_init_builtin_types_module(void) {
+	extern ModuleFunc builtin_types_module;
+	extern ModuleFunc mercury__mercury_builtin__init;
+
+	builtin_types_module();
+
+	/* 
+	** We had better call this init() because we use the
+	** labels for the special preds of int, float, pred, 
+	** character and string. If they aren't initialized,
+	** we might initialize the base_type_info with
+	** garbage
+	*/
+	mercury__mercury_builtin__init();
+
+	MR_INIT_BUILTIN_BASE_TYPE_INFO(
+		mercury_data___base_type_info_int_0, _int_);
+	MR_INIT_BUILTIN_BASE_TYPE_INFO(
+		mercury_data___base_type_info_float_0, _float_);
+	MR_INIT_BUILTIN_BASE_TYPE_INFO(
+		mercury_data___base_type_info_pred_0, _pred_);
+	MR_INIT_BUILTIN_BASE_TYPE_INFO(
+		mercury_data___base_type_info_character_0, _character_);
+	MR_INIT_BUILTIN_BASE_TYPE_INFO(
+		mercury_data___base_type_info_string_0, _string_);
+	MR_INIT_BASE_TYPE_INFO_WITH_PRED(
+		mercury_data___base_type_info_void_0, mercury__unused_0_0);
+}
+
 ").
 
 	% This is used by the code that the compiler generates for compare/3.
 compare_error :-
 	error("internal error in compare/3").
 
-term__context_init(term__context("", 0)).
-
-:- type var_supply == int.
-:- type var == int.
-
-% create a new supply of variables
-term__init_var_supply(0).
-
-% We number variables using sequential numbers.
-term__create_var(VarSupply0, VarSupply, VarSupply) :-
-		VarSupply is VarSupply0 + 1.
-
-term__var_to_int(Var, Var).
-
 %-----------------------------------------------------------------------------%
 
-/* copy/2
-	:- pred copy(T, T).
-	:- mode copy(ui, uo) is det.
-	:- mode copy(in, uo) is det.
+/* unsafe_promise_unique/2
+	:- pred unsafe_promise_unique(T, T).
+	:- mode unsafe_promise_unique(in, uo) is det.
 */
 
-	% XXX note that this is *not* deep copy, and so it is unsafe!
-
 /* This doesn't work, due to the lack of support for aliasing.
-:- pragma(c_code, copy(X::ui, Y::uo), "Y = X;").
-:- pragma(c_code, copy(X::in, Y::uo), "Y = X;").
+:- pragma(c_code, unsafe_promise_unique(X::in, Y::uo), "Y = X;").
 */
 
-:- external(copy/2).
+:- external(unsafe_promise_unique/2).
 :- pragma(c_code, "
-Define_extern_entry(mercury__copy_2_0);
-Define_extern_entry(mercury__copy_2_1);
+Define_extern_entry(mercury__unsafe_promise_unique_2_0);
 
-BEGIN_MODULE(copy_module)
-	init_entry(mercury__copy_2_0);
-	init_entry(mercury__copy_2_1);
+BEGIN_MODULE(unsafe_promise_unique_module)
+	init_entry(mercury__unsafe_promise_unique_2_0);
 BEGIN_CODE
 
-Define_entry(mercury__copy_2_0);
-Define_entry(mercury__copy_2_1);
+Define_entry(mercury__unsafe_promise_unique_2_0);
 #ifdef	COMPACT_ARGS
 	r1 = r2;
 #else
@@ -659,11 +674,90 @@ Define_entry(mercury__copy_2_1);
 
 END_MODULE
 
+/* Ensure that the initialization code for the above module gets run. */
+/*
+INIT sys_init_unsafe_promise_unique_module
+*/
+void sys_init_unsafe_promise_unique_module(void);
+	/* extra declaration to suppress gcc -Wmissing-decl warning */
+void sys_init_unsafe_promise_unique_module(void) {
+	extern ModuleFunc unsafe_promise_unique_module;
+	unsafe_promise_unique_module();
+}
+
+").
+
+%-----------------------------------------------------------------------------%
+
+/* copy/2
+	:- pred copy(T, T).
+	:- mode copy(ui, uo) is det.
+	:- mode copy(in, uo) is det.
+*/
+
+/*************
+Using `pragma c_code' doesn't work, due to the lack of support for
+aliasing, and in particular the lack of support for `ui' modes.
+:- pragma c_code(copy(Value::ui, Copy::uo), "
+	save_transient_registers();
+	Copy = deep_copy(Value, TypeInfo_for_T, NULL, NULL);
+	restore_transient_registers();
+").
+:- pragma c_code(copy(Value::in, Copy::uo), "
+	save_transient_registers();
+	Copy = deep_copy(Value, TypeInfo_for_T, NULL, NULL);
+	restore_transient_registers();
+").
+*************/
+
+:- external(copy/2).
+
+:- pragma(c_header_code, "#include ""deep_copy.h""").
+
+:- pragma(c_code, "
+Define_extern_entry(mercury__copy_2_0);
+Define_extern_entry(mercury__copy_2_1);
+
+BEGIN_MODULE(copy_module)
+	init_entry(mercury__copy_2_0);
+	init_entry(mercury__copy_2_1);
+BEGIN_CODE
+
+#ifdef PROFILE_CALLS
+  #define fallthru(target, caller) { tailcall((target), (caller)); }
+#else
+  #define fallthru(target, caller)
+#endif
+
+Define_entry(mercury__copy_2_0);
+fallthru(ENTRY(mercury__copy_2_1), ENTRY(mercury__copy_2_0))
+Define_entry(mercury__copy_2_1);
+{
+	Word value, copy, type_info;
+
+	type_info = r1;
+	value = r2;
+
+	save_transient_registers();
+	copy = deep_copy(value, type_info, NULL, NULL);
+	restore_transient_registers();
+
+#ifdef	COMPACT_ARGS
+	r1 = copy;
+#else
+	r3 = copy;
+#endif
+
+	proceed();
+}
+END_MODULE
+
 /* Ensure that the initialization code for the above module gets run. */
 /*
 INIT sys_init_copy_module
 */
-void sys_init_copy_module(void); /* suppress gcc -Wmissing-decl warning */
+void sys_init_copy_module(void);
+	/* extra declaration to suppress gcc -Wmissing-decl warning */
 void sys_init_copy_module(void) {
 	extern ModuleFunc copy_module;
 	copy_module();
@@ -674,7 +768,79 @@ void sys_init_copy_module(void) {
 %-----------------------------------------------------------------------------%
 
 % The type c_pointer can be used by predicates which use the C interface.
-:- type c_pointer == int.
+
+:- pragma(c_code, "
+
+/*
+ * c_pointer has a special value reserved for its layout, since it needs to
+ * be handled as a special case.
+ */
+
+#ifdef  USE_TYPE_LAYOUT
+
+const struct mercury_data_mercury_builtin__base_type_layout_c_pointer_0_struct {
+	TYPE_LAYOUT_FIELDS
+} mercury_data_mercury_builtin__base_type_layout_c_pointer_0 = {
+	make_typelayout_for_all_tags(TYPELAYOUT_CONST_TAG, 
+		mkbody(TYPELAYOUT_C_POINTER_VALUE))
+};
+
+const struct
+mercury_data_mercury_builtin__base_type_functors_c_pointer_0_struct {
+	Integer f1;
+} mercury_data_mercury_builtin__base_type_functors_c_pointer_0 = {
+	MR_TYPEFUNCTORS_SPECIAL
+};
+
+#endif
+
+Define_extern_entry(mercury____Unify___mercury_builtin__c_pointer_0_0);
+Define_extern_entry(mercury____Index___mercury_builtin__c_pointer_0_0);
+Define_extern_entry(mercury____Compare___mercury_builtin__c_pointer_0_0);
+
+BEGIN_MODULE(unify_c_pointer_module)
+	init_entry(mercury____Unify___mercury_builtin__c_pointer_0_0);
+	init_entry(mercury____Index___mercury_builtin__c_pointer_0_0);
+	init_entry(mercury____Compare___mercury_builtin__c_pointer_0_0);
+
+BEGIN_CODE
+Define_entry(mercury____Unify___mercury_builtin__c_pointer_0_0);
+	/*
+	** For c_pointer, we assume that equality and comparison
+	** can be based on object identity (i.e. using address comparisons).
+	** This is correct for types like io__stream, and necessary since
+	** the io__state contains a map(io__stream, filename).
+	** However, it might not be correct in general...
+	*/
+	unify_output = (unify_input1 == unify_input2);
+	proceed();
+
+Define_entry(mercury____Index___mercury_builtin__c_pointer_0_0);
+	index_output = -1;
+	proceed();
+
+Define_entry(mercury____Compare___mercury_builtin__c_pointer_0_0);
+	compare_output = (compare_input1 == compare_input2 ? COMPARE_EQUAL :
+			  compare_input1 < compare_input2 ? COMPARE_LESS :
+			  COMPARE_GREATER);
+	proceed();
+
+END_MODULE
+
+/* Ensure that the initialization code for the above module gets run. */
+/*
+INIT sys_init_unify_c_pointer_module
+*/
+extern ModuleFunc unify_c_pointer_module;
+void sys_init_unify_c_pointer_module(void);
+	/* duplicate declaration to suppress gcc -Wmissing-decl warning */
+void sys_init_unify_c_pointer_module(void) {
+	unify_c_pointer_module();
+}
+
+").
+
+
 
 :- end_module mercury_builtin.
 
