@@ -1,4 +1,7 @@
 %-----------------------------------------------------------------------------%
+% Copyright (C) 1995, 1997 The University of Melbourne.
+% This file may only be copied under the terms of the GNU Library General
+% Public License - see the file COPYING.LIB in the Mercury distribution.
 %-----------------------------------------------------------------------------%
 %
 %  Red-black tree module.
@@ -35,7 +38,7 @@
 :- module rbtree.
 :- interface.
 
-:- import_module list, std_util, assoc_list.
+:- import_module list, assoc_list.
 
 :- type rbtree(Key, Value).
 
@@ -105,7 +108,7 @@
 
 :- implementation.
 
-:- import_module int, require.
+:- import_module int, require, std_util.
 
 
 :- type rbtree(K,V)	 --->	empty
