@@ -1,5 +1,5 @@
 %---------------------------------------------------------------------------%
-% Copyright (C) 1995 University of Melbourne.
+% Copyright (C) 1995-1997 The University of Melbourne.
 % This file may only be copied under the terms of the GNU Library General
 % Public License - see the file COPYING.LIB in the Mercury distribution.
 %------------------------------------------------------------------------------%
@@ -18,7 +18,7 @@
 :- module relation.
 
 :- interface.
-:- import_module list, set, set_bbbtree, std_util, assoc_list.
+:- import_module list, set, set_bbbtree, assoc_list.
 
 :- type relation(T).
 
@@ -65,7 +65,8 @@
 :- mode relation__add_assoc_list(in, in, out) is det.
 
 	% relation__remove removes an element from the relation.
-:- pred relation__remove(relation(T), relation_key, relation_key, relation(T)).
+:- pred relation__remove(relation(T), relation_key, relation_key,
+		relation(T)).
 :- mode relation__remove(in, in, in, out) is det.
 
 	% relation__remove_assoc_list removes a list of elements
@@ -168,8 +169,9 @@
 	% providing we have already visited Visit0 nodes, 
 	% ie the reverse of Dfs from relation__dfs/5.
 	% Visit is Visit0 + DfsRev.
-:- pred relation__dfsrev(relation(T), relation_key,  set_bbbtree(relation_key), 
-		set_bbbtree(relation_key), list(relation_key)).
+:- pred relation__dfsrev(relation(T), relation_key,
+		set_bbbtree(relation_key), set_bbbtree(relation_key),
+		list(relation_key)).
 :- mode relation__dfsrev(in, in, in, out, out) is det.
 
 	% relation__is_dag(R) is true iff R is a directed acyclic graph.
@@ -941,12 +943,18 @@ relation__rtc(Rel, Rtc) :-
 	relation__domain_sorted_list(Rel, DomList),
 	list__length(DomList, DomLen),
 	map__init(Map0),
-	rtc__init_map(Map0, DomList, Map),
-	relation__init(RtcIn),
+	relation__rtc_init_map(Map0, DomList, Map),
+
+	% Make a new relation using the same domain as the old one.
+	Rel   = relation(NextElement, ElMap, _, _),
+	map__init(FwdMap),
+	map__init(BwdMap),
+	RtcIn = relation(NextElement, ElMap, FwdMap, BwdMap),
+
 	Inf is (DomLen + 2) * 2,	% This is close enough to infinity.
 	relation__rtc_2(Inf, Rel, DomList, Map, RtcIn, Rtc).
 
-	% rtc__init_map takes a domain Xs and creates a map
+	% relation__rtc_init_map takes a domain Xs and creates a map
 	% Map such that:
 	%
 	%	all [X]
@@ -961,13 +969,13 @@ relation__rtc(Rel, Rtc) :-
 	% and a "map" would be the ease of initialisation,
 	% and the semi-static nature of the domain.  (Arrays
 	% can be resized.)
-:- pred rtc__init_map(map(relation_key, int), list(relation_key),
+:- pred relation__rtc_init_map(map(relation_key, int), list(relation_key),
 		map(relation_key, int)).
-:- mode rtc__init_map(in, in, out) is det.
-rtc__init_map(Map, [], Map).
-rtc__init_map(MapIn, [ X | Xs ], MapOut) :-
+:- mode relation__rtc_init_map(in, in, out) is det.
+relation__rtc_init_map(Map, [], Map).
+relation__rtc_init_map(MapIn, [ X | Xs ], MapOut) :-
 	map__det_insert(MapIn, X, 0, Map1),
-	rtc__init_map(Map1, Xs, MapOut).
+	relation__rtc_init_map(Map1, Xs, MapOut).
 
 :- pred relation__rtc_2(int, relation(T), list(relation_key),
 		map(relation_key, int), relation(T), relation(T)).
