@@ -1,5 +1,5 @@
 %------------------------------------------------------------------------------%
-% Copyright (C) 1995 University of Melbourne.
+% Copyright (C) 1995-1997 The University of Melbourne.
 % This file may only be copied under the terms of the GNU Library General
 % Public License - see the file COPYING.LIB in the Mercury distribution.
 %------------------------------------------------------------------------------%
@@ -16,7 +16,7 @@
 
 :- interface.
 
-:- import_module bool, list, std_util.
+:- import_module bool, list.
 
 :- type set_bbbtree(T).
 
@@ -212,7 +212,8 @@
 	% `set_bbbtree__intersect(SetA, SetB, Set)' is true iff `Set' is the
 	% intersection of `SetA' and `SetB'.
 
-:- pred set_bbbtree__intersect(set_bbbtree(T), set_bbbtree(T), set_bbbtree(T)).
+:- pred set_bbbtree__intersect(set_bbbtree(T), set_bbbtree(T),
+					set_bbbtree(T)).
 :- mode set_bbbtree__intersect(in, in, out) is det.
 
 
@@ -220,7 +221,7 @@
 	% interscetion of the sets in `Sets'.
 
 :- pred set_bbbtree__power_intersect(set_bbbtree(set_bbbtree(T)),
-								set_bbbtree(T)).
+					set_bbbtree(T)).
 :- mode set_bbbtree__power_intersect(in, out) is det.
 
 
@@ -228,19 +229,20 @@
 	%  set containing all the elements of `SetA' except those that
 	% occur in `SetB'.
 
-:- pred set_bbbtree__difference(set_bbbtree(T), set_bbbtree(T), set_bbbtree(T)).
+:- pred set_bbbtree__difference(set_bbbtree(T), set_bbbtree(T),
+					set_bbbtree(T)).
 :- mode set_bbbtree__difference(in, in, out) is det.
 
 
 	% `set_bbbtree__subset(SetA, SetB)' is true iff all the elements of
-	% `SetB' are also elements of `SetA'.
+	% `SetA' are also elements of `SetB'.
 
 :- pred set_bbbtree__subset(set_bbbtree(T), set_bbbtree(T)).
 :- mode set_bbbtree__subset(in, in) is semidet.
 
 
 	% `set_bbbtree__superset(SetA, SetB)' is true iff all the elements of
-	% `SetA' are also elements of `SetB'.
+	% `SetB' are also elements of `SetA'.
 
 :- pred set_bbbtree__superset(set_bbbtree(T), set_bbbtree(T)).
 :- mode set_bbbtree__superset(in, in) is semidet.
@@ -248,6 +250,10 @@
 %------------------------------------------------------------------------------%
 %------------------------------------------------------------------------------%
 
+:- implementation.
+
+:- import_module int, require.
+
 % Implementation based on "Functional Pearls: Efficient sets - a balancing act"
 % by Stephen Adams, J. Functional Programming 3 (4): 553-561, Oct 1993.
 %
@@ -288,10 +294,6 @@
 %
 %------------------------------------------------------------------------------%
 
-:- implementation.
-
-:- import_module int, require.
-
 :- type set_bbbtree(T) --->
 				empty
 			;	tree(T, int, set_bbbtree(T), set_bbbtree(T)).
@@ -392,7 +394,7 @@ set_bbbtree__equal(SetA, SetB) :-
 set_bbbtree__insert(Set0, X, Set) :-
 	set_bbbtree__def_ratio(Ratio),
 	set_bbbtree__insert_r(Set0, X, Set1, Ratio),
-	copy(Set1, Set).
+	unsafe_promise_unique(Set1, Set).
 
 /* Uncomment this once destructive input and unique modes are fixed and detele
    the one above.
@@ -486,7 +488,7 @@ set_bbbtree__delete(Set0, X, Set) :-
 	;
 		Set2 = Set0
 	),
-	copy(Set2, Set).
+	unsafe_promise_unique(Set2, Set).
 
 %------------------------------------------------------------------------------%
 
@@ -864,7 +866,7 @@ set_bbbtree__difference_r(tree(V, _N, LL, LR), R, Set, Ratio) :-
 %------------------------------------------------------------------------------%
 
 set_bbbtree__subset(SetA, SetB) :-
-	set_bbbtree__difference(SetB, SetA, Set),
+	set_bbbtree__difference(SetA, SetB, Set),
 	set_bbbtree__empty(Set).
 
 %------------------------------------------------------------------------------%
@@ -888,7 +890,7 @@ set_bbbtree__build_node(X, L, R, Tree) :-
 	set_bbbtree__size(R, RSize),
 	N is 1 + LSize + RSize,
 	Tree0 = tree(X, N, L, R),
-	copy(Tree0, Tree).
+	unsafe_promise_unique(Tree0, Tree).
 
 %------------------------------------------------------------------------------%
 
