@@ -3,7 +3,7 @@ INIT mercury_sys_init_call
 ENDINIT
 */
 /*
-** Copyright (C) 1995 University of Melbourne.
+** Copyright (C) 1995-1997 The University of Melbourne.
 ** This file may only be copied under the terms of the GNU Library General
 ** Public License - see the file COPYING.LIB in the Mercury distribution.
 */
@@ -29,7 +29,6 @@ ENDINIT
 */
 
 #include "imp.h"
-#include "type_info.h"
 
 Define_extern_entry(do_call_det_closure);
 Declare_label(det_closure_return);
@@ -45,10 +44,6 @@ Define_extern_entry(mercury__compare_3_1);
 Define_extern_entry(mercury__compare_3_2);
 Define_extern_entry(mercury__compare_3_3);
 Declare_label(mercury__compare_3_0_i1);
-Define_extern_entry(mercury__term_to_type_2_0);
-Declare_label(mercury__term_to_type_2_0_i1);
-Define_extern_entry(mercury__type_to_term_2_0);
-Declare_label(mercury__type_to_term_2_0_i1);
 
 BEGIN_MODULE(call_module)
 	init_entry(do_call_det_closure);
@@ -65,10 +60,6 @@ BEGIN_MODULE(call_module)
 	init_entry(mercury__compare_3_2);
 	init_entry(mercury__compare_3_3);
 	init_label(mercury__compare_3_0_i1);
-	init_entry(mercury__term_to_type_2_0);
-	init_label(mercury__term_to_type_2_0_i1);
-	init_entry(mercury__type_to_term_2_0);
-	init_label(mercury__type_to_term_2_0_i1);
 BEGIN_CODE
 
 Define_entry(do_call_det_closure);
@@ -96,8 +87,9 @@ Define_entry(do_call_det_closure);
 		}
 	} /* else do nothing because i == 3 */
 
-	for (i = 1; i <= num_in_args; i++) 
+	for (i = 1; i <= num_in_args; i++) {
 		virtual_reg(i) = field(0, closure, i+1); /* copy args */
+	}
 
 	restore_registers();
 
@@ -116,8 +108,9 @@ Define_label(det_closure_return);
 #else
 	save_registers();
 
-	for (i = 1; i <= num_out_args; i++)
+	for (i = 1; i <= num_out_args; i++) {
 		virtual_reg(i) = virtual_reg(i+num_in_args);
+	}
 
 	restore_registers();
 #endif
@@ -151,8 +144,9 @@ Define_entry(do_call_semidet_closure);
 		}
 	} /* else do nothing because i == 3 */
 
-	for (i = 1; i <= num_in_args; i++) 
+	for (i = 1; i <= num_in_args; i++) {
 		virtual_reg(i) = field(0, closure, i+1); /* copy args */
+	}
 #else
 	if (num_in_args < 2) {
 		for (i = 1; i <= num_extra_args; i++) {
@@ -164,8 +158,9 @@ Define_entry(do_call_semidet_closure);
 		}
 	} /* else do nothing because i == 2 */
 
-	for (i = 1; i <= num_in_args; i++) 
+	for (i = 1; i <= num_in_args; i++) {
 		virtual_reg(i+1) = field(0, closure, i+1); /* copy args */
+	}
 #endif
 
 	restore_registers();
@@ -185,8 +180,9 @@ Define_label(semidet_closure_return);
 #else
 	save_registers();
 
-	for (i = 1; i <= num_out_args; i++)
+	for (i = 1; i <= num_out_args; i++) {
 		virtual_reg(i+1) = virtual_reg(i+1+num_in_args);
+	}
 
 	restore_registers();
 #endif
@@ -220,8 +216,9 @@ Define_entry(do_call_nondet_closure);
 		}
 	} /* else do nothing because i == 3 */
 
-	for (i = 1; i <= num_in_args; i++) 
+	for (i = 1; i <= num_in_args; i++) {
 		virtual_reg(i) = field(0, closure, i+1); /* copy args */
+	}
 
 	restore_registers();
 
@@ -239,8 +236,9 @@ Define_label(nondet_closure_return);
 #else
 	save_registers();
 
-	for (i = 1; i <= num_out_args; i++)
+	for (i = 1; i <= num_out_args; i++) {
 		virtual_reg(i) = virtual_reg(i+num_in_args);
+	}
 
 	restore_registers();
 #endif
@@ -274,37 +272,23 @@ Define_entry(mercury__unify_2_0);
 	Word	x, y;
 	int	i;
 
-#ifdef	ONE_OR_TWO_CELL_TYPE_INFO
 	Word	base_type_info;
 
 	x = mercury__unify__x;
 	y = mercury__unify__y;
 
 	base_type_info = field(0, mercury__unify__typeinfo, 0);
-	if (base_type_info == 0)
-	{
+	if (base_type_info == 0) {
 		type_arity = 0;
 		unify_pred = (Code *) field(0, mercury__unify__typeinfo,
 				OFFSET_FOR_UNIFY_PRED);
 		/* args_base will not be needed */
-	}
-	else
-	{
+	} else {
 		type_arity = field(0, base_type_info, OFFSET_FOR_COUNT);
 		unify_pred = (Code *) field(0, base_type_info,
 				OFFSET_FOR_UNIFY_PRED);
 		args_base = mercury__unify__typeinfo;
 	}
-#else
-	x = mercury__unify__x;
-	y = mercury__unify__y;
-
-	type_arity = field(0, mercury__unify__typeinfo, OFFSET_FOR_COUNT);
-	unify_pred = (Code *) field(0, mercury__unify__typeinfo,
-			OFFSET_FOR_UNIFY_PRED);
-	args_base = (Word) ((Word *) mercury__unify__typeinfo
-			- 1 + OFFSET_FOR_ARG_TYPE_INFOS);
-#endif
 
 	save_registers();
 
@@ -349,30 +333,20 @@ Define_entry(mercury__index_2_0);
 	Word	x;
 	int	i;
 
-#ifdef	ONE_OR_TWO_CELL_TYPE_INFO
 	Word	base_type_info;
 
 	x = r2;
 	base_type_info = field(0, r1, 0);
-	if (base_type_info == 0)
-	{
+	if (base_type_info == 0) {
 		type_arity = 0;
 		index_pred = (Code *) field(0, r1, OFFSET_FOR_INDEX_PRED);
 		/* args_base will not be needed */
-	}
-	else
-	{
+	} else {
 		type_arity = field(0, base_type_info, OFFSET_FOR_COUNT);
 		index_pred = (Code *) field(0, base_type_info,
 				OFFSET_FOR_INDEX_PRED);
 		args_base = r1;
 	}
-#else
-	x = r2;
-	type_arity = field(0, r1, OFFSET_FOR_COUNT);
-	index_pred = (Code *) field(0, r1, OFFSET_FOR_INDEX_PRED);
-	args_base = (Word) ((Word *) r1 - 1 + OFFSET_FOR_ARG_TYPE_INFOS);
-#endif
 
 	save_registers();
 
@@ -464,37 +438,23 @@ Define_entry(mercury__compare_3_3);
 	Word	x, y;
 	int	i;
 
-#ifdef	ONE_OR_TWO_CELL_TYPE_INFO
 	Word	base_type_info;
 
 	x = mercury__compare__x;
 	y = mercury__compare__y;
 
 	base_type_info = field(0, mercury__compare__typeinfo, 0);
-	if (base_type_info == 0)
-	{
+	if (base_type_info == 0) {
 		type_arity = 0;
 		compare_pred = (Code *) field(0, mercury__compare__typeinfo,
 				OFFSET_FOR_COMPARE_PRED);
 		/* args_base will not be needed */
-	}
-	else
-	{
+	} else {
 		type_arity = field(0, base_type_info, OFFSET_FOR_COUNT);
 		compare_pred = (Code *) field(0, base_type_info,
 				OFFSET_FOR_COMPARE_PRED);
 		args_base = mercury__compare__typeinfo;
 	}
-#else
-	x = mercury__compare__x;
-	y = mercury__compare__y;
-
-	type_arity = field(0, mercury__compare__typeinfo, OFFSET_FOR_COUNT);
-	compare_pred = (Code *) field(0, mercury__compare__typeinfo,
-			OFFSET_FOR_COMPARE_PRED);
-	args_base = (Word) ((Word *) mercury__compare__typeinfo
-			- 1 + OFFSET_FOR_ARG_TYPE_INFOS);
-#endif
 
 	save_registers();
 
@@ -534,227 +494,6 @@ Define_label(mercury__compare_3_0_i1);
 	proceed();
 #endif
 }
-
-/*
-** mercury__term_to_type_2_0 is called as `term_to_type(TypeInfo, Term, X)'
-** in the mode `term_to_type(in, in, out) is semidet'.
-**
-** With the simple parameter convention, the inputs are in r2 and r3,
-** the success/failure indication in r1, and the output in r4.
-**
-** With the compact parameter convention, the inputs are in r1 and r2,
-** the success/failure indication in r1, and the output in r2.
-**
-** We call the type-specific term_to_type routine as
-** `TermToTypePred(...ArgTypeInfos..., Term, X)' is semidet.
-**
-** With the simple parameter convention, the inputs are in r2, ... rN+2,
-** the success/failure indication in r1, and the output in rN+3.
-**
-** With the compact parameter convention, the inputs are in r1, ... rN+1,
-** the success/failure indication in r1, and the output in r2.
-**
-** With the compact convention, we can make the call to the type-specific
-** routine a tail call, and we do so. With the simple convention, we can't.
-*/
-
-Define_entry(mercury__term_to_type_2_0);
-{
-#if OFFSET_FOR_ARG_TYPE_INFOS != 6
-	fatal_error("type_to_term/2 and term_to_type/2 not implemented");
-#else
-
-	Code	*term_to_type_pred;	/* address of the term_to_type pred */
-					/* for this type */
-	int	type_arity;	/* number of type_info args */
-	Word	args_base;	/* the address of the word before the first */
-				/* type_info argument */
-	Word	term;
-	int	i;
-
-#ifdef	ONE_OR_TWO_CELL_TYPE_INFO
-	Word	base_type_info;
-
-	term = mercury__term_to_type__term;
-
-	base_type_info = field(0, mercury__term_to_type__typeinfo, 0);
-	if (base_type_info == 0)
-	{
-		type_arity = 0;
-		term_to_type_pred = (Code *) field(0,
-				mercury__term_to_type__typeinfo,
-				OFFSET_FOR_TERM_TO_TYPE_PRED);
-		/* args_base will not be needed */
-	}
-	else
-	{
-		type_arity = field(0, base_type_info, OFFSET_FOR_COUNT);
-		term_to_type_pred = (Code *) field(0, base_type_info,
-				OFFSET_FOR_TERM_TO_TYPE_PRED);
-		args_base = mercury__term_to_type__typeinfo;
-	}
-#else
-	term = mercury__term_to_type__term;
-
-	type_arity = field(0, mercury__term_to_type__typeinfo,
-			OFFSET_FOR_COUNT);
-	term_to_type_pred = (Code *) field(0,
-			mercury__term_to_type__typeinfo,
-			OFFSET_FOR_TERM_TO_TYPE_PRED);
-	args_base = (Word) ((Word *) mercury__term_to_type__typeinfo
-			- 1 + OFFSET_FOR_ARG_TYPE_INFOS);
-#endif
-
-	save_registers();
-
-	/* we call 'TermToTypePred(...ArgTypeInfos..., Term, X)' */
-	for (i = 1; i <= type_arity; i++) {
-		virtual_reg(i + mercury__term_to_type__offset) =
-			field(0, args_base, i);
-	}
-	virtual_reg(type_arity + mercury__term_to_type__offset + 1) = term;
-
-	restore_registers();
-
-#ifdef	COMPACT_ARGS
-	tailcall(term_to_type_pred, LABEL(mercury__term_to_type_2_0));
-#else
-	push(succip);
-	push(type_arity);
-	call(term_to_type_pred, LABEL(mercury__term_to_type_2_0_i1),
-		LABEL(mercury__term_to_type_2_0));
-#endif
-#endif
-}
-/*
-** Since mod2c declares this label, we must define it,
-** even though it is not needed with COMPACT_ARGS.
-*/
-Define_label(mercury__term_to_type_2_0_i1);
-{
-#if OFFSET_FOR_ARG_TYPE_INFOS != 6
-	fatal_error("type_to_term/2 and term_to_type/2 not implemented");
-#else
-#ifdef	COMPACT_ARGS
-	fatal_error("mercury__term_to_type_2_0_i1 reached in COMPACT_ARGS mode");
-#else
-	/* r1 already contains the truth result of the semidet pred */
-	/* mercury__term_to_type_2_0 so r1 does not have to be updated. */
-
-	int	type_arity;
-	
-	type_arity = pop();
-	succip = pop();
-	save_registers();
-	r4 = virtual_reg(type_arity + 3);
-	proceed();
-#endif
-#endif
-}
-
-/*
-** mercury__type_to_term_2_0 is called as `type_to_term(TypeInfo, X, Term)'
-** in the mode `type_to_term(in, in, out) is det'.
-**
-** With both conventions, the inputs are in r1 and r2.
-** With the simple parameter convention, the output is in r3;
-** with the compact parameter convention, the output is in r1.
-**
-** We call the type-specific type_to_term routine as
-** `TypeToTermPred(...ArgTypeInfos..., X, Term)' is det.
-**
-** With both conventions, the inputs are in r1, ... rN.
-** With the simple parameter convention, the output is in rN+1;
-** with the compact parameter convention, the output is in r1.
-**
-** With the compact convention, we can make the call to the type-specific
-** routine a tail call, and we do so. With the simple convention, we can't.
-*/
-
-Define_entry(mercury__type_to_term_2_0);
-{
-#if OFFSET_FOR_ARG_TYPE_INFOS != 6
-	fatal_error("type_to_term/2 and term_to_type/2 not implemented");
-#else
-
-	Code	*type_to_term_pred;	/* address of the type_to_term pred */
-					/* for this type */
-	int	type_arity;	/* number of type_info args */
-	Word	args_base;	/* the address of the word before the first */
-				/* type_info argument */
-	Word	x;
-	int	i;
-
-#ifdef	ONE_OR_TWO_CELL_TYPE_INFO
-	Word	base_type_info;
-
-	x = r2;
-
-	base_type_info = field(0, r1, 0);
-	if (base_type_info == 0)
-	{
-		type_arity = 0;
-		type_to_term_pred = (Code *) field(0, r1,
-				OFFSET_FOR_TYPE_TO_TERM_PRED);
-		/* args_base will not be needed */
-	}
-	else
-	{
-		type_arity = field(0, base_type_info, OFFSET_FOR_COUNT);
-		type_to_term_pred = (Code *) field(0, base_type_info,
-				OFFSET_FOR_TYPE_TO_TERM_PRED);
-		args_base = r1;
-	}
-#else
-	x = r2;
-
-	type_arity = field(0, r1, OFFSET_FOR_COUNT);
-	type_to_term_pred = (Code *) field(0, r1, OFFSET_FOR_TYPE_TO_TERM_PRED);
-	args_base = (Word) ((Word *) r1 - 1 + OFFSET_FOR_ARG_TYPE_INFOS);
-#endif
-
-	save_registers();
-
-	/* we call 'TypeToTermPred(...ArgTypeInfos..., X, Term)' */
-	for (i = 1; i <= type_arity; i++) {
-		virtual_reg(i) = field(0, args_base, i);
-	}
-	virtual_reg(type_arity + 1) = x;
-
-	restore_registers();
-
-#ifdef	COMPACT_ARGS
-	tailcall(type_to_term_pred, LABEL(mercury__type_to_term_2_0));
-#else
-	push(succip);
-	push(type_arity);
-	call(type_to_term_pred, LABEL(mercury__type_to_term_2_0_i1),
-		LABEL(mercury__type_to_term_2_0));
-#endif
-#endif
-}
-/*
-** Since mod2c declares this label, we must define it,
-** even though it is not needed with COMPACT_ARGS.
-*/
-Define_label(mercury__type_to_term_2_0_i1);
-{
-#if OFFSET_FOR_ARG_TYPE_INFOS != 6
-	fatal_error("type_to_term/2 and term_to_type/2 not implemented");
-#else
-#ifdef	COMPACT_ARGS
-	fatal_error("mercury__type_to_term_2_0_i1 reached in COMPACT_ARGS mode");
-#else
-	int	type_arity;
-	
-	type_arity = pop();
-	succip = pop();
-	save_registers();
-	r3 = virtual_reg(type_arity + 2);
-	proceed();
-#endif
-#endif
-}
 END_MODULE
 void mercury_sys_init_call(void); /* suppress gcc warning */
 void mercury_sys_init_call(void) {
