@@ -1,10 +1,11 @@
 /*
-** Copyright (C) 1995 University of Melbourne.
+** Copyright (C) 1993-1995, 1997 The University of Melbourne.
 ** This file may only be copied under the terms of the GNU Library General
 ** Public License - see the file COPYING.LIB in the Mercury distribution.
 */
 
 #include "dummy.h"
+#include "imp.h"	/* we need libmer_globals.h for Windows DLLs */
 
 /*
 ** This dummy function is in a file of its own to ensure
@@ -16,4 +17,5 @@ void	*global_pointer_2;
 
 void dummy_function_call(void)
 {
+	return;
 }
