@@ -1,16 +1,21 @@
 /*
-** Copyright (C) 1995 University of Melbourne.
+** Copyright (C) 1994-1997 The University of Melbourne.
 ** This file may only be copied under the terms of the GNU Library General
 ** Public License - see the file COPYING.LIB in the Mercury distribution.
 */
 
+/*
+** label.h defines the interface to the label table, which is a pair of
+** hash tables mapping from procedure names to addresses and vice versa.
+*/
+
 #ifndef	LABEL_H
 #define	LABEL_H
 
-#include	"dlist.h"
+#include "mercury_types.h"	/* for `Code *' */
+#include "dlist.h"		/* for `List' */
 
-typedef struct s_label
-{
+typedef struct s_label {
 	const char	*e_name;   /* name of the procedure	     */
 	Code		*e_addr;   /* address of the code	     */
 } Label;
@@ -25,4 +30,4 @@ extern  int 	entry_table_size;
 	/* expected number of entries in the table */
 	/* we allocate 8 bytes per entry */
 
-#endif
+#endif /* not LABEL_H */
