@@ -1,12 +1,16 @@
 /*
-** Copyright (C) 1995 University of Melbourne.
+** Copyright (C) 1995-1997 The University of Melbourne.
 ** This file may only be copied under the terms of the GNU Library General
 ** Public License - see the file COPYING.LIB in the Mercury distribution.
 */
 
+/* prof.h -- definitions for profiling */
+
 #ifndef PROF_H
 #define PROF_H
 
+#include "mercury_types.h"	/* for `Code *' */
+
 /*
 ** this variable holds the address of the "current" procedure so that
 ** when a profiling interrupt occurs, the profiler knows where we are,
@@ -42,4 +46,4 @@ extern	void	prof_output_addr_pair_table(void);
 extern	void	prof_output_addr_decls(const char *, const Code *);
 extern	void	prof_output_addr_table(void);
 
-#endif
+#endif	/* not PROF_H */
