@@ -1,20 +1,20 @@
 /*
-** Copyright (C) 1995 University of Melbourne.
+** Copyright (C) 1993-1995, 1997 The University of Melbourne.
 ** This file may only be copied under the terms of the GNU Library General
 ** Public License - see the file COPYING.LIB in the Mercury distribution.
 */
 
 /*
-**	Definitions for the table module.
+** table.h - defines the interface to the hash table module.
 */
 
 #ifndef	TABLE_H
 #define	TABLE_H
 
-#include	"dlist.h"
+#include "std.h"	/* for bool */
+#include "dlist.h"	/* for List */
 
-typedef	struct	s_table
-{
+typedef	struct	s_table {
 	int		ta_size;
 	List		**ta_store;
 	const void *	(*ta_key)(const void *);    /* applied to entries */
@@ -39,4 +39,4 @@ extern	bool	tab_insert_table(const Table *, void *);
 extern	List	*tab_get_all_entries(const Table *);
 extern	int	tab_str_to_int(const char *);
 
-#endif
+#endif /* not TABLE_H */
